import re
import signal

import celery
import kombu
from celery import signals

import agent.config
from common.utils import celeryutil
from common.utils import configutil
from common.utils import logutil
from common.utils.facts import fact

if fact["os"]["name"] != "windows":
    import ctypes

    libc = ctypes.CDLL("libc.so.6")
else:
    ctypes = None
    libc = None


LOG = logutil.get_logger(__name__)


class AgentCelery(celery.Celery):
    """Custom celery application for agent."""

    task_cls = "agent.utils.celeryutil:AgentTask"
    events_cls = "agent.utils.celeryutil:Events"

    def __init__(self, *args, **kwargs):
        super(AgentCelery, self).__init__(*args, **kwargs)

    def send_task(self, name, args=(), kwargs=None, **opts):  # pylint: disable=arguments-differ
        kwargs = kwargs or {}
        reply_to = celeryutil.agent_results_queue(agent.config.workflow_engine["server_id"])
        opts["reply_to"] = opts.get("reply_to", reply_to)
        if self.conf.task_always_eager:
            task = self.tasks[name]
            return task.apply(args, kwargs, **opts)
        else:
            return super(AgentCelery, self).send_task(name, args, kwargs, **opts)


@signals.worker_process_init.connect
def _kill_hook(*args, **kwargs):
    if libc:
        libc.prctl(1, signal.SIGKILL)


@signals.before_task_publish.connect
def _set_user_id(properties=None, headers=None, **kwds):
    """Fill message.properties.user_id attribute with server_id which is equal
    to Rabbitmq username for the server. Rabbitmq guarantees this property is valid.

    https://www.rabbitmq.com/validated-user-id.html
    """
    properties["user_id"] = agent.config.workflow_engine["server_id"]


def create_app():
    LOG.debug("Creating celery app")
    app = AgentCelery()
    app.amqp.Consumer.auto_declare = False
    app.amqp.Producer.auto_declare = False

    kombu.Producer.auto_declare = False
    kombu.Consumer.auto_declare = False
    kombu.Exchange.passive = True
    kombu.Queue.passive = True

    LOG.debug("App %s created", app)
    return app


def register_facts():
    LOG.debug("Registering facts")
    fact.includes += [
        "agent.tasks.sys",
        "agent.tasks.update",
        "agent.tasks.storage",
        "agent.tasks.filesystem",
        "common.tasks.storage",
        "common.tasks.libcloud",
    ]
    fact.load()
    LOG.debug("Facts registered")


def configure_worker():
    LOG.debug("Configuring worker app")
    configutil.configure_ssl(agent.config)
    app = celery.current_app
    app.conf.update(agent.config.workflow_engine["celery"])
    # strip RabbitMQ password from log.
    broker_password = re.split(":|@", agent.config.workflow_engine["celery"]["broker_url"])[2]
    LOG.debug(
        "App %s configured with %s",
        app,
        str(agent.config.workflow_engine["celery"]).replace(broker_password, "******"),
    )


# Initialization
agent.config.reload()
if not logutil.root.handlers:
    # Not under agent.scalarizr.app entrypoint, configure own logging.
    logutil.basic_config(agent.config.workflow_engine["log_level"])

app = create_app()
app.set_default()
register_facts()
configure_worker()

LOG.debug("App %s loaded.", app)
