"""
.. module:: config
:platform: All
:synopsis: Agent configuration

.. moduleauthor:: Marat Komarov <marat@scalr.com>
"""
import os
import sys

is_windows = sys.platform == "win32"

workflow_engine = {}

"""
Scalr server-id. Required. Override.
"""
workflow_engine["server_id"] = None

"""
Configuration files directory.
"""
workflow_engine["etc_dir"] = "C:\\opt\\scalarizr\\etc" if is_windows else "/etc/scalr"

"""
Agent Home directory.
More at https://scalr-labs.atlassian.net/wiki/display/FAM/Agent+Home+Directory
"""
workflow_engine["home_dir"] = "C:\\opt\\scalarizr\\var\\lib" if is_windows else "/var/lib/scalarizr"


"""
Directory to store scalr-init cloud data
"""
workflow_engine["cloud_dir"] = (
    "C:\\opt\\scalarizr\\var\\lib\\cloud" if is_windows else "/var/lib/scalarizr/cloud"
)

"""
Directory to store scalr-init result symlink
"""
workflow_engine["run_dir"] = "C:\\opt\\scalarizr\\var\\run" if is_windows else "/var/run/scalarizr"

"""
Path to agent.json. Default is $ETC_DIR/agent.json
"""
workflow_engine["agent_json_path"] = os.path.join(workflow_engine["etc_dir"], "agent.json")

"""
Cache directory
"""
workflow_engine["cache_dir"] = os.path.join(workflow_engine["home_dir"], "cache")


"""
Cache directory to store downloaded fatmouse agent packages
"""
workflow_engine["pkgmgr_cache_dir"] = os.path.join(workflow_engine["cache_dir"], "pkgmgr")

"""
Agent log level.
Allowed values are ('debug', 'info', 'warn', 'error')
"""
workflow_engine["log_level"] = "DEBUG"

"""
Progress reports tick time. (currently unused)
Time consuming operation should report it's progress at least every PROGRESS_TICK seconds.
"""
workflow_engine["progress_tick"] = 1

"""
Name of the agent package
"""
workflow_engine["package_name"] = "scalarizr"

"""
Name of the agent service
"""
workflow_engine["service_name"] = workflow_engine["package_name"]

"""
Global timeout in seconds.
Maximum timeout for all time consuming, blocking operations
Default: 40 minutes (same as HostInit / RebootFinish timeout in Scalr)
"""
workflow_engine["global_timeout"] = 2400

"""
The interval for garbage collection (messages, tasks data, etc..)
"""
workflow_engine["gc_interval"] = 3600  # 1 hour

"""
Time to keep stale/done data (messages, tasks data, orchestration logs, etc...)
"""
workflow_engine["gc_keep_time"] = 86400  # 1 day

"""
Maximum length of a string field in result, after which it will be ellipsed.
:see: :func:`common.utils.mixutil.ellipsis()`
"""
workflow_engine["ellipsis_length"] = 100

"""
Maximum number of bytes to read by :func:`common.utils.sysutil.tail()`
"""
workflow_engine["tail_num_bytes"] = 20000

"""
Maximum number of lines to be propagated from 3rd party sources:  logs, stderr etc.
:see: :func:`common.utils.sysutil.tail()`
"""
workflow_engine["tail_num_lines"] = 600

"""
Celery config section

result_expires has no meaning in agent config - it affects x-expires argument
for results queue, and fatmouse uses result exchange instead
"""
workflow_engine["celery"] = {
    "accept_content": ["json"],
    "broker_connection_timeout": 5,
    "broker_url": "amqp://guest:guest@rabbit:5671/",
    "broker_use_ssl": True,
    "broker_heartbeat": 60,
    "result_backend": "agent.utils.celeryutil:RPCBackend",
    "result_serializer": "json",
    "result_persistent": True,
    "task_acks_late": True,
    "task_serializer": "json",
    "task_always_eager": False,
    "task_send_sent_event": True,
    "worker_consumer": "agent.utils.celeryutil:CeleryConsumer",
    "worker_max_tasks_per_child": 1,
    "worker_enable_remote_control": False,
    "worker_send_task_events": False,
    "include": [
        "agent.tasks.agent",
        "agent.tasks.augeas",
        "agent.tasks.chef",
        "agent.tasks.execute",
        "agent.tasks.firewall",
        "agent.tasks.image",
        "agent.tasks.misc",
        "agent.tasks.pkgmgr",
        "agent.tasks.service",
        "agent.tasks.storage",
        "agent.tasks.sys",
        "agent.tasks.update",
        "common.tasks.libcloud",
        "common.tasks.repo",
    ],
}

"""
RabbitMQ config section
"""
workflow_engine["rabbitmq"] = {
    "ssl_verify": True,
    "ssl_cacert": None,
    # ssl_cacert_store_path - path to the file where raw CA Certificate will be saved.
    # Because it is required to pass a path to cacert.pem in ssl modules.
    "ssl_cacert_store_path": os.path.join(workflow_engine["etc_dir"], "cacert.pem"),
}

workflow_engine["devbox"] = {"enabled": False, "keep_task_data": False}

"""
Agent statistics section
"""
stat = {"push_interval": 60, "metrics": [], "metrics_ignore": []}


def get_stat_metrics():
    """Return list of metrics that can be pushed by agent or pulled by scalr server."""
    return list(set(stat.get("metrics", [])) - set(stat.get("metrics_ignore", [])))


def reload():
    # local configutil import to avoid circular importing
    from common.utils import configutil

    return configutil.merge(
        sys.modules[__name__], cfg_file=workflow_engine["agent_json_path"], cfg_env="FAM_AGENT_CFG"
    )
