import base64
import collections
import fnmatch
import json
import os
import re
import sys

from six.moves import BaseHTTPServer
from six.moves import socketserver as SocketServer
from six.moves import urllib_parse

from agent.scalarizr import types
from agent.scalarizr.hub import hub
from common.utils import dockerutil
from common.utils import logutil
from common.utils import subprocess2
from common.utils import sysutil

is_windows = sys.platform == "win32"
if not is_windows:
    from six.moves.socketserver import UnixStreamServer


__all__ = ["UnixHTTPRequestHandler", "PolicyEngine", "HTTPServer", "install_plugin"]

PolicyResult = collections.namedtuple("PolicyResult", "resolution, errors")

LOG = logutil.get_logger(__name__)

PLUGIN_NAME = "scalr-auth"
SPEC_PATH = "/etc/docker/plugins/{}.spec".format(PLUGIN_NAME)
CONF_PATH = "/etc/docker/daemon.json"


class Policy(object):
    """
    Parent Policy class that implements general policy check logic.
    The logic is to call all methods on policy class which names starts with "_check"
    and ensure every one of them returns PolicyEngine.ALOW.
    A single PolicyEngine.DENY means the entire Policy.check result is PolicyEngine.DENY.
    """

    def __init__(self):
        super(Policy, self).__init__()
        self.url_params = None
        self.request = None
        self.errors = []

    def check(self, request, url_params=None):
        self.url_params = url_params
        self.request = request
        self.errors = []
        checks = [getattr(self, method)() for method in dir(self) if method.startswith("_check")]
        if not all(check == PolicyEngine.ALLOW for check in checks):
            return PolicyResult(PolicyEngine.DENY, self.errors)
        return PolicyResult(PolicyEngine.ALLOW, self.errors)


class PortContainer(tuple):
    """
    Examples:
    PortContainer(8080)
    PortContainer("8080")
    PortContainer("8080-8082")
    """

    def __new__(cls, value):
        if isinstance(value, int) or "-" not in value:
            return super(PortContainer, cls).__new__(cls, (int(value),))
        min_, max_ = value.split("-")
        return super(PortContainer, cls).__new__(
            cls, [int(port) for port in range(int(min_), int(max_) + 1)]
        )


class CreateContainerPolicy(Policy):
    """
    Enforces policy for /containers/create
    https://docs.docker.com/engine/api/v1.25/#operation/ContainerCreate
    """

    def __init__(self, allow_privileged_create=None, allowed_host_ports=None, allowed_mounts=None):
        super(CreateContainerPolicy, self).__init__()
        self.allow_privileged_create = allow_privileged_create
        allowed_mounts = allowed_mounts if allowed_mounts is not None else []
        self.allowed_mounts = [os.path.normpath(mount) for mount in allowed_mounts]
        self.allowed_host_ports = sum(
            [PortContainer(port_str) for port_str in allowed_host_ports or ()], ()
        )

    def _check_allow_privileged(self):
        if self.allow_privileged_create is not None:
            if self.request["HostConfig"].get("Privileged") and self.allow_privileged_create:
                self.errors.append("Privileged mode is not allowed")
                return PolicyEngine.DENY
        return PolicyEngine.ALLOW

    def _check_allowed_host_ports(self):
        """
        Note that -p allows user to pass ContainerPort only (i.e. docker run -it -p 8080 busybox sh)
        When configured like that docker makes its own choice for HostPort number to map ContainerPort to.
        but does that AFTER the request to authz plugin is made.
        So the only option is to DENY such requests completely.
        """
        if self.allowed_host_ports:
            port_bindings = self.request["HostConfig"].get("PortBindings") or {}
            for container_port, options in port_bindings.items():
                for option in options:
                    host_port = option.get("HostPort")
                    if not host_port:
                        self.errors.append(
                            "It's not allowed to publish port "
                            "{} on a host's random port".format(container_port)
                        )
                        return PolicyEngine.DENY
                    elif not set(self.allowed_host_ports).issuperset(PortContainer(host_port)):
                        self.errors.append(
                            "It's not allowed to bind the {} port".format(
                                option["HostPort"] or container_port
                            )
                        )
                        return PolicyEngine.DENY
        return PolicyEngine.ALLOW

    def _check_allowed_mounts(self):
        if self.allowed_mounts:
            binds = self.request["HostConfig"].get("Binds") or []
            for bind in binds:
                host_path = bind.split(":")[0]
                host_path = os.path.normpath(host_path)
                check_mounts = [
                    fnmatch.fnmatch(host_path, allowed_mount)
                    for allowed_mount in self.allowed_mounts
                ]
                if not any(check_mounts):
                    self.errors.append("The directory {} is forbidden to mount".format(host_path))
                    return PolicyEngine.DENY
        return PolicyEngine.ALLOW


class CreateImagePolicy(Policy):
    """
    Enforces policy for /images/create
    https://docs.docker.com/engine/api/v1.25/#operation/ImageCreate
    """

    def __init__(self, allowed_sources=None, allowed_images=None):
        super(CreateImagePolicy, self).__init__()
        # XXX: Think for a good reason to have these separate.
        self.allowed_sources = allowed_sources if allowed_sources else []
        allowed_images = allowed_images if allowed_images else []
        self.allowed_images = []
        for img in allowed_images:
            if ":" not in img:
                self.allowed_images.append(img + ":*")
            else:
                self.allowed_images.append(img)

    def _check_allowed(self):
        try:
            # Pull
            image = self.request["fromImage"][0]
            tags = self.request.get("tag")  # Asserting image tags
            if tags and tags[0]:
                image = "%s:%s" % (image, tags[0])
        except KeyError:
            # Import
            image = self.request["fromSrc"][0]
        if not any(
            fnmatch.fnmatch(image, pattern)
            for pattern in self.allowed_sources + self.allowed_images
        ):
            self.errors.append("It's not allowed to create this image: {}".format(image))
            return PolicyEngine.DENY
        return PolicyEngine.ALLOW


class ContainerExecPolicy(Policy):
    """
    Enforces policy for /containers/{id}/exec
    https://docs.docker.com/engine/api/v1.25/#operation/ContainerExec
    """

    def __init__(self, limit_escalating=False):
        super(ContainerExecPolicy, self).__init__()
        self.limit_escalating = limit_escalating

    def _check_allow_privileged(self):
        if self.request.get("Privileged", False) and self.limit_escalating:
            self.errors.append("Privileged mode is not allowed")
            return PolicyEngine.DENY
        return PolicyEngine.ALLOW


class ContainerAttachPolicy(Policy):
    """
    Enforces policy for /containers/{id}/attach
    https://docs.docker.com/engine/api/v1.25/#operation/ContainerAttach
    """

    def __init__(self, limit_escalating=False):
        super(ContainerAttachPolicy, self).__init__()
        self.limit_escalating = limit_escalating

    def _is_container_privileged(self, container_id):
        stats = dockerutil.get_container_stat(container_id)
        container_stats = stats[container_id]
        return container_stats.get("privileged")

    def _check_allow_privileged(self):
        is_container_privileged = self._is_container_privileged(self.url_params["id"])
        if is_container_privileged and self.limit_escalating:
            self.errors.append("Privileged mode is not allowed")
            return PolicyEngine.DENY
        return PolicyEngine.ALLOW


def convert_policies(policies):
    """Converts Scalr Governance Policies to fatmouse internal representation
    :type policies: dict
    """
    dp = types.DockerPolicies()
    docker_policies = {
        policy: json.loads(value[0])["rules"][0]
        for policy, value in policies.items()
        if policy in types.DockerPoliciesNames.values()
    }
    allow_privileged_create = docker_policies.get(
        types.DockerPoliciesNames.allow_privileged_create, {}
    ).get("value", None)
    allowed_host_ports = docker_policies.get(types.DockerPoliciesNames.allowed_host_ports, {}).get(
        "value", None
    )
    allowed_mounts = docker_policies.get(types.DockerPoliciesNames.allowed_mounts, {}).get(
        "value", None
    )
    no_container_policies = (allow_privileged_create, allowed_host_ports, allowed_mounts) == (
        None,
        None,
        None,
    )
    dp.create_container_policy = (
        None
        if no_container_policies
        else types.CreateContainerDockerPolicy(
            dict(
                allow_privileged_create=allow_privileged_create,
                allowed_host_ports=allowed_host_ports,
                allowed_mounts=allowed_mounts,
            )
        )
    )

    image_policies = docker_policies.get(types.DockerPoliciesNames.image_policies, {}).get(
        "value", None
    )
    if image_policies:
        allowed_images = []
        allowed_sources = []
        for policy in image_policies:
            if policy["type"] == "registry":
                allowed_images.append(policy["source"])
            elif policy["type"] == "url":
                allowed_sources.append(policy["source"])
        dp.create_image_policy = types.CreateImageDockerPolicy(
            dict(allowed_images=allowed_images, allowed_sources=allowed_sources)
        )
    else:
        dp.create_image_policy = None

    limit_escalating = docker_policies.get(types.DockerPoliciesNames.allow_privileged_exec, {}).get(
        "value", None
    )
    if limit_escalating:
        dp.exec_container_policy = dict(limit_escalating=limit_escalating)
    else:
        dp.exec_container_policy = None
    return dp


class PolicyEngine(object):
    """
    Docker policy checker class that dispatches docker api requests on corresponding Policy classes
    """

    _dp = None
    _api_routes = None
    DENY, ALLOW = range(2)

    def is_empty(self):
        policies = [
            self._dp.create_container_policy,
            self._dp.create_image_policy,
            self._dp.exec_container_policy,
        ]
        return all(policy is None for policy in policies)

    def __init__(self, policies):
        self._api_routes = {}
        self._dp = convert_policies(policies)

        if not self.is_empty():
            if self._dp.create_container_policy:
                self._api_routes[r"/containers/create"] = CreateContainerPolicy(
                    **self._dp.create_container_policy
                )

            if self._dp.create_image_policy:
                self._api_routes[r"/images/create"] = CreateImagePolicy(
                    **self._dp.create_image_policy
                )

            if self._dp.exec_container_policy:
                self._api_routes[r"/containers/.+/exec"] = ContainerExecPolicy(
                    **self._dp.exec_container_policy
                )
                self._api_routes[r"/containers/(?P<id>.+)/attach"] = ContainerAttachPolicy(
                    **self._dp.exec_container_policy
                )

    def check(self, route, request):
        # Optional api version.
        api_prefix_re = r"(/v\d+.\d+)?"

        for route_re, policy in self._api_routes.items():
            result = re.match(api_prefix_re + route_re, route)
            if result:
                policy_result = policy.check(request, result.groupdict())
                if policy_result.resolution == PolicyEngine.DENY:
                    return policy_result
        return PolicyResult(PolicyEngine.ALLOW, [])


if not is_windows:

    class UnixHTTPServer(UnixStreamServer):
        allow_reuse_address = 1

        def server_bind(self):
            """Override server_bind to store the server name.
            Only one instance exists at the time so
            Safely cleaning junk from previous runs"""
            try:
                os.unlink(self.server_address)
            except OSError:
                pass
            UnixStreamServer.server_bind(self)
            self.server_name = self.socket.getsockname()
            self.server_port = 0


else:

    class UnixHTTPServer:
        """
        TODO: https://github.com/Microsoft/Virtualization-Documentation/issues/442
        """

        def __init__(self, *args, **kwargs):
            raise NotImplementedError()


class UnixBaseHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    def address_string(self):
        """Return the client address formatted for logging.
        This version looks up the full hostname using gethostbyaddr(),
        and tries to find a name that contains at least one dot.
        """
        return self.client_address

    def log_message(self, format, *args):  # pylint: disable=redefined-builtin
        pass


class UnixHTTPRequestHandler(UnixBaseHTTPRequestHandler):
    def _plugin_response(self, payload):
        self.send_response(200)
        self.end_headers()
        self.wfile.write(json.dumps(payload).encode("utf-8"))

    def _plugin_request(self):
        content_len = int(self.headers.get("Content-Length", 0))
        req = self.rfile.read(content_len).decode()
        if not req:
            return None
        r = json.loads(req)
        try:
            r["RequestBody"] = json.loads(base64.b64decode(r["RequestBody"]).decode())
        except KeyError:
            pass
        r["RequestParams"] = urllib_parse.parse_qs(urllib_parse.urlsplit(r["RequestUri"]).query)
        return r

    def do_POST(self):
        try:
            path = urllib_parse.urlsplit(self.path).path
            plugin_methods = {
                "/Plugin.Activate": self.activate,
                "/AuthZPlugin.AuthZReq": self.auth_request,
                "/AuthZPlugin.AuthZRes": self.auth_response,
            }
            self._plugin_response(plugin_methods[path](self._plugin_request()))
        except Exception:
            LOG.exception("Docker authz plugin reqest failed: ")

    def activate(self, request):
        return {"Implements": ["authz"]}

    def auth_request(self, request):
        if self.server.health_check:
            if "docker-authz-health-check" in request["RequestUri"]:
                self.server.health_check["passed"] = True

        check = self.server.policy_engine.check(
            request["RequestUri"], request.get("RequestBody") or request["RequestParams"]
        )
        if check.resolution == PolicyEngine.DENY:
            return {"Allow": False, "Msg": "\n".join(check.errors), "Err": "Request forbidden"}
        return {"Allow": True, "Msg": "Request allowed", "Err": ""}

    def auth_response(self, request):
        return {"Allow": True, "Msg": "Request allowed", "Err": ""}


class HTTPServer(SocketServer.ThreadingMixIn, UnixHTTPServer):

    _health_check = False

    def is_active(self):
        self._health_check = {"passed": False}
        try:
            subprocess2.run("docker ps -f name=docker-authz-health-check", shell=True)
        except BaseException as e:
            LOG.debug("Exception during docker pluging health check: %s", e)
        finally:
            passed = self._health_check["passed"]
            self._health_check = False
        return passed

    @property
    def health_check(self):
        return self._health_check


def install_plugin():
    LOG.debug("Installing docker authz plugin")

    if not os.path.exists(SPEC_PATH):
        sysutil.mkdir_p(os.path.dirname(SPEC_PATH))
        with open(SPEC_PATH, "w") as f:
            f.write("unix://%s" % hub.docker_authz_socket)
        LOG.info("Created spec file for docker authz plugin: %s", SPEC_PATH)

    conf = {}
    if os.path.exists(CONF_PATH):
        with open(CONF_PATH, "r") as f:
            try:
                conf = json.load(f)
            except ValueError:
                pass
    if conf.get("authorization-plugins"):
        if PLUGIN_NAME not in conf["authorization-plugins"]:
            conf["authorization-plugins"].append(PLUGIN_NAME)
    else:
        conf["authorization-plugins"] = [PLUGIN_NAME]
    with open(CONF_PATH, "w") as f:
        json.dump(conf, f)

    LOG.info("Created JSON config file for docker authz plugin: %s", CONF_PATH)
    LOG.debug("Docker authz plugin has been configured and installed.")


def uninstall_plugin():
    """
    :return: bool Uninstall Result
    """
    uninstalled = False
    LOG.debug("Checking docker authz plugin installation status")
    if os.path.exists(SPEC_PATH):
        os.remove(SPEC_PATH)
        LOG.debug("%s removed", SPEC_PATH)

    if os.path.exists(CONF_PATH):
        with open(CONF_PATH, "r") as f:
            try:
                conf = json.load(f)
            except ValueError:
                LOG.warning("Cannot uninstall authz plugin: malformed %s", CONF_PATH)
                return
        if "authorization-plugins" in conf:
            active_plugins = conf["authorization-plugins"]
            if isinstance(active_plugins, list) and PLUGIN_NAME in active_plugins:
                active_plugins.remove(PLUGIN_NAME)
                with open(CONF_PATH, "w") as f:
                    json.dump(conf, f)
                    uninstalled = True
                LOG.debug("%s updated", CONF_PATH)
                LOG.debug("Uninstalled Docker authz plugin.")
    return uninstalled


if __name__ == "__main__":
    install_plugin()
    httpd = HTTPServer("/var/run/scalr_docker_authz_debug.socket", UnixHTTPRequestHandler)
    sa = httpd.socket.getsockname()
    httpd.serve_forever()
