# isort:skip_file
import argparse
import collections
import contextlib
import ctypes
import functools
import logging.handlers
import os
import platform
import pprint
import signal
import socket
import sys
import threading
import time
import wsgiref.simple_server

import celery
import psutil
from schematics.exceptions import ValidationError

# Create celery application:
from .hub import hub
from common.utils import logutil
from common.utils import six

if not hasattr(sys, "_under_pytest"):
    # Disabled in devbox to avoid creating Scalarizr log files,
    # or having [Errno 13] Permission denied
    logutil.file_config(six.moves.cStringIO(hub.log_config))
LOG = logutil.get_logger("scalarizr")
# The import below creates default celery application and MUST go before any common.tasks / agent.tasks imports.
if not hasattr(sys, "_under_pytest"):
    # Disabled in devbox cause breakes module reload trick in test_app.py
    LOG.debug("[pid: %s] Loading scalarizr %s (corev2)", os.getpid(), hub.agent_version)
import agent.app

#

from agent.dockerauthz import plugin as dockerauthz_plugin
from agent.scalarizr import postinst
from agent.scalrinit import app as scalrinit_app
from agent.scalrinit import cloudconfig
from agent.scalrinit import datadir
from agent.tasks import execute as execute_tasks
from agent.tasks import service as service_tasks
from agent.utils import celeryutil
from common.utils import cryptoutil
from common.utils import exc as common_exc
from common.utils import mixutil
from common.utils import rpcutil
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

import agent.config

from . import base
from . import exc
from . import metadata
from . import msg
from . import rpc
from . import stat
from . import types
from .utils import apputil

if sys.platform == "win32":  # pragma: no cover
    linuxutil = None
    pwd = None
else:
    from common.utils import linuxutil
    import pwd

if sys.platform == "win32":  # pragma: no cover
    # pylint: disable=import-error
    libc = None
    import servicemanager
    import win32api
    import win32event
    import win32service
    import win32serviceutil
else:
    libc = ctypes.CDLL("libc.so.6")
    servicemanager = None
    win32api = None
    win32event = None
    win32service = None
    win32serviceutil = None


class Error(Exception):
    """ Class for general errors. """


@contextlib.contextmanager
def init_error_handling():
    """Handles exceptions by either sending OperationResult to Scalr if the agent is
    initializing (this will mark sever as `Failed` in the Scalr UI) or by sending RuntimeError
    (which does nothing by now, but will show warnings in the UI in the future).
    """
    try:
        yield
    except (SystemExit, KeyboardInterrupt):
        raise
    except BaseException as e:
        e_message = "{}".format(e)
        if hub.state in ("bootstrapping", "initializing"):
            LOG.exception("Error in server initialization: %s: %s", e.__class__.__name__, e_message)
            msg.send(
                msg.create(
                    "OperationResult",
                    body={
                        "id": str(mixutil.sortable_uuid()),
                        "name": "system.init",
                        "status": "error",
                        "error": e_message,
                    },
                )
            )
        else:
            LOG.exception(
                "Error in %s server configuration: %s: %s",
                hub.state,
                e.__class__.__name__,
                e_message,
            )
            msg.send(msg.create("RuntimeError", {"message": e_message}))


@contextlib.contextmanager
def systemd_notification():
    try:
        yield
    except Exception:
        if fact["os"]["init"] == "systemd":
            # Notify systemd about an error in initialization
            linuxutil.sd_notify("ERRONO=1")
        raise
    else:
        if fact["os"]["init"] == "systemd":
            # Notify systemd about the end of initialization
            linuxutil.sd_notify("READY=1")


@mixutil.error_logged("Error in main")
def main():
    with systemd_notification():
        # Low-level initialization.
        prepare()
        args = commandline_parse()
        commandline_handle(args)

    # Daemon startup.
    LOG.info("[pid: %s] Starting scalarizr %s (corev2)", os.getpid(), hub.agent_version)
    base.container_host_add_ssh_authorized_key()
    if hub.state == "importing":
        base.clean_previous_server(chef=False)  # Customer may have well configured chef-client,
        # that he don't want to lose, just because
        # we're making scalarizr image.
        hub.setup_from_commandline(args)
    else:
        apputil.boot_wait()
        if not update.client.running:
            update.client.start()
        update.wait()
        metadata.wait(timeout=60)
        user_data = metadata.user_data()
        server_id = user_data.get("serverid")

        # If `server_id` is empty, it means that we bootstrapping a new server.
        # see: SCALARIZR-2533
        if not server_id or hub.general.server_id != server_id:
            if hub.state == "discovering":
                LOG.info("Deploying agent on discovered server (%s)", server_id)
                base.clean_previous_server(
                    chef=False
                )  # Customer may have well configured chef-client.
            else:
                LOG.info(
                    "Scalr server-id changed! " "Bootstrapping new server (%s, previous=%s)",
                    server_id,
                    hub.general.server_id,
                )
                hub.state = "bootstrapping"
                base.clean_previous_server()
            base.container_host_add_ssh_authorized_key(force=True)
            hub.setup_from_user_data(user_data)
            hub.boot_id = sysutil.get_boot_id()
        hub.sync()
        update.restart_if_needed()

    # Starting core services and threads, doing basic setup common for all states.
    with init_error_handling():
        base.sync_time()
        hub.start_time = time.time()
        start()
        base.setup_firewall()
        base.sync_global_variables()
        base.mount_all()
        base.cloudinit_enable_root_login()
        base.add_root_ssh_authorized_key()
        base.provision_ansible_host()
        reboot_finish()
        update_finish()

        # State dependent setup.
        if hub.state in ("discovering", "running"):
            base.check_behaviors()
            base.setup_hostname()
        if hub.state == "discovering":
            import_server()
            hub.state = "running"
        if hub.state == "importing":
            import_server()
        if hub.state == "bootstrapping":
            bootstrap()
        if hub.state == "initializing":
            initialize()
    # Daemon main loop.
    msg.handler.resume()

    # Importing on windows runs as standalone executable and not as a service
    # so we require main loop there.
    if fact["os"]["family"] != "windows" or hub.state == "importing":
        apputil.daemon_main(shutdown)


def prepare():
    """Low-level application initialization."""
    if fact["os"]["family"] != "windows" and os.getegid():
        logutil.die('Scalarizr must run under "root"')
    # TODO: rewrite for logging.config.dictConfig()
    logutil.file_config(six.moves.cStringIO(hub.log_config))
    Fatmouse.setup_local()
    setup_environ()


def start():
    """Starting services and background threads."""
    if hub.state != "importing":
        adjust_ports()

    msg.handler.set_callback(HandlerCallback())
    _start_api()
    _start_messaging()
    hub.periodic_executor.start()
    fatmouse.setup()
    fatmouse.start()
    if fatmouse.was_setup():
        stat.collector.start()
    docker_plugin.start()
    hub.fire("start")  # application started


def _start_api():
    """Starts API server (1 daemon thread)."""
    # pylint: disable=no-name-in-module,import-error
    from scalarizr import api

    class ThreadingWSGIServer(
        six.moves.socketserver.ThreadingMixIn, wsgiref.simple_server.WSGIServer
    ):
        pass

    app = rpcutil.Server(rpc.RequestHandler(api.api_routes), hub.crypto_key_path)

    api_server = wsgiref.simple_server.make_server(
        "0.0.0.0", hub.base.api_port, app, server_class=ThreadingWSGIServer
    )

    LOG.info("Starting API on *:%s", hub.base.api_port)
    t = threading.Thread(target=api_server.serve_forever, name="api-server")
    t.daemon = True
    t.start()


def _start_messaging():
    """Starts message consumer and handler."""
    # start consumer thread(s)
    msg.consumer.start()

    # start handler thread
    msg.handler.suspend()
    handler_thread = threading.Thread(
        target=handles_fatal(msg.handler.serve_forever), name="message-handler"
    )
    handler_thread.daemon = True
    handler_thread.start()


def bootstrap():
    """HostInit message handling."""
    base.check_behavior_deprecation(hub.behaviors)
    base.update_pkgmgr_db()
    base.setup_os_license()
    hi_out = msg.create("HostInit")
    hi_out.body.crypto_key = cryptoutil.keygen().decode()
    hi_out.body.base.hostname = fact["hostname"]  # pylint: disable=no-member
    hi_out.body.seconds_since_start = float("{0:.2f}".format(time.time() - hub.start_time))
    hi_out.body.seconds_since_boot = float("{0:.2f}".format(sysutil.uptime()))
    hub.bootstrap_metrics = {"boot_seconds": sysutil.uptime(), "boot_time": hub.start_time}
    with msg.consumer.suspending():
        msg.send(hi_out)
        hub.crypto_key = hi_out.body.crypto_key
        with msg.consumer.wait_for("HostInit") as hi_in:
            hub.state = "initializing"
            reboot = bool(base.setup_hostname(hi_out, reboot=False))
            msg.handler.handle(hi_in)
        if hub.base.reboot_after_hostinit_phase or reboot:
            LOG.info("Rebooting server (Cause base.reboot_after_hostinit_phase is enabled)")

            # FAM-747
            # This will block until message consumer finishes handling current
            # ongoing request. If we skip the wait, reboot can happen before
            # successful response is sent to Scalr and HostInit will be
            # delivered again after reboot.
            msg.consumer.shutdown()

            reboot_now()


def initialize():
    """HostInitResponse message handling."""
    with msg.consumer.wait_for("HostInitResponse") as hir:
        msg.handler.handle(hir)
        docker_plugin.setup()
        host_up = msg.create("HostUp")
        hub.fire("host_init_response", hir)
        hub.fire("before_host_up", host_up)
        # setup_hostname() is useless here, because it has already been called in bootstrap()
        # and the second call will return cached value without host_up data update.
        host_up.body.base.hostname = fact["hostname"]  # pylint: disable=no-member
        msg.send(msg.create("BeforeHostUp"))
        with msg.consumer.wait_for("BeforeHostUp") as m:
            msg.handler.handle(m)
        msg.send(host_up)
    hub.state = "running"
    stat.push_bootstrap_metrics()


def import_server():
    if fact["manage_container_host"]:
        os_info = fact["container_host"]["os"]
    else:
        os_info = fact["os"]
    base.update_pkgmgr_db()
    hello = msg.create("Hello")
    hello.body.behaviour = base.get_supported_behaviors()
    hello.body.dist = {
        "distributor": os_info["name"],
        "release": str(os_info["release"]),
        "codename": os_info["codename"],
    }
    hello.body.os = {
        "arch": os_info["arch"],
        "version": "{} {} {}".format(
            os_info["name"], os_info["release"], os_info["codename"]
        ).strip(),
        "string_version": " ".join(platform.uname()).strip(),
    }
    hello.body.architecture = hello.body.os["arch"]
    hello.body.base.hostname = fact["hostname"]  # pylint: disable=no-member
    msg.send(hello)


def reboot_finish(force=False):
    cur_boot_id = sysutil.get_boot_id()
    if hub.state != "bootstrapping" and hub.boot_id and cur_boot_id != hub.boot_id:
        LOG.info("OS boot-id changed! (%s, previous: %s)", cur_boot_id, hub.boot_id)
        if not hub.halted:
            hub.rebooted = True
    hub.boot_id = cur_boot_id
    # handle
    if hub.rebooted or hub.halted or force:
        if hub.rebooted:
            LOG.info("Scalarizr resumed after reboot")
        elif hub.halted:
            LOG.info("Scalarizr resumed after server stop (same as reboot)")
        try:
            m = msg.create("RebootFinish")
            base.setup_hostname(m)
            hub.fire("before_reboot_finish", m)
            msg.send(m)
        finally:
            del hub.rebooted
            del hub.halted
            hub.fire("reboot_finish")


def update_finish():
    if not hub.updated:
        LOG.debug("Skipping HostUpdate fire, status: %s", hub.updateclient.status)
        return
    try:
        msg.send(msg.create("HostUpdate", body={"scalarizr": {"version": hub.agent_version}}))
    finally:
        del hub.updated


def commandline_parse():
    """Parse commandline arguments."""
    argv = list(sys.argv)
    p = argparse.ArgumentParser(description="Scalr agent on cloud server")
    p.add_argument(
        "-v", "--version", dest="version", action="store_true", help="Show version information"
    )
    p.add_argument(
        "-m",
        "--import",
        dest="import_server",
        action="store_true",
        default=False,
        help="Create image/role from non-Scalr server",
    )
    p.add_argument("-o", dest="cnf", action="append", help="Runtime .ini option key=value")
    p.add_argument("--debug", dest="debug", action="store_true", help="Debug mode")
    p.add_argument("--install", dest="install", action="store_true", help="Post install actions")
    if fact["os"]["name"] == "windows":
        p.add_argument(
            "--install-win-services",
            dest="install_win_services",
            action="store_true",
            default=False,
            help="Install Scalarizr service (Windows)",
        )
        p.add_argument(
            "--uninstall-win-services",
            dest="uninstall_win_services",
            action="store_true",
            default=False,
            help="Uninstall Scalarizr service (Windows)",
        )
    else:
        p.add_argument(
            "-z",
            dest="daemonize",
            action="store_true",
            default=False,
            help="Daemonize process (Linux)",
        )

    # Candidates for removal
    p.add_argument(
        "-c", "--etc-path", dest="etc_path", help="Configuration directory path (Deprecated)"
    )
    p.add_argument(
        "-n",
        "--configure",
        dest="configure",
        action="store_true",
        default=False,
        help="Configure Scalarizr. Use '-y -o' to configure Scalarizr"
        " non-interactively (Deprecated)",
    )
    p.add_argument(
        "-y",
        dest="yesall",
        action="store_true",
        default=False,
        help='Answer "yes" to all questions (Deprecated)',
    )

    # XXX: Code smell. This actually does a wrong thing: replaces region with availability_zone
    # Import somehow depends from this side effect
    # and we can't remove this until we find out why and rewrite
    if ("cloud-location=" in argv or "region=" in argv) and "platform=ec2" in argv:
        region = fact["ec2"]["placement_availability-zone"][:-1]
        try:
            argv[sys.argv.index("region=")] += region
        except ValueError:
            argv += ["-o", "region=" + region]
    return p.parse_args(argv[1:])


def commandline_handle(args):
    """Handle basic command line arguments."""
    # -v, --version
    if args.version:
        print("Scalarizr {}".format(hub.agent_version))
        raise SystemExit()
    # --debug
    if args.debug:
        logutil.set_console_log_level("scalarizr", logutil.DEBUG)

    # --install
    if args.install:
        LOG.info("Running post installation actions...")
        postinst.run()
        raise SystemExit()

    # -m, --import
    if args.import_server:
        LOG.info("Creating image from non-Scalr server...")
        LOG.warning("Don't terminate Agent until Scalr will create the new image!")
        if not args.debug:
            logutil.set_console_log_level("scalarizr", logutil.INFO)
        hub.state = "importing"
    elif hub.state == "importing":
        # Importing state valid to set only from command line.
        # So if agent was launched without --import, and stored state is 'importing',
        # it's a first start on a new server.
        del hub.state

    # OS specific options
    if fact["os"]["name"] == "windows":
        # install/uninstall windows service
        if args.install_win_services:
            # --install-win-services
            sys.argv = [sys.argv[0], "--startup", "manual", "install"]
            win32serviceutil.HandleCommandLine(WindowsService)
            time.sleep(1)  # related to SCALARIZR-2054
            raise SystemExit()
        elif args.uninstall_win_services:
            # --uninstall-win-services
            sys.argv = [sys.argv[0], "remove"]
            win32serviceutil.HandleCommandLine(WindowsService)
            raise SystemExit()
    else:
        if args.daemonize:
            # -z, --daemonize
            linuxutil.daemonize()
    _make_pid_file()


def _make_pid_file():
    if fact["os"]["name"] != "windows":
        other_pid = hub.pid
        this_pid = os.getpid()
        if other_pid and other_pid != this_pid:
            try:
                other = psutil.Process(other_pid)
                if other.name() == "scalarizr":
                    raise common_exc.DuplicateError(
                        "Other Scalarizr process already running: {}".format(other_pid)
                    )
            except psutil.NoSuchProcess:
                pass
        hub.pid = this_pid


def _before_halt():
    LOG.info("Server is shutting down...")
    hub.halted = True
    host_down = msg.create("HostDown")
    try:
        hub.fire("before_host_down", host_down)
    finally:
        msg.send(host_down)
    hub.fire("host_down")


def _before_reboot():
    LOG.info("Server is rebooting...")
    hub.rebooted = True
    reboot_start = msg.create("RebootStart")
    try:
        hub.fire("before_reboot_start", reboot_start)
    finally:
        msg.send(reboot_start)
    hub.fire("reboot_start")


class DockerPlugin(object):
    """
    Docker plugin management.
    """

    def __init__(self):
        self._server = None

    @property
    @mixutil.memorize()
    def policy_engine(self):
        return dockerauthz_plugin.PolicyEngine(hub.base.policies)

    def start(self):
        if os.path.exists(dockerauthz_plugin.SPEC_PATH):
            LOG.info("Starting HTTPServer for docker authz plugin")
            self._server = dockerauthz_plugin.HTTPServer(
                hub.docker_authz_socket, dockerauthz_plugin.UnixHTTPRequestHandler
            )
            self._server.policy_engine = self.policy_engine
            t = threading.Thread(target=self._server.serve_forever, name="docker-plugin")
            t.daemon = True
            t.start()
            LOG.debug("HTTPServer started.")

    def shutdown(self):
        if self._server:
            LOG.debug("Trying to shut down Docker plugin HTTPServer.")
            self._server.shutdown()
            LOG.debug("Docker plugin HTTPServer shutdown complete.")

    def is_active(self):
        if self._server:
            status = self._server.is_active()
            LOG.debug("HTTPServer is active: %s", status)
            return status
        else:
            return False

    def reload(self):
        if self._server:
            self._server.policy_engine = dockerauthz_plugin.PolicyEngine(hub.base.policies)
            LOG.debug("HTTPServer reloaded.")

    def is_docker_installed(self):
        status = bool(subprocess2.which("docker"))
        LOG.debug("docker is installed: %s", status)
        return status

    def setup(self):
        """
        Install and start Docker AuthZ plugin and then restart Docker if Docker is installed
        """
        try:
            if fact["os"]["name"] == "windows":
                return  # pragma: no cover
            if self.is_docker_installed() and not self.policy_engine.is_empty():
                LOG.info("Setting up docker authz plugin")
                dockerauthz_plugin.install_plugin()
                self.start()
                if service_tasks.Service("docker").running:
                    service_tasks.restart("docker")  # XXX: destroys all running containers!
                LOG.info("Docker authz plugin setup complete.")
        except Exception as e:
            LOG.error("Docker authz plugin setup failed! Reason: %s", e)


docker_plugin = DockerPlugin()


def _systemd_before_shutdown(targets_info):
    if any(t in targets_info for t in ("poweroff.target", "halt.target")):
        _before_halt()
    elif "reboot.target" in targets_info:
        _before_reboot()


def shutdown():
    """Shutdown main loop.
    - send notification to Scalr server
    - stop messaging
    - stop optional services: bollard, fatmouse
    """
    if fact["os"]["family"] is not "windows":
        if fact["manage_container_host"]:
            targets_info_path = os.path.join(
                agent.config.workflow_engine["run_dir"],
                "scalarizr.systemctl.list-units.type.target",
            )
            if os.path.exists(targets_info_path):
                with open(targets_info_path) as targets_info:
                    _systemd_before_shutdown(targets_info.read())
            else:
                LOG.warning(
                    "File '%s' doesn't exist. " "Scalarizr shutdown may be incorrectly processed",
                    targets_info_path,
                )
        elif fact["os"]["init"] == "systemd":
            # pylint: disable=unsupported-membership-test
            targets_info = subprocess2.check_output_host(
                "systemctl list-units --type target", shell=True
            )
            _systemd_before_shutdown(targets_info)
        else:
            runlevel = linuxutil.runlevel()
            if runlevel == 0:
                _before_halt()
            elif runlevel == 6:
                _before_reboot()

    hub.fire("shutdown")
    msg.consumer.shutdown()
    fatmouse.shutdown()
    stat.collector.shutdown()
    docker_plugin.shutdown()
    del hub.pid
    LOG.info("[pid: %s] Scalarizr terminated", os.getpid())


class Fatmouse(object):
    def __init__(self):
        self._proc = None

    def was_setup(self):
        """Whether fatmouse agent was set up?"""
        retval = hub.base.fatmouse and os.path.exists(hub.fatmouse_config_path)
        if not retval:
            LOG.debug("Workflow engine agent is not configured.")
        return retval

    def enabled(self):
        retval = hub.base.scalr_labs_rabbitmq_tunnel and hub.base.fatmouse
        if not retval:
            LOG.debug("Launch workflow engine agent is not enabled")
        return retval

    @classmethod
    def setup_local(cls):
        LOG.debug("Setting up local mode for workflow engine")
        celery.current_app.conf.task_always_eager = True
        # Rotate tasks
        hub.periodic_executor.add_task(
            celeryutil.TaskLocalDir.gc,
            agent.config.workflow_engine["gc_interval"],
            "Rotate finished tasks data.",
        )

    def setup(self):
        sysutil.rm_rf(hub.fatmouse_config_path)  # remove old configuration file.
        if not hub.base.fatmouse:
            LOG.info("No workflow engine configuration")
            return
        LOG.info("Setting up workflow engine agent")
        try:
            psr = cloudconfig.parse_script(hub.base.fatmouse.user_data)
            cloudconfig.run_script(datadir.instance_script(psr), psr.script_type)
        except Exception as e:
            LOG.warning(
                "Workflow engine agent configuration failed: %s", e, exc_info=sys.exc_info()
            )
        agent.config.reload()
        agent.app.configure_worker()
        celery.current_app.conf.task_always_eager = True  # app.send_task() depends from this.

    def _start_process(self):
        if not self.enabled() or not self.was_setup():
            return
        # Find debug log handler
        hdlr = None
        for h in logutil.get_logger("scalarizr").handlers:
            if (
                isinstance(h, logging.handlers.RotatingFileHandler)
                and os.path.basename(h.baseFilename) == "scalarizr_debug.log"
            ):
                hdlr = h
                break
        # Create fatmouse process
        LOG.info("Starting workflow engine agent")
        os.environ["SCALARIZR"] = "yes"
        if fact["os"]["name"] == "windows":
            self._proc = subprocess2.Popen(
                (
                    "C:\\opt\\scalarizr\\current\\embedded\\Scripts\\celery.exe",
                    "worker",
                    "-A",
                    "agent.app",
                    "--pool",
                    "threads",
                    "--maxtasksperchild",
                    "50",
                    "--concurrency",
                    "5",
                    "-l",
                    "DEBUG",
                ),
                stdout=os.dup(hdlr and hdlr.stream.fileno() or sys.stdout.fileno()),
                stderr=subprocess2.STDOUT,
            )
        else:
            self._proc = subprocess2.Popen(
                (
                    "/opt/scalarizr/embedded/bin/celery",
                    "worker",
                    "-A",
                    "agent.app",
                    "--pool",
                    "threads",
                    "--maxtasksperchild",
                    "50",
                    "-l",
                    "DEBUG",
                    "--concurrency",
                    "5",
                ),
                stdout=os.dup(hdlr and hdlr.stream.fileno() or sys.stdout.fileno()),
                stderr=subprocess2.STDOUT,
                preexec_fn=functools.partial(libc.prctl, 1, signal.SIGKILL),
            )
        # Start communicate() thread
        t = threading.Thread(target=self._proc.communicate, name="fatmouse-agent-communicate")
        t.setDaemon(True)
        t.start()

    def _supervisor(self):
        while True:
            try:
                if not self._proc or self._proc.poll() is not None:
                    self._start_process()
            except Exception as e:
                LOG.warning("Error in workflow engine agent start: %s", e, exc_info=sys.exc_info())
            finally:
                time.sleep(1)

    def start(self):
        if not self.enabled() or not self.was_setup():
            return
        t = threading.Thread(target=self._supervisor, name="fatmouse-agent")
        t.daemon = True
        t.start()

    @mixutil.error_logged("Error in workflow engine agent shutdown", reraise=False)
    def shutdown(self):
        if self._proc:
            self._proc.kill()


fatmouse = Fatmouse()


def adjust_ports():
    """Switch to alternative ports for api and messaging if the default ones are busy."""

    def port_busy(port):
        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            # pylint: disable=no-member
            return s.connect_ex(("0.0.0.0", port)) == 0

    updated = False
    for port_name, default_port, spare_port in types.SCALARIZR_PORTS:
        if hub.base[port_name] == default_port and port_busy(default_port):
            hub.base[port_name] = spare_port
            updated = True

    if updated:
        msg.send(
            msg.create(
                "UpdateControlPorts",
                {"api": hub.base.api_port, "messaging": hub.base.messaging_port},
            )
        )


def setup_environ():
    """Setup process ENV variables."""
    if fact["os"]["name"] == "windows":
        return
    pwr = pwd.getpwuid(os.getuid())
    os.environ["USER"] = pwr.pw_name
    os.environ["HOME"] = pwr.pw_dir
    # Ignore boto system config file
    os.environ["BOTO_CONFIG"] = ""


class Update(object):
    def __init__(self):
        self.client = service_tasks.Service(
            "ScalrUpdClient" if fact["os"]["name"] == "windows" else "scalr-upd-client"
        )
        self.needs_restart = False

    def _wait_actionable_state(self):
        try:
            if fact["manage_container_host"]:
                updater_host = "scalr-upd-client"
            else:
                updater_host = "127.0.0.1"
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(0.2)
            s.connect((updater_host, hub.base.update.api_port))
            s.close()
        except socket.error:
            LOG.debug("UpdateClient API (port: %s) connection timeouted", hub.base.update.api_port)
            return
        return hub.updateclient.status["state"] not in ("noop", "unknown")

    @mixutil.error_logged("Error in communication with UpdateClient", reraise=False)
    def wait(self):
        LOG.info("Polling UpdateClient state...")
        mixutil.try_until(
            self._wait_actionable_state,
            timeout=600,
            interval=1,
            error_msg="UpdateClient not reachable or in a wrong state. {}".format(
                exc.message_check_updateclient
            ),
        )
        s = hub.updateclient.status
        LOG.info("UpdateClient state: %s", s["state"])

        if s["state"].startswith("in-progress"):
            if s["state"] == "in-progress/rollback":
                version = s["service_version"]
            else:
                version = (s["candidate"] or s["installed"]).split("-")[0]
            if version != hub.agent_version:
                LOG.info("Update is in-progress. Exiting")
                raise SystemExit()
            mixutil.try_until(
                lambda: not hub.updateclient.status["state"].startswith("in-progress"),
                timeout=600,
                interval=1,
                error_msg="UpdateClient stucked in-progress",
            )
        if hub.updateclient.status["state"] == "completed/wait-ack":
            # Special case for bootstrapping from old agent versions without
            # latest scalrinit implementation.
            previous_version = hub.updateclient.status.get("service_version")
            if previous_version and mixutil.Version(previous_version) < mixutil.Version("4.6.0"):
                scalrinit_app.main_start()

            self.needs_restart = True

    @mixutil.error_logged("Error when restarting UpdateClient", reraise=False)
    def restart_if_needed(self):
        """Restarts UpdateClient if it was in the completed/wait-ack state
        after successful update.
        """
        if self.needs_restart:
            LOG.info("UpdateClient completed update and should be restarted. Restarting")
            self.client.restart()
            if hub.state == "bootstrapping":
                # Wait until UpdateClient is started completely after bootstrap, to prevent
                # race condition when it tries to connect to Scalr with new crypto key
                # genereated in HostInit, before Scalr processes HostInit.
                mixutil.try_until(
                    lambda: hub.updateclient.status["state"] == "completed",
                    timeout=600,
                    interval=1,
                    error_msg="UpdateClient didn't change its state to `completed` after restart",
                )


update = mixutil.Proxy(Update)


class HandlerCallback(object):
    """This a :class:`msg.Handler` callback, that handles incoming messages

    For ANY message the following steps are invoked:

      1. `base` section is synced into :attr:`hub`
      2. Cloud credentials are forwared to Platform.set_access_data()
      3. Global Variables flushed into /etc/profile.d
      4. the chain of Handler objects is called.

    The 4) step in a nutshell:

    Handler objects are scalarizr objects defined in scalarizr/handlers/*
    They are loaded by _load_handlers() from modules listed in /etc/scalr/*/*.ini
    under [handlers] section.

    The handling process is asking each Handler object via h.accept()
    - whether it could handle the message? and if True the subsequent
    h.on_<MessageName>() is called.
    """

    def __init__(self):
        self._modules = []
        self._handlers = []
        self._collect_modules()
        self._load_handlers()
        self._sort_handlers()
        LOG.debug("Message handlers chain:\n%s", pprint.pformat(self._handlers))
        hub.fire("init")

    def __call__(self, message):
        act = "handle-callback {} message ({})".format(message.name, message.id)
        LOG.debug("Begin %s", act)
        self._handle_sync_hub(message)
        self._handle_cloud_credentials(message)
        self._handle_global_variables(message)
        self._handle_with_handlers(message)
        LOG.debug("End %s", act)

    @mixutil.error_logged("Error in syncing data from message")
    def _handle_sync_hub(self, message):
        # TODO: small test - any message having 'base' section updates hub.base
        if isinstance(message.body, types.StateChanging):
            if message.body._initial.get("base"):  # pylint: disable=protected-access
                # Access model's protected '_initial'
                # for do not override hub.base with empty BaseData
                # when incoming message has no 'base' property
                hub.base.import_data(message.body.base.to_primitive())
            if message.body._initial.get("volumes"):  # pylint: disable=protected-access
                hub.volumes = message.body.volumes

    @mixutil.error_logged("Error in setting cloud credentials from message")
    def _handle_cloud_credentials(self, message):
        # set cloud credentials to platform object
        # TODO: small test - platform.set_access_data called with platform_access_data
        if message.body.platform_access_data:
            hub.platform.set_access_data(message.body.platform_access_data)

    @mixutil.error_logged("Error in setting global variables from message")
    def _handle_global_variables(self, message):
        if message.body.global_variables:
            base.sync_global_variables_from_message(message.body.global_variables)

    def _handle_with_handlers(self, message):
        accepted = False
        for h in self._handlers:
            try:
                if h.accept(
                    message, "control", behaviour=hub.behaviors, platform=hub.platform.name
                ):
                    accepted = True
                    act = "handle {} message ({}) with {}".format(
                        message.name, message.id, h.__class__.__name__
                    )
                    LOG.debug("Begin %s", act)
                    h(message)
                    LOG.debug("End %s", act)
            except Exception as e:
                # This is a special case, when HostInit or BeforeHost script triggered by self
                # and exited with non-zero code, should interrupt handlers chain and bubble up.
                # TODO: cover by medium
                if (
                    hub.base.abort_init_on_script_fail
                    and isinstance(e, (execute_tasks.ExecuteError, ValidationError))
                    and message.name in ("HostInit", "BeforeHostUp")
                    and message.body.server_id == hub.general.server_id
                ):
                    six.reraise(
                        exc.AbortInitializationError,
                        exc.AbortInitializationError(e),
                        sys.exc_info()[2],
                    )
                LOG.exception(e)
        if not accepted:
            LOG.debug("No one could handle '%s'", message.name)

    def _collect_modules(self):
        self._modules = hub.get_handlers()

    def _load_handlers(self):
        """This loads handler modules, discovers handler objects in them
        and constructs `self._handlers` list.
        """
        LOG.debug("Collecting message handlers...")
        # load handler objects
        self._handlers = []
        for module_s in self._modules:
            try:
                __import__(module_s)
                self._handlers.extend(sys.modules[module_s].get_handlers())
            except Exception as e:
                six.reraise(
                    Error,
                    Error("Can't load handlers (module {!r}): {}".format(module_s, e)),
                    sys.exc_info()[2],
                )

    def _sort_handlers(self):
        """This sorts handlers in `self._handlers` in the following order
        - IpListBuilder
        - block device handlers (i.e. BlockDeviceHandler and subclasses)
        - the rest
        It also applys the same handlers order for Hub "init" and "start" event listeners
        """

        def cls_priority(obj):
            cls = obj.__class__.__name__
            if cls == "IpListBuilder":
                return 1
            elif cls in ("EbsHandler", "BlockDeviceHandler"):
                return 10
            else:
                return 100

        # sort handlers
        self._handlers = list(sorted(self._handlers, key=cls_priority))
        # sort event listeners
        for ev in ("init", "start"):
            # pylint: disable=protected-access
            if ev in hub._listeners:
                hub._listeners[ev] = list(sorted(hub._listeners[ev], key=cls_priority))

    def get_handlers(self):
        # pylint: disable=protected-access
        return collections.OrderedDict(
            {
                "{}.{}".format(h.__class__.__module__, h.__class__.__name__): h
                for h in self._handlers
            }
        )


def reboot_now():
    """Reboot server right now."""
    sysutil.reboot("Rebooted by Scalr")
    threading.Event().wait(600)  # wait for service stop


def handles_fatal(fn):
    """Decorator to wrap function, any unhandled exception in which, should terminate application.
    Handles exception and interrupt main thread.

    :decorator:
    """

    @six.wraps(fn)
    def inner(*args, **kwargs):
        try:
            fn(*args, **kwargs)
        except BaseException as e:
            LOG.exception("Shutting down, cause: %s: %s", e.__class__.__name__, e)
            if servicemanager:
                servicemanager.LogMsg(
                    servicemanager.EVENTLOG_ERROR_TYPE,
                    servicemanager.PYS_SERVICE_STOPPED,
                    ("Scalarizr", " {}".format(e)),
                )
            # TODO: test interruption on Linux/Windows.
            # Windows probably must be stopped via module:`servicemanager` api
            six.moves._thread.interrupt_main()  # pylint: disable=protected-access

    return inner


if win32serviceutil:

    class WindowsService(win32serviceutil.ServiceFramework, object):
        """Scalarizr windows service."""

        _svc_name_ = "Scalarizr"
        _svc_display_name_ = "Scalarizr"
        _svc_deps_ = ["EventLog"]

        def __init__(self, args=None):

            # When run as a Windows service sys.stdout and sys.stderr are None,
            # but some modules, like standard python's http server, expect it to be
            # a valid file (https://bugs.python.org/issue28294).
            sys.stdout = sys.stderr = open("nul", "w")

            if args is not None:
                super(WindowsService, self).__init__(args)

            def handler(*args):
                return True

            win32api.SetConsoleCtrlHandler(handler, True)
            self.h_stop = win32event.CreateEvent(None, 0, 0, None)

        @mixutil.error_logged("Error in windows service main")
        def SvcDoRun(self):
            """
            Windows service main loop.
            """
            try:
                servicemanager.LogMsg(
                    servicemanager.EVENTLOG_INFORMATION_TYPE,
                    servicemanager.PYS_SERVICE_STARTED,
                    (self._svc_name_, ""),
                )
            except Exception as e:
                LOG.warning(
                    "Got exception while sending Service Started message to Windows Log: %s", e
                )
            try:
                main()
            except SystemExit:
                servicemanager.LogMsg(
                    servicemanager.EVENTLOG_INFORMATION_TYPE,
                    servicemanager.PYS_SERVICE_STOPPED,
                    (self._svc_name_, ""),
                )
                shutdown()
                return
            LOG.info("Service main loop (Windows)")
            while True:
                rc = win32event.WaitForSingleObject(self.h_stop, 3600)
                if rc == win32event.WAIT_OBJECT_0:
                    try:
                        servicemanager.LogMsg(
                            servicemanager.EVENTLOG_INFORMATION_TYPE,
                            servicemanager.PYS_SERVICE_STOPPED,
                            (self._svc_name_, ""),
                        )
                    except Exception as e:
                        LOG.warning(
                            "Got exception while sending Service Stopped message to Windows Log: %s",
                            e,
                        )
                    break
            shutdown()

        @mixutil.error_logged("Error in windows service stop")
        def SvcStop(self):
            """
            Windows service stop actions.
            """
            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
            win32event.SetEvent(self.h_stop)

        @mixutil.error_logged("Error in windows service shutdown")
        def SvcShutdown(self):
            """
            Windows service shutdown actions.
            """
            if hub.state != "running":
                LOG.debug('Skiped Win_HostDown, cause state is "%s"', hub.state)
                return
            if hub.rebooted:
                LOG.debug("Shutdown already handled, cause .reboot flag is set")
                return
            hub.rebooted = True
            msg.send(msg.create("Win_HostDown"))


if __name__ == "__main__":
    main()
