import copy
import os
import re
import sys

import common.types
from agent.dockerauthz import plugin as dockerauthz_plugin
from agent.tasks import firewall
from agent.tasks import pkgmgr
from agent.tasks import service as service_tasks
from agent.tasks import sys as sys_tasks
from common.utils import exc
from common.utils import imputil
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

from . import msg
from . import types
from .hub import hub

try:
    from common.utils import linuxutil
except ImportError:
    linuxutil = None

if fact["os"]["name"] != "windows":
    from common.utils import linuxutil
else:
    linuxutil = None


SCALR_USERNAMES = ["root", "scalr", "ubuntu", "centos", "ec2-user"]  # FAM-1196

LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    pass


# pylint: disable=locally-disabled,too-many-branches
def _update_ssh_authorized_keys(
    home_dir, keys_to_add=None, keys_to_remove=None, tags_to_remove=None
):
    """Adds or removes sets of SSH keys from ~/.ssh/authorized_keys for single home directory
    :param keys_to_add: list of keys to add
    :param keys_to_remove: list of keys to remove
    :param tags_to_remove: list of tags, keys with that tags will be removed
    :param home_dir: home directory, to manage authorized_keys file into. default is '~/'
    """
    if not os.path.exists(home_dir):
        return

    owner_uid = os.stat(home_dir).st_uid
    owner_gid = os.stat(home_dir).st_gid

    authorized_keys_path = os.path.join(home_dir, ".ssh/authorized_keys")

    # ensure authorized_keys directory
    authorized_keys_dir = os.path.dirname(authorized_keys_path)
    if not os.path.exists(authorized_keys_dir):
        LOG.debug("Creating directory %s", authorized_keys_dir)
        sysutil.mkdir_p(authorized_keys_dir)
        os.chown(authorized_keys_dir, owner_uid, owner_gid)

    # read authorized_keys
    authorized_keys_file = ""

    if os.path.exists(authorized_keys_path):
        os.chmod(authorized_keys_path, 0o600)
        LOG.debug("Reading SSH keys from %s", authorized_keys_path)
        with open(authorized_keys_path) as fp:
            authorized_keys_file = fp.read()

    # remove keys
    if keys_to_remove and authorized_keys_file:
        LOG.info("Removing %s ssh key(s) from %s", len(keys_to_remove), authorized_keys_path)
        for key in keys_to_remove:
            authorized_keys_file = authorized_keys_file.replace(key, "")
    elif not authorized_keys_file:
        LOG.debug("No keys found. Keys file %s is empty", authorized_keys_path)

    # remove keys by tag
    if tags_to_remove and authorized_keys_file:
        LOG.info("Removing ssh key(s) by tag %s from %s", len(tags_to_remove), authorized_keys_path)
        key_record_list = authorized_keys_file.split("\n")
        for key_record in key_record_list:
            listed_record = key_record.split()
            if len(listed_record) == 3 and listed_record[2] in tags_to_remove:
                key_record_list.remove(key_record)
        authorized_keys_file = "\n".join(key_record_list)

    # add keys
    if keys_to_add:
        LOG.info("Adding %s ssh key(s) to %s", len(keys_to_add), authorized_keys_path)
        for key in keys_to_add:
            if key not in authorized_keys_file:
                authorized_keys_file += "\n%s\n" % key
            else:
                # handle 'Please login as the user "ubuntu" rather than the user "root".'
                authorized_keys_file = authorized_keys_file.replace(key, "\n%s\n" % key)

    while "\n\n" in authorized_keys_file:
        authorized_keys_file = authorized_keys_file.replace("\n\n", "\n")

    if not authorized_keys_file.endswith("\n"):
        authorized_keys_file += "\n"

    try:
        with open(authorized_keys_path, "w") as fp:
            fp.write(authorized_keys_file)
    except IOError:
        six.reraise(
            Error,
            Error("Error writing SSH authorized keys file {}".format(authorized_keys_path)),
            sys.exc_info(),
        )
    os.chown(authorized_keys_path, owner_uid, owner_gid)
    os.chmod(authorized_keys_path, 0o600)


def update_ssh_authorized_keys(
    keys_to_add=None,
    keys_to_remove=None,
    tags_to_remove=None,
    home_dir=None,
    update_all_users=False,
):
    """Add or removes sets of SSH keys from ~/.ssh/authorized_keys.

    :param keys_to_add: list of keys to add;
    :param keys_to_remove: list of keys to remove;
    :param tags_to_remove: list of tags, keys with that tags will be removed;
    :param home_dir: home directory, to manage authorized_keys file into.
                     default is '~/'.
    :param update_all_users: When True ssh keys will be updated for every linux user managed by Scalr
    """
    if fact["os"]["family"] == "windows":
        return

    if not keys_to_add and not keys_to_remove:
        LOG.debug("Empty keys lists. Nothing to do.")
        return

    home_dirs = []
    updated = []

    if update_all_users:
        home_dirs = [
            os.path.expanduser("~" + username)
            for username in SCALR_USERNAMES
            if os.path.exists(os.path.expanduser("~" + username))
        ]

    if home_dir:
        if not os.path.exists(home_dir):
            LOG.warning("SSH authorized keys has not been updated: %s does not exist.", home_dir)
        elif home_dir not in home_dirs:
            home_dirs.append(home_dir)

    for directory in home_dirs:
        _update_ssh_authorized_keys(directory, keys_to_add, keys_to_remove, tags_to_remove)
        updated.append(directory)

    if not updated:
        LOG.warning("SSH authorized keys has not been updated.")


def add_root_ssh_authorized_key():
    """Insert root's public key to authorized_keys file."""
    if hub.general.platform == "ec2":
        LOG.info("Provision SSH public key")
        pub_key = hub.platform.get_ssh_pub_key()
        if fact["manage_container_host"]:
            update_ssh_authorized_keys(keys_to_add=(pub_key,), home_dir="/rootfs/root")
        else:
            update_ssh_authorized_keys(
                keys_to_add=(pub_key,), home_dir="/root", update_all_users=True
            )


def generate_ssh_keypair(path, force=None):
    """Generates ssh key pair.

    :param path: full path where key pair will be stored
    :param force: if True overwrite existed key pair
    """
    if force and os.path.isfile(path):
        LOG.debug("SSH key pair will be recreated")
        sysutil.rm_rf(path)

    if not os.path.isfile(path):
        LOG.info("Generating SSH key pair (path: %s)", path)
        ssh_dir = "/".join(path.split("/")[:-1])
        sysutil.mkdir_p(ssh_dir, mode=0o755)
        cmd = 'ssh-keygen -b 2048 -t rsa -C "scalr-agent" -f {} -q -N ""'.format(path)
        subprocess2.check_output(cmd, shell=True)
        LOG.debug("SSH key pair was created (path: %s)", path)
    else:
        LOG.debug("SSH key pair already exists (path: %s)", path)

    def _read_file(file_path):
        with open(file_path) as f:
            return f.read().strip()

    private_key = _read_file(path)
    public_key = _read_file(path + ".pub")
    return private_key, public_key


def _install_ssh_keypair(key_path):
    root_priv_path = "/root/.ssh/id_rsa"
    root_pub_path = "/root/.ssh/id_rsa.pub"
    sysutil.mkdir_p("/root/.ssh", mode=0o755)
    if not os.path.islink(root_priv_path):
        sysutil.symlink(key_path, root_priv_path)
        LOG.debug("Link %s to %s", key_path, root_priv_path)
    if not os.path.islink(root_pub_path):
        sysutil.symlink(key_path + ".pub", root_pub_path)
        LOG.debug("Link %s.pub to %s", key_path + ".pub", root_pub_path)


# pylint: disable=locally-disabled,invalid-name
def container_host_add_ssh_authorized_key(force=False):
    """
    If scalarizr run inside container will add ssh key to host machine for management purposes
    """
    if fact["manage_container_host"]:
        LOG.info("Adding management ssh key to host machine")
        key_path = "/etc/scalr/private.d/.ssh/id_rsa"
        public_key = generate_ssh_keypair(key_path, force)[1]
        _install_ssh_keypair(key_path)
        update_ssh_authorized_keys(
            keys_to_add=[public_key], tags_to_remove=["scalr-agent"], home_dir="/rootfs/root"
        )


def check_behaviors():
    # pylint: disable=no-name-in-module,import-error
    from scalarizr import api
    from scalarizr import exceptions

    if fact["os"]["name"] == "windows":
        return
    LOG.info(
        "Checking that assigned role automations %s " "are compatible with current environment",
        hub.behaviors,
    )

    installed_packages = {p.name: p.version for p in pkgmgr.get_installed()}
    for behavior in hub.behaviors:
        if behavior in ["base", "mongodb"] or behavior not in api.api_routes.keys():
            continue
        try:
            api_cls = imputil.import_attr(api.api_routes[behavior])
            api_cls.check_software(installed_packages)
            check_behavior_deprecation(behavior)
        except exceptions.NotFound as e:
            LOG.error(e)
        except exceptions.UnsupportedBehavior as e:
            if e.args[0] == "chef":
                # We pass it, cause a lot of roles has chef behavior without chef installed on them
                continue
            err = ("Role automation '{}' is no more compatible " "with the environment: {}").format(
                e.args[0], e.args[1]
            )
            msg.send("RuntimeError", {"code": "UnsupportedBehavior", "message": err})
            six.reraise(exc.NotSupportedError, exc.NotSupportedError(err), sys.exc_info()[2])


def check_behavior_deprecation(behaviors):
    """
    Raises :class:`exceptions.UnsupportedBehavior` for deprecated behavior(s).

    :type behaviors: str or list
    :raises :class:`exceptions.UnsupportedBehavior`
    """
    # pylint: disable=no-name-in-module,import-error
    from scalarizr import exceptions

    if isinstance(behaviors, str):
        behaviors = [behaviors]

    if (fact["os"]["name"] == "ubuntu" and fact["os"]["release"] >= (18, 4)) or (
        fact["os"]["name"] == "amazon" and fact["os"]["release"] == 2
    ):
        supported = ("base", "chef", "router")
        for behavior in behaviors or []:
            if behavior not in supported:
                raise exceptions.UnsupportedBehavior(
                    behavior,
                    (
                        "{0} role automation is deprecated for Ubuntu 18.04, "
                        "Amazon Linux 2 or later. Supported automations: {1}."
                    ).format(behavior, ", ".join(supported)),
                )


def clean_previous_server(chef=True, systemd=True, docker=True, etc_dir=True):
    """Clean previous server data."""
    LOG.info("Cleaning previous server data...")
    if chef:
        _cleanup_chef()
    if systemd:
        _cleanup_systemd()
    if docker:
        _cleanup_docker()
        _cleanup_docker_plugin()
    if etc_dir:
        _cleanup_etc_dir()


def _cleanup_docker():
    """Cleanup failed scalr containers"""
    LOG.info("Cleaning previous scalr managed containers")
    if fact["manage_container_host"]:
        cmd = 'docker ps -q -f "status=exited" -f "label=com.scalr.agent.version"'
        container_to_dell = subprocess2.check_output_host(cmd)
        for container in container_to_dell.split():
            subprocess2.check_output_host("docker rm %s" % container)


def _cleanup_docker_plugin():
    """Cleanup previous AUTHZ setup"""
    uninstalled = dockerauthz_plugin.uninstall_plugin()
    if uninstalled and service_tasks.Service("docker").running:
        service_tasks.restart("docker")  # XXX: destroys all running containers!


def _cleanup_chef():
    """Cleanup chef-client configuration"""
    LOG.info("Cleaning chef-client configuration")
    if os.path.exists("/etc/chef/client.pem"):
        os.remove("/etc/chef/client.pem")
    if os.path.exists("/etc/chef/client.rb"):
        os.remove("/etc/chef/client.rb")


def _cleanup_systemd():
    """Remove systemd generated fstab units for scalr-managed disks."""
    if fact["os"]["name"] != "windows":
        LOG.info("Cleaning previous scalr managed disks configuration")
        if fact["manage_container_host"]:
            root_fs = "/rootfs"
        else:
            root_fs = "/"
        fstab = linuxutil.fstab()
        systemd = os.path.realpath("/sbin/init").endswith("systemd")
        entries = [entry for entry in fstab if "comment=scalr" in entry.options]
        for entry in entries:
            linuxutil.umount(
                entry.device
            )  # TODO: why scalr disks umount associated with systemd routine?
            if systemd:
                mpoint = entry.mpoint
                unit_name = "{}.mount".format(mpoint.replace("/", "-")[1:])
                LOG.debug("Removing systemd runtime unit %s", unit_name)
                sysutil.rm_rf(
                    os.path.join(
                        root_fs, "run/systemd/generator/local-fs.target.wants/{}".format(unit_name)
                    )
                )
                sysutil.rm_rf(os.path.join(root_fs, "run/systemd/generator/{}".format(unit_name)))
        if systemd:
            subprocess2.call_host("systemctl daemon-reload", shell=True)


def _cleanup_etc_dir():
    """Delete scalarizr previous configuration."""
    LOG.info("Deleting previous agent configuration")
    sysutil.rm_rf(hub.fatmouse_config_path)
    for file in os.listdir(hub.private_dir):
        # protect user-data and UpdateClient status
        # keys/default maybe already refreshed by UpdateClient
        if file in (".ssh", ".state", ".user-data", "update.status", "keys", "scalr_labs_corev2"):
            continue
        path = os.path.join(hub.private_dir, file)
        if not fact["os"]["name"] == "windows":
            linuxutil.chmod_r(path, 0o700)
        try:
            sysutil.rm_rf(path)
        except Exception as e:
            if fact["os"]["name"] == "windows" and isinstance(e, OSError):
                # ScalrUpdClient locks db.sqlite
                LOG.debug(e)
            else:
                raise
    if fact["os"]["name"] != "windows":
        subprocess2.check_call("sync", shell=True)


def cloudinit_enable_root_login():
    """Enable ssh root loging in cloud-init configuration."""
    if fact["os"]["name"] != "windows" and hub.general.platform != "gce":
        for path in ("/etc/cloud/cloud.cfg", "/etc/ec2-init/ec2-config.cfg"):
            if os.path.exists(path):
                LOG.info("Enabling SSH root login in Cloud-Init (%s)", path)
                c = None
                with open(path, "r") as fp:
                    c = fp.read()
                c = re.sub(re.compile(r"^disable_root[^:=]*([:=]).*", re.M), r"disable_root\1 0", c)
                with open(path, "w") as fp:
                    fp.write(c)


def get_supported_behaviors():
    # pylint: disable=no-name-in-module,import-error
    from scalarizr import api
    from scalarizr import exceptions

    LOG.info("Detecting supported behaviors...")
    if fact["os"]["name"] != "windows":
        installed_packages = {p.name: p.version for p in pkgmgr.get_installed()}
        behaviors = types.Behavior.values()
    else:
        installed_packages = []
        behaviors = ("base", "chef")

    msg = (
        "Scalr built-in automation: checking for supported software.\n"
        "If installed software isn't detected, "
        "review the Scalr Wiki: https://scalr-wiki.atlassian.net/wiki/x/IoB1"
    )
    LOG.info(msg)

    supported_behaviors = []
    for behavior in behaviors:
        if behavior in ["base", "mongodb"] or behavior not in api.api_routes.keys():
            continue
        try:
            api_cls = imputil.import_attr(api.api_routes[behavior])
            installed = api_cls.check_software(installed_packages)
            check_behavior_deprecation(behavior)

            LOG.info("%s: Available. Installed version: %s", behavior, installed[1])
            supported_behaviors.append(behavior)
        except (exceptions.NotFound, exceptions.UnsupportedBehavior, ImportError) as e:
            if isinstance(e, exceptions.UnsupportedBehavior):
                LOG.info("%s: %s", behavior, e.args[1])
            else:
                LOG.info("%s: %s", behavior, e)
            continue
    return supported_behaviors


def mount_all():
    if fact["os"]["name"] != "windows":
        subprocess2.run(("mount", "-a"))


def setup_firewall():
    """Open Scalarizr ports in firewall (when firewall enabled)"""
    if hub.base.disable_firewall_management:
        LOG.debug("Firewall management is disabled in Scalr. Skipping firewall rules setup")
        return
    if not firewall.enabled():
        LOG.debug("Firewall is disabled in OS. Skipping firewall rules setup")
        return
    rule = dict(
        action="allow",
        direction="in",
        protocol="tcp",
        from_port=8008,
        to_port=8014,
        source="0.0.0.0/0",
        comment="Scalr agent",
    )
    firewall.ensure(rule)


def update_pkgmgr_db():
    """Update OS package manager database (Linux-only).

    Catches any Exception and logs with WARN level.
    """
    if fact["os"]["name"] != "windows" and not fact["manage_container_host"]:
        try:
            pkgmgr.updatedb()
        except Exception as e:
            LOG.warning(
                "Error in OS package manager database update: %s: %s",
                e.__class__.__name__,
                e,
                exc_info=sys.exc_info(),
            )


@mixutil.memorize()
def setup_hostname(message=None, reboot=False):
    """Assign sanitized hostname to server (run once)"""
    if hub.base.hostname:
        hostname = hub.base.hostname
    elif (
        hub.general.platform == "ec2"
        and fact["os"]["name"] != "windows"
        and hub.ec2.hostname_as_pubdns
        and hub.ec2.hostname_as_pubdns.isdigit()
        and int(hub.ec2.hostname_as_pubdns)
    ):
        hostname = hub.platform.get_public_hostname()
    else:
        hostname = None

    if hub.base.domain_name:
        domain_name = hub.base.domain_name
    else:
        domain_name = None

    if message is not None:
        message.body.base.hostname = hostname or fact["hostname"]
        message.body.base.domain_name = domain_name or fact["domain_name"]

    if hub.base.disable_hostname_management:
        LOG.debug("Hostname management is disabled in Scalr. Skipping hostname setup")
        return

    if not hostname and not domain_name:
        LOG.debug("Setup hostname skipped: FarmRole hostname and domain name settings are empty.")
        return

    try:
        return sys_tasks.set_hostname(hostname, domain_name=domain_name, reboot=reboot)
    except Exception as e:
        err_msg = "Error setting up server hostname: {} (hostname={}), (domain_name={})".format(
            e, hostname, domain_name
        )
        six.reraise(Error, Error(err_msg), sys.exc_info()[2])


def setup_os_license():
    """Setup licensing (currently RedHat-only)"""
    if fact["os"]["name"] == "redhat" and hub.general.platform == "ec2":
        LOG.debug("Enabling RedHat subscription")
        is_enabled = fact["ec2"]["dynamic"]["instance-identity_document"]
        LOG.debug("RedHat License is enabled: %s", is_enabled)


def sync_global_variables():
    if hub.state in ("importing", "discovering"):
        LOG.debug(
            (
                "Sync global variables skipped: "
                'Scalr does not implement QueryEnv.list_global_variables() for server state "%s"'
            ),
            hub.state,
        )
        return
    gvs = []
    for name, value in hub.queryenv.list_global_variables()["public"].items():
        gv = common.types.GlobalVariable()
        gv.name = name
        gv.value = value
        gv.hidden = False
        gvs.append(gv)
    LOG.info("Synchronizing GVs from Scalr")
    sys_tasks.set_global_variables(gvs)


def sync_global_variables_from_message(message_gvs):
    gvs = []
    for gv0 in copy.deepcopy(message_gvs):
        gv = common.types.GlobalVariable()
        gv.name = gv0["name"]
        gv.value = gv0["value"]
        gv.hidden = gv0.get("private", False)
        gvs.append(gv)
    LOG.debug("Updating GVs from message")
    sys_tasks.set_global_variables(gvs)


def sync_time():
    """Sync server datetime"""
    if hub.base.disable_ntp_management:
        LOG.debug("NTP management is disabled in Scalr. Time sync skipped")
        return
    try:
        sys_tasks.sync_time()
    except exc.NotSupportedError as e:
        LOG.warning("Error synchronizing server time: %s", e)
    except Exception as e:
        # Time sync involves 3rd party (e.g, pool.ntp.org)
        # Catch all prevents server bootstrap failure if the 3rd party is down.
        LOG.warning("Error synchronizing server time: %s", e, exc_info=sys.exc_info())


def provision_ansible_host():
    """
    Prepare instance for ansible management.
    """

    if not hub.base.ansible:
        return

    def _setup_windows():
        for credentials in hub.base.ansible.machine_credentials:
            LOG.info('Creating user "%s"', credentials.username)
            sys_tasks.add_win_user(
                username=credentials.username,
                password=credentials.password,
                groups=["Administrators"],
            )

    def _setup_linux():
        base_home_dir = "/rootfs/home" if fact["manage_container_host"] else "/home"
        sshd_config_requirements = {"RSAAuthentication": "yes", "PubkeyAuthentication": "yes"}
        if sys_tasks.update_sshd_config(sshd_config_requirements):
            LOG.info("Enabling public key authentication in sshd")
            if not fact["manage_container_host"]:
                service_tasks.Service("sshd").reload()
        for credentials in hub.base.ansible.machine_credentials:
            sudo = True if credentials.become_method == "sudo" else None
            LOG.info('Creating user "%s"', credentials.username)
            if sudo:
                LOG.info('Adding user "%s" into sudoers', credentials.username)
            sys_tasks.add_user(username=credentials.username, sudoers=sudo)
            update_ssh_authorized_keys(
                keys_to_add=[credentials.ssh_public_key],
                home_dir=os.path.join(base_home_dir, credentials.username),
            )

    LOG.info("Prepare instance for ansible management")
    if fact["os"]["name"] == "windows":
        _setup_windows()
    else:
        _setup_linux()
