import os
import shutil

from common.utils import logutil
from common.utils import mixutil
from common.utils import pkgutil
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

from .. import config
from .hub import hub

LOG = logutil.get_logger("agent.scalarizr.compat")

SINGLE_PACKAGE_FOR_LINUX_VER = "6.3.0"
SCALR_VER_QUERYENV_LFRP_HAS_VOLUMES = (7, 7, 6)  # SCALRCORE-7274


def _uninstall_platform_package_maybe():  # pylint: disable=invalid-name
    """
    Uninstalls scalarizr-{platform} package if that package is not required.

    Note: for deb/rpm only.
    """
    os_family = fact["os"]["family"]
    os_family_map = {"redhat": "rpm", "suse": "suse", "debian": "deb"}

    if fact["manage_container_host"] or os_family not in os_family_map:
        return

    package_mgr_cls = pkgutil.get_pkgmgr(os_family_map[os_family])
    package_name = config.workflow_engine["package_name"]
    platform_package_name = "{0}-{1}".format(package_name, hub.general.platform)

    installed_version = package_mgr_cls.get_installed_version(package_name)
    if mixutil.Version(installed_version) < mixutil.Version(SINGLE_PACKAGE_FOR_LINUX_VER):
        return

    installed_platform_version = package_mgr_cls.get_installed_version(platform_package_name)
    if installed_platform_version is not None:
        if os_family == "debian":
            LOG.debug(
                "Marking %s package as manually installed before uninstalling "
                "deprecated %s package...",
                package_name,
                platform_package_name,
            )
            subprocess2.run(["apt-mark", "manual", "scalarizr", "--option", "dir::cache=/tmp"])

        LOG.debug("Uninstalling deprecated platform package %s...", platform_package_name)
        package_mgr_cls.uninstall(platform_package_name)


def _migrate_platform_package_cache_maybe():  # pylint: disable=invalid-name
    """
    Migrates platform package cache.
    """
    cache_dir = hub.updateclient.cache_dir
    index_dir = os.path.join(cache_dir, ".index")
    package_name = config.workflow_engine["package_name"]
    platform_package_name = "{0}-{1}".format(package_name, hub.general.platform)
    packages_cache_dir = os.path.join(cache_dir, package_name)
    platform_packages_cache_dir = os.path.join(cache_dir, platform_package_name)

    if not os.path.exists(platform_packages_cache_dir):
        return
    LOG.debug("Moving %s package cache to %s directory...", platform_package_name, package_name)
    sysutil.mkdir_p(index_dir, mode=0o755)
    sysutil.mkdir_p(packages_cache_dir, mode=0o755)

    for package_hash in os.listdir(platform_packages_cache_dir):
        package_src = os.path.join(platform_packages_cache_dir, package_hash)
        package_dst = os.path.join(packages_cache_dir, package_hash)
        package_index_link = os.path.join(index_dir, package_hash)

        # move directory and make a new symlink
        if not os.path.exists(package_dst):
            shutil.move(package_src, package_dst)
        if os.path.lexists(package_index_link):
            os.remove(package_index_link)
        os.symlink(package_dst, package_index_link)
        LOG.debug("Moved %s to %s.", package_src, package_dst)
    shutil.rmtree(platform_packages_cache_dir)


def handle_platform_package_deprecation():  # pylint: disable=invalid-name
    """
    Try to uninstall outdated scalarizr-{platform} package and migrate cache.
    """
    try:
        _uninstall_platform_package_maybe()
    except Exception as e:
        LOG.warning("Cannot uninstall scalarizr platform meta-package, reason: %s", e)
    try:
        _migrate_platform_package_cache_maybe()
    except Exception as e:
        LOG.warning("Cannot migrate cache of scalarizr platform meta-package, reason: %s", e)
