"""All Scalarizr configuration settings from all places under the single roof.
"""
from __future__ import absolute_import

import contextlib
import glob
import io
import json
import os
import pprint
import sqlite3
import stat
import sys
import textwrap

import agent.config
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils import sysutil
from common.utils.facts import fact

from . import metadata
from . import types
from .utils import mixutil as agent_mixutil

LOG = logutil.get_logger(__name__)


class TempChmod(object):
    def __init__(self, fn):
        self.file_name = fn

    def __enter__(self):
        if os.path.exists(self.file_name):
            os.chmod(self.file_name, 0o600)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if os.path.exists(self.file_name):
            os.chmod(self.file_name, 0o400)


class _SectionProxy(object):
    """This class implements attribute access to a single section of a ConfigParser"""

    def __init__(self, section):
        object.__setattr__(self, "_section", section)

    def set_cp(self, cp):
        object.__setattr__(self, "_cp", cp)

    def __getattribute__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            pass

        try:
            return object.__getattribute__(self, "_cp").get(
                object.__getattribute__(self, "_section"), key
            )
        except six.moves.configparser.NoOptionError:
            return None

    def __setattr__(self, key, value):
        def setter():
            object.__getattribute__(self, "_cp").set(
                object.__getattribute__(self, "_section"), key, str(value)
            )
            object.__setattr__(self, key, value)

        try:
            setter()
        except six.moves.configparser.NoSectionError:
            object.__getattribute__(self, "_cp").addsection(
                object.__getattribute__(self, "_section")
            )
            setter()


class _SectionProperty(object):
    """This is a property returning :class:`_SectionProxy`"""

    def __init__(self, section):
        self.section = section
        self._proxy = _SectionProxy(self.section)

    def __get__(self, obj, objtype):
        # pylint: disable=protected-access
        self._proxy.set_cp(obj._cp)
        return self._proxy


class _FileProperty(object):
    def __init__(self, filename, default=None):
        self._filename = filename
        self._default = default

    def __get__(self, obj, objtype):
        try:
            with open(self._filename) as fp:
                return fp.read().strip()
        except IOError:
            # TODO: log why
            return self._default

    def __set__(self, obj, value):
        sysutil.mkdir_p(os.path.dirname(self._filename))
        with open(self._filename, "w+") as fp:
            return fp.write(str(value).strip())

    def __delete__(self, obj):
        if os.path.exists(self._filename):
            os.unlink(self._filename)


class _StateFileProperty(_FileProperty):
    def __set__(self, obj, value):
        LOG.info("State: %s", value)
        return super(_StateFileProperty, self).__set__(obj, value)


class _JsonFileProperty(_FileProperty):
    def __get__(self, obj, objtype):
        v = super(_JsonFileProperty, self).__get__(obj, objtype)
        if v:
            try:
                data = json.loads(v)
            except (ValueError, TypeError):
                data = self._default
            return data
        else:
            return self._default

    def __set__(self, obj, value):
        return super(_JsonFileProperty, self).__set__(obj, json.dumps(value))


class _CryptoKeyProperty(_FileProperty):
    """This is a getter/setter for Scalarizr crypto key"""

    def __set__(self, obj, value):
        with TempChmod(self._filename):
            super(_CryptoKeyProperty, self).__set__(obj, value)


class _FileFlagProperty(_FileProperty):
    """This class represents a boolean property backed by a filename"""

    def __get__(self, obj, objtype):
        return os.path.exists(self._filename)

    def __set__(self, obj, value):
        sysutil.touch(self._filename)


class _PidFileProperty(_FileProperty):
    """This class represent an intener property backed by a file"""

    def __get__(self, obj, objtype):
        v = super(_PidFileProperty, self).__get__(obj, objtype)
        try:
            return int(v)
        except (TypeError, ValueError):
            return 0


class DictProxy(dict):
    """Class implements attribute access to dictionary"""

    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


class _Hub(mixutil.Observable):
    """All Scalarizr configuration settings from all places under the single roof"""

    if fact["os"]["family"] == "windows":
        root_dir = "C:\\opt\\scalarizr"
        install_dir = os.path.join(root_dir, "current")
        etc_dir = os.path.join(root_dir, "etc")
        lib_dir = os.path.join(root_dir, "var", "lib")
        log_dir = os.path.join(root_dir, "var", "log")
        scripting_log_dir = os.path.join(log_dir, "scripting")
        run_dir = os.path.join(root_dir, "var", "run")
    else:
        install_dir = "/opt/scalarizr"
        etc_dir = "/etc/scalr"
        lib_dir = "/var/lib/scalarizr"
        log_dir = "/var/log"
        scripting_log_dir = os.path.join(log_dir, "scalarizr", "scripting")
        run_dir = "/var/run"
        home_dir = "/root/.scalr"  # required by configuration presets
        embedded_bin_dir = os.path.join(install_dir, "embedded", "bin")
    # TODO: do and test it later
    # agent.config.workflow_engine['etc_dir'] = etc_dir
    # agent.config.workflow_engine['home_dir'] = lib_dir
    # agent.config.workflow_engine['cloud_dir'] = os.path.join(lib_dir, 'cloud')
    # agent.config.workflow_engine['cache_dir'] = os.path.join(lib_dir, 'cache')
    # agent.config.workflow_engine['run_dir'] = run_dir

    private_dir = os.path.join(etc_dir, "private.d")
    public_dir = os.path.join(etc_dir, "public.d")
    scripts_dir = os.path.join(install_dir, "scripts")
    crypto_key_path = os.path.join(private_dir, "keys", "default")
    crypto_key = _CryptoKeyProperty(crypto_key_path)
    db_path = os.path.join(private_dir, "db.sqlite")
    farm_crypto_key_path = os.path.join(private_dir, "keys", "farm")
    farm_crypto_key = _CryptoKeyProperty(farm_crypto_key_path)
    fatmouse_config_path = os.path.join(etc_dir, "agent.json")
    halted = _FileFlagProperty(os.path.join(private_dir, ".halt"))
    rebooted = _FileFlagProperty(os.path.join(private_dir, ".reboot"))
    updated = _FileFlagProperty(os.path.join(private_dir, ".update"))
    state = _StateFileProperty(os.path.join(private_dir, ".state"), default="unknown")
    pid_file = os.path.join(run_dir, "scalarizr.pid")
    pid = _PidFileProperty(pid_file)
    boot_id = _FileProperty(os.path.join(private_dir, "boot_id"))
    general = _SectionProperty("general")
    messaging = _SectionProperty("messaging")
    messaging_p2p = _SectionProperty("messaging_p2p")
    ec2 = _SectionProperty("ec2")
    openstack = _SectionProperty("openstack")
    ip_list_builder = _SectionProperty("ip_list_builder")
    base = types.BaseData()
    volumes = None
    apache = _SectionProperty("app")
    nginx = _SectionProperty("www")
    tomcat = _SectionProperty("tomcat")
    docker_authz_socket = "/var/run/scalr_docker_authz.sock"
    fatmouse_proc = None
    access_data = {}
    start_time = None
    # Saving these metrics into a file so they can survive a reboot during bootstrap
    bootstrap_metrics = _JsonFileProperty(
        os.path.join(private_dir, "bootstrap_metrics"),
        default={"boot_seconds": None, "boot_time": None},
    )

    @property
    def queryenv(self):
        # pylint: disable=no-name-in-module,import-error
        from scalarizr import queryenv

        kwargs = dict(url=self.general.queryenv_url, key_path=self.crypto_key_path, autoretry=True)
        if hub.general.server_id:
            kwargs.update(server_id=self.general.server_id)
        elif self.platform == "ec2":
            kwargs.update(
                farm_role_id=self.general.farm_role_id, cloud_instance_id=metadata.instance_id
            )
        else:
            raise exc.Error(
                "Unable to initialize QueryEnv client. The config.ini option "
                "general.server_id is empty. Most likely no Scalr Server Id "
                "was passed in User-Data to this VM"
            )
        return queryenv.QueryEnvService.create_latest(**kwargs)

    @property
    @mixutil.memorize()
    def scalr_url(self):
        pr = six.moves.urllib.parse.urlparse(hub.queryenv.url)
        # pylint: disable=too-many-function-args
        return six.moves.urllib.parse.urlunparse((pr.scheme, pr.netloc, "", "", "", ""))

    @property
    @mixutil.memorize()
    def scalr_logger(self):
        if self.base.logger.enabled:  # pylint: disable=no-member
            url = "{proto}://{path}:{port}".format(**hub.base.logger)
            auth = (
                (hub.base.logger.username, hub.base.logger.password)
                if hub.base.logger.username
                else None
            )
            return logutil.ScalrLogger(
                url=url,
                tag=hub.base.logger.tag,
                timeout=hub.base.logger.timeout,
                backend=hub.base.logger.backend,
                auth=auth,
                ssl_verify=hub.base.logger.ssl_verify,
            )

    @property
    @mixutil.memorize()
    def platform(self):
        # pylint: disable=no-name-in-module,import-error
        from scalarizr.platform import PlatformFactory

        return PlatformFactory().new_platform(hub.general.platform)

    @property
    def agent_version(self):
        # pylint: disable=no-member
        try:
            import scalarizr

            return scalarizr.__version__
        except ImportError:
            # Error happends only in fatmouse small / medium tests.
            return None

    @property
    def behaviors(self):  # TODO: rename current `__node__[behavior]` across codebase
        """
        This is a behaviors list, read from general/behaviour section of config.ini
        """
        try:
            rv = set(self.general.behaviour.strip().split(",")) - {
                "mysql",
                "mysql2",
                "percona",
                "mariadb",
                "postgresql",
                "redis",
                "rabbitmq",
                "mongodb",
            }
            return tuple(rv)
        except (six.moves.configparser.Error, AttributeError):
            return ()

    @property
    @mixutil.memorize(threadsafe=True)
    def db(self):
        # pylint: disable=no-name-in-module,import-error
        from scalarizr.util import sqlite_server

        def connect_db():
            conn = sqlite3.connect(hub.db_path, 5.0)
            conn.row_factory = sqlite3.Row
            conn.text_factory = sqlite3.OptimizedUnicode
            return conn

        if not os.path.exists(hub.db_path) or not os.stat(hub.db_path).st_size:
            LOG.debug("Creating SQLite database")
            conn = connect_db()
            try:
                with open(os.path.join(hub.share_dir, "db.sql")) as fp:
                    conn.executescript(fp.read())
                conn.commit()
            finally:
                conn.close()

        # Configure database connection pool
        LOG.debug("Initializing database connection")
        t = sqlite_server.SQLiteServerThread(connect_db)
        t.setDaemon(True)
        t.start()
        sqlite_server.wait_for_server_thread(t)
        return t.connection

    @property
    @mixutil.memorize()
    def share_dir(self):
        for d in (
            os.path.join(self.install_dir, "share"),
            "/usr/share/scalr",
            "/usr/local/share/scalr",
        ):
            if os.access(d, os.F_OK):
                return d
        return os.path.join(self.install_dir, "share")

    @property
    def log_config(self):
        config = r"""
[loggers]
keys=root,scalarizr,agent,common

[handlers]
keys=console,user_log,debug_log

[formatters]
keys=debug,user

[logger_root]
level=DEBUG
handlers=console,user_log,debug_log

[logger_scalarizr]
level=DEBUG
qualname=scalarizr
handlers=console,user_log,debug_log
propagate=0

[logger_agent]
level=DEBUG
qualname=agent
handlers=console,user_log,debug_log
propagate=0

[logger_common]
level=DEBUG
qualname=common
handlers=console,user_log,debug_log
propagate=0

[handler_console]
class=StreamHandler
level=ERROR
formatter=user
args=(sys.stderr,)

[handler_user_log]
class=common.utils.logutil.RotatingFileHandler
level=INFO
formatter=user
args=(r'LOG_PATH', 'a+', 5242880, 5, 0o600, 'utf-8')

[handler_debug_log]
class=common.utils.logutil.RotatingFileHandler
level=DEBUG
formatter=debug
args=(r'LOG_DEBUG_PATH', 'a+', 5242880, 5, 0o600, 'utf-8')

[formatter_debug]
format=%(asctime)s - %(levelname)s - %(name)s - %(message)s
class=common.utils.logutil.DebugFormatter

[formatter_user]
format=%(asctime)s - %(levelname)-5s - %(message)s
datefmt=%Y-%m-%d %H:%M:%S
class=common.utils.logutil.UserFormatter
"""
        config = config.replace("LOG_PATH", os.path.join(self.log_dir, "scalarizr.log"))
        config = config.replace("LOG_DEBUG_PATH", os.path.join(self.log_dir, "scalarizr_debug.log"))
        return config

    @property
    @mixutil.memorize()
    def updateclient(self):
        """UpdateClient Hub section."""

        class _RetClass(object):
            # TODO: move cache_dir into $lib_dir\\updateclient\\pkgmgr: think how
            # to handle postinst migration
            if fact["os"]["family"] == "windows":
                cache_dir = self.root_dir + "\\var\\cache\\scalr\\updclient\\pkgmgr"
            else:
                cache_dir = "/var/cache/scalr/updclient/pkgmgr"
            status_file = os.path.join(self.private_dir, "update.status")
            status = _JsonFileProperty(status_file, default={"state": "unknown"})
            pid_file = os.path.join(self.run_dir, "scalr-upd-client.pid")
            pid = _PidFileProperty(pid_file)
            service_name = "ScalrUpdClient" if sys.platform == "win32" else "scalr-upd-client"

            log_config = r"""
[loggers]
keys=root,scalarizr,agent,common

[handlers]
keys=console,debug_log

[formatters]
keys=debug

[logger_root]
level=DEBUG
handlers=console,debug_log

[logger_scalarizr]
level=DEBUG
qualname=scalarizr
handlers=console,debug_log
propagate=0

[logger_agent]
level=DEBUG
qualname=agent
handlers=console,debug_log
propagate=0

[logger_common]
level=DEBUG
qualname=common
handlers=console,debug_log
propagate=0

[handler_console]
class=StreamHandler
level=DEBUG
formatter=debug
args=(sys.stderr,)

[handler_debug_log]
class=common.utils.logutil.RotatingFileHandler
level=DEBUG
formatter=debug
args=(r'LOG_DEBUG_PATH', 'a+', 5242880, 5, 0o600, 'utf-8')

[formatter_debug]
format=%(asctime)s - %(levelname)s - %(name)s - %(message)s
class=common.utils.logutil.DebugFormatter
"""
            log_config = log_config.replace(
                "LOG_DEBUG_PATH", os.path.join(self.log_dir, "scalarizr_update.log")
            )

        return _RetClass()

    def __init__(self):
        super(_Hub, self).__init__()
        self._in_flushing_context = False
        self._cp = six.moves.configparser.ConfigParser()
        self._load()
        self.periodic_executor = agent_mixutil.PeriodicExecutor()

    def setup_from_commandline(self, args=None):
        """Setup from command-line arguments.

        When agent is launched with --import argv, this resets server state to 'importing' and
        initializes crypto_key, server_id, platform and queryenv_url from command line arguments
        """
        try:
            cnf = dict(i.strip().split("=", 1) for i in args.cnf)
            self.crypto_key = cnf["crypto-key"]
            self.general.env_id = cnf["env-id"]
            self.general.platform = cnf["platform"]
            self.general.queryenv_url = cnf["queryenv-url"]
            self.general.region = cnf["region"]
            self.general.role_name = cnf["role-name"]
            self.general.scalr_id = cnf["scalr-id"]
            self.general.server_id = cnf["server-id"]
            self.messaging_p2p.producer_url = cnf["messaging-p2p.producer-url"]
        except KeyError as e:
            raise exc.MalformedError(
                "Required option is missing: {}.\n"
                "If you're creating image/role from non-Scalr server, "
                "make sure that command correctly copied".format(e.args[0])
            )

    def setup_from_user_data(self, user_data):
        """Setup from user-data.

        When server_id in config.ini not match one from user-data, this resets server state
        to 'bootstrapping' and initializes crypto_key, server_id, platform and queryenv_url
        """
        g = lambda key: (user_data.get(key, "") or "").strip()
        # A minimum configuration required for startup.
        self.crypto_key = g("szr_key")
        self.general.farm_role_id = g("farm_roleid")
        self.general.platform = g("platform") or "ec2"
        self.general.queryenv_url = g("queryenv_url")
        self.general.server_id = g("serverid")
        self.general.scalr_version = g("scalr_version")

    def sync(self, flush=True):
        """Synchronize Hub attributes from QueryEnv and optionally flush INI changes to disk."""
        LOG.info("Synchronizing my configuration from Scalr")
        with self.flushing(dry_run=not flush):
            self.sync_global_config()
            self.sync_user_data()
            self.sync_farm_role_params()  # new farm_role_id from sync_user_data() is used.
            self.fire("sync")

    def sync_user_data(self):
        """This calls QueryEnv.get_server_user_data(), applies result data to Hub attributes, and
        flushes private.d/config.ini to disk.

        Most attributes comes to 'general' section, except 'p2p_producer_endpoint' and 'message_format' -
        they are applied to 'messaging_p2p' section and used by :class:`msg.Producer`
        and :class`msg.Consumer`
        """
        user_data = self.queryenv.get_server_user_data()
        user_data_safe_copy = types.remove_sensitive_data(user_data)
        LOG.debug("User-data (QueryEnv):\n%s", pprint.pformat(user_data_safe_copy))

        g = lambda key: (user_data.get(key, "") or "").strip()
        self.general.behaviour = g("behaviors")
        self.general.cloud_storage_path = g("cloud_storage_path")
        self.general.env_id = g("env_id")
        self.general.farm_id = g("farmid")
        self.general.farm_role_id = g("farm_roleid")
        self.general.owner_email = g("owner_email")
        self.general.platform = g("platform") or "ec2"
        self.general.queryenv_url = g("queryenv_url")
        self.general.region = g("region")
        self.general.role_id = g("roleid")
        self.general.role_name = g("realrolename")
        self.general.server_id = g("serverid")
        self.general.server_index = g("server_index")
        self.messaging_p2p.message_format = g("message_format") or "json"
        self.messaging_p2p.producer_url = g("p2p_producer_endpoint")

        self.flush()

    def sync_global_config(self):
        """This calls QueryEnv.get_global_config(), applies result data to Hub attributes, and
        flushes private.d/config.ini to disk
        """
        global_config = self.queryenv.get_global_config()["params"]
        self.general.scalr_id = global_config["scalr.id"]
        self.general.scalr_version = mixutil.Version(global_config["scalr.version"])
        self.flush()

    def sync_farm_role_params(self):
        """This calls QueryEnv.list_farm_role_params() and updates :attr:`base` with data"""
        farm_role_params = self.queryenv.list_farm_role_params(self.general.farm_role_id)["params"]
        params_base = farm_role_params["base"]

        # set vm id to private.d/config.ini/general/cloud_server_id
        gv = dict((gv["name"], gv["value"]) for gv in farm_role_params["global_variables"])
        self.general.cloud_server_id = gv.get("SCALR_CLOUD_SERVER_ID")

        # import roles data from queryenv
        self.base.import_data(params_base)
        self.volumes = farm_role_params.get("volumes")
        # feature toggles
        for k in params_base:
            if k.startswith("scalr_labs") and params_base[k]:
                sysutil.touch(os.path.join(hub.private_dir, k))

        # create update.status file for stable <= 4.12.x
        s = self.updateclient.status
        s["server_id"] = s["system_id"] = self.general.server_id
        s["system_id_is_server_id"] = True
        s["downgrades_enabled"] = True
        s["queryenv_url"] = self.general.queryenv_url
        s["farm_role_id"] = self.general.farm_role_id
        self.updateclient.status = s
        # agent.config overrides
        agent.config.workflow_engine["gc_keep_time"] = self.base.keep_scripting_logs_time

        self.flush()

    @contextlib.contextmanager
    def flushing(self, dry_run=False):
        """Context manages to flush INI settings on exit"""
        try:
            self._in_flushing_context = True
            yield
            self._in_flushing_context = False
            if not dry_run:
                self.flush()
        finally:
            self._in_flushing_context = False

    def flush(self):
        """Flushes INI settings to private.d/config.ini"""
        if self._in_flushing_context:
            return
        config_ini_path = os.path.join(self.private_dir, "config.ini")
        if os.path.exists(config_ini_path) and not os.access(config_ini_path, os.W_OK):
            os.chmod(config_ini_path, stat.S_IWRITE)
        with open(config_ini_path, "w+") as fp:
            self._cp.write(fp)

    def _load(self):
        filenames = glob.glob(os.path.join(self.public_dir, "*.ini"))
        filenames += glob.glob(os.path.join(self.private_dir, "*.ini"))
        fp = io.StringIO(
            six.u(
                textwrap.dedent(
                    """
              [general]
              crypto_key_path = private.d/keys/default

              [messaging]
              adapter = p2p

              [messaging_p2p]
              producer_retries_progression = 1,2,5,10,20,30,60
              consumer_url = http://0.0.0.0:8013

              [app]
              vhosts_path =
              apache_conf_path =

              [ec2]
              ec2_cert_path = public.d/keys/ec2_cert.pem
              cert_path = private.d/keys/ec2_cert.pem
              pk_path = private.d/keys/ec2_pk.pem
              hostname_as_pubdns = 1

              [ip_list_builder]
              base_path=$etc_path/private.d/hosts

              [memcached]
              cache_size = 1536

              [script_executor]
              exec_dir_prefix=/usr/local/bin/scalr-scripting.
              logs_dir=/var/log/scalarizr/scripting
              logs_truncate_over=20K

              [www]
              binary_path = /usr/sbin/nginx
              app_include_path = /etc/nginx/app-servers.include
              https_include_path = /etc/nginx/https.include
              app_port = 80

              """
                )
            )
        )
        self._cp.read_file(fp)
        self._cp.read(filenames)
        self._cp.remove_section("handlers")  # handlers no more configurable

    def get_configparser(self):
        return self._cp

    def get_handlers(self):
        """
        Return a list of handler import strings in a form
        ['scalarizr.handlers.gce.lifecycle', 'scalarizr.handlers.gce.lifecycle']
        """
        package = "scalarizr.handlers"

        # core modules
        modules = ["ip_list_builder", "ssh_auth_keys", "union.base"]

        # platform defining modules
        if hub.general.platform == "ec2":
            modules.extend(("ec2.ebs", "ec2.lifecycle", "ec2.rebundle"))
        if hub.general.platform == "azure":
            modules.extend(("azure.vol",))
        elif hub.general.platform == "gce":
            modules.extend(("gce.lifecycle", "gce.rebundle", "gce.vol"))
        elif hub.general.platform == "openstack":
            modules.extend(("openstack.rebundle", "openstack.vol"))
        elif hub.general.platform == "vmware":
            modules.append("vmware.vol")

        # behavior defining modules
        behavior_modules = [
            "apache",
            "haproxy",
            "memcached",
            "nginx",
            "router",
            "tomcat",
        ]
        for module in behavior_modules:
            if module in hub.behaviors:
                modules.append(module)
        # behavior aliases
        if "app" in hub.behaviors:
            modules.append("apache")
        if "www" in hub.behaviors:
            modules.append("nginx")

        return [".".join([package, module]) for module in modules]


hub: _Hub = mixutil.Proxy(_Hub)
