import json
import logging
import re

from agent.scalrinit import datadir
from agent.scalrinit import metadata
from common.utils import exc
from common.utils import mixutil

from . import exc as scalarizr_exc

LOG = logging.getLogger()
TIMEOUT = 20


class Error(Exception):
    pass


class Userdata(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setdefault("scalr_version", "0.0.0")  # added in Scalr 7.7.10

    @classmethod
    def from_string(cls, data):
        try:
            # openstack/configdrive legacy
            return Userdata(json.loads(data)["meta"])
        except ValueError:
            return Userdata(re.findall("([^=]+)=([^;]*);?", data))
        except KeyError:
            raise Error(
                "Invalid OpenStack/ConfigDrive locally cached JSON user-data: "
                "A 'meta' keyword with Scalr data is missing! "
                "This is very rarery possible when you've upgraded from Agent < 4.11.6."
                "Please delete /var/lib/scalarizr/cloud and restart scalr-upd-client service"
            )


@mixutil.memorize()
def instance_id():
    """Returns instance_id from cloud/data/instance-id

    Raises :func:`wait()` errors
    """
    wait(TIMEOUT)
    return datadir.instance_id()


@mixutil.memorize()
def user_data():
    """Parses raw/openstack format cloud/instance/user-data.txt into into k->v dict.

    Raises Error on empty user-data.
    Raises :func:`wait()` errors
    """
    wait(TIMEOUT)
    ud = datadir.user_data()
    if not ud:
        raise Error("No User-data found")
    return Userdata.from_string(ud)


@mixutil.memorize()
def wait(timeout=0):
    """Waits for user-data file in `cloud/instance/user-data.txt`

    Raises :class:`Error` on timeout (30 seconds)
    """
    try:
        datadir.result.wait(timeout)
    except exc.TimeoutError:
        raise Error(
            "UpdateClient(scalr-init) completion flag {} "
            "still not available after {} seconds.\n{}".format(
                datadir.result.link_path, timeout, scalarizr_exc.message_check_updateclient
            )
        )
    # Prevent one more metadata lookup in agent.scalrinit.metadata
    metadata.set_datasource(datadir.datasource())
