"""
Communication with Scalarizr.
"""
import binascii
import contextlib
import copy
import datetime
import enum
import hashlib
import hmac
import itertools
import json
import socket
import sqlite3
import sys
import threading
import time
import typing as tp
import urllib.parse

import requests
from requests import codes

import agent.config
from common.utils import cryptoutil
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import six

from . import types
from .hub import hub

LOG = logutil.get_logger(__name__)

# Messages rotation settings
ROTATE_KEEP_SIZE = 50

# HTTP status codes, don't retry on
# pylint: disable=E1101
SEND_FAIL_CODES = [codes.BANDWIDTH_LIMIT_EXCEEDED, codes.BAD_REQUEST, codes.FORBIDDEN]

# Message direction
INCOMING = 1
OUTGOING = 0

# Message status
HANDLED = 1
UNHANDLED = 0

# Message authorization
AUTH_V2_SCALR_VERSION = (7, 13, 0)

# Stores authorization version
# See: :meth:`get_auth_version` and :meth:`set_auth_version``
_auth_version = None  # type: tp.Optional[int]


@enum.unique
class AuthVersion(enum.IntEnum):
    """
    Supported authorization versions.
    """

    V1 = 1
    V2 = 2


class Error(exc.Error):
    """Exception for messaging module."""


class InvalidSignatureError(exc.Error):
    """When inbound message has invalid signature."""


def create(name, body=None, meta=None):
    """Create new message."""
    is_internal = name.startswith("Int")
    # see SCALARIZR-2437.
    #
    # Internal messages like IntServerReboot and IntServerHalt are firing when
    # network interfaces are terminating, and queries to get private/public ip
    # may not succeed.
    #
    # All below information actually is for Scalr and Scalarizr doesn't need
    # geting it from itself.

    body = body or {}
    if not is_internal:
        body["local_ip"] = hub.platform.get_private_ip()
        body["remote_ip"] = hub.platform.get_public_ip()
        body["behaviour"] = hub.behaviors
        body["role_name"] = hub.general.role_name
        body["server_id"] = hub.general.server_id
        if issubclass(Message.body_cls_for_name(name), types.StateChanging):
            body["base"] = hub.base.to_primitive()

    meta = meta or {}
    if not is_internal:
        meta["server_id"] = hub.general.server_id
        meta["szr_version"] = hub.agent_version

    return Message(name, body, meta)


def send(message, queue=None, endpoint=None, crypto_key=None, retry=True):
    """
    Send message.

    :rtype :class:`requests.Response`
    """
    LOG.debug("Sending message %s", message.name)
    queue = queue or "control"
    endpoint = endpoint or hub.messaging_p2p.producer_url
    crypto_key = crypto_key or hub.crypto_key
    if retry:
        progression = hub.messaging_p2p.producer_retries_progression.split(",")
        progression = list(six.moves.map(float, progression))
        retry_time_progression = itertools.chain(progression, itertools.repeat(progression[-1]))

        for retry_time in retry_time_progression:
            response = _try_send(message, queue, endpoint=endpoint, crypto_key=crypto_key)
            if response:
                return response
            elif (
                response is not None
                and response.status_code == codes.UNAVAILABLE
                and response.headers.get("retry-after")
            ):
                retry_time = int(response.headers.get("retry-after"))
                LOG.debug(
                    "Scalr returned 503 response code with 'Retry-After' header, "
                    "retry in %s seconds...",
                    retry_time,
                )
                time.sleep(retry_time)
                return send(
                    message, queue=queue, endpoint=endpoint, crypto_key=crypto_key, retry=retry
                )
            else:
                LOG.debug("Retry in %s seconds...", retry_time)
                time.sleep(retry_time)
    else:
        return _try_send(message, queue, endpoint=endpoint, crypto_key=crypto_key)


def send_internal(message, host, **kwargs):
    """Send internal message to another host."""
    endpoint = "http://{}:{}".format(host, hub.base.int_messaging_port)
    crypto_key = hub.farm_crypto_key
    return send(message, endpoint=endpoint, crypto_key=crypto_key, **kwargs)


def _try_send(message, queue, endpoint, crypto_key):
    """Function that actually does single send attempt.

    1. Encrypt & sign message
    2. Fire 'message_before_send' event
    3. Invoke POST /<queue> to endpoint with encrypted message as HTTP body

    :type message: :class:`Message`
    :type queue: str
    :type endpoint: str
    :type crypto_key: str
    """
    if message.name not in ("Log", "OperationResult", "RebundleLog"):
        log_msg = "Delivering message %s (%s, queue=%s)\n%s"
        LOG.debug(log_msg, message.name, message.id, queue, message)
    message.meta["timestamp"] = datetime.datetime.utcnow().strftime("%a %d %b %Y %H:%M:%S")
    endpoint_uri = "{}/{}".format(endpoint, queue)
    auth_version = get_auth_version()
    response = None

    def _do_post_request(auth_version=auth_version, fire_events=True):
        """
        Encrypts the message and makes a POST request to the Scalr messaging service.
        """
        data, headers = encrypt_message(
            message, crypto_key=crypto_key, auth_version=auth_version, endpoint=endpoint
        )
        if fire_events:
            hub.fire("message_before_send", queue, message)
        return requests.post(endpoint_uri, data=data, headers=headers, verify=False)

    try:
        response = _do_post_request()

        # try to send message via auth v1 if request failed with an error 401
        if (
            response.status_code in (codes.BAD_REQUEST, codes.UNAUTHORIZED)
            and auth_version == AuthVersion.V2
        ):
            try:
                response_v1 = _do_post_request(auth_version=AuthVersion.V1.value, fire_events=False)
                if response_v1.ok:
                    set_auth_version(AuthVersion.V1.value)
                    response = response_v1
                    LOG.warning(
                        "Messaging v2 authentication failed. " "Successfully downgraded to v1."
                    )
                else:
                    response_v1.raise_for_status()
            except Exception as err:
                LOG.warning(
                    "Messaging v2 authentication failed. " "Unable to downgrade to v1 (%s)",
                    str(err),
                )
                raise

        if response.status_code in SEND_FAIL_CODES:
            raise Error(
                ("Unable to send message, server responded with {}").format(response.status_code)
            )
        response.raise_for_status()

        LOG.info("Delivered %s message (%s)", message.name, message.id)
        _store(message, OUTGOING)
    except (requests.HTTPError, requests.ConnectionError, requests.RequestException) as err:
        LOG.warning(
            "Message '%s' not delivered (%s), cause: %s",
            message.name,
            message.id,
            str(err),
            exc_info=sys.exc_info(),
        )

    return response


def _store(message, direction):
    conn = hub.db
    cur = conn.cursor()
    try:
        sql = """
            INSERT INTO p2p_message
            (id, message, message_id, message_name, queue, is_ingoing, in_is_handled, in_consumer_id, format)
            VALUES (NULL, ?, ?, ?, NULL, ?, ?, NULL, ?)
            """
        query = [
            message.to_json(sanitized=True),
            message.id,
            message.name,
            direction,
            UNHANDLED,
            "json",
        ]
        cur.execute(sql, query)
        if "request_id" in message.meta:
            cur.execute(
                "UPDATE p2p_message SET response_uuid = ? WHERE message_id = ?",
                [message.id, message.meta["request_id"]],
            )
        conn.commit()
        LOG.debug("Dumped %s message (%s) into database", message.name, message.id)
        return True
    except sqlite3.IntegrityError:
        LOG.debug("Skipping %s message (%s), cause it's already exists", message.name, message.id)
        return False
    finally:
        cur.close()


def get_auth_version() -> int:
    """
    Returns the authorization version number for the messaging services.
    """
    if _auth_version is None:
        if hub.state == "importing":
            # there is no 'scalr_version' while importing server,
            # so use auth v2 as default
            return set_auth_version(AuthVersion.V2.value)

        scalr_version = hub.general.scalr_version
        if not scalr_version:
            return set_auth_version(AuthVersion.V2.value)

        if not isinstance(scalr_version, mixutil.Version):
            scalr_version = mixutil.Version(scalr_version)
        if scalr_version < AUTH_V2_SCALR_VERSION:
            return set_auth_version(AuthVersion.V1.value)
        return set_auth_version(AuthVersion.V2.value)

    return _auth_version


def set_auth_version(version: int):
    """
    Updates the authorization version number for the messaging services.
    """
    global _auth_version
    _auth_version = version
    return _auth_version


def prepare_encryption_headers(
    payload: str,
    auth_headers: dict,
    crypto_key: tp.Optional[str] = None,
    auth_version: tp.Optional[int] = None,
) -> dict:
    """
    Returns encryption headers for the message.

    Headers:
     - ``Content-Type``: application/json
     - ``X-Scalr-Agent-Version``: x.y.z version of the agent
     - ``X-Scalr-AuthVersion``: 1 (for scalr<7.10.0) or 2 (for scalr>=7.10.0)
     - ``X-Signature``: SHA-1 signature for auth v1 (optional)
     - ``X-Signature-Sha256``: SHA-256 signature for auth v2
     - ``Date``: signing date

    :param payload: The data to sign.
    :type payload: str | dict
    :param auth_headers: Headers to authorize the request, acceptable names:
        `'X-Server-Id` or `X-Scalr-Farm-Role-Id` and `X-Scalr-Cloud-Server-Id`.
    :type auth_headers: dict
    :param crypto_key: The cryptographic key (optional).
    :type crypto_key: str | None
    :param auth_version: The authorization version (optional).
    :type auth_version: int | None

    :return: The encryption headers.
    :rtype: dict
    """
    if auth_version is None:
        auth_version = get_auth_version()

    if auth_headers and set(auth_headers.keys()) not in (
        {"X-Server-Id"},
        {"X-Scalr-Farm-Role-Id", "X-Scalr-Cloud-Server-Id"},
    ):
        raise ValueError("Invalid auth headers")

    headers = {
        "Content-Type": "application/json",
    }

    # create authorization signatures for the message body
    timestamp = time.gmtime()
    crypto_key_bin = binascii.a2b_base64(crypto_key or hub.crypto_key)
    signature_v1, _ = (
        cryptoutil.sign_http_request(
            payload, crypto_key_bin, timestamp=timestamp, date_fmt=cryptoutil.RFC_3339_UTC_DATE_FMT,
        )
        if auth_version == AuthVersion.V1
        else (None, None)
    )
    signature_v2, sign_date = cryptoutil.sign_http_request(
        payload,
        crypto_key_bin,
        headers=auth_headers,
        timestamp=timestamp,
        version=2,
        digestmod=hashlib.sha256,
        date_fmt=cryptoutil.RFC_3339_UTC_DATE_FMT,
    )

    # setting message headers
    headers["X-Scalr-Agent-Version"] = hub.agent_version
    headers["X-Scalr-AuthVersion"] = str(auth_version)
    if signature_v1:
        headers["X-Signature"] = signature_v1
    headers["X-Signature-Sha256"] = signature_v2
    headers["Date"] = sign_date

    headers.update(auth_headers)
    return headers


def _encrypt_data(data, crypto_key, auth_version=None):
    crypto_key = binascii.a2b_base64(crypto_key)
    if auth_version is None or auth_version == AuthVersion.V1:
        return cryptoutil.encrypt_scalarizr(data, crypto_key)
    elif auth_version == AuthVersion.V2:
        return cryptoutil.encrypt_scalr(data, crypto_key)
    else:
        raise Error("Unknown auth version: {}".format(auth_version))


def _decrypt_data(data, crypto_key, auth_version=None):
    crypto_key = binascii.a2b_base64(crypto_key)
    if auth_version is None or auth_version == AuthVersion.V1:
        return cryptoutil.decrypt_scalarizr(data, crypto_key)
    elif auth_version == AuthVersion.V2:
        return cryptoutil.decrypt_scalr(data, crypto_key).decode()
    else:
        raise Error("Unknown auth version: {}".format(auth_version))


def encrypt_message(
    message,
    crypto_key: tp.Optional[str] = None,
    auth_version: tp.Optional[int] = None,
    endpoint: tp.Optional[str] = None,
) -> tp.Tuple[bytes, dict]:
    """
    Encrypt the outcoming message.

    :param message: The message to encrypt.
    :type message: :class:`Message`
    :param crypto_key: The cryptographic key (optional).
    :type crypto_key: str | None
    :param auth_version: The authorization version (optional).
    :type auth_version: int | None

    :return: Returns encrypted message with encryption/authorization headers.
    :rtype: (bytes, dict)
    """
    if auth_version is None:
        auth_version = get_auth_version()
    if crypto_key is None:
        crypto_key = hub.crypto_key
    auth_headers = {"X-Server-Id": hub.general.server_id}

    if endpoint:
        # do not include auth headers into signature for the local or internal messaging
        parse_result = urllib.parse.urlparse(endpoint)
        if parse_result.hostname == "0.0.0.0" or parse_result.port == hub.base.int_messaging_port:
            auth_headers = {}

    LOG.debug("Encrypting message (auth_version=%d)...", auth_version)
    encrypted_message = _encrypt_data(
        message.to_json().encode("utf-8"), crypto_key, auth_version=auth_version
    )

    LOG.debug("Setting headers (auth_version=%d)", auth_version)
    headers = prepare_encryption_headers(
        encrypted_message,
        auth_headers=auth_headers,
        crypto_key=crypto_key,
        auth_version=auth_version,
    )

    return encrypted_message, headers


def decrypt_message(payload: str, headers: dict, crypto_key: tp.Optional[str] = None):
    """
    Decrypt the incoming message.

    :param payload: The data to decrypt.
    :type payload: str
    :param headers: The `POST` request headers.
    :type headers: dict
    :param crypto_key: The cryptographic key (optional).
    :type crypto_key: str | None

    :rtype: :class:`Message`
    """
    auth_version = str(headers.get("X-Scalr-AuthVersion", ""))
    auth_version = int(auth_version) if auth_version.isdigit() else AuthVersion.V1.value
    crypto_key = crypto_key or hub.crypto_key

    LOG.debug("Decrypting message (auth_version=%d)...", auth_version)
    if auth_version == AuthVersion.V2:
        signature = headers.get("X-Signature-Sha256")
        sign_date = headers.get("Date")
        if not signature:
            raise Error("Empty 'X-Signature-Sha256' header")
        if not sign_date:
            raise Error("Empty 'Date' header")

        messaging_date_fmt = cryptoutil.RFC_3339_UTC_DATE_FMT
        expected_signature, _ = cryptoutil.sign_http_request(
            payload,
            binascii.a2b_base64(crypto_key),
            timestamp=time.strptime(sign_date, messaging_date_fmt),
            version=2,
            digestmod=hashlib.sha256,
            date_fmt=messaging_date_fmt,
        )
        if not hmac.compare_digest(signature, expected_signature):
            raise InvalidSignatureError("Signature doesn't match.")

    try:
        message = _decrypt_data(payload, crypto_key, auth_version=auth_version)
    except:
        LOG.debug("Decryption error", exc_info=sys.exc_info())
        LOG.debug("Crypto key: %s", crypto_key)
        LOG.debug("Encrypted message: %s", payload)
        raise Error("Can't decrypt message (details in debug log)")

    if auth_version > get_auth_version():
        set_auth_version(auth_version)

    return message


class Message(object):
    """This is a Scalr / Scalarizr communication message.
    Both ends could produce and consume messages.
    Message has name, metadata and body.
    Messages passed on a wire encoded into JSON and encrypted.
    """

    # pylint: disable=redefined-builtin
    def __init__(self, name=None, body=None, meta=None, id=None, **kwargs):
        """Creates object.

        :param name: Message name
        :type name: str

        :param body: Message body object
        :type body: dict or :class:`types.MessageBody` subclass

        :param meta: Metadata
        :type meta: dict

        :param id: Message ID
        :type id: uuid string
        """
        assert name is not None, "Message name cannot be None"
        self.name = name
        if isinstance(body, types.MessageBody):
            self.body = body
        else:
            # ensure there will be no serialization variance errors later
            body_data = json.loads(json.dumps(mixutil.map_recursive(body or {}, self.binary2utf8)))
            body_cls = Message.body_cls_for_name(name)
            self.body = body_cls(body_data, strict=False)
        self.meta = meta or {}
        self.id = id or str(mixutil.sortable_uuid())  # pylint: disable=invalid-name

    def __eq__(self, other):
        return self.__dict__ == other

    def __getattr__(self, key):
        return getattr(self.body, key, None)

    def __setattr__(self, key, value):
        if key in ("id", "name", "meta"):
            object.__setattr__(self, key, value)
        elif key == "body":
            if isinstance(value, types.MessageBody):
                object.__setattr__(self, key, value)
            else:
                raise exc.Error("The message body must be {}".format(types.MessageBody))
        else:
            self.body[key] = value

    def __contains__(self, key):
        return key in self.body

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        """Return logging-safe representation"""
        return self.to_json(sanitized=True, indent=4)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.ack()

    @staticmethod
    def body_cls_for_name(name):
        try:
            return getattr(types, name)
        except AttributeError:
            return types.DefaultMessageBody

    @staticmethod
    def sanitize(message):
        message = copy.deepcopy(message)
        message["body"] = types.remove_sensitive_data(message["body"])
        return message

    def ack(self):
        """Marks message as handled."""
        conn = hub.db
        cur = conn.cursor()

        try:
            sql = """
                UPDATE p2p_message
                SET in_is_handled = ?, message = ?, out_last_attempt_time = datetime("now")
                WHERE message_id = ? AND is_ingoing = ?"""
            cur.execute(sql, [HANDLED, self.to_json(sanitized=True), self.id, INCOMING])
            conn.commit()
        finally:
            cur.close()
            LOG.debug("Marked %s message (%s) done", self.name, self.id)

    @staticmethod
    def binary2utf8(value):
        """Encodes binary value to UTF-8."""
        if isinstance(value, six.binary_type):
            if six.PY2:
                return value.encode("utf-8", errors="replace")
            return value.decode("utf-8", errors="replace")
        else:
            return value

    @staticmethod
    def from_json(value):
        """Creates message object from JSON data."""
        return Message(**json.loads(value))

    def to_json(self, sanitized=False, indent=None):
        """Serializes message object to a JSON formatted string."""
        return json.dumps(self.to_primitive(sanitized=sanitized), indent=indent)

    def to_primitive(self, sanitized=False):
        """Sanitizes message."""
        sanitize_fn = self.sanitize if sanitized else lambda x: x
        return sanitize_fn(
            {"id": self.id, "name": self.name, "body": self.body.to_primitive(), "meta": self.meta,}
        )


class _RequestHandler(six.moves.BaseHTTPServer.BaseHTTPRequestHandler):
    """
    A RequestHandler class for BaseHTTPServer(s).

    We use BaseHTTPServer(s) listening:
     - 8013 port to receive ingoing messages from Scalr
     - 8012 port (optional) to receive internal messages from another agent
    Only POST method is implemented.
    """

    timeout = 10
    consumer = None

    def log_message(self, format, *args):  # pylint: disable=redefined-builtin
        LOG.debug(format, *args)

    def decode_message(self, msg_str):
        """Decode message."""
        LOG.debug("Decoding message...")
        mime_type = self.headers.get("Content-Type", "application/json")
        if "application/json" not in mime_type:
            raise Error("Not acceptable Content-Type: {}".format(mime_type))
        try:
            # log what we actually received, but escape known secrets
            data = json.loads(msg_str)
            msg = Message(**data)
            LOG.debug("Raw message is:\n%s", json.dumps(Message.sanitize(data), indent=4))
        except:
            LOG.debug("Decoding error", exc_info=sys.exc_info())
            LOG.debug("Encoded message: %s", msg_str)
            raise Error("Can't decode message (details in debug log)")
        return msg

    # pylint: disable=invalid-name
    def do_POST(self):
        """
        Process incoming payload
        """
        try:
            c = self.consumer
            LOG.debug(
                "POST (server_address: *:%s running: %s, waiting: %s)",
                self.server.server_address[1],
                c.running,
                c.waiting,
            )
            if c.running or c.waiting:
                try:
                    is_internal = hub.base.messaging_port != self.server.server_address[1]
                    payload = self.rfile.read(int(self.headers["Content-length"]))
                    message = self.decode_message(
                        decrypt_message(
                            payload,
                            headers=self.headers,
                            crypto_key=hub.farm_crypto_key if is_internal else hub.crypto_key,
                        )
                    )
                except Exception as e:
                    LOG.exception("Consumer can't decrypt/decode message: %s", str(e))
                    response = codes.BAD_REQUEST  # pylint: disable=E1101
                else:
                    LOG.info("Received %s message (%s)", message.name, message.id)
                    # Always receive IntBlockDeviceUpdated cause that message can be
                    # rejected on bootstrap. See: FAM-1406
                    if c.waiting and not isinstance(message.body, types.IntBlockDeviceUpdated):
                        if c.receive_waiting_for(message):
                            # FAM-639: prevent double handling
                            _store(message, INCOMING)
                            response = codes.CREATED  # pylint: disable=E1101
                        else:
                            LOG.info(
                                (
                                    "Rejecting %s message (%s), cause it's unexpected. "
                                    "Scalr will re-deliver message in a minute"
                                ),
                                message.name,
                                message.id,
                            )
                            response = codes.SERVICE_UNAVAILABLE  # pylint: disable=E1101
                    elif c.running:
                        c.put(message)
                        response = codes.CREATED  # pylint: disable=E1101
            else:
                LOG.debug("Consumer loop is not running, rejecting message")
                response = codes.SERVICE_UNAVAILABLE
        except Exception:
            LOG.exception("Internal error")
            response = codes.INTERNAL_SERVER_ERROR

        self.send_response(response)
        self.end_headers()
        return


class Consumer(object):
    """This object is a message consumer on Agent side.
    It implements HTTP server, message decoder and an ingoing queue
    that should be connected to :class:`Handler`
    """

    def __init__(self):
        self._http_servers = []
        self._request_handler_cls = _RequestHandler
        self._request_handler_cls.consumer = self
        self._running = False

        # primitives for queue lookup
        self._message = None  # slot for awaited message
        self._criteria = None  # criteria function(msg)
        self._wait_for_done = threading.Event()
        self._waiting = False

        self.queue = six.moves.queue.Queue()
        hub.periodic_executor.add_task(
            self.rotate, agent.config.workflow_engine["gc_interval"], "Rotate stored messages"
        )

    def start(self):
        """Start HTTP server(s)"""
        self._load_unhandled()
        if self._running:
            raise exc.DuplicateError("Message consumer already running")
        self._running = True

        self._run_threaded_server("message-consumer", hub.base.messaging_port)

    def _run_threaded_server(self, name, port):
        LOG.info("Starting message consumer server on *:%s", port)
        http_server = six.moves.BaseHTTPServer.HTTPServer(
            ("0.0.0.0", port), self._request_handler_cls
        )
        thread = threading.Thread(target=http_server.serve_forever, name=name)
        thread.daemon = True
        thread.start()
        self._http_servers.append(http_server)

    def shutdown(self):
        """Shutdown HTTP server(s)"""
        self._running = False
        LOG.debug("Message consumer will shutdown.")
        for http_server in self._http_servers:
            try:
                http_server.shutdown()
                http_server.socket.shutdown(socket.SHUT_RDWR)
                http_server.socket.close()
            except Exception as e:
                LOG.debug("Error shutting down message consumer: %s", six.text_type(e))

    @contextlib.contextmanager
    def suspending(self):
        """Suspend HTTP server(s)"""
        try:
            self._running = False
            LOG.debug("Consumer suspended")
            yield
        finally:
            LOG.debug("Consumer resumed")
            self._running = True

    @property
    def running(self):
        """Property that checks if server is running."""
        return self._running

    @property
    def waiting(self):
        """Property that checks if server is waiting for specific message."""
        return self._waiting

    def put(self, message):
        """Puts message in the queue."""
        if _store(message, INCOMING):
            self.queue.put(message)

    def receive_waiting_for(self, message):
        assert self._criteria
        # we're now receving specific message
        LOG.debug("Checking that %s matches wait criteria ...", message.name)
        if self._criteria(message):
            LOG.debug("Alright, %s matches wait criteria", message.name)
            self._message = message
            self._waiting = False
            self._wait_for_done.set()  # signal .wait_for()
            return True

    def wait_for(self, name):
        """Wait for incoming message, matching criterias. Blocking.

        The criterias are:
        - message.name is equal to `name`
        - body.server_id is equal to server_id in config.ini

        :param name: Message name
        :type name: str
        """
        LOG.info("Waiting for %s message ...", name)
        try:

            def criteria(message):
                return message.name == name and message.server_id == hub.general.server_id

            self._criteria = criteria
            self._waiting = True

            # Lookup message in local queue
            LOG.debug("Looking up for %s message in local queue", name)
            for (
                m
            ) in (
                self.queue.queue
            ):  # Queue stores it's items in `self.queue: collections.deque` attribute.
                if self.receive_waiting_for(m):
                    self.queue.queue.remove(m)
                    break
            else:
                LOG.debug("Nothing has found in local queue, will check every incoming message")

            # Passing timeout to make Event.wait() non-blocking. timeout must never expire.
            self._wait_for_done.wait()
            LOG.debug("wait_for() got message %s", self._message.name)
            return self._message
        finally:
            self._criteria = None
            self._message = None
            self._wait_for_done.clear()

    def _load_unhandled(self):
        LOG.debug("Loading unhandled messages from database into consumer queue.")
        conn = hub.db
        cur = conn.cursor()
        try:
            cur.execute(
                "SELECT message FROM p2p_message WHERE in_is_handled = ? AND is_ingoing = ?",
                (UNHANDLED, INCOMING),
            )
            unhandled_msgs = cur.fetchall()
            restart_explained = False

            for m in unhandled_msgs:
                message = Message.from_json(m[0])

                for script in message.body.get("scripts") or []:
                    if not restart_explained:
                        LOG.warning(
                            "Scalarizr has been restarted while it had scripts in the queue. "
                            "It is not possible to requeue them because they had been sanitized."
                        )
                        restart_explained = True
                    LOG.warning(
                        "  Script %s from the message %s is skipped.", script["name"], message.name
                    )

                if "scripts" in message.body:
                    message.body["scripts"] = []
                self.queue.put(message)
            LOG.debug("%d messages loaded.", len(unhandled_msgs))
        finally:
            cur.close()

    @staticmethod
    def rotate():
        """
        Deletes unnecessary messages.
        """
        conn = hub.db
        cur = conn.cursor()
        try:
            cur.execute(
                "SELECT * FROM p2p_message ORDER BY id DESC " "LIMIT {}, 1".format(ROTATE_KEEP_SIZE)
            )
            row = cur.fetchone()
            if row:
                LOG.debug("Deleting messages older then message_id: %s", row["message_id"])
                cur.execute("DELETE FROM p2p_message WHERE id <= ?", (row["id"],))
        finally:
            cur.close()
        conn.commit()


class Handler(object):
    """
    This object is a message event handler.
    """

    def __init__(self):
        self._running = False
        self._resumed = threading.Event()
        self._resumed.set()
        self._callback = None

    @contextlib.contextmanager
    def suspending(self):
        """Context manager that suspends event handler."""
        try:
            self.suspend()
            yield
        finally:
            self.resume()

    def suspend(self):
        """Suspend event handler."""
        LOG.debug("Suspending event handler")
        self._resumed.clear()

    def resume(self):
        """Resume event handler."""
        LOG.debug("Resuming event handler")
        self._resumed.set()

    @property
    def running(self):
        """Property that checks if handler is running."""
        return self._running

    def serve_forever(self):
        """Start event handler loop."""
        if self._running:
            return
        LOG.debug("Starting handler loop")
        self._running = True
        q = consumer.queue
        while True:
            self._resumed.wait()  # suspend/resume
            try:
                msg = q.get_nowait()
            except six.moves.queue.Empty:
                # XXX: https://bugs.python.org/issue26958
                # Queue.not_empty is private in fact, and has limited application.
                with q.not_empty:
                    q.not_empty.wait()
                continue
            else:
                with msg:
                    self.handle(msg)

    def handle(self, message):
        """Handle message."""
        if not self._callback:
            return
        try:
            return self._callback(message)
        except Exception as e:
            LOG.exception("Message handling error: %s: %s", e.__class__.__name__, str(e))

    def set_callback(self, callback):
        """
        Sets callback for message handler.
        """
        self._callback = callback


# pylint: disable=invalid-name
handler = mixutil.Proxy(Handler)
consumer = mixutil.Proxy(Consumer)
