"""
.. module:: postinst
   :platform: Linux
   :synopsis: Post install action's

.. moduleauthor:: Scalr Inc. <http://scalr.com>

"""
import os

import agent.config
from agent.scalarizr.hub import hub
from common.utils import logutil
from common.utils import sysutil
from common.utils.facts import fact

fact.includes.append("agent.tasks.sys")
fact.load()

LOG = logutil.get_logger(__name__)

# We mount /dev because of https://github.com/docker/docker/issues/16160
DOCKER_SERVICE_TEMPLATE = """
[Unit]
Description={service_name}
Documentation=https://github.com/Scalr/scalarizr
After=docker.service docker.socket network-online.target local-fs.target
Requires=docker.service docker.socket network-online.target local-fs.target

[Service]
Restart=on-failure
RestartSec=30
RestartPreventExitStatus=5
TimeoutStartSec=120
KillMode=none
ExecStartPre=-/usr/bin/docker kill {service_name}
ExecStartPre=-/usr/bin/docker rm {service_name}
ExecStartPre=-/usr/bin/docker network create \\
                --driver=bridge \\
                -o com.docker.network.bridge.enable_ip_masquerade=true \\
                -o com.docker.network.bridge.enable_icc=true \\
                --subnet=172.19.5.0/16 \\
                --ip-range=172.19.5.0/24 \\
                --gateway=172.19.5.1 \\
                scalr-int

ExecStart=/usr/bin/docker run --name {service_name} \\
                --privileged \\
                --net=scalr-int \\
                --entrypoint={entrypoint} \\
                --volume=/etc/scalr:/etc/scalr \\
                --volume=/var/lib/scalarizr:/var/lib/scalarizr \\
                --volume=/var/log:/var/log \\
                --volume=/var/run/scalarizr:/var/run/scalarizr \\
                --volume=/dev:/dev \\
                --volume=/:/rootfs \\
                -e FAM_USER_DATA_PROVIDER='personality' \\
                {ports} \\
                {image}

ExecStop={exec_stop}
Environment=LANG=en_US.UTF-8

[Install]
WantedBy=multi-user.target
"""

STOP_ACTION_TEMPLATE = """#!/usr/bin/env bash
systemctl list-units --type target &> {targets_info_path}
docker stop {service_name}
"""

DOCKER_CLI_TEMPLATE = """
#!/usr/bin/env bash

AGENT_RUN=$(docker ps --filter "name={scalarizr_container}" | grep {scalarizr_container})
if [ -z "$AGENT_RUN" ]
then
      docker run \\
        --rm \\
        --privileged \\
        --net=scalr-int \\
        --entrypoint={entrypoint} \\
        --volume=/etc/scalr:/etc/scalr \\
        --volume=/var/lib/scalarizr:/var/lib/scalarizr \\
        --volume=/var/log:/var/log \\
        --volume=/var/run/scalarizr:/var/run/scalarizr \\
        --volume=/dev:/dev \\
        --volume=/:/rootfs \\
        {ports} \\
        {image} $@
else
      docker exec {scalarizr_container} {entrypoint} $@
fi
"""


class DockerPostInst(object):
    """
    Implement post install action in case when scalarizr inside container
    """

    ROOTFS = "/rootfs"
    SERVICE_FILE_PATH = os.path.join(ROOTFS, "etc/systemd/system")
    SCALARIZR_PRIVATE_DIR = os.path.join(ROOTFS, hub.private_dir.strip("/"))
    SCALARIZR_BINARY_PATH = "/opt/scalarizr/bin"
    AGENT_NAME = "scalarizr"
    AGENT_PORTS = ["8010", "8013"]
    UPDATER_NAME = "scalr-upd-client"
    UPDATER_PORTS = ["8008"]
    IMAGE = "scalr/scalarizr"
    SCALARIZR_CLI = "scalarizr"
    SZRADM_CLI = "szradm"
    CLI_BINARY_PATH = os.path.join(ROOTFS, "opt/bin")
    SCALARIZR_STOP_SCRIPT = "/opt/bin/_scalarizr-coreos-stop.sh"

    def __init__(self):
        self.agent_service_file_name = self.AGENT_NAME + ".service"
        self.updater_service_file_name = self.UPDATER_NAME + ".service"
        self.agent_ports_list = " ".join(["-p " + p + ":" + p for p in self.AGENT_PORTS])
        self.updater_ports_list = " ".join(["-p " + p + ":" + p for p in self.UPDATER_PORTS])
        self.scalarizr_cli_path = os.path.join(self.CLI_BINARY_PATH, self.SCALARIZR_CLI)
        self.szradm_cli_path = os.path.join(self.CLI_BINARY_PATH, self.SZRADM_CLI)
        self.agent_service_file_path = os.path.join(
            self.SERVICE_FILE_PATH, self.agent_service_file_name
        )
        self.updater_service_file_path = os.path.join(
            self.SERVICE_FILE_PATH, self.updater_service_file_name
        )
        self.updater_stop_cmd = "/usr/bin/docker stop %s" % self.UPDATER_NAME
        self.targets_info_path = os.path.join(
            agent.config.workflow_engine["run_dir"], "scalarizr.systemctl.list-units.type.target"
        )

    @staticmethod
    def _generate_service_file(service_name, image, entrypoint, ports, exec_stop):
        return DOCKER_SERVICE_TEMPLATE.format(
            image=image,
            service_name=service_name,
            entrypoint=entrypoint,
            ports=ports,
            exec_stop=exec_stop,
        )

    @staticmethod
    def _generate_cli_file(image, entrypoint, ports, scalarizr_container):
        return DOCKER_CLI_TEMPLATE.format(
            image=image, entrypoint=entrypoint, ports=ports, scalarizr_container=scalarizr_container
        )

    @staticmethod
    def _generate_stop_file(targets_info_path, service_name):
        return STOP_ACTION_TEMPLATE.format(
            targets_info_path=targets_info_path, service_name=service_name
        )

    @staticmethod
    def _write_file(filename, data, mod=0o644):
        if os.path.exists(filename):
            LOG.info("File %s exist.", filename)
            LOG.info("Update %s", filename)
        else:
            LOG.info("Create %s", filename)

        with open(filename, "w") as f:
            f.write(data)

        os.chmod(filename, mod)

    def _write_executable_file(self, filename, data):
        self._write_file(filename, data, mod=0o755)

    def _append_user_path(self):
        sysutil.mkdir_p("/rootfs/etc/bash/bashrc.d")
        self._write_file("/rootfs/etc/bash/bashrc.d/scalr-profile.sh", 'PATH="$PATH:/opt/bin"')

    def run(self):
        """
        Entry point for DockerPostInst class
        """
        updater_service_data = self._generate_service_file(
            self.UPDATER_NAME,
            self.IMAGE,
            self.UPDATER_NAME,
            self.updater_ports_list,
            self.updater_stop_cmd,
        )

        agent_service_data = self._generate_service_file(
            self.AGENT_NAME,
            self.IMAGE,
            self.AGENT_NAME,
            self.agent_ports_list,
            self.SCALARIZR_STOP_SCRIPT,
        )

        scalarizr_cli_data = self._generate_cli_file(
            self.IMAGE, self.SCALARIZR_CLI, self.agent_ports_list, self.AGENT_NAME
        )

        szradm_cli_data = self._generate_cli_file(
            self.IMAGE, self.SZRADM_CLI, self.agent_ports_list, self.AGENT_NAME
        )

        stop_file_data = self._generate_stop_file(self.targets_info_path, self.AGENT_NAME)

        sysutil.mkdir_p(self.CLI_BINARY_PATH, mode=0o755)
        self._write_executable_file(self.scalarizr_cli_path, scalarizr_cli_data)
        self._write_executable_file(self.szradm_cli_path, szradm_cli_data)
        self._write_executable_file(self.ROOTFS + self.SCALARIZR_STOP_SCRIPT, stop_file_data)
        self._write_file(self.updater_service_file_path, updater_service_data)
        self._write_file(self.agent_service_file_path, agent_service_data)
        self._append_user_path()
        sysutil.mkdir_p(self.SCALARIZR_PRIVATE_DIR)
        # Set updated flag.
        sysutil.touch(os.path.join(self.SCALARIZR_PRIVATE_DIR, ".update"))
        # Set CoreV2.
        sysutil.touch(os.path.join(self.SCALARIZR_PRIVATE_DIR, "scalr_labs_corev2"))


def run():
    """
    Execute post install actions
    triggered by scalarizr --install
    """
    if fact["manage_container_host"]:
        DockerPostInst().run()
