"""RPC implementation for scalarizr and update client.
"""
import collections
import json
import sys
import traceback

from common.utils import exc
from common.utils import imputil
from common.utils import logutil
from common.utils import rpcutil
from common.utils import six

from . import types
from .hub import hub

LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    """General rpc error."""


def service_method(func):
    """Marks service RPC method.

    :decorator:
    """
    # pylint: disable=protected-access
    func._jsonrpc = True
    return func


def query_method(func):
    """Marks query RPC method.

    :decorator:
    """
    # pylint: disable=protected-access
    func._jsonrpc = "query"
    return func


def command_method(func):
    """Marks command RPC method.

    :decorator:
    """
    # pylint: disable=protected-access
    func._jsonrpc = "command"
    return func


Request = collections.namedtuple("Request", ["id", "method", "params"])


class RequestHandler(rpcutil.ABCRequestHandler):
    """RPC api request handler. Used in conjunction with common.utils.rpcutil."""

    def __init__(self, services):
        if not isinstance(services, collections.Mapping):
            services = {None: services}
        self.services = services

    def handle_request(self, data, namespace=None):
        """RPC request handling logic.
        Finds target API service and method in given namespace and then executes the method.

        :param str data: Request data
        :param str namespace: Services namespace
        """
        error = None
        exc_info = (None, None, None)
        result = ""
        log_it = False
        try:
            request = self._parse_request(data)
            self._handle_meta_params(request)
            service = self._find_service(namespace)
            rpc_method = self._find_method(service, request.method)

            # pylint: disable=protected-access
            log_it = rpc_method._jsonrpc == "command"
            if log_it:
                LOG.debug("Request: %s", self._clear_request_data(request))
            result = self._invoke_method(rpc_method, request.params)
            # Test json serializarion before serializing the whole result.
            json.dumps(result)

        except rpcutil.ServiceError as e:
            error = e.as_dict()

        except Exception as e:
            error = "{}: {}".format(e.__class__.__name__, e)
            exc_info = sys.exc_info()

        finally:
            if error:
                if isinstance(error, six.string_types):
                    error_message = error

                    if isinstance(exc_info[1], (KeyError, IndexError)):
                        # file/line/def where exception occurred formatted like exception stacktrace
                        where = traceback.format_list([traceback.extract_tb(exc_info[2])[-1]])[
                            0
                        ].strip()
                        error_message = "{} in {}".format(error_message, where)

                    #  SCALARIZR-2344
                    exc_info = False if isinstance(exc_info[1], exc.NotSupportedError) else exc_info

                    LOG.debug("Caught API exception. %s", error_message, exc_info=exc_info)

                    error = {
                        "code": rpcutil.ServiceError.INTERNAL,
                        "message": error_message,
                        "data": None,
                    }

                response = {"error": error}
            else:
                response = {"result": result}
            response["id"] = request.id
        response_json = json.dumps(response)
        if log_it:
            LOG.debug("Response: %s", response_json)
        return response_json.encode("utf-8")

    @staticmethod
    def _clear_request_data(request):
        request_copy = request._asdict()
        request_copy["params"] = types.remove_sensitive_data(request_copy["params"])
        return Request(**request_copy)

    @staticmethod
    def _parse_request(data):
        try:
            data = json.loads(data)
        except Exception as e:
            raise rpcutil.ParseError(six.text_type(e))

        if data.get("params", None) is None:
            raise rpcutil.InvalidRequestError("Request doesn't have the params attribute")
        return Request(data["id"], data["method"], data["params"])

    @staticmethod
    def _handle_meta_params(request):
        if "_platform_access_data" in request.params:
            hub.platform.set_access_data(request.params["_platform_access_data"])
            del request.params["_platform_access_data"]

    def _find_service(self, namespace):
        try:
            service = self.services[namespace]
        except KeyError:
            six.reraise(
                rpcutil.NamespaceNotFoundError,
                rpcutil.NamespaceNotFoundError(namespace),
                sys.exc_info()[2],
            )
        else:
            if isinstance(service, six.string_types):
                service = self.services[namespace] = imputil.import_object(service)
        return service

    @staticmethod
    def _find_method(service, name):
        method = getattr(service, name, None)
        if method and getattr(method, "_jsonrpc", None):
            return method
        else:
            raise rpcutil.MethodNotFoundError(name)

    @staticmethod
    def _invoke_method(method, params):
        return method(**params)
