import datetime
import sys
import time

import celery

import agent.config
import common.types
from agent.tasks import sys as sys_tasks
from common.utils import logutil
from common.utils import mixutil

from .hub import hub

LOG = logutil.get_logger(__name__)


def get_server_metadata():
    """Returns server metadata collected from Hub."""
    return {
        "farm_id": hub.general.farm_id,
        "farm_role_id": hub.general.farm_role_id,
        "server_id": hub.general.server_id,
        "server_index": hub.general.server_index,
        "agent_version": hub.agent_version,
        "env_id": hub.general.env_id,
        "cloud_location": hub.general.region,
        "cloud": hub.general.platform,
    }


def _push_agent_stat(stat_data: common.types.AgentStatEvent) -> None:
    """
    Push ``agent-stat`` to the Scalr.
    """
    with celery.current_app.producer_pool.acquire(block=True) as prod:
        prod.connection.ensure_connection(max_retries=1)
        event_dispatcher = celery.current_app.events.Dispatcher(
            connection=prod.connection,
            hostname=None,
            enabled=True,
            channel=prod.channel,
            buffer_while_offline=False,
        )
        event_dispatcher.send("agent-stat", **stat_data.to_primitive())
        event_dispatcher.close()


class StatCollector:
    """
    Statistics collector on Agent side.
    """

    def __init__(self) -> None:
        self.push_stat = mixutil.retries_error(
            Exception,
            interval=int(agent.config.stat["push_interval"] * 0.15),  # type: ignore
            times=2,
            silent=True,
        )(_push_agent_stat)
        self._connection_lost_at = None
        self._pushes_lost = 0
        self._last_error_repr = None

    def start(self):
        """Add self to application periodical tasks."""
        LOG.debug("Collecting server metrics: %s", agent.config.get_stat_metrics())
        hub.periodic_executor.add_task(
            self, agent.config.stat["push_interval"], "Push server statistics"
        )

    def shutdown(self):
        """Remove self from the application periodical tasks."""
        hub.periodic_executor.remove_task(self)

    def __call__(self):
        """Collect statistics and push as message (celery control) to Scalr."""
        server_meta = get_server_metadata()
        stat_data = common.types.AgentStatEvent(server_meta)
        metrics = agent.config.get_stat_metrics()
        stat_data.facts = sys_tasks.get_facts.run(metrics, raise_on_fact_error=False)

        try:
            # send message to scalr.
            self.push_stat(stat_data)
            # log connection repair
            if self._connection_lost_at:
                LOG.warning(
                    "Metrics collector connection to %s was repaired, "
                    "%s statistics push(es) lost since %s",
                    celery.current_app.connection().as_uri(),
                    self._pushes_lost,
                    self._connection_lost_at.strftime("%Y-%m-%d %H:%M:%S"),
                )
                self._reset_error_counters()
        except Exception as e:
            # log first error, and unique subsequent errors.
            if not self._connection_lost_at:
                # logging goes in local time.
                self._connection_lost_at = datetime.datetime.now()

            if not self._last_error_repr or self._last_error_repr != repr(e):
                LOG.warning(
                    "Metrics collector connection to %s is lost: %s",
                    celery.current_app.connection().as_uri(),
                    e,
                    exc_info=sys.exc_info(),
                )

            self._last_error_repr = repr(e)
            self._pushes_lost += 1

    def _reset_error_counters(self):
        self._connection_lost_at = None
        self._pushes_lost = 0
        self._last_error_repr = None


def push_bootstrap_metrics():
    """Sends agent bootstrap metrics to Scalr server.

    The metics are:
        - boot_seconds: uptime in seconds at the time of the first start of Scalarizr
        - configure_seconds: seconds between the first start of Scalarizr
            and sending of the HostUp message (including possible reboots)
    """
    try:
        metadata = get_server_metadata()
        agent_stat = common.types.AgentStatEvent(metadata)
        boot_seconds = hub.bootstrap_metrics["boot_seconds"]
        configure_seconds = time.time() - hub.bootstrap_metrics["boot_time"]
        agent_stat.facts = {
            "stat.bootstrap": {
                "boot_seconds": float("{0:.2f}".format(boot_seconds)),
                "configure_seconds": float("{0:.2f}".format(configure_seconds)),
            }
        }
        _push_agent_stat(agent_stat)
    except Exception as e:
        LOG.warning("Error pushing bootstrap metrics to Scalr: %s", e, exc_info=sys.exc_info())


# pylint: disable=invalid-name
collector = mixutil.Proxy(StatCollector)
