"""Core scalarizr types and constants.

Scalr-Scalarizr messaging protocol is defined here.
"""
from __future__ import absolute_import

import copy
import re

# pylint: disable=import-error
from schematics import models
from schematics.types import BaseType
from schematics.types import BooleanType
from schematics.types import FloatType
from schematics.types import IntType
from schematics.types import StringType
from schematics.types.compound import DictType
from schematics.types.compound import ListType
from schematics.types.compound import ModelType
from six.moves import Mapping

import common.types
from common.utils import six
from common.utils.facts import fact

# pylint: enable=import-error


class DictCompatMixin(object):
    """Scalarizr handlers expect message.body and its items to be dicts,
    so we add dict compatibility methods which are not implemented by Schematics models.
    """

    def copy(self):
        """Implements dict.copy."""
        return {key: value for key, value in self._data.items() if value is not None}

    def pop(self, key, default=None):
        """Implements dict.pop."""
        if key in self._data:
            try:
                return self._data.pop(key, default)
            finally:
                self._data[key] = None
        return default

    def __contains__(self, item):
        return item in self._data and self._data[item] is not None


class Model(DictCompatMixin, models.Model):
    """Base class for all our models."""

    def __init__(self, *args, **kwargs):
        kwargs["strict"] = False
        super(Model, self).__init__(*args, **kwargs)

    class Options:
        """Schematics options."""

        serialize_when_none = False


class LooseMixin(object):
    """Fallback to defaults when values are missing."""

    def to_native(self, value, context=None):
        if value == "" and not isinstance(self, StringType):
            return self.default
        elif value is False and not isinstance(self, BooleanType):
            return self.default
        else:
            return super(LooseMixin, self).to_native(value, context)


class LooseIntType(LooseMixin, IntType):
    """IntType with fallback to default."""


class LooseBooleanType(LooseMixin, BooleanType):
    """BooleanType with fallback to default."""


class LooseStringType(LooseMixin, StringType):
    """StringType with fallback to default."""


class LooseListType(LooseMixin, ListType):
    """ListType with fallback to default."""


class LooseDictType(LooseMixin, DictType):
    """DictType with fallback to default."""


class FatmouseData(Model):
    """Fatmouse part of BaseData."""

    user_data = LooseStringType()


class UpdateData(Model):
    """UpdateClient part of BaseData."""

    server_url = LooseStringType()
    api_port = LooseIntType(default=8008)
    repository = LooseStringType()
    deb_repo_url = LooseStringType()
    rpm_repo_url = LooseStringType()
    suse_repo_url = LooseStringType()
    win_repo_url = LooseStringType()
    docker_repo_url = LooseStringType()
    ssl_verify_peer = LooseBooleanType(default=True)


# Scalar Governance policies
class DockerPoliciesNames(common.types.Enum):
    allow_privileged_create = "docker.container.privileged.usage"
    allowed_host_ports = "docker.network.ports"
    allowed_mounts = "docker.volume.mounts"
    image_policies = "docker.image.sources"
    allow_privileged_exec = "docker.container.privileged.escalating_exec_usage"


class GovernanceRulesData(Model):
    type = StringType()
    value = ListType(DictType(BaseType))


class GovernancePolicyValue(Model):
    conditions = DictType(StringType)
    rules = ListType(GovernanceRulesData)


class GovernancePolicyData(Model):
    type = StringType()
    value = ModelType(GovernancePolicyValue)


class CreateContainerDockerPolicy(Model):
    allow_privileged_create = BooleanType()
    allowed_host_ports = ListType(StringType)
    """
    TODO: allowed_host_ports = StringType()
    """
    allowed_mounts = ListType(StringType)


class ExecContainerDockerPolicy(Model):
    limit_escalating = BooleanType()


class CreateImageDockerPolicy(Model):
    allowed_images = ListType(StringType)
    allowed_sources = ListType(StringType)


class DockerPolicies(Model):
    create_container_policy = ModelType(CreateContainerDockerPolicy)
    create_image_policy = ModelType(CreateImageDockerPolicy)
    exec_container_policy = ModelType(ExecContainerDockerPolicy)


class InstallData(Model):
    """Install software.

    see scalarizr.union.base
    """

    behaviors = LooseListType(StringType, required=True, default=[])


class AnsibleTower(Model):
    """Class for encapsulating ansible related data"""

    if fact["os"]["name"] == "windows":

        class _MachineCredential(Model):
            username = StringType(required=True)
            """Machine Credentials username"""

            password = StringType(required=True)
            """Machine Credentials password"""

    else:

        class _MachineCredential(Model):
            username = StringType(required=True)
            """Machine Credentials username"""

            ssh_public_key = StringType(required=True)
            """SSH Public Key"""

            become_method = StringType(choices=("sudo",))
            """Privilege escalation mode"""

    machine_credentials = ListType(ModelType(_MachineCredential))
    """List of credential settings"""


class ScalrLogger(Model):
    """Scalr Log part of BaseData."""

    proto = LooseStringType(default="http")
    path = LooseStringType()
    port = LooseIntType()
    tag = LooseStringType(default="agent")
    backend = LooseStringType()
    username = LooseStringType()
    password = LooseStringType()
    ssl_verify = BooleanType(default=True)
    enabled = BooleanType(default=False)
    timeout = LooseIntType(default=1)


class BaseData(Model):
    """Basic server configuration that comes from Scalr."""

    abort_init_on_script_fail = LooseBooleanType(default=False)
    ansible = ModelType(AnsibleTower, strict=False)
    api_port = LooseIntType(default=8010)
    disable_firewall_management = LooseBooleanType(default=False)
    disable_hostname_management = LooseBooleanType(default=False)
    disable_ntp_management = LooseBooleanType(default=False)
    domain_name = LooseStringType()
    fatmouse = ModelType(FatmouseData, strict=False)
    hostname = LooseStringType()
    install = ModelType(InstallData, strict=False)
    int_messaging_port = LooseIntType(default=8012)
    keep_scripting_logs_time = LooseIntType(default=86400)
    logger = ModelType(ScalrLogger, strict=False, default=ScalrLogger)
    messaging_port = LooseIntType(default=8013)
    policies = LooseDictType(ListType(StringType))
    reboot_after_hostinit_phase = LooseBooleanType(default=False)
    scalr_labs_corev2 = LooseBooleanType(default=False)
    scalr_labs_rabbitmq_tunnel = LooseBooleanType(default=False)
    scalr_labs_fatmouse_update = LooseBooleanType(default=False)
    tags = LooseDictType(StringType)
    union_script_executor = LooseBooleanType(default=False)
    update = ModelType(UpdateData, strict=False, required=True, default=UpdateData)


class ChefData(Model):
    """Chef config."""

    client_rb_template = LooseStringType()
    cookbook_url = LooseStringType()  # For chef-solo
    cookbook_url_type = LooseStringType(choices=("git", "http"))  # For chef-solo
    daemonize = LooseBooleanType(default=False)
    environment = LooseStringType(default=None)
    event_name = LooseStringType(default="HostInit")
    execution_id = LooseStringType()
    first_bootstrap = LooseBooleanType()
    json_attributes = LooseStringType()  # JSON encoded attributes
    log_level = LooseStringType(default="auto")
    node_name = LooseStringType()
    relative_path = LooseStringType()  # For chef-solo
    role = LooseStringType()  # Deprecated. Chef role (use run_list instead)
    run_list = LooseStringType()  # JSON encoded run_list
    script_name = LooseStringType(default="[Scalr built-in] Chef bootstrap")
    server_url = LooseStringType()
    solo_rb_template = LooseStringType()
    ssh_private_key = LooseStringType()  # For chef-solo
    ssl_verify_mode = LooseStringType()
    validator_key = LooseStringType()
    validator_name = LooseStringType()


class HaproxyData(Model):
    """Haproxy config."""

    proxies = LooseListType(DictType(BaseType))


class NginxData(Model):
    """Haproxy config."""

    proxies = LooseListType(DictType(BaseType))
    preset = LooseListType(BaseType)


class ApacheData(Model):
    """
    Apache config.
    """

    virtual_hosts = LooseListType(BaseType)
    preset = LooseListType(BaseType)


class RouterData(Model):
    """
    Router config.
    """

    cidr = LooseStringType()
    scalr_addr = LooseStringType()
    whitelist = LooseListType(StringType)


class MessageBody(Model):
    """Minimal message body."""

    event_id = LooseStringType()  # All orchestrable events must have it


class DefaultMessageBody(MessageBody):
    """Default message body, most messages contain these fields."""

    behaviour = LooseListType(StringType)
    cloud_location = LooseStringType()
    crypto_key = LooseStringType()
    farm_crypto_key = LooseStringType()
    farm_role_alias = LooseStringType()
    farm_role_id = LooseStringType()
    params = LooseDictType(StringType)
    event_name = LooseStringType()
    global_variables = LooseListType(BaseType)
    local_ip = LooseStringType()
    platform_access_data = BaseType()
    remote_ip = LooseStringType()
    role_name = LooseStringType()
    scripts = LooseListType(BaseType)
    server_id = LooseStringType()
    server_index = LooseIntType(default=0)  # Rebundle comes with ''
    db_type = LooseStringType()


class StateChanging(DefaultMessageBody):
    """Messages that change server state."""

    chef = ModelType(ChefData, strict=False)
    haproxy = ModelType(HaproxyData, strict=False)
    apache = ModelType(ApacheData, strict=False)
    www = ModelType(NginxData, strict=False)
    base = ModelType(BaseData, strict=False, default=BaseData)
    router = ModelType(RouterData, strict=False)
    volumes = LooseListType(BaseType)


class Hello(StateChanging):
    """Agent sends this message on server import."""

    os = BaseType()
    dist = BaseType()
    architecture = LooseStringType()  # obsolete


class HostDown(StateChanging):
    """Message fired on server shutdown."""


class HostInit(StateChanging):
    """Message that starts server initialization."""

    seconds_since_start = BaseType()
    seconds_since_boot = BaseType()


class HostInitResponse(StateChanging):
    """Message with data needed for server initialization."""


class BeforeHostUp(StateChanging):
    """Message used as a trigger for scripts and code that
    needs to be run at the end of initialization.
    """


class HostUp(StateChanging):
    """Message that concludes server initialization."""


class ScalarizrVersion(Model):
    """Agent version."""

    version = LooseStringType()


class HostUpdate(StateChanging):
    """Agent to Scalr message to notify about changes in server state."""

    scalarizr = ModelType(ScalarizrVersion)


class ConvertVolume(MessageBody):
    volume = LooseDictType(BaseType)


class ConvertVolumeResult(MessageBody):
    status = LooseStringType()
    last_error = LooseStringType()
    volume = LooseDictType(BaseType)


class OperationResult(MessageBody):
    """OperationAPI result. See int-scalarizr/src/scalarizr/api/operation.py
    Also used to mark server as failed if an error occurs during initialization.
    """

    id = BaseType(required=True)
    name = LooseStringType(required=True)
    status = LooseStringType(required=True)
    result = BaseType()
    error = LooseStringType()


class RebootStart(DefaultMessageBody):
    """Notifies about reboot start."""


class RebootFinish(StateChanging):
    """Notifies about reboot finish."""


class Rebundle(DefaultMessageBody):
    """Scalr to agent message to initiate rebundle."""

    bundle_task_id = BaseType(required=True)
    excludes = LooseListType(StringType)  # in EC2 bundle, Scalr never uses
    root_volume_template = BaseType()  # EC2 rootfs volume, new way
    volume_size = LooseIntType()  # EC2 rootfs volume size, old way


class RebundleLog(MessageBody):
    """Agent to Scalr log messages during rebundle."""

    bundle_task_id = BaseType(required=True)
    message = LooseStringType()


class RuntimeError(MessageBody):  # pylint: disable=redefined-builtin
    """Agent to Scalr message with runtime error."""

    code = LooseStringType()
    message = LooseStringType(required=True)


class RebundleResult(DefaultMessageBody):
    """Agent to Scalr message with rebundle result."""

    bundle_task_id = BaseType(required=True)
    dist = BaseType()
    os = BaseType()
    snapshot_id = LooseStringType(required=True)
    status = LooseStringType(choices=("ok", "error"), required=True)
    last_error = LooseStringType()


class UpdateControlPorts(DefaultMessageBody):
    """Agent sends this message to Scalr when it adjusts
    its ports to non-default values.
    """

    api = LooseIntType()
    messaging = LooseIntType()


class UpdateSshAuthorizedKeys(DefaultMessageBody):
    add = LooseListType(StringType)
    remove = LooseListType(StringType)


class BlockDeviceAttached(DefaultMessageBody):
    """Agent to Scalr message to notify about device attached."""

    device_name = LooseStringType()


class BlockDeviceDetached(DefaultMessageBody):
    """Agent to Scalr message to notify about device detached."""

    device_name = LooseStringType()


class BlockDeviceMounted(DefaultMessageBody):
    """Agent to Scalr message to notify about device mounted."""

    device_name = LooseStringType()
    volume_id = LooseStringType()
    mountpoint = LooseStringType()


class IntBlockDeviceUpdated(DefaultMessageBody):
    """Internal message which is sent by scripts/udev.py."""

    action = LooseStringType()
    devname = LooseStringType()


class ExecScript(DefaultMessageBody):
    """Scalr to agent message to initiate script execution."""

    event_name = LooseStringType()


class ExecScriptResult(MessageBody):
    """Agent to Scalr message with script execution result."""

    event_name = LooseStringType()
    event_server_id = LooseStringType()
    execution_id = LooseStringType()
    return_code = LooseIntType()
    run_as = LooseStringType()
    script_name = LooseStringType()
    script_path = LooseStringType()
    script_id = LooseIntType()
    script_version = LooseIntType()
    stderr = LooseStringType()
    stderr_log_path = LooseStringType()
    stdout = LooseStringType()
    stdout_log_path = LooseStringType()
    time_elapsed = FloatType()
    timeout = LooseIntType()


class Behavior(common.types.Enum):
    """List of possible behaviours."""

    apache = "app"
    chef = "chef"
    haproxy = "haproxy"
    memcached = "memcached"
    mongodb = "mongodb"  # tombstone
    mysql = "mysql"  # tombstone
    mysql2 = "mysql2"  # tombstone
    mariadb = "mariadb"  # tombstone
    percona = "percona"  # tombstone
    nginx = "www"
    postgresql = "postgresql"  # tombstone
    rabbitmq = "rabbitmq"  # tombstone
    redis = "redis"  # tombstone
    tomcat = "tomcat"


def remove_sensitive_data(data):
    """
    Creates a copy of message body and removes all sensitive data from it including
    passwords, keys and private global variables.
    Can also be used for filtering similar structures returned by queryenv client.
    :param dict msg_dict: message body
    :return: dict
    """
    if not data or not isinstance(data, Mapping):
        return data

    msg_dict_copy = copy.deepcopy(data)
    mask = "******"

    # filtering upper-level data
    if "platform_access_data" in msg_dict_copy:
        if isinstance(msg_dict_copy["platform_access_data"], Mapping):
            for k in msg_dict_copy["platform_access_data"]:
                msg_dict_copy["platform_access_data"][k] = mask
        else:
            msg_dict_copy["platform_access_data"] = mask

    if "scripts" in msg_dict_copy and isinstance(msg_dict_copy["scripts"], list):
        for script in msg_dict_copy["scripts"]:
            if isinstance(script, Mapping) and "body" in script:
                script["body"] = mask
            if "repo" in script:
                if "password" in script["repo"]:
                    script["repo"]["password"] = mask
                if "ssh_key" in script["repo"]:
                    script["repo"]["ssh_key"] = mask
            if "ansibletower" in script:
                if "password" in script["ansibletower"]:
                    script["ansibletower"]["password"] = mask

    if "global_variables" in msg_dict_copy and isinstance(msg_dict_copy["global_variables"], list):
        for i, gv in enumerate(msg_dict_copy["global_variables"]):
            if gv.get("private") or gv.get("hidden"):
                msg_dict_copy["global_variables"][i]["value"] = mask

    for key in ("farm_crypto_key", "crypto_key", "szr_key", "pma_password"):
        if key in msg_dict_copy:
            msg_dict_copy[key] = mask

    if "base" in msg_dict_copy and "fatmouse" in msg_dict_copy["base"]:
        if "user_data" in msg_dict_copy["base"]["fatmouse"]:
            user_data = msg_dict_copy["base"]["fatmouse"]["user_data"]
            pattern = r"(amqp:.+:)(.+)(@.+:\d+)"
            msg_dict_copy["base"]["fatmouse"]["user_data"] = re.sub(
                pattern, r"\1{}\3".format(mask), user_data
            )

    if "base" in msg_dict_copy and "logger" in msg_dict_copy["base"]:
        if "password" in msg_dict_copy["base"]["logger"]:
            msg_dict_copy["base"]["logger"]["password"] = mask

    # filtering software-specific data
    excludes = {
        "chef": ["validator_name", "validator_key"],
        # Sanitize removed behaviors just in case scalr-server will send any of their messages
        "rabbitmq": ["password"],
        "redis": ["master_password", "passwords"],
        "postgresql": ["root_ssh_public_key", "root_ssh_private_key", "root_password"],
        "mysql2": ["stat_password", "repl_password", "root_password"],
        "mariadb": ["stat_password", "repl_password", "root_password"],
        "percona": ["stat_password", "repl_password", "root_password"],
    }
    for service, sensitive_keys in excludes.items():
        if service in msg_dict_copy:
            for key in sensitive_keys:
                if key in (msg_dict_copy.get(service) or {}):
                    if isinstance(msg_dict_copy[service][key], six.string_types):
                        msg_dict_copy[service][key] = mask
                    elif isinstance(msg_dict_copy[service][key], list):
                        msg_dict_copy[service][key] = [mask,] * len(msg_dict_copy[service][key])
    return msg_dict_copy


# Constants

# Default and fallback ports of scalarizr services.
SCALARIZR_PORTS = (("api_port", 8010, 8009), ("messaging_port", 8013, 8011))
