"""Updateclient entry point.
"""
# isort:skip_file
import argparse
import os
import sys
import threading
import time
import wsgiref.simple_server

import psutil

import agent.config
import agent.scalrinit.app as scalrinit_app
import common.types
from agent.scalarizr import base  # pylint: disable=locally-disabled,wrong-import-order
from agent.tasks import service as service_tasks
from agent.tasks import sys as sys_tasks
from agent.tasks import pkgmgr as pkgmgr_tasks
from common.tasks import update as update_tasks
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import pkgutil
from common.utils import rpcutil
from common.utils import six
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

from .. import config
from . import compat
from . import metadata
from . import rpc
from .hub import hub
from .utils import apputil

if sys.platform == "win32":
    linuxutil = None
else:
    from common.utils import linuxutil
if sys.platform == "win32":
    # pylint: disable=import-error
    import servicemanager
    import win32api
    import win32event
    import win32service
    import win32serviceutil
else:
    servicemanager = None
    win32api = None
    win32event = None
    win32service = None
    win32serviceutil = None


LOG = logutil.get_logger("agent.scalarizr.updateclient")


class Error(Exception):
    """UpdateClient error."""


def remove_py2_artifacts():
    """
    Removes PY2 artifacts that are cannot be removed in postinstall sequence.
    """
    if sys.platform == "win32":
        # This link is created in postinstall sequence for PY2 to be able
        # to finish update on embedded PY3 successfully. We need to remove it as soon as we can.
        sysutil.unlink(os.path.join(hub.install_dir, "embedded", "lib", "encodings"))


@mixutil.error_logged("Error in main")
def main():
    """ "UpdateClient main funciton."""
    # PY2 artifacts removal
    remove_py2_artifacts()

    # Low-level initialization.
    prepare()
    commandline_handle()

    # Daemon startup.
    LOG.info("[pid: %s] Starting UpdateClient %s (corev2)", os.getpid(), hub.agent_version)
    base.container_host_add_ssh_authorized_key()
    apputil.boot_wait()

    if hub.updateclient.status["state"] not in ("error", "rollbacked"):
        _set_state("completed")

    # We do new server probe after each reboot. This might look excessive,
    # but there is no other way to reliably detect first start after rebundle.
    if hub.boot_id != sysutil.get_boot_id():
        LOG.debug(
            "Server boot-id changed! Re-scaning Meta-data to "
            "figure out whether it's a new server bootstrap"
        )
        scalrinit_app.main_start()
        user_data = metadata.user_data()
        server_id = user_data.get("serverid")

        # If `server_id` is empty, it means that we bootstrapping a new server.
        # see: SCALARIZR-2533
        if not server_id or hub.general.server_id != server_id:
            apputil.wait_vmware_customization()
            _set_state("noop")
            LOG.info(
                "Scalr server-id changed! " "Bootstrapping new server (%s, previous=%s)",
                server_id,
                hub.general.server_id,
            )
            if hub.state != "discovering":
                service_tasks.stop("scalarizr")
                service_tasks.enable(hub.updateclient.service_name)
            hub.setup_from_user_data(user_data)
            if hub.state != "discovering":
                api.update(bootstrap=True)
            else:
                _set_state("completed")

    # Save service version in status file.
    s = hub.updateclient.status
    s["service_version"] = hub.agent_version
    hub.updateclient.status = s

    sync()
    start()

    # try to uninstall outdated scalarizr-{platform} package and migrate cache
    compat.handle_platform_package_deprecation()

    _ensure_agent_running()
    if sys.platform != "win32":
        # ... if program is still running, entering linux daemon main loop
        apputil.daemon_main(shutdown)


def sync():
    """Sync configuration from Scalr."""
    try:
        hub.sync(flush=False)
    except Exception as e:
        err = (
            "Communication with Scalr is lost: {}. "
            "Check that crypto key in {} matches scalarizr.key property "
            "on Server Dashboard"
        ).format(e, hub.crypto_key_path)
        raise Error(err)
    public_global_variables = (
        common.types.GlobalVariable({"name": name, "value": value})
        for (name, value) in hub.queryenv.list_global_variables()["public"].items()
    )
    sys_tasks.apply_proxy_global_variables(public_global_variables)

    if fact["os"]["family"] != "windows":
        channel = api.update_channel
        repo = channel.repos[0]

        pkgmgr_tasks.repository(
            name=channel["name"],
            url=repo["url"],
            ssl_verify=hub.base.update.ssl_verify_peer,
            prefix="scalr-agent",
        )


def prepare():
    """Low-level application initialization."""
    if fact["os"]["family"] != "windows" and os.getegid():
        logutil.die('UpdateClient must run under "root"')

    logutil.file_config(six.moves.cStringIO(hub.updateclient.log_config))

    from agent.app import app

    app.conf.task_always_eager = True


def start():
    """Starts API server (1 daemon thread)."""

    class ThreadingWSGIServer(
        six.moves.socketserver.ThreadingMixIn, wsgiref.simple_server.WSGIServer
    ):
        """ThreadingWSGIServer."""

    app = rpcutil.Server(rpc.RequestHandler(api), hub.crypto_key_path)
    api_server = wsgiref.simple_server.make_server(
        "0.0.0.0", hub.base.update.api_port, app, server_class=ThreadingWSGIServer
    )

    LOG.info("Starting API on *:%s", hub.base.update.api_port)
    t = threading.Thread(target=api_server.serve_forever, name="api-server")
    t.daemon = True
    t.start()


def shutdown():
    """Shutdown updclient."""
    del hub.updateclient.pid
    LOG.info("[pid: %s] UpdateClient terminated", os.getpid())


def commandline_parse():
    """Parse commandline arguments.

    :rtype: tuple
    :returns: A tuple with two elements: first is parsed known args, second is all unknown ones.
        Second is always None on Linux and is only relevant for Windows.
    """
    argv = list(sys.argv)
    p = argparse.ArgumentParser(description="Scalr agent auto-update service")
    p.add_argument("-d", "--daemonize", action="store_true", help="daemonize process")
    p.add_argument(
        "-P", "--pid-file", default=hub.updateclient.pid_file, help="file to store PID in"
    )

    # deprecated and removed options
    p.add_argument(
        "--get-system-id",
        action="store_true",
        default=False,
        help="print system-id and exit [removed]",
    )
    p.add_argument(
        "--make-status-file",
        action="store_true",
        default=False,
        help="make status file with current state and exit [removed]",
    )
    p.add_argument(
        "--downgrades-disabled",
        action="store_true",
        default=False,
        help="works only with --make-status-file "
        "(introduced for migration to new update system) [removed]",
    )
    p.add_argument(
        "-l",
        "--log-file",
        default=os.path.join(hub.log_dir, "scalarizr_update.log"),
        help="log file [removed]",
    )
    p.add_argument(
        "-v", "--verbose", action="store_true", default=True, help="verbose logging [removed]"
    )
    p.add_argument("-r", "--set-repository", help="[option removed]")
    if fact["os"]["name"] == "windows":
        return p.parse_known_args(argv[1:])
    return (p.parse_args(argv[1:]), None)


def commandline_handle():

    """Handle basic command line arguments"""
    args, unknown_args = commandline_parse()

    if fact["os"]["name"] == "windows" and unknown_args:
        unknown_args.insert(0, sys.argv[0])
        win32serviceutil.HandleCommandLine(WindowsService, argv=unknown_args)
        raise SystemExit(0)

    if args.set_repository:
        print("-r|--set-repository removed - updates are now controlled by Scalr")
        raise SystemExit()
    elif args.get_system_id:
        print("--get-system-id removed")
        raise SystemExit(0)
    elif args.make_status_file:
        print(
            "--make-status-file removed - migration to new update "
            "system completed between v2.7.0 and v4.7.14"
        )
        raise SystemExit(0)
    elif args.downgrades_disabled:
        print(
            "--downgrades-disabled removed - migration to new update "
            "system completed between v2.7.0 and v4.7.14"
        )
        raise SystemExit(0)

    if fact["os"]["name"] != "windows":
        if args.daemonize:
            linuxutil.daemonize()
        _make_pid_file()


def _make_pid_file():
    if sys.platform != "win32":
        other_pid = hub.updateclient.pid
        this_pid = os.getpid()
        if other_pid and other_pid != this_pid:
            try:
                other = psutil.Process(other_pid)
                if other.name() == hub.updateclient.service_name:
                    raise exc.DuplicateError(
                        "Other UpdateClient process already running: {}".format(other_pid)
                    )
            except psutil.NoSuchProcess:
                pass
        hub.updateclient.pid = this_pid


class API(object):
    """Updateclient API methods that can be called via HTTP RPC."""

    def __init__(self):
        self._update_lock = threading.Lock()

    @property
    def update_channel(self):
        scalarizr_package_name = config.workflow_engine["package_name"]
        ud = hub.base.update
        repo_family_map = {
            "windows": ("win", ud.win_repo_url),
            "debian": ("deb", ud.deb_repo_url),
            "redhat": ("rpm", ud.rpm_repo_url),
            "suse": ("suse", ud.suse_repo_url),
            "docker": ("docker", ud.docker_repo_url),
        }

        repo = common.types.Repository()
        repo.packages = []
        if fact["manage_container_host"]:
            repo_type, repo_url = repo_family_map["docker"]
            scalarizr_package_name = ud.docker_repo_url.split("/")[-1]
        else:
            repo_type, repo_url = repo_family_map[fact["os"]["family"]]
            if fact["os"]["family"] != "windows":
                repo.packages.append(
                    common.types.Package(
                        {
                            "name": "{0}-{1}".format(scalarizr_package_name, hub.general.platform),
                            "deprecated_since_version": compat.SINGLE_PACKAGE_FOR_LINUX_VER,
                        }
                    )
                )

        repo.repo_type = repo_type
        repo.url = repo_url
        repo.packages.append(common.types.Package({"name": scalarizr_package_name}))
        channel = common.types.UpdateChannel()
        channel.name = ud.repository
        channel.repos = [repo]
        return channel

    @staticmethod
    def _cleanup_rpmdb():
        """Cleanup RPMDb from old entries. See SCALARIZR-1797."""
        if fact["os"]["family"] in ("redhat", "suse"):
            out = subprocess2.run(
                ("rpm", "-qa", "scalarizr*", "--queryformat", "%{NAME}|%{VERSION}\n")
            ).stdout
            for line in out.strip().split("\n"):
                if "|" in line:  # https://bugzilla.redhat.com/show_bug.cgi?id=918184
                    name, version = line.strip().split("|")
                    if not version == hub.agent_version:
                        subprocess2.run(
                            ("rpm", "-e", "--nodeps", "--justdb", "{0}-{1}".format(name, version))
                        )

    @rpc.command_method
    def restart(self, force=False):
        # XXX: force is not implemented
        service_tasks.restart("scalarizr")

    @rpc.query_method
    def status(self, cached=False):
        """Updateclient status."""
        status = hub.updateclient.status

        # FAM-842. Set fields needed by periodic update service.
        update_channel = self.update_channel
        status["repository"] = update_channel["name"]
        status["repo_url"] = update_channel["repos"][0][
            "url"
        ]  # pylint: disable=unsubscriptable-object
        status["service_version"] = hub.agent_version
        status["service_status"] = "unknown"
        package_name = agent.config.workflow_engine["package_name"]
        ud = hub.base.update
        if not cached:
            try:
                sync()
                status["service_status"] = service_tasks.status("scalarizr")
                repo_url = status["repo_url"]

                if fact["manage_container_host"]:
                    agent_repo_type = "docker"
                    package_name = ud.docker_repo_url.split("/")[-1]
                elif fact["os"]["family"] == "redhat":
                    agent_repo_type = "rpm"
                elif fact["os"]["family"] == "suse":
                    agent_repo_type = "suse"
                elif fact["os"]["family"] == "debian":
                    agent_repo_type = "deb"
                else:
                    agent_repo_type = "win"

                pkgmgr_cls = pkgutil.get_pkgmgr(agent_repo_type)
                if agent_repo_type in ("rpm", "suse"):
                    repo_url = pkgutil.RpmManager.normalize_repo_string(
                        repo_url, fact["os"]["name"], str(fact["os"]["release"]), fact["os"]["arch"]
                    )

                pkgmgr = pkgmgr_cls(
                    repo_string=repo_url,
                    storage_dir=hub.updateclient.cache_dir,
                    ssl_verify_peer=hub.base.update.ssl_verify_peer,
                    arch=fact["os"]["arch"],
                )
                pkgmgr.updatedb()
                pkginfo = pkgmgr.status(package_name)
            except Exception:
                LOG.exception("Couldn't get uncached UpdateClient status")
            else:
                status["installed"] = pkginfo["installed"]
                status["candidate"] = pkginfo["candidate"]
        return status

    @rpc.command_method
    def update(self, force=False, bootstrap=False, **kwargs):
        """Update command.

        :param bool force: Bypass UpdateServer check before updating
        :param bool bootstrap: Perform initial update on a new server

        :rtype: dict
        :returns: UpdateClient status after update.
        """
        # try to delete deprecated meta-package before upgrading
        compat.handle_platform_package_deprecation()

        def do_update():
            # save the time when update started
            status = hub.updateclient.status
            status["executed_at"] = time.strftime("%a %d %b %Y %H:%M:%S UTC", time.gmtime())
            hub.updateclient.status = status

            if bootstrap:
                self._cleanup_rpmdb()
            sync()
            ud = hub.base.update
            if ud.server_url:
                update_server = common.types.UpdateServer()
                update_server.url = ud.server_url
                update_server.crypto_key = hub.crypto_key
                update_server.scalr_id = hub.general.scalr_id
                update_server.scalr_version = hub.general.scalr_version
            else:
                update_server = None

            # TODO: update.update(fork=True|False, force=False|True)
            _set_state("in-progress")
            try:
                update_result = update_tasks.update_agent(
                    server_id=hub.general.server_id,
                    channel=self.update_channel.to_primitive(),
                    cache_dir=hub.updateclient.cache_dir,
                    ssl_verify_peer=ud.ssl_verify_peer,
                    update_server=update_server.to_primitive() if update_server else None,
                    bootstrap=bootstrap,
                    force=force,
                    do_update_inproc=True,
                )
            except Exception as e:
                LOG.exception("Update failed with the following error:")
                status = hub.updateclient.status
                status["error"] = "{}".format(e)
                hub.updateclient.status = status
                _set_state("error")
                raise
            else:
                if update_result:
                    LOG.debug(
                        "Successfully completed agent update to %s version",
                        update_result["installed_version"],
                    )
                    _set_state("completed/wait-ack")
                    _ensure_agent_running()
                else:
                    _set_state("completed")
            return self.status()

        def invoke():
            try:
                return do_update()
            finally:
                self._update_lock.release()

        acquired = self._update_lock.acquire(False)  # pylint: disable=assignment-from-no-return
        if not acquired:
            raise Error("Operation is already in progress")
        if kwargs.get("async"):
            return threading.Thread(target=invoke).start()
        else:
            return invoke()


api = API()


def _ensure_agent_running():
    try:
        service_tasks.start("scalarizr")
    except Exception as e:
        LOG.error("Error starting Scalarizr: %s", e)


def _set_state(value):
    s = hub.updateclient.status
    LOG.debug("State transition: %s -> %s", s["state"], value)
    s["prev_state"] = s["state"]
    s["state"] = value
    hub.updateclient.status = s


if win32serviceutil:

    class WindowsService(win32serviceutil.ServiceFramework):
        """UpdateClient windows service."""

        _svc_name_ = hub.updateclient.service_name
        _svc_display_name_ = hub.updateclient.service_name
        _svc_deps_ = ["EventLog"]

        def __init__(self, args=None):
            if args is not None:
                win32serviceutil.ServiceFramework.__init__(self, args)

            def handler(*args):
                return True

            win32api.SetConsoleCtrlHandler(handler, True)
            self.h_stop = win32event.CreateEvent(None, 0, 0, None)

        @mixutil.error_logged("Error in windows service main")
        def SvcDoRun(self):
            servicemanager.LogMsg(
                servicemanager.EVENTLOG_INFORMATION_TYPE,
                servicemanager.PYS_SERVICE_STARTED,
                (self._svc_name_, ""),
            )
            try:
                main()
            except SystemExit:
                servicemanager.LogMsg(
                    servicemanager.EVENTLOG_INFORMATION_TYPE,
                    servicemanager.PYS_SERVICE_STOPPED,
                    (self._svc_name_, ""),
                )
                shutdown()
                return
            LOG.info("Service main loop (Windows)")
            while True:
                rc = win32event.WaitForSingleObject(self.h_stop, 3600)
                if rc == win32event.WAIT_OBJECT_0:
                    servicemanager.LogMsg(
                        servicemanager.EVENTLOG_INFORMATION_TYPE,
                        servicemanager.PYS_SERVICE_STOPPED,
                        (self._svc_name_, ""),
                    )
                    break
            shutdown()

        @mixutil.error_logged("Error in windows service stop")
        def SvcStop(self):
            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
            win32event.SetEvent(self.h_stop)


if __name__ == "__main__":
    main()
