"""Common functionality for agent and update client applications.
"""
import errno
import select
import signal
import sys

from agent import scalrinit
from agent.scalarizr import metadata
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils.facts import fact

if sys.platform == "win32":
    # pylint: disable=import-error
    from common.utils import winutil
    import win32event
else:
    winutil = None
    win32event = None


LOG = logutil.get_logger(__name__)


def daemon_main(shutdown=None):
    """daemon main loop.

    :param func shutdown: Optional shutdown funciton.
    """

    def on_sigterm(*args):
        # pylint: disable=protected-access
        six.moves._thread.interrupt_main()

    def on_sigint(*args):
        """Reimplement default Python SIGINT handler, or else get an error. Sometimes.
        https://bugs.python.org/issue23395
        """
        raise KeyboardInterrupt

    signal.signal(signal.SIGTERM, on_sigterm)
    signal.signal(signal.SIGINT, on_sigint)

    is_windows = fact["os"]["family"] == "windows"
    LOG.info("Service main loop (%s)", "Windows" if is_windows else "Linux")
    if is_windows:
        windows_event = win32event.CreateEvent(None, 0, 0, None)
    while True:
        try:
            if is_windows:
                # [FAM-481] 3 empty sequences are not allowed in select.select on Windows:
                # https://docs.python.org/2/library/select.html
                win32event.WaitForSingleObject(windows_event, 3600)
            else:
                select.select([], [], [], 30)
        except select.error as e:
            # handle interrupted syscall
            if e.args[0] != errno.EINTR:
                raise
        except (KeyboardInterrupt, SystemExit):
            if shutdown is not None:
                shutdown()
            break


def boot_wait():
    """Wait operating system boot (Windows only)."""
    if fact["os"]["name"] != "windows":
        return

    winutil.wait_boot()
    LOG.info("Windows is ready!")


def _vmware_is_customized():
    vmware_provider = scalrinit.metadata.VMwareProvider()
    ud = metadata.Userdata.from_string(vmware_provider.user_data)
    return ud.get("customization_completed") == "1"


def wait_vmware_customization():
    """Wait for VM to get fully customized.
    Relevant only for VMware (FAM-1419).
    """
    ud = metadata.user_data()
    if ud["platform"] != "vmware":
        return

    if mixutil.Version(ud["scalr_version"]) < mixutil.Version("7.7.10"):
        LOG.warning(
            "Scalr >= 7.7.10 has notification to Scalarizr about VM customization completion. "
            "On your Scalr version if unexpected reboot will interrupt Scalarizr before you've seen "
            "server Running in Scalr, initialization will stuck. Consider upgrade to new version"
        )
        return

    customized = ud.get("customization_completed")
    if customized == "1" or customized is None:
        return
    # Initiate polling only for new Scalr versions (attribute is set but not 1).
    LOG.info("Waiting VM customization completion...")
    timeout = 600
    mixutil.try_until(
        _vmware_is_customized,
        error_msg=(
            "No `customization_completed` in user-data after {} seconds. "
            "It may indicate that VM customization has failed on Scalr side"
        ).format(timeout),
        timeout=timeout,
    )
