from __future__ import print_function

import argparse
import functools
import logging
import pprint
import sys
from datetime import datetime

import monotonic

import agent.config
from agent.scalrinit import cloudconfig
from agent.scalrinit import datadir
from agent.scalrinit import metadata
from common.utils import logutil
from common.utils import sysutil

LOG = logutil.get_logger("agent.scalrinit.app")


def main_start(force=False):
    time0 = monotonic.monotonic()
    LOG.info("Scalr-Init started. Uptime: {:.3f}s".format(sysutil.uptime()))

    datadir.result.unlink()

    # 1. check current cloud server id
    try:
        instance_id = metadata.instance_id()
    except Exception as e:
        LOG.error("No Instance-Id: %s", e)
        raise
    LOG.info("Instance-Id: %s", instance_id)
    boot_finished = datadir.find_instance_boot(instance_id)
    if not force and boot_finished and datadir.user_data():
        datadir.result.symlink()
        metadata.remove_local_seeds()  # Fix for previous versions
        LOG.info("Scalr-Init already finished at %s", boot_finished)
        return

    # 2. init current instance
    LOG.info("New instance is booting up...")
    has_user_data = False
    try:
        user_data = metadata.user_data()
        sanitized_user_data = ";".join(
            "szr_key=******" if value.startswith("szr_key") else value
            for value in user_data.split(";")
        )
        LOG.info("User-data: %s", sanitized_user_data)
        has_user_data = True
    except metadata.NoUserdataError as e:
        user_data = ""
        LOG.warning("No User-data found")
    datadir.push_instance(instance_id)
    datadir.datasource.set(metadata.get_datasource())
    datadir.user_data.set(user_data or "")
    if has_user_data:
        metadata.remove_local_seeds()

    # 3. execute cloud-config
    errors = []
    try:
        spr = cloudconfig.parse_script(user_data)
        path = datadir.instance_script(spr)
        cloudconfig.run_script(path, spr.script_type)
    except cloudconfig.NotAScriptError:
        # ignore non script user-data
        pass
    except Exception as e:
        errors.append(str(e))
    finally:
        datadir.boot_finished.set(
            {"datetime": datetime.now().isoformat(), "elapsed_time": monotonic.monotonic() - time0}
        )
        result = {
            "datasource": datadir.datasource()["user_data"],  # Compat before SCALARIZR-2247
            "errors": errors,
        }
        datadir.result.set(result)
        LOG.info("Result: %s", pprint.pformat(result))
        LOG.info("Scalarizr scalr-init finished at %s", datadir.boot_finished())


def main_status():
    print_ = functools.partial(print, sep=": ")

    print("Scalr-Init status", end="\n\n")
    print_("boot_finished", datadir.boot_finished())
    print_("datasource", datadir.datasource())
    print_("instance_id", datadir.instance_id())
    print_("previous_instance_id", datadir.previous_instance_id())
    print_("previous_datasource", datadir.previous_datasource())
    print_("result", datadir.result())


def main(argv=None):
    logutil.basic_config(getattr(logging, agent.config.workflow_engine["log_level"]))
    LOG.debug("Logging configured in %s", __file__)

    parser = argparse.ArgumentParser(
        prog="scalr-init", description="A tiny subset of cloud-init, to initialize Scalarizr"
    )
    # TODO: extend parser arguments

    argv = argv or sys.argv
    if len(argv) < 2:
        argv.append("start")
    if argv[1] in ("start", "force-start"):
        main_start(force=argv[1] == "force-start")
    elif argv[1] == "status":
        main_status()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()
