"""
CloudInit directory layout:
  /var/lib/scalrarizr/cloud/
      - data/
         - instance-id
         - previous-instance-id
         - datasource
         - datasource.instance_id
         - previous-datasource
         - previous-datasource.instance_id
         - result.json
      - instance
      - instances/
          i-00000XYZ/
            - boot-finished
            - user-data.txt
"""
import json
import logging
import os

import agent.config
from common.utils import mixutil
from common.utils import sysutil

LOG = logging.getLogger(__name__)
base_dir = agent.config.workflow_engine["cloud_dir"]


class FileProperty(object):
    def __init__(self, path):
        self.path = path

    def abspath(self):
        return os.path.join(base_dir, self.path)

    def set(self, value):
        path = self.abspath()
        parent = os.path.dirname(path)
        sysutil.mkdir_p(parent)
        LOG.debug("Writing %s", path)
        with open(path, "w+") as fp:
            fp.write(value)

    def get(self):
        path = self.abspath()
        if os.path.exists(path):
            with open(path) as fp:
                return fp.read()

    def __call__(self):
        return self.get()

    def delete(self):
        sysutil.rm_rf(self.abspath())

    def rename_into(self, newpath):
        path = self.abspath()
        if os.path.exists(path):
            LOG.debug("Renaming %s -> %s", path, newpath)
            newpath = os.path.join(base_dir, newpath)
            if os.path.exists(newpath):
                os.remove(newpath)  # for Windows
            os.rename(path, newpath)

    def exists(self):
        return os.path.exists(self.abspath())


class JsonProperty(FileProperty):
    def set(self, value):
        super(JsonProperty, self).set(json.dumps(value))

    def get(self):
        return json.loads(super(JsonProperty, self).get() or "null")


class ResultProperty(JsonProperty):
    def __init__(self, path):
        super(ResultProperty, self).__init__(path)
        self.link_path = os.path.join(
            agent.config.workflow_engine["run_dir"], os.path.basename(path)
        )

    def set(self, value):
        super(ResultProperty, self).set(value)
        self.symlink()

    def unlink(self):
        if os.path.exists(self.link_path):
            LOG.debug("Unlinking %s", self.link_path)
            os.remove(self.link_path)

    def symlink(self):
        if not os.path.exists(agent.config.workflow_engine["run_dir"]):
            os.makedirs(agent.config.workflow_engine["run_dir"])
        if not os.path.exists(self.link_path):
            LOG.debug("Linking %s -> %s", self.link_path, self.abspath())
            sysutil.symlink(self.abspath(), self.link_path)

    def wait(self, timeout=0):
        mixutil.try_until(
            lambda: os.path.exists(self.link_path),
            try_msg="{} not exists".format(self.link_path),
            timeout=timeout,
        )
        LOG.debug("Result exists: %s", self.link_path)


class BootFinishedProperty(FileProperty):
    def set(self, value):
        super(BootFinishedProperty, self).set(
            "{} - {:.3f}s".format(value.get("datetime"), value.get("elapsed_time"))
        )

    def get(self):
        data = super(BootFinishedProperty, self).get().split(" - ")
        return dict(zip(("datetime", "elapsed_time"), data))


class DatasourceProperty(FileProperty):
    def __init__(self, path):
        super(DatasourceProperty, self).__init__(path)
        self.instance_id = FileProperty(path + ".instance_id")

    def set(self, value):
        if isinstance(value, dict):
            super(DatasourceProperty, self).set(value.get("user_data"))
            self.instance_id.set(value.get("instance_id"))
        else:
            super(DatasourceProperty, self).set(value)

    def get(self):
        user_data = super(DatasourceProperty, self).get()
        if self.instance_id.exists():
            instance_id = self.instance_id()
        else:
            instance_id = user_data
        return {"user_data": user_data, "instance_id": instance_id}

    def delete(self):
        super(DatasourceProperty, self).delete()
        self.instance_id.delete()

    def rename_into(self, newpath):
        super(DatasourceProperty, self).rename_into(newpath)
        if self.instance_id.exists():
            self.instance_id.rename_into(newpath + ".instance_id")


boot_finished = BootFinishedProperty("instance/boot-finished")
datasource = DatasourceProperty("data/datasource")
instance_id = FileProperty("data/instance-id")
previous_datasource = DatasourceProperty("data/previous-datasource")
previous_instance_id = FileProperty("data/previous-instance-id")
result = ResultProperty("data/result.json")
user_data = FileProperty("instance/user-data.txt")


def push_instance(inst_id):
    instance_link = os.path.join(base_dir, "instance")
    instance_dir = os.path.join("instances", inst_id)
    instance_dir_abs = os.path.join(base_dir, instance_dir)
    sysutil.rm_rf(instance_dir_abs)
    sysutil.mkdir_p(instance_dir_abs)
    sysutil.unlink(instance_link)
    sysutil.symlink(instance_dir, instance_link, target_is_directory=True)

    g = globals()
    if g["instance_id"].get() != inst_id:
        g["instance_id"].rename_into("data/previous-instance-id")
        g["datasource"].rename_into("data/previous-datasource")
        g["instance_id"].set(inst_id)


def find_instance_boot(inst_id):
    boot_finished = os.path.join(base_dir, "instances", inst_id, "boot-finished")
    if os.path.exists(boot_finished):
        return BootFinishedProperty(boot_finished)


def instance_script(script):
    path = os.path.join(base_dir, "instance", "scripts")
    sysutil.mkdir_p(path)
    sysutil.clean_dir(path)

    path = os.path.join(path, "part-001")
    if script.script_type == "cmd":
        path += ".cmd"
    if script.script_type == "powershell":
        path += ".ps1"
    with open(path, "w+") as fp:
        fp.write(script.content)
    os.chmod(path, 0o700)
    return path
