"""
Metadata managing
"""

from __future__ import absolute_import

import binascii
import functools
import json
import logging
import multiprocessing
import multiprocessing.pool
import os
import posixpath
import re
import sys
import time
import typing as tp
from xml.etree import ElementTree

import monotonic
import requests
import requests.exceptions

from common.utils import exc
from common.utils import facts
from common.utils import mixutil
from common.utils import six
from common.utils import subprocess2
from common.utils import sysutil

if facts.WINDOWS:
    import wmi
    from common.utils import winutil
else:
    from common.utils import linuxutil

    wmi = winutil = None


LOG = logging.getLogger(__name__)
LINUX_PERSONALITY_PATH = "/etc/scalr/private.d/.user-data"
LINUX_PERSONALITY_ALT_PATH = "/etc/.scalr-user-data"
LINUX_PERSONALITY_CONTAINER_PATH = "/rootfs/etc/.scalr-user-data"
WINDOWS_PERSONALITY_PATH = "C:\\Program Files\\Scalarizr\\etc\\private.d\\.user-data"
WINDOWS_PERSONALITY_ALT_PATH = "C:\\.scalr-user-data"
TIMEOUT = int(os.environ.get("FAM_META_DATA_TIMEOUT", 10))
UUID_LEN = 36


class Error(Exception):
    """Module exception"""


class NoProviderError(Error):
    """No User-Data providers that matches environment"""


class NoUserdataError(Error):
    """Empty User-Data returned by provider"""


retries_timeout = mixutil.retries_error(
    (requests.Timeout, OSError), times=10  # [Errno 101] Network is unreachable
)


class Provider(object):
    """
    Describes metadata provider interface
    """

    @property
    def name(self):
        """ Name of provider. Usually it equals to the name of the cloud """
        raise NotImplementedError()

    _base_url = None

    iteration_timeout = 1

    @property
    def LOG(self):
        """ Logger property """
        return logging.getLogger("{}.{}".format(__name__, self.name))

    def requests_get(self, url, *args, **kwds):
        """ Wrapper for requests.get """
        kwds_ = {"timeout": self.iteration_timeout}
        kwds_.update(kwds)
        return requests.get(self._build_url(url), **kwds_)

    def _build_url(self, url):
        if self._base_url and not re.search(r"^http(s)?\://", url):
            url = posixpath.join(self._base_url, url)
        return url

    def fit(self):
        """
        Tries to access metadata storage.
        Returns True if accesed successfully. False otherwise or raises an Error exception.
        """
        raise NotImplementedError()

    @property
    def user_data(self):
        """ Searches for user_data in metadata storage """
        raise NotImplementedError()

    @property
    def instance_id(self):
        """ Searches for instance_id in metadata storage """
        raise NotImplementedError()

    def __str__(self):
        return "<{}>".format(self.__class__.__name__)

    __repr__ = __str__


class Ec2Provider(Provider):
    """
    Provider for EC2 metadata
    """

    name = "ec2"
    _base_url = "http://169.254.169.254/latest"
    retry_interval = 5 if facts.WINDOWS else 1
    wait_timeout = 5 * 60 if facts.WINDOWS else 10  # 5*60 = 5 min
    metadata_token: tp.Optional[str] = None

    if facts.WINDOWS:

        @winutil.coinitialized
        def fit(self):
            # https://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/identify_ec2_instances.html#inspect-uuid
            # Detecting by the first three letters in serial number (EC2), the number might be in big or little endian.
            serial_number = wmi.WMI().Win32_BIOS()[0].SerialNumber.lower().split("-")[0]
            serial_number_little_endian = "{:x}".format(
                int.from_bytes(bytes.fromhex(serial_number), byteorder="little")
            )
            return serial_number.startswith("ec2") or serial_number_little_endian.startswith("ec2")

    else:

        def fit(self):
            return self.requests_get("meta-data/instance-id", headers=self._get_token_headers()).ok

    def _get_token_headers(self):
        if not self.metadata_token:
            LOG.debug("Fetching metadata token")
            args = {
                "timeout": self.iteration_timeout,
                "headers": {"X-aws-ec2-metadata-token-ttl-seconds": "21600"},
            }
            self.metadata_token = requests.put(super()._build_url("api/token"), **args).text

        return {"X-aws-ec2-metadata-token": self.metadata_token}

    def _get_user_data(self):
        try:
            r = self.requests_get("user-data", headers=self._get_token_headers())
            if r.status_code == 404:
                raise NoUserdataError(
                    "User-data is not available at {}/user-data".format(self._base_url)
                )
            return r.text if r.ok else None
        except (requests.Timeout, OSError):
            return None

    @property
    def user_data(self):
        try:
            return mixutil.try_until(
                self._get_user_data,
                try_msg="Trying to get User-data.",
                error_msg="Could not get User-data.",
                log=self.LOG,
                interval=self.retry_interval,
                timeout=self.wait_timeout,
            )
        except exc.Error:
            raise NoUserdataError(
                "User-data is not available at {}/user-data".format(self._base_url)
            )

    def _get_instance_id(self):
        try:
            return self.requests_get(
                "meta-data/instance-id", headers=self._get_token_headers()
            ).text
        except (requests.Timeout, OSError):
            return None

    @property
    def instance_id(self):
        return mixutil.try_until(
            self._get_instance_id,
            try_msg="Trying to get Instance-Id.",
            error_msg="Could not get Instance-Id.",
            log=self.LOG,
            interval=self.retry_interval,
            timeout=self.wait_timeout,
        )


class GceProvider(Provider):
    """
    Provider for GCE metadata
    """

    name = "gce"

    def __init__(self):
        if facts.fact["manage_container_host"]:
            with open("/rootfs/etc/hosts") as fp:
                for line in fp.readlines():
                    if not line.startswith("#") and "metadata" in line:
                        self._base_url = "http://%s/computeMetadata/v1" % line.split()[0]
                        break
        else:
            self._base_url = "http://metadata.google.internal/computeMetadata/v1"

    def requests_get(self, url, *args, **kwds):
        kwds_ = {"headers": {"Metadata-Flavor": "Google"}}
        kwds_.update(kwds)
        return super(GceProvider, self).requests_get(url, *args, **kwds_)

    def fit(self):
        r = self.requests_get(self._base_url)
        return r.headers.get("Metadata-Flavor") == "Google"

    @property
    @retries_timeout
    def user_data(self):
        r = self.requests_get("instance/attributes/scalr")
        if not r.ok:
            raise NoUserdataError("User-data is not available at {}".format(r.url))
        return r.text

    @property
    @retries_timeout
    def instance_id(self):
        return self.requests_get("instance/id").text


class FileUserDataMixin(object):
    """
    Mixin for common methods that are used to get metadata from files.
    """

    def __init__(self):
        self.path = None

    def _find_personality_path(self):
        if self.path:
            return
        possible_paths = None
        if facts.WINDOWS:
            possible_paths = [WINDOWS_PERSONALITY_PATH, WINDOWS_PERSONALITY_ALT_PATH]
        else:
            possible_paths = [
                LINUX_PERSONALITY_PATH,
                LINUX_PERSONALITY_ALT_PATH,
                LINUX_PERSONALITY_CONTAINER_PATH,
            ]

        for path in possible_paths:
            if not os.path.exists(path):
                continue
            if not os.access(path, os.R_OK):
                continue
            self.path = path
            break
        else:
            raise NoUserdataError(
                "User-data files {} do not exist or not accessible.".format(possible_paths)
            )

    @property
    def user_data(self):
        """ Searches for user_data in metadata storage """
        self._find_personality_path()
        with open(self.path) as fp:
            return fp.read()

    def __str__(self):
        return "<{} (path: {})>".format(self.__class__.__name__, self.path)

    __repr__ = __str__


class PersonalityProvider(FileUserDataMixin, Provider):
    """
    Provider for metadata stored in personality file
    """

    name = "personality"

    def fit(self):
        self._find_personality_path()
        return self.path and os.access(self.path, os.R_OK)

    @property
    def instance_id(self):
        raise NotImplementedError()


class DockerProvider(FileUserDataMixin, Provider):
    """
    Provider for docker metadata
    """

    name = "docker"

    def fit(self):
        return len(self.docker_cgroups()) > 0

    @classmethod
    def docker_cgroups(cls):
        """ Returns Docker cgroups """
        retval = []
        if os.path.exists("/proc/1/cgroup"):
            with open("/proc/1/cgroup") as fp:
                cgroups = set((line.rsplit(":")[-1].strip() for line in fp))
            retval = [cg for cg in cgroups if cg.startswith("/docker")]
        return retval

    @property
    def instance_id(self):
        return self.docker_cgroups()[0].split("/")[-1]


class OpenStackProvider(Provider):
    """
    Provider for OpenStack metadata
    """

    name = "openstack"
    iteration_timeout = 5
    _user_data_url = "http://169.254.169.254/openstack/latest/user_data"
    _meta_data_url = "http://169.254.169.254/openstack/latest/meta_data.json"

    def fit(self):
        return self.requests_get(self._user_data_url).ok

    @property
    @retries_timeout
    def user_data(self):
        r = self.requests_get(self._user_data_url)
        if not r.ok:
            raise NoUserdataError("User-data is not available at {}".format(r.url))
        return r.text

    @property
    @retries_timeout
    def instance_id(self):
        return self.requests_get(self._meta_data_url).json()["uuid"]


class OpenStackMixinLegacy(object):
    """
    Container for methods that are used in OpenStack providers.

    Deprecated in favor of new implementation, see FAM-744.
    """

    @property
    def meta_data(self):
        """ Returns OpenStack metadata """
        raise NotImplementedError()

    @property
    def user_data(self):
        """ Searches for user_data in metadata storage """
        try:
            if isinstance(self.meta_data["meta"], dict):
                return ";".join("{}={}".format(*kv) for kv in self.meta_data["meta"].items())
            return self.meta_data["meta"]
        except KeyError:
            raise NoUserdataError(
                'User-data is not available (i.e. no "meta" property in Meta-data)'
            )
        except NotImplementedError as e:
            raise NoUserdataError(str(e))

    @property
    def instance_id(self):
        """ Searches for instance in metadata storage """
        try:
            return self.meta_data["uuid"]
        except KeyError:
            raise NotImplementedError()


class OpenStackProviderLegacy(OpenStackMixinLegacy, Provider):
    """Legacy provider to OpenStack metadata that is accessed by http

    Deprecated in favor of new implementation, see FAM-744.
    """

    name = "openstack_legacy"
    iteration_timeout = 5
    _url = "http://169.254.169.254/openstack/latest/meta_data.json"

    def fit(self):
        return self.requests_get(self._url).ok

    @property
    @retries_timeout
    def meta_data(self):
        r = self.requests_get(self._url)
        try:
            r.raise_for_status()
            return json.loads(r.text)
        except requests.exceptions.HTTPError:
            six.reraise(
                NoUserdataError,
                NoUserdataError("Meta-data is not accessible. URL: {}".format(self._url)),
                sys.exc_info()[2],
            )
        except ValueError:
            six.reraise(
                NoUserdataError,
                NoUserdataError("Could not decode JSON Meta-data"),
                sys.exc_info()[2],
            )


class DriveMixin(object):
    """Implements helper methods for user data providers that attach as a disk."""

    DRIVE_LABEL = ""
    DEFAULT_MOUNT_POINT = ""

    def __init__(self, drive_device=None):
        if facts.WINDOWS:
            self._drive_letter = None
        else:
            self._drive_device = drive_device or "/dev/disk/by-label/" + self.DRIVE_LABEL
        self._mounted = False

    if facts.WINDOWS:

        @winutil.coinitialized
        def _mount(self):
            if not self._mounted:
                c = wmi.WMI()
                vol = c.Win32_Volume(Label=self.DRIVE_LABEL)
                if len(vol) == 0:
                    raise Error(
                        "Disk drive with User-data not found (label: {})".format(self.DRIVE_LABEL)
                    )
                self._drive_letter = vol[0].DriveLetter
                self._mounted = True

        def _absolute_path(self, relpath):
            return os.path.join(self._drive_letter, relpath.replace("/", "\\"))

    else:

        def _mount(self):
            if not self._mounted:
                if not os.path.exists(self._drive_device):
                    raise Error(
                        "Disk drive with User-data not found (label: {})".format(self.DRIVE_LABEL)
                    )
                linuxutil.mount(self._drive_device, self.DEFAULT_MOUNT_POINT)
                self._mounted = True

        def _absolute_path(self, relpath):
            return os.path.join(self.DEFAULT_MOUNT_POINT, relpath)


class ConfigDriveProvider(DriveMixin, Provider):
    """OpenStack configuration drive provider"""

    name = "configdrive"
    DRIVE_LABEL = "config-2"
    DEFAULT_MOUNT_POINT = "/mnt/config"
    USERDATA_RELPATH = "openstack/latest/user_data"
    MEATADATA_RELPATH = "openstack/latest/meta_data.json"

    def fit(self):
        self._mount()
        return os.path.exists(self._absolute_path(self.USERDATA_RELPATH))

    @property
    def user_data(self):
        self._mount()
        userdata_path = self._absolute_path(self.USERDATA_RELPATH)
        if not os.path.exists(userdata_path):
            raise NoUserdataError(
                "User-data file {} does not exist or not accessible.".format(userdata_path)
            )
        if not os.path.getsize(userdata_path):
            raise NoUserdataError("User-data fille {} is empty".format(userdata_path))
        with open(userdata_path) as fp:
            return fp.read()

    @property
    def instance_id(self):
        self._mount()
        meta_data_path = self._absolute_path(self.MEATADATA_RELPATH)
        with open(meta_data_path) as fp:
            return json.load(fp)["uuid"]


class ConfigDriveProviderLegacy(OpenStackMixinLegacy, DriveMixin, Provider):
    """Legacy OpenStack configuration drive provider

    Deprecated in favor of new implementation, see FAM-744.
    """

    name = "configdrive_legacy"
    DRIVE_LABEL = "config-2"
    DEFAULT_MOUNT_POINT = "/mnt/config"
    USERDATA_RELPATH = "openstack/latest/meta_data.json"

    def fit(self):
        self._mount()
        return True

    @property
    def meta_data(self):
        self._mount()
        userdata_path = self._absolute_path(self.USERDATA_RELPATH)
        if not os.path.exists(userdata_path):
            raise NoUserdataError(
                "User-data file {} does not exist or not accessible.".format(userdata_path)
            )
        with open(userdata_path) as fp:
            return json.load(fp)

    @property
    def instance_id(self):
        raise NotImplementedError()


class VMwareProvider(Provider):
    """ VMware configuration provider through guestinfo """

    name = "vmware"

    def __init__(self):
        self._user_data = None
        self._vmtoolsd_path = None
        self._get_cmd = '--cmd="info-get guestinfo.scalr.user-data"'

    def _check_vmtoolsd(self):
        """ Tries to find vmtoolsd and eturns True if vmtoolsd is found, False otherwise. """
        self._vmtoolsd_path = subprocess2.which("vmtoolsd")

        if self._vmtoolsd_path is None and facts.WINDOWS:
            self._vmtoolsd_path = "C:\\Program Files\\VMware\\VMware Tools\\vmtoolsd.exe"
            return os.access(self._vmtoolsd_path, os.X_OK)

        return self._vmtoolsd_path is not None

    def fit(self):
        if not self._check_vmtoolsd():
            raise Error("VMware Tools are not installed.")

        cmd = '"{}" {}'.format(self._vmtoolsd_path, self._get_cmd)
        subprocess2.check_call(cmd, shell=True)
        return True

    @property
    def user_data(self):
        if not self._user_data:
            if not self._check_vmtoolsd():
                raise NoUserdataError("VMware Tools are not installed.")

            cmd = '"{}" {}'.format(self._vmtoolsd_path, self._get_cmd)
            vmtoolsd_call = subprocess2.run(cmd, shell=True)
            if vmtoolsd_call.returncode:
                raise NoUserdataError("User-data does not exist or not accessible.")

            self._user_data = binascii.a2b_base64(vmtoolsd_call.stdout).decode()

        return self._user_data

    @property
    def instance_id(self):
        raise NotImplementedError()


class AzureProvider(Provider):
    """
    Provider for Azure Instance Metadata Service.
    """

    name = "azure_metadata"

    def __init__(self):
        # consider https://docs.microsoft.com/en-us/azure/virtual-machines/windows/instance-metadata-service#service-availability
        # before bumping `api-version`
        # NOTE: customData is available starting from 2019-02-01.
        self._base_url = "http://169.254.169.254/metadata/instance?api-version=2019-02-01"

    @functools.lru_cache(maxsize=1)
    def _fetch(self):
        return self.requests_get(self._base_url, headers={"Metadata": "true"}).json()

    def fit(self):
        try:
            meta = self._fetch()
            if "azEnvironment" not in meta.get("compute", {}):
                LOG.debug(
                    "Metadata Service response has no compute.azEnvironment inside. "
                    "This is not the Azure Cloud"
                )
                return False
        except json.JSONDecodeError:
            return False
        else:
            return True

    @property
    @retries_timeout
    def user_data(self):
        meta = self._fetch()
        if not meta["compute"].get("customData"):
            raise NoUserdataError(
                "Azure Instance Metadata Service returned empty compute.customData"
            )
        return binascii.a2b_base64(meta["compute"]["customData"]).decode()

    @property
    @retries_timeout
    def instance_id(self):
        meta = self._fetch()
        return meta["compute"]["vmId"]


class AzureProviderLegacy(Provider):
    """
    Provider for Azure previous metadata delivery method
    from WAAgent on Linux and C:\\WindowsAzure\\CustomData.bin on Windows
    """

    name = "azure"
    wait_timeout = 5 * 60  # 5 minutes
    retry_interval = 5
    waagent_win_dir = "C:\\WindowsAzure"

    def __init__(self):
        self._user_data = None
        self._path = (
            os.path.expandvars("$SYSTEMDRIVE\\AzureData\\CustomData.bin")
            if facts.WINDOWS
            else "/var/lib/waagent/ovf-env.xml"
        )

    def fit(self):
        return os.path.exists(
            self.waagent_win_dir if facts.WINDOWS else os.path.dirname(self._path)
        )

    def load_user_data_linux(self):
        """ Loads userdata. Linux method """
        ns = {
            "azure": "http://schemas.microsoft.com/windowsazure",
            "ovf": "http://schemas.dmtf.org/ovf/environment/1",
        }
        custom_data_xp = (
            "azure:ProvisioningSection/azure:LinuxProvisioningConfigurationSet/azure:CustomData"
        )
        with open(self._path) as fp:
            xml = ElementTree.fromstring(fp.read())
            custom_data = xml.find(custom_data_xp, ns)
            if custom_data is None:
                raise Exception(
                    "File {} has no user-data node {!r}".format(self._path, custom_data_xp)
                )
            return binascii.a2b_base64(custom_data.text).decode("utf-8")

    def load_user_data_windows(self):
        """ Loads userdata. Windows method """
        with open(self._path) as fp:
            return fp.read()

    def _load_user_data(self):
        try:
            self._user_data = (
                self.load_user_data_windows() if facts.WINDOWS else self.load_user_data_linux()
            )
            return True
        except Exception as e:
            self.LOG.debug("Could not get userdata. Error:\n%s", e)
            return False

    @property
    def user_data(self):
        if self._user_data is None:
            mixutil.try_until(
                self._load_user_data,
                try_msg="Trying to read userdata from file.",
                error_msg="Could not read userdata from {}.".format(self._path),
                log=self.LOG,
                interval=self.retry_interval,
                timeout=self.wait_timeout,
            )
            self.LOG.debug("user_data found")

        return self._user_data

    @property
    def instance_id(self):
        raise NotImplementedError()


class DefaultProvider(Provider):
    """
    General provider of metadata
    """

    name = "default"

    def fit(self):
        return True

    @property
    def instance_id(self):
        """Returns System-ID, or at least Boot-ID.

        So that DefaultProvider's instance_id is guaranteed constant until reboot.
        """
        try:
            LOG.debug("Getting System-Id...")
            return system_id()
        except Exception as e:
            LOG.debug("No System-Id: %s", e, exc_info=sys.exc_info())
        try:
            LOG.debug("Getting Boot-Id...")
            return sysutil.get_boot_id()
        except Exception as e:
            LOG.debug("No Boot-Id: %s", e, exc_info=sys.exc_info())

        raise Error("No one local Instance-Id seed succeed (tried: System-Id, Boot-Id)")

    @property
    def user_data(self):
        raise NoUserdataError("No User-data provider that fits this cloud environment is available")


# Providers are sorted by rank from most suitable to less suitable
_providers = [
    # private clouds
    ConfigDriveProvider,
    ConfigDriveProviderLegacy,
    VMwareProvider,
    OpenStackProvider,
    OpenStackProviderLegacy,
    # public clouds
    GceProvider,
    AzureProvider,
    AzureProviderLegacy,
    Ec2Provider,  # detection is not 100% robust on Windows
    # local seeds
    DockerProvider,
    PersonalityProvider,
    DefaultProvider,
]


def create_provider(name):
    """
    Creates provider for given name.
    Raises NoProviderError if no provider is found.
    """
    if name is None:
        return DefaultProvider()
    for cls in _providers:
        if getattr(cls, "name") == name:
            return cls()
    raise NoProviderError("No Meta-data provider: {}".format(name))


@mixutil.memorize()
def list_fitting_providers():
    """
    Returns list of providers that are fit to current server, i.e. can provide user_data.

    :return list(Provider): list of providers.
    """
    pvds = [p() for p in _providers]

    def fit(pvd):
        """This calls pvd.fit() and immediately return provider if it fits.
        Otherwise retries check and repeat until reaches `TIMEOUT`.
        Maintains `Provider.iteration_timeout` interval between fit() calls.
        """
        time0 = monotonic.monotonic()
        while True:
            pvd.LOG.debug("Fitting...")
            time1 = monotonic.monotonic()
            try:
                fits = pvd.fit()
                pvd.LOG.debug("Fits" if fits else "Does not fit")
                if fits:
                    return pvd
            except Exception as e:
                pvd.LOG.debug("Does not fit: %s", e)
            time2 = monotonic.monotonic()
            if time2 - time0 <= TIMEOUT - pvd.iteration_timeout:
                interval = pvd.iteration_timeout - (time2 - time1)
                if interval > 0:
                    time.sleep(interval)
            else:
                break

    fitting_providers = []
    pool = multiprocessing.pool.ThreadPool(processes=len(pvds))
    try:
        result = pool.map_async(fit, pvds).get()
        fitting_providers = [pvd for pvd in result if pvd]
    finally:
        pool.close()

    if not fitting_providers:
        raise Error("Found no metadata providers that fit.")

    fitting_providers.sort(key=lambda p: _providers.index(p.__class__))
    LOG.info("Meta-data providers that fit: %s", fitting_providers)

    return fitting_providers


_datasource = {}


def set_datasource(datasource):
    """
    Manually sets data source for metadata
    """
    global _datasource
    _datasource = datasource


def get_instance_id_datasource(providers=None):
    """
    Returns name of the provider that can provide instance_id.
    Checks that instance_id is actually accessible.

    :param list(Provider) providers: list of the providers to search on.
        If None, the function will call list_fitting_providers() to get one.

    :return str: Name of the provider that can provide instance_id.
    """
    global _datasource
    if "instance_id" in _datasource:
        return _datasource["instance_id"]

    if providers is None:
        providers = list_fitting_providers()

    for pvd in providers:
        try:
            getattr(pvd, "instance_id")
            break
        except NotImplementedError:
            continue

    # pylint: disable=undefined-loop-variable
    _datasource["instance_id"] = pvd.name
    LOG.info("Best match for Instance-Id: %s", pvd.name)
    return pvd.name


def get_user_data_datasource(providers=None):
    """
    Returns name of the provider that can provide user_data.
    Checks that user_data is actually accessible.

    :param list(Provider) providers: list of the providers to search on.
        If None, the function will call list_fitting_providers() to get one.

    :return str: Name of the provider that can provide user_data.
    """
    global _datasource
    if "user_data" in _datasource:
        return _datasource["user_data"]

    if providers is None:
        providers = list_fitting_providers()

    if os.environ.get("FAM_USER_DATA_PROVIDER"):
        LOG.debug(
            "User-data provider recommended with $FAM_USER_DATA_PROVIDER=%s",
            os.environ["FAM_USER_DATA_PROVIDER"],
        )
        ud_providers = [
            pvd for pvd in providers if pvd.name == os.environ["FAM_USER_DATA_PROVIDER"]
        ]
        if ud_providers:
            providers = ud_providers
        else:
            LOG.debug(
                'Recommended user-data provider "%s" not found among matched',
                os.environ["FAM_USER_DATA_PROVIDER"],
            )

    for pvd in providers:
        try:
            getattr(pvd, "user_data")
            break
        except (NoUserdataError, NotImplementedError):
            continue

    # pylint: disable=undefined-loop-variable
    _datasource["user_data"] = pvd.name
    LOG.info("Best-match for User-data: %s", pvd.name)
    return pvd.name


def get_datasource():
    """
    Finds both datasources for instance_id and user_data.

    :return dict: dict of datasources in format:
        {
            'instance_id': <provider_name1>,
            'user_data': <provider_name2>
        }
    """
    retval = {}
    fitting_providers = list_fitting_providers()
    retval["instance_id"] = get_instance_id_datasource(fitting_providers)
    retval["user_data"] = get_user_data_datasource(fitting_providers)
    return retval


if facts.WINDOWS:

    @winutil.coinitialized
    def system_id():
        """
        .. py:function:: system_id()

           Returns a machine serial number, which we use as an instance_id, when
           there is no opportunity to get it from cloud meta-data

           :rtype: str
        """
        try:
            return wmi.WMI().Win32_BIOS()[0].SerialNumber
        except IndexError:
            raise Error("WMI returns empty UUID")


else:

    def system_id():
        """
        .. py:function:: system_id()

           Returns a machine serial number, which we use as an instance_id, when
           there is no opportunity to get it from cloud meta-data

           :rtype: str
        """
        if not subprocess2.which("dmidecode"):
            raise Error("dmidecode is not available")
        retval = subprocess2.check_output("dmidecode -s system-uuid", shell=True).strip()
        if not retval:
            raise Error("dmidecide returns empty UUID")
        elif len(retval) != UUID_LEN:
            raise Error("dmidecode returns invalid UUID: {}".format(retval))
        return retval


@mixutil.memorize()
def instance_id():
    """ Returns instance_id of current server """
    return create_provider(get_instance_id_datasource()).instance_id


@mixutil.memorize()
def user_data():
    """ Returns user data for current server """
    return create_provider(get_user_data_datasource()).user_data


def remove_local_seeds():
    """Removes all personality seeds from local filesystem.

    This is the MUST CALL after scalrinit successfully initialized current instance datadir.
    """
    LOG.debug("Removing local seeds (personality)")
    for path in (
        LINUX_PERSONALITY_PATH,
        LINUX_PERSONALITY_ALT_PATH,
        LINUX_PERSONALITY_CONTAINER_PATH,
        WINDOWS_PERSONALITY_PATH,
        WINDOWS_PERSONALITY_ALT_PATH,
    ):
        try:
            sysutil.rm_rf(path)
        except EnvironmentError as e:
            LOG.warning(e)
