import base64
import json
import os
import sys
import time
import typing

import requests

import common.types
from agent.utils import celeryutil
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import sysutil

from . import execute

DEFAULT_JOB_TIMEOUT = 120
MIN_REFRESH_INTERVAL = 5
JOB_STATUSES = ("successful", "failed", "canceled")

LOG = logutil.get_logger(__name__)


class AnsibleTowerJobExecutor(execute.ABCActionExecutor):
    """Implements AnsibleTower job execution process
    The class is supervised and stores metainfo as JSON data like: {'job_id': 123}
    Details about AnsibleTower job execution see in: docs/ansible.rst
    """

    cacert_file = None
    verify = False
    auth_token = None
    job_id = None
    job_template_name = None

    def __init__(self, action, task):
        super().__init__(action, task)
        self.task = task
        LOG.debug("AnsibleTowerJobExecutor starts with action %s", action)
        if action.get("ssl_cacert"):
            self.cacert_file = os.path.join(task.local_dir_ext.data_dir, "ansibletower_cacert.pem")
            sysutil.create_file(self.cacert_file, action["ssl_cacert"])
            self.verify = action.get("ssl_verify", True)
        self.timeout = action.get("timeout", DEFAULT_JOB_TIMEOUT)
        self.action = common.types.AnsibleTowerJobAction(action, strict=False)
        self.action.validate()
        self.auth_header = self._get_basic_auth_header()
        self.json_request = {
            "credentials": [self.action.machine_credentials],
            "limit": self.action.host,
            "inventory": self.action.inventory,
            "timeout": self.action.timeout,
        }
        if action.get("variables"):
            self.json_request["extra_vars"] = action.get("variables")
        if action.get("scm_branch"):
            self.json_request["scm_branch"] = action.get("scm_branch")
        self.response = {}

        self.return_stdout = self.action.return_stdout
        self.return_stderr = self.action.return_stderr
        self.stdout_symlink_path = self.action.stdout_symlink_path
        self.stderr_symlink_path = self.action.stderr_symlink_path
        self.stdout_log_path = os.path.join(self.task.local_dir_ext.log_dir, "stdout.log")
        self.stderr_log_path = os.path.join(self.task.local_dir_ext.log_dir, "stderr.log")
        self._link_logs(self.stdout_log_path, self.stdout_symlink_path)
        self._link_logs(self.stderr_log_path, self.stderr_symlink_path)

    def _get_basic_auth_header(self):
        """Constructs header for Basic authorization

        :return: header value
        """
        creds = "%s:%s" % (self.action.username, self.action.password)
        return "Basic %s" % base64.b64encode(creds.encode()).decode()

    def _ansibletower_request(self, url, method="get", kwargs=None):
        """Unified request to AnsibleTower server

        :param url: AnsibleTower URL
        :param method: post or get method
        :param kwargs: dict for request body
        :return: respond dict
        """
        headers = {"Authorization": self.auth_header}

        if method == "post":
            response = requests.post(
                url, json=kwargs, headers=headers, cert=self.cacert_file, verify=self.verify
            )
        else:
            response = requests.get(url, headers=headers, cert=self.cacert_file, verify=self.verify)

        msg = response.content.decode()
        try:
            response.raise_for_status()
        except Exception as e:
            raise type(e)("{}: {}".format(msg, str(e))).with_traceback(sys.exc_info()[2])
        return json.loads(msg)

    def _get_token(self):
        """Gets AnsibleTower authorization token. AnsibleTower binds this token to caller IP
        and User-Agent. So it should be requested on place.

        :return: auth_token string
        """
        url = "{}v2/authtoken/".format(self.action.api_url)
        json_req = {
            "password": self.action.password,
            "username": self.action.username,
        }
        return self._ansibletower_request(url, "post", json_req).get("token")

    def _launch_job(self):
        """Launch AnsibleTower job on predefined AnsibleTower server

        :return: job_id
        """
        url = "{}v2/job_templates/{}/launch/".format(self.action.api_url, self.action.job_template)
        response = self._ansibletower_request(url, "post", kwargs=self.json_request)
        job_id = response.get("id")
        self.job_template_name = response.get("name")
        if job_id is None or not isinstance(job_id, int):
            msg = 'Incorrect or missing field "job_id" in Ansible Tower response'
            log_msg = msg + ": %s", response
            LOG.error(log_msg)
            raise exc.MalformedError(msg)
        self._write_execution_metadata({"job_id": job_id, "timestamp": time.time()})
        return job_id

    def _get_job(self):
        """Checks AnsibleTower job status

        :return: server response dict
        """
        url = "{}v2/jobs/{}/".format(self.action.api_url, self.job_id)
        response = self._ansibletower_request(url)
        if response.get("status") in JOB_STATUSES:
            LOG.debug(
                'Job %s status is "%s". Job result: %s', self.job_id, response["status"], response
            )
            return response
        LOG.debug('Job %s status is "%s"', self.job_id, response.get("status"))

    def _get_awx_stdout(self):
        """Gets AWX/AT stdout

        :return: str stdout text
        """
        url = "{}v2/jobs/{}/stdout/?format=txt_download".format(self.action.api_url, self.job_id)
        response = requests.get(
            url,
            headers={"Authorization": self.auth_header},
            cert=self.cacert_file,
            verify=self.verify,
        ).content.decode()
        LOG.debug("Fetched AWX/AT job stdout %s", response)
        return response

    def _get_result(self):
        """Return result structure

        :return: common.types.AnsibleTowerJobResult object
        """
        response = self.response
        response["return_code"] = response.get("return_code", 1)
        if response.get("status") and response["status"] in JOB_STATUSES:
            response["return_code"] = JOB_STATUSES.index(response["status"])
        command = ["tower-cli job launch"]
        command += ["--%s=%s" % ("job_template", str(self.action.job_template))]
        command += ["--%s=%s" % (key, str(value)) for key, value in self.json_request.items()]

        stdout = stderr = ""
        if self.return_stdout:
            stdout = response.get("result_stdout") or self._get_awx_stdout()
            with open(self.stdout_log_path, "wb") as f:
                f.write(stdout.encode())

        if self.return_stderr:
            stderr = response.get("stderr", "")
            with open(self.stderr_log_path, "wb") as f:
                f.write(stderr.encode())

        result_data = dict(
            # generic ExecutedScript fields
            interpreter=self.action.api_url,
            path=self.action.host,
            command=command,
            run_as=str(self.action.machine_credentials),
            timeout=self.action.timeout,
            success_codes=[0],
            return_code=response["return_code"],
            stdout=stdout,
            stdout_total_bytes=os.path.getsize(self.stdout_log_path),
            stderr=stderr,
            stderr_total_bytes=os.path.getsize(self.stderr_log_path),
            stdout_log_path=self.stdout_log_path,
            stderr_log_path=self.stderr_log_path,
            time_elapsed=response.get("elapsed", 0),
            # AnsibleTowerJobResult specific fields
            job_id=self.job_id or 0,
            job_template_name=self.job_template_name or "None",
            status=response.get("status", ""),
            inventory=self.action.inventory,
        )
        return common.types.AnsibleTowerJobResult(result_data)

    def _wait_response(self, timeout):
        """Wait job execution response within provided timeout

        :param timeout: execution timeout
        :raise exc.TimeoutError: if script finished with timeout
        """

        def interval_generator(
            step: int = MIN_REFRESH_INTERVAL, max_step: int = DEFAULT_JOB_TIMEOUT / 2
        ) -> typing.Generator:
            interval = step
            while True:
                yield interval
                if interval < max_step:
                    interval = 2 * interval
                if interval > max_step:
                    interval = max_step

        msg = 'Checking "%s" job' % str(self.job_id)
        if timeout < MIN_REFRESH_INTERVAL:
            self.response = self._get_job()
            if not self.response:
                raise exc.TimeoutError()
            return
        self.response = mixutil.try_until(
            self._get_job,
            try_msg=msg,
            error_msg=msg,
            log=LOG,
            timeout=timeout,
            interval=interval_generator(),
        )

    def _run(self):
        """Normal run and handle exception

        :raise exc.TimeoutError: if job timeout exceeded
        """
        self.job_id = self._launch_job()
        self._wait_response(self.timeout)

    def _supervise(self):
        """Run supervision case and handle exception

        :raise exc.TimeoutError: if script finished with timeout
        """
        self.job_id = self._metadata["job_id"]
        spent_time = int(time.time()) - int(self._metadata["timestamp"])
        timeout = self.timeout - spent_time
        self._wait_response(timeout)

    def run_or_supervise(self):
        """Main entry to run or supervise execution process

        :return: common.types.AnsibleTowerJobResult object
        """
        try:
            if self._metadata:
                LOG.info("Supervising AnsibleTower job '%s' executed", self.job_id)
                self._supervise()
            else:
                LOG.info("Executing AnsibleTower job template '%s'", self.action.job_template)
                self._run()
        except Exception as e:
            exc_cls = execute.ExecuteError
            self.response = {} if self.response is None else self.response
            if isinstance(e, exc.TimeoutError):
                # pylint: disable=bad-string-format-type
                msg = "Executing AnsibleTower job '%d' timeout exceeded (%d seconds)." % (
                    self.job_id,
                    self.action.timeout,
                )
                self.response["stderr"] = msg
                self.response["return_code"] = execute.ExecuteTimeoutExpired.RETURN_CODE
                LOG.error(msg)
                exc_cls = execute.ExecuteTimeoutExpired
                result = self._get_result().to_primitive()
            else:
                # pylint: disable=bad-string-format-type
                msg = "Error in executing AnsibleTower job template '%s': %r" % (
                    self.action.job_template,
                    e,
                )
                self.response["stderr"] = msg
                self.response["return_code"] = 1
                result = self._get_result().to_primitive()
                LOG.exception(
                    "Error executing AnsibleTower job template '%s': %s, %s",
                    self.action.job_template,
                    e,
                    result,
                )
            raise exc_cls(msg, **result)
        else:
            if self.response.get("status") != JOB_STATUSES[0]:
                msg = "Error in executing AnsibleTower job '%s'" % str(self.job_id)
                self.response["stderr"] = msg
                self.response["return_code"] = 1
                result = self._get_result().to_primitive()
                LOG.warning("Error in executing AnsibleTower job '%s': %s", self.job_id, result)
                raise execute.ExecuteError(msg, **result)
            return self._get_result()


@celeryutil.agent_task(bind=True, name="ansibletower.job")
def job(self, action, **kwargs):
    """
    Entrypoint for AnsibleTower job execution logic

    :param self: celery task object
    :param action: job action data
    :return: result of job execution
    """
    return AnsibleTowerJobExecutor(action, self).run_or_supervise()
