# pylint: disable=no-member
"""
Tasks for installing and running chef-client.

.. moduleauthor:: Marat Komarov
.. moduleauthor:: Victor Ognev
.. moduleauthor:: Roman Koshel
"""
from __future__ import absolute_import

import copy
import errno
import filecmp
import json
import os
import re
import shutil
import signal
import sys
import tempfile

import chef as pychef  # That why absolute import is here
import cryptography.hazmat.backends
import cryptography.hazmat.primitives.serialization
import requests

import common.types
from agent.tasks import execute
from agent.tasks import misc
from agent.tasks import pkgmgr
from agent.tasks import service
from agent.utils import celeryutil
from common.tasks import repo
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

LOG = logutil.get_logger(__name__)
if fact["os"]["family"] == "windows":
    CHEF_ETC_DIR = "C:\\chef"
    CHEF_INSTALL_DIR = "C:\\opscode\\chef"
    CHEFDK_INSTALL_DIR = "C:\\opscode\\chefdk"

    if not os.path.exists(CHEF_INSTALL_DIR) and os.path.exists(CHEFDK_INSTALL_DIR):
        CHEF_BIN_DIR = "C:\\opscode\\chefdk\\embedded\\bin"
    else:
        CHEF_BIN_DIR = "C:\\opscode\\chef\\bin"
else:
    CHEF_ETC_DIR = "/etc/chef"
    CHEF_INSTALL_DIR = "/opt/chef"
    CHEFDK_INSTALL_DIR = "/opt/chefdk"

    if not os.path.exists(CHEF_INSTALL_DIR) and os.path.exists(CHEFDK_INSTALL_DIR):
        CHEF_BIN_DIR = "/opt/chefdk/embedded/bin"
    else:
        CHEF_BIN_DIR = "/usr/bin"

CLIENT_RB_PATH = os.path.join(CHEF_ETC_DIR, "client.rb")
CLIENT_KEY_PATH = os.path.join(CHEF_ETC_DIR, "client.pem")
CHEF_CLIENT_BINARY = os.path.join(CHEF_BIN_DIR, "chef-client")
CHEF_SOLO_BINARY = os.path.join(CHEF_BIN_DIR, "chef-solo")
CLIENT_RB_TEMPLATE = """
log_level        :{log_level}
log_location     STDOUT
chef_server_url  '{server_url}'
validation_client_name '{validation_client_name}'
validation_key   '{validation_key}'
node_name        '{node_name}'
ssl_verify_mode  :{ssl_verify_mode}
"""
SOLO_RB_TEMPLATE = """
log_level        :{log_level}
log_location     STDOUT
cookbook_path    '{cookbook_path}'
role_path        '{role_path}'
data_bag_path    '{data_bag_path}'
environment_path '{environment_path}'
file_cache_path  '{file_cache_path}'
"""


class Error(exc.Error):
    pass


ChefClientCmd = subprocess2.new_command("ChefClientCmd", binary=CHEF_CLIENT_BINARY)
ChefSoloCmd = subprocess2.new_command("ChefSoloCmd", binary=CHEF_SOLO_BINARY)


if fact["os"]["family"] != "windows" and not os.access("/etc/init.d/chef-client", os.X_OK):
    # TODO(marat) move below functions into utils
    def get_shell_vars(script_path):
        """
        Starts /bin/sh,
        includes script `script_path`,
        dumps ENV variables
        and returns them a dict
        """
        return {}

    def get_profile_gv():
        profile_path = "/etc/profile.d/scalr_globals.sh"
        if os.path.exists(profile_path):
            return get_shell_vars(profile_path)
        return {}

    def kill_proc(pid_file, signum=signal.SIGTERM):
        if not os.path.exists(pid_file):
            return
        with open(pid_file) as fp:
            try:
                pid = int(fp.read().strip())
            except ValueError:
                return
            os.kill(pid, signum)
            return True

    class ChefClientServiceProvider(service.ABCServiceProvider):
        has_reload = False
        has_restart = False

        log_file = "/var/log/chef-client.log"
        pid_file = "/var/run/chef-client.pid"

        def start_service(self, name):
            if fact["os"]["init"] == "systemd":
                subprocess2.call(["systemctl", "start", "chef-client.service"])
            else:
                cmd = ChefClientCmd().args(daemonize=True, logfile=self.log_file, pid=self.pid_file)
                try:
                    subprocess2.check_call(
                        cmd.build(),
                        preexec_fn=os.setsid,
                        stdout=subprocess2.DEVNULL,
                        stderr=subprocess2.DEVNULL,
                        env=get_profile_gv(),
                    )
                except subprocess2.CalledProcessError as e:
                    if e.returncode == 255:
                        LOG.debug("chef-client daemon already started")
                    else:
                        raise service.Error("Service {!r} start failed: {}".format(name, e))
                except IOError as e:
                    if e.errno == errno.ENOENT:
                        raise service.NotInstalledError(name)
                    raise

        def stop_service(self, name):
            if fact["os"]["init"] == "systemd":
                subprocess2.call(["systemctl", "stop", "chef-client.service"])
            else:
                kill_proc(self.pid_file)

        def get_service_enabled(self, name):
            if fact["os"]["init"] == "systemd":
                return subprocess2.call(["systemctl", "is-enabled", "chef-client.service"]) == 0
            return True

        def set_service_enabled(self, name, enabled):
            source = "/opt/scalarizr/init/systemd/chef-client.service"
            target = "/lib/systemd/system/chef-client.service"

            if fact["os"]["init"] == "systemd":
                install_unit = False

                # Remove sysvinit script if exists
                if os.path.exists("/etc/init.d/chef-client"):
                    os.unlink("/etc/init.d/chef-client")
                    subprocess2.call(["systemctl", "daemon-reload"])

                # Install/reinstall chef-client systemd unit
                chef_service = [
                    i
                    for i in subprocess2.check_output(["systemctl"]).splitlines()
                    if i.startswith("chef-client")
                ]
                if not chef_service:
                    install_unit = True
                if os.path.exists(target) and not filecmp.cmp(source, target):
                    install_unit = True
                if install_unit:
                    shutil.copy(source, target)
                    subprocess2.call(["systemctl", "daemon-reload"])

                # Enable chef-client autostart
                if subprocess2.call(["systemctl", "is-enabled", "chef-client.service"]) != 0:
                    subprocess2.call(["systemctl", "enable", "chef-client.service"])

        def get_service_status(self, name):
            if fact["os"]["init"] == "systemd":
                return subprocess2.call(["systemctl", "is-active", "chef-client.service"]) == 0
            else:
                try:
                    return kill_proc(self.pid_file, 0)
                except OSError:
                    return False

    service.name_to_provider["chef-client"] = ChefClientServiceProvider


@celeryutil.agent_task(name="chef.install")
def install(version="auto", package_url=None, **kwargs):
    """
    Check an os type and install an appropriate version of chef.
    """
    regex = r"\d{1,2}\.\d{1,2}\.\d{1,2}"
    match = re.match(regex, version)
    if version not in ("latest", "auto") and not match:
        raise exc.MalformedError(
            'Version "%s" is not valid. '
            'Possible values are "<x>.<y>.<z>", "latest", or "auto"' % version
        )
    if version == "auto":
        try:
            installed_version()
            return
        except exc.NotFoundError:
            pass
        version = "latest"

    if fact["os"]["family"] != "windows":
        os.setuid(os.getuid())
    try:
        if fact["os"]["family"] == "windows":
            _install_windows(version, package_url)
        else:
            _install_linux(version, package_url)
    except subprocess2.CalledProcessError as e:
        raise Error("Chef installation failed with the following error: %s" % e)


def _install_linux(version, package_url=None):
    if package_url:
        mixutil.validate_url(package_url)
        pkgmgr.install(source=package_url)
    else:
        ver_flag = ""
        if version is not "latest":
            ver_flag = "-s -- -v {} ".format(version)
        command = "curl -L {} | bash {}".format("https://www.opscode.com/chef/install.sh", ver_flag)
        subprocess2.check_output(command, shell=True)


def default_windows_package_url(version=None):
    LOG.debug("Finding package url for chef installer")
    if not version or version == "latest":
        version = ""
    metadata_url = (
        "https://www.opscode.com/chef/metadata?"
        "v={}&prerelease=false&nightlies=false&p=windows&pv=2008r2&m=x86_64".format(version)
    )
    metadata = requests.get(metadata_url)
    LOG.debug("Chef installer metadata: %s, response code: %s", metadata.text, metadata.status_code)
    for word in metadata.text.split():
        if word.startswith("http") or word.startswith("https"):
            LOG.debug("Package url for chef installer: %s", word)
            return word


def _install_windows(version, package_url=None):
    package_url = package_url or default_windows_package_url(version)
    if not package_url:
        raise exc.NotFoundError(
            "Couldn't find download url for chef installer version {}".format(version)
        )
    tmp_dir = tempfile.gettempdir()
    pkg_path = misc.fetch(package_url, tmp_dir)
    subprocess2.check_output(
        ["msiexec", "/i", pkg_path, "/qn", "/l*", os.path.join(tmp_dir, "chef_install.log")],
        shell=True,
    )


@celeryutil.agent_task(bind=True, name="chef.bootstrap")
def bootstrap(self, action, **kwargs):
    """Run chef-client bootstrap action

    :param action: A chef-client action to run
    :type action: Serialized :class:`common.types.ChefClientBootstrapAction`

    :rtype: :class:`common.types.BootstrappedChefClient`

    Example:

        chef.bootstrap(dict(
            server_url='https://chef.io/organization/myorg',
            json_attributes={
                'run_list': ['role[applicatiom]', 'role[monitoring]'],
                'application': {
                    'install_dir': '/opt/app/current'}
                },
            validation_key='--- BEGIN PRIVATE KEY ---...',
            validation_client_name='myorg-validator',
            daemonize=True))
    """
    a = common.types.ChefClientBootstrapAction(action)
    a.validate()

    # create chef-client configuration files
    a.node_name = a.node_name or fact["hostname"]
    a.validation_key = sysutil.create_file(
        os.path.join(self.local_dir_ext.data_dir, "validation.pem"), a.validation_key
    )
    if not a.ssl_verify_mode:
        a.ssl_verify_mode = "verify_none" if installed_version() < 12 else "verify_peer"
    verify_peer = a.ssl_verify_mode == "verify_peer"
    check_key(a.validation_key)
    check_client_name(
        a.node_name, a.server_url, a.validation_client_name, a.validation_key, verify_peer
    )
    sysutil.rm_rf(CLIENT_KEY_PATH)
    vrs = copy.copy(locals())
    vrs.update(a.to_primitive())
    client_rb = CLIENT_RB_TEMPLATE.format(**vrs)
    if a.client_rb_template:
        client_rb = "{}\n{}".format(a.client_rb_template, client_rb)
    if a.environment:
        client_rb = "{}\nenvironment      '{}'".format(client_rb, a.environment)
    sysutil.create_file(CLIENT_RB_PATH, client_rb)

    # stop chef before configuring
    try:
        service.stop("chef-client")
    except (service.NotInstalledError, service.NotRunningError, service.NotRunningError):
        pass

    # execute chef-client
    cmd = ChefClientCmd()
    json_attr_path = os.path.join(self.local_dir_ext.data_dir, "attr.json")
    sysutil.create_file(json_attr_path, json.dumps(a.json_attributes), 0o644)
    cmd.args("-j", json_attr_path)

    executed_script = create_and_run_chef_script(self, a, cmd)

    # maybe delete chef client and node
    if a.delete_node or a.delete_client:
        try:
            erase_client(a.node_name, a.server_url, verify_peer, a.delete_node, a.delete_client)
        except Exception as e:
            msg = 'Unable to delete chef client: "{0}" \n due to: {1}'.format(a.node_name, e)
            LOG.warning(msg, exc_info=sys.exc_info())

    # maybe daemonize chef client
    if a.daemonize:
        try:
            service.enable("chef-client")
            service.start("chef-client")
        except service.NotInstalledError:
            LOG.warning(
                "Can't daemonize Chef, cause 'chef-client' is not a registered Windows Service.\n"
                "Most likely you haven't selected Chef Service option in Chef installer."
            )
        except Exception as e:
            msg = "Can't daemonize Chef: {}".format(e)
            LOG.warning(msg, exc_info=sys.exc_info())

    return common.types.BootstrappedChefClient(
        dict(executed_script=executed_script, node_name=a.node_name)
    )


@celeryutil.agent_task(bind=True, name="chef.override")
def override(self, action, **kwargs):
    """Run chef-client override action

    :param action: A chef-client action to run
    :type action: Serialized :class:`common.types.ChefClientOverrideAction`

    :rtype: :class:`common.types.ExecutedScript`
    """
    a = common.types.ChefClientOverrideAction(action)
    a.validate()
    check_client_rb()

    cmd = ChefClientCmd()
    if a.json_attributes:
        json_attr_path = os.path.join(self.local_dir_ext.data_dir, "attr.json")
        sysutil.create_file(json_attr_path, json.dumps(a.json_attributes), 0o644)
        cmd.args("-j", json_attr_path)
    # pylint: disable=not-an-iterable
    for item in a.run_list:
        cmd.args("-o", mixutil.quote(item))

    return create_and_run_chef_script(self, a, cmd)


@celeryutil.agent_task(bind=True, name="chef.reconverge")
def reconverge(self, action, **kwargs):
    """Run chef-client reconverge action

    :param action: A chef-client action to run
    :type action: Serialized :class:`common.types.ChefClientReconvergeAction`

    :rtype: :class:`common.types.ExecutedScript`
    """
    a = common.types.ChefClientReconvergeAction(action)
    a.validate()
    check_client_rb()

    cmd = ChefClientCmd()
    return create_and_run_chef_script(self, a, cmd)


def create_and_run_chef_script(self, action, cmd):
    script_code = create_chef_script(" ".join(cmd.build()))

    LOG.info("Executing chef ...")
    # pylint: disable=no-value-for-parameter
    try:
        retval = execute.script(
            action=dict(
                action_type="script",
                name="chef",
                code=script_code,
                timeout=action.timeout,
                cwd=action.cwd,
                env=action.env,
                run_as=action.run_as,
                success_codes=action.success_codes,
                return_stdout=action.return_stdout,
                return_stderr=action.return_stderr,
                stdout_symlink_path=action.stdout_symlink_path,
                stderr_symlink_path=action.stderr_symlink_path,
            )
        )

        LOG.info("Successfully executed chef")
        return retval
    except execute.ExecuteError as e:
        # this error can be tricky to read, that is why we combine a chef stacktrace
        # together with stderr tail
        stacktrace_path = mixutil.firstmatched(
            os.path.exists,
            (
                os.path.join(self.local_dir_ext.data_dir, "cache", "chef-stacktrace.out"),
                "/var/chef/cache/chef-stacktrace.out",
                "C:\\chef\\cache\\chef-stacktrace.out",
            ),
        )
        if stacktrace_path is not None:
            err_msg = "{}\nChef trace:\n{}".format(
                e.get_default_pretty_error(), sysutil.tail(stacktrace_path, from_head=True)
            )
            e.set_pretty_error(err_msg)
        e.command = cmd.binary
        raise


@celeryutil.agent_task(bind=True, name="chef.solo")
def solo(self, action, **kwargs):
    """Run chef-solo orchestration action

    :param action: chef-solo action to run
    :type action: :class:`common.types.ChefSoloAction`

    :rtype: :class:`common.types.ExecutedScript`

    Example:
    Bootstrap from GitHub public repository::

        chef.solo(dict(
            json_attributes={
                'run_list': ['role[applicatiom]', 'role[monitoring]'],
                'application': {
                    'install_dir': '/opt/app/current'}
                },
            cookbooks={
                'source_type': 'repository',
                'repo': {
                    'url': 'git+https://github.com/MyOrg/application_devops@develop'},
                'relative_path': 'relative/path/in/repository'}))

    Example:
    Bootstrap from GitHub private repository::

        chef.solo(dict(
            ...
            cookbooks={
                'source_type': 'repository',
                'repo': {
                    'url': 'git+git@github.com:Scalr/cookbooks.git',
                    'ssh_key': '--- BEGIN PRIVATE KEY ...'}}))

    Example:
    Bootstrap from URL::

        chef.solo(
            ...
            cookbooks={
                'source_type': 'url',
                'url': 'https://myorg.example.com/application_devops/cookbooks.tar.gz'})

    Example:
    Bootstrap from local path::

        chef.solo(
            ...
            cookbooks={
                'source_type': 'path',
                'path': '/path/to/cookbooks/directory'})

    """
    a = common.types.ChefSoloAction(action)
    a.validate()
    Url = common.types.CookbookUrlSource
    Repo = common.types.CookbookRepositorySource
    Path = common.types.CookbookPathSource

    # pylint: disable=possibly-unused-variable
    file_cache_path = os.path.join(self.local_dir_ext.data_dir, "cache")
    cookbook_path = None

    if isinstance(a.cookbooks, (Url, Repo)):
        cookbook_path = os.path.join(self.local_dir_ext.data_dir, "cookbooks")
        sysutil.mkdir_p(cookbook_path)
        if isinstance(a.cookbooks, Url):
            # BTW(marat) chef-solo can download cookbooks by it's own
            # (`recipe_url` option in solo.rb)
            misc.unarchive(misc.fetch(a.cookbooks.url, cookbook_path), cookbook_path)
        elif isinstance(a.cookbooks, Repo):
            a.cookbooks.repo.work_dir = cookbook_path
            # pylint: disable=no-value-for-parameter
            repo.sync(a.cookbooks.repo)
        if isinstance(a.cookbooks, (Repo, Url)) and a.cookbooks.relative_path:
            # cookbook directory
            cookbook_path = os.path.join(cookbook_path, a.cookbooks.relative_path)
    elif isinstance(a.cookbooks, Path):
        cookbook_path = a.cookbooks.path

    if not cookbook_path:
        raise exc.MalformedError("No `cookbook_path` to operate with")
    if not os.path.exists(cookbook_path):
        raise exc.NotFoundError("Cookbooks directory not exists: {!r}".format(cookbook_path))

    if "cookbooks" in os.listdir(cookbook_path) and not (
        isinstance(a.cookbooks, Repo) and a.cookbooks.relative_path
    ):
        # chef repository
        repo_dir = cookbook_path
        cookbook_path = os.path.join(repo_dir, "cookbooks")
        # pylint: disable=possibly-unused-variable
        role_path = os.path.join(repo_dir, "roles")
        data_bag_path = os.path.join(repo_dir, "data_bags")
        environment_path = os.path.join(repo_dir, "environments")
    else:
        # default locations
        # pylint: disable=possibly-unused-variable
        role_path = "/var/chef/roles"
        data_bag_path = "/var/chef/data_bags"
        environment_path = "/var/chef/environments"

    solo_rb_path = os.path.join(self.local_dir_ext.data_dir, "solo.rb")
    vrs = copy.copy(locals())
    vrs.update(a.to_primitive())
    solo_rb = SOLO_RB_TEMPLATE.format(**vrs)
    if a.solo_rb_template:
        solo_rb = "{}\n{}".format(a.solo_rb_template, solo_rb)
    sysutil.create_file(solo_rb_path, solo_rb)
    attr_json_path = os.path.join(self.local_dir_ext.data_dir, "attr.json")
    sysutil.create_file(attr_json_path, json.dumps(a.json_attributes))

    # execute chef-solo
    cmd = ChefSoloCmd()
    cmd.args("-c", solo_rb_path, "-j", attr_json_path)
    if installed_version() >= (12, 10, 54):
        cmd.args("--legacy-mode")  # see https://github.com/chef/chef/pull/4919
    return create_and_run_chef_script(self, a, cmd)


def installed_version() -> mixutil.Version:
    """
    Return installed Chef Client version.

    :raises: :class:`exc.NotFoundError`
    """
    if os.path.exists(CHEF_INSTALL_DIR):
        version_manifest_json = os.path.join(CHEF_INSTALL_DIR, "version-manifest.json")
        if os.path.exists(version_manifest_json):
            with open(os.path.join(version_manifest_json)) as fp:
                return mixutil.Version(json.load(fp)["build_version"])

        version_manifest_txt = os.path.join(CHEF_INSTALL_DIR, "version-manifest.txt")
        if os.path.exists(version_manifest_txt):
            with open(version_manifest_txt) as fp:
                version_string = fp.readline().strip()
                return mixutil.Version(version_string.split()[-1])

        raise exc.NotFoundError(
            (
                "Chef version manifest is not found in following locations: {}. "
                "Is Chef installed?"
            ).format([version_manifest_json, version_manifest_txt])
        )

    # The Chef Development Kit includes chef-client tool
    if os.path.exists(CHEFDK_INSTALL_DIR):
        try:
            version_cmd = ChefClientCmd().args(version=True)
            version_stdout = version_cmd.check_output(shell=fact["os"]["family"] == "windows")
            version_se = re.search(r"(\d+\.?\d+\.?\d+)", version_stdout or "")
            if not version_se:
                raise exc.Error(("Invalid version string: {}").format(version_stdout))
            version_str = version_se.group(0)
        except (subprocess2.CalledProcessError, exc.Error) as err:
            raise exc.NotFoundError(
                ("Unable to detect chef-client version in Chef Development " "Kit: {}").format(err)
            )
        return mixutil.Version(version_str)

    raise exc.NotFoundError(
        ("Chef installation not found in following locations: {0}, {1}").format(
            CHEF_INSTALL_DIR, CHEFDK_INSTALL_DIR
        )
    )


def create_chef_script(shell_cmd):
    if fact["os"]["family"] == "windows":
        return "#!cmd\n@echo off\ncall {}\n".format(shell_cmd)
    else:
        return "#!/bin/bash\n{}\n".format(shell_cmd)


def check_key(filename):
    """
    Check if a key exists at a given path and  has a correct format.
    """
    try:
        with open(filename, "rb") as fp:
            cryptography.hazmat.primitives.serialization.load_pem_private_key(
                fp.read(), password=None, backend=cryptography.hazmat.backends.default_backend()
            )
    except Exception as e:
        msg = (
            "Unable to load your key (with Cryptography): {}\n"
            "Please ensure, that location you're passing is a valid PEM-encoded key\n{}"
        ).format(filename, repr(e))
        raise Error(msg)


def check_client_name(node_name, server_url, validation_client_name, validation_key, verify_peer):
    """
    Check if given node name is unique.
    """
    LOG.debug("Chef node_name validation skipped (see FAM-771)")
    return


def check_client_rb():
    # reconverge and override expect already configured chef-client
    if not os.path.exists(CLIENT_RB_PATH):
        msg = ("No validation key and server url provided " "and no client config at {!r}").format(
            CLIENT_RB_PATH
        )
        raise exc.NotFoundError(msg)


def erase_client(node_name, server_url, verify_peer, delete_node=False, delete_client=False):
    """
    Connect to chef api and delete node, client or both if corresponding flags
    have been passed to chef client.
    """
    LOG.info(
        "Deleting client %r (delete_client: %s delete_node: %s)",
        node_name,
        delete_client,
        delete_node,
    )
    with pychef.ChefAPI(server_url, CLIENT_KEY_PATH, node_name, ssl_verify=verify_peer):
        client_ = pychef.Client(node_name)
        node = pychef.Node(node_name)
        if delete_node:
            node.delete()
        if delete_client:
            client_.delete()
