# pylint: disable=no-member
"""
Tasks to execute scripts on Linux/Windows.

.. moduleauthor:: Marat Komarov
.. moduleauthor:: Victor Ognev
"""
from __future__ import absolute_import

import abc
import json
import os
import posixpath
import re
import time
from collections import Iterable

import celery.exceptions
import monotonic
import psutil
import requests

import agent.config
import common.types
from agent.utils import celeryutil
from common.tasks import repo
from common.utils import configutil
from common.utils import exc
from common.utils import logutil
from common.utils import six
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

try:
    from common.utils import linuxutil
except ImportError:
    linuxutil = None

LOG = logutil.get_logger(__name__)
WINDOWS_VIRTUAL_INTERPRETERS = ("powershell", "cmd", "cscript")


class ExecuteError(exc.Error):
    command = None
    return_code = None
    stderr = stderr_log_path = None
    stdout = stdout_log_path = None
    _pretty_error = None

    def get_pretty_error(self):
        return self._pretty_error or self.get_default_pretty_error()

    def set_pretty_error(self, message):
        self._pretty_error = message

    def get_default_pretty_error(self):
        tail = None
        if self.stderr:
            tail = sysutil.tail_s(self.stderr)
        elif self.stderr_log_path:
            tail = sysutil.tail(self.stderr_log_path)
        return "Stderr tail:\n{}".format(tail) if tail else ""

    def get_pretty_command(self):
        if (
            not isinstance(self.command, six.string_types)
            and isinstance(self.command, Iterable)
            and len(self.command) >= 1
        ):
            # pylint: disable=not-an-iterable, unsubscriptable-object
            return " ".join(self.command)
        return "{}".format(self.command)

    def __str__(self):
        super(ExecuteError, self).__str__()
        return 'Command "{}" exited with code {} \n{}'.format(
            self.get_pretty_command(), self.return_code, self.get_pretty_error()
        )


class ExecuteTimeoutExpired(ExecuteError):
    RETURN_CODE = 130
    timeout = None

    def __str__(self):
        super(ExecuteTimeoutExpired, self).__str__()
        return 'Command "{}" timed out after {} seconds.\n{}'.format(
            self.get_pretty_command(), self.timeout, self.get_pretty_error()
        )


class Shebang(object):
    """
    Class for managing regular and 'virtual' shebang.
    """

    INTERPRETERS = {
        ".ps1": "powershell",
        ".bat": "cmd",
        ".cmd": "cmd",
    }

    def __init__(self, path=None, code=None):
        """
        Reads given code and parses shebang.
        """
        assert path or code
        if path and sysutil.is_binary(path):
            self.interpreter = path
            return

        if path:
            with open(path, encoding="utf-8") as fp:
                self.lines = [fp.readline(), fp.readline()]
        else:
            self.lines = code.splitlines()[:2]
            if len(self.lines) < 2:
                self.lines.append("")
        shebang = re.search(r"^#!(\S+.+)", self.lines[0])
        if not shebang:
            if fact["os"]["family"] == "windows" and path:
                script_ext = os.path.splitext(path)[1].lower()
                self.interpreter = self.INTERPRETERS.get(script_ext, "cscript")
                return
            raise exc.MalformedError(
                "First line of the script must be in a form of shebang:\n"
                "#!<interpreter> [optional-arg]"
            )

        self.interpreter = [item.strip() for item in shebang.group(1).split()][0]
        if fact["os"]["family"] != "windows" and not os.path.exists(self.interpreter):
            raise exc.NotFoundError("Interpreter not found {!r}".format(self.interpreter))

    def normalize_script_name(self, script_name):
        """Modifies script name in accordance with OS requirements.

        :see: WindowsShebang.normalize_script_name()
        """
        return script_name

    def normalize_script_code(self, script_code):
        """Modifies script code in accordance with OS requirements.

        :see: WindowsShebang.normalize_script_code()
        """
        return script_code

    def build_args(self, path, user=None):
        """
        Builds subprocess.Popen args.
        """
        if user and user != "root":
            # preserving environment with -E and setting HOME to the homedir of user with -H
            return [["sudo", "-E", "-H", "-u", user, path]]
        return [[path]]


class WindowsShebang(Shebang):
    def __init__(self, path=None, code=None):
        """
        Parses seconod line with virtual shebang of given code to get windows launch parameters.
        Vitual shebang formats:
        #!windows command=<path_to_binary>
        #!windows extension=<extension>
        """
        self.lines = None
        super(WindowsShebang, self).__init__(path, code)
        self.extension = None
        self.params = None

        shebang = re.search(r"^#!windows(.+)", self.lines[1])
        if not shebang:
            old_style_shebang = re.search(r"^#!(\S+.+)", self.lines[1])
            if old_style_shebang:
                self.interpreter = old_style_shebang.group(1).strip()
            if self.interpreter not in WINDOWS_VIRTUAL_INTERPRETERS:
                raise exc.NotFoundError(
                    "The only supported interpreters on Windows in first shebang are {}. Got {}".format(
                        WINDOWS_VIRTUAL_INTERPRETERS, self.interpreter
                    )
                )
            return

        parameters = shebang.group(1).strip()
        name, val = (x.strip() for x in parameters.split("=", 1))
        val = val.strip('"')
        if name == "command":
            val = val.split()
            self.interpreter, self.params = val[0], val[1:]
        elif name == "extension":
            self.extension = val
            self.interpreter = "registered_app"
        else:
            raise exc.MalformedError(
                "Second shebang of the script must be in a following format:\n"
                "#!windows (command|extension)=<value>"
            )

    def normalize_script_name(self, script_name):
        """
        Normalizes extensions.
        """
        if not script_name:
            return None
        ext = os.path.splitext(script_name)[1]
        if self.interpreter == "powershell" and ext not in (".ps1", ".psm1"):
            return script_name + ".ps1"
        elif self.interpreter == "cmd" and ext not in (".cmd", ".bat"):
            return script_name + ".bat"
        elif self.interpreter == "registered_app" and ext[1:] != self.extension:
            return script_name + "." + self.extension
        return script_name

    def normalize_script_code(self, script_code):
        """Removes first lines containing virtual shebang."""
        if script_code.startswith("#!"):
            # windows shebang
            lines = script_code.splitlines()
            start = 1
            if len(lines) > 1 and lines[1].startswith("#!"):
                # cross-platform script with linux and windows shebangs
                start = 2
            return "\n".join(lines[start:])
        return script_code

    def build_args(self, path, user=None):
        """
        Builds subprocess.Popen args.
        """
        assert self.interpreter
        # self.params is None and self.params == [] is different.
        # In the first case we received alias as interpreter and need to add default params.
        # In the second one we received interpreter and no params.
        if self.interpreter == "powershell" and self.params is None:
            return [
                [
                    "powershell.exe",
                    "-NoProfile",
                    "-NonInteractive",
                    "-ExecutionPolicy",
                    "RemoteSigned",
                    "-File",
                    path,
                ]
            ]
        if (
            self.interpreter == "cmd" and self.params is None
        ) or self.interpreter == "registered_app":
            return [["cmd.exe", "/C", path]]
        self.params = self.params or []
        return [[self.interpreter] + self.params + [path]]


@six.add_metaclass(abc.ABCMeta)
class ABCActionExecutor(object):
    """Base abstract class to provide supervising functionality."""

    def __init__(self, action, task):
        self.execution_metadata_file = os.path.join(
            task.local_dir_ext.data_dir, "execution_metadata.json"
        )
        self._metadata = self._read_execution_metadata()

    @abc.abstractmethod
    def _supervise(self):
        """
        Run supervision case and handle exception
        :raise ExecuteError: if return code not in success codes
        :raise exc.TimeoutError: if script finish with timeout
        """

    @abc.abstractmethod
    def _run(self):
        """
        Implement normal run and handle exception
        :raise ExecuteError: if return code not in success codes
        :raise exc.TimeoutError: if script finish with timeout
        """

    def _read_execution_metadata(self):
        """
        Read execution metadata file
        :return: metadata dict
        """
        if os.path.exists(self.execution_metadata_file):
            LOG.debug("Execution metadata file was found. Proceed supervision")
            with open(self.execution_metadata_file) as mf:
                return json.load(mf)
        else:
            LOG.debug("Execution metadata file not found. Proceed normal run")

    def _write_execution_metadata(self, metadata):
        """
        Run execution metadata file that needs for Supervision
        """
        LOG.debug("Write execution metadata to: %s", self.execution_metadata_file)
        sysutil.mkdir_p(os.path.dirname(self.execution_metadata_file), mode=0o755)
        with open(self.execution_metadata_file, "w") as metadata_file:
            json.dump(metadata, metadata_file)

    def _link_logs(self, log_path, symlink_path):
        """
        Link logs, need for user purposes (SCALARIZR-2327)
        :param log_path: actual path to logs
        :param symlink_path: path to symlink
        """
        if not symlink_path:
            return
        if not os.path.exists(log_path):
            sysutil.touch(log_path)
        if not os.path.exists(symlink_path):
            sysutil.mkdir_p(os.path.dirname(symlink_path))
            sysutil.symlink(log_path, symlink_path)


class BaseScriptExecutor(ABCActionExecutor):
    """Base executor class that implements all base
    functionality - parse initial data, create script file
    """

    exc_cls = ExecuteError
    raise_exc = False
    exc_message = ""
    return_code = 1
    rootfs_dir = "/"
    time0 = None
    base_name = None
    script_url = None
    return_stdout = None
    return_stderr = None
    stdout_symlink_path = None
    stderr_symlink_path = None
    bin_dir = None
    stdout_log_path = None
    stderr_log_path = None
    cwd = None
    success_codes = None
    timeout = None
    env = None
    user = None
    shebang = None
    script_name = None
    script_code = None
    script_path = None
    rootfs_script_path = None

    def __init__(self, action, task):
        """
        Parse initial data, create script file
        :param action: script action data
        :param task: celery task object
        """
        LOG.debug("Initialize %s", self.__class__.__name__)
        self.windows = fact["os"]["family"] == "windows"
        self.action = common.types.ScriptAction(action)
        self.action.validate()
        self.task = task
        self._init_base_data()
        self._link_logs(self.stdout_log_path, self.stdout_symlink_path)
        self._link_logs(self.stderr_log_path, self.stderr_symlink_path)
        if self.action.repo.url:
            self.script_path = self._fetch_script_repo()
            self.base_name = os.path.basename(self.script_path)
        self._init_shebang()
        self._init_script_data()
        self._create_script_file()
        super(BaseScriptExecutor, self).__init__(action, task)

    def _init_script_data(self):
        """Define script name, code and local path and store it into object attributes"""
        # pylint: disable=access-member-before-definition
        script_name = self.base_name
        if self.script_url:
            script_name = script_name or os.path.basename(self.script_url)
        script_name = self.shebang.normalize_script_name(script_name)
        script_code = (
            self.shebang.normalize_script_code(self.script_code) if self.script_code else None
        )
        script_path = self.script_path
        if script_code and not script_path:
            script_path = os.path.join(self.bin_dir, sysutil.sanitize_filename(script_name))

        self.script_name = script_name
        self.script_code = script_code
        self.script_path = script_path

    def _init_shebang(self):
        """Define shebang object"""
        # pylint: disable=not-callable
        if self.windows:
            self.shebang_cls = WindowsShebang
        else:
            self.shebang_cls = Shebang
        if self.script_code:
            self.shebang = self.shebang_cls(code=self.script_code)
        elif self.script_path:
            self.rootfs_script_path = (
                os.path.join(self.rootfs_dir, self.script_path[1:])
                if not self.windows
                else self.script_path
            )
            if not os.path.exists(self.rootfs_script_path):
                raise exc.NotFoundError("Script not found {!r}".format(self.rootfs_script_path))
            self.shebang = self.shebang_cls(self.rootfs_script_path)
        else:
            msg = (
                'Task "execute" requires either "script_path" '
                'or "script_code" argument to be set.'
            )
            raise exc.ConstraintError(msg)

    def _get_script_code(self):
        """
        Download script(if need) from url
        :return: script code
        """
        script_url = self.action.url
        script_code = self.action.code
        if script_url:
            LOG.info("Fetching script code from url %s", script_url)
            resp = requests.get(script_url)
            if not resp.ok:
                msg = "Script URL {} returned HTTP code {}".format(script_url, resp.status_code)
                raise exc.NotFoundError(msg)
            return resp.text
        return script_code

    def _init_base_data(self):
        """
        Just handle action and task objects and store all needed data it into object attributes
        """
        self.script_path = self.action.path
        self.script_url = self.action.url
        if self.script_path:
            self.base_name = os.path.basename(self.script_path)
        elif self.script_url:
            self.base_name = self.script_url.split("/")[-1]
        else:
            self.base_name = self.action.name
        self.return_stdout = self.action.return_stdout
        self.return_stderr = self.action.return_stderr
        self.stdout_symlink_path = self.action.stdout_symlink_path
        self.stderr_symlink_path = self.action.stderr_symlink_path
        self.bin_dir = self.task.local_dir_ext.bin_dir
        self.stdout_log_path = os.path.join(self.task.local_dir_ext.log_dir, "stdout.log")
        self.stderr_log_path = os.path.join(self.task.local_dir_ext.log_dir, "stderr.log")
        self.cwd = self.action.cwd or os.getcwd()
        self.success_codes = self.action.success_codes or [0]
        self.script_code = self._get_script_code()
        self.timeout = self.action.timeout or agent.config.workflow_engine["global_timeout"]
        self.env = self.action.env
        self.user = self.action.run_as

    def _fetch_script_repo(self):
        """
        Fetch script repo from provided repository
        """
        self.action.repo.work_dir = os.path.join(self.task.local_dir_ext.data_dir, "repo")
        os.mkdir(self.action.repo.work_dir)

        # pylint: disable=no-value-for-parameter
        repo.sync(self.action.repo)
        relative_path = self.action.relative_path
        relative_path = relative_path[1:] if relative_path.startswith("/") else relative_path
        relative_path = relative_path.replace("/", "\\") if self.windows else relative_path
        script_path = os.path.join(self.action.repo.work_dir, relative_path)
        if not os.path.exists(script_path):
            raise exc.NotFoundError(
                "Requested script %s is not found in repo %s"
                % (self.action.relative_path, self.action.repo.url)
            )
        return script_path

    def _create_script_file(self):
        """
        Create script file if needed
        """
        if self.script_code and not os.path.exists(self.script_path):
            LOG.debug(
                "Writing script '%s' into %s (%s bytes)",
                self.script_name,
                self.script_path,
                len(self.script_code),
            )
            with open(self.script_path, "wb+") as fp:
                fp.write(self.script_code.encode("utf-8"))
        if not self.windows and any([self.script_code, self.action.repo.url]):
            path = self.rootfs_script_path or self.script_path
            os.chmod(path, 0o755)
            if self.user:
                linuxutil.chown_r(path, self.user)

    def _prepare_script_result(self):
        """
        Prepare result structure
        :return: common.types.ExecutedScript object
        """
        stdout = sysutil.tail(self.stdout_log_path) if self.return_stdout else ""
        stderr = sysutil.tail(self.stderr_log_path) if self.return_stderr else ""

        result_data = dict(
            interpreter=self.shebang.interpreter,
            path=self.script_path,
            run_as=self.user,
            timeout=self.timeout,
            success_codes=self.success_codes,
            return_code=self.return_code,
            stdout=stdout,
            stdout_total_bytes=os.path.getsize(self.stdout_log_path),
            stderr=stderr,
            stderr_total_bytes=os.path.getsize(self.stderr_log_path),
            stdout_log_path=self.stdout_log_path,
            stderr_log_path=self.stderr_log_path,
            time_elapsed=max(monotonic.monotonic() - self.time0, 0),
        )
        return common.types.ExecutedScript(result_data)

    def _get_result(self):
        """
        Return result (only if the script was executed)
        or raise error if script execution was failed
        :return: common.types.ExecutedScript or raise error
        """
        if self.time0:
            result_data = self._prepare_script_result()
            if self.raise_exc:
                raise self.exc_cls(self.exc_message, **result_data.to_primitive())
            else:
                return result_data

    def run_or_supervise(self):
        """
        Main entry to run or supervise execution process
        :return: common.types.ExecutedScript or raise error
        """
        try:
            if self._metadata:
                LOG.info("Supervising script '%s' executed at %s", self.script_name, self.time0)
                self._supervise()
            else:
                LOG.info("Executing script '%s'", self.script_name)
                self._run()
        except Exception as e:
            self.raise_exc = True
            if isinstance(e, exc.TimeoutError):
                self.exc_cls = ExecuteTimeoutExpired
                self.return_code = ExecuteTimeoutExpired.RETURN_CODE
                LOG.warning(
                    "Executing script '%s' timeout expired (%s seconds).",
                    self.script_name,
                    self.timeout,
                )

        return self._get_result()

    def _run(self):
        raise NotImplementedError()

    def _supervise(self):
        raise NotImplementedError()


class SimpleScriptExecutor(BaseScriptExecutor):
    """Implements Script Execution Process for non systemd os"""

    def __init__(self, action, task):
        super(SimpleScriptExecutor, self).__init__(action, task)
        self.popen_kwargs = {"cwd": self.cwd}
        self.popen_process = None
        self.psutil_process = None
        self.popen_kwargs["env"] = self.env
        self.popen_args = self.shebang.build_args(self.script_path, self.user)
        self.process = None

    def _wait_script_result(self, timeout: int) -> None:
        """
        Wait for process to terminate and update :var:`self.return_code`.

        If *timeout* (in seconds) is specified and process is still
        alive raise `psutil.psutil.TimeoutExpired` or :class:`subprocess2.TimeoutExpired`.

        :raises :class:`psutil.TimeoutExpired`: When timeout expired (when `self._metadata` exists).
        :raises :class:`subprocess2.TimeoutExpired`: When timeout expired (when no metadata).
        """
        if self._metadata:
            self.psutil_process.wait(timeout)
        else:
            self.return_code = self.popen_process.run(timeout=timeout).returncode

    def _prepare_script_result(self):
        result_data = super(SimpleScriptExecutor, self)._prepare_script_result()
        result_data.command = self.popen_args[0] if self.popen_args else None
        return result_data

    def _run(self):
        try:
            self.time0 = monotonic.monotonic()
            with open(self.stdout_log_path, "a+") as fp_out:
                with open(self.stderr_log_path, "a+") as fp_err:
                    self.popen_kwargs.update({"stdout": fp_out, "stderr": fp_err})
                    self.popen_process = subprocess2.Popen(*self.popen_args, **self.popen_kwargs)
                    self._write_execution_metadata(
                        {"pid": self.popen_process.pid, "time0": self.time0}
                    )
                    self._wait_script_result(self.timeout)
        except (subprocess2.TimeoutExpired, celery.exceptions.SoftTimeLimitExceeded):
            raise exc.TimeoutError

        # pylint: disable=unsupported-membership-test
        if self.return_code not in self.success_codes:
            raise ExecuteError()

    def _supervise(self):
        try:
            self.time0 = self._metadata["time0"]
            self.return_code = None
            timeout = max(self.timeout - (monotonic.monotonic() - self.time0), 0)
            self.psutil_process = psutil.Process(self._metadata["pid"])
            self.psutil_process.wait(timeout)
        except psutil.TimeoutExpired:
            os.kill(self._metadata["pid"], 9)
            raise exc.TimeoutError
        except psutil.NoSuchProcess:
            LOG.error("Executing script '%s' NoSuchProcess.", self.script_name)


class SystemdScriptExecutor(BaseScriptExecutor):
    """
    Implements Script Execution Process using Systemd service file for run and supervise stages.
    """

    rootfs_dir = "/rootfs"
    _service_status = None
    # run systemd script from privileged user
    systemd_user = "root"

    def __init__(self, action, task):
        super(SystemdScriptExecutor, self).__init__(action, task)
        self.unit = configutil.UnitFile()
        self.service_file_folder = os.path.join(self.rootfs_dir, "etc/systemd/system")
        self.service_name = ".".join(
            ("scalr-agent", self.task.local_dir_ext.task_id, self.base_name, "service")
        )
        self.service_file_path = posixpath.join(self.service_file_folder, self.service_name)
        command = " ".join(self.shebang.build_args(self.script_path, self.user)[0])
        self.exec_string = "/bin/bash -c '{} 2> {} 1> {}'".format(
            command, self.stderr_log_path, self.stdout_log_path
        )

    def _prepare_unit(self):
        """
        Configure configutil.UnitFile object that represent, complete structure of service file
        """
        self.unit.set_description(
            "Scalr orchestration script '{name}' execution (task: {id})".format(
                name=self.script_name, id=self.task.local_dir_ext.task_id
            )
        )
        self.unit.set_type("oneshot")
        self.unit.set_remain_after_exit("true")
        self.unit.set_user(self.systemd_user)
        self.unit.set_start_timeout(self.timeout)
        self.unit.set_success_exit_status(self.success_codes)
        self.unit.set_start_exec(self.exec_string)
        for name, value in self.env.items():
            self.unit.set_env_variable(name.encode("utf-8"), value.encode("utf-8"))

    def _write_unit_file(self):
        with open(self.service_file_path, "w") as service_file:
            self.unit.write(service_file)

    def _purge_service(self):
        if os.path.exists(self.service_file_path):
            linuxutil.systemctl_stop(self.service_name)
            os.remove(self.service_file_path)
            linuxutil.systemctl_daemon_reload()
            LOG.debug("Clear %s entry.", self.service_file_path)

    def _wait_script_result(self, timeout=0):
        check_interval = 10
        while self._get_status(cached=False) == "running" and timeout > 0:
            time.sleep(min(check_interval, timeout))
            timeout -= check_interval
            LOG.debug("Script %s timeout wait left %s sec", self.script_name, timeout)
        self.return_code = self._service_status.exit_code

    def _prepare_script_result(self):
        result_data = super(SystemdScriptExecutor, self)._prepare_script_result()
        result_data.command = [[self.exec_string]]
        return result_data

    def _run(self):
        try:
            self.time0 = monotonic.monotonic()
            self._write_execution_metadata({"time0": self.time0})
            self._prepare_unit()
            self._write_unit_file()
            linuxutil.systemctl_daemon_reload()
            linuxutil.systemctl_start(self.service_name)
        except subprocess2.CalledProcessError:
            if self._get_status() == "timeout_killed":
                raise exc.TimeoutError
            elif self._get_status() == "failed":
                raise ExecuteError
        finally:
            self._wait_script_result()
            self._purge_service()

    def _supervise(self):
        self.time0 = self._metadata["time0"]
        timeout = max(self.timeout - (monotonic.monotonic() - self.time0), 0)
        self._wait_script_result(timeout)
        self._purge_service()
        if self._get_status() == "timeout_killed":
            raise exc.TimeoutError
        elif self._get_status() == "failed":
            raise ExecuteError

    def _get_status(self, cached=True):
        try:
            if cached:
                self._service_status = self._service_status or linuxutil.systemctl_status(
                    self.service_name
                )
            else:
                self._service_status = linuxutil.systemctl_status(self.service_name)
            if (
                self._service_status.active_state == "activating"
                and self._service_status.sub_state == "start"
            ):
                return "running"
            if self._service_status.result == "timeout":
                return "timeout_killed"
            if (
                self._service_status.active_state == "inactive"
                and self._service_status.sub_state == "dead"
            ):
                self._service_status.exit_code = 1
                return "failed"
            # pylint: disable=unsupported-membership-test
            if (
                self._service_status.result == "exit-code"
                and self._service_status.exit_code not in self.success_codes
            ):
                return "failed"
        except subprocess2.CalledProcessError:
            LOG.warning("Couldn't get status of %s", self.service_name)
        return "unknown"


@celeryutil.agent_task(bind=True, name="execute.script")
def script(self, action, **kwargs):
    """
    Entrypoint for script execution logic
    :param self: celery task object
    :param action: script action data
    :return: result of script execution
    """
    if fact["manage_container_host"]:
        executor_cls = SystemdScriptExecutor
    else:
        executor_cls = SimpleScriptExecutor

    executor = executor_cls(action, self)
    result = executor.run_or_supervise()
    return result


LOG_TYPES = ("execute.script", "ansibletower.job")


@celeryutil.agent_task(name="execute.get_logs")
def get_logs(task_id, **kwargs):
    """
    Retrieve logs for specified execute.script task

    :keyword task_id: Identifier of task
    :type task_id: str

    :rtype: :class:`common.types.ScriptLogs`
    """
    local_dir_ext = celeryutil.TaskLocalDir()
    local_dir_ext.task_id = task_id
    logs = common.types.ScriptLogs()

    def get_single_log(stream_name):
        log_path = os.path.join(local_dir_ext.log_dir, stream_name + ".log")
        LOG.debug("Fetching log file: %s", log_path)
        if os.path.exists(log_path):
            message = six.u("Log file truncated. See the full log in ")
            retval = []
            if os.path.getsize(log_path) > agent.config.workflow_engine["tail_num_bytes"]:
                retval += [message, log_path, "\n"]
            retval += sysutil.tail(log_path)
            return "".join(retval)

    for name in LOG_TYPES:
        if local_dir_ext.set_current_func(name):
            logs.stdout = get_single_log("stdout")
            logs.stderr = get_single_log("stderr")
            break

    return logs
