"""
NTFS filesystem implementation.
"""
from __future__ import absolute_import

import typing as tp

import agent.tasks.filesystem.tasks as filesystem
from common import types
from common.utils import logutil
from common.utils import subprocess2
from common.utils.facts import fact

if fact["os"]["family"] == "windows":
    from common.utils import winutil  # pragma: no cover
else:
    winutil = None


LOG = logutil.get_logger(__name__)


class NTFSFileSystem(filesystem.ABCFileSystem):
    """
    NT file system management.
    """

    freezable = False
    extendable = True
    umount_on_resize = False

    def mkfs(self, device: tp.Union[types.SCSIAddress, str], *args, **kwargs):
        """
        Build a NT File System on a device.

        :param device: SCSI address of the ``Win32_DiskDrive``.
        """
        if isinstance(device, str):
            device = types.SCSIAddress.from_packed_address(device)
        disk_index = winutil.get_disk_drive_index(device)
        LOG.info(
            "Creating filesystem on the Win32_DiskDrive (index=%s, scsi_address=%s)",
            disk_index,
            device,
        )
        commands = """select disk {}
               online disk noerr
               attributes disk clear readonly noerr
               clean
               convert basic noerr
               convert gpt noerr
               create partition primary
               format fs=ntfs quick
               exit
            """.format(
            disk_index
        )
        try:
            output = winutil.call_diskpart(stdin=commands).stdout
        except subprocess2.CalledProcessError as e:
            raise filesystem.MkfsError(device, e)

        # pylint: disable=unsupported-membership-test
        if output and "The disk you specified is not valid" in output:
            raise filesystem.MkfsError(disk_index, "Diskpart can't find disk")
        if output and "The volume you selected is not valid or does not exist" in output:
            raise filesystem.MkfsError(disk_index, "Error occurred while creating volume")
        # pylint: enable=unsupported-membership-test

    def extend(self, device: tp.Union[types.SCSIAddress, str], *args, **kwargs):
        """
        Extends FS on the device. It assumes the disk has only 1 partition

        :param device: SCSI address of the ``Win32_DiskDrive``.
        """
        if isinstance(device, str):
            device = types.SCSIAddress.from_packed_address(device)
        disk_index = winutil.get_disk_drive_index(device)
        part_index = winutil.get_partition_index(disk_index)
        LOG.info(
            "Extending filesystem on the Win32_DiskDrive (index=%s, scsi_address=%s)",
            disk_index,
            device,
        )
        commands = """
            rescan
            select disk {}
            select partition {}
            extend
            exit
            """.format(
            disk_index, part_index
        )
        try:
            output = winutil.call_diskpart(stdin=commands).stdout
        except subprocess2.CalledProcessError as e:
            raise filesystem.ExtendfsError(device, e)

        # pylint: disable=unsupported-membership-test
        if output and "The disk you specified is not valid" in output:
            raise filesystem.ExtendfsError(disk_index, "Diskpart can't find disk")
        if output and "The volume you selected is not valid or does not exist" in output:
            raise filesystem.ExtendfsError(disk_index, "Error occurred while extending volume")

    def set_label(self, device, label):
        """
        Set label for the ``Win32_LogicalDisk``.

        :param device: SCSI address of the ``Win32_LogicalDisk``.
        :type device: :class:`types.SCSIAddress`

        :param label: The label for the ``Win32_LogicalDisk``.
        :type label: str
        """
        return winutil.set_volume_label(device, label)

    def get_label(self, device):
        """
        Get label from the ``Win32_LogicalDisk``.

        :param device: SCSI address of the ``Win32_LogicalDisk``.
        :type device: :class:`types.SCSIAddress`

        :return: The label of the ``Win32_LogicalDisk``.
        :rtype: str
        """
        return winutil.get_volume_label(device)
