"""
Filesystem tasks.
"""
from __future__ import absolute_import

import abc
import os
import sys

from agent.tasks import pkgmgr
from agent.utils import celeryutil
from common.utils import exc
from common.utils import imputil
from common.utils import logutil
from common.utils import six
from common.utils import subprocess2
from common.utils.facts import fact

if fact["os"]["family"] != "windows":
    from common.utils import linuxutil
else:
    linuxutil = None  # pragma: no cover


__all__ = [
    "make",
    "extend",
    "FileSystem",
    "ABCFileSystem",
    "LinuxFileSystem",
    "Error",
    "MkfsError",
    "FsckError",
    "ExtendfsError",
    "LabelSetError",
    "LabelGetError",
    "FreezeError",
    "UnfreezeError",
    "MountRequiredError",
]


LOG = logutil.get_logger(__name__)

providers = {
    "ext3": "agent.tasks.filesystem.ext.Ext3FileSystem",
    "ext4": "agent.tasks.filesystem.ext.Ext4FileSystem",
    "xfs": "agent.tasks.filesystem.xfs.XfsFileSystem",
    "ntfs": "agent.tasks.filesystem.ntfs.NTFSFileSystem",
}


class Error(exc.Error):
    """
    Exception class for filesystem tasks.
    """


def new_exc(name, message, **kwds):
    """
    Factory function to create filesystem exception classes.
    """

    def __init__(self, device, *args):
        super(type(self), self).__init__(message, device, *args)

    kwds["__init__"] = __init__
    return type(name, (Error,), kwds)


MkfsCmd = subprocess2.new_command("MkfsCmd", binary="/sbin/mkfs")

MkfsError = new_exc("MkfsError", "Error during filesystem creation on device")
FsckError = new_exc("FsckError", "Error during filesystem check on device")
ExtendfsError = new_exc("ExtendfsError", "Error during filesystem extend")
LabelSetError = new_exc("LabelSetError", "Error while setting label for device")
LabelGetError = new_exc("LabelGetError", "Error while getting label for device")
FreezeError = new_exc("FreezeError", "Error during filesystem freeze on device")
UnfreezeError = new_exc("UnfreezeError", "Error during filesystem un-freeze on device")
MountRequiredError = new_exc(
    "MountRequiredError", "Device should be mounted to perform this operation"
)


def FileSystem(fstype, *args, **kwds):  # pylint: disable=invalid-name
    """
    Factory function to create filesystem provider.
    """
    try:
        imp_name = providers[fstype]
    except KeyError:
        raise exc.NotSupportedError("No filesystem class for name {0!r}".format(fstype))
    return imputil.import_object(imp_name, fstype, *args, **kwds)


@six.add_metaclass(abc.ABCMeta)
class ABCFileSystem(object):
    """
    Abstract class for filesystem implementations.
    """

    freezable = False
    extendable = False
    umount_on_resize = False

    name = None
    packages = []

    def __init__(self, name, *args, **kwargs):
        self.name = name

    @abc.abstractmethod
    def mkfs(self, device, *args, **kwargs):
        """
        Creates filesystem on the device.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set_label(self, device, label):
        """
        Returns label for the device.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_label(self, device):
        """
        Applies label from the device.
        """
        raise NotImplementedError()

    def extend(self, device, *args, **kwargs):
        """
        Extends filesystem on given device partition size.
        """
        if not self.extendable:
            raise exc.NotSupportedError()
        raise NotImplementedError()

    def freeze(self, device):
        """
        Freezes a filesystem on the device.
        """
        if not self.freezable:
            raise exc.NotSupportedError()
        raise NotImplementedError()

    def unfreeze(self, device):
        """
        Unfreezes a filesystem on the device.
        """
        if not self.freezable:
            raise exc.NotSupportedError()
        raise NotImplementedError()


@six.add_metaclass(abc.ABCMeta)
class LinuxFileSystem(ABCFileSystem):
    """
    Base class for linux filesystem implementations.
    """

    def __init__(self, name, *args, **kwargs):
        super(LinuxFileSystem, self).__init__(name, *args, **kwargs)
        mkfs_exec = "{}.{}".format(MkfsCmd.binary, name)
        if not os.path.exists(mkfs_exec):
            linuxutil.modprobe(name)
            if self.packages:
                for package in self.packages:
                    pkgmgr.install(package)
            if not os.path.exists(mkfs_exec):
                raise Error(
                    (
                        "The operation on the filesystem {} failed because '{}' binary was not found."
                    ).format(name, mkfs_exec)
                )

    def mkfs(self, device, *args, **kwargs):
        cmd = MkfsCmd()
        cmd.args("-t", self.name)
        cmd.args(*args, **kwargs)
        cmd.args(device)
        try:
            # TODO: progress callback
            LOG.info("Creating filesystem on %s.", device)
            return cmd.check_output()
        except subprocess2.CalledProcessError as e:
            raise MkfsError(device, e)

    def set_label(self, device, label):
        """
        Returns label for the linux device.
        """
        raise NotImplementedError()

    def get_label(self, device):
        """
        Applies label from the linux device.
        """
        raise NotImplementedError()


@celeryutil.agent_task(name="filesystem.make")
def make(fstype, device, *args, **kwargs):
    """
    Creates filesystem on the given device.

    :param fstype: The type of the file system (example: ``ext4``, ``ntfs``)
    :type: str

    :param device: The path to device (e.g: /dev/sda) or :class:`types.SCSIAddress` (for windows).
    :type device: str or :class:`types.SCSIAddress`
    """
    # TODO: maybe move code from storage, that checks for already created filesystem
    # TODO: maybe add force flag to ignore any checks
    try:
        return FileSystem(fstype).mkfs(device, *args, **kwargs)
    except MkfsError as e:
        msg = "Error creating filesystem (device={} args={} kwargs={}). " "Reason: {}".format(
            device, args, kwargs, e
        )
        six.reraise(Error, Error(msg), sys.exc_info()[2])


@celeryutil.agent_task(name="filesystem.extend")
def extend(fstype, device, *args, **kwds):
    """Extends a logical volume to take all the free space on the device."""
    try:
        return FileSystem(fstype).extend(device, *args, **kwds)
    except Error as e:
        msg = "Error resizing filesystem (device={} args={} kwargs={}). " "Reason: {}".format(
            device, args, kwds, e
        )
        six.reraise(Error, Error(msg), sys.exc_info()[2])
