"""
Managing firewalld.service
"""
from __future__ import absolute_import

import shlex

import agent.tasks.firewall.tasks as firewall
from agent.tasks import service
from agent.tasks.firewall import iptables
from common.utils import exc
from common.utils import logutil
from common.utils import subprocess2

LOG = logutil.get_logger(__name__)


class FirewallCmd(subprocess2.Command):
    """
    `firewall-cmd` command wrapper.
    """

    binary = subprocess2.which("firewall-cmd")

    @classmethod
    def validate(cls):
        """
        Check if binary exists.

        :raises :class:`exc.NotFoundError`
        """
        if not cls.binary:
            raise exc.NotFoundError(
                "firewall-cmd binary not found in PATH environment "
                "variable, check that `firewalld` package is installed."
            )


class FirewallDProvider(firewall.ABCFirewall):
    """
    Provides methods to manage firewalld.service.
    """

    protocol = "ipv4"
    """
    Choose: {ipv4|ipv6|eb}.
    """

    table = "filter"
    """
    Specifying which iptables table is to be used.
    """

    def __init__(self):
        FirewallCmd.validate()
        self.service = service.Service("firewalld")

    def enable(self):
        """
        Enable firewalld service.
        """
        return self.service.enable()

    def enabled(self, *args, **kwargs):
        """
        Check that firewalld service is enabled.
        """
        return self.service.enabled

    def disable(self):
        """
        Disable firewalld.service.
        """
        return self.service.disable()

    def ensure(self, rule, append=False):
        """
        Add firewalld rule.

        :type rule: :class:`common.types.data.FirewallRule`
        :param rule: firewall rule

        :type append: bool
        :param append: add rule to end of table
        """
        super(FirewallDProvider, self).ensure(rule)
        rule_priority = len(self._get_all_rules()) if append else 0
        rule_chain, rule_kwargs = iptables.rule_to_iptables_kwargs(rule)

        if rule not in self.list_rules():
            # create permanent and runtime rule
            self._add_rule(rule_chain, rule_priority, rule_kwargs, permanent=True)
            self._add_rule(rule_chain, rule_priority, rule_kwargs)
            LOG.debug("Firewall rule '%s' was added.", rule.comment)
        else:
            LOG.debug("Firewall rule '%s' already exists.", rule.comment)

    def remove(self, rule):
        """
        Remove firewalld rule.

        :type rule: :class:`common.types.data.FirewallRule`
        :param rule: firewall rule
        """
        super(FirewallDProvider, self).remove(rule)
        for rule_args in self._get_all_rules():
            if rule_args[:2] != [self.protocol, self.table]:
                continue
            existing_rule = self._parse_rule(rule_args)
            if rule == existing_rule:
                # remove permanent and runtime rule
                self._remove_rule(rule_args, permanent=True)
                self._remove_rule(rule_args)
                LOG.debug("Firewall rule '%s' has been removed.", existing_rule.comment)
                return
        LOG.debug("Firewall rule '%s' was not found.", rule.comment)

    def list_rules(self):
        """
        Get all rules.

        :rtype: list of :class:`common.types.data.FirewallRule`
        """
        rules = []
        for rule_args in self._get_all_rules():
            rule = self._parse_rule(rule_args)
            if rule:
                rules.append(rule)
        return rules

    def _add_rule(self, chain, rule_priority, rule_kwargs, permanent=False):
        """
        Add runtime or permanent firewalld rule.
        """
        add_cmd = self._new_direct_cmd(permanent=permanent, add_rule=True)
        add_cmd.args(self.protocol, self.table, chain, rule_priority)
        # XXX Remove on python>=3.6, use `add_cmd.args(**rule_kwargs)` instead.
        # Reason: https://www.python.org/dev/peps/pep-0468/
        # pylint: disable=protected-access
        add_cmd._format_args(long_args=rule_kwargs)
        add_cmd.check_output()

    def _remove_rule(self, rule_args, permanent=False):
        """
        Remove runtime or permanent firewalld rule.
        """
        remove_cmd = self._new_direct_cmd(permanent=permanent, remove_rule=True)
        remove_cmd.args(*rule_args)
        remove_cmd.check_output()

    def _get_all_rules(self):
        """
        Return all firewalld permanent rules.

        :return: firewalld rules, rule format: [<protocol> <table> <chain>
                 <priority> *<iptables_args>]
        :rtype list of list[str]
        """
        list_cmd = self._new_direct_cmd(get_all_rules=True, permanent=True)
        output = list_cmd.check_output() or ""
        return [shlex.split(line) for line in output.strip().split("\n") if line.strip()]

    @staticmethod
    def _new_direct_cmd(*args, **kwargs):
        """
        The direct options give a more direct access to the firewall.
        """
        cmd = FirewallCmd()
        if kwargs.pop("permanent", False):
            cmd.args(permanent=True)
        return cmd.args(direct=True).args(*args, **kwargs)

    @staticmethod
    def _parse_rule(args):
        """
        Create :class:`common.types.data.FirewallRule` object
        from a firewalld rule record.

        :type args: list[str]
        :param args: firewalld rule arguments, format: [<protocol>
                     <table> <chain> <priority> *<iptables_args>]

        :rtype :class:`common.types.data.FirewallRule`
        """
        _, _, chain, _ = args[:4]  # <protocol> <table> <chain> <priority>
        iptables_args = args[4:]  # <iptables_args>
        try:
            return iptables.iptables_kwargs_to_rule(chain, iptables_args)
        except Exception as e:
            LOG.debug("Parser skipped rule `%s`. Cause: %s", iptables_args, e)
