"""
Managing iptables.service
"""
from __future__ import absolute_import

import argparse
import collections
import functools
import os
import shlex

from agent.tasks import service
from agent.tasks.firewall import tasks as firewall
from common import types
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import subprocess2

__all__ = ["IptablesFirewall", "rule_to_iptables_kwargs", "iptables_kwargs_to_rule"]

LOG = logutil.get_logger(__name__)

ACTION_TO_POLICY = {
    "allow": "ACCEPT",
    "deny": "DROP",
}

DIRECTION_TO_CHAIN = {
    "in": "INPUT",
    "out": "OUTPUT",
}

LOCKED_IPTABLES_RETRY_INTERVAL = 1
LOCKED_IPTABLES_RETRY_TIMES = 3


def rule_to_iptables_kwargs(rule):
    """
    Create iptables arguments from
    :class:`common.types.data.FirewallRule` object.

    :type rule: :class:`common.types.data.FirewallRule`
    :param rule: firewall rule

    :rtype: (str, :class:`collections.OrderedDict`)
    :return: chain and dict of keyword arguments.
    """
    values = [
        ("match", "comment"),
        ("comment", rule.comment),
        ("source", rule.source),
        ("protocol", rule.protocol),
        ("jump", ACTION_TO_POLICY[rule.action]),
    ]
    if rule.from_port != rule.to_port:
        dport = "{}:{}".format(rule.from_port, rule.to_port)
    else:
        dport = str(rule.from_port)
    values.append(("dport", dport))

    chain = DIRECTION_TO_CHAIN[rule.direction]
    args = collections.OrderedDict(values)
    return chain, args


def iptables_kwargs_to_rule(chain, args):
    """
    Create :class:`common.types.data.FirewallRule` object from a
    iptables arguments, e.g: "-A INPUT -p tcp -m tcp --dport 80 -j ACCEPT"
    All unknown arguments will be ignored.

    :type chain: str
    :param chain: iptables chain

    :type args: list[str]
    :param args: iptables arguments

    :rtype: :class:`common.types.data.FirewallRule`
    :return: firewall rule
    """
    # setup rule parser
    values = (
        ("--comment",),
        ("--source", "-s"),
        ("--protocol", "-p"),
        ("--jump", "-j"),
        ("--dport",),
    )
    parser = argparse.ArgumentParser()
    for value in values:
        parser.add_argument(*value)

    # parse iptables arguments to dict
    try:
        known_args, _ = parser.parse_known_args(args)
    except:
        raise exc.Error("Could not parse arguments")

    # validate parsed data
    rule_data = {
        "comment": known_args.comment or "",
        "source": known_args.source or "0.0.0.0/0",
        "protocol": known_args.protocol or "all",
    }

    # validate rule chain
    try:
        rule_data["direction"] = mixutil.inverted_dict(DIRECTION_TO_CHAIN)[chain]
    except KeyError:
        raise exc.Error("Chain '{}' is not supported by Scalr agent.".format(chain))

    # validate rule target
    if not known_args.jump:
        raise exc.Error("No target specified.")
    try:
        rule_data["action"] = mixutil.inverted_dict(ACTION_TO_POLICY)[known_args.jump]
    except KeyError:
        raise exc.Error("Target '{}' is not supported by Scalr agent.".format(known_args.jump))

    # validate port range
    if known_args.dport is None:
        from_port, to_port = 0, 65535
    elif ":" in known_args.dport:
        from_port, to_port = known_args.dport.split(":", 1)
    else:
        from_port = to_port = known_args.dport
    rule_data["from_port"], rule_data["to_port"] = from_port, to_port

    # create :class:`common.types.data.FirewallRule` object from parsed data
    rule = types.data.FirewallRule(rule_data)
    rule.validate()
    return rule


retry_locked_iptables = functools.partial(
    mixutil.retries_error,
    except_expr=subprocess2.CalledProcessError,
    condition_func=lambda e: e.returncode != 0,
    interval=LOCKED_IPTABLES_RETRY_INTERVAL,
    times=LOCKED_IPTABLES_RETRY_TIMES,
)


class IptablesCmd(subprocess2.Command):
    """
    `iptables` command wrapper.
    """

    binary = subprocess2.which("iptables")

    def args(self, *args, **long_args):
        parent_args = super(IptablesCmd, self).args
        for key in ("protocol", "match"):
            if key in long_args:
                parent_args(**{key: long_args.pop(key)})
        return parent_args(*args, **long_args)


class IptablesFirewall(firewall.ABCFirewall):
    """
    Provides methods to manage iptables service.
    """

    def __init__(self):
        self.service = service.Service("iptables")

    def enabled(self, *args, **kwargs):
        """
        Check that iptables service is enabled.

        :rtype: bool
        """
        return os.access(IptablesCmd.binary, os.X_OK) if IptablesCmd.binary else False

    def enable(self):
        """
        Enables iptables service.
        """
        self.service.enable()

    def disable(self):
        """
        Disables iptables service.
        """
        self.service.disable()

    @retry_locked_iptables()
    def ensure(self, rule, append=False):
        """
        Add iptables rule.

        :param rule: firewall rule
        :type rule: :py:class:`common.types.data.FirewallRule`
        :param bool append: add rule to end of chain
        """
        super(IptablesFirewall, self).ensure(rule)
        rule_chain, rule_kwargs = rule_to_iptables_kwargs(rule)
        existing_rules = self._get_chain_rules(rule_chain)

        if rule not in existing_rules:
            rule_kwargs["append" if append else "insert"] = rule_chain
            add_cmd = IptablesCmd().args(**rule_kwargs)
            add_cmd.check_output()
            LOG.debug("Firewall rule '%s' was added.", rule.comment)
        else:
            LOG.debug("Firewall rule '%s' already exists.", rule.comment)

    @retry_locked_iptables()
    def remove(self, rule):
        """
        Remove iptables rule.

        :param rule: firewall rule
        :type rule: :py:class:`common.types.data.FirewallRule`
        """
        super(IptablesFirewall, self).ensure(rule)
        rule_chain, rule_kwargs = rule_to_iptables_kwargs(rule)
        existing_rules = self._get_chain_rules(rule_chain)

        for existing_rule in existing_rules:
            if rule == existing_rule:
                rule_kwargs["comment"] = existing_rule.comment
                remove_cmd = IptablesCmd().args(delete=rule_chain).args(**rule_kwargs)
                remove_cmd.check_output()
                LOG.debug("Firewall rule '%s' has been removed.", existing_rule.comment)
                return
        LOG.debug("Firewall rule '%s' was not found.", rule.comment)

    @retry_locked_iptables()
    def list_rules(self):
        """
        Get all valid rules.

        :rtype: list of :class:`common.types.data.FirewallRule`
        """
        chains = DIRECTION_TO_CHAIN.values()
        return sum([self._get_chain_rules(chain) for chain in chains], [])

    @staticmethod
    def _get_chain_rules(chain_name):
        """
        Return iptables rules for specified chain.

        :rtype: list of :class:`common.types.data.FirewallRule`
        """
        list_cmd = IptablesCmd().args(list_rules=True).args(chain_name)
        stdout = list_cmd.check_output() or ""
        rules = []
        for line in stdout.strip().splitlines():
            args = shlex.split(line)
            if "-P" in args or "-N" in args:
                continue  # dull rules
            try:
                rule = iptables_kwargs_to_rule(chain_name, args)
            except Exception as e:
                LOG.debug("Parser skipped rule `%s`. Cause: %s", args, e)
            else:
                rules.append(rule)
        return rules
