from __future__ import absolute_import

import abc

from agent.tasks import service
from agent.utils import celeryutil
from common import types
from common.utils import exc
from common.utils import imputil
from common.utils import logutil
from common.utils import six
from common.utils.facts import fact

__all__ = [
    "Error",
    "Firewall",
    "ABCFirewall",
    "enable",
    "enabled",
    "disable",
    "ensure",
    "remove",
    "list_rules",
]


LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    pass


def _is_firewalld_running():
    """
    Checks firewalld is installed and service is running.

    :rtype bool
    """
    try:
        return service.Service("firewalld").running
    except service.NotInstalledError:
        return False


def Firewall(*args, **kwds):  # pylint: disable=invalid-name
    if fact["os"]["family"] == "windows":
        imp_name = "windows_firewall.WindowsFirewall"
    elif _is_firewalld_running():
        imp_name = "firewalld.FirewallDProvider"
    else:
        imp_name = "iptables.IptablesFirewall"
    imp_name = "agent.tasks.firewall." + imp_name
    cls = imputil.import_object(imp_name, *args, **kwds)
    return cls


def instance():
    # pylint: disable=protected-access
    if not getattr(instance, "_instance", None):
        instance._instance = Firewall()
    return instance._instance


@six.add_metaclass(abc.ABCMeta)
class ABCFirewall(object):
    @abc.abstractmethod
    def enable(self):
        """
        Enable platform firewall.
        """

    @abc.abstractmethod
    def enabled(self, *args, **kwargs):
        """
        Returns True if firewall is enabled in OS.
        """

    @abc.abstractmethod
    def disable(self):
        """
        Disable platform firewall.
        """

    @abc.abstractmethod
    def ensure(self, rule, append=False):
        """
        Insert or append rule.

        :type rule: :py:class:`common.types.data.FirewallRule`
        :param append: append rule to end of table.
        """
        rule.validate()

    @abc.abstractmethod
    def remove(self, rule):
        """
        Removes rule from the default chain.
        """
        rule.validate()

    @abc.abstractmethod
    def list_rules(self):
        """
        Lists all rules that are currently applied to the default (INPUT) chain.

        :returns: current state of the default chain, i.e. parsed output of 'iptables -nL'.
        :rtype: dict
        """


@celeryutil.agent_task(name="firewall.enable")
def enable(**kwargs):
    instance().enable()


@celeryutil.agent_task(name="firewall.enabled")
def enabled(**kwargs):
    return instance().enabled()


@celeryutil.agent_task(name="firewall.disable")
def disable(**kwargs):
    instance().disable()


@celeryutil.agent_task(name="firewall.ensure")
def ensure(rule, append=False, **kwargs):
    """
    :param rule: :py:class:`common.types.data.FirewallRule` parameters.
    :type rule: dict

    :param append: append rule to end of table.
    """
    rule = types.FirewallRule(rule)
    LOG.info("Opening firewall ports %s-%s", rule.from_port, rule.to_port)
    instance().ensure(rule, append=append)


@celeryutil.agent_task(name="firewall.remove")
def remove(rule, **kwargs):
    """
    :param rule: :py:class:`common.types.data.FirewallRule` parameters.
    :type rule: dict
    """
    instance().remove(types.FirewallRule(rule))


@celeryutil.agent_task(name="firewall.list_rules")
def list_rules(**kwargs):
    """
    :rtype list
    """
    return instance().list_rules()
