"""
Module that provides class for managing firewall on Windows
"""
from __future__ import absolute_import

import sys

import agent.tasks.service.tasks as service
from agent.tasks.firewall import tasks as firewall
from agent.tasks.service import windows as windows_service
from common.utils import exc
from common.utils import six
from common.utils import subprocess2

# pylint: disable=invalid-name
NetshCmd = subprocess2.new_command("NetshCmd", binary="C:\\Windows\\system32\\netsh.exe")


class Error(exc.Error):
    """
    Subclass for this module's exceptions
    """


class WindowsFirewall(firewall.ABCFirewall):
    """
    Provides methods to manage firewall on Windows
    """

    def enabled(self, *args, **kwargs):
        # make sure that the mpssvc service is running
        win_service_provider = windows_service.WindowsServiceProvider()
        if win_service_provider.get_service_status("mpssvc") != service.ServiceStatus.running:
            return False

        # this will show status for public, private and domain firewall profiles
        cmd = NetshCmd().args("advfirewall", "show", "all", "state")
        out = cmd.check_output()
        profiles_enabled = [("State" in line and "ON" in line) for line in out.splitlines()]
        return any(profiles_enabled)

    def disable(self):
        raise NotImplementedError()

    def enable(self):
        raise NotImplementedError()

    def list_rules(self):
        raise NotImplementedError()

    def remove(self, rule):
        raise NotImplementedError()

    def ensure(self, rule, append=False):
        """
        Builds and executes command of a form netsh advfirewall
        firewall add rule action=allow dir=in localport=8008-8013 name=Scalr
        """
        super(WindowsFirewall, self).ensure(rule)

        assert not append, "WindowsFirewall.ensure() does not implement append=True"

        from_port = str(rule.from_port)
        to_port = str(rule.to_port or from_port)
        values = dict(
            dir=rule.direction,
            name=rule.comment,
            action=rule.action,
            protocol=rule.protocol,
            localport="-".join([from_port, to_port]),
        )

        cmd = NetshCmd()
        cmd.args("advfirewall", "firewall", "add", "rule")
        cmd.args(*map("=".join, values.items()))
        cmd.build()
        try:
            return cmd.check_call()
        except subprocess2.CalledProcessError as e:
            six.reraise(Error, Error(e), sys.exc_info()[2])
