"""
Tasks for making server snapshots on EC2.
"""
from __future__ import absolute_import

import os
import shutil
import time

from libcloud.compute.base import Node
from libcloud.compute.base import StorageVolume

from agent.tasks import misc
from agent.tasks.pkgmgr import tasks as pkgmgr_tasks
from agent.utils import celeryutil
from common import types
from common.utils import exc
from common.utils import libcloudutil
from common.utils import logutil
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

if fact["os"]["family"] != "windows":
    from common.utils import augeasutil
    from common.utils import linuxutil
    from agent.tasks.storage.utils import linuxutil as storage_linuxutil
else:
    linuxutil = None
    storage_linuxutil = None


__all__ = ["Error", "prepare_image_create", "sysprep"]


LOG = logutil.get_logger(__name__)


AMI_TOOLS_PACKAGE = "ec2-ami-tools-1.5.6"
AMI_TOOLS_PATCH_PATH = "./ami_tools.patch"
SCALR_RUBY_DIR = "/opt/scalr-ruby/embedded/"


class Error(exc.Error):
    pass


def _instance_store_prepare(credentials, image_creation, cloud_server):
    bundle_locals = types.Ec2BundleLocals(image_creation.bundle_creation.flatten())
    bundle_locals.validate(partial=True)

    bundle_locals.name = image_creation.name + "-" + time.strftime("%Y%m%d%H%M%S")
    bundle_locals.manifest_path = (
        os.path.join(bundle_locals.destination, bundle_locals.name) + ".manifest.xml"
    )

    for dev in linuxutil.df():
        if dev.mpoint == "/":
            bundle_locals.size = dev.size // 1024
            break
    else:
        raise Error("Can't find root volume")

    bundle_locals.environ = {
        "EC2_CERT": credentials.certificate_path,
        "EC2_PRIVATE_KEY": credentials.private_key_path,
        "EC2_USER_ID": credentials.account_id,
        "AWS_ACCESS_KEY": credentials.access_key,
    }

    return types.Ec2ImageCreationLocals(
        {
            "bundle": bundle_locals,
            "cloud_server": cloud_server,
            "image_creation": image_creation,
            "libcloud_credentials": credentials.to_libcloud(cloud_server.location),
        }
    )


@celeryutil.agent_task(name="ec2.prepare_image_create")
def prepare_image_create(credentials=None, image_creation=None, cloud_server=None, **kwargs):
    """Gathers necessary data for image creation.
    :param dict credentials: Cloud credentials.

    :param dict image_creation: Image creation settings.

    :param dict cloud_server: Cloud server information.
    """
    image_creation = types.Ec2ImageCreation(image_creation)
    image_creation.validate()
    cloud_server = types.CloudServer(cloud_server)
    cloud_server.validate()
    credentials = types.AwsCloudCredentials(credentials)
    credentials.validate()

    if image_creation.bundle_creation:
        return _instance_store_prepare(credentials, image_creation, cloud_server)

    libcloud_credentials = credentials.to_libcloud(cloud_server.location)

    block_device_mapping = []

    server = libcloudutil.find_one(
        Node, credentials=libcloud_credentials, api_call_kwds={"ex_node_ids": [cloud_server.id]}
    )

    if image_creation.root_device_mapping:
        root_mapping = types.Ec2BlockDeviceMapping(
            {
                "device_name": server.extra["root_device_name"],
                "volume_size": image_creation.root_device_mapping.get("size"),
                "volume_type": image_creation.root_device_mapping.get("volume_type"),
                "iops": image_creation.root_device_mapping.get("iops"),
                "throughput": image_creation.root_device_mapping.get("throughput"),
            }
        )
    else:
        dev_name = server.extra["root_device_name"]

        root_vol = libcloudutil.find_one(
            StorageVolume,
            driver=server.driver,
            query={"device": dev_name},
            api_call_kwds={"node": server},
        )
        root_mapping = types.Ec2BlockDeviceMapping(
            {
                "device_name": dev_name,
                "volume_size": root_vol.size,
                "volume_type": root_vol.extra.get("volume_type"),
                "iops": root_vol.extra.get("iops"),
                "throughput": root_vol.extra.get("throughput"),
            }
        )

    block_device_mapping.append(root_mapping)

    return types.Ec2ImageCreationLocals(
        {
            "block_device_mapping": block_device_mapping,
            "libcloud_credentials": libcloud_credentials,
            "cloud_server": cloud_server,
            "image_creation": image_creation,
        }
    )


@celeryutil.agent_task(name="ec2.sysprep")
def sysprep(**kwargs):
    """Performs ec2config sysprep"""
    subprocess2.check_output("ec2config sysprep", shell=True)


def _get_version(tools_dir):
    version = tools_dir.split("-")[-1]
    version = tuple(int(x) for x in version.split("."))
    return version


def _remove_old_tools(tools_dir):
    for item in os.listdir(tools_dir):
        if item.startswith("ec2-ami-tools"):
            os.removedirs(os.path.join(tools_dir, item))


def _fetch_unarchive(src, dst):
    return misc.unarchive(misc.fetch(src, dst), dst)


def _patch_ami_tools_centos5(ami_bin_dir):
    # patching ami tools so /dev/root is determinated as valid device
    file_to_patch = os.path.join(os.path.dirname(ami_bin_dir), "lib/ec2/platform/linux/image.rb")
    subprocess2.check_output(("patch", file_to_patch, "-i", AMI_TOOLS_PATCH_PATH))

    # updating mkfs because of filesystem option setting bug
    pkgmgr_tasks.install("texinfo")

    _fetch_unarchive(
        "https://www.kernel.org/pub/linux/kernel/"
        "people/tytso/e2fsprogs/v1.42.5/e2fsprogs-1.42.5.tar.gz",
        "/tmp",
    )

    e2fs_dir = "/tmp/e2fsprogs-1.42.5"
    build_dir = os.path.join(e2fs_dir, "build")
    os.mkdir(build_dir)
    subprocess2.check_output(("../configure"), cwd=build_dir)
    subprocess2.check_output(("make"), cwd=build_dir)
    subprocess2.check_output(("make", "install"), cwd=build_dir)


def _install_ami_tools(tools_dir="/usr/local/scalr/ec2-tools"):
    sysutil.mkdir_p(os.path.dirname(tools_dir))
    sysutil.mkdir_p(tools_dir, 0o755)

    # check if tools are already installed
    for item in os.listdir(tools_dir):
        if item.startswith("ec2-ami-tools"):
            return os.path.join(tools_dir, "ec2-ami-tools-1.5.6", "bin", "ec2-bundle-vol")

    pkgmgr_tasks.install("scalr-ruby")

    _fetch_unarchive(
        "http://s3.amazonaws.com/ec2-downloads/{}.zip".format(AMI_TOOLS_PACKAGE), tools_dir
    )
    ami_bin_dir = os.path.join(tools_dir, AMI_TOOLS_PACKAGE, "bin")

    if fact["os"]["name"] == "CentOS" and fact["os"]["release"] < (6, 0):
        _patch_ami_tools_centos5(ami_bin_dir)
    subprocess2.check_output(("chmod", "-R", "0755", tools_dir))

    return os.path.join(ami_bin_dir, "ec2-bundle-vol")


@celeryutil.agent_task(name="ec2.install_ami_tools")
def install_ami_tools(image_creation_locals=None, **kwargs):
    pkgmgr_tasks.updatedb()
    environ = None
    if fact["os"]["name"] == "Amazon":
        pkgmgr_tasks.install("aws-amitools-ec2-1.5.3")
        bundle_vol_cmd = "/opt/aws/bin/ec2-bundle-vol"
    else:
        bundle_vol_cmd = _install_ami_tools()
        environ = os.environ.copy()
        environ["PATH"] = environ["PATH"] + (":%s/bin" % SCALR_RUBY_DIR)
        environ["MY_RUBY_HOME"] = SCALR_RUBY_DIR

    if fact["os"]["family"] == "redhat":
        pkgmgr_tasks.install("parted")
    pkgmgr_tasks.install("kpartx")

    return types.Ec2BundleLocals({"cmd_path": bundle_vol_cmd, "environ": environ})


def _generate_fstab(node, original_path="/etc/fstab", edited_path="/tmp/fstab"):
    shutil.copy(original_path, edited_path)
    devices = (
        storage_linuxutil.ec2_name_to_device(entry["device_name"])  # pylint: disable=no-member
        for entry in node.extra["block_device_mapping"]
    )
    devices = (
        storage_linuxutil.ec2_get_persistent_device_path(entry)  # pylint: disable=no-member
        for entry in devices
    )
    rm_commands = ['rm */spec[.="{}"]/..'.format(entry) for entry in devices]
    rm_commands.append("save")
    augeasutil.execute(
        rm_commands, context="/files" + edited_path, lens="Fstab.lns", incl=[edited_path]
    )


@celeryutil.agent_task(name="ec2.bundle_volume")
def bundle_volume(
    credentials=None,
    cloud_server=None,
    name=None,
    size=None,
    environ=None,
    bundle_vol_cmd=None,
    destination=None,
    excludes=None,
    **kwargs,
):
    credentials = types.AwsCloudCredentials(credentials)
    cloud_server = types.CloudServer(cloud_server)
    libcloud_credentials = credentials.to_libcloud(cloud_server.location)

    sysutil.mkdir_p(destination)

    node = libcloudutil.find_one(
        Node, credentials=libcloud_credentials, api_call_kwds={"ex_node_ids": [cloud_server.id]}
    )

    fstab_path = "/tmp/fstab"
    _generate_fstab(node, edited_path=fstab_path)

    cmd = (
        bundle_vol_cmd,
        "--cert",
        credentials.certificate_path,
        "--privatekey",
        credentials.private_key_path,
        "--user",
        credentials.account_id,
        "--arch",
        fact["os"]["arch"],
        "--size",
        str(size),
        "--destination",
        destination,
        "--prefix",
        name,
        "--volume",
        "/",
        "--fstab",
        fstab_path,
        "--debug",
    )

    excludes_str = fstab_path

    if excludes:
        excludes_str = excludes_str + "," + ",".join(excludes)

    cmd += ("--exclude", excludes_str)

    LOG.debug("Image prepare command: %s", logutil.LazyJoin(" ", cmd))
    out = subprocess2.check_output(
        cmd, env=environ, stdout=subprocess2.PIPE, stderr=subprocess2.STDOUT
    )
    LOG.debug("Image prepare command out: %s", out)


@celeryutil.agent_task(name="ec2.upload_bundle")
def upload_bundle(
    credentials=None,
    bundle_vol_cmd=None,
    environ=None,
    manifest_path=None,
    ami_cloudfs_path=None,
    **kwargs,
):
    credentials = types.AwsCloudCredentials(credentials)

    LOG.debug("Uploading image (with ec2-upload-bundle)")

    if ami_cloudfs_path.startswith("s3://"):
        ami_cloudfs_path = ami_cloudfs_path[len("s3://") :]

    cmd = (
        os.path.join(os.path.dirname(bundle_vol_cmd), "ec2-upload-bundle"),
        "--bucket",
        ami_cloudfs_path,
        "--access-key",
        credentials.access_key,
        "--secret-key",
        credentials.secret_key,
        "--manifest",
        manifest_path,
    )

    subprocess2.check_output(cmd, env=environ)


@celeryutil.agent_task(name="ec2.ex_register_image")
def register_image(
    credentials=None,
    node_id=None,
    name=None,
    description=None,
    manifest_path=None,
    ami_cloudfs_path=None,
    **kwargs,
):
    LOG.debug("Registering image")
    s3_manifest_path = "%s/%s" % (ami_cloudfs_path, os.path.basename(manifest_path))
    LOG.debug("Registering image '%s'", s3_manifest_path)

    credentials = types.LibcloudCredentials(credentials)
    driver = libcloudutil.create_driver(credentials)
    node = libcloudutil.find_one(Node, api_call_kwds={"ex_node_ids": [node_id]}, driver=driver)

    img = driver.ex_register_image(
        name,
        description=description,
        architecture=fact["os"]["arch"],
        image_location=s3_manifest_path,
        kernel_id=node.extra["kernel_id"],
        ramdisk_id=node.extra["ramdisk_id"],
        virtualization_type=node.extra["virtualization_type"],
    )

    LOG.debug("Image is registered.")
    LOG.debug("Image %s is available", img.id)
    return img.id


@celeryutil.agent_task(name="ec2.bundle_cleanup")
def bundle_cleanup(name=None, destination=None, **kwargs):
    subprocess2.check_output("rm -f {}/{}*".format(destination, name), shell=True)
    subprocess2.check_output("rm -f /tmp/fstab", shell=True)
