"""
Miscellaneous masks

.. moduleauthor:: Nick Demyanchuk
.. moduleauthor:: Marat Komarov
"""
from __future__ import absolute_import

import contextlib
import os
import tarfile
import zipfile

import requests

from agent.utils import celeryutil
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil

__author__ = "spike"

LOG = logutil.get_logger(__name__)


@celeryutil.agent_task(name="misc.unarchive")
def unarchive(src, dst=".", **kwargs):
    display_src = os.path.basename(src)
    LOG.info("Extracting archive %r ...", display_src)
    LOG.debug("Extracting archive  %r ->  %r", src, dst)
    if src.endswith(".zip"):
        LOG.debug("Detected zip")
        opener, mode = zipfile.ZipFile, "r"
    elif src.endswith(".tar.gz") or src.endswith(".tgz"):
        LOG.debug("Detected tar+gzip")
        opener, mode = tarfile.open, "r:gz"
    elif src.endswith(".tar.bz2") or src.endswith(".tbz"):
        LOG.debug("Detected tar+bzip2")
        opener, mode = tarfile.open, "r:bz2"
    else:
        msg = (
            "Archive {!r} has an unsupported file extension. "
            "Supported: zip, tar.gz, tgz, tar.bz2, tbz"
        ).format(display_src)
        raise exc.NotSupportedError(msg)
    cwd = os.getcwd()
    os.chdir(dst)
    try:
        file = opener(src, mode)
        try:
            file.extractall()
        finally:
            file.close()
    finally:
        os.chdir(cwd)
    LOG.info("Successfully extracted %r", display_src)


@celeryutil.agent_task(name="misc.fetch")
def fetch(url, destination, chunk_size=8192, report_steps=5, **kwargs):
    mixutil.validate_url(url)
    if os.path.isdir(destination):
        destination = os.path.join(destination, os.path.basename(url))
    LOG.info("Downloading %r ...", url)
    LOG.debug("Downloading %r -> %r (chunk_size: %s)", url, destination, chunk_size)

    if report_steps != 0:
        r = requests.head(url)
        r.raise_for_status()
        content_size = float(r.headers.get("content-length", 0))
    else:
        content_size = 0
    with contextlib.closing(requests.get(url, stream=True)) as r:
        r.raise_for_status()
        with open(destination, "wb+") as fp:
            read_bytes = 0.0
            step = 1.0
            for chunk in r.iter_content(chunk_size):
                read_bytes += len(chunk)
                fp.write(chunk)
                # TODO(marat) restore PROGRESS when logs will be done
                if (
                    report_steps
                    and content_size != 0
                    and read_bytes / content_size >= step / report_steps
                ):
                    LOG.debug("%d%% downloaded...", step * 100.0 / report_steps)
                    step += 1
    LOG.info("Successfully downloaded %r", url)
    return destination
