from __future__ import absolute_import

import errno
import glob
import os
import re

import agent.tasks.pkgmgr.tasks as pkgmgr
from common.utils import logutil
from common.utils import mixutil
from common.utils import subprocess2

LOG = logutil.get_logger(__name__)


class AptPackageMgr(pkgmgr.ABCPackageMgr):
    def get_installed(self):
        output = subprocess2.check_output(
            ("dpkg-query", "-W", "-f=${Status}|${Package}|${Version}\n")
        )
        return (
            pkgmgr.Package(line.split("|")[1], line.split("|")[2].split(":")[-1])
            for line in output.splitlines()
            if line.split("|")[0] in ("install ok installed", "hold ok installed")
        )

    def updatedb(self, **kwds):
        def clean_lists(path):
            if os.path.isdir(path):
                for name in os.listdir(path):
                    filename = os.path.join(path, name)
                    if name != "lock" and os.path.isfile(filename):
                        try:
                            os.remove(filename)
                        except OSError as e:
                            if e.errno != errno.ENOENT:
                                raise

        clean_lists("/var/lib/apt/lists/partial")
        clean_lists("/var/lib/apt/lists")

        cmd = []
        if kwds.get("apt_repository"):
            cmd.extend(
                [
                    "--no-list-cleanup",
                    "-o",
                    "Dir::Etc::sourcelist=sources.list.d/{0}.list".format(kwds["apt_repository"]),
                    "-o",
                    "Dir::Etc::sourceparts=-",
                ]
            )
        cmd.append("update")
        apt_get(cmd)

    def repos(self):
        files = glob.glob("/etc/apt/sources.list.d/*.list")
        names = [os.path.basename(os.path.splitext(f)[0]) for f in files]
        return names

    def remove(self, name, purge=False):
        action = "purge" if purge else "remove"
        apt_get((action, name))

    def _installed_and_candidate(self, name):
        installed_re = re.compile(r"^\s{2}Installed: (.+)$")
        candidate_re = re.compile(r"^\s{2}Candidate: (.+)$")
        installed = candidate = None

        apt_cache = ("/usr/bin/apt-cache", "policy", name)
        for line in subprocess2.check_output(apt_cache).splitlines():
            m = installed_re.match(line)
            if m:
                installed = m.group(1)
                if installed == "(none)":
                    installed = None
                continue

            m = candidate_re.match(line)
            if m:
                candidate = m.group(1)
                continue

        return installed, candidate

    def _format_name_version(self, name, version):
        if version:
            name += "={0}".format(version)
        return name

    def _install_package(self, name_version, **kwds):
        apt_get(("install", name_version))

    def _install_package_list(self, package_list, **kwds):
        apt_get(["install"] + package_list)

    def _install_file(self, filename):
        subprocess2.check_output(("dpkg", "-i", filename))


def apt_get(args, **kwds):
    env = kwds.setdefault("env", os.environ)
    env.update({"DEBIAN_FRONTEND": "noninteractive", "DEBIAN_PRIORITY": "critical"})
    command = [
        "/usr/bin/apt-get",
        "--quiet",
        "--yes",
        "--force-yes",
        "-o Dpkg::Options::=--force-confold",
    ]
    command += list(args)

    def try_apt_get():
        try:
            return subprocess2.check_output(command, **kwds)
        except subprocess2.CalledProcessError as e:
            if "is another process using it?" in e.stderr or "Could not get lock" in e.stderr:
                LOG.debug("Dpkg is locked by another process")
                return False
            else:
                raise

    return mixutil.try_until(try_apt_get, interval=10)
