import os
import re

import agent.tasks.pkgmgr.tasks as pkgmgr
from common.utils import subprocess2


class ZypperPackageMgr(pkgmgr.ABCPackageMgr):
    def get_installed(self):
        output = subprocess2.check_output(("rpm", "-qa", "--queryformat", "%{NAME}|%{VERSION}\n"))
        return (pkgmgr.Package(*line.split("|")) for line in output.splitlines() if line)

    def updatedb(self, **kwds):
        ZypperCmd().args("ref").check_output()

    def remove(self, name, purge=False):
        ZypperCmd().args("remove", name).check_output()

    def repos(self):
        ret = []
        out = ZypperCmd().args("repos", show_enabled_only=True).check_output()
        repo_re = re.compile(r"^\s?\d+\s?\|[^|]*\|\s([^|]*)\s")
        for line in out.splitlines():
            match = repo_re.search(line)
            if match:
                ret.append(match.group(1)).lower().rstrip()
        return ret

    def _format_name_version(self, name, version):
        if version:
            return "{}-{}".format(name, version)
        return name

    def _installed_and_candidate(self, name):
        out = ZypperCmd().args("info", name).check_output().strip()
        status, installed, latest = "", None, None

        latest_match = re.search("Version: (.*)", out)
        if latest_match:
            latest = latest_match.group(1)
        status_match = re.search("Status: (.*)", out)
        if status_match:
            status = status_match.group(1)

        if status == "up-to-date":
            installed = latest
        elif "out-of-date" in status:
            match = re.search(r"version (.*?) installed", status)
            if match:
                installed = match.group(1)
        return installed, latest

    def _install_package(self, name_version, **kwds):
        self._install_package_list([name_version], **kwds)

    def _install_package_list(self, package_list, **kwds):
        ZypperCmd().args("install", *package_list, no_confirm=True).check_output()

    def _install_file(self, filename):
        subprocess2.check_output(("rpm", "--install", "--nodeps", filename))


class ZypperCmd(subprocess2.Command):
    """Wraps 'zypper' command.

    Execute only for current architecture and disable prompts.
    """

    binary = "/usr/bin/zypper"

    def __init__(self):
        super(ZypperCmd, self).__init__()
        self.args(non_interactive=True)

    def _normalize_env_keyword(self, kwds):
        env = kwds.pop("env", os.environ)
        env.pop("PYTHONPATH", None)
        kwds["env"] = env

    def check_output(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(ZypperCmd, self).check_output(**kwds)

    def check_call(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(ZypperCmd, self).check_call(**kwds)

    def call(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(ZypperCmd, self).call(**kwds)

    def run(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(ZypperCmd, self).run(**kwds)
