from __future__ import absolute_import

import abc
import collections
import glob
import os
import tempfile

from agent.tasks import misc
from agent.utils import celeryutil
from common.utils import exc
from common.utils import imputil
from common.utils import logutil
from common.utils import six
from common.utils import sysutil
from common.utils.facts import fact

__all__ = [
    "ABCPackageMgr",
    "Error",
    "get_installed",
    "install",
    "PackageMgr",
    "remove",
    "repository",
    "updatedb",
]


LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    pass


def PackageMgr(*args, **kwds):
    """
    Factory function for package manager creation
    """
    if fact["os"]["family"] == "redhat":
        imp_name = "yum.YumPackageMgr"
    elif fact["os"]["family"] == "debian":
        imp_name = "apt.AptPackageMgr"
    elif fact["os"]["family"] == "suse":
        imp_name = "suse.ZypperPackageMgr"
    else:
        raise exc.NotSupportedError(
            ("No package manager interface for " "operation system family {0!r}").format(
                fact["os"]["family"]
            )
        )
    imp_name = "agent.tasks.pkgmgr." + imp_name
    return imputil.import_object(imp_name, *args, **kwds)


def instance():
    # pylint: disable=protected-access
    if not getattr(instance, "_instance", None):
        instance._instance = PackageMgr()
    return instance._instance


@six.add_metaclass(abc.ABCMeta)
class ABCPackageMgr(object):
    """
    Does package fetch, install, update and removal.
    Compares two package versions.
    """

    def info(self, name):
        """
        Returns info about package
        Example:
                {'installed': '2.6.7-ubuntu1',
                'candidate': '2.6.7-ubuntu5'}
        installed is None, if package not installed
        candidate is None, if latest version of package is installed
        """
        installed, candidate = self._installed_and_candidate(name)
        return {
            "installed": installed,
            "candidate": candidate if installed and installed != candidate else None,
        }

    def install(self, name, version=None, updatedb=False, **kwds):
        """
        Installs a `version` of package `name`
        """
        name_version = self._format_name_version(name, version)
        if updatedb:
            self.updatedb()

        self._install_package(name_version, **kwds)

    def install_list(self, packages, updatedb=False, **kwds):
        """
        Installs list of packages.

        :param iterable(str) or iterable(tuple(str, str)) packages: List of packages to install.
        Can contain either name of package or tuple in format (<name>, <version>).
        """
        unified_packages = []
        for package in packages:
            # pylint: disable=no-member
            if isinstance(package, six.string_types):
                unified_packages.append(self._format_name_version(package, None))
            elif isinstance(package, six.moves.Iterable):
                unified_packages.append(self._format_name_version(package[0], package[1]))
            else:
                raise Error(
                    "Unknown packages container {}, packages: {}".format(type(packages), packages)
                )

        if updatedb:
            self.updatedb()

        self._install_package_list(unified_packages, **kwds)

    def install_file(self, path_or_url):
        """
        Installs package from path or URL
        """
        assert isinstance(path_or_url, six.string_types)
        tmp_path = path = None
        if path_or_url.startswith("http"):
            tmp_path = misc.fetch(path_or_url, tempfile.mkdtemp())
        elif not os.path.exists(path_or_url):
            raise exc.NotFoundError("Package file not exists: {!r}".format(path_or_url))
        try:
            path = tmp_path or path_or_url
            LOG.info("Installing package %s", path)
            self._install_file(path)
        finally:
            if tmp_path:
                sysutil.rm_rf(os.path.dirname(tmp_path))

    @abc.abstractmethod
    def get_installed(self):
        """Return an installed :class:`Package`-s list."""

    @abc.abstractmethod
    def remove(self, name, purge=False):
        """Removes package with given name."""

    @abc.abstractmethod
    def updatedb(self, **kwds):
        """Updates package manager internal database."""

    @abc.abstractmethod
    def repos(self):
        """List enabled repositories."""

    @abc.abstractmethod
    def _format_name_version(self, name, version):
        pass

    @abc.abstractmethod
    def _installed_and_candidate(self, name):
        pass

    @abc.abstractmethod
    def _install_package(self, name_version, **kwds):
        pass

    @abc.abstractmethod
    def _install_package_list(self, package_list, **kwds):
        pass

    @abc.abstractmethod
    def _install_file(self, filename):
        pass


class Repository(object):
    """
    Manages system repositories.
    """

    def __init__(self, name, url, prefix=None, ssl_verify=True):
        """
        :param str name: Name of the repository
        :param str url: Repository url
        :param str prefix: Prefix to the name of the repository,
            as in "scalr-stable".
        """
        filename_tpl, config_tpl = self._get_templates()
        self.prefix = prefix
        self.ssl_verify = ssl_verify
        self.name = "{}-{}".format(prefix, name) if prefix else name
        self.url = url
        kwds = {"name": self.name, "url": self.url, "ssl_verify": int(self.ssl_verify)}
        self.filename = filename_tpl.format(**kwds)
        self.config = config_tpl.format(**kwds)

    def _get_templates(self):
        """
        Returns appropriate repository templates for supported platforms.
        """
        if fact["os"]["family"] == "debian":
            filename_tpl = "/etc/apt/sources.list.d/{name}.list"
            config_tpl = "deb {url}"
        elif fact["os"]["family"] in ("redhat", "suse"):
            if fact["os"]["family"] == "suse":
                filename_tpl = "/etc/zypp/repos.d/{name}.repo"
            else:
                filename_tpl = "/etc/yum.repos.d/{name}.repo"
            config_tpl = (
                "[{name}]\n"
                "name={name}\n"
                "baseurl={url}\n"
                "sslverify={ssl_verify}\n"
                "enabled=1\n"
                "gpgcheck=1\n"
                "repo_gpgcheck=1\n"
                "gpgkey={url}/scalr.gpg\n"
            )

        else:
            raise exc.NotSupportedError(
                (
                    "Repository management is not available for " "operation system family {0!r}"
                ).format(fact["os"]["family"])
            )
        return filename_tpl, config_tpl

    def ensure(self):
        """
        Ensures repository config file.
        """
        assert self.filename and self.config
        # Delete previous repositories
        # SCALARIZR-2531 - self.prefix is always 'scalr-agent' but if
        # the repo files were not found, the code tries clearing repo files using 'scalr*' mask
        # (in generic case this could cause a clash with Scalr repo file)
        if self.prefix:
            path = os.path.dirname(self.filename)
            filenames = glob.glob(os.path.join(path, self.prefix + "*")) or glob.glob(
                os.path.join(path, "scalr*")
            )
            for filename in filenames:
                if os.path.isfile(filename):
                    os.remove(filename)
        # Ensure new repository
        LOG.info("Creating repository '%s' (%s, url=%s)", self.name, self.filename, self.url)
        with open(self.filename, "w+") as fp:
            fp.write(self.config)


Package = collections.namedtuple("Package", "name,version")


def get_installed():
    """Returns installed :class:`Package`-es list.

    NOTE! not a task
    """
    return instance().get_installed()


@celeryutil.agent_task(name="pkgmgr.updatedb")
def updatedb(**kwargs):
    """
    Updates package manager database
    """
    instance().updatedb()


@celeryutil.agent_task(name="pkgmgr.install")
def install(name=None, version=None, source=None, updatedb=False, **kwds):
    """
    Ensure that package installed.
    version='latest' will each time ensure that latest package is installed
    """
    assert name or source, "Install requires `name` or `source` set"
    mgr = instance()
    # pylint: disable=no-member
    if source:
        mgr.install_file(source)
    elif isinstance(name, six.string_types):
        if updatedb:
            mgr.updatedb()
        info = mgr.info(name)
        if version == "latest" and info["candidate"]:
            LOG.info("Installing package %s=%s", name, info["candidate"])
            mgr.install(name, info["candidate"], **kwds)
        elif not info["installed"]:
            version_str = "" if version is None else "=%s" % version
            LOG.info("Installing package %s%s", name, version_str)
            mgr.install(name, version, **kwds)
    elif isinstance(name, six.moves.Iterable):
        if updatedb:
            mgr.updatedb()

        # Looking through every version of package in the iterable and gathering candidate
        # for each `latest` version
        def pkg_iterator():
            for package in name:
                # pylint: disable=no-member
                if isinstance(package, six.string_types):
                    yield package
                elif isinstance(package, six.moves.Iterable):
                    if package[1] == "latest":
                        candidate = mgr.info(package[0])["candidate"]
                        # if there are no candidate, skipping install completely
                        if candidate:
                            yield package[0], candidate
                    else:
                        yield package[0], package[1]

        mgr.install_list(pkg_iterator(), **kwds)


@celeryutil.agent_task(name="pkgmgr.remove")
def remove(name, purge=False, **kwargs):
    """
    Ensure that package removed (purged)
    """
    mgr = instance()
    installed = mgr.info(name)["installed"]
    if purge or installed:
        if installed:
            LOG.info("Uninstalling package %s=%s", name, installed)
        else:
            LOG.info("Uninstalling package %s", name)
        mgr.remove(name, purge)


@celeryutil.agent_task(name="pkgmgr.repository")
def repository(name, url, prefix=None, ssl_verify=True, **kwargs):
    """
    Ensure repository file
    """
    if not fact["manage_container_host"]:
        repo = Repository(name, url, prefix, ssl_verify)
        repo.ensure()
