from __future__ import absolute_import

import os
import re

import agent.tasks.pkgmgr.tasks as pkgmgr
from common.utils import subprocess2
from common.utils.facts import fact


class YumPackageMgr(pkgmgr.ABCPackageMgr):
    def get_installed(self):
        output = subprocess2.check_output(("rpm", "-qa", "--queryformat", "%{NAME}|%{VERSION}\n"))
        return (pkgmgr.Package(*line.split("|")) for line in output.splitlines() if line)

    def updatedb(self, **kwds):
        YumCmd().args("clean", "expire-cache").check_output()

    def remove(self, name, purge=False):
        YumCmd().args("remove", name).check_output()

    def repos(self):
        ret = []
        repo_re = re.compile(r"Repo-id\s+:\s(.*)")
        out = YumCmd().args("repolist", "--verbose").check_output()
        for line in out.splitlines():
            m = repo_re.search(line)
            if m:
                ret.append(m.group(1))
        return [s.lower() for s in ret]

    def _format_name_version(self, name, version):
        if version:
            return "{}-{}".format(name, version)
        return name

    def _installed_and_candidate(self, name):
        version_re = re.compile(r"[^\s]+\s+([^\s]+)")
        out = YumCmd().args("list", "--showduplicates", name).check_output().strip()
        lines = [s.strip() for s in out.splitlines()]
        try:
            line = lines[lines.index("Installed Packages") + 1]
            installed = version_re.match(line).group(1)
        except ValueError:
            installed = None
        if "Available Packages" in lines:
            versions = [
                version_re.match(line).group(1)
                for line in lines[lines.index("Available Packages") + 1 :]
            ]
        else:
            versions = [None]
        return installed, versions[-1]

    def _install_package(self, name_version, **kwds):
        self._install_package_list([name_version], **kwds)

    def _install_package_list(self, package_list, **kwds):
        out = YumCmd().args("install", *package_list).check_output()
        re_scriptlet = r"(Non-fatal|Error in) (PREIN|PRERM|POSTIN|POSTRM) scriptlet"
        if kwds.get("rpm_raise_scriptlet_errors") and re.search(re_scriptlet, out):
            msg = "Got scriplet errors during {0} install: {1}".format(" ".join(package_list), out)
            raise pkgmgr.Error(msg)

    def _install_file(self, filename):
        subprocess2.check_output(("rpm", "-i", filename))


class YumCmd(subprocess2.Command):
    """Wraps 'yum' command.

    Execute only for current architecture and disable prompts.
    """

    binary = "/usr/bin/yum"

    def __init__(self, debug_level=0):
        super(YumCmd, self).__init__()
        # --assumeyes and --debuglevel don't work on Centos5
        self.args("-y", "-d{}".format(debug_level))
        self._strict_arch()

    def _strict_arch(self):
        if fact["kernel"]["machine"] == "x86_64":
            exclude_what = ["*.i386", "*.i486", "*.i686"]
        else:
            exclude_what = ["x86_64"]
        for exclude in exclude_what:
            self.args(exclude=exclude)

    def _normalize_env_keyword(self, kwds):
        env = kwds.pop("env", os.environ)
        env.pop("PYTHONPATH", None)
        kwds["env"] = env

    def check_output(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(YumCmd, self).check_output(**kwds)

    def check_call(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(YumCmd, self).check_call(**kwds)

    def call(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(YumCmd, self).call(**kwds)

    def run(self, **kwds):
        self._normalize_env_keyword(kwds)
        return super(YumCmd, self).run(**kwds)
