"""Abstract Linux service module.
"""
from __future__ import absolute_import

import abc
import os

import agent.tasks.service.tasks as service
from common.utils import six
from common.utils import subprocess2

SERVICE_RUNNING_CODE = 0
SERVICE_STOPPED_CODE = 3


class NotInstalledError(service.NotInstalledError):
    """Service not found exception."""

    message_tpl = "Service {!r} not found. Make sure that init script {!r} exists and executable"


# pylint: disable=abstract-method
@six.add_metaclass(abc.ABCMeta)
class LinuxServiceProvider(service.ABCServiceProvider):
    """Abstract Linux service code."""

    initd_dir = "/etc/init.d"
    has_restart = True
    has_reload = True
    has_configtest = True
    has_kill = False

    def _check_action(self, action):
        has_action = getattr(self, "has_{}".format(action), True)
        if not has_action:
            raise NotImplementedError(
                "Service action {!r} is not implemented "
                "in provider {!r}".format(action, self.__class__.__name__)
            )

    def _run_service_action(self, name, action):
        init_path = os.path.join(self.initd_dir, name)
        if not os.access(init_path, os.X_OK):
            raise NotInstalledError(name, init_path)
        self._check_action(action)
        cmd = "{} {}".format(init_path, action)
        return subprocess2.run(cmd, shell=True)

    def start_service(self, name):
        self._run_service_action(name, "start").check_returncode()

    def stop_service(self, name):
        self._run_service_action(name, "stop").check_returncode()

    def restart_service(self, name):
        self._run_service_action(name, "restart").check_returncode()

    def reload_service(self, name):
        self._run_service_action(name, "reload").check_returncode()

    def configtest_service(self, name):
        self._run_service_action(name, "configtest").check_returncode()

    def get_service_status(self, name):
        ret_code = self._run_service_action(name, "status").returncode
        if ret_code == SERVICE_RUNNING_CODE:
            return service.ServiceStatus.running
        elif ret_code == SERVICE_STOPPED_CODE:
            return service.ServiceStatus.stopped
        else:
            return service.ServiceStatus.unknown
