from __future__ import absolute_import

import sys

import agent.tasks.service.tasks as service
from common.utils import logutil
from common.utils import mixutil
from common.utils import subprocess2
from common.utils.facts import fact

LOG = logutil.get_logger(__name__)


class SystemdServiceProvider(service.ABCServiceProvider):
    has_restart = True
    has_reload = True
    has_configtest = False
    has_kill = True
    START_TIMEOUT = 10
    SERVICE_RUNNING_CODE = 0
    SERVICE_STOPPED_CODE = 3
    SERVICE_ENABLED_CODE = 0

    def ctl(self, name, action):
        result = subprocess2.run("systemctl {} {}.service".format(action, name), shell=True)
        if "No such file or directory" in result.stderr:
            raise service.NotInstalledError(name)
        return result

    def start_service(self, name):
        self.ctl(name, "start").check_returncode()

    def stop_service(self, name):
        self.ctl(name, "stop").check_returncode()

    def kill_service(self, name):
        self.ctl(name, "kill").check_returncode()

    def set_service_enabled(self, name, enabled):
        self.ctl(name, "enable" if enabled else "disable").check_returncode()

    def get_service_enabled(self, name):
        return self.ctl(name, "is-enabled").returncode == self.SERVICE_ENABLED_CODE

    def get_service_status(self, name):
        result = self.ctl(name, "show -q")
        stdout = result.stdout.splitlines()
        if "LoadState=not-found" in stdout:
            raise service.NotInstalledError(name)
        elif "ActiveState=active" in stdout:
            return service.ServiceStatus.running
        elif "ActiveState=inactive" in stdout:
            return service.ServiceStatus.stopped
        return service.ServiceStatus.unknown

    def _wait_running(self, name):
        LOG.debug("Waiting for service '%s' to start running", name)

        error_msg = "Service '{}' failed to reload in {} seconds".format(name, self.START_TIMEOUT)
        mixutil.try_until(
            lambda: self.get_service_status(name) == service.ServiceStatus.running,
            error_msg=error_msg,
            log=LOG,
            interval=0.1,
            timeout=self.START_TIMEOUT,
        )

    def reload_service(self, name):
        LOG.debug("Reloading service %s", name)
        self.ctl(name, "reload").check_returncode()
        self._wait_running(name)

    def restart_service(self, name):
        LOG.debug("Restarting service %s", name)
        self.ctl(name, "restart").check_returncode()
        self._wait_running(name)


# pylint: disable=locally-disabled,abstract-method
class RemoteSystemdServiceProvider(SystemdServiceProvider):
    """
    Extend SystemdServiceProvider to work in case when scalarizr inside container
    """

    def ctl(self, name, action):
        cmd = "systemctl {} {}.service".format(action, name)
        try:
            rsh = subprocess2.RemoteCommand(
                hostname=fact["routes"]["default"]["gw"], host_key_checking="no"
            )
            rsh.args(cmd)
            result = rsh.run()
            result.check_returncode()
            return result
        except subprocess2.CalledProcessError:
            e = sys.exc_info()[1]
            if "could not be found" in e.stderr:
                raise service.NotInstalledError(name)
            else:
                raise service.Error(e.stderr)
