from __future__ import absolute_import

import abc
import time

from agent.utils import celeryutil
from common.utils import exc
from common.utils import imputil
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils.facts import fact

__all__ = [
    "Error",
    "AlreadyRunningError",
    "NotInstalledError",
    "NotRunningError",
    "Service",
    "ServiceStatus",
    "ABCServiceProvider",
    "name_to_provider",
    "enable",
    "disable",
    "start",
    "stop",
    "restart",
    "reload",
    "status",
]


LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    pass


class AlreadyRunningError(Error):
    message_fmt = "Service {!r} is already running"


class NotInstalledError(Error):
    message_fmt = "Service {!r} is not installed"


class NotRunningError(Error):
    message_fmt = "Service {!r} is not running"


# Custom service to provider mapping. For example, if we need on a specific
# platform to have a custom service provider class, we need to set:
# name_to_provider[<service_name>] = <CustomProviderClass>
name_to_provider = {}


# Pool of created providers. For each service name we need only one provider
# instance, so we store them in this dict after creation
_provider_pool = {}


@mixutil.memorize()
def default_provider():
    """Selects service provider class for the current init system
    and creates its instance via load_provider()

    :rtype: :class:`ABCServiceProvider`
    """

    if fact["manage_container_host"]:
        imp_name = "systemd.RemoteSystemdServiceProvider"
    elif fact["os"]["init"] == "systemd":
        imp_name = "systemd.SystemdServiceProvider"
    elif fact["os"]["name"] == "ubuntu":
        imp_name = "upstart.UpstartServiceProvider"
    elif fact["os"]["family"] == "debian":
        imp_name = "debian.DebianServiceProvider"
    elif fact["os"]["family"] in ("redhat", "suse"):
        imp_name = "redhat.RedHatServiceProvider"
    elif fact["os"]["family"] == "windows":
        imp_name = "windows.WindowsServiceProvider"
    else:
        raise exc.NotSupportedError(
            ("No service control manager interface for " "operation system {0!r}").format(
                fact["os"]["name"]
            )
        )
    return load_provider("agent.tasks.service." + imp_name)


def load_provider(provider):
    """
    :rtype: :class:`ABCServiceProvider`
    """
    if isinstance(provider, six.string_types):
        provider = imputil.import_attr(provider)
    if issubclass(provider, ABCServiceProvider):
        provider = provider()
    return provider


class ServiceStatus(object):
    starting = "starting"
    running = "running"
    stopping = "stopping"
    stopped = "stopped"
    unknown = "unknown"


class Service(object):
    name = None
    provider = None

    def __init__(self, name, **kwds):
        if name in _provider_pool:
            self.provider = _provider_pool[name]
        else:
            if name in name_to_provider:
                self.provider = load_provider(name_to_provider[name])
            else:
                self.provider = default_provider()
            _provider_pool[name] = self.provider

        self.name = name
        self.__dict__.update(kwds)

    def status(self):
        return self.provider.get_service_status(self.name)

    @property
    def running(self):
        return self.provider.get_service_status(self.name) == ServiceStatus.running

    @property
    def enabled(self):
        return self.provider.get_service_enabled(self.name)

    def enable(self):
        self.provider.set_service_enabled(self.name, True)

    def disable(self):
        self.provider.set_service_enabled(self.name, False)

    def start(self, already_running_ok=True):
        try:
            self.provider.start_service(self.name)
        except AlreadyRunningError:
            LOG.debug("Service %s is already running.", self.name)
            if not already_running_ok:
                raise
        LOG.info("Service started: %s", self.name)

    def stop(self):
        self.provider.stop_service(self.name)
        LOG.info("Service stopped: %s", self.name)

    def restart(self):
        if self.provider.has_restart:
            self.provider.restart_service(self.name)
        else:
            self.stop()
            time.sleep(1)
            self.start()

    def reload(self):
        if self.provider.has_reload:
            LOG.info("Service reload: %s", self.name)
            self.provider.reload_service(self.name)

    def configtest(self):
        if self.provider.has_configtest:
            self.provider.configtest_service(self.name)

    def kill(self):
        if self.provider.has_kill:
            self.provider.kill_service(self.name)


@six.add_metaclass(abc.ABCMeta)
class ABCServiceProvider(object):
    @abc.abstractmethod
    def start_service(self, name):
        raise NotImplementedError()

    @abc.abstractmethod
    def stop_service(self, name):
        raise NotImplementedError()

    @abc.abstractmethod
    def set_service_enabled(self, name, enabled):
        pass

    @abc.abstractmethod
    def get_service_enabled(self, name):
        pass

    @abc.abstractmethod
    def get_service_status(self, name):
        pass

    # Optinal service actions
    def reload_service(self, name):
        pass

    def restart_service(self, name):
        pass

    def configtest_service(self, name):
        pass


@celeryutil.agent_task(name="service.enable")
def enable(name, *args, **kwds):
    """
    Enable service auto-start on boot

    :param str name: Service name

    :raises NotInstalledError: when service is not installed in OS
    """
    return Service(name).enable(*args, **kwds)


@celeryutil.agent_task(name="service.disable")
def disable(name, *args, **kwds):
    """
    Disable service auto-start on boot

    :param str name: Service name

    :raises NotInstalledError: when service is not installed in OS
    """
    return Service(name).disable(*args, **kwds)


@celeryutil.agent_task(name="service.start")
def start(name, *args, **kwds):
    """
    Start service.
    If service is already running, silently returns

    :param str name: Service name

    :raises NotInstalledError: when service is not installed in OS
    """
    return Service(name).start(*args, **kwds)


@celeryutil.agent_task(name="service.stop")
def stop(name, *args, **kwds):
    """
    Stop service.

    If service is alrealy stopped, silently returns

    :raises NotInstalledError: when service is not installed in OS
    """
    return Service(name).stop(*args, **kwds)


@celeryutil.agent_task(name="service.restart")
def restart(name, *args, **kwds):
    """
    Restart service.

    If service is not running, just start it without warinings

    :raises NotInstalledError: when service is not installed in OS
    """
    return Service(name).restart(*args, **kwds)


@celeryutil.agent_task(name="service.reload")
def reload(name, *args, **kwds):
    """
    Reload service.

    :param str name: Service name

    :raises NotInstalledError: when service is not installed in OS

    :raises NotRunningError: when service is not running

    :raises common.utils.exc.NotSupportedError: if action is not supported by
        underliyng service, or service provider
    """
    return Service(name).reload(*args, **kwds)


@celeryutil.agent_task(name="service.status")
def status(name, *args, **kwds):
    """
    Get service status.

    Status is one of the following:
    ['running', 'stopped', 'starting', 'stopping', 'unknown']

    :param str name: Service name

    :raises NotInstalledError: when service is not installed in OS
    """
    return Service(name).status(*args, **kwds)
