from __future__ import absolute_import

import collections
import contextlib

# pylint: disable=import-error
import pywintypes
import win32service

import agent.tasks.service.tasks as service
from common.utils import logutil
from common.utils import mixutil

# pylint: enable=import-error


LOG = logutil.get_logger(__name__)
NOT_INSTALLED_RETURNCODE = 1060
ALREADY_STARTED_RETURNCODE = 1056
NOT_RUNNING_RETURNCODE = 1062
STATUS_CHANGE_TIMEOUT = 20
STATUS_POLL_INTERVAL = 0.1


status_map = collections.defaultdict(lambda: service.ServiceStatus.unknown)
status_map.update(
    {
        win32service.SERVICE_RUNNING: service.ServiceStatus.running,
        win32service.SERVICE_START_PENDING: service.ServiceStatus.starting,
        win32service.SERVICE_CONTINUE_PENDING: service.ServiceStatus.starting,
        win32service.SERVICE_STOP_PENDING: service.ServiceStatus.stopping,
        win32service.SERVICE_PAUSE_PENDING: service.ServiceStatus.stopping,
        win32service.SERVICE_STOPPED: service.ServiceStatus.stopped,
        win32service.SERVICE_PAUSED: service.ServiceStatus.stopped,
    }
)


start_mode_map = collections.defaultdict(lambda: "unknown")
start_mode_map.update(
    {
        win32service.SERVICE_BOOT_START: "Boot",
        win32service.SERVICE_SYSTEM_START: "System",
        win32service.SERVICE_AUTO_START: "Automatic",
        win32service.SERVICE_DEMAND_START: "Manual",
        win32service.SERVICE_DISABLED: "Disabled",
    }
)


@contextlib.contextmanager
def service_handle(
    name,
    desired_access=win32service.SERVICE_ALL_ACCESS,
    scm_desired_access=win32service.SC_MANAGER_ALL_ACCESS,
):
    # pylint: disable=no-member
    """
    Context for automatic Windows service handle managing and typical error reporting.
    """
    handle = None
    try:
        handle = win32service.OpenService(
            win32service.OpenSCManager(None, None, scm_desired_access), name, desired_access
        )
        yield handle
    except pywintypes.error as e:
        if e.winerror == NOT_RUNNING_RETURNCODE:
            raise service.NotRunningError(name)
        if e.winerror == ALREADY_STARTED_RETURNCODE:
            raise service.AlreadyRunningError(name)
        if e.winerror == NOT_INSTALLED_RETURNCODE:
            raise service.NotInstalledError(name)
        raise service.Error("Could not complete command.\n{}".format(e))
    finally:
        if handle:
            win32service.CloseServiceHandle(handle)


def _get_service_status(handle):
    """
    Returns Windows service status mapped to Fatmouse's inner states.
    """
    internal_state = win32service.QueryServiceStatusEx(handle)["CurrentState"]
    state = status_map[internal_state]
    if state == service.ServiceStatus.unknown:
        LOG.warning("Service is in an unknown state: '%s'", internal_state)
    return state


def _wait_status(name, status, handle):
    """
    Waits for service provided by handle to enter given status.
    name param is needed for logging only.
    """
    LOG.debug(
        "Waiting for service '%s' to become '%s' (timeout=%s sec)...",
        name,
        status,
        STATUS_CHANGE_TIMEOUT,
    )
    error_msg = (
        f"'{name}' service failed to become '{status}' "
        f"after {STATUS_CHANGE_TIMEOUT}-second timeout."
    )

    def check_service_status():
        """Checks that the service status has reached the expected status.
        """
        current_status = _get_service_status(handle)
        LOG.debug(
            "Checking '%s' service status, current status '%s', expected '%s'.",
            name,
            current_status,
            status,
        )
        return current_status == status

    mixutil.try_until(
        check_service_status,
        error_msg=error_msg,
        log=LOG,
        interval=STATUS_POLL_INTERVAL,
        timeout=STATUS_CHANGE_TIMEOUT,
    )


class WindowsServiceProvider(service.ABCServiceProvider):
    has_restart = True
    has_reload = False
    has_kill = False
    has_configtest = False

    def start_service(self, name):
        with service_handle(name) as handle:
            LOG.debug("Starting service '%s' ...", name)
            try:
                win32service.StartService(handle, None)
            except pywintypes.error as e:
                if e.winerror == ALREADY_STARTED_RETURNCODE:
                    LOG.debug("Service '%s' is already started. Nothing to do", name)
                    return
                raise
            _wait_status(name, service.ServiceStatus.running, handle)
            LOG.debug("Service '%s' is successfully started", name)

    def stop_service(self, name):
        with service_handle(name) as handle:
            LOG.debug("Stopping service '%s' ...", name)
            try:
                win32service.ControlService(handle, win32service.SERVICE_CONTROL_STOP)
            except pywintypes.error as e:
                if e.winerror == NOT_RUNNING_RETURNCODE:
                    LOG.debug("Service '%s' is already stopped. Nothing to do", name)
                    return
                raise
            _wait_status(name, service.ServiceStatus.stopped, handle)
            LOG.debug("Service '%s' is successfully stopped", name)

    def restart_service(self, name):
        if self.get_service_status(name) != service.ServiceStatus.stopped:
            self.stop_service(name)
        self.start_service(name)

    def set_service_enabled(self, name, enabled):
        if enabled:
            start_mode = win32service.SERVICE_AUTO_START
        else:
            start_mode = win32service.SERVICE_DEMAND_START

        with service_handle(name) as handle:
            win32service.ChangeServiceConfig(
                handle,
                win32service.SERVICE_NO_CHANGE,
                start_mode,
                win32service.SERVICE_NO_CHANGE,
                None,
                None,
                0,
                None,
                None,
                None,
                None,
            )

    def get_service_enabled(self, name):
        with service_handle(
            name, win32service.SERVICE_QUERY_CONFIG
        ) as handle:  # pylint: disable=no-member
            cfg = win32service.QueryServiceConfig(handle)
            return start_mode_map[cfg[1]] in ("Boot", "System", "Automatic")

    def get_service_status(self, name):
        with service_handle(
            name, win32service.SERVICE_QUERY_STATUS
        ) as handle:  # pylint: disable=no-member
            return _get_service_status(handle)
