"""
Manages Amazon Elastic File System storage.
"""
from __future__ import absolute_import

# pylint: disable=import-error
from libcloud.compute.drivers import ec2 as ec2_compute

from agent.tasks import pkgmgr
from agent.tasks.storage import base as storage_base
from agent.tasks.storage import utils as storage_utils
from common import types
from common.utils import exc
from common.utils import logutil
from common.utils.facts import fact

# pylint: enable=import-error


LOG = logutil.get_logger(__name__)


class EFSStorage(storage_base.ABCStorage):
    """
    Amazon Elastic File System storage.
    """

    supports_attach = False
    supports_detach = False
    supports_create_volume = False
    supports_create_snapshot = False

    volume_type_cls = types.EFSVolume
    cloud_server_cls = types.EC2CloudServer
    cloud_credentials_cls = types.AwsCloudCredentials
    context_cls = types.StorageContext

    def prepare_storage(self, ctx):

        if ctx.volume.id is None:
            raise exc.StorageError("Amazon EFS file system ID is not specified.")

        if fact["os"]["family"] == "windows":
            raise exc.NotSupportedError("Amazon EFS is not supported on Microsoft Windows.")

        if not fact["manage_container_host"]:  # nfs-utils is installed in coreOS by default
            if fact["os"]["family"] == "redhat":
                pkgmgr.install("nfs-utils")
            else:
                pkgmgr.install("nfs-common")

        result = types.StorageLocals(dict(volume=ctx.volume, libcloud_credentials=None))

        # collecting mount target ips in every available subnet
        subnet_ips = {
            value["SubnetId"]: value["IpAddress"] for value in self._list_mount_targets(ctx)
        }

        # finding the first subnet with the same availability zone our instance is running in
        # https://docs.aws.amazon.com/efs/latest/ug/accessing-fs.html
        for subnet in ctx.driver.ex_list_subnets(subnet_ids=subnet_ips.keys()):
            if subnet.extra["zone"] == ctx.cloud_server.availability_zone:
                ctx.volume.mount_target = "{host}:/".format(host=subnet_ips[subnet.id])
                result.mount_filesystem = True
                return result

        raise exc.StorageError(
            "Mount target for {} not found in availability zone {}.".format(
                ctx.volume, ctx.cloud_server.availability_zone
            )
        )

    @staticmethod
    def _list_mount_targets(ctx):
        """
        Returns the descriptions of all the current mount
        targets for a file system.

        :rtype: list[dict]
        """
        efs_driver = ec2_compute.EFSDriver(
            *ctx.libcloud_credentials.args, **ctx.libcloud_credentials.kwargs
        )
        return efs_driver.describe_mount_targets(file_system_id=ctx.volume.id)

    @staticmethod
    def mount_volume(volume, fstab=True, raise_already_mounted=False):
        """
        Mounts the ``StorageVolume`` and ensure that
        OS will do re-mount after reboot.

        :param volume: A volume to mount.
        :type volume: :class:`common.types.Volume`

        :param fstab: re-mount after reboot.
        :type fstab: bool

        :param raise_already_mounted: If True, raises an error if device is already mounted.
        :type raise_already_mounted: bool

        :rtype :class:`common.types.StorageLocals`
        """
        if not (volume.mpoint and volume.mount_target and volume.fstype):
            raise AttributeError(
                (
                    "{} has 'mpoint', 'fstype' or 'mount_target' attribute unset.\n"
                    "(mpoint={}, mount_target={}, fstype={})"
                ).format(volume, volume.mpoint, volume.mount_target, volume.fstype)
            )

        storage_utils.mount_device(
            volume.mount_target,
            volume.mpoint,
            volume.fstype,
            raise_already_mounted=raise_already_mounted,
            mount_options=volume.mount_options,
            fstab=fstab,
        )
        return types.StorageLocals(dict(volume=volume))

    def _is_volume_attached_to_node(self, volume, node):
        raise exc.NotSupportedError()

    def _should_clone_volume(self, volume, node):
        raise exc.NotSupportedError()

    def _get_attachment_state(self, volume):
        raise exc.NotSupportedError()

    def _get_volume_device(self, volume, node):
        raise exc.NotSupportedError()

    @classmethod
    def _find_node(cls, ctx):
        raise exc.NotSupportedError()

    @classmethod
    def _find_volume(cls, ctx):
        raise exc.NotSupportedError()

    @classmethod
    def _find_snapshot(cls, ctx):
        raise exc.NotSupportedError()

    @classmethod
    def _find_location(cls, ctx):
        raise exc.NotSupportedError()
