"""
Manages azure storage.
"""
from __future__ import absolute_import

import os
import re
import uuid

from libcloud.common import exceptions as libcloud_exc
from libcloud.compute import base as compute_base

from agent.tasks.storage import base as storage_base
from agent.tasks.storage import utils
from common import types
from common.utils import exc
from common.utils import libcloudutil
from common.utils import logutil
from common.utils.facts import fact

LOG = logutil.get_logger(__name__)


class AzureStorage(storage_base.ABCStorage):
    """
    Azure (ARM) Managed Disk storage.
    """

    supports_attach = True
    supports_detach = True
    supports_create_volume = True
    supports_create_snapshot = True

    volume_type_cls = types.AzureVolume
    cloud_server_cls = types.AzureCloudServer
    cloud_credentials_cls = types.AzureCloudCredentials
    context_cls = types.StorageContext

    case_insensitive_ids = True

    ATTACHMENT_STATE_MAP = {
        "attached": storage_base.StorageVolumeAttachmentState.ATTACHED,
        "unattached": storage_base.StorageVolumeAttachmentState.DETACHED,
        "activesas": storage_base.StorageVolumeAttachmentState.ACCESSING,
        "reserved": storage_base.StorageVolumeAttachmentState.RESERVED,
    }

    @classmethod
    def _find_node(cls, ctx):
        try:
            return ctx.driver.ex_get_node(ctx.cloud_server.id)
        except libcloud_exc.BaseHTTPError as e:
            if re.search(libcloudutil.AZURE_NOT_FOUND_CODE_REGEX, e.message):
                raise exc.NotFoundError(e.message)
            raise

    @classmethod
    def _find_volume(cls, ctx):
        try:
            return ctx.driver.ex_get_volume(ctx.volume.id)
        except libcloud_exc.BaseHTTPError as e:
            if re.search(libcloudutil.AZURE_NOT_FOUND_CODE_REGEX, e.message):
                raise exc.NotFoundError(e.message)
            raise

    @classmethod
    def _find_snapshot(cls, ctx):
        try:
            return ctx.driver.ex_get_snapshot(ctx.volume.snap.id)
        except libcloud_exc.BaseHTTPError as e:
            if re.search(libcloudutil.AZURE_NOT_FOUND_CODE_REGEX, e.message):
                raise exc.NotFoundError(e.message)
            raise

    @classmethod
    def _find_location(cls, ctx):
        return cls._find_resource(
            ctx, compute_base.NodeLocation, query={"id": ctx.cloud_server.location}
        )

    def _is_volume_attached_to_node(self, volume, node):
        managed_by = volume.extra["properties"].get("managedBy")
        if managed_by and node.id:
            return managed_by.lower() == node.id.lower()
        return False

    def _should_clone_volume(self, volume, node):
        if (
            self._get_attachment_state(volume)
            == storage_base.StorageVolumeAttachmentState.ACCESSING
        ):
            LOG.debug("%s is in use via SAS Uri.", volume)
            return True

        volume_location = volume.extra["location"].lower()
        node_location = node.extra["location"].lower()

        if volume_location != node_location:
            # TODO implement volume migration between locations
            raise storage_base.StorageError(
                (
                    "StorageVolume (id={}, location={}) and Instance (id={}, location={}) "
                    "are in different locations.\n"
                    "Currently volume migration between different locations "
                    "is not supported by Azure storage provider."
                ).format(volume.id, volume_location, node.id, node_location)
            )

        return False

    def _get_attachment_state(self, volume):
        disk_state = volume.extra["properties"].get("diskState", "").lower()
        return self.ATTACHMENT_STATE_MAP.get(
            disk_state, storage_base.StorageVolumeAttachmentState.UNKNOWN
        )

    def _get_volume_device(self, volume, node):
        """
        Return device of attached volume.
        """
        for disk in node.extra["properties"]["storageProfile"]["dataDisks"]:
            if disk["name"] == volume.name:
                volume_lun = int(disk["lun"])
                return utils.azure_get_device(volume_lun)

        raise storage_base.StorageError(
            ("Could not determine the device name. {} is not attached to the instance.").format(
                volume
            )
        )

    def _call_create_volume(self, ctx):
        """
        Call `create_volume` via libcloud driver.
        """
        location = self._find_location(ctx)
        return ctx.driver.create_volume(
            ctx.volume.size,
            ctx.volume.name,
            location=location,
            snapshot=ctx.volume.snap,
            ex_resource_group=os.path.basename(ctx.volume.resource_group),
            # In the new Azure API the name of disk type parameter is sku
            # but Scalr still stores it as `account_type`
            ex_sku_name=ctx.volume.account_type,
            ex_tags=ctx.volume.tags,
            ex_zones=ctx.volume.zones,
            ex_disk_encryption_set=ctx.volume.disk_encryption_set or None,
        )

    def _call_create_volume_snapshot(self, ctx, **kwargs):
        """
        Call `create_volume_snapshot` via libcloud driver.
        """
        location = self._find_location(ctx)
        return ctx.driver.create_volume_snapshot(
            ctx.volume,
            name="snap-{}".format(uuid.uuid4()),
            location=location,
            ex_resource_group=os.path.basename(ctx.volume.resource_group),
            ex_tags=ctx.volume.snap.tags if ctx.volume.snap else None,
        )

    def _call_detach_volume(self, ctx):
        """
        Call `detach_volume` via libcloud driver.
        """
        libcloud_volume = self._find_volume(ctx)
        if libcloud_volume.state != compute_base.StorageVolumeState.INUSE:
            LOG.debug("%s is unattached, skip detaching", ctx.volume)
            return

        owner_node = self._find_resource(
            ctx, compute_base.Node, query={"id": libcloud_volume.extra["properties"]["managedBy"]}
        )
        if owner_node.state == compute_base.NodeState.TERMINATED:
            LOG.debug("%s owner node is already in deletion state, skip detaching", ctx.volume)
            return

        ctx.driver.detach_volume(libcloud_volume, ex_node=owner_node)

    def _wait_volume_update(self, ctx):
        pass

    @staticmethod
    def mount_volume(volume, fstab=True, raise_already_mounted=False):
        if fact["os"]["family"] != "windows":
            # On Azure volume devices may be scrambled after server reboot.
            # To work around that we update volume.device and mount
            # volumes by /dev/disk/by-uuid link
            volume.device = utils.azure_get_persistent_device_path(volume.device)
        return super(AzureStorage, AzureStorage).mount_volume(
            volume, fstab=fstab, raise_already_mounted=raise_already_mounted
        )
