"""
Manages azure VHD storage.
"""
from __future__ import absolute_import

import os
import re

# pylint: disable=import-error
from libcloud.common import exceptions as libcloud_exc
from libcloud.compute import base as compute_base
from libcloud.storage import types as storage_types
from libcloud.storage.drivers import azure_blobs

from agent.tasks.storage import base as storage_base
from agent.tasks.storage import utils
from common import types
from common.utils import exc
from common.utils import libcloudutil
from common.utils import logutil
from common.utils import mixutil
from common.utils.facts import fact

# pylint: enable=import-error


LOG = logutil.get_logger(__name__)


class AzureVHDStorage(storage_base.ABCStorage):
    """
    Azure VHD (unmanaged) storage.
    """

    supports_attach = True
    supports_detach = True
    supports_create_volume = False
    supports_create_snapshot = False

    volume_type_cls = types.AzureVHDVolume
    cloud_server_cls = types.AzureCloudServer
    cloud_credentials_cls = types.AzureCloudCredentials
    context_cls = types.StorageContext

    case_insensitive_ids = True

    def prepare_storage(self, ctx):
        result = types.StorageLocals(
            dict(
                volume=ctx.volume,
                libcloud_credentials=ctx.libcloud_credentials,
                attach_volume=True,
                detach_volume=False,
                create_snapshot=False,
                create_volume=False,
            )
        )

        # wait for node provision state to be 'running'
        self._wait_node_state_change(
            ctx, to_state=compute_base.NodeState.RUNNING, abort_on=(compute_base.NodeState.ERROR,)
        )
        libcloud_node = self._find_node(ctx)

        if ctx.volume.id is None:
            ctx.volume.id = mixutil.generate_uid(prefix="vol-{}-".format(ctx.volume.type))
        else:
            LOG.debug("Checking that VHD blob (uri=%s) exists.", ctx.volume.vhd_uri)
            if not self._vhd_blob_exists(ctx):
                if ctx.volume.recreate_if_missing:
                    LOG.warning(
                        'VHD blob (uri=%s) does not exist. "Recreate missing volumes" '
                        "option is checked, so a new volume will be made "
                        "from the initial storage configuration.",
                        ctx.volume.vhd_uri,
                    )
                    if ctx.volume.template:
                        result.volume = self.volume_type_cls(ctx.volume.template)
                    if result.volume.id is None:
                        result.volume.id = mixutil.generate_uid(
                            prefix="vol-{}-".format(ctx.volume.type)
                        )
                    result.attach_volume = True
                    return result
                else:
                    raise exc.NotFoundError(
                        "VHD blob (uri={}) does not exist.".format(ctx.volume.vhd_uri)
                    )
            else:
                LOG.debug("VHD blob (uri=%s) already exists.", ctx.volume.vhd_uri)

        try:
            LOG.debug("Checking that %s exists.", ctx.volume)
            libcloud_volume = self._find_volume(ctx)
        except exc.NotFoundError:
            LOG.debug("%s will be created and attached.", ctx.volume)
            result.attach_volume = True
            return result
        else:
            LOG.debug("%s already exists.", ctx.volume)

        if self._is_volume_attached_to_node(libcloud_volume, libcloud_node):
            LOG.debug("%r already attached.", ctx.volume)
            result.volume.device = self._get_volume_device(libcloud_volume, libcloud_node)
            result.volume.managed_by = libcloud_node.id
            result.attach_volume = False
        else:
            LOG.debug("%s in use by different instance.", ctx.volume)
            result.volume.managed_by = libcloud_volume.extra["properties"]["managedBy"]
            result.detach_volume = True

        return result

    def _wait_for_volume_attachment(self, ctx):
        # azure unmanaged volumes does not have an attachment state,
        # when the volume is attaching to the node, his state becomes 'updating'
        self._wait_node_state_change(
            ctx, to_state=compute_base.NodeState.RUNNING, abort_on=(compute_base.NodeState.ERROR,)
        )

    def _wait_for_volume_detachment(self, ctx):
        # azure unmanaged volumes does not have an attachment state,
        # when the volume is detaching from the node, his state becomes 'updating'
        self._wait_owner_node_state_change(
            ctx, to_state=compute_base.NodeState.RUNNING, abort_on=(compute_base.NodeState.ERROR,)
        )

    def _call_attach_volume(self, ctx):
        """
        Call `attach_volume` via libcloud driver.
        """
        node = self._find_node(ctx)
        ctx.driver.attach_volume(
            node,
            ctx.volume,
            ex_vhd_uri=ctx.volume.vhd_uri,
            ex_vhd_create=not self._vhd_blob_exists(ctx),
        )

    def _call_detach_volume(self, ctx):
        """
        Call `detach_volume` via libcloud driver.
        """
        libcloud_volume = self._find_volume(ctx)
        owner_node = self._find_owner_node(ctx)
        ctx.driver.detach_volume(libcloud_volume, ex_node=owner_node)

    def _is_volume_attached_to_node(self, volume, node):
        return self._get_attached_volume_lun(volume, node) is not None

    def _should_clone_volume(self, volume, node):
        raise exc.NotSupportedError()

    def _get_attachment_state(self, volume):
        raise exc.NotSupportedError()

    def _get_volume_device(self, volume, node):
        """
        Return device of attached volume.

        :type volume: :class:`compute_base.StorageVolume`
        :type node: :class:`compute_base.Node`

        :rtype: str or :class:`types.SCSIAddress` or None
        """
        volume_lun = self._get_attached_volume_lun(volume, node)
        if volume_lun is not None:
            return utils.azure_get_device(volume_lun)
        raise storage_base.StorageError(
            ("Could not determine the device name. " "{} is not attached to the instance.").format(
                volume
            )
        )

    def _get_attached_volume_lun(self, volume, node):
        """
        Returns a LUN if volume attached to a node, otherwise ``None``.

        :type volume: :class:`compute_base.StorageVolume`
        :type node: :class:`compute_base.Node`

        :rtype: int or None
        """
        for disk in node.extra["properties"]["storageProfile"]["dataDisks"]:
            if "vhd" in disk and disk["vhd"].get("uri") == volume.extra["vhd"]["uri"]:
                return int(disk["lun"])

    @classmethod
    def _vhd_blob_exists(cls, ctx):
        """
        Check if volume VHD blob exists.

        :rtype: ``bool``
        """
        # Example URI:
        # https://{storage_account}.blob.core.windows.net/{blob_container}/{blob_name}.vhd
        uri = ctx.volume.vhd_uri.split("/")
        storage_host = uri[2]
        storage_account = uri[2].split(".", 1)[0]
        blob_container = uri[3]
        blob_name = "/".join(uri[4:])
        try:
            secret = ctx.driver.ex_get_storage_account_keys(
                os.path.basename(ctx.volume.resource_group), storage_account
            )["key1"]
            proxy_url = ctx.libcloud_credentials.kwargs.get("proxy_url")

            blob_driver = azure_blobs.AzureBlobsStorageDriver(
                storage_account, secret=secret, host=storage_host, proxy_url=proxy_url, timeout=60
            )
            blob_driver.get_object(blob_container, blob_name)
            return True
        except (storage_types.ObjectDoesNotExistError, storage_types.ContainerDoesNotExistError):
            return False

    @classmethod
    def _find_node(cls, ctx):
        try:
            return ctx.driver.ex_get_node(ctx.cloud_server.id)
        except libcloud_exc.BaseHTTPError as e:
            if re.search(libcloudutil.AZURE_NOT_FOUND_CODE_REGEX, e.message):
                raise exc.NotFoundError(e.message)
            raise

    @classmethod
    def _find_owner_node(cls, ctx):
        try:
            return ctx.driver.ex_get_node(ctx.volume.managed_by)
        except libcloud_exc.BaseHTTPError as e:
            if re.search(libcloudutil.AZURE_NOT_FOUND_CODE_REGEX, e.message):
                raise exc.NotFoundError(e.message)
            raise

    @classmethod
    def _find_volume(cls, ctx):
        for node in ctx.driver.list_nodes():
            for disk in node.extra["properties"]["storageProfile"]["dataDisks"]:
                if "vhd" not in disk:
                    continue
                if disk["vhd"].get("uri") == ctx.volume.vhd_uri:
                    return compute_base.StorageVolume(
                        id=None,  # unmanaged volumes do not have IDs
                        name=disk["name"],
                        size=disk.get("diskSizeGB"),
                        driver=ctx.driver,
                        state=compute_base.StorageVolumeState.INUSE,
                        extra=dict(disk, properties={"managedBy": node.id}),
                    )
        raise exc.NotFoundError("Could not find the {} on the cloud.".format(ctx.volume))

    def _wait_owner_node_state_change(self, ctx, to_state=None, ignore_errors=None, **kwds):
        return libcloudutil.wait_state_change(
            object_type=compute_base.Node,
            object_getter=lambda: self._find_owner_node(ctx),
            to_state=to_state,
            ignore_errors=ignore_errors,
            case_insensitive_id=self.case_insensitive_ids,
            driver=ctx.driver,
            **kwds,
        )

    def _wait_volume_update(self, ctx):
        pass

    @staticmethod
    def mount_volume(volume, fstab=True, raise_already_mounted=False):
        if fact["os"]["family"] != "windows":
            # On Azure volume devices may be scrambled after server reboot.
            # To work around that we update volume.device and mount
            # volumes by /dev/disk/by-uuid link
            volume.device = utils.azure_get_persistent_device_path(volume.device)
        return super(AzureVHDStorage, AzureVHDStorage).mount_volume(
            volume, fstab=fstab, raise_already_mounted=raise_already_mounted
        )
