"""
This module provides abstract class :class:`ABCStorage`.
"""
from __future__ import absolute_import

import abc

# pylint: disable=import-error
from libcloud.compute import base as compute_base
from libcloud.compute import types as compute_types

from agent.tasks.filesystem import tasks as filesystem_tasks
from agent.tasks.storage import utils as storage_utils
from common import types
from common.utils import exc
from common.utils import libcloudutil
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils import sysutil
from common.utils.facts import fact

# pylint: enable=import-error


if fact["os"]["family"] == "windows":
    from common.utils.winutil import umount as umount_device  # pragma: no cover
else:
    from common.utils.linuxutil import umount as umount_device


LOG = logutil.get_logger(__name__)


class StorageError(exc.Error):
    """
    Exception class for storage tasks.
    """


class StorageVolumeAttachmentState(object):
    """
    Attachment states of a StorageVolume.
    """

    ATTACHING = "attaching"
    ATTACHED = "attached"
    DETACHING = "detaching"
    DETACHED = "detached"
    ACCESSING = "accessing"
    RESERVED = "reserved"
    UNKNOWN = "unknown"


# pylint: disable=no-self-use
@six.add_metaclass(abc.ABCMeta)
class ABCStorage(object):
    """
    Abstract class for storage implementations.
    """

    supports_attach = False
    supports_detach = False
    supports_create_volume = False
    supports_create_snapshot = False

    midstates = (
        compute_base.StorageVolumeState.CREATING,
        compute_base.StorageVolumeState.UPDATING,
        compute_base.StorageVolumeState.DELETING,
        compute_base.StorageVolumeState.MIGRATING,
        compute_base.StorageVolumeState.ATTACHING,
    )

    abort_states = (compute_base.StorageVolumeState.ERROR, compute_base.StorageVolumeState.DELETED)

    attachment_midstates = (
        StorageVolumeAttachmentState.ATTACHING,
        StorageVolumeAttachmentState.DETACHING,
    )

    ATTACHMENT_STATE_MAP = {}  # type: dict

    volume_type_cls = types.Volume
    snap_type_cls = types.Snapshot
    cloud_server_cls = types.CloudServer
    cloud_credentials_cls = types.CloudCredentials
    context_cls = types.StorageContext

    case_insensitive_ids = False

    def prepare_storage(self, ctx):
        """
        Configures :class:`types.StorageLocals` before ``StorageVolume`` creation.

        Prepare steps:
         - Set defaults:\n
           - ``attach_volume = True``
           - ``detach_volume = False``
           - ``create_snapshot = False``
           - ``create_volume = False``

         - If ``volume.id is None`` set:\n
           - ``create_volume = True``
           and return.

         - Validate existing ``volume.id``, and when volume doesn't
           exist in the cloud and ``volume.recreate_if_missing is True`` set:\n
           - ``create_volume = True``
           and return, else raise exception.

         - Wait if volume in midstates (creating, updating, attaching, etc.).

         - If volume is already attached to current ``ctx.cloud_server``:\n
           - ``attach_volume = False``

         - If volume in use in the same location:\n
           - ``detach_volume = True``

         - If volume in use in another location:\n
           - ``create_snapshot = True``
           - ``create_volume = True``

        :type ctx: :class:`common.types.StorageContext`

        :raises: :class:`exc.NotFoundError`
        :rtype: :class:`common.types.StorageLocals`
        """
        # set defaults
        result = types.StorageLocals(
            dict(
                volume=ctx.volume,
                libcloud_credentials=ctx.libcloud_credentials,
                attach_volume=True,
                detach_volume=False,
                create_snapshot=False,
                create_volume=False,
            )
        )

        # return if `volume.id is None`
        if ctx.volume.id is None:
            ctx.volume.id = mixutil.generate_uid(prefix="vol-{}-".format(ctx.volume.type))
            result.create_volume = True
            return result

        # validate existing `volume.id`
        try:
            LOG.debug("Checking that %s exists.", ctx.volume)
            libcloud_volume = self._find_volume(ctx)
        except exc.NotFoundError as e:
            if ctx.volume.recreate_if_missing:
                LOG.warning(
                    '%s does not exist. "Recreate missing volumes" '
                    "option is checked, so a new volume will be made "
                    "from the initial storage configuration.",
                    ctx.volume,
                )
                if ctx.volume.template:
                    result.volume = self.volume_type_cls(ctx.volume.template)
                if result.volume.id is None:
                    result.volume.id = mixutil.generate_uid(
                        prefix="vol-{}-".format(ctx.volume.type)
                    )
                result.create_volume = True
                return result
            else:
                raise e
        else:
            LOG.debug("%s already exists.", ctx.volume)

        # wait if volume state in `self.midstates`
        if libcloud_volume.state in self.midstates:
            self._wait_volume_state_change(
                ctx, from_state=libcloud_volume.state, abort_on=self.abort_states
            )
            LOG.debug(
                "%s provision state transition completed.\n" "Updating volume object.", ctx.volume
            )
            libcloud_volume = self._find_volume(ctx)

        # wait if volume attachment state in `self.attachment_midstates`
        attachment_state = self._get_attachment_state(libcloud_volume)
        if attachment_state in self.attachment_midstates:
            self._wait_attachment_state_change(ctx, from_state=attachment_state)
            LOG.debug("%s disk state transition completed.\n" "Updating volume object.", ctx.volume)
            libcloud_volume = self._find_volume(ctx)

        libcloud_node = self._find_node(ctx)

        # volume is already attached to an instance
        if self._is_volume_attached_to_node(libcloud_volume, libcloud_node):
            LOG.debug("%s already attached.", ctx.volume)
            result.volume.device = self._get_volume_device(libcloud_volume, libcloud_node)
            result.attach_volume = False

        # volume should be cloned
        elif self._should_clone_volume(libcloud_volume, libcloud_node):
            LOG.debug("%s will be cloned.", ctx.volume)
            result.create_snapshot = True
            result.create_volume = True

        # volume in use by different instance
        elif libcloud_volume.state == compute_types.StorageVolumeState.INUSE:
            LOG.debug("%s in use by different instance.", ctx.volume)
            result.detach_volume = True

        return result

    def create_volume(self, ctx, wait_available=True):
        """
        Creates the ``StorageVolume`` on a cloud.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :param wait_available: Waiting for volume to become `available`.
        :param wait_available: bool

        :rtype :class:`common.types.StorageLocals`
        """
        if not self.supports_create_volume:
            raise exc.NotSupportedError(
                ("Creating a volume is not supported by '{}' provider.").format(
                    self.__class__.__name__
                )
            )

        created_volume = self._call_create_volume(ctx)
        ctx.volume.id = created_volume.id
        LOG.info("%s created.", ctx.volume)

        if wait_available:
            LOG.debug("Wait available state of %s.", ctx.volume)
            self._wait_volume_state_change(
                ctx,
                to_state=compute_types.StorageVolumeState.AVAILABLE,
                ignore_errors=(exc.NotFoundError,),
            )
            LOG.debug("%s is available.", ctx.volume)

        return types.StorageLocals(dict(volume=ctx.volume))

    def create_snapshot(self, ctx, wait_available=True, **kwargs):
        """
        Creates the ``VolumeSnapshot`` on a cloud.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :param wait_available: Waiting for ``VolumeSnapshot`` to become `available`.
        :param wait_available: bool

        :rtype :class:`common.types.StorageLocals`
        """
        if not self.supports_create_snapshot:
            raise exc.NotSupportedError(
                ("Creating a snapshot is not supported by '{}' provider.").format(
                    self.__class__.__name__
                )
            )

        created_snap = self._call_create_volume_snapshot(ctx, **kwargs)
        ctx.volume.snap = self.snap_type_cls.from_libcloud_snapshot(
            created_snap, volume_id=ctx.volume.id
        )
        LOG.debug("%s created.", ctx.volume.snap)

        if wait_available:
            self._wait_snapshot_state_change(
                ctx,
                to_state=compute_types.VolumeSnapshotState.AVAILABLE,
                ignore_errors=(exc.NotFoundError,),
            )
            created_snap = self._find_snapshot(ctx)
            ctx.volume.snap.state = created_snap.state
            ctx.volume.snap.size = created_snap.size
            ctx.volume.snap.created = created_snap.created

        return types.StorageLocals(dict(volume=ctx.volume))

    def delete_snapshot_maybe(self, ctx):
        """
        Deletes the ``VolumeSnapshot``.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :rtype :class:`common.types.StorageLocals`
        """
        if ctx.volume.snap:
            self._call_destroy_volume_snapshot(ctx)
            ctx.volume.snap = None
        return types.StorageLocals(dict(volume=ctx.volume))

    def attach_volume(self, ctx):
        """
        Attaches the ``StorageVolume`` to the ``Instance``.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :rtype :class:`common.types.StorageLocals`
        """
        if not self.supports_attach:
            raise exc.NotSupportedError(
                ("Device attach is not supported by '{}' provider.").format(self.__class__.__name__)
            )
        result = types.StorageLocals(dict(volume=ctx.volume))

        with self._attaching_device(ctx) as attachment:
            LOG.info("Attaching %s to the %s.", ctx.volume, ctx.cloud_server)
            self._call_attach_volume(ctx)
            self._wait_for_volume_attachment(ctx)

        result.volume.device = attachment.device_name
        LOG.info("%s mapped to device: %s", ctx.volume, result.volume.device)

        result = self._check_volume_filesystem(result)
        result.mount_filesystem = True
        return result

    def _check_volume_filesystem(self, result):
        """
        Checks for partition table and file system type.
        """
        volume = result.volume
        if volume.snap and sysutil.has_partition_table(volume.device):
            raise exc.Error(
                (
                    "{} source snapshot {} was taken from a "
                    "partitioned device. Scalr Storage does not support "
                    "devices with partition table."
                ).format(volume, volume.snap)
            )

        fs_type = sysutil.get_fstype(volume.device)
        if fs_type is None:
            result.create_filesystem = True
        elif fs_type != volume.fstype:
            warn_msg = (
                "Device %s already has %r filesystem on it. Ignoring "
                "%r fstype passed in %s configuration."
            )
            LOG.warning(warn_msg, volume.device, fs_type, volume.fstype, volume)
            result.volume.fstype = fs_type
            result.create_filesystem = False
        return result

    def detach_volume(self, ctx):
        """
        Detaches the ``StorageVolume`` from the ``Instance``.

        Do nothing if volume is available and doesn't have
        any attachment state (i.e. is not attached).

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :rtype :class:`common.types.StorageLocals`
        """
        if not self.supports_detach:
            raise exc.NotSupportedError(
                ("Device detach is not supported by '{}' provider.").format(self.__class__.__name__)
            )

        result = types.StorageLocals(dict(volume=ctx.volume))

        # ensure that we have unmounted the storage volume that we want to detach
        if ctx.volume.device is not None:
            libcloud_volume = self._find_volume(ctx)
            libcloud_node = self._find_node(ctx)
            if self._is_volume_attached_to_node(libcloud_volume, libcloud_node):
                LOG.info("Unmounting %s before detaching.", ctx.volume)
                umount_device(ctx.volume.device)

        # XXX: when we attempt to detach a storage volume that is mounted,
        # the node to which the volume is mounted may come to a fault state
        LOG.info("Trying to detach %s.", ctx.volume)
        self._call_detach_volume(ctx)
        self._wait_for_volume_detachment(ctx)

        result.attach_volume = True
        LOG.info("%s detached.", ctx.volume)

        return result

    def prepare_volume_update(self, ctx, update_parameters):
        """
        Prepares volume for update. Checks that update is possible.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :param update_parameters: Parameters to update.
        :type update_parameters: :class:`common.types.VolumeModification`

        :rtype: :class:`common.types.StorageLocals`
        """
        result = types.StorageLocals({"volume": ctx.volume})
        self._prepare_volume_update(ctx, update_parameters)
        LOG.info("Volume %s is prepared to be updated.", ctx.volume.id)

        return result

    def update_volume(self, ctx, update_parameters):
        """
        Updates volume parameters on clouds that support volume modification.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :param update_parameters: Parameters to update.
        :type update_parameters: ``dict``

        :rtype: :class:`common.types.StorageLocals`
        """
        result = types.StorageLocals({"volume": ctx.volume})
        self._update_volume(ctx, update_parameters)
        LOG.info("Volume %s updated.", ctx.volume.id)

        return result

    def wait_volume_update(self, ctx):
        """
        Monitors volume after update and exits as soon as it's ready.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`

        :rtype: :class:`common.types.StorageLocals`
        """
        result = types.StorageLocals({"volume": ctx.volume})
        self._wait_volume_update(ctx)
        LOG.info("Volume %s is ready.", ctx.volume.id)

        return result

    @staticmethod
    def mount_volume(volume, fstab=True, raise_already_mounted=False):
        """
        Mounts the ``StorageVolume`` and ensure that
        OS will do re-mount after reboot.

        :param volume: A volume to mount.
        :type volume: :class:`common.types.Volume`

        :param fstab: re-mount after reboot (linux-only).
        :type fstab: bool

        :param raise_already_mounted: If True, raises an error if device is already mounted.
        :type raise_already_mounted: bool

        :rtype :class:`common.types.StorageLocals`
        """
        if not (volume.mpoint and volume.device and volume.fstype):
            raise AttributeError(
                (
                    "{} has 'mpoint', 'fstype' or 'device' attribute unset.\n"
                    "(mpoint={}, device={}, fstype={})"
                ).format(volume, volume.mpoint, volume.device, volume.fstype)
            )

        if volume.fstype not in filesystem_tasks.providers:
            raise exc.StorageError(
                (
                    "{volume} has unknown filesystem {fstype} and can't be mounted to {mpoint}."
                ).format(volume=volume, fstype=volume.fstype, mpoint=volume.mpoint)
            )

        LOG.info("Mounting %s to %s", volume.device, volume.mpoint)

        storage_utils.mount_device(
            volume.device,
            volume.mpoint,
            volume.fstype,
            raise_already_mounted=raise_already_mounted,
            mount_options=volume.mount_options,  # linux-only option
            fstab=fstab,  # linux-only option
            label=volume.label,  # windows-only option
        )

        return types.StorageLocals(dict(volume=volume))

    @classmethod
    def init_context(cls, volume, cloud_server, credentials):
        """
        Initializes a storage context.

        :param volume: A volume to operate on.
        :type volume: :class:`common.types.Volume`

        :param credentials: A credentials for `cloud_server`.
        :type credentials: :class:`common.types.CloudCredentials`

        :param cloud_server: A cloud server to operate on.
        :type cloud_server: :class:`common.types.CloudServer`

        :rtype: :class:`common.types.StorageContext`
        """
        LOG.debug("Initializing %s context.", cls.__name__)

        # validate volume
        volume = cls.volume_type_cls(volume)
        volume.validate()
        LOG.debug("StorageVolume to operate on within context %s.", volume.to_primitive())

        # validate cloud server
        cloud_server = cls.cloud_server_cls(cloud_server)
        cloud_server.validate()

        # validate cloud credentials
        credentials = cls.cloud_credentials_cls(credentials)
        credentials.validate()

        # initialize libcloud driver
        libcloud_credentials = credentials.to_libcloud(location=cloud_server.location)
        driver = libcloudutil.create_driver(libcloud_credentials, api="compute")

        return cls.context_cls(
            dict(
                driver=driver,
                volume=volume,
                cloud_server=cloud_server,
                libcloud_credentials=libcloud_credentials,
            )
        )

    @staticmethod
    def _attaching_device(ctx, timeout=None, polling_interval=None):
        """
        A context manager to handle & get attaching device information.

        It knows how to poll for an actual new device name that
        appears in OS (both Linux/Windows).

        You should supervise any ``libcloud_driver.attach_volume()``
        with this context manager.

        Example:

        .. code-block::

            with self._attaching_device(ctx) as attachment:
                ctx.driver.attach_volume(node, volume)
                # now it's an actual attached device name
                attachment.device_name

        :rtype: :class:`storage_utils.attaching_device`
        """
        return storage_utils.attaching_device(timeout=timeout, polling_interval=polling_interval)

    def _call_create_volume(self, ctx):
        """
        Call `create_volume` via libcloud driver.
        """
        location = self._find_location(ctx)
        libcloud_volume = ctx.driver.create_volume(
            ctx.volume.size, ctx.volume.id, location=location, snapshot=ctx.volume.snap
        )
        if ctx.volume.tags:
            ctx.driver.ex_create_tags(libcloud_volume, ctx.volume.tags)
        return libcloud_volume

    def _call_create_volume_snapshot(self, ctx, **kwargs):
        """
        Call `create_volume_snapshot` via libcloud driver.
        """
        return ctx.driver.create_volume_snapshot(ctx.volume)

    # pylint: disable=invalid-name
    def _call_destroy_volume_snapshot(self, ctx):
        """
        Call `destroy_volume_snapshot` via libcloud driver.
        """
        return ctx.driver.destroy_volume_snapshot(ctx.volume.snap)

    def _call_attach_volume(self, ctx):
        """
        Call `attach_volume` via libcloud driver.
        """
        node = self._find_node(ctx)
        ctx.driver.attach_volume(node, ctx.volume)

    def _call_detach_volume(self, ctx):
        """
        Call `detach_volume` via libcloud driver.
        """
        return ctx.driver.detach_volume(ctx.volume)

    def _wait_for_volume_attachment(self, ctx):
        """
        Waits until the volume state is attached.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`
        """
        self._wait_attachment_state_change(ctx, to_state=StorageVolumeAttachmentState.ATTACHED)

    def _wait_for_volume_detachment(self, ctx):
        """
        Waits until the volume state is detached.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`
        """
        self._wait_attachment_state_change(ctx, to_state=StorageVolumeAttachmentState.DETACHED)

    @classmethod
    def _find_node(cls, ctx):
        raise NotImplementedError()

    @classmethod
    def _find_volume(cls, ctx):
        raise NotImplementedError()

    @classmethod
    def _find_snapshot(cls, ctx):
        pass

    @classmethod
    def _find_location(cls, ctx):
        pass

    @classmethod
    def _find_resource(cls, ctx, object_type, query=None, api_call_kwds=None):
        return libcloudutil.find_one(
            object_type=object_type,
            query=query,
            ignore_case=cls.case_insensitive_ids,
            driver=ctx.driver,
            api_call_kwds=api_call_kwds,
        )

    def _wait_node_state_change(self, ctx, to_state=None, ignore_errors=None, **kwds):
        return libcloudutil.wait_state_change(
            object_type=compute_base.Node,
            object_getter=lambda: self._find_node(ctx),
            to_state=to_state,
            ignore_errors=ignore_errors,
            object_id=ctx.cloud_server.id,
            case_insensitive_id=self.case_insensitive_ids,
            driver=ctx.driver,
            **kwds,
        )

    def _wait_attachment_state_change(
        self, ctx, from_state=None, object_getter=None, to_state=None, **kwargs
    ):
        object_getter = lambda: self._find_volume(ctx)

        return libcloudutil.wait_state_change(
            object_type=compute_base.StorageVolume,
            object_getter=object_getter,
            from_state=from_state,
            to_state=to_state,
            state_getter=self._get_attachment_state,
            object_id=ctx.volume.id,
            case_insensitive_id=self.case_insensitive_ids,
            driver=ctx.driver,
            **kwargs,
        )

    def _wait_volume_state_change(
        self, ctx, from_state=None, to_state=None, abort_on=None, ignore_errors=None, **kwds
    ):
        return libcloudutil.wait_state_change(
            object_type=compute_base.StorageVolume,
            object_getter=lambda: self._find_volume(ctx),
            from_state=from_state,
            to_state=to_state,
            abort_on=abort_on,
            ignore_errors=ignore_errors,
            object_id=ctx.volume.id,
            case_insensitive_id=self.case_insensitive_ids,
            driver=ctx.driver,
            **kwds,
        )

    def _wait_snapshot_state_change(
        self, ctx, to_state=None, abort_on=None, ignore_errors=None, **kwds
    ):
        return libcloudutil.wait_state_change(
            object_type=compute_base.VolumeSnapshot,
            object_getter=lambda: self._find_snapshot(ctx),
            to_state=to_state,
            abort_on=abort_on,
            ignore_errors=ignore_errors,
            object_id=ctx.volume.snap.id,
            case_insensitive_id=self.case_insensitive_ids,
            driver=ctx.driver,
            **kwds,
        )

    @abc.abstractmethod
    def _is_volume_attached_to_node(self, volume, node):
        """
        Checks if volume is attached to the specified node.

        :type volume: :class:`compute_base.StorageVolume`
        :type node: :class:`compute_base.Node`

        :rtype: bool
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _should_clone_volume(self, volume, node):
        """
        If the volume cannot be attached (reattached) to node.

        :type volume: :class:`compute_base.StorageVolume`
        :type node: :class:`compute_base.Node`

        :rtype: bool
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _get_attachment_state(self, volume):
        """
        Return volume attachment state.

        :type volume: :class:`compute_base.StorageVolume`

        :return: value from :class:`StorageVolumeAttachmentState`
        :rtype: str
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def _get_volume_device(self, volume, node):
        """
        Return device of attached volume.

        :type volume: :class:`compute_base.StorageVolume`
        :type node: :class:`compute_base.Node`

        :rtype: str or :class:`types.SCSIAddress`
        """
        raise NotImplementedError()

    def _prepare_volume_update(self, ctx, parameters):
        """
        Prepares volume update.
        """

    def _update_volume(self, ctx, parameters):
        """
        Updates volume parameters on clouds that support volume modification.
        """

    def _wait_volume_update(self, ctx):
        """
        Waits for a volume to be ready
        """
