"""
Manages EBS storage.
"""
from __future__ import absolute_import

# pylint: disable=import-error
from libcloud.compute import base as compute_base
from libcloud.compute import types as compute_types

from agent.tasks.storage import base
from agent.tasks.storage import utils as storage_utils
from common import types
from common.utils import exc
from common.utils import libcloudutil
from common.utils import logutil
from common.utils import mixutil
from common.utils.facts import fact

# pylint: enable=import-error


LOG = logutil.get_logger(__name__)


class EbsStorage(base.ABCStorage):
    """
    Amazon Elastic Block Store class.
    """

    supports_attach = True
    supports_detach = True
    supports_create_volume = True
    supports_create_snapshot = True

    volume_type_cls = types.EbsVolume
    snap_type_cls = types.EbsSnapshot
    cloud_server_cls = types.CloudServer
    cloud_credentials_cls = types.AwsCloudCredentials
    context_cls = types.StorageContext

    update_polling_interval = 1
    update_timeout = 300

    ATTACHMENT_STATE_MAP = {
        "attaching": base.StorageVolumeAttachmentState.ATTACHING,
        "attached": base.StorageVolumeAttachmentState.ATTACHED,
        "detaching": base.StorageVolumeAttachmentState.DETACHING,
        "detached": base.StorageVolumeAttachmentState.DETACHED,
    }

    @classmethod
    def _find_node(cls, ctx):
        return cls._find_resource(
            ctx, compute_base.Node, api_call_kwds={"ex_node_ids": [ctx.cloud_server.id]}
        )

    @classmethod
    def _find_volume(cls, ctx):
        return cls._find_resource(
            ctx, compute_base.StorageVolume, api_call_kwds={"ex_volume_ids": [ctx.volume.id]}
        )

    @classmethod
    def _find_snapshot(cls, ctx):
        return cls._find_resource(
            ctx, compute_base.VolumeSnapshot, api_call_kwds={"snapshot": ctx.volume.snap}
        )

    @classmethod
    def _find_location(cls, ctx):
        return cls._find_resource(
            ctx, compute_base.NodeLocation, query={"id": ctx.cloud_server.location}
        )

    def _wait_for_volume_detachment(self, ctx):
        """
        Waits until the volume state is detached.

        :param ctx: Context for storage tasks.
        :type ctx: :class:`common.types.StorageContext`
        """
        libcloud_volume = self._find_volume(ctx)
        volume_owner = libcloud_volume.extra.get("instance_id")
        if volume_owner is None:
            return

        self._wait_attachment_state_change(
            ctx,
            object_getter=lambda: self._find_resource(
                ctx, compute_base.Node, api_call_kwds={"ex_node_ids": [volume_owner]}
            ),
            to_state=base.StorageVolumeAttachmentState.DETACHED,
        )

    def _wait_attachment_state_change(
        self, ctx, from_state=None, object_getter=None, to_state=None, **kwargs
    ):
        if object_getter is None:
            object_getter = lambda: self._find_node(ctx)

        return libcloudutil.wait_state_change(
            object_type=compute_base.StorageVolume,
            object_getter=object_getter,
            from_state=from_state,
            to_state=to_state,
            state_getter=lambda node: self._get_attachment_state_from_bdm(node, ctx.volume),
            object_id=ctx.volume.id,
            case_insensitive_id=self.case_insensitive_ids,
            driver=ctx.driver,
            **kwargs,
        )

    def _is_volume_attached_to_node(self, volume, node):
        return volume.extra["instance_id"] == node.id

    def _should_clone_volume(self, volume, node):
        volume_zone = volume.extra.get("zone")
        node_zone = node.extra.get("availability")

        if volume_zone != node_zone:
            msg = (
                "EbsVolume (id=%s, zone=%s) and the Instance "
                "(id=%s, zone=%s) are in different availability zones."
            )
            LOG.debug(msg, volume.id, volume_zone, node.id, node_zone)
            return True

        return False

    def _get_attachment_state_from_bdm(self, node, volume):
        for block_device in node.extra["block_device_mapping"]:
            if block_device["ebs"]["volume_id"] == volume.id:
                return self.ATTACHMENT_STATE_MAP.get(
                    block_device["ebs"]["status"], base.StorageVolumeAttachmentState.UNKNOWN
                )
        # If volume was not found in node's block_device_mapping assume it is detached
        return base.StorageVolumeAttachmentState.DETACHED

    def _get_attachment_state(self, volume):
        attachment_status = self.ATTACHMENT_STATE_MAP.get(
            volume.extra["attachment_status"], base.StorageVolumeAttachmentState.UNKNOWN
        )

        # FAM-889
        # Actual observed behavior on AWS: after volume is detached, attachment
        # state is empty (not DETACHED) and volume state is AVAILABLE.
        if (
            attachment_status == base.StorageVolumeAttachmentState.UNKNOWN
            and volume.state == compute_types.StorageVolumeState.AVAILABLE
        ):
            attachment_status = base.StorageVolumeAttachmentState.DETACHED

        return attachment_status

    def _get_volume_device(self, volume, node):
        if fact["os"]["family"] == "windows":
            # Trying to determine scsi address by serial number (volume id)/
            # https://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/nvme-ebs-volumes.html#identify-nvme-ebs-device
            # Volume id has format "vol-<uid>" and in WMI it has format "vol<uid>.*"
            try:
                return storage_utils.ec2_volume_id_to_device(volume.id)
            except exc.StorageError:
                return storage_utils.ec2_name_to_device(volume.extra["device"])
        else:
            device = storage_utils.ec2_name_to_device(volume.extra["device"])
            return storage_utils.ec2_get_persistent_device_path(device)

    def _call_create_volume(self, ctx):
        """
        Call `create_volume` via libcloud driver.
        """
        # find ec2 specific `EC2NodeLocation`
        node = self._find_node(ctx)
        location = self._find_resource(
            ctx, compute_base.NodeLocation, query={"name": node.extra["availability"]}
        )

        if ctx.volume.kms_key_id and ctx.volume.snap and ctx.volume.snap.encrypted:
            # Cannot use encryption key ID when creating volume from
            # encrypted snapshot.
            kms_key_id = None
        else:
            kms_key_id = ctx.volume.kms_key_id

        libcloud_volume = ctx.driver.create_volume(
            ctx.volume.size,
            None,
            location=location,
            snapshot=ctx.volume.snap,
            ex_volume_type=ctx.volume.volume_type,
            ex_iops=ctx.volume.iops,
            ex_throughput=ctx.volume.throughput,
            ex_encrypted=ctx.volume.encrypted,
            ex_kms_key_id=kms_key_id,
            ex_tags=ctx.volume.tags,
        )
        return libcloud_volume

    def _call_create_volume_snapshot(self, ctx, **kwargs):
        """
        Call `create_volume_snapshot` via libcloud driver.
        """
        libcloud_snap = ctx.driver.create_volume_snapshot(
            ctx.volume, name=kwargs.get("description")
        )
        if kwargs.get("tags"):
            ctx.driver.ex_create_tags(libcloud_snap, kwargs["tags"])
            libcloud_snap.extra["tags"].update(kwargs["tags"])
        return libcloud_snap

    def _call_attach_volume(self, ctx):
        """
        Call `attach_volume` via libcloud driver.
        """
        node = self._find_node(ctx)
        virtualization_type = node.extra.get("virtualization_type")
        device_name = storage_utils.ec2_get_free_device_name(
            virtualization_type=virtualization_type,
            taken_device_names=[bdm["device_name"] for bdm in node.extra["block_device_mapping"]],
        )

        ctx.driver.attach_volume(node, ctx.volume, device=device_name)
        ctx.volume.name = device_name

    def _call_detach_volume(self, ctx):
        """
        Call `detach_volume` via libcloud driver.
        """
        ctx.driver.detach_volume(ctx.volume, ex_force=True)

    def _prepare_volume_update(self, ctx, parameters):
        if parameters.size and int(parameters.size) < ctx.volume.size:
            raise exc.MalformedError(
                "New volume size ({}) must be greater than the old one ({}).".format(
                    parameters.size, ctx.volume.size
                )
            )
        if ctx.volume.volume_type == "standard":
            raise exc.MalformedError(
                'Modifications are not supported by "standard" EBS volume type.'
            )
        if parameters.volume_type == "standard":
            raise exc.MalformedError('New volume type cannot be "standard".')

    def _update_volume(self, ctx, parameters):
        parameters = {
            mixutil.camelize(param, True): value
            for param, value in parameters.items()
            if value is not None
        }
        ctx.driver.ex_modify_volume(ctx.volume, parameters)
        self._wait_volume_update(ctx)

    def _check_modification_state(self, ctx):
        volume_modifications = ctx.driver.ex_describe_volumes_modifications(
            volume_ids=[ctx.volume.id]
        )
        if not volume_modifications:
            # if there is no volume modification states, it means the volume is ready
            return True

        volume_modification = volume_modifications[0]
        state = volume_modification.modification_state
        if state == "optimizing" or state == "completed":
            return True
        if state == "failed":
            raise base.StorageError(
                "Volume modification failed: {}".format(volume_modification.status_message)
            )

    def _wait_volume_update(self, ctx):
        mixutil.try_until(
            lambda: self._check_modification_state(ctx),
            try_msg='Waiting for volume to transition from "modifying" state.',
            interval=self.update_polling_interval,
            timeout=self.update_timeout,
        )

    @staticmethod
    def mount_volume(volume, fstab=True, raise_already_mounted=False):
        if fact["os"]["family"] == "windows":
            try:
                # Updating volume.device using volume.id if it's NVMe type volumes.
                # The volumes could be attached to a new SCSI addresses on reboot.
                volume.device = storage_utils.ec2_volume_id_to_device(volume.id)
                LOG.info("Volume %s mapped to %s", volume.id, volume.device)
            except exc.StorageError:
                # `name` is the only right argument here.
                # Fallback to `device` exists for older Scalrs
                # see https://scalr-labs.atlassian.net/browse/SCALRCORE-15649
                volume.device = storage_utils.ec2_name_to_device(volume.name or volume.device)
                LOG.info(
                    "Volume %s (%s) mapped to %s",
                    volume.id,
                    volume.name or volume.device,
                    volume.device,
                )
        else:
            # Don't even think about changing the argument below
            # from `volume.device` to `volume.name or volume.device`
            # SNAFU: https://scalr-labs.atlassian.net/browse/SES-4175
            device = storage_utils.ec2_name_to_device(volume.device)
            volume.device = storage_utils.ec2_get_persistent_device_path(device)

        return super(EbsStorage, EbsStorage).mount_volume(
            volume, fstab=fstab, raise_already_mounted=raise_already_mounted
        )
