"""
Module that provides tasks to manage storage.
"""
from __future__ import absolute_import

from agent.utils import celeryutil
from common import types
from common.utils import exc
from common.utils import imputil
from common.utils import logutil

__all__ = [
    "prepare_storage",
    "create_volume",
    "create_snapshot",
    "delete_snapshot_maybe",
    "attach_volume",
    "detach_volume",
    "mount_volume",
    "update_volume",
]

LOG = logutil.get_logger(__name__)


providers = {
    "ebs": "agent.tasks.storage.ec2.EbsStorage",
    "azure": "agent.tasks.storage.azure.AzureStorage",
    "azure_vhd": "agent.tasks.storage.azure_vhd.AzureVHDStorage",
    "aws_efs": "agent.tasks.storage.aws_efs.EFSStorage",
}


# pylint: disable=invalid-name
def Storage(storage_volume_type):
    """
    Factory function to find & create storage provider for the volume.

    :param str storage_volume_type: one of (ebs, azure).
    """
    try:
        imp_name = providers[storage_volume_type]
    except KeyError:
        raise exc.NotFoundError(
            "No Storage provider class for storage volume type {!r}.".format(storage_volume_type)
        )
    return imputil.import_object(imp_name)


@celeryutil.agent_task(name="storage.prepare_storage")
def prepare_storage(volume=None, cloud_server=None, credentials=None, **kwargs):
    """
    Plans workflow further steps: whether to snapshot, create, attach or mount?

    :param volume: A volume to prepare.
    :type volume: :class:`common.types.Volume`

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.prepare_storage(ctx)


@celeryutil.agent_task(name="storage.create_volume")
def create_volume(volume=None, cloud_server=None, credentials=None, wait_available=True, **kwargs):
    """
    Create volume on cloud.

    Returns it without waiting for completion.

    :param volume: A volume to create.
    :type volume: :class:`common.types.Volume` or subclass

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :param wait_available: Waiting for volume to become `available`.
    :param wait_available: bool

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.create_volume(ctx, wait_available=wait_available)


@celeryutil.agent_task(name="storage.create_snapshot")
def create_snapshot(
    volume: types.Volume = None,
    cloud_server: types.CloudServer = None,
    credentials: types.CloudCredentials = None,
    wait_available: bool = True,
    description: str = None,
    tags: dict = None,
    **kwargs,
) -> types.StorageLocals:
    """
    Creates snapshot on a cloud.

    Returns it without waiting for completion

    :param volume: A volume to snapshot.
    :param cloud_server: A cloud server (Node/Instance).
    :param credentials: credentials that are used to create cloud driver.
    :param wait_available: Waiting for snapshot to become `available`.
    :param description: Snapshot description
    :param tags: Snapshot tags (if provider supports tagging)
    """
    provider = Storage(volume.get("type"))  # type: ignore
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.create_snapshot(
        ctx, wait_available=wait_available, description=description, tags=tags, **kwargs
    )


@celeryutil.agent_task(name="storage.delete_snapshot_maybe")
def delete_snapshot_maybe(volume=None, cloud_server=None, credentials=None, **kwargs):
    """
    Delete temporary snapshot used for volume creation.

    :param volume: A volume.
    :type volume: :class:`common.types.Volume`

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.delete_snapshot_maybe(ctx)


@celeryutil.agent_task(name="storage.attach_volume")
def attach_volume(volume=None, cloud_server=None, credentials=None, **kwargs):
    """
    Attach volume to cloud server.

    Waits for volume attachment. Grabs device name.
    Updates :attr:`volume.device` in result volume.

    :param volume: A volume to attach.
    :type volume: :class:`common.types.Volume` or subclass

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    :return: ``StorageLocals`` partial with ``create_filesystem`` and
        ``volume`` attributes.
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.attach_volume(ctx)


@celeryutil.agent_task(name="storage.detach_volume")
def detach_volume(volume=None, cloud_server=None, credentials=None, **kwargs):
    """
    Detach volume from cloud server.

    :param volume: A volume to detach.
    :type volume: :class:`common.types.Volume` or subclass

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.detach_volume(ctx)


@celeryutil.agent_task(name="storage.mount_volume")
def mount_volume(volume=None, fstab=True, raise_already_mounted=False, **kwargs):
    """
    Mount volume and ensure that OS will do re-mount after reboot.

    :param volume: A volume to mount.
    :type volume: :class:`common.types.Volume`

    :param fstab: re-mount after reboot (linux-only).
    :type fstab: bool

    :param raise_already_mounted: If True, raises an error if device is already mounted.
    :type raise_already_mounted: bool

    :rtype: :class:`common.types.StorageLocals`
    """
    return Storage(volume.get("type")).mount_volume(
        volume=volume, fstab=fstab, raise_already_mounted=raise_already_mounted
    )


@celeryutil.agent_task(name="storage.prepare_volume_update")
def prepare_volume_update(
    volume=None, update_parameters=None, cloud_server=None, credentials=None, **kwargs
):
    """
    Prepares volume for update. Checks that update is possible.

    :param volume: A volume to update.
    :type volume: :class:`common.types.Volume` or subclass

    :param update_parameters: A update_parameters config with volume parameters to update.
    :type update_parameters: :class:`common.types.VolumeModification` or subclass

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.prepare_volume_update(ctx, update_parameters)


@celeryutil.agent_task(name="storage.update_volume")
def update_volume(
    volume=None, update_parameters=None, cloud_server=None, credentials=None, **kwargs
):
    """
    Updates volume with new config on clouds that support volume modification.

    :param volume: A volume to update.
    :type volume: :class:`common.types.Volume` or subclass

    :param update_parameters: A update_parameters config with volume parameters to update.
    :type update_parameters: :class:`common.types.VolumeModification` or subclass

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.update_volume(ctx, update_parameters)


@celeryutil.agent_task(name="storage.wait_volume_update")
def wait_volume_update(volume=None, cloud_server=None, credentials=None, **kwargs):
    """
    Monitors volume after modification and exits as soon as it's ready.

    :param volume: A volume to update.
    :type volume: :class:`common.types.Volume` or subclass

    :param cloud_server: A cloud server (Node/Instance).
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: credentials that are used to create cloud driver.
    :type credentials: :class:`common.types.Credentials`

    :rtype: :class:`common.types.StorageLocals`
    """
    provider = Storage(volume.get("type"))
    ctx = provider.init_context(volume, cloud_server, credentials)
    return provider.wait_volume_update(ctx)
