"""
Storage utils for Linux.
"""
from __future__ import absolute_import

import collections
import ctypes
import glob
import itertools
import os
import string  # pylint: disable=deprecated-module
import sys

from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils import subprocess2
from common.utils.facts import fact

if fact["os"]["family"] != "windows":
    import fcntl
    from common.utils import linuxutil
else:
    raise ImportError("Not available on Windows")


LOG = logutil.get_logger(__name__)

__all__ = [
    # common
    "attaching_device",
    "mount_device",
    "umount_device",
    "get_uuid_device_path",
    # ec2
    "ec2_name_to_device",
    "ec2_get_free_device_name",
    "ec2_get_persistent_device_path",
    "ec2_volume_id_to_device",
    "ec2_get_nvme_info",
    # azure
    "azure_get_device",
    "azure_get_persistent_device_path",
]


Ec2NvmeInfo = collections.namedtuple("Ec2NvmeInfo", ("device", "volume_id", "manufacturer"))


class UdevAdmCmd(subprocess2.Command):
    """
    `udevadm` command wrapper.

    Note: The udevadm binary can be located in the `/sbin` or `/bin` directories.
    See: https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=852580
    """

    binary = subprocess2.which("udevadm")


#: Retry interval for checking `/dev/disk/by-uuid/*` link.
UDEV_LINK_POLLING_INTERVAL = 1

#: Time assumed for `/dev/disk/by-uuid/*` link to appear after disk formatting.
UDEV_LINK_POLLING_TIMEOUT = 10


class attaching_device(object):  # pylint: disable=invalid-name
    """
    A context manager to handle & get attaching device information.
    """

    _timeout = 300
    _polling_interval = 10

    def __init__(self, timeout=None, polling_interval=None):
        if timeout is not None:
            self._timeout = timeout
        if polling_interval is not None:
            self._polling_interval = polling_interval

        self.taken_device_names = None
        self.device_name = None

    def __enter__(self):
        self.taken_device_names = _list_taken_device_names()
        LOG.debug("Devices before attach %s", self.taken_device_names)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            six.reraise(exc_type, exc_value, exc_traceback)

        mixutil.try_until(
            self._check_if_new_disks_appeared,
            try_msg="Waiting for new block devices in the operating system.",
            interval=self._polling_interval,
            timeout=self._timeout,
        )

        new_devices = [
            device for device in _list_taken_device_names() if device not in self.taken_device_names
        ]
        if not new_devices:
            raise exc.StorageError("No new devices appeared after the attach.")

        # [sda1, sda, sda2, sdc, sdc2, sdm2] -> [sda, sdc, sdm2]
        if len(new_devices) > 1:
            exclude_pattern = lambda x: any(i for i in new_devices if i != x and x.startswith(i))
            new_devices = [d for d in new_devices if not exclude_pattern(d)]

        if len(new_devices) == 1:
            self.device_name = new_devices[0]
            LOG.debug(
                "Found new device '%s' among Linux devices. " "Attachment is OK.", self.device_name
            )
        else:
            raise exc.StorageError(
                (
                    "While polling for attached device, got multiple new "
                    "devices: {}. Unable to determine which one to select."
                ).format(new_devices)
            )

    def _check_if_new_disks_appeared(self):
        LOG.debug("Rescanning SCSI hosts.")
        scsi_host = "/sys/class/scsi_host"
        if os.path.exists(scsi_host):
            for name in os.listdir(scsi_host):
                scsi_scan_file = os.path.join(scsi_host, name, "scan")
                LOG.debug("Poking %s", scsi_scan_file)
                with open(scsi_scan_file, "w") as fp:
                    # The three values stand for channel, SCSI target ID, and LUN.
                    # The dashes act as wildcards meaning "rescan everything"
                    fp.write("- - -")
        LOG.debug("SCSI hosts scanned. Checking updated device list.")
        return len(self.taken_device_names) < len(_list_taken_device_names())


@mixutil.memorize()
def _ec2_rhel_device_ordering_bug():
    """
    Check if system is affected by devices ordering bug.

    see: https://bugzilla.redhat.com/show_bug.cgi?id=729340
    """
    if fact["os"]["family"] == "redhat":
        entries = linuxutil.mounts().entries
        root_entries = (entry for entry in entries if entry.mpoint == "/")
        return any(entry.device.startswith("/dev/xvde") for entry in root_entries)


def _list_taken_device_names():
    devices = set(glob.glob("/dev/xvd*") + glob.glob("/dev/[hs]d*") + glob.glob("/dev/nvme*n*"))
    return list(set(six.moves.map(os.path.realpath, devices)))


def _device_letters(start, end):
    sequence = string.ascii_letters
    for s in sequence[sequence.index(start) : sequence.index(end) + 1]:
        yield s


def ec2_get_nvme_info(device_path):
    """
    Returns an EC2 specific information about the nvme device.

    :param device_path: The path to the nvme device, example: `/dev/nvme2n1`.
    :type device_path: str

    :return: :class:`Ec2NvmeInfo`
    """
    if not device_path.startswith("/dev/nvme"):
        raise ValueError(
            "Invalid device path, expected '/dev/nvme*', " "got '{}'.".format(device_path)
        )
    buf = ctypes.create_string_buffer(0x1000)
    cmd = linuxutil.nvme_admin_cmd(
        opcode=0x06, nsid=0, addr=ctypes.addressof(buf), data_len=len(buf), cdw10=1
    )
    with open(device_path) as fp:
        fcntl.ioctl(fp, linuxutil.NVME_IOCTL_ADMIN_CMD, cmd)

    # see struct nvme_id_ctrl https://github.com/linux-nvme/nvme-cli/blob/master/linux/nvme_ioctl.h
    # char sn[20], example: `vol0cc482cb1e4670c52`
    volume_id = buf.raw[4:24].decode().strip()
    if volume_id.startswith("vol"):
        volume_id = "vol-{}".format(volume_id[3:])

    # char mn[40], example: `Amazon Elastic Block Store`
    manufacturer = buf.raw[24:64].decode().strip()

    # char device[32] (aws extension), example: `sda1` or `/dev/xvda`
    device = buf.raw[3072:3104].decode().strip()
    if not device.startswith("/dev/"):
        device = "/dev/{}".format(device)

    return Ec2NvmeInfo(device, volume_id, manufacturer)


def ec2_name_to_device(name):
    """
    Converts ``sdX`` or ``/dev/sdX`` to ``/dev/xvdX``.
    """
    if not name.startswith("/dev"):
        name = os.path.join("/dev", name)
    if name.startswith("/dev/xvd"):
        return name
    if _ec2_rhel_device_ordering_bug() or os.path.exists("/dev/xvda1"):
        name = name.replace("/sd", "/xvd")
    if _ec2_rhel_device_ordering_bug():
        name = name[0:8] + chr(ord(name[8]) + 4) + name[9:]
    return name


def ec2_get_persistent_device_path(name: str) -> str:
    """
    Return persistent device path.

    Reboot changes how EBS volumes are mapped to `/dev/nvme*` names.
    Which means: volume may obtain a device name, previously assigned to different EBS.
    There is no AWS documented solution for this problem (01/2018), but we've found
    that symlink `/dev/disk/by-id/nvme-Amazon_Elastic_Block_Store_<volume_id>` is persistent.

    NVME EBS and Instance Store volumes use IDs:
      - EBS: `/dev/disk/by-id/nvme-Amazon_Elastic_Block_Store_volXXX`
      - Instance Store: `/dev/disk/by-id/nvme-Amazon_EC2_NVMe_Instance_Storage_AWSXXX`
    Other volumes use regular `/dev/disk/by-uuid/*` links.

    :param name: The path to the volume (e.g. /dev/sda, /dev/xvda1, /dev/nvme1n1)
    """
    if name.startswith("/dev/disk/by-id") or name.startswith("/dev/disk/by-uuid"):
        return name

    def _check_for_udev_symlink(symlink):
        if not os.path.exists(symlink):
            # updates /dev/disk/* symlinks
            subprocess2.check_output(["udevadm", "trigger"])
            return False
        return True

    nvme_aliases = {
        ec2_get_nvme_info(device_name).device: device_name
        for device_name in _list_taken_device_names()
        if device_name.startswith("/dev/nvme")
    }
    if name.startswith("/dev/nvme") or name in nvme_aliases:
        nvme_info = ec2_get_nvme_info(nvme_aliases.get(name) or name)

        if nvme_info.manufacturer and "instance storage" in nvme_info.manufacturer.lower():
            volume_id = nvme_info.volume_id
            name_by_id = "/dev/disk/by-id/nvme-Amazon_EC2_NVMe_Instance_Storage_{}".format(
                volume_id
            )
        else:
            volume_id = nvme_info.volume_id.replace("vol-", "vol", 1)
            name_by_id = "/dev/disk/by-id/nvme-Amazon_Elastic_Block_Store_{}".format(volume_id)

        mixutil.try_until(
            lambda: _check_for_udev_symlink(name_by_id),
            error_msg="The device symlink {} was not found (device={}).".format(name_by_id, name),
            interval=UDEV_LINK_POLLING_INTERVAL,
            timeout=UDEV_LINK_POLLING_TIMEOUT,
        )
        return name_by_id
    return get_uuid_device_path(name)


def ec2_get_free_device_name(virtualization_type=None, taken_device_names=None):
    """
    Return available device name for Linux instance on EC2.

    :param string virtualization_type: Virtualization type
    :param list taken_device_names: winutil compatibility param.

    See: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/device_naming.html
    """
    start = "f"
    if fact["os"]["name"] == "ubuntu" and fact["os"]["release"] >= (14, 4):
        # Ubuntu 14.04 returns 'Attachment point /dev/sdf is already in used'
        start = "g"
    if fact["os"]["family"] == "redhat":
        # RHEL 6 returns "Null body" for /dev/sdf and /dev/sdg
        start = "h"

    if virtualization_type == "hvm":
        # /dev/xvd[f-p]
        # /dev/xvd[b-c][a-z]
        available_device_names = itertools.chain(
            ("/dev/xvd{}".format(s) for s in _device_letters(start, "p")),
            ("/dev/xvd{}{}".format(a, b) for a in "bc" for b in string.ascii_lowercase),
        )
    else:
        available_device_names = ("/dev/sd{}".format(s) for s in _device_letters(start, "p"))

    taken_device_names = []
    for device_name in _list_taken_device_names():
        # converts nvme device names to common naming
        if device_name.startswith("/dev/nvme"):
            device_name = ec2_get_nvme_info(device_name).device
            if virtualization_type == "hvm":
                device_name = device_name.replace("/sd", "/xvd")
        taken_device_names.append(device_name)

    is_partition_of = lambda dev, part: part[-1].isdigit() and part[:-1] == dev

    # skip /dev/sdXN if /dev/sdX device already in use and vise versa
    for device_name in available_device_names:
        device = ec2_name_to_device(device_name)
        if any(
            device == taken_device
            or is_partition_of(device, taken_device)
            or is_partition_of(taken_device, device)
            for taken_device in taken_device_names
        ):
            continue
        return device_name

    raise exc.StorageError("All device names already used.")


def ec2_volume_id_to_device(*args, **kwargs):
    """ Winutil only function """
    raise NotImplementedError("ec2_volume_id_to_device() is Windows-only function.")


def azure_get_persistent_device_path(device: str) -> str:
    """
    Returns by-uuid device link from given `/dev/sd*` device path.

    :param device: device path
    :return: by-uuid path
    """
    return get_uuid_device_path(device)


def azure_get_device(scsi_lun):
    """
    Searches for device path by logical unit number.

    See: https://docs.microsoft.com/en-us/azure/storage/storage-about-disks-and-vhds-linux

    :param scsi_lun: Logical unit number of attached disk drive.
    :type scsi_lun: int

    :rtype: str
    """
    device_path = None
    for value in glob.glob("/sys/block/sd*/device"):
        device_name = value.split("/")[-2]

        # scsi address - `host:channel:target:lun`
        hctl = os.path.basename(os.path.realpath(value))
        device_lun = int(hctl.split(":")[-1])

        if device_lun == scsi_lun:
            if device_path is not None:
                raise exc.StorageError(
                    (
                        "Found multiple disks with same logical unit number ({}). "
                        "Unable to determine which one to select."
                    ).format(device_lun)
                )
            device_path = os.path.join("/dev", device_name)

    if device_path is None:
        raise exc.StorageError(
            ("Could not find attached disk with logical unit number {}.").format(scsi_lun)
        )
    return azure_get_persistent_device_path(device_path)


def mount_device(
    device, mpoint, fstype, mount_options=None, fstab=False, raise_already_mounted=False, **kwargs
):
    """
    Mounts device to the mount point path.

    :param device: Path to the device (e.g: ``/dev/sda``).
    :type device: str

    :param mpoint: Path to the mount point.
    :type mpoint: str

    :param fstype: Type of the file system (example: ``xfs``, ``ext4``)
    :type fstype: str

    :param mount_options: List of mount options.
    :type mount_options: list[str] or None

    :param fstab: Re-mount device after reboot.
    :type fstab: bool

    :param raise_already_mounted: If True, raises an error if device is already mounted.
    :type raise_already_mounted: bool
    """
    fstab_mgr = linuxutil.fstab()

    def ensure_fstab():
        """
        Add volume mount point to `/etc/fstab`.
        """
        if not fstab:
            return
        try:
            fstab_mgr.put(device, mpoint, fstype=fstype, options=mount_options)
        except (exc.Error, IOError) as e:
            err_msg = (
                "Error adding fstab record (device={}, mpoint={}, fstype={}, "
                "options={}).\n"
                "Reason: {}"
            ).format(device, mpoint, fstype, mount_options, e)
            six.reraise(exc.StorageError, exc.StorageError(err_msg), sys.exc_info()[2])

    # prepare mount options
    mount_options = mount_options or ["defaults"]
    fact_os = fact["os" if not fact["manage_container_host"] else "container_host.os"]
    if fact_os["name"] == "ubuntu" and fact_os["release"] < (16, 4):
        mount_options.append("nobootwait")
    elif fact_os["family"] != "coreos":
        # coreos can't boot with a 'nofail' option
        mount_options.append("nofail")
    mount_options.append("comment=scalr")

    # mount device to mpoint
    try:
        linuxutil.mount(
            device,
            mpoint,
            options=",".join(set(mount_options) - {"nofail"}),
            type=fstype,
            raise_already_mounted=raise_already_mounted,
        )
    except subprocess2.CalledProcessError as e:
        msg = (
            "Error mounting volume (device={}, mpoint={}, fstype={}, options={}).\n" "Reason: {}"
        ).format(device, mpoint, fstype, mount_options, e)
        six.reraise(exc.StorageError, exc.StorageError(msg), sys.exc_info()[2])
    except exc.AlreadyMountedError:
        ensure_fstab()
        raise

    ensure_fstab()


def umount_device(device, fstab=False):
    """
    Umounts device and removes it's record from fstab.

    :param str device: Device path. (e.g: ``/dev/sda``).

    :param bool fstab: If True edit /etc/fstab record.
    """
    linuxutil.umount(device)
    if fstab:
        linuxutil.fstab().remove_device(device)


def get_uuid_device_path(device: str) -> str:
    """
    Returns by-uuid device link from given `/dev/sd*` device path.

    See:
     - https://wiki.archlinux.org/index.php/persistent_block_device_naming#by-uuid

    :param device: device path
    :return: by-uuid path (`/dev/disk/by-uuid/{uuid4}`).
    """

    def get_persistent_link():
        UdevAdmCmd().args(
            "trigger", action="change", sysname_match=os.path.basename(device)
        ).check_output()

        for uuid_ in glob.glob("/dev/disk/by-uuid/*"):
            if uuid_ == device:
                # if it's already by-uuid link
                return device
            if os.path.realpath(uuid_) == device:
                return uuid_

    uuid_device: str = mixutil.try_until(
        get_persistent_link,
        try_msg="getting /dev/disk/by-uuid/* device link for {}".format(device),
        log=LOG,
        interval=UDEV_LINK_POLLING_INTERVAL,
        timeout=UDEV_LINK_POLLING_TIMEOUT,
    )
    return uuid_device
