"""
Storage utils for Windows.
"""
from __future__ import absolute_import

import re
import string  # pylint: disable=deprecated-module
import threading
import typing as tp

from agent.tasks import filesystem
from common import types
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import six
from common.utils.facts import fact

if fact["os"]["family"] == "windows":
    # pylint: disable=ungrouped-imports
    import wmi
    import pythoncom
    from common.utils import winutil
else:
    raise ImportError("Not available on Linux")


LOG = logutil.get_logger(__name__)

__all__ = [
    # common
    "attaching_device",
    "mount_device",
    "get_uuid_device_path",
    # ec2
    "ec2_name_to_device",
    "ec2_get_free_device_name",
    "ec2_volume_id_to_device",
    # azure
    "azure_get_device",
    "azure_get_persistent_device_path",
]


class attaching_device(object):  # pylint: disable=invalid-name
    """
    A context manager to handle & get attaching device information.
    """

    _timeout = 300
    _polling_interval = 10
    _bring_online_timeout = 60
    _bring_online_polling_interval = 1

    def __init__(self, timeout=None, polling_interval=None):
        if timeout is not None:
            self._timeout = timeout
        if polling_interval is not None:
            self._polling_interval = polling_interval

        self._wmi_api = None
        self.taken_scsi_addrs = None
        self.device_name = None

    def __enter__(self):
        self._coinitialize()
        self._wmi_api = wmi.WMI()

        self.taken_scsi_addrs = self._get_taken_scsi_addrs()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type:
            self._couninitialize()
            six.reraise(exc_type, exc_value, exc_traceback)

        try:
            mixutil.try_until(
                self._check_if_new_disks_appeared,
                try_msg="Waiting for new Win32_DiskDrive to be available in system.",
                timeout=self._timeout,
                interval=self._polling_interval,
                error_msg="New Win32_DiskDrive is not available.",
            )

            # find difference and assert it as a disk attached
            new_scsi_addrs = self._get_taken_scsi_addrs()
            if len(new_scsi_addrs) <= len(self.taken_scsi_addrs):
                raise exc.StorageError("No new Win32_DiskDrive appeared after the attach.")

            attached_scsi_addr = (new_scsi_addrs - self.taken_scsi_addrs).pop()
            LOG.debug(
                "New Win32_DiskDrive (scsi_address=%s) appeared. Assuming it is a "
                "newly attached disk.",
                attached_scsi_addr,
            )
            self.device_name = self._bring_online(attached_scsi_addr)
        finally:
            self._couninitialize()

    @staticmethod
    def _coinitialize():
        if threading.current_thread().name != "MainThread":
            pythoncom.CoInitialize()

    @staticmethod
    def _couninitialize():
        if threading.current_thread().name != "MainThread":
            pythoncom.CoUninitialize()

    def _get_taken_scsi_addrs(self):
        """
        Returns taken SCSI addresses.

        :rtype: set of :class:`types.SCSIAddress`
        """
        taken_scsi_addrs = set(
            types.SCSIAddress(disk_drive=d) for d in self._wmi_api.Win32_DiskDrive()
        )
        LOG.debug(
            "Polling Win32_DiskDrive WMI objects to get taken SCSI addresses " "(addresses=%s).",
            list(taken_scsi_addrs),
        )
        return taken_scsi_addrs

    def _check_if_new_disks_appeared(self):
        return len(self.taken_scsi_addrs) < len(self._get_taken_scsi_addrs())

    @staticmethod
    def _check_if_disk_online(scsi_address):
        try:
            winutil.bring_disk_online(scsi_address)
            return True
        except exc.Error as e:
            LOG.debug(
                "Error when bringing Win32_DiskDrive (scsi_address=%s) " "online: %s.",
                scsi_address,
                e,
            )
            return False

    def _bring_online(self, scsi_address):
        """
        Performs actions to bring attached disk online.

        Should be ran only in ``coinitialized_context``.
        """
        # perform all the actions that are needed to bring disk online
        LOG.debug("Bringing Win32_DiskDrive (scsi_address=%s) online.", scsi_address)
        mixutil.try_until(
            lambda: self._check_if_disk_online(scsi_address),
            timeout=self._bring_online_timeout,
            interval=self._bring_online_polling_interval,
            error_msg="Cannot bring online Win32_DiskDrive (scsi_address={})".format(scsi_address),
        )

        # onlining can restore unwanted volume letters and mount points
        winutil.umount(scsi_address)

        return scsi_address


def _ec2_device_name_to_scsi_target_id(device_name):  # pylint: disable=invalid-name
    """
    Returns SCSITargetId for corresponding EC2 device name (e.g. ``xvda``, ``/dev/xvdt``).

    Formula from here:
    https://aws.amazon.com/blogs/developer/stripe-windows-ephemeral-disks-at-launch/
    """
    device_name = device_name.split("/")[-1]
    device_suffix = device_name[3:]

    if len(device_suffix) == 1:
        return ord(device_suffix) - 97
    elif len(device_suffix) == 2:
        return (ord(device_suffix[0]) - 96) * 26 + ord(device_suffix[1]) - 97
    else:
        raise exc.StorageError("Wrong EC2 disk device name format: {}".format(device_name))


def ec2_device_name_to_scsi_address(
    device_name_or_target_id: tp.Union[str, int]
) -> types.SCSIAddress:
    """
    Converts EC2 device name to SCSI address.

    :param device_name_or_target_id: EC2 device name (e.g. ``/dev/xvda``) or SCSI target ID.
    """
    try:
        target_id = int(device_name_or_target_id)
    except ValueError:
        target_id = _ec2_device_name_to_scsi_target_id(device_name_or_target_id)

    try:
        with winutil.coinitialized_context():
            wmi_api = wmi.WMI()
            for disk_drive in wmi_api.Win32_DiskDrive():
                if disk_drive.SCSITargetId == target_id:
                    return types.SCSIAddress(disk_drive=disk_drive)
    except Exception as err:
        LOG.debug("Unable to find SCSI device for target ID %s. " "Reason: %s", target_id, str(err))

    return types.SCSIAddress({"target_id": target_id})


def ec2_scsi_target_id_to_device_name(scsi_target_id):  # pylint: disable=invalid-name
    """
    Returns EC2 device name (e.g. ``xvda``) by the SCSI target ID.

    :param scsi_target_id: The SCSI address target ID.
    :type scsi_target_id: int

    :rtype: str
    """
    return (
        "xvd{}".format(chr(scsi_target_id + 97))
        if scsi_target_id < 26
        else "xvd{}{}".format(chr(scsi_target_id // 26 + 96), chr(scsi_target_id % 26 + 97))
    )


def ec2_name_to_device(name):
    """
    Return SCSI address of device by EC2 device name (e.g. ``sdX`` or ``xvdX``).

    :return: :class:`types.SCSIAddress`
    """
    if isinstance(name, types.SCSIAddress):
        return name

    if name.startswith("sd") or name.startswith("xvd"):
        return ec2_device_name_to_scsi_address(re.sub(r"^sd", "xvd", name))

    try:
        return types.SCSIAddress.from_packed_address(name)
    except exc.Error:
        pass

    raise exc.StorageError("Unknown device name format '{}'.".format(name))


def _ec2_get_taken_device_names():
    """
    Seraches block device mapping metadata and returns list of used device names.

    See: https://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/block-device-mapping-concepts.html#bdm-instance-metadata
    and https://scalr-labs.atlassian.net/browse/FAM-1204
    and https://scalr-labs.atlassian.net/browse/FAM-1428

    :return list(str): List of taken device names
    """
    device_names = set()
    for k, v in fact["ec2"].items():
        if k.startswith("block-device-mapping"):
            device_names.add(v.split("/")[-1])
    return list(device_names)


def ec2_get_free_device_name(virtualization_type=None, taken_device_names=None):
    """
    Return list of available device names for Windows instance on EC2.

    :param string virtualization_type: Not used,
        it's needed only in ``linuxutil.ec2_get_free_device_name()``
    :param list taken_device_names: List of device names not to choose from.

    See: http://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/device_naming.html
    """
    taken_device_names = taken_device_names or []

    # xvd[f-p]
    # xvd[b-c][a-z]
    available_device_names = ["xvd{}".format(s) for s in string.ascii_lowercase[5:-10]] + [
        "xvd{}{}".format(a, b) for a in "bc" for b in string.ascii_lowercase
    ]

    taken_device_names.extend(_ec2_get_taken_device_names())
    LOG.debug("Device names taken by block device mapping (%s)", taken_device_names)
    with winutil.coinitialized_context():
        wmi_api = wmi.WMI()
        taken_device_names.extend(
            ec2_scsi_target_id_to_device_name(d.SCSITargetId) for d in wmi_api.Win32_DiskDrive()
        )

    free_device_names = [name for name in available_device_names if name not in taken_device_names]
    if len(free_device_names) > 0:
        return free_device_names[0]
    else:
        raise exc.StorageError("All device names already used.")


def azure_get_persistent_device_path(device):
    """
    Linuxutil only function.
    """
    raise NotImplementedError("azure_get_persistent_device_path() is Linux-only function")


def azure_get_device(scsi_lun):
    """
    Searches for SCSI address by logical unit number.

    See: https://docs.microsoft.com/en-us/azure/storage/storage-about-disks-and-vhds-windows

    :param scsi_lun: Logical unit number of attached disk drive.
    :type scsi_lun: int

    :rtype: :class:`types.SCSIAddress`
    """
    with winutil.coinitialized_context():
        wmi_api = wmi.WMI()
        disk_drives = list(set(types.SCSIAddress(disk_drive=d) for d in wmi_api.Win32_DiskDrive()))

    # exclude disks with SCSIPort=0, that port for system and temporary disk
    disk_drives = [
        disk_drive
        for disk_drive in disk_drives
        if disk_drive.port != 0 and disk_drive.logical_unit == scsi_lun
    ]

    if len(disk_drives) == 1:
        return disk_drives[0]
    elif len(disk_drives) > 1:
        raise exc.StorageError(
            (
                "Found multiple disks with same logical unit number ({}). "
                "Unable to determine which one to select."
            ).format(scsi_lun)
        )
    else:
        raise exc.StorageError(
            ("Could not find attached disk with logical unit number {}.").format(scsi_lun)
        )


def mount_device(scsi_address, mpoint, fstype, label=None, raise_already_mounted=False, **kwargs):
    """
    Mounts device to the mount point path.

    :param scsi_address: The SCSI address of the ``Win32_DiskDrive`` (example: ``3:0:0:1``).
    :type scsi_address: str

    :param mpoint: The mount point (drive letter or directory) for the ``Win32_Volume``.
    :type mpoint: str

    :param fstype: The type of the file system (example: ``xfs``, ``ext4``).
    :type fstype: str

    :param label: The label for the ``Win32_Volume``.
    :type label: str or None

    :param raise_already_mounted: If True, raises an error if device is already mounted.
    :type raise_already_mounted: bool
    """
    if isinstance(scsi_address, six.string_types):
        scsi_address = types.SCSIAddress.from_packed_address(scsi_address)

    winutil.mount(scsi_address, mpoint, raise_already_mounted=raise_already_mounted)
    if not winutil.logical_disk_exists(scsi_address) and winutil.disk_drive_exists(scsi_address):
        # If logical disk (Win32_Volume) is still missing after mount attempt
        # and actual volume device (Win32_DiskDrive) is present,
        # it means that device went offline for some reason.
        # In that case we try to bring it online again.
        # This may happen when changing instance type on EC2.
        winutil.bring_disk_online(scsi_address)
        winutil.mount(scsi_address, mpoint, raise_already_mounted=raise_already_mounted)

    # assign label to the Win32_Volume on the Win32_DiskDrive
    if label is not None:
        fs = filesystem.FileSystem(fstype)
        LOG.debug("Setting label '%s' for Win32_Volume (scsi_address=%s).", label, scsi_address)
        fs.set_label(scsi_address, label)


def ec2_volume_id_to_device(volume_id):
    """
    Finds SCSI address of a volume by its ID.

    :param str volume_id: ID of a volume.
    :return types.SCSIAddress: SCSI address of a Windows disk.
    """
    # Trying to determine scsi address by serial number (volume id)/
    # https://docs.aws.amazon.com/AWSEC2/latest/WindowsGuide/nvme-ebs-volumes.html#identify-nvme-ebs-device
    # Volume id has format "vol-<uid>" and in WMI it has format "vol<uid>.*"
    return winutil.serial_number_to_scsi_address(volume_id.split("-")[-1])


def get_uuid_device_path(device) -> str:
    raise NotImplementedError("get_uuid_device_path() is Linux-only function")
