"""System tasks.
"""
from __future__ import absolute_import

import codecs
import copy
import json
import logging
import os
import platform
import posixpath
import re
import socket
import sys
import time
from multiprocessing import pool

import psutil
import requests
import six

import common.types
from agent.scalrinit import datadir
from agent.tasks import service as service_tasks
from agent.utils import celeryutil
from common.utils import dockerutil
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

if fact["os"]["family"] == "windows":
    import wmi
    from common.utils import winutil
    import ctypes
    import ctypes.wintypes

    augeasutil = None
else:
    from common.utils import augeasutil

    wmi = None
    winutil = None
    ctypes = None

LOG = logutil.get_logger(__name__)
NSSWITCH_CONF_PATH = "/etc/nsswitch.conf"
SYSCONFIG_NETWORK_PATH = "/etc/sysconfig/network"
ETC_HOSTS = "/etc/hosts"
ETC_HOSTS_LENS_DIR = os.path.dirname(__file__)
NTP_CONF_PATH = "/etc/ntp.conf"
HOSTNAME_MAX_LEN_WIN = 63
PRIVATE_IP_PREFIXES = ("10.", "172.", "192.168.")
AWS_METADATA_HOST = "169.254.169.254"
GCE_SET_HOSTNAME_HOOK_DEBIAN = "/etc/dhcp/dhclient-exit-hooks.d/set-hostname"
GCE_SET_HOSTNAME_HOOK_CENTOS = "/etc/dhcp/dhclient-exit-hooks"

# Possible dhcp config paths and GCE set_hostname commands
GCE_DHCP_CONFIG_PATHS = ("/etc/dhcp/dhclient.d/gce.sh", "/etc/dhcp/dhclient.d/google_hostname.sh")
GCE_SET_HOSTNAME_CMDS = ("/usr/share/google/set-hostname", "set_hostname")


class WIN_API_COMPUTER_NAME_FORMAT(common.types.Enum):
    """
    Specifies a type of computer name.
    https://docs.microsoft.com/en-us/windows/desktop/api/sysinfoapi/ne-sysinfoapi-_computer_name_format
    """

    ComputerNameNetBIOS = 0
    ComputerNameDnsHostname = 1
    ComputerNameDnsDomain = 2
    ComputerNameDnsFullyQualified = 3
    ComputerNamePhysicalNetBIOS = 4
    ComputerNamePhysicalDnsHostname = 5
    ComputerNamePhysicalDnsDomain = 6
    ComputerNamePhysicalDnsFullyQualified = 7


@celeryutil.agent_task(name="sys.sync_time")
def sync_time(ntp_servers=None, **kwargs):
    """Sync system time."""
    LOG.info("Synchronizing server time")
    if not ntp_servers:
        ntp_servers = [str(i) + ".pool.ntp.org" for i in range(0, 3)]
    LOG.debug("Using NTP pool: %s", ntp_servers)
    if fact["os"]["family"] == "windows":
        service_tasks.Service("w32time").start(already_running_ok=True)
        subprocess2.check_output(
            "w32tm "
            "/config "
            "/manualpeerlist:{} ".format(",".join(ntp_servers)) + "/syncfromflags:manual "
            "/update",
            shell=True,
        )
        subprocess2.check_output("w32tm /resync", shell=True)
    else:
        if subprocess2.which("ntpdate"):
            LOG.debug("Use ntpdate for sync")
            subprocess2.check_output(["ntpdate", "-u"] + ntp_servers)
        elif subprocess2.which("ntpd"):
            LOG.debug("Use ntpd for sync")
            # Ok ntpdate is not installed so we could try ntpd
            try:
                # Step 1. Stop NTP daemon. ntpd can't sync time with started NTP
                if service_tasks.Service("ntp").running:
                    service_tasks.Service("ntp").stop()
                # Step 2. Sync itself
                subprocess2.check_output(["ntpd", "-gq"])
            finally:
                # Step 3. Start NTP
                service_tasks.Service("ntp").start()

        elif fact["os"]["init"] == "systemd":
            # if this is OS that uses systemd init system that we should use timedatectl
            # Note: in fact if in OS with systemd ntpdate / ntp are installed timedatectl
            # steps back to let you keep your old setup. That's why this should be in last if.
            # more info https://help.ubuntu.com/lts/serverguide/NTP.html
            LOG.debug("Use timedatectl for sync")
            output = subprocess2.check_output(["timedatectl", "status"])
            match = re.search(r"(:?Network time on:|NTP enabled:) \b(?P<ntp_enabled>.+?)\b", output)
            if match and match.group("ntp_enabled").lower() != "yes":
                subprocess2.check_output(["timedatectl", "set-ntp", "true"])
        else:
            raise exc.NotSupportedError(
                "Unable to synchronize time because ntpdate/ntpd binary is not found in $PATH or "
                "timedatectl is not supported."
            )


@celeryutil.agent_task(name="sys.get_facts")
def get_facts(keys=None, invalidate_bindings=None, raise_on_fact_error=True, **kwargs):
    ret = {}
    error_args = []
    invalidate_bindings = [] if invalidate_bindings is None else invalidate_bindings
    # TODO: enforce this via Schematics
    # pylint: disable=no-member
    if not isinstance(keys, six.moves.Iterable):
        error_args.append("keys")
    if not isinstance(invalidate_bindings, six.moves.Iterable):
        error_args.append("invalidate_bindings")
    if error_args:
        raise exc.MalformedError(
            "Argument(s) {!r} should be iterable".format(", ".join(error_args))
        )

    for binding in invalidate_bindings:
        try:
            fact.invalidate(binding)
        except KeyError:
            msg = "Invalid binding name to invalidate: {!r}"
            six.reraise(
                exc.MalformedError, exc.MalformedError(msg.format(binding)), sys.exc_info()[2]
            )

    for key in keys:
        try:
            value = fact[key]
            if isinstance(value, dict):
                val = {}
                for subkey in value.keys():
                    try:
                        val[subkey] = copy.deepcopy(value[subkey])
                    except Exception as e:
                        kw = "['%s']['%s']" % (key, subkey)
                        if raise_on_fact_error:
                            msg = "Error occurred while querying fact {}: {}"
                            six.reraise(exc.Error, exc.Error(msg.format(kw, e)), sys.exc_info()[2])
                        else:
                            LOG.warning("Error occurred while querying fact %s: %s", kw, e)
                value = val
        except KeyError:
            msg = "Invalid fact name: '{}'"
            six.reraise(exc.MalformedError, exc.MalformedError(msg.format(key)), sys.exc_info()[2])
        except Exception as e:
            if raise_on_fact_error:
                msg = "Error occurred while querying fact '{}': {}"
                six.reraise(exc.Error, exc.Error(msg.format(key, e)), sys.exc_info()[2])
            else:
                LOG.warning("Error occurred while querying fact %s: %s", key, e)
        else:
            ret[key] = value
    return ret


@celeryutil.agent_task(name="sys.get_fact")
def get_fact(key, **kwargs):
    return get_facts([key]).values()[0]


def _edit_hosts_file(hostname: str, canonical: bool = False):
    """
    Edits /etc/hosts and sets canonical or alias of the record with private IP.
    Creates the record if it's missing.

    :param hostname: Hostname or FQDN to be added to record.
    :param canonical: If true, hostname will be added as canonical (first in list). If false,
        as alias (last in list). Naturally, if the record is missing hostname can only be canonical.

    Example:
        canonical is True
          /etc/hosts before editing:
            <private_ip> fqdn1 alias1
            ...

          /etc/hosts after editing:
            <private_ip> <hostname> fqdn1 alias1
            ...

        canonical is False
          /etc/hosts before editing:
            <private_ip> fqdn1 alias1
            ...

          /etc/hosts after editing:
            <private_ip> fqdn1 alias1 <hostname>
            ...
    """
    aug = augeasutil.augeas(
        context="/files" + ETC_HOSTS,
        loadpath=ETC_HOSTS_LENS_DIR,
        lens="Hosts.lns",
        incl=[ETC_HOSTS],
    )
    record_xpath = '*/ipaddr[.="{}"]'.format(fact["private_ip"])
    get_record_cmd = "get {}".format(record_xpath)

    has_private_ip = augeasutil.execute([get_record_cmd], aug=aug)[0]
    canonical_name = augeasutil.execute([get_record_cmd + "/../canonical"], aug=aug)[0]
    has_alias = augeasutil.execute(
        [get_record_cmd + '/../alias[.="{}"]'.format(hostname)], aug=aug
    )[0]

    if not has_private_ip:
        # if there is no record, creating one with our hostname set as canonical
        augeasutil.execute(
            changes=[
                "ins 01 before ./1",
                "set ./01/ipaddr {}".format(fact["private_ip"]),
                "set ./01/canonical {}".format(hostname),
                "save",
            ],
            aug=aug,
        )
        return

    if not canonical_name:
        # if the record is broken and there is only ip addr without any hostname,
        # setting hostname as canonical
        augeasutil.execute(
            changes=["set {}/../canonical {}".format(record_xpath, hostname), "save"], aug=aug
        )
        return

    if canonical and canonical_name != hostname:
        # if the canonical name is different, setting canonical to hostname and moving
        # previous canonical to aliases
        changes = ["set {}/../canonical {}".format(record_xpath, hostname)]
        if has_alias:
            # if canonical happened to be in alias, removing it from there
            changes += ['rm {}/../alias[.="{}"]'.format(record_xpath, hostname)]

        changes += [
            "ins alias after {}/../canonical".format(record_xpath),
            "set {}/../alias[1] {}".format(record_xpath, canonical_name),
            "save",
        ]
        augeasutil.execute(changes=changes, aug=aug)
        return

    if not canonical and not has_alias and canonical_name != hostname:
        # if the hostname should be alias and there isn't one, adding it to aliases.
        # Also, if hostname is canonical, we should skip setting an alias. We don't know
        # what to move into canonical from existing aliases, it's better to just skip this.
        augeasutil.execute(
            changes=[
                "ins alias after {}/../canonical".format(record_xpath),
                "set {}/../alias[1] {}".format(record_xpath, hostname),
                "save",
            ],
            aug=aug,
        )


@celeryutil.agent_task(name="sys.set_hostname")
def set_hostname(hostname, domain_name=None, reboot=True, **kwargs):
    """Apply a new value for hostname - Linux / computer name - Windows

    :param str hostname: Hostname / computer name to apply.
    :param str domain_name: Domain suffix.
    :param bool reboot: Whether to reboot server after computer name is applied. (Windows-only)
    """
    if not hostname and not domain_name:
        raise exc.Error("Computer name cannot be empty")
    if not hostname:
        # This way it's possible to use cloud-generated hostname with custom domain_name
        hostname = get_hostname()

    # Work around the GV with spaces substitution, done by Scalr.
    hostname = hostname.replace(" ", "")

    if fact["os"]["family"] == "windows":
        if len(hostname) > HOSTNAME_MAX_LEN_WIN:
            raise exc.Error("Computer name is maximum %s characters" % HOSTNAME_MAX_LEN_WIN)
        pattern = re.compile(r"[\\\/\:\*\?\"\<\>\|]")
        if pattern.search(hostname):
            msg = "Computer name contains invalid characters: %s" % pattern.pattern[1:-1].replace(
                "\\", ""
            )
            raise exc.Error(msg)
        if domain_name is not None and pattern.search(hostname):
            msg = "Computer name contains invalid characters: %s" % pattern.pattern[1:-1].replace(
                "\\", ""
            )
            raise exc.Error(msg)

        LOG.info(
            "Changing computer name to %s%s",
            hostname,
            " and domain name to {}".format(domain_name) if domain_name else "",
        )

        hostname_changed = False
        if get_hostname() != hostname:
            ctypes.cdll.kernel32.SetComputerNameExA.argtypes = [
                ctypes.wintypes.CHAR,
                ctypes.wintypes.LPSTR,
            ]
            hostname = ctypes.wintypes.LPSTR(hostname.encode("utf-8"))
            exit_code = ctypes.cdll.kernel32.SetComputerNameExA(
                ctypes.wintypes.CHAR(WIN_API_COMPUTER_NAME_FORMAT.ComputerNamePhysicalDnsHostname),
                hostname,
            )
            if exit_code != 1:
                raise exc.Error("Cannot change computer name to {0}".format(hostname))
            hostname_changed = True

        if domain_name is not None and get_domain_name() != domain_name:
            domain_name = ctypes.wintypes.LPSTR(domain_name.encode("utf-8"))
            exit_code = ctypes.cdll.kernel32.SetComputerNameExA(
                ctypes.wintypes.CHAR(WIN_API_COMPUTER_NAME_FORMAT.ComputerNamePhysicalDnsDomain),
                domain_name,
            )
            if exit_code != 1:
                raise exc.Error("Cannot change domain name to {0}".format(domain_name))
            hostname_changed = True

        # We have to reboot instance to apply new hostname
        if reboot and hostname_changed:
            sysutil.reboot(reason="Changing hostname")
        else:
            # SCALRCORE-12637 For windows we need reboot server correctly after HostInit
            # phase to apply new hostname. So we return boolean hostname_changed status.
            return hostname_changed

    elif fact["manage_container_host"]:
        cmd = "hostnamectl set-hostname %s" % hostname
        subprocess2.check_output_host(cmd)
    else:
        LOG.info(
            "Changing hostname to %s%s",
            hostname,
            " and domain name to {}".format(domain_name) if domain_name else "",
        )
        # Remove various dhcp hooks (those may override hostname later) on GCE
        if fact["scalrinit"]["result"] and fact["scalrinit"]["datasource"]["instance_id"] == "gce":
            sysutil.rm_rf(GCE_SET_HOSTNAME_HOOK_DEBIAN)
            sysutil.rm_rf(GCE_SET_HOSTNAME_HOOK_CENTOS)
            for dhcp_conf_path in GCE_DHCP_CONFIG_PATHS:
                if os.path.exists(dhcp_conf_path):
                    with open(dhcp_conf_path, "r") as fp:
                        contents = fp.read()
                    with open(dhcp_conf_path, "w") as fp:
                        for cmd_to_replace in GCE_SET_HOSTNAME_CMDS:
                            contents = contents.replace(cmd_to_replace, ":")
                        fp.write(contents)

        try:
            # Set runtime hostname
            subprocess2.check_output("hostname {}".format(hostname), shell=True)
        except subprocess2.CalledProcessError as e:
            if "name too long" in e.stderr:
                host_name_max = subprocess2.check_output("getconf HOST_NAME_MAX", shell=True)
                six.reraise(
                    exc.Error,
                    exc.Error(
                        'Hostname "{}" is too long. Maximum length is {}.'.format(
                            hostname, host_name_max
                        )
                    ),
                    sys.exc_info()[2],
                )
            raise

        # Set permanent hostname
        with open("/etc/hostname", "w+") as fp:
            fp.write(hostname)

        # Apply hostname to /etc/hosts
        if domain_name is not None:
            # If domain_name is present, set fqdn first
            _edit_hosts_file(".".join((hostname, domain_name)), canonical=True)
        _edit_hosts_file(hostname, canonical=False)

        # Edit sysconfig on RedHat-based
        if os.path.isfile(SYSCONFIG_NETWORK_PATH):
            # (FAM-1488) Removing HOSTNAME definitions manually,
            # because a client reported that in some cases
            # the file contains multiple HOSTNAME definitions and augeas cannot handle it.
            with open(SYSCONFIG_NETWORK_PATH, "r") as fp:
                lines = fp.readlines()
            with open(SYSCONFIG_NETWORK_PATH, "w") as fp:
                fp.write("".join(line for line in lines if not line.startswith("HOSTNAME")))
            augeasutil.edit(
                path=SYSCONFIG_NETWORK_PATH,
                lens="Shellvars.lns",
                changes=["set HOSTNAME {}".format(hostname)],
            )
        # Reload syslog service
        LOG.debug("Reloading syslog service")
        subprocess2.run("pkill -HUP syslogd", shell=True)


@fact(binding="domain_name", no_cache=True)
def get_domain_name():
    """
    Returns [hostname on Linux | computer name on Windows]

    :rtype: str
    """
    if fact["os"]["name"] == "windows":
        ctypes.cdll.kernel32.GetComputerNameExA.argtypes = [
            ctypes.wintypes.CHAR,
            ctypes.wintypes.LPSTR,
            ctypes.wintypes.LPDWORD,
        ]
        domain_name = ctypes.wintypes.LPSTR(b"")

        ctypes.cdll.kernel32.GetComputerNameExA(
            ctypes.wintypes.CHAR(WIN_API_COMPUTER_NAME_FORMAT.ComputerNamePhysicalDnsDomain),
            domain_name,
            ctypes.wintypes.LPDWORD(ctypes.c_ulong(HOSTNAME_MAX_LEN_WIN)),
        )
        return domain_name.value.decode()  # pylint: disable=no-member
    else:
        fqdn_parts = fact["hostname"].split(".", 1)  # pylint: disable=no-member
        return fqdn_parts[1] if len(fqdn_parts) > 1 else None


def _readlines(path):
    with open(path) as fp:
        return fp.read().splitlines()


@fact(binding="hostname", no_cache=True)
def get_hostname():
    """Returns [hostname on Linux | computer name on Windows]

    :rtype: str
    """
    if fact["manage_container_host"]:
        status_data = subprocess2.check_output_host("hostnamectl status")
        for l in status_data.splitlines():
            if "Static hostname" in l:
                hostname = l.split(":")[1].strip()
                return hostname
    if fact["os"]["name"] == "windows":
        ctypes.cdll.kernel32.GetComputerNameExA.argtypes = [
            ctypes.wintypes.CHAR,
            ctypes.wintypes.LPSTR,
            ctypes.wintypes.LPDWORD,
        ]
        hostname = ctypes.wintypes.LPSTR(b"")

        ctypes.cdll.kernel32.GetComputerNameExA(
            ctypes.wintypes.CHAR(WIN_API_COMPUTER_NAME_FORMAT.ComputerNamePhysicalDnsHostname),
            hostname,
            ctypes.wintypes.LPDWORD(ctypes.c_ulong(HOSTNAME_MAX_LEN_WIN)),
        )
        return hostname.value.decode()  # pylint: disable=no-member
    else:
        return socket.gethostname()


@fact(binding="scalrinit")
def scalrinit():
    return {
        "result": datadir.result(),
        "user_data": datadir.user_data(),
        "datasource": datadir.datasource(),
    }


@fact(binding="stat.load_average", no_cache=True)
def load_average():
    """
    :return list: Load average [1, 5, 15] in 3 items list.

    Example::

        [
           0.0,      // LA1
           0.01,     // LA5
           0.05]     // LA15
    """
    return [] if fact["os"]["name"] == "windows" else list(os.getloadavg())


def _get(array_, index):
    """
    Access a list element by index avoiding 'index out of range' error.
    """
    return array_[index] if len(array_) > index else None


def _get_token():
    ttl_seconds = 21600
    now = time.time()
    expired = now - getattr(_get_token, "_expires_at_", 0) > ttl_seconds

    if expired:
        LOG.debug("Fetching metadata token...")
        conn = six.moves.http_client.HTTPConnection(AWS_METADATA_HOST, 80, timeout=1)

        conn.request(
            "PUT",
            "/latest/api/token",
            headers={"X-aws-ec2-metadata-token-ttl-seconds": str(ttl_seconds)},
        )
        response = conn.getresponse()
        if response.status != 200:
            raise IOError("Unable to fetch metadata token")

        setattr(_get_token, "_expires_at_", now + ttl_seconds)
        setattr(_get_token, "_retval_", response.read())

    return getattr(_get_token, "_retval_", None)


def _call_aws(url):
    """
    Call AWS via httplib. Require correct path.
    Host: 169.254.169.254
    """
    conn = six.moves.http_client.HTTPConnection(AWS_METADATA_HOST, 80, timeout=1)
    conn.request("GET", url, headers={"X-aws-ec2-metadata-token": _get_token()})
    response = conn.getresponse()
    if response.status == 200:
        return response.read()


def _get_ec2_hostinfo(basepath="/latest/meta-data", path="", data=None):
    """
    Recursive function that walks the EC2 metadata available in each data-subgroup.
    """
    data = data if data is not None else {}
    for line in _call_aws(posixpath.join(basepath, path)).splitlines():
        if isinstance(line, bytes):
            line = line.decode()
        if line[-1] != "/":
            call_response = _call_aws(posixpath.join(basepath, path, line))
            if call_response is not None:
                data[path.replace("/", "_") + line] = call_response.decode()
            else:
                data[path.replace("/", "_")[:-1]] = line
        else:
            _get_ec2_hostinfo(basepath, posixpath.join(path, line), data=data)
    return data


@fact(binding="ec2")
def cloud_fact():
    """
    Returns a dictionary containig cloud service info
    """
    try:
        # First check that the AWS magic URL works. If it does
        # we are running in AWS and will try to get more data.
        _call_aws("/")
    except (socket.timeout, socket.error, IOError):
        return {}
    try:
        ec2_info = _get_ec2_hostinfo()
        ec2_info["region"] = ec2_info["placement_availability-zone"][:-1]
        return ec2_info
    except socket.timeout as serr:
        logging.info("Could not read EC2 data (timeout): %s", serr)
        return {}


@fact(binding="ec2.dynamic")
def cloud_fact_dynamic():
    return _get_ec2_hostinfo("/latest/dynamic")


def lsmod():
    try:
        output = subprocess2.check_output(["lsmod"], shell=True)
        return tuple(line.split()[0] for line in output.splitlines())
    except subprocess2.CalledProcessError:
        return tuple()


@fact(binding="kernel")
def modules():
    """
    Return a dictionary containing os module list.
    """
    if fact["os"]["name"] != "windows":
        return {"release": platform.release(), "machine": platform.machine(), "modules": lsmod()}
    else:
        return {}


def is_private_ip(ipaddr):
    return any(ipaddr.startswith(prefix) for prefix in PRIVATE_IP_PREFIXES)


@fact(binding="routes")
def host_routes():
    """
    :return: dict: host routing table

    Data format::
                {
                    '169.254.0.0/16': {'dev': 'docker0', 'gw': '0.0.0.0'},
                    '172.17.0.0/16': {'dev': 'docker0', 'gw': '0.0.0.0'},
                    '192.168.1.0/24': {'dev': 'enp3s0', 'gw': '192.168.88.2'},
                    '192.168.88.0/24': {'dev': 'enp3s0', 'gw': '0.0.0.0'},
                    'default': {'dev': 'enp3s0', 'gw': '192.168.88.1'}
                }
    """

    if fact["os"]["family"] == "windows":
        raise NotImplementedError()

    routes = {}
    try:
        route_list = subprocess2.check_output("ip route", shell=True)
    except subprocess2.CalledProcessError:
        err = sys.exc_info()[1]
        LOG.error("Could not fetch host routes. Error: %s", err.stderr)
        return None
    for entry in route_list.split("\n"):
        if entry:
            data = entry.split()
            route = data[0]
            if data[1] == "via":
                gw = data[2]
                dev = data[4]
            else:
                gw = "default"
                dev = data[2]
            routes[route] = {"gw": gw, "dev": dev}
    return routes


@fact(binding="")
def interfaces():
    """
    :return dict: Network interfaces info and default interface ip.

    Data format::
        {
            'interfaces': [
                {'default': False, 'iface': 'lo', 'ipv4': '127.0.0.1', 'ipv6': '::1'},
                {'default': False, 'iface': 'docker0', 'ipv4': '172.17.0.1', 'ipv6': None},
                {'default': True, 'iface': 'eth0', 'ipv4': '10.0.2.15', 'ipv6': 'fe80::a00:27ff:fefc:2e27%eth0'}
            ],
            'private_ip': '10.0.2.15'
        }
    """
    default = dict()
    res = []
    addr_iface = {}
    for iface, snics in psutil.net_if_addrs().items():
        ipv4 = None
        ipv6 = None
        for snic in snics:
            if snic.family == socket.AF_INET:
                ipv4 = snic.address
                addr_iface[snic.address] = iface
            elif snic.family == socket.AF_INET6:
                ipv6 = snic.address
        res.append({"iface": iface, "ipv4": ipv4, "ipv6": ipv6})

    if fact["os"]["family"] == "windows":
        # Retrieving routes from command line
        with logutil.lower_level():
            output = subprocess2.check_output("route print 0.0.0.0", shell=True)
        next_is_route = False
        for line in output.splitlines():
            if line.strip().endswith("Metric"):
                next_is_route = True
                continue
            elif line.strip().endswith("==="):
                next_is_route = False
                continue
            if next_is_route:
                fields = line.strip().split()
                if fields[-2] in addr_iface:
                    default[addr_iface[fields[-2]]] = True

    else:
        # Retrieving routes from /proc interface
        with open("/proc/net/route") as f:
            for line in f.readlines():
                fields = line.strip().split()
                if fields[1] == "00000000" and fields[7] == "00000000":
                    default[fields[0]] = True

    private_ip = None
    for iface in res:
        iface["default"] = default.get(iface["iface"], False)
        if iface["default"]:
            private_ip = iface["ipv4"]

    return {"interfaces": res, "private_ip": private_ip}


@fact(binding="cpu")
def cpu_info():
    """
    :return dict: CPU info from /proc/cpuinfo

    Example::

        {'cpu':
            {0: {
                "bogomips":"5319.98",
                "hlt_bug":"no",
                "fpu_exception":"yes",
                "stepping":"10",
                "cache_alignment":"64",
                "clflush size":"64",
                "microcode":"0xa07",
                "coma_bug":"no",
                "cache size":"6144 KB",
                "cpuid level":"13",
                "fpu":"yes",
                "model name":"Intel(R) Xeon(R) CPU           E5430  @ 2.66GHz",
                "address sizes":"38 bits physical, 48 bits virtual",
                "f00f_bug":"no",
                "cpu family":"6",
                "vendor_id":"GenuineIntel",
                "wp":"yes",
                "fdiv_bug":"no",
                "power management":"",
                "flags":"fpu tsc msr pae cx8",
                "model":"23",
                "cpu MHz":"2659.994"}}}
    """
    if fact["os"]["name"] == "windows":
        return {}

    with open("/proc/cpuinfo") as fp:
        allcpudata = fp.read()
    res = {}
    for cpudata in allcpudata.split("processor"):
        try:
            lines = cpudata.split("\n")
            res[lines[0].split(":")[1]] = dict(
                [
                    [s.strip() for s in line.split(":")]
                    for line in lines[1:]
                    if line != "" and ":" in line
                ]
            )
        except IndexError:
            continue
    return res


@fact(binding="stat.cpu", no_cache=True)
def cpu_stat():
    """
    :return dict: CPU stats

    Example::

        {'user': 8416,
         'nice': 0,
         'system': 6754,
         'idle': 147309}

    """
    cpu_times = psutil.cpu_times()
    nice = 0
    try:
        nice = cpu_times.nice
    except AttributeError:
        pass
    return {
        "user": int(cpu_times.user * 100),
        "system": int(cpu_times.system * 100),
        "idle": int(cpu_times.idle * 100),
        "nice": int(nice * 100),
    }


@fact(binding="stat.memory", no_cache=True)
def mem_info():
    """
    :return dict: Memory information from /proc/meminfo.

    Example::

        {'total_swap': 0,
         'avail_swap': 0,
         'total_real': 604364,
         'total_free': 165108,
         'total_avail': 499696,
         'shared': 168,
         'buffer': 17832,
         'cached': 316756,}
    """
    if fact["os"]["name"] == "windows":
        with winutil.coinitialized_context():
            c = wmi.WMI()
            meminfo = c.Win32_PerfFormattedData_PerfOS_Memory()[0]
            sysinfo = c.Win32_ComputerSystem()[0]

        return {
            "total_swap": int(meminfo.CommitLimit) / 1024,
            "avail_swap": (int(meminfo.CommitLimit) - int(meminfo.CommittedBytes)) / 1024,
            "total_real": int(sysinfo.TotalPhysicalMemory) / 1024,
            "total_free": int(meminfo.AvailableKBytes),
            "total_avail": int(meminfo.AvailableKBytes) + int(meminfo.CacheBytes) / 1024,
        }

    info = {}
    lines = _readlines("/proc/meminfo")
    for line in lines:
        key, value = line.split(":", 1)
        info[key] = int(value.strip().split(" ")[0])

    if "MemAvailable" in info:
        total_avail = info["MemAvailable"]
    else:
        # case for kernel <3.2
        total_avail = info["MemFree"] + info["Active(file)"] + info["Inactive(file)"]
        total_avail += info["SReclaimable"]

    return {
        "total_swap": info["SwapTotal"],
        "avail_swap": info["SwapFree"],
        "total_real": info["MemTotal"],
        "total_free": info["MemFree"],
        "total_avail": total_avail,
        "shared": info.get("Shmem", 0),
        "buffer": info["Buffers"],
        "cached": info["Cached"],
    }


@fact(binding="stat.disk", no_cache=True)
def disk_stats():
    """
    :return dict: Disks I/O statistics.

    Data format::

        {
            <device>: {
                <read>: {
                    <num>: total number of reads completed successfully
                    <sectors>: total number of sectors read successfully
                    <bytes>: total number of bytes read successfully}
                <write>: {
                    <num>: total number of writes completed successfully
                    <sectors>: total number of sectors written successfully
                    <bytes>: total number of bytes written successfully},
        ...
        }

    For Linux see more at http://www.kernel.org/doc/Documentation/iostats.txt,
    https://www.kernel.org/doc/Documentation/ABI/testing/procfs-diskstats

    For MS Windows https://msdn.microsoft.com/en-us/library/aa394308(v=vs.85).aspx
    https://scalr-labs.atlassian.net/browse/FAM-1129
    """
    sector_size = 512

    if fact["os"]["name"] == "windows":
        with winutil.coinitialized_context():
            connection = wmi.WMI()
            disks = {
                disk.Name: {
                    "read": {
                        "num": int(disk.DiskReadsPerSec),
                        "bytes": int(disk.DiskReadBytesPerSec),
                    },
                    "write": {
                        "num": int(disk.DiskWritesPerSec),
                        "bytes": int(disk.DiskWriteBytesPerSec),
                    },
                }
                for disk in connection.Win32_PerfRawData_PerfDisk_LogicalDisk()
                if disk.Name != "_Total"
            }
        return disks

    lines = _readlines("/proc/diskstats")
    devicelist = {}
    for value in lines:
        # Discard the "major" and "minor" numbers because they are useless
        #
        # These and other columns are explained in the kernel documentation,
        # see `Documentation/ABI/testing/procfs-diskstats`
        params = value.split()[2:]

        device = params[0]
        params_num = len(params)
        for i in range(1, params_num - 1):
            params[i] = int(params[i])

        # Before Linux 4.19, there were 14 columns in this file
        # Since Linux 4.19, there are 18 columns
        #
        # See commit bdca3c87fb7ad1cc61d231d37eb0d8f90d001e0c in git.kernel.org
        #
        # The kernel 4.18 shipped with RHEL 8 / CentOS 8 includes this patch as well.
        if params_num in (12, 16):
            read = {"num": params[1], "sectors": params[3], "bytes": params[3] * sector_size}
            write = {"num": params[5], "sectors": params[7], "bytes": params[7] * sector_size}
        # 7 columns may occur on kernels older than 2.6.25
        #
        # See commit 28f39d553ee242000e62f6c589ee3dc6de3f9aaa in git.kernel.org
        elif params_num == 5:
            read = {"num": params[1], "sectors": params[2], "bytes": params[2] * sector_size}
            write = {"num": params[3], "sectors": params[4], "bytes": params[4] * sector_size}
        else:
            raise Exception(
                "Unexpected number of columns in /proc/diskstats: {}".format(params_num + 2)
            )
        devicelist[device] = {"write": write, "read": read}
    return devicelist


def _get_docker_mounts():
    """Return mount points that has been mounted by host for scalarizr that runs in container."""
    ret = []
    if fact["manage_container_host"]:
        ret = ["/etc/hosts", "/etc/hostname", "/etc/resolv.conf"]
        cmd = "docker inspect -f '{{ json .Mounts }}' scalarizr"
        mounts_data = subprocess2.check_output_host(cmd)
        mounts = json.loads(mounts_data)
        for m in mounts:
            ret.append(m["Destination"])
    return ret


def _filter_mounts(mounts):
    """Return mounts that are 'interesting' for Scalr users.

    :arg mounts: iterable with mount points.
    :type mounts: linuxutil.FstabLikeDict for Linux or list of strings for Windows.
    :return: list of mount point names(str).
    """
    if fact["os"]["family"] == "windows":
        regex = re.compile(r"^[a-zA-Z]:")
        result = filter(regex.match, mounts)
    else:
        skip_mpoint_re = re.compile(r"/(sys|proc|dev|selinux|rootfs|var/lib/docker)")
        skip_fstypes = ("tmpfs", "devfs", "cgroup", "fuse.lxcfs", "aufs", "nsfs", "proc")
        docker_mounts = _get_docker_mounts()

        def _valid_mount(mount):
            valid_fstype = mount.fstype not in skip_fstypes
            valid_mpoint = not skip_mpoint_re.search(mount.mpoint)
            docker_mount = mount.mpoint in docker_mounts
            return mount.mpoint and valid_fstype and valid_mpoint and not docker_mount

        result = [m.mpoint for m in filter(_valid_mount, mounts)]
    return result


@fact(binding="stat.mounts", no_cache=True)
def disk_usage():
    """Disks total/free bytes(usage).

    :return dict in format:
    {
        <mount_point>: {
            'total': int total bytes,
            'free': int free bytes
        },..
    }
    """
    mounts = sysutil.mounts()
    mounts = _filter_mounts(mounts)
    return sysutil.statvfs(mounts)


@fact(binding="stat.net", no_cache=True)
def net_stats():
    """
    :return dict: Network I/O statistics.

    Data format::

        {
            '<iface>': {
                'default': is this default route interface,
                'receive': {
                    'bytes': total received bytes
                    'packets': total received packets
                    'errors': total receive errors},
                'transmit': {
                    'bytes': total transmitted bytes
                    'packets': total transmitted packets
                    'errors': total transmit errors}},
            ...}
    """
    default = dict()
    for interface in interfaces()["interfaces"]:
        default[interface["iface"]] = interface["default"]

    return {
        iface: {
            "default": default.get(iface, False),
            "receive": {"bytes": s.bytes_recv, "packets": s.packets_recv, "errors": s.errin},
            "transmit": {"bytes": s.bytes_sent, "packets": s.packets_sent, "errors": s.errout},
        }
        for iface, s in psutil.net_io_counters(pernic=True).items()
    }


DOCKER_VERSION = ()  # type: tuple
CONTAINER_STATS = {}  # type: dict


@fact(binding="stat.container", no_cache=True)
def docker_stats(pool_size=8):
    """
    :param int pool_size: size of the ThreadPool which performs multiple `docker inspect` calls

    :return list(dict): docker container stats. Example:
        [
            {
            'container_id': '12587a2f455c1d0b8b82eca2dbe7d4176497717a906cd26e7f9abd507038617b',
            'created': '2016-05-17 18:19:52',
            'image': 'alpine:stable',
            'network': 'default',
            'name': 'fugazzi',
            'privileged': False,
            'labels': {
                'build-date': '20160701',
                'license': 'GPLv2',
                'name': 'CentOS Base Image',
                'vendor': 'CentOS'
            },
            'volumes': [{'source': '/mnt/media', 'destination': '/mnt/media'}],
            'ports': {
                'tcp/80': [
                    {'port': 8080, 'host_ip': '0.0.0.0'},
                     {'port': 9090, 'host_ip': '1.1.1.1'}],
                '443/tcp': None}},
            {
            'container_id': '21ea45e5946323066d9bd55e6ad5fe351cc18f15d9ba180925fecd67a9cc2a6f',
            'created': '2016-06-03 11:13:34',
            'image': '04c5d3b7b0656168630d3ba35d8889bd0e9caafcaeb3004d2bfbc47e7c5d35d2',
            'network': 'default',
            'name': 'spy',
            'privileged': True,
            'labels': {
                'build-date': '20150208',
                'license': 'GPLv2'
            },
            'volumes': [],
            'ports': {}},
        ]
    """
    if fact["os"]["name"] == "windows":
        return []

    try:
        global DOCKER_VERSION
        if not DOCKER_VERSION and subprocess2.which("docker"):
            # docker --version output is "Docker version {version}, build {build}"
            version_out = subprocess2.check_output_host("docker --version", shell=True).split()
            if len(version_out) < 3:
                LOG.warning(
                    "docker --version output format is unknown, "
                    "trying to gather container stats anyway"
                )
            else:
                DOCKER_VERSION = mixutil.Version(version_out[2].rstrip(","))
        if DOCKER_VERSION and DOCKER_VERSION < dockerutil.SUPPORTED_DOCKER_VERSION:
            LOG.debug(
                "Docker runtime is not supported: docker version < %s",
                dockerutil.SUPPORTED_DOCKER_VERSION,
            )
            return []
        with logutil.lower_level():
            containers = subprocess2.check_output_host("docker ps -q", shell=True).split()
    except subprocess2.TimeoutExpired:
        return []
    except subprocess2.CalledProcessError as e:
        if e.returncode == dockerutil.RC_NOT_FOUND:
            return []
        elif "Cannot connect to the Docker daemon" in e.stderr:
            return []
        raise exc.Error(e.stderr)

    prev_containers, current_containers = set(containers), set(CONTAINER_STATS.keys())

    if prev_containers != current_containers:
        removed, added = current_containers - prev_containers, prev_containers - current_containers

        for container in removed:
            del CONTAINER_STATS[container]

        thread_pool = pool.ThreadPool(pool_size)
        try:
            stats = thread_pool.map(dockerutil.get_container_stat, added)
            for item in stats:
                CONTAINER_STATS.update(item)
        finally:
            thread_pool.close()

    return list(CONTAINER_STATS.values())


@fact(binding="block_devices", no_cache=True)
def block_devices():
    """
    :return list: List of block devices including ramX and loopX
    """
    if fact["os"]["name"] == "windows":
        with winutil.coinitialized_context():
            c = wmi.WMI()
            devicelist = [
                d.Name for d in c.Win32_PerfRawData_PerfDisk_LogicalDisk() if d.Name != "_Total"
            ]
        return devicelist

    lines = _readlines("/proc/diskstats")
    devicelist = []
    for value in lines:
        devicelist.append(value.split()[2])
    return devicelist


@fact(binding="os.boot_id")
def boot_id():
    if fact["os"]["name"] == "windows":
        with winutil.coinitialized_context():
            c = wmi.WMI()
            return c.Win32_OperatingSystem()[-1].LastBootUpTime
    else:
        with open("/proc/sys/kernel/random/boot_id") as fp:
            return fp.read().strip()


def apply_proxy_global_variables(global_variables):
    """
    Sets HTTP_PROXY, HTTPS_PROXY and co. environment variables from Scalr global variables.
    Unlike other global variables these ones are relevant for the scalarizr process too,
    so require special treatment.

    :param Iterable(common.types.GlobalVariable) global_variables: List of Scalr global variables
    """
    http_proxy = no_proxy = None
    for gv in global_variables:
        if http_proxy and no_proxy:
            break
        elif not http_proxy and gv["name"] in ("HTTP_PROXY", "http_proxy"):
            http_proxy = gv["value"]
        elif not no_proxy and gv["name"] in ("NO_PROXY", "no_proxy"):
            no_proxy = gv["value"]
    if http_proxy:
        LOG.debug("Setting $HTTP_PROXY = %s", http_proxy)
        for name in ("HTTP_PROXY", "HTTPS_PROXY", "FTP_PROXY", "RSYNC_PROXY"):
            os.environ[name] = http_proxy
            os.environ[name.lower()] = http_proxy
    if no_proxy:
        LOG.debug("Setting $NO_PROXY = %s", no_proxy)
        os.environ["NO_PROXY"] = os.environ["no_proxy"] = no_proxy


def _set_windows_global_variables(global_variables):
    applied = []
    env_vars = []
    for gv in global_variables:
        if not gv["hidden"]:
            env_vars.append([gv["name"], gv["value"]])
            applied.append(common.types.GlobalVariable(gv))

    if env_vars:
        import winreg  # pylint: disable=import-error

        key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            "SYSTEM\\CurrentControlSet\\Control" "\\Session Manager\\Environment",
            0,
            winreg.KEY_ALL_ACCESS,
        )
        for name, value in env_vars:
            winreg.SetValueEx(key, name, 0, winreg.REG_EXPAND_SZ, value)
            LOG.debug("Set environment variable %s = %s", name, value)
        winreg.CloseKey(key)
    return applied


def _set_linux_global_variables(global_variables):
    def _shield_characters(txt):
        characters = ["\\", '"', "`"]
        for character in characters:
            txt = txt.replace(character, "\\" + character)
        return txt

    applied = []
    globals_path = "/etc/profile.d/scalr_globals.sh"

    with open(globals_path, "wb") as f:
        for gv in global_variables:
            if not gv["hidden"]:
                v = _shield_characters(gv["value"])
                line = 'export {}="{}"\n'.format(gv["name"], v)
                f.write(codecs.encode(line, "utf-8"))
                LOG.debug("Set environment variable (%s: %s)", gv["name"], v)
                applied.append(common.types.GlobalVariable(gv))
    if applied:
        os.chmod(globals_path, 0o644)
    return applied


def _add_user_to_sudo(username, sudo_file_path):
    """
    Add sudo permission for user
    :param username: name of user
    :param sudo_file_path: file path where setting will be saved
    """
    LOG.debug("Add sudo permission for user (%s)", username)
    changes = (
        "set spec[user='{username}']/user {username}",
        "set spec[user='{username}']/host_group/host ALL",
        "set spec[user='{username}']/host_group/command ALL",
        "set spec[user='{username}']/host_group/command/runas_user ALL",
        "set spec[user='{username}']/host_group/command/tag NOPASSWD",
    )

    augeasutil.edit(
        path=sudo_file_path,
        lens="Sudoers.lns",
        changes=[line.format(username=username) for line in changes],
    )


@celeryutil.agent_task(name="sys.add_user")
def add_user(username, sudoers=None, sudo_file_path="etc/sudoers.d/scalarizr", **kwargs):
    """
    Create local system user
    :param username: name of user
    :param sudoers: determines whether a user will have sudo permission
    :param sudo_file_path: file path where setting will be saved
    """
    LOG.debug("Create local user (%s)", username)
    if fact["manage_container_host"]:
        root_fs = "/rootfs"
    else:
        root_fs = "/"
    subprocess2.call_host("useradd -m %s" % username, shell=True)
    if sudoers:
        _add_user_to_sudo(username=username, sudo_file_path=os.path.join(root_fs, sudo_file_path))


@celeryutil.agent_task(name="sys.add_win_user")
def add_win_user(username, password, groups=None, **kwargs):
    """
    Creates a new windows user account.

    :param username: The name for windows user.
    :type username: str
    :param password: The password associated with the username.
    :type: str
    :param groups: A list of user groups.
    :type groups: list
    """
    LOG.debug("Create local user (%s)", username)
    try:
        subprocess2.check_output(
            'net user /add /y "{username}" "{password}"'.format(
                username=username, password=password,
            ),
            shell=True,
        )
    except subprocess2.CalledProcessError as e:
        if "already exists" in str(e):
            LOG.debug("Local user already exists (%s)", username)
        else:
            raise

    for groupname in groups or []:
        LOG.debug("Add local user to group (%s, %s)", username, groupname)
        try:
            subprocess2.check_output(
                'net localgroup /add "{groupname}" "{username}"'.format(
                    groupname=groupname, username=username,
                ),
                shell=True,
            )
        except subprocess2.CalledProcessError as e:
            if "already a member" in str(e):
                LOG.debug("User is already a member of the group (%s, %s)", username, groupname)
            else:
                raise


@celeryutil.agent_task(name="sys.update_sshd_config")
def update_sshd_config(changes_dict, **kwargs):
    """
    Change sshd settings by managing sshd_config
    :param changes_dict: settings that need to be changed
    :return: number of changed files or None
    """
    aug_changes_list = [" ".join(("set", key, value)) for key, value in changes_dict.items()]
    result = augeasutil.edit(path="/etc/ssh/sshd_config", lens="Sshd.lns", changes=aug_changes_list)
    if result[-1]:
        LOG.debug("Update /etc/ssh/sshd_config with values %s", changes_dict)
        return result[-1]


@celeryutil.agent_task(name="sys.set_global_variables")
def set_global_variables(global_variables, **kwargs):
    """
    Apply Global Variables to OS global environment.

    :param global_variables: A list of global variables to apply
    :type: global_variables: list(common.types.GlobalVariable)

    :return list: list(common.types.GlobalVariable) returns applied GV list
    """

    gvs = [common.types.GlobalVariable(gv) for gv in global_variables]

    for gv in gvs:
        gv.validate()

    apply_proxy_global_variables(gvs)

    if fact["os"]["family"] == "windows":
        return _set_windows_global_variables(gvs)

    return _set_linux_global_variables(gvs)


@celeryutil.agent_task(name="sys.httpcat", acks_late=False)
def httpcat(port, method, headers, body, path=None, options=None, **kwargs):
    """
    Sends data to port on localhost and returns the answer.

    :param int|str port: port number
    :param str method: HTTP method.
    :param dict headers: json-packed string of headers.
    :param str body: data to send in body.
    :param str path: request path.
    :param dict options: other request options.
    :return dict(str): localhost's response and status code
    """
    options = options or {}
    if fact["os"]["family"] == "windows":
        url = "http://127.0.0.1:{}"
    else:
        url = "http://0.0.0.0:{}"

    try:
        url = six.moves.urllib.parse.urljoin(url.format(port), path)
        req = requests.Request(method, url, headers, data=body, **options)
        prepared = req.prepare()
        session = requests.Session()
        resp = session.send(prepared)

        LOG.debug("sys.httpcat response is %s %s", resp.text, resp.status_code)
        return {"text": resp.text, "status_code": resp.status_code}
    except Exception as e:
        LOG.debug("sys.httpcat request failed with error: %s", e)
        # pylint: disable=no-member
        return {"text": "{}".format(e), "status_code": requests.codes.internal_server_error}
