"""
Agent update task and a supportive utility which performs the actual update.

isort:skip_file
"""
from __future__ import absolute_import
import os
import sys
from unittest import mock
import json
import subprocess

if __name__ == "__main__":
    # Workaround to avoid name collisions with files in the `agent/tasks`
    # directory during import. Only relevant when the file runs as standalone utility.
    del sys.path[0]
from agent.tasks import service
from agent.utils import celeryutil
from common.utils import exc
from common.utils import logutil
from common.utils import pkgutil
from common.utils import six
from common.utils.facts import fact
import common.types
import agent.config

import celery


LOG = logutil.get_logger("update" if __name__ == "__main__" else __name__)
DOWNGRADE_BLACKLIST_PATH = os.path.join(os.path.dirname(__file__), "downgrade_blacklist.json")


@fact(binding="agent")
def agent_package():
    os_families = {"redhat": "rpm", "suse": "suse", "debian": "deb", "windows": "win"}
    if fact["manage_container_host"]:
        agent_repo_type = "docker"
    else:
        agent_repo_type = os_families[fact["os"]["family"]]

    # TODO(decoratorz): Think of a better way to do it, without leaking high level
    # app logic inside a task.
    from agent.scalarizr.hub import hub

    service_version = hub.agent_version

    pkgmgr_cls = pkgutil.get_pkgmgr(agent_repo_type)
    installed_version = pkgmgr_cls.get_installed_version(
        agent.config.workflow_engine["package_name"]
    )

    if os.path.exists(DOWNGRADE_BLACKLIST_PATH):
        with open(DOWNGRADE_BLACKLIST_PATH, "r") as fp:
            downgrade_blacklist = json.load(fp)
    else:
        downgrade_blacklist = []

    package = common.types.AgentPackage(
        {
            "name": agent.config.workflow_engine["package_name"],
            "installed_version": installed_version,
            "service_version": service_version,
            "repo_type": agent_repo_type,
            "downgrade_blacklist": downgrade_blacklist,
        }
    )
    return {"package": package.to_primitive()}


@celeryutil.agent_task(name="update.do_update", bind=True)
def do_update(
    self, repository, rollback, cache_dir=None, ssl_verify_peer=True, inproc=False, **kwargs
):
    """
    :type repository: Repository
    :param repository: A repository to update from

    :type rollback: bool
    :param rollback: Is rollback allowed. Default is True.
        When False, our package manager doesn't try to rollback to the latest working version.

    :type cache_dir: str
    :param cache_dir: Package manager cache directory

    :type ssl_verify_peer: bool
    :param ssl_verify_peer: Package manager ssl verify setting

    :rtype: AgentUpdateResult
    :return: Update result structure

    :raises UpdateFailureError: When update failed, and rollback also failed or wasn't possible.

    Example: Update on bootstrap
    >>> update(
    ...    repository={'repo_type': 'deb',
    ...                'url': 'http://repo.example.com/latest/apt /',
    ...                'packages': [{'name': 'fatmouse-agent', 'arch': 'x86_64', 'version': '1.0'}]}
    ...    rollback=False)

    Note: This task doesn't actually implement this, it delegates it to the utility.
          It's done this way so we can return task result after worker is restarted during update.
    """
    if cache_dir is None:
        cache_dir = agent.config.workflow_engine["pkgmgr_cache_dir"]

    if inproc:
        return update(repository, cache_dir, ssl_verify_peer, rollback)
    else:
        update_util = os.path.realpath(__file__)
        args = {
            "task_id": self.request.id,
            "reply_to": self.request.reply_to,
            "correlation_id": self.request.correlation_id,
            "repository": repository,
            "cache_dir": cache_dir,
            "ssl_verify_peer": ssl_verify_peer,
            "rollback": rollback,
        }
        p = subprocess.Popen([sys.executable, update_util], stdin=subprocess.PIPE)
        p.stdin.write(json.dumps(args).encode())
        p.stdin.close()
        LOG.debug(
            "Delegating update to a standalone utility, "
            "this task will be ignored by the current worker."
        )
        raise celery.exceptions.Ignore


def update(repository, cache_dir, ssl_verify_peer, rollback):
    """
    :type repository: Repository
    :param repository: A repository to update from

    :type cache_dir: str
    :param cache_dir: Package manager cache directory

    :type ssl_verify_peer: bool
    :param ssl_verify_peer: Package manager ssl verify setting

    :type rollback: bool
    :param rollback: Is rollback allowed. Default is True.
        When False, our package manager doesn't try to rollback to the latest working version.

    :rtype: bool
    :return: True on success
    """
    repo_url = repository["url"]
    repo_type = repository["repo_type"]
    package = repository["packages"][0]
    dependencies = repository["packages"][1:]

    try:
        pkgmgr = pkgutil.get_pkgmgr(repo_type)(
            repo_string=repo_url,
            storage_dir=cache_dir,
            ssl_verify_peer=ssl_verify_peer,
            arch=package["arch"],
        )
        LOG.debug("Fetching repository information")
        pkgmgr.updatedb()

        # searching for a package to rollback within the cached packages
        pkginfo = pkgmgr.status(package["name"])
        backup = pkginfo.get("installed_hash")

        LOG.debug("Fetching agent packages")
        fetched = pkgmgr.fetch(package["name"], package["version"], dependencies)
    except (
        pkgutil.PackageNotFoundError,
        pkgutil.PackageDownloadError,
        pkgutil.Error,
        Exception,
    ) as e:
        LOG.error("Unable to fetch agent package from repository.")
        six.reraise(exc.UpdateRepositoryError, exc.UpdateRepositoryError(str(e)), sys.exc_info()[2])

    try:
        LOG.debug("Installing agent package")

        # TODO(decoratorz): Think of a better way to do it, without leaking high level
        # app logic inside a task.
        from agent.scalarizr.hub import hub

        s = hub.updateclient.status
        s["candidate"] = package["version"]
        hub.updateclient.status = s

        # Ignore dependencies when installing scalarizr, because the upgrade can lead to a
        # dependency error if deprecated scalarizr meta package (scalarizr-%platform%)
        # has not been removed. See: FAM-1493
        pkgmgr.install(fetched, no_deps=not bool(dependencies))
        service.restart(agent.config.workflow_engine["service_name"])
    except (pkgutil.PackageInstallError, service.Error, Exception) as e:
        e_msg = "{}".format(e)
        if rollback and backup:
            LOG.error("Error occurred during new agent package installation, trying to rollback...")

            # TODO(decoratorz): Think of a better way to do it, without leaking high level
            # app logic inside a task.
            from agent.scalarizr.updateclient import _set_state

            _set_state("in-progress/rollback")

            pkgmgr.install(backup, no_deps=not bool(dependencies))
            service.restart(agent.config.workflow_engine["service_name"])
            six.reraise(
                exc.UpdateRollbackedError, exc.UpdateRollbackedError(e_msg), sys.exc_info()[2]
            )
        LOG.error(
            "Error occurred during new agent package installation and we cannot rollback:\n%s",
            e_msg,
        )
        six.reraise(
            exc.UpdateRollbackFailureError, exc.UpdateRollbackFailureError(e_msg), sys.exc_info()[2]
        )
    return True


def update_cli():
    """Internal update utility which is run by agent.update task to do the
    actual work. Waits for arguments on stdin in json format.
    Arguments are: 'task_id', 'reply_to' and 'correlation_id', 'repository' and 'rollback'.
    """
    from agent.app import app

    args = json.loads(sys.stdin.read().strip())
    task_id = args.pop("task_id")
    request = mock.Mock()
    request.reply_to = args.pop("reply_to")
    request.correlation_id = args.pop("correlation_id")
    request.children = []
    try:
        result = update(**args)
        app.backend.mark_as_done(task_id=task_id, result=result, request=request)
    except Exception as e:
        app.backend.mark_as_failure(task_id=task_id, exc=e, request=request)


if __name__ == "__main__":
    update_cli()
