"""
Module that provides celery utils for agent.
"""
import threading

import amqp.exceptions
import celery
import celery.app.events
import celery.backends.rpc as rpc_backend
import celery.bootsteps
import celery.events.dispatcher
import celery.exceptions
import celery.worker.consumer as celery_consumer
import kombu
from celery.utils.time import humanize_seconds
from kombu.utils import retry_over_time

import agent.config
from common.utils import celeryutil
from common.utils import logutil
from common.utils.celeryutil import TaskLocalDir
from common.utils.facts import fact

LOG = logutil.get_logger(__name__)

DELIVERY_MODE_PERSISTENT = 2


class AgentTask(celery.Task):
    """Class that implements celery task for agent."""

    _local_dir_ext = threading.local()

    @property
    def local_dir_ext(self):
        """Returns celery task extension."""
        if not hasattr(self._local_dir_ext, "instance"):
            self._local_dir_ext.instance = TaskLocalDir()
        return self._local_dir_ext.instance

    @local_dir_ext.deleter
    def local_dir_ext(self):
        """Deleter for celery task extension."""
        delattr(self._local_dir_ext, "instance")

    def __call__(self, *args, **kwds):
        if not self.local_dir_ext.stack:
            task_id = self.request.id or TaskLocalDir.generate_task_id()
            if celery.current_app.conf.task_always_eager:
                LOG.info("Calling task %s locally (%s)", self.name, task_id)
            self.local_dir_ext.task_pre_run(task_id)
        self.local_dir_ext.func_pre_run(self.name)
        try:
            # pylint: disable=no-member
            return super(AgentTask, self).__call__(*args, **kwds)
        finally:
            self.local_dir_ext.func_post_run()
            if not self.local_dir_ext.stack:
                self.local_dir_ext.task_post_run()
                del self.local_dir_ext

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        """
        Error handler.
        This is run by the worker when the task fails.
        """
        LOG.debug(
            "Task (name=%s, task_id=%s) failed. " "Reason: %s\n%s", self.name, task_id, exc, einfo
        )

    def run(self, *args, **kwargs):
        raise NotImplementedError()


def agent_task(*args, **kwds):
    """Decorator for agent tasks."""
    kwds["base"] = AgentTask
    return celery.task(*args, **kwds)


class CeleryConsumer(celery_consumer.Consumer):
    """Custom celery consumer."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("without_mingle", True)
        kwargs.setdefault("without_gossip", True)
        kwargs.setdefault("without_heartbeat", True)
        super(CeleryConsumer, self).__init__(*args, **kwargs)

    # [FAM-697]
    def _establish_connection(self):
        def on_error(exc, intervals, _):
            msg = "Can't connect to broker: %s. Retry %s"
            interval = next(intervals)
            LOG.error(msg, exc, humanize_seconds(interval, "in", " "))
            return interval

        connect = super(CeleryConsumer, self).connect
        return retry_over_time(connect, amqp.exceptions.AccessRefused, errback=on_error)

    def connect(self):
        """Establish the broker connection."""
        updated = agent.config.reload()
        if updated:
            LOG.debug("Configuration was updated, reloading application.")
            self.app.close()
            self.app.conf.update(agent.config.workflow_engine["celery"])

            LOG.debug("Restarting worker pool")
            self.pool.terminate()
            self.pool.start()

        connection = self._establish_connection()

        # pylint: disable=no-member
        self.app.send_task(
            "on.agent.online",
            kwargs={"boot_id": fact["os"]["boot_id"]},
            exchange="agent.celery",
            routing_key="server",
            declare=False,
            connection=connection,
        )
        return connection


class RPCBackend(rpc_backend.RPCBackend):
    """Publishes results by sending messages."""

    def __init__(self, *args, **kwargs):
        """This prevents agent from attempts to declare AMQP entities."""
        self.Consumer.auto_declare = False
        self.Producer.auto_declare = False
        self.Exchange.passive = True
        self.Queue.passive = True
        super(RPCBackend, self).__init__(*args, **kwargs)

    def _create_exchange(
        self, name, type="direct", delivery_mode=2
    ):  # pylint: disable=redefined-builtin
        return self.Exchange("agent.results")

    # pylint: disable = no-self-use
    def on_reply_declare(self, task_id):
        """Optional list of required entities that must have been
        declared before publishing the message.
        """
        # All declarations already done on callee side
        return []

    def prepare_exception(self, exc, serializer=None):
        """Prepare exception for serialization."""
        exc_s = super(RPCBackend, self).prepare_exception(exc, serializer)
        if isinstance(exc_s, dict):
            exc_s["exc_data"] = dict(exc.__dict__)
        return exc_s

    def delete_group(self, group_id):
        raise NotImplementedError()

    def ensure_chords_allowed(self):
        raise NotImplementedError()

    def reload_group_result(self, task_id):
        raise NotImplementedError()

    def reload_task_result(self, task_id):
        raise NotImplementedError()

    def restore_group(self, group_id, cache=True):
        raise NotImplementedError()

    def save_group(self, group_id, result):
        raise NotImplementedError()


class Events(celery.app.events.Events):
    """
    This class overrides `dispatcher_cls` with custom `EventDispatcher` class
    """

    dispatcher_cls = "agent.utils.celeryutil:EventDispatcher"


class EventDispatcher(celery.events.dispatcher.EventDispatcher):
    """
    This class overrides events dispatcher for agent celery app to use a custom exchange for events
    """

    def __init__(self, *args, **kwargs):
        kwargs["delivery_mode"] = DELIVERY_MODE_PERSISTENT
        super(EventDispatcher, self).__init__(*args, **kwargs)
        self.exchange = kombu.Exchange("agent.celeryev", type="topic")

    @property
    def oid(self):
        """override standard oid for agent"""
        return celeryutil.agent_results_queue(agent.config.workflow_engine["server_id"])
