from __future__ import absolute_import

import celery
from libcloud.compute.base import Node

from common import types
from common.utils import exc
from common.utils import imputil
from common.utils import libcloudutil
from common.utils import logutil

LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    pass


@celery.current_app.task(name="_libcloud.compute.create_image")
def create_image(
    credentials=None,
    node_id=None,
    name=None,
    description=None,
    no_reboot=False,
    block_device_mapping=None,
):
    """Wrapper aroud libcloud driver.create_image() method
    :param dict credentials: Cloud credentials compatible with libcloud (must have provider
    type and location. Example: result of AwsCloudCredentials.for_libcloud() method).

    :param str node_id: ID of cloud server.

    :param str name: Name of image.

    :param str description: Image description.

    :param bool no_reboot: If True server will not be rebooted after image creation.

    :param list block_device_mapping: Block device mappings.
    """
    block_device_mapping = [
        types.Ec2BlockDeviceMapping(bdm).for_libcloud() for bdm in block_device_mapping
    ]

    credentials = types.LibcloudCredentials(credentials)
    driver = libcloudutil.create_driver(credentials)
    node = libcloudutil.find_one(Node, api_call_kwds={"ex_node_ids": [node_id]}, driver=driver)

    return driver.create_image(
        node, name, description, block_device_mapping=block_device_mapping, reboot=not no_reboot
    ).id


@celery.current_app.task(name="_libcloud.compute.wait_state_change")
def wait_state_change(
    object_type=None,
    object_id=None,
    from_state=None,
    to_state=None,
    credentials=None,
    abort_on=None,
    timeout=None,
    state_getter=None,
    ignore_errors=None,
):
    """
    Waits a libcloud object by polling for a given state, then exits.
    If objects transits to state in abort_on list, then Exception is raised.

    :param class or str object_type: A liblcoud.comput.base class, representing
        one of base libcloud object types. Can be specified via class name
        string.
    :param dict credentials: Cloud credentials compatible with libcloud (must
        have provider type and location.
        Example: result of AwsCloudCredentials.for_libcloud() method).
    :param str object_id: object's  unique identificator (typically `id`
        attribute)
    :param str state: State to wait for.
    :param list[str] abort_on: List of states that will be considered failed.
    :param function state_getter: A state field getter function
    """
    if isinstance(object_type, str):
        object_type = imputil.import_attr("libcloud.compute.base.{}".format(object_type))

    if not isinstance(credentials, types.LibcloudCredentials):
        credentials = types.LibcloudCredentials(credentials)
    driver = libcloudutil.create_driver(credentials)

    return libcloudutil.wait_state_change(
        object_type=object_type,
        object_id=object_id,
        from_state=from_state,
        to_state=to_state,
        abort_on=abort_on,
        timeout=timeout,
        driver=driver,
        state_getter=state_getter,
        ignore_errors=ignore_errors,
    )
