"""
Tasks for managing git repository.
"""
import abc
import ctypes
import os
import re
import socket
import sys
import typing as tp

import celery
import pygit2
import requests

import common.types
from common.utils import cryptoutil
from common.utils import exc
from common.utils import logutil
from common.utils import six
from common.utils import sysutil

try:
    import gevent
except ImportError:
    gevent = None


try:
    import server.config as config
except ImportError:
    import agent.config as config  # type: ignore


if sys.platform == "darwin":  # for small tests
    ssh2 = ctypes.cdll.LoadLibrary("libssh2.dylib")
elif sys.platform == "win32":
    ssh2 = ctypes.cdll.LoadLibrary("libssh2.dll")
else:
    ssh2 = ctypes.cdll.LoadLibrary("libssh2.so")

LOG = logutil.get_logger(__name__)


# problem descrition: https://scalr-labs.atlassian.net/browse/SCALRCORE-12161
if pygit2.settings.search_path[pygit2.GIT_CONFIG_LEVEL_GLOBAL]:
    pygit2.settings.search_path[pygit2.GIT_CONFIG_LEVEL_GLOBAL] = ""


def get_http_proxies_from_env() -> tp.Dict[str, str]:
    """
    Returns HTTP(S) proxies from the environment variables.
    """
    result = {}
    http_proxy = os.environ.get("http_proxy") or os.environ.get("HTTP_PROXY")
    https_proxy = os.environ.get("https_proxy") or os.environ.get("HTTPS_PROXY")
    if http_proxy:
        result["http"] = http_proxy
    if https_proxy:
        result["https"] = https_proxy
    return result


class Error(exc.Error):
    pass


class Ssh2Error(Exception):
    pass


class GitURL:
    """
    Module responsible for parsing a git URL.
    """

    POSSIBLE_REGEX: tuple = (
        # GIT+ Protocol
        # - git+https://host.xz/owner/repo.git
        # - git+ssh://host.xz/owner/repo.git
        # - git+https://dev.azure.com/perepyolka/test/_git/terraform-aws-vpc@master
        re.compile(
            r"(git\+)?"
            r"((?P<protocol>\w+)://)"
            r"((?P<username>\w+)@)?"
            r"((?P<host>[\w\.\-]+))"
            r"(:(?P<port>\d+))?"
            r"(?P<path>(\/(?P<owner>[\w\-]+)/)?"
            r"((?P<project>[\w\-]+)\/)?(\_git\/)?"  # project/_git/ used for Azure DevOps providers
            r"(\/?(?P<name>[\w\-]+)(\.git)?)?)"
            r"(?:@(?P<commit_or_branch>.+))*$"
        ),
        # HTTPS/GIT/SSH Protocols
        # - ssh://user@host.xz:123/owner/repo.git
        # - https://host.xz/owner/repo.git/
        # - git://host.xz/owner/repo.git@master
        # - https://dev.azure.com/perepyolka/test/_git/terraform-aws-vpc@master
        re.compile(
            r"^(?P<protocol>https?|git|ssh)\://"
            r"(?:(?P<username>.+)@)*"
            r"(?P<host>[a-z0-9_.-]*)"
            r"[:/]*"
            r"(?P<port>[\d]+){0,1}"
            r"(?P<path>\/((?P<owner>[\w\-]+)\/)?"
            r"((?P<project>[\w\-]+)\/)?(\_git\/)?"  # project/_git/ used for Azure DevOps providers
            r"((?P<name>[\w\-\.]+?)(\.git)?)?)"
            r"(?:@(?P<commit_or_branch>.+))*$"
        ),
        # SSH Protocol
        # - user@host.xz:22:owner/repo.git
        # - host.xz/owner/repo.git
        re.compile(
            r"^(?:(?P<username>.+)@)*"
            r"(?P<host>[a-z0-9_.-]*)[:/]*"
            r"(?P<port>[\d]+){0,1}"
            r"[:](?P<path>\/?(?P<owner>.+)/(?P<name>.+).git)"
            r"(?:@(?P<commit_or_branch>.+))*$"
        ),
        # SSH Protocol
        # - user@host.xz/path
        # - host.xz:path
        re.compile(
            r"((?P<username>\w+)@)?"
            r"((?P<host>[\w\.\-]+))"
            r"[\:\/]{1,2}"
            r"(?P<path>((?P<owner>[\w\-]+)/)?"
            r"((?P<name>[\w\-]+)(\.git)?)?)"
            r"(?:@(?P<commit_or_branch>.+))*$"
        ),
    )

    def __init__(self, url: str) -> None:
        if not url:
            raise Error("Empty git URL")
        self._url = url.rstrip("/")

        url_values = dict.fromkeys(("protocol", "username", "host", "port", "path"))
        for regex in self.POSSIBLE_REGEX:
            match = regex.search(self._url)
            if match:
                url_values.update(match.groupdict())
                break
        else:
            raise Error(f"Unable to parse git URL '{self._url}'")

        self.protocol: str = url_values["protocol"] or "ssh"
        self.username: str = url_values["username"] or "git"
        self.host: str = url_values["host"]
        self.port: str = url_values["port"]
        self.path: str = url_values["path"]
        self.branch: tp.Optional[str] = None
        self.commit_hash: tp.Optional[str] = None

        if not (self.host and self.path):
            raise Error(f"Unable to parse git URL '{self._url}'")
        self.host = self.host.strip("/")
        self.path = self.path.strip("/")

        if self.host == "github.com":
            # force 'git' username for github.com
            # see: https://github.com/libgit2/pygit2/issues/428
            self.username = "git"

        commit_or_branch = url_values["commit_or_branch"]
        if commit_or_branch:
            if self._is_sha1_hash(commit_or_branch):
                self.commit_hash, self.branch = commit_or_branch, None
            else:
                self.commit_hash, self.branch = None, commit_or_branch

    def get_url(self) -> str:
        """
        Returns original git repo URL without branch (or commit hash).
        """
        commit_or_branch = self.commit_hash or self.branch
        url = self._url[: -1 - len(commit_or_branch)] if commit_or_branch else self._url
        return url[4:] if url.startswith("git+") else url

    def as_https_url(self) -> str:
        """
        Returns HTTPS-formatted git repo URL.
        """
        if self.protocol in ("http", "https") and not self._url.startswith("git+"):
            return self.get_url()
        return f"https://{self.host}/{self.path}"

    def get_preferred_url(self) -> str:
        """
        Returns https URL if HTTP proxy is enabled, otherwise
        returns original URL.
        """
        return self.as_https_url() if get_http_proxies_from_env() else self.get_url()

    @staticmethod
    def _is_sha1_hash(maybe_sha: str) -> bool:
        """
        Checks if a string is SHA1 Hash.
        """
        if len(maybe_sha) < 5 or len(maybe_sha) > 40:
            return False
        try:
            int(maybe_sha, 16)
        except ValueError:
            return False
        return True

    def __str__(self) -> str:
        return self._url


def git_proxy_init_options(
    certificate_cb: tp.Optional[tp.Callable] = None,
    credentials_cb: tp.Optional[tp.Callable] = None,
    proxy: tp.Optional[common.types.ProxySettings] = None,
    git_repo_url: tp.Optional[GitURL] = None,
):
    """
    Initialize and return ``git_proxy_options`` structure.
    """

    def get_url_for_logging(url):
        """
        Get url without sensitive data for logging
        #
        :param url: proxy url in some format:
        1. http://testuser:testpass@<proxy-url>:3128
        2. http://<proxy-url>:3128
        :return:
        """
        if "@" in url:
            url = "http://{}".format(url.split("@")[1])
        return url

    # load proxy setting
    env_proxies = get_http_proxies_from_env()
    env_proxy = env_proxies.get("http") or env_proxies.get("https")
    if proxy:
        proxy_url = proxy.for_requests()["http"]
        proxy_url_no_auth = proxy.for_logger()
    elif env_proxy:
        proxy_url = env_proxy
        proxy_url_no_auth = get_url_for_logging(proxy_url)
    else:
        proxy_url = None

    # initialize 'git_proxy_options' struct
    proxy_opts = pygit2.ffi.new("git_proxy_options *")
    pygit2.C.git_proxy_init_options(proxy_opts, pygit2.C.GIT_PROXY_OPTIONS_VERSION)

    # configure 'git_proxy_options' struct
    if proxy_url:
        try:
            requests.head(
                git_repo_url.get_url(), proxies={"http": proxy_url, "https": proxy_url}, timeout=2
            )
        except (requests.exceptions.ProxyError, requests.exceptions.ConnectTimeout):
            raise Error(
                "Cannot connect to proxy server {}. Check url and credentials.".format(
                    proxy_url_no_auth
                )
            )
        except Exception as e:
            LOG.debug("Cannot connect to %s. Reason: %s", git_repo_url.get_url(), e)

        if proxy_url.startswith("https://"):
            raise Error("SSL connections to proxy are not supported")
        proxy_url = proxy_url.rstrip("/") + "/"  # XXX: pygit2 requires one trailing shash

        LOG.debug("Connecting via proxy server %s...", proxy_url_no_auth)

        proxy_opts.url = pygit2.ffi.from_buffer(proxy_url.encode())
        proxy_opts.type = pygit2.C.GIT_PROXY_SPECIFIED
        if certificate_cb is not None:
            proxy_opts.certificate_check = certificate_cb
        if credentials_cb is not None:
            proxy_opts.credentials = credentials_cb
    else:
        # do not attempt to connect through a proxy
        proxy_opts.type = pygit2.C.GIT_PROXY_NONE

    return proxy_opts[0]


class GitRemoteCallbacks(pygit2.RemoteCallbacks):
    """
    Module for pygit2 remote callbacks.
    """

    def __init__(
        self,
        credentials: tp.Optional[tp.Union[pygit2.Keypair, pygit2.UserPass]] = None,
        certificate: tp.Optional[str] = None,
        git_repo_url: tp.Optional[GitURL] = None,
        validate: bool = True,
        proxy: tp.Optional[common.types.ProxySettings] = None,
    ) -> None:
        self.proxy = proxy
        self.validate = validate
        self._credentials = credentials
        self._auth_call_count = 0
        self._git_repo_url = git_repo_url
        super(GitRemoteCallbacks, self).__init__(None, certificate)

    def certificate_check(self, certificate, valid, host) -> bool:
        """
        Certificate callback. Function to determine whether to accept
        the server's certificate.

        Returns: `True` to connect, `False` to abort.
        """
        return True if not self.validate else valid

    def credentials(self, url, username_from_url, allowed_types):
        """
        Credentials callback. If the remote server requires authentication
        and credentials parameter is not passed to RemoteCallbacks,
        this function will be called and its return value used for authentication
        Without this method git pull with wrong creds is stuck
        :return: git credentials object
        """
        if self._auth_call_count > 2:
            raise Error("Max retries exceeded for authentication to {}".format(url))
        self._auth_call_count += 1
        return self._credentials

    def _fill_fetch_options(self, fetch_opts) -> None:
        """
        Update `git_fetch_options` reference options.

        See:
         - https://libgit2.org/libgit2/#HEAD/type/git_fetch_options
         - https://libgit2.org/libgit2/#HEAD/type/git_proxy_options

        :param fetch_opts: `git_fetch_options` structure wrapped in CFFI.
        """
        # pylint: disable=no-member
        super(GitRemoteCallbacks, self)._fill_fetch_options(fetch_opts)
        fetch_opts.proxy_opts = git_proxy_init_options(
            proxy=self.proxy,
            certificate_cb=self._certificate_cb,
            credentials_cb=self._credentials_cb,
            git_repo_url=self._git_repo_url,
        )
        # pylint: enable=no-member


@six.add_metaclass(abc.ABCMeta)
class ABCProvider(object):
    @abc.abstractmethod
    def sync_repo(self, repo):
        """Update existing working copy, or get a new clone."""
        raise NotImplementedError()


class GitProvider(ABCProvider):
    """A Git repository operations provider."""

    def __init__(self, data_dir: str) -> None:
        self.data_dir = data_dir

    def check_basic_auth_valid(self, repo):
        """Check whether basic auth credentials (username + password) are valid

        :param repo: Git repository data
        :type repo: :class:`common.types.GitRepository`
        """
        response = requests.get(
            "{}/info/refs?service=git-upload-pack".format(repo.url),
            auth=(repo.username, repo.password),
        )
        # retry in case user did not add .git in the end of url (console git client handles this)
        if response.status_code != 200 and repo.url[-4:] != ".git":
            response = requests.get(
                "{}.git/info/refs?service=git-upload-pack".format(repo.url),
                auth=(repo.username, repo.password),
            )
        if response.status_code == 200:
            LOG.debug("Successful basic authorization to %s", repo.url)
        else:
            LOG.error("Unsuccessful basic authorization to %s", repo.url)
            raise Error("Unsuccessful basic authorization: {}".format(response.status_code))

    def check_ssh_auth_valid(self, repo):
        """Check whether ssh auth credentials (ssh_key) are valid

        :param repo: Git repository data
        :type repo: :class:`common.types.GitRepository`
        """
        git_url = GitURL(repo.url)

        def libssh2_session_last_error(sess):
            errmsg_len = ctypes.c_int(100)
            errmsg = ctypes.c_char_p(bytes(bytearray(errmsg_len.value)))
            ssh2.libssh2_session_last_error(sess, ctypes.byref(errmsg), ctypes.byref(errmsg_len), 0)
            # pylint: disable=no-member
            return errmsg.value.decode()

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((git_url.host, 22))
        sess = ssh2.libssh2_session_init_ex(None, None, None, None)
        try:
            if ssh2.libssh2_session_handshake(sess, sock.fileno()) < 0:
                raise Ssh2Error(libssh2_session_last_error(sess))

            ssh_key_path = os.path.join(self.data_dir, "ssh.key")
            sysutil.create_file(ssh_key_path, repo.ssh_key, 0o400)
            ssh_public_key = cryptoutil.generate_openssh_public_key(repo.ssh_key.encode()).decode()
            ssh_public_key_path = os.path.join(self.data_dir, "ssh_public.key")
            sysutil.create_file(ssh_public_key_path, ssh_public_key, 0o400)
            priv_key = ctypes.c_char_p(ssh_key_path.encode())
            username = ctypes.c_char_p(git_url.username.encode())
            if (
                ssh2.libssh2_userauth_publickey_fromfile_ex(
                    sess, username, len(username.value), None, priv_key, None
                )
                < 0
            ):
                raise Ssh2Error(libssh2_session_last_error(sess))
            LOG.debug("Successful ssh authentication to %s", repo.url)
        except Exception as e:
            LOG.error("Unsuccessful ssh authentication to %s: %s", repo.url, e)
            raise Error("Unsuccessful ssh authentication: {}".format(e))
        finally:
            ssh2.libssh2_session_disconnect_ex(
                sess, ctypes.c_int(1), ctypes.c_char_p(b"description"), ctypes.c_char_p(b"lang")
            )
            ssh2.libssh2_session_free(sess)

    def check_connection(self, repo):
        """
        Decrypt git repository authentication data and call appropriate method (for ssh/basic authorization)
        to check if the credentials are correct
        :param repo: dictionary with git repository data
        :return: None or connection error.
        """
        if repo.get("ssh_key"):
            repo["ssh_key"] = cryptoutil.decrypt_scalr(repo["ssh_key"], config.crypto_key).decode(
                "utf-8"
            )
            repo = common.types.GitRepository(repo)
            return self.check_ssh_auth_valid(repo)
        else:
            repo["username"] = cryptoutil.decrypt_scalr(repo["username"], config.crypto_key).decode(
                "utf-8"
            )
            repo["password"] = cryptoutil.decrypt_scalr(repo["password"], config.crypto_key).decode(
                "utf-8"
            )
            repo = common.types.GitRepository(repo)
            return self.check_basic_auth_valid(repo)

    def sync_repo(self, repo: common.types.GitRepository) -> str:
        """
        Synchronizes repository into a local directory.

        Respects the proxy environment variables named ``$HTTP_PROXY``
        & ``$HTTPS_PROXY``.

        :param repo: Repository to sync.
        """
        if not repo.work_dir:
            # XXX: an empty value may cause pygit2 crash
            raise Error(f"Unable to sync repository {repo.url}: no 'repo.work_dir' is provided")
        git_url = GitURL(repo.url)

        ssh_key_path = ssh_public_key_path = None
        if repo.ssh_key:
            repo.validate_ssl_certificate = False
            ssh_key_path = os.path.join(self.data_dir, "ssh.key")
            sysutil.create_file(ssh_key_path, repo.ssh_key, 0o400)
            ssh_public_key = cryptoutil.generate_openssh_public_key(repo.ssh_key.encode()).decode()
            ssh_public_key_path = os.path.join(self.data_dir, "ssh_public.key")
            sysutil.create_file(ssh_public_key_path, ssh_public_key, 0o400)

        if repo.username and repo.password:
            credentials = pygit2.UserPass(repo.username, repo.password)
        elif repo.username and repo.ssh_key:
            credentials = pygit2.Keypair(repo.username, ssh_public_key_path, ssh_key_path, "")
        elif repo.ssh_key:
            credentials = pygit2.Keypair(git_url.username, ssh_public_key_path, ssh_key_path, "")
        elif repo.token:
            credentials = pygit2.UserPass(
                repo.username_substitute if repo.username_substitute is not None else "oauth2",
                repo.token,
            )
        else:
            credentials = None

        # get_proxy_settings function available within server config only, on the agent proxy setting taken from env var
        try:
            proxy_settings = config.get_proxy_settings()
        except AttributeError:
            proxy_settings = None

        # shallow copy not supported by git: https://github.com/libgit2/libgit2/issues/3058
        callbacks = GitRemoteCallbacks(
            credentials=credentials,
            validate=repo.validate_ssl_certificate,
            proxy=proxy_settings,
            git_repo_url=git_url,
        )

        def _git_clone(url):
            _clone = lambda url: pygit2.clone_repository(
                url, repo.work_dir, callbacks=callbacks, checkout_branch=git_url.branch
            )
            if gevent:
                return gevent.hub.get_hub().threadpool.apply(_clone, args=(url,))
            return _clone(url)

        def git_clone(url):
            nonlocal git_url
            try:
                return _git_clone(url)
            except KeyError as err:
                error_re = re.compile(r"^.*?reference.*not found.*?$", re.IGNORECASE)
                if error_re.match(str(err)):
                    # we cannot distinguish branch and tag until we'll try to clone repository if branch not found
                    # Consider "branch" value was tag
                    git_url.commit_hash, git_url.branch = git_url.branch, None
                    return _git_clone(url)
                else:
                    raise

        LOG.info(
            "Preparing to clone Git repository at %s, revision '%s'",
            git_url.get_url(),
            git_url.branch or git_url.commit_hash,
        )
        try:
            try:
                LOG.debug("Trying to clone repository %s by URL: %s", repo.url, git_url.get_url())
                repository = git_clone(git_url.get_url())
            except pygit2.GitError as err:
                if not get_http_proxies_from_env() or git_url.protocol in ("http", "https"):
                    raise
                # try one more time via http proxy
                LOG.debug(
                    'Git clone failed with "%s", Trying to clone repository ' "%s by URL: %s",
                    err,
                    repo.url,
                    git_url.as_https_url(),
                )
                repository = git_clone(git_url.as_https_url())
        except pygit2.GitError as err:
            raise Error(f"Failed to clone git repository at {repo.url!r}. Reason: {str(err)}.")

        try:
            if git_url.commit_hash:
                # for tags we must get real oid to reset. Works for commits the same way
                oid = repository.revparse_single(git_url.commit_hash).oid
                repository.reset(oid, pygit2.GIT_RESET_HARD)
            else:
                repository.free()
        except KeyError as err:
            raise Error(
                f"Failed to checkout git repository to {git_url.commit_hash} branch/revision. "
                f"Reason: {str(err)} branch/revision does not exist."
            )
        except pygit2.GitError as err:
            raise Error(
                f"Failed to checkout git repository to {git_url.commit_hash} revision. "
                f"Reason: {str(err)}."
            )

        LOG.info("Successful clone for %s", repo.url)
        return str(repo.work_dir)


@celery.current_app.task(bind=True, name="repo.sync", local_dir=True)
def sync(self, repo):
    """Synchronizes repository into a local directory

    :param data_dir: data directory to store ssh keys
    :param repo: Repository to sync
    :type repo: Serialized :class:`common.types.GitRepository`
    """
    if not isinstance(repo, common.types.GitRepository):
        repo = common.types.GitRepository(repo)
    return GitProvider(data_dir=self.local_dir_ext.data_dir).sync_repo(repo)


@celery.current_app.task(bind=True, name="repo.validate", local_dir=True)
def validate(self, repo):
    """
    Validate git repository credentials.
    :param repo: dictionary containing git repository url and auth data
    :return: empty string - if credentials are valid
             string with error message - if credentials are invalid
    """
    try:
        GitProvider(data_dir=self.local_dir_ext.data_dir).check_connection(repo)
    except Exception as e:
        return str(e)
    else:
        return ""
