"""
Storage workflow tasks.
"""
import sys

# pylint: disable=import-error
import celery

from common import types
from common.utils import exc
from common.utils import logutil
from common.utils import six
from common.utils import sysutil

# pylint: enable=import-error


LOG = logutil.get_logger(__name__)


def _ensure_volume(locs, globs):
    """
    Ensures that the volume exists on the cloud.

    :rtype: :class:`common.types.StorageLocals`
    """
    app = celery.current_app

    locs = celery.current_app.send_task(
        "storage.prepare_storage",
        kwargs={
            "volume": locs.volume,
            "credentials": globs.credentials,
            "cloud_server": globs.cloud_server,
        },
        server_id=globs.server_id,
    ).get()
    locs = types.StorageLocals(locs)

    if locs.create_snapshot:
        result = app.send_task(
            "storage.create_snapshot",
            kwargs={
                "volume": locs.volume,
                "credentials": globs.credentials,
                "cloud_server": globs.cloud_server,
                "wait_available": True,
            },
            server_id=globs.server_id,
        ).get()
        locs.volume.snap = result.volume.snap

    if locs.create_volume:
        create_data = app.send_task(
            "storage.create_volume",
            kwargs={
                "volume": locs.volume,
                "credentials": globs.credentials,
                "cloud_server": globs.cloud_server,
                "wait_available": True,
            },
            server_id=globs.server_id,
        ).get()
        locs.volume = create_data.volume

        if locs.create_snapshot:
            app.send_task(
                "storage.delete_snapshot_maybe",
                kwargs={
                    "volume": locs.volume,
                    "credentials": globs.credentials,
                    "cloud_server": globs.cloud_server,
                },
                queue="server",
            )
            locs.volume.snap = None

    if locs.detach_volume:
        detach_data = app.send_task(
            "storage.detach_volume",
            kwargs={
                "volume": locs.volume,
                "credentials": globs.credentials,
                "cloud_server": globs.cloud_server,
            },
            server_id=globs.server_id,
        ).get()
        locs.volume = detach_data.volume
        locs.attach_volume = detach_data.attach_volume

    if locs.attach_volume:
        attach_data = app.send_task(
            "storage.attach_volume",
            kwargs={
                "volume": locs.volume,
                "credentials": globs.credentials,
                "cloud_server": globs.cloud_server,
            },
            server_id=globs.server_id,
        ).get()
        locs.import_data(attach_data)

    return locs


def _ensure_filesystem(locs, globs):
    """
    Ensures that the disk formatted.
    """
    return celery.current_app.send_task(
        "filesystem.make",
        kwargs={"device": locs.volume.device, "fstype": locs.volume.fstype},
        server_id=globs.server_id,
    ).get()


def _ensure_mount(locs, globs, fstab=False):
    """
    Ensures that the disk mounted.
    """
    app = celery.current_app

    if not locs.volume.mpoint:
        LOG.warning("Can't mount device %s, no mount point specified.", locs.volume.device)
        return locs

    try:
        result = app.send_task(
            "storage.mount_volume",
            kwargs={"volume": locs.volume, "fstab": fstab, "raise_already_mounted": True,},
            server_id=globs.server_id,
        ).get()
        locs.volume = result.volume
    except exc.AlreadyMountedError as err:
        LOG.debug("Device already mounted: %s", err)
        if err.device:
            locs.volume.device = err.device
        return locs

    # XXX: FAM-969
    import agent.scalarizr.msg as msg

    msg.send(
        msg.create(
            "BlockDeviceMounted",
            {
                "device_name": six.text_type(locs.volume.device),
                "volume_id": locs.volume.id,
                "mountpoint": locs.volume.mpoint,
            },
        )
    )
    return locs


@celery.current_app.task(name="workflows.storage.create", workflow=True)
def create(
    server_id, volume, cloud_server=None, credentials=None, mount=True, mkfs=False, fstab=True
):
    """Workflow that configures storage volume on cloud server.

    It creates a volume on the cloud if needed.

    For some clouds (i.e. EC2) it may take a snapshot of the original volume
    and recreate it in the other availability zone.

    If :attr:`volume.fstype` is passed and volume doesn't have a filesystem,
    workflow will create one.

    Mounts volume if :attr:`volume.mpoint` is specified.

    :param str server_id: A Scalr-managed server ID

    :param volume: A volume configuration to create
    :type volume: :class:`common.types.Volume`

    :param cloud_server: A Scalr-managed cloud server object
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: Credentials for the cloud
    :type credentials: :class:`common.types.CloudCredentials`

    :param bool mount: If set, volume will be mounted to it's mount point.
    :param bool mkfs: If set, volume will be formatted in corresponding filesystem.
    :param bool fstab: Re-mount device after reboot.

    :rtype: :class:`common.types.Volume`
    """
    globs = types.StorageCreateGlobals(
        {
            "server_id": server_id,
            "volume": volume,
            "cloud_server": cloud_server,
            "credentials": credentials,
        }
    )
    globs.validate()

    locs = types.StorageLocals(
        {"volume": volume, "create_filesystem": False, "mount_filesystem": False}
    )

    try:
        if globs.credentials:
            locs = _ensure_volume(locs, globs)
        else:
            LOG.info("Skipped ensure %s. Cloud credentials are not available.", locs.volume)
            if locs.volume.device:
                locs.create_filesystem = sysutil.get_fstype(locs.volume.device) is None
                locs.mount_filesystem = True

        if locs.volume.fstype:
            if mkfs and locs.create_filesystem:
                _ensure_filesystem(locs, globs)
            if mount and locs.mount_filesystem:
                locs = _ensure_mount(locs, globs, fstab=fstab)

        return locs.volume
    except Exception as e:
        e = exc.StorageError(message=str(e), volume=locs.volume)
        six.reraise(type(e), e, sys.exc_info()[2])


@celery.current_app.task(name="workflows.storage.update", workflow=True)
def update(server_id, volume, update_parameters, cloud_server=None, credentials=None):
    """Workflow that updates storage volume on cloud server.

    Depending on the cloud it can update size, type or iops of the volume.

    :param str server_id: A Scalr-managed server ID

    :param volume: Configuration that will be used to find a volume
    :type volume: :class:`common.types.Volume`

    :param update_parameters: Parameters that needs to be updated
    :type update_parameters: :class:`common.types.VolumeModification`

    :param cloud_server: A Scalr-managed cloud server object
    :type cloud_server: :class:`common.types.CloudServer`

    :param credentials: Credentials for the cloud
    :type credentials: :class:`common.types.CloudCredentials`
    """
    app = celery.current_app

    globs = types.StorageUpdateGlobals(
        {
            "server_id": server_id,
            "volume": volume,
            "update_parameters": update_parameters,
            "cloud_server": cloud_server,
            "credentials": credentials,
        }
    )
    globs.validate()

    app.send_task(
        "storage.prepare_volume_update",
        kwargs={
            "volume": globs.volume,
            "update_parameters": globs.update_parameters,
            "credentials": globs.credentials,
            "cloud_server": globs.cloud_server,
        },
        server_id=globs.server_id,
    ).get()

    app.send_task(
        "storage.update_volume",
        kwargs={
            "volume": globs.volume,
            "update_parameters": globs.update_parameters,
            "credentials": globs.credentials,
            "cloud_server": globs.cloud_server,
        },
        queue="server",
    ).get()

    # pylint: disable=no-member
    if volume.size:
        app.send_task(
            "filesystem.extend",
            kwargs={"fstype": globs.volume.fstype, "device": globs.volume.device},
            server_id=globs.server_id,
        ).get()
