"""Update workflow tasks.
"""
import itertools

import celery

import common.types
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import pkgutil
from common.utils import rpcutil
from common.utils import six

LOG = logutil.get_logger(__name__)
SLES_MIN_VERSION = "11.3"


def find_compatibility_breaks(candidate, downgrade_blacklist):
    """
    Checks that candidate package version doesn't break backward compatibility.

    :param (str|mixutil.Version) candidate: version of Scalarizr to check compatibility breaks on
    :param list(dict) downgrade_blacklist: list of versions where compatibility is broken

    :returns list: Empty if there were no compatibility breaks,
        reasons and versions in dicts otherwise.

    Format of downgrade_blacklist:
        [{
            "version": "5.1.0",
            "reason": "Unified OpenStack local meta-data format"
        }, {
            ...
        }]
    """
    if not isinstance(candidate, mixutil.Version):
        candidate = mixutil.Version(candidate)
    sorted_blacklist = sorted(
        downgrade_blacklist, key=lambda x: mixutil.Version(x["version"]), reverse=True
    )
    compatibility_breaks = itertools.takewhile(
        lambda x: candidate < mixutil.Version(x["version"]), sorted_blacklist
    )
    return list(compatibility_breaks)


def _get_repo_string(repo_type, repo_url):
    if repo_type in (common.types.RepositoryType.rpm, common.types.RepositoryType.suse):
        if repo_type == common.types.RepositoryType.rpm:
            os_name = "centos"
            releasever = "7"
        else:
            releasever = SLES_MIN_VERSION
            os_name = "suse"

        repo_string = pkgutil.RpmManager.normalize_repo_string(
            repo_url, releasever=releasever, os_name=os_name, arch="x86_64"
        )
    else:
        repo_string = repo_url
    return repo_string


@celery.current_app.task(name="update.sync_latest_versions")
def sync_latest_versions(repository, ssl_verify_peer=True, proxies=None):
    """Update repository with package's latest versions.

    :type repository: :class:`common.types.Repository`
    :param repository: The channel or repository which we are interested in.

    :type ssl_verify_peer: bool
    :param ssl_verify_peer: Instructs package manager to perform ssl peer
        verify during sync.

    :type proxies: list
    :param proxies: list of proxies understood by Requests

    :rtype: :class:`common.types.Repository`
    :returns: The repository with filled `version` fields.
    """
    repository = common.types.Repository(repository)
    repository.validate()

    repo_string = _get_repo_string(repository.repo_type, repository.url)

    for package in repository.packages:  # pylint: disable=not-an-iterable
        package_mgr = pkgutil.get_pkgmgr(repository.repo_type)(
            repo_string=repo_string,
            ssl_verify_peer=ssl_verify_peer,
            proxies=proxies,
            arch=package.arch,
            agent=False,
        )
        search_result = package_mgr.search(package.name)
        package.version = search_result.version
    return repository


def repo_from_channel(channel, repo_type, agent_os, ssl_verify_peer=True):
    """Extracts repository from channel based on repo_type and syncs its
    packages to the latest version.

    :param :class:`common.types.UpdateChannel` channel: Update channel.
    :param str repo_type: Type of the repository: rpm, deb, win.
    :param dict agent_os: Agent OS fact.
    :param bool ssl_verify_peer: Package manager ssl verify setting

    :rtype: :class:`common.types.Repository`
    :returns: Repository filled with latest package versions.
    """
    channel = common.types.UpdateChannel(channel)
    channel.validate()

    # pylint: disable=not-an-iterable
    repositories = [repo for repo in channel.repos if repo.repo_type == repo_type]
    if not repositories:
        raise exc.Error(
            ("Update channel '{channel_name}' has no {repo_type} repositories").format(
                channel_name=channel.name, repo_type=repo_type
            )
        )
    repository = repositories[0]  # type: common.types.Repository

    for package in repository.packages:
        package["arch"] = agent_os["arch"]
    repository = sync_latest_versions(repository, ssl_verify_peer=ssl_verify_peer)

    if repo_type in (common.types.RepositoryType.rpm, common.types.RepositoryType.suse):
        repository["url"] = pkgutil.RpmManager.normalize_repo_string(
            repository["url"],
            # XXX: we might refactor the os fact in the future and this cast
            # won't be needed. FAM-704
            releasever=six.text_type(agent_os["release"]),
            os_name=agent_os["name"],
            arch=agent_os["arch"],
        )

    return repository


@celery.current_app.task(name="update.update_agent", workflow=True)
def update_agent(
    server_id,
    channel,
    cache_dir=None,
    ssl_verify_peer=True,
    update_server=None,
    bootstrap=False,
    force=False,
    do_update_inproc=False,
):
    """Entry point into the update workflow.
    Decides if we should perform an update on the given server.

    :param str server_id: Id of the server we want to update
    :param UpdateChannel channel: Channel we want to update from
    :param str cache_dir: Package manager cache dir
    :param bool ssl_verify_peer: Package manager ssl verify setting
    :param UpdateServer updserver: UpdateServer used to check if updates are allowed
    :param bool bootstrap: Used to differentiate between first and subsequent updates
    :param bool do_update_inproc: Call do_update task locally, not as a Celery task

    :rtype (AgentPackage|None):
    :returns: AgentPackage with updated package version or None if update didn't happen.
    """
    app = celery.current_app
    channel = common.types.UpdateChannel(channel)
    channel.validate()

    def _get_agent_facts():
        facts = app.send_task(
            "sys.get_facts",
            kwargs={
                "keys": ["agent.package", "os", "container_host.os"],
                "invalidate_bindings": ["agent"],
            },
            server_id=server_id,
        ).get()

        return facts["container_host.os"] or facts["os"], facts["agent.package"]

    def _update_repository_config():
        if agent_os["family"] != "windows":
            app.send_task(
                "pkgmgr.repository",
                kwargs={
                    "name": channel["name"],
                    "url": repo["url"],
                    "ssl_verify": ssl_verify_peer,
                    "prefix": "scalr-agent",
                },
                server_id=server_id,
            ).get()

    update_server_api = (
        rpcutil.Client(update_server["url"], update_server["crypto_key"], sign_only=True)
        if update_server
        else None
    )

    # get agent facts BEFORE update
    agent_os, agent_package = _get_agent_facts()

    # XXX: Non-iterable value repo['packages'] is used in an iterating context
    # XXX: Value 'repo['packages']' is unsubscriptable (unsubscriptable-object)
    # pylint: disable=not-an-iterable,unsubscriptable-object
    repo = repo_from_channel(channel, agent_package["repo_type"], agent_os, ssl_verify_peer)
    repo["packages"] = [
        pkg
        for pkg in repo["packages"]  # pylint: disable=not-an-iterable
        if (
            not pkg["deprecated_since_version"]
            or mixutil.Version(pkg["version"]) < mixutil.Version(pkg["deprecated_since_version"])
        )
    ]

    repo_package = repo["packages"][0]  # pylint: disable=unsubscriptable-object

    # XXX: temporary fix for HS
    installed_version, available_version = pkgutil.fix_version_format(
        agent_package["installed_version"], repo_package["version"]
    )

    installed_version = mixutil.Version(installed_version)
    available_version = mixutil.Version(available_version)

    if installed_version == available_version:
        LOG.debug(
            "Available version %s is the same as installed %s", available_version, installed_version
        )
        LOG.debug("Updating repository config..")
        _update_repository_config()
        return None

    if not bootstrap and installed_version >= available_version:
        LOG.debug(
            "Available version %s is older than installed %s.", available_version, installed_version
        )
        LOG.debug("Updating repository config..")
        _update_repository_config()
        return None

    if bootstrap:
        break_points = find_compatibility_breaks(
            available_version, agent_package["downgrade_blacklist"]
        )
        if break_points:
            reasons = "\n".join(
                "{} - {}".format(break_point["version"], break_point["reason"])
                for break_point in break_points
            )
            LOG.warning(
                "Downgrade from %s to %s is not supported. "
                "There are backward incompatible changes in:\n%s",
                agent_package["installed_version"],
                repo_package["version"],
                reasons,
            )
            return None

    if not force and update_server_api:
        allowed = update_server_api.update_allowed(
            server_id=server_id, version=repo_package["version"], package=repo_package["name"]
        )
        if not allowed:
            raise exc.UpdateDeniedError("Update server denied update")

    error = None
    try:
        app.send_task(
            "update.do_update",
            kwargs={
                "repository": repo,
                "cache_dir": cache_dir,
                "ssl_verify_peer": ssl_verify_peer,
                "rollback": not bootstrap,
                "inproc": do_update_inproc,
            },
            server_id=server_id,
        ).get()

        # get agent facts AFTER update
        agent_os, agent_package = _get_agent_facts()

        # check installed version
        expected_version = mixutil.Version(
            repo["packages"][0]["version"]
        )  # pylint: disable=unsubscriptable-object
        agent_version = mixutil.Version(agent_package["installed_version"])
        if expected_version != agent_version:
            raise exc.UpdateIntegrityError(
                "Current installed version {} differs from expected version {} "
                "after update".format(agent_version, expected_version)
            )

        LOG.debug("Updating repository config..")
        _update_repository_config()
        return common.types.AgentPackage(agent_package)
    except Exception as e:
        error = repr(e)
        raise
    finally:
        if update_server_api and not bootstrap:
            LOG.debug("Sending report...")
            # pylint: disable=unsubscriptable-object
            update_server_api.report(
                ok=error is None,
                package=repo["packages"][0]["name"],
                version=repo["packages"][0]["version"],
                server_id=server_id,
                scalr_id=update_server["scalr_id"],
                scalr_version=update_server["scalr_version"],
                phase="",
                dist="{name} {release} {codename}".format(**agent_os),
                error=error,
                package_type="omnibus",
            )
