"""
Data types, base for all others

.. py:module:: common.types.base
   :synopsis: Data types, base for all others

.. py:moduleauthor:: Marat Komarov <marat@scalr.com>
"""
import datetime
import distutils
import json
import uuid

import kombu.utils.json
import peewee
import playhouse.shortcuts
import schematics
import schematics.transforms
from schematics import models
from schematics.models import validate
from schematics.transforms import wholelist
from schematics.types import BaseType
from schematics.types import StringType
from schematics.types.compound import DictType
from schematics.types.serializable import Serializable

from common.utils import facts
from common.utils import six


# TODO fix after update to schematics 2.x(FAM-1185)
# class Role(schematics.role.Role)
class Role(schematics.transforms.Role):
    """Extended class with ordered_fields attribute"""

    def __init__(self, function, fields):
        self.ordered_fields = fields
        super(Role, self).__init__(function, fields)


# TODO fix after update to schemtaics 2.x(FAM-1185)
schematics.transforms.Role = Role


class OverridableModelMeta(models.ModelMeta):
    """This metaclass for Model allows having base model class, and override some field values in subclass.
    Value defined in subclass is validated against it's field, and then field is transformed into
    :class:`schematics.types.serializable.Serializable`, holding the given value.

    Example::

        class Foo(Model):
            cloud = StringType(required=True)

        class Bar(Foo):
            cloud = 'aws'

        >> bar = Bar()
        >> bar.to_primitive()
        >> {'cloud': 'aws'}
    """

    def __new__(mcs, name, bases, attrs):
        origin_attrs = attrs.copy()  # preserve original attrs from class definition.
        # pylint: disable=too-many-function-args
        klass = models.ModelMeta.__new__(mcs, name, bases, attrs)

        fields = getattr(klass, "_fields")
        serializables = getattr(klass, "_serializables")
        for key, value in six.iteritems(origin_attrs):
            # test that we've got subclass value for the super field.
            if key in fields and not isinstance(value, BaseType):
                # goodbye field
                field = fields.pop(key)
                field.validate(value)
                # hello serializable
                srlzbl = Serializable(
                    lambda self, value=value: value,
                    type=field,
                    serialized_name=field.serialized_name,
                    serialize_when_none=field.serialize_when_none,
                )
                serializables[key] = srlzbl
                setattr(klass, key, srlzbl)
        return klass


@six.add_metaclass(OverridableModelMeta)
class Model(models.Model):
    """Base class for all models."""

    def __init__(self, *args, **kwargs):
        kwargs["strict"] = False
        super(Model, self).__init__(*args, **kwargs)

    def validate(self, **kwds):  # pylint: disable=arguments-differ
        role_name = kwds.pop("role", "default" if "default" in self.Options.roles else None)
        if role_name:
            try:
                role = self.Options.roles[role_name]
            except KeyError:
                msg = "Unsupported validation role %s" % role_name
                raise schematics.exceptions.ModelValidationError(msg)
            data_to_validate = {}
            data_to_skip = {}
            for field, _, value in self.atoms():
                if not role(field, value):
                    data_to_validate[field] = value
                else:
                    data_to_skip[field] = value
            try:
                data = validate(self.__class__, data_to_validate, context=data_to_skip)
            except schematics.exceptions.ValidationError as e:
                raise schematics.exceptions.ModelValidationError(e)
            self._data.update(**data)
        else:
            return super(Model, self).validate(**kwds)

    def __hash__(self):
        return str(tuple(self.to_primitive().items())).__hash__()

    class Options:
        roles = {
            # public role - role to export sanitized data
            "public": wholelist(),
            # celery role - role to export model data as celery task arguments
            "celery": wholelist(),
        }


class Enum(object):
    """Basic enum object."""

    @classmethod
    def values(cls):
        return [
            v
            for k, v in cls.__dict__.items()
            if not (callable(v) or isinstance(v, classmethod) or k.startswith("_"))
        ]


class Error(Model):
    """Class that represents error."""

    exc_type = StringType(required=True)
    """An error type (Example: "NotFoundError")

    :type: str
    """

    exc_message = StringType(required=True)
    """An error details message.

    This message is user faced

    :type: str
    """

    exc_data = DictType(StringType)
    """An error details data, the depends from :attr:`Error.exc_type`

    This data is machine oriented.

    :type: dict
    """

    @classmethod
    def from_celery(cls, celery_exc):
        """Construct :class:`Error` from Celery serialized error.

        :param dict celery_exc: Celery exception
            that :meth:`celery.backends.base.BaseBackend.prepare_exception` (or a subclass) returns
        """
        assert isinstance(celery_exc, dict)
        return Error(celery_exc)

    @classmethod
    def from_exception(cls, exc):
        """Construct :class:`Error` from Python exception object

        :type exc: :class:`Exception` or :class:`BaseException` subclass
        """
        assert isinstance(exc, (Exception, BaseException))
        e = Error()
        e.exc_type = type(exc).__name__
        e.exc_message = str(exc)
        e.exc_data = {k: v for k, v in exc.__dict__.items() if not k.startswith("__")}
        return e


def patch_json():
    def default(self, o):
        # pylint: disable=missing-docstring
        # pylint: disable=method-hidden
        if isinstance(o, models.Model):
            return o.to_primitive()
        elif isinstance(o, peewee.Model):
            return playhouse.shortcuts.model_to_dict(o, recurse=False)
        elif isinstance(o, datetime.datetime):
            return o.isoformat()
        elif isinstance(o, distutils.version.LooseVersion):
            return o.vstring
        elif isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, facts.CachedFunction):
            return o()
        elif isinstance(o, bytes):
            return o.decode()
        else:
            return super(self.__class__, self).default(o)

    class JSONEncoder(json.JSONEncoder):
        """JSON encoder, that knows how to encode
        :class:`schematics.model.Model`, :class:`datetime.datetime` and :class:`distutils.version.LooseVersion`
        """

        def default(self, o):
            # pylint: disable=method-hidden
            return default(self, o)

    class KombuJSONEncoder(kombu.utils.json.JSONEncoder):
        """Kombu JSON encoder, that knows how to encode
        :class:`schematics.model.Model`, :class:`datetime.datetime` and :class:`distutils.version.LooseVersion`
        """

        # pylint: disable=arguments-differ
        def default(self, o, **kwargs):
            # pylint: disable=method-hidden
            return default(self, o)

    # monkey patch kombu json encoder
    # pylint: disable=W0212
    kombu.utils.json._default_encoder = KombuJSONEncoder

    # monkey patch json encoder
    # pylint: disable=W0212
    json._default_encoder = JSONEncoder(
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        indent=None,
        separators=None,
        default=None,
    )
