"""
.. py:module:: common.types.data
   :synopsis: Data types shared between server and agent, server and phpclient

.. py:moduleauthor:: Marat Komarov <marat@scalr.com>
"""
import enum
import json
import os
import typing as tp

import httplib2
from libcloud.compute import types as compute_types
from libcloud.storage import types as storage_types
from schematics.exceptions import ValidationError
from schematics.transforms import whitelist
from schematics.types import BaseType
from schematics.types import BooleanType
from schematics.types import DateTimeType
from schematics.types import FloatType
from schematics.types import IntType
from schematics.types import StringType
from schematics.types import URLType
from schematics.types.compound import DictType
from schematics.types.compound import ListType
from schematics.types.compound import ModelType
from schematics.types.compound import PolyModelType

from common.types.base import Enum
from common.types.base import Model
from common.utils import cryptoutil
from common.utils import exc
from common.utils import httplib2util
from common.utils import libcloudutil
from common.utils import mixutil
from common.utils import six

try:
    import server.config as config
except ImportError:
    import agent.config as config


class ActionType(Enum):
    """Enum of action types."""

    ansibletower_job = "ansibletower.job"
    chef_bootstrap = "chef.bootstrap"
    chef_override = "chef.override"
    chef_reconverge = "chef.reconverge"
    chef_solo = "chef.solo"
    script = "script"

    @classmethod
    def class_for(cls, value):
        try:
            return {
                cls.ansibletower_job: AnsibleTowerJobAction,
                cls.chef_bootstrap: ChefClientBootstrapAction,
                cls.chef_override: ChefClientOverrideAction,
                cls.chef_reconverge: ChefClientReconvergeAction,
                cls.chef_solo: ChefSoloAction,
                cls.script: ScriptAction,
            }[value]
        except KeyError:
            raise ValidationError("Unknown action type: {!r}".format(value))


class Action(Model):
    """A base class for orchestration actions."""

    action_type = StringType(required=True, choices=ActionType.values())
    """An action discriminator.

    Allowed values are :class:`ActionType` attributes
    """

    timeout = IntType()
    """An execution timeout in seconds.

    :type: int
    """

    run_as = StringType()
    """User to run process under.

    Default is Linux - root / Windows - Administrator

    :type: str
    """

    env = DictType(StringType)
    """Process environment.

    Default is server GV + agent environment

    :type: dict
    """

    cwd = StringType()
    """A process working directory

    Default is $HOME

    :type: str
    """

    success_codes = ListType(IntType, default=[0])
    """Success return codes.

    Default is [0]

    :type: list
    """

    return_stdout = BooleanType(default=True)
    """Whether return stdout in action result?

    Default is True

    :type: bool
    """

    return_stderr = BooleanType(default=True)
    """Whether return stderr in action result?

    Default is True

    :type: bool
    """

    stdout_symlink_path = StringType()
    """Symlink to stdout.log

    The main usecase - create symlink in /var/log/scalarizr/scripting

    :type: str
    """

    stderr_symlink_path = StringType()
    """Symlink to stderr.log

    The main usecase - create symlink in /var/log/scalarizr/scripting

    :type: str
    """


class ChefAction(Action):
    """A mixin for chef actions family."""

    log_level = StringType(
        choices=["auto", "info", "debug", "warn", "error", "fatal"], default="auto"
    )
    """Chef log level.
    """

    ssl_verify_mode = StringType(choices=["verify_peer", "verify_none"])
    """Chef SSL verify mode.
    """

    # Successful run (0) and Reboot Scheduled (35) are expected exit codes.
    # see https://www.rubydoc.info/gems/chef/Chef/Application/ExitCode
    success_codes = ListType(IntType, default=[0, 35])


class ChefClientBootstrapAction(ChefAction):
    """Bootstrap chef-client action.

    Configures client.rb, runs chef-client
    """

    action_type = ActionType.chef_bootstrap

    json_attributes = DictType(BaseType, default={})
    """Chef attributes.

    A run list should be passed in ``json_attributes['run_list']``

    :type: dict
    """

    server_url = URLType()
    """Chef Server URL.

    :type: str
    """

    client_rb_template = StringType()
    """A client.rb template.

    This is a client.rb template string, understood by string.format().
    see `client.rb <https://docs.chef.io/config_rb_client.html>`_ for available
    configuration options

    :type: str
    """

    validation_key = StringType()
    """A validation.pem certificate string.

    :type: str
    """

    validation_client_name = StringType(default="chef-validator")
    """A validator chef client.

    Default is 'chef-validator'

    :type: str
    """

    environment = StringType(default=None)
    """A chef environment to operate in.

    :type: str
    """

    node_name = StringType()
    """Chef node_name.

    Default is a ``hostname`` fact

    :type: str
    """

    delete_node = BooleanType(default=False)
    """Whether delete a chef node after chef execution?

    Default is False

    :type: bool
    """

    delete_client = BooleanType(default=False)
    """Whether delete a chef client after chef execution?

    Default is False

    :type: bool
    """

    daemonize = BooleanType(default=False)
    """Daemonize chef client after bootstrap was executed

    Default is False

    :type: bool
    """


class ChefClientOverrideAction(ChefAction):
    """An override run_list chef-client action.

    Executes chef-client with override run_list.
    chef-client should be previously configured with :class:`ChefClientBootstrapAction`
    """

    action_type = ActionType.chef_override

    json_attributes = DictType(BaseType)
    """An object with Chef attributes.

    :type: dict
    """

    run_list = ListType(StringType, default=[])
    """An override run list.

    :type: list(str)
    """


class ChefClientReconvergeAction(ChefAction):
    """A chef-client reconverge action.

    Re-executes chef-client.
    chef-client should be previously configured with :class:`ChefClientBootstrapAction`
    """

    action_type = ActionType.chef_reconverge


class GitRepository(Model):
    """A Git repository parameters."""

    url = StringType(required=True)  # TODO: validator
    """A full qualified URL with revision/branch

    Example::
        https://github.com/Scalr/pecha.git@4c3ab41
        git@github.com:Scalr/pecha.git@stable
        ssh://user@server/project.git@mybranch

    :type: str
    """

    ssh_key = StringType()  # TODO: validator
    """Private SSH authorization PEM-encoded key

    :type: str
    """

    username = StringType()
    """Basic/SSH authorization username

    :type: str
    """

    password = StringType()
    """Basic authorization password

    :type: str
    """

    token = StringType()
    """Oauth2 token

    :type: str
    """

    username_substitute = StringType()
    """Username substitute for Oauth2 authorization

    :type: str
    """

    work_dir = StringType()
    """A working directory to sync into

    :type: str
    """

    validate_ssl_certificate = BooleanType(default=True)
    """Should be remote server certificate validated?

    Default is True

    :type: bool
    """


class CookbookSourceType(Enum):
    """Enum of cookbook source types"""

    repository = "repository"
    url = "url"
    path = "path"

    @classmethod
    def class_for(cls, value):
        try:
            return {
                cls.repository: CookbookRepositorySource,
                cls.path: CookbookPathSource,
                cls.url: CookbookUrlSource,
            }[value]
        except KeyError:
            raise ValidationError("Unknown cookbook source type: {!r}".format(value))


class CookbookSource(Model):
    """A base class for chef cookbooks source."""

    source_type = StringType(required=True, choices=CookbookSourceType.values())
    """A cookbooks source discriminator.

    :type: str
    """


class CookbookPathSource(CookbookSource):
    """Chef cookbooks from a local path."""

    source_type = CookbookSourceType.path

    path = StringType(required=True)
    """Path to a local cookbooks / or chef-repo directory.

    :type: str
    """


class CookbookRepositorySource(CookbookSource):
    """Chef cookbooks from Git repository."""

    source_type = CookbookSourceType.repository

    repo = ModelType(GitRepository, required=True)
    """Repository with cookbooks.

    It could follow chef repo structure (ie. environments/ cookbooks/ roles/ data_bags/),
    or only contain cookbooks.

    :type: :class:`GitRepository`
    """

    relative_path = StringType()
    """Relative directory with cookbooks inside repository.

    This parameter is using only when repository doesn't follow chef repo structure.

    :type: str
    """


class CookbookUrlSource(CookbookSource):
    """Chef cookbooks.tar.gz from URL."""

    source_type = CookbookSourceType.url

    url = URLType(required=True)
    """An http(s) url to cookbooks.tar.gz.
    """

    relative_path = StringType()
    """Relative directory with cookbooks inside the archive.

    This parameter is using only when repository doesn't follow chef repo structure.

    :type: str
    """


class ChefSoloAction(ChefAction):
    """Execute chef-solo action."""

    action_type = ActionType.chef_solo

    json_attributes = DictType(BaseType, default={})
    """An object with Chef attributes.

    A run list should be passed in ``json_attributes['run_list']``

    :type: str
    """

    solo_rb_template = StringType()
    """A solo.rb template string.

    A solo.rb template string, understood by string.format().
    see `solo.rb <https://docs.chef.io/config_rb_solo.html>`_ for available
    configuration options

    :type: str
    """

    cookbooks = PolyModelType(
        CookbookSource,
        claim_function=lambda model, data: CookbookSourceType.class_for(data.get("source_type")),
        required=True,
    )
    """Cookbooks

    :type: :class:`CookbookSource`
    """


class ScriptAction(Action):
    action_type = ActionType.script

    code = StringType()
    """A script body to save into a local file and run.

    :type: str
    """

    name = StringType()
    """A script name.

    :type: str
    """

    path = StringType()
    """Local script path to run.

    :type: str
    """

    repo = ModelType(GitRepository, default=GitRepository({"url": ""}))
    """Git repository data structure

    :type: GitRepository
    """

    relative_path = StringType()
    """Script relative path in Git repository.

    :type: str
    """

    url = StringType()
    """URL to download script from.

    :type: str
    """

    def validate(self, **kwds):
        super(ScriptAction, self).validate(**kwds)
        # pylint: disable=no-member
        if not any(
            [self.code and self.name, self.repo.url and self.relative_path, self.path, self.url]
        ):
            raise ValidationError(
                "Script code+name, repo_url+relative_path, " "path, or url is required."
            )


class ExecutedScript(Model):
    interpreter = StringType(required=True)
    """A path to interpreter that was using to execute script

    Example: '/bin/bash'

    :type: str
    """

    path = StringType(required=True)
    """A script local path

    :type: str
    """

    command = ListType(StringType, required=True)
    """An executed process args.

    :type: list
    """

    run_as = StringType(required=True)
    """An OS user the process was executed under

    :type: str
    """

    success_codes = ListType(StringType, required=True)
    """Success return codes.

    :type: list
    """

    return_code = IntType(required=True)
    """Exit status of the process.

    A value -130 indicates that process was killed by timeout.

    :type: int
    """

    stdout = StringType(required=True, default="")
    """the tail() of captured stdout from the process.

    :type: str
    """

    stdout_total_bytes = IntType(required=True)
    """A total bytes from captured stdout.

    :type: int
    """

    stdout_log_path = StringType(required=True)
    """A local path to captured stdout.

    :type: str
    """

    stderr = StringType(required=True, default="")
    """the tail() of captured stderr from the process.

    :type: str
    """

    stderr_total_bytes = IntType(required=True)
    """A total bytes from captured stderr.

    :type: int
    """

    stderr_log_path = StringType(required=True)
    """A local path to captured stderr.

    :type: str
    """

    timeout = IntType(required=True)
    """An execution timeout in seconds.

    :type: int
    """

    time_elapsed = FloatType(required=True)
    """Execution time in seconds

    :type: float
    """


class AnsibleTowerJobAction(Action):
    api_url = StringType(required=True)
    """Ansible Tower installation URL

    :type: str
    """

    username = StringType(required=True)
    """User name for Ansible Tower credentials

    :type: str
    """

    password = StringType(required=True)
    """Password for Ansible Tower credentials

    :type: str
    """

    organization = IntType()
    """Inventory organization id

    :type: int
    """

    inventory = IntType(required=True)
    """Inventory id

    :type: int
    """

    machine_credentials = IntType(required=True)
    """Machine credentials id

    :type: int
    """

    job_template = IntType(required=True)
    """Job template id

    :type: int
    """

    host = StringType(required=True)
    """Ansible Tower managed host name

    :type: str
    """

    scm_branch = StringType()
    """Branch to execute job template against

    :type: string
    """

    def validate(self, *args, **kwargs):
        """Target host validation"""
        # pylint: disable=arguments-differ
        super(AnsibleTowerJobAction, self).validate(*args, **kwargs)
        # pylint: disable=no-member
        if not self.host.strip():
            raise ValidationError("Managed host name is required.")


class AnsibleTowerJobResult(ExecutedScript):
    job_id = IntType(required=True)
    """Job id

    :type: int
    """

    job_template_name = StringType(required=True)
    """Job template name

    :type: int
    """

    inventory = IntType(required=True)
    """Inventory id

    :type: int
    """

    status = StringType(required=True)
    """Job execution status

    :type: str
    """


class BootstrappedChefClient(Model):
    executed_script = ModelType(ExecutedScript, required=True)
    """An executed bootstrap chef-client script

    :type: :class:`ExecutedScript`
    """

    node_name = StringType(required=True)
    """A computed Chef node name

    :type: str
    """


class Rule(Model):
    """A class that incaplulates Scalr orchestration rule."""

    action = PolyModelType(
        Action,
        claim_function=lambda model, data: ActionType.class_for(data.get("action_type")),
        required=True,
    )
    """What action to execute?

    :type: :class:`Action`
    """

    blocking = BooleanType(default=True)
    """Should all next orchestration rules in a chain wait
    this action completion?

    Default is True

    :type: bool
    """


class GlobalVariable(Model):
    """A Scalr global variable object."""

    name = StringType(required=True)
    """Global variable name.

    :type: str
    """

    value = StringType()
    """Global variable value.

    :type: str
    """

    hidden = BooleanType(default=False)
    """Should this variable be persisted on a server?

    :type: bool
    """


class ServerRulesMapping(Model):
    """A server to orchestration rules mapping."""

    server_id = StringType(required=True)
    """A Scalr managed server_id

    :type: str
    """

    rule_indexes = ListType(IntType, required=True)
    """Indexes of the orchestration rules (from orchestration.rules list)
    that matching this server

    :type: list(int)
    """

    global_variables = ListType(ModelType(GlobalVariable))
    """Global variables to apply to this server.

    :type list(:class:`GlobalVariable`)
    """


class OrchestrationGlobals(Model):
    """Orcheatration workflow globals."""

    rules = ListType(ModelType(Rule), required=True)
    """Rules involved into this orchestration.

    :type: list(:class:`Rule`)
    """

    server_rules_mappings = ListType(ModelType(ServerRulesMapping), required=True)
    """A list of :class:`ServerRulesMapping` for orchestrating servers.

    :type: list(:class:`ServerRulesMapping`)
    """


class OrchestrationServerThreadLocals(Model):
    """Orchestration workflow a single server orchestration thread locals."""

    server_id = StringType(required=True)
    global_variables = ListType(ModelType(GlobalVariable))
    rules = ListType(ModelType(Rule), required=True)


class OrchestrationLocals(Model):
    """Orchestration workflow main thread locals."""

    num_servers = IntType(required=True)
    server_thread_locals_list = ListType(ModelType(OrchestrationServerThreadLocals))


class ChefInit(Model):
    install = StringType()
    bootstrap = ModelType(ChefClientBootstrapAction)
    solo = ModelType(ChefSoloAction)


class InitializedChef(BootstrappedChefClient):
    pass


class ServerInit(Model):
    hostname = StringType()
    sync_time = BooleanType()
    chef = ModelType(ChefInit)
    global_variables = ListType(ModelType(GlobalVariable))
    # volumes = ListType()


class InitializedServer(Model):
    hostname = StringType(required=True)
    """The server applied hostname

    :type: str
    """

    chef = ModelType(InitializedChef)
    """An executed bootstrap chef-client script

    :type: :class:`InitializedChef`
    """

    global_variables = ListType(ModelType(GlobalVariable))
    """Global variables applied on the server.

    :type list(:class:`GlobalVariable`)
    """
    # volumes = None


class InitGlobals(Model):
    server_id = StringType(required=True)
    """Scalr-managed server ID

    :type: str
    """

    init = ModelType(ServerInit, required=True)
    """Server initialization data.

    :type: :class:`ServerInit`
    """


class InitLocals(InitGlobals):
    pass


class LibcloudCredentials(Model):
    """
    Container for parameters that are used to create libcloud driver.
    """

    provider = StringType(required=True)
    """Used to choose libcloud provider.
    """

    args = BaseType(default=tuple())
    """Passed as positional arguments when creating driver.
    """

    kwargs = BaseType(default={})
    """Passed as named arguments when creating driver.
    """


class CloudServer(Model):
    """Server in a Cloud (aka Node, Instance)"""

    id = StringType(required=True)
    """Cloud server ID

    Example: i-1ee871a

    :type: str
    """

    location = StringType(required=True, serialized_name="cloud_location")
    """Cloud location

    Example: us-west-1

    :type: str
    """

    instance_type = StringType()
    """ Type of the server

    Example: m3.medium
    """

    def __str__(self):
        return "Instance (id={id}, location={location})".format(id=self.id, location=self.location)


class EC2CloudServer(CloudServer):
    """Server in EC2."""

    availability_zone = StringType(required=True)
    """Specific availability zone in the current cloud location.

    :type: str
    """


class AzureCloudServer(CloudServer):
    """Server in Azure Cloud."""

    resource_group = StringType(required=True)
    """The name of the resource group.

    :type: str
    """


class ProxyType(Enum):
    """Enum of supported proxy types.
    https://github.com/Scalr/int-scalr/blob/master/app/etc/config.yml-sample#L16-L27
    """

    http = 0
    socks4 = 4
    socks5 = 5


class ProxyAuthType(Enum):
    """Enum of proxy auth types.
    https://github.com/Scalr/int-scalr/blob/master/app/etc/config.yml-sample#L16-L27
    """

    basic = 1
    digest = 2
    gssneg = 4
    ntlm = 8
    any = -1


class ProxySettings(Model):
    """Proxy settings container
    See https://github.com/Scalr/int-scalr/blob/master/app/etc/config.yml-sample#L16-L27
    """

    _SCHEMAS = {4: "socks4", 5: "socks5", 0: "http"}

    host = StringType(required=True, default="localhost")
    """Proxy host, Default: localhost"""

    password = StringType(deserialize_from="pass")
    """Auth password"""

    port = IntType(required=True, default=3128)
    """Proxy port"""

    type = IntType(required=True, default=ProxyType.http, choices=ProxyType.values())
    """Proxy type. One of: http, socks4, socks5"""

    user = StringType()
    """Auth user"""

    authtype = IntType(default=ProxyAuthType.basic, choices=ProxyAuthType.values())
    """Auth type. One of: 'Basic', 'Digest', 'GSSNeg', 'NTLM', 'any'"""

    no_proxy = StringType()
    """The comma separated list of hosts to exclude from proxying"""

    @property
    def url(self) -> str:
        """Return the proxy URL."""
        scheme = self._SCHEMAS[self.type]
        if self.user and self.password:
            return f"{scheme}://{self.user}:{self.password}@{self.host}:{self.port}"
        return f"{scheme}://{self.host}:{self.port}"

    def for_libcloud(self) -> str:
        """Returns proxy url for libcloud driver creation."""
        if self.authtype not in (ProxyAuthType.basic, ProxyAuthType.any):
            raise exc.NotSupportedError(
                "Unsupported auth type for libcloud proxy settings: {}".format(self.authtype)
            )

        if self.type != ProxyType.http:
            raise exc.NotSupportedError(
                "Libcloud supports only http proxy, type given: {}".format(self.type)
            )

        libcloud_proxy_params = {
            "scheme": self._SCHEMAS[self.type],
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password,
        }
        return mixutil.build_url(**libcloud_proxy_params)

    def for_requests(self) -> tp.Dict[str, str]:
        """Returns proxy urls for requests library.

        Note: The requests library does not support "no_proxy" option - https://github.com/psf/requests/issues/5000
        """
        proxy_params = {
            "scheme": self._SCHEMAS[self.type],
            "host": self.host,
            "port": self.port,
            "user": self.user,
            "password": self.password,
        }
        proxy_url = mixutil.build_url(**proxy_params)
        return {"http": proxy_url, "https": proxy_url, "no_proxy": self.no_proxy}

    def for_logger(self) -> str:
        """
        Returns proxy url without sensitive data
        """
        looging_params = {
            "scheme": self._SCHEMAS[self.type],
            "host": self.host,
            "port": self.port,
        }
        return mixutil.build_url(**looging_params)

    def for_httplib2(self) -> httplib2.ProxyInfo:
        """Returns proxy settings for the httplib2.

        :raises :class:`exc.NotSupportedError`
        """
        if self.authtype not in (ProxyAuthType.basic, ProxyAuthType.any):
            basic_auth_code = ProxyAuthType.basic.value  # type: ignore
            raise exc.NotSupportedError(
                f"Unsupported auth type for httplib2 proxy settings: {self.authtype}. "
                f"Only basic auth ({basic_auth_code}) is supported."
            )

        if self.type == ProxyType.socks4:
            proxy_type = httplib2.socks.PROXY_TYPE_SOCKS4
        elif self.type == ProxyType.socks5:
            proxy_type = httplib2.socks.PROXY_TYPE_SOCKS5
        elif self.type == ProxyType.http:
            proxy_type = httplib2.socks.PROXY_TYPE_HTTP
        else:
            available = [f"{i.value} ({i.name})" for i in tuple(ProxyType)]  # type: ignore
            raise exc.NotSupportedError(
                f"Unknown proxy type: {self.type}, available: {', '.join(available)}"
            )

        return httplib2util.ProxyInfo(
            proxy_type,
            self.host,
            self.port,
            proxy_user=self.user,
            proxy_pass=self.password,
            no_proxy=self.no_proxy,
        )

    def for_shell(self) -> tp.Dict[str, str]:
        """Returns proxy settings for the shell environment."""
        # pylint: disable=no-member

        # This workaround removes leading dots from all domains to make the
        # no_proxy format more consistent across different HTTP clients like
        # curl or net/http (golang),
        # e.g. no_proxy='.example.com' should work the same as no_proxy='example.com'
        no_proxy = ",".join(u.lstrip(".") for u in (self.no_proxy or "").split(","))

        return {
            "HTTPS_PROXY": self.url,
            "HTTP_PROXY": self.url,
            "NO_PROXY": no_proxy,
            "https_proxy": self.url,
            "http_proxy": self.url,
            "no_proxy": no_proxy,
        }


class CloudCredentials(Model):
    """A base class for cloud credentials family."""

    _encrypted_fields = []
    """List of fields with encrypted values"""

    _unique_key_fields = []
    """List of fields for 'uniqueness' calculation"""

    _subject_key = None

    account_id = None
    auth_url = None
    auth_url_key = None
    proxy = ModelType(ProxySettings)

    id = StringType(required=False, serialized_name="cloud_credentials_id")
    """Cloud Credentials Id

        :type: str
    """

    cloud = StringType(required=False, default=None)
    """Cloud

        :type: str
    """

    env_id = IntType(required=False)
    """Environment ID

        :type: int
    """

    def __init__(self, *args, **kwargs):
        super(CloudCredentials, self).__init__(*args, **kwargs)
        if not self.proxy and self.cloud and getattr(config, "get_proxy_settings", None):
            self.proxy = config.get_proxy_settings(self.cloud)

    def to_libcloud(self, location=None, api="compute", retry_delay=None, timeout=None):
        """Returns credentials for libcloud

        :rtype: LibcloudCredentials
        """
        kwargs = {}
        if self.proxy:
            # pylint: disable=no-member
            kwargs["proxy_url"] = self.proxy.for_libcloud()

        kwargs["retry_delay"] = retry_delay or (5, 10, 20, 40, 60)
        kwargs["backoff"] = 1
        if timeout is None:
            kwargs["timeout"] = config.workflow_engine["global_timeout"]
        else:
            kwargs["timeout"] = timeout

        return LibcloudCredentials(dict(kwargs=kwargs))

    @classmethod
    def from_db(cls, data, **kwds):
        """Returns model instance from raw(database) data.

        :param list data: list of dicts with raw data from database
        :param dict kwds: additional keyword arguments for `schematics.Model` class
        """
        normalized_data = {row["name"]: row["value"] for row in data}
        for field in cls._encrypted_fields:
            if field in normalized_data:
                normalized_data[field] = cryptoutil.decrypt_scalr(
                    normalized_data[field], config.crypto_key
                ).decode()
        if data:
            normalized_data["cloud_credentials_id"] = data[0].get("cloud_credentials_id")
            normalized_data["env_id"] = data[0].get("env_id")
        return cls(normalized_data, **kwds)

    def to_db(self):
        """Serializes model into database-ready form.

        The reverse of ``from_db``
        """
        retval = self.to_primitive()
        for field in self._encrypted_fields:
            if field in retval and retval.get(field) is not None:
                retval[field] = cryptoutil.encrypt_scalr(
                    retval[field].encode(), config.crypto_key
                ).decode()
        return tuple({"name": name, "value": value} for name, value in retval.items())

    def field_to_db(self, field):
        """Returns single field in the database-ready form.

        :param str field: Field name
        :return: Dict with 2 keys: ``name`` and ``value``
        :rtype: Dict[str, str]
        """
        # cast to primitive, cause self._encrypted_fields is defined
        # for serialized field name, that may be differ from attribute name.
        value = self.to_primitive().get(field)

        if field in self._encrypted_fields and value is not None:
            value = cryptoutil.encrypt_scalr(value.encode(), config.crypto_key).decode()
        return {"name": field, "value": value}

    @property
    def subject_key(self):
        """Returns the name of the subject field of this cloud credentials
        or raises NotImplementedError.
        """
        if self._subject_key:
            return self._subject_key
        raise NotImplementedError()

    @property
    def unique_key(self):
        """Builds unique string for current cloud credentials

        :rtype: str
        """
        # pylint: disable=protected-access
        fields = [
            str(getattr(self, field)) if getattr(self, field) is not None else ""
            for field in self._unique_key_fields
        ]
        fields = [self.cloud] + fields
        return cryptoutil.sha256("|".join(fields))

    @property
    def subject(self):
        """Returns the subject of this cloud credentials (AWS: access_key, Azure: client_id, etc...)"""
        return getattr(self, self.subject_key)

    @property
    def cloud_account(self):
        """Returns Scalr cloud account"""
        return None

    @property
    def cloud_id(self):
        """ Returns unique id of the cloud """
        if self.auth_url:
            return six.moves.urllib.parse.urlparse(self.auth_url).hostname

    def __str__(self):
        data = self.to_primitive(role="public")
        if self.cloud_id:
            data["cloud_id"] = self.cloud_id
        if self.auth_url:
            data["auth_url"] = self.auth_url
        return "{} {}".format(self.__class__.__name__, data)

    def __repr__(self):
        return self.__str__()

    class Options:
        roles = {
            # public role - role to export sanitized data
            "public": whitelist("id", "cloud", "env_id"),
            # celery role - role to export CloudCredentials as celery task arguments
            "celery": whitelist("id"),
        }


class AwsCloudCredentials(CloudCredentials):
    """Credentials from Amazon Web Services."""

    _encrypted_fields = ["access_key", "secret_key"]

    _unique_key_fields = ["account_id", "access_key", "account_type"]

    _subject_key = "access_key"

    _s3_region_provider_map = {
        # 'cn-northwest-1': TODO: update Libcloud,
        # 'eu-west-3': TODO: update Libcloud,
        "ap-northeast-1": storage_types.Provider.S3_AP_NORTHEAST,
        "ap-northeast-2": storage_types.Provider.S3_AP_NORTHEAST1,
        "ap-northeast-3": storage_types.Provider.S3_AP_NORTHEAST2,
        "ap-south-1": storage_types.Provider.S3_AP_SOUTH,
        "ap-southeast-1": storage_types.Provider.S3_AP_SOUTHEAST,
        "ap-southeast-2": storage_types.Provider.S3_AP_SOUTHEAST2,
        "ca-central-1": storage_types.Provider.S3_CA_CENTRAL,
        "cn-north-1": storage_types.Provider.S3_CN_NORTH,
        "eu-central-1": storage_types.Provider.S3_EU_CENTRAL,
        "eu-west-1": storage_types.Provider.S3_EU_WEST,
        "eu-west-2": storage_types.Provider.S3_EU_WEST2,
        "sa-east-1": storage_types.Provider.S3_SA_EAST,
        "us-east-1": storage_types.Provider.S3,
        "us-east-2": storage_types.Provider.S3_US_EAST2,
        "us-gov-west-1": storage_types.Provider.S3_US_GOV_WEST,
        "us-west-1": storage_types.Provider.S3_US_WEST,
        "us-west-2": storage_types.Provider.S3_US_WEST_OREGON,
    }

    cloud = "ec2"

    account_id = StringType(required=True)
    """Amazon Account ID

    :type: str
    """

    account_type = StringType(
        choices=["regular", "gov-cloud", "cn-cloud"], default="regular", required=True
    )
    """Amazon Account Type

    :type: str
    """

    access_key = StringType(required=True)
    """Amazon Access Key ID

    :type: str
    """

    secret_key = StringType(required=True)
    """Amazon Secret Key

    :type: str
    """

    private_key_path = StringType()

    certificate_path = StringType()

    ec2_url = URLType()

    detailed_billing_enabled = BooleanType(serialized_name="detailed_billing.enabled")
    """AWS Detailed Billing enable flag

    :type: bool
    """

    detailed_billing_region = StringType(serialized_name="detailed_billing.region")
    detailed_billing_bucket = StringType(serialized_name="detailed_billing.bucket")
    detailed_billing_credentials_id = StringType(serialized_name="detailed_billing.credentials_id")

    cost_and_usage_enabled = BooleanType(serialized_name="cost_and_usage.enabled")
    """AWS Cost and Usage enable flag

    :type: bool
    """

    cost_and_usage_region = StringType(serialized_name="cost_and_usage.region")
    cost_and_usage_bucket = StringType(serialized_name="cost_and_usage.bucket")
    cost_and_usage_report_prefix = StringType(serialized_name="cost_and_usage.report_prefix")
    cost_and_usage_report_name = StringType(serialized_name="cost_and_usage.report_name")
    cost_and_usage_credentials_id = StringType(serialized_name="cost_and_usage.credentials_id")

    @property
    def cloud_account(self):
        """Returns Scalr cloud account"""
        return self.account_id

    def to_libcloud(self, location=None, api="compute", retry_delay=None, timeout=None):
        """Returns credentials for libcloud

        :param str location: AWS region. Required.
        :param str api: libcloud api type ('compute', 'storage', 'dns' etc..)

        :rtype: ``LibcloudCredentials``
        """
        if location is None:
            raise ValueError("Must provide `location` value.")
        creds = super(AwsCloudCredentials, self).to_libcloud(
            location=location, retry_delay=retry_delay, timeout=timeout
        )
        kwargs = {"region": location}
        if self.ec2_url:
            url = six.moves.urllib.parse.urlparse(self.ec2_url)
            kwargs.update({"secure": url.scheme == "https", "host": url.hostname, "port": url.port})

        if api == "compute":
            creds.provider = compute_types.Provider.EC2
        elif api == "storage":
            if location not in self._s3_region_provider_map:
                raise exc.NotSupportedError("S3 Region '{}' is not supported.".format(location))
            creds.provider = self._s3_region_provider_map[location]
        else:
            raise exc.NotSupportedError("Libcloud API '{}' is not supported.".format(api))

        creds.args = (self.access_key, self.secret_key)
        # pylint: disable=no-member
        creds.kwargs.update(kwargs)
        return creds


class GceCloudCredentials(CloudCredentials):
    """Credentials from Google Compute Engine"""

    _encrypted_fields = ["client_id", "account_id", "project_id"]

    _unique_key_fields = ["client_id", "project_id", "account_id"]

    cloud = "gce"

    account_id = StringType(required=True)
    """Service account email (Example: )

    :type: str
    """

    client_id = StringType(required=True)
    """Client ID

    :type: str
    """

    project_id = StringType(required=True)
    """Project ID

    :type: str
    """

    private_key = StringType(required=True)
    """Private key contents (p12 or json)

    :type: str
    """

    json_key = DictType(BaseType, required=True)
    """
    dictionary with service account data including the keys:
    private_key - Private key contents (p12 or json)
    private_key_id - Private key id
    token_uri - URI to oauth2 token generation
    account_type - service_account
    """

    dataset_name = StringType(serialized_name="detailed_billing.gce.dataset_name")
    """ BigQuery data set name"""

    detailed_billing_enabled = BooleanType(serialized_name="detailed_billing.enabled")
    """GCE Detailed Billing enable flag

    :type: bool
    """

    @property
    def cloud_account(self):
        """Returns Scalr cloud account"""
        return self.project_id

    def to_libcloud(self, location=None, api="compute", retry_delay=None, timeout=None):
        """Returns credentials for libcloud

        :param str location: GCE datacenter
        :param str api: libcloud api type ('compute', 'storage', 'dns' etc..)

        :rtype: ``LibcloudCredentials``
        """

        creds = super(GceCloudCredentials, self).to_libcloud(
            location=location, retry_delay=retry_delay, timeout=timeout
        )
        # pylint: disable=unsubscriptable-object
        kwargs = {
            "user_id": self.account_id,
            "key": self.private_key,
            "project": self.project_id,
            "datacenter": location,
        }
        if api == "compute":
            creds.provider = compute_types.Provider.GCE
        elif api == "storage":
            creds.provider = storage_types.Provider.GOOGLE_STORAGE
        else:
            raise exc.NotSupportedError("Libcloud API '{}' is not supported.".format(api))

        # pylint: disable=no-member
        creds.kwargs.update(kwargs)
        return creds

    @classmethod
    def from_db(cls, data, **kwds):
        normalized_data = {row["name"]: row["value"] for row in data}

        json_key = normalized_data.get("json_key")
        if json_key:
            json_key = cryptoutil.decrypt_scalr(json_key, config.crypto_key).decode()
            key_file_data = json.loads(json_key)
            data = data + [
                {"name": "json_key", "value": key_file_data},
                {"name": "private_key", "value": key_file_data["private_key"]},
            ]

        service_account_name = normalized_data.get("service_account_name")
        if service_account_name:
            data = data + [{"name": "account_id", "value": service_account_name}]
        return super(GceCloudCredentials, cls).from_db(data, **kwds)


class VMwareCloudCredentials(CloudCredentials):
    """Credentials for VMware services."""

    _encrypted_fields = ["username", "password"]

    _unique_key_fields = ["url", "username"]

    cloud = "vmware"

    auth_url_key = "url"

    username = StringType(required=True)
    """Username.

    :type: str
    """

    password = StringType(required=True)
    """Password

    :type: str
    """

    url = StringType(required=True)
    """URL to your VmWare endpoint.

    Example: https://1.2.3.4/sdk

    :type: str
    """

    host = StringType(required=True)
    """VmWare vSphere host address.

    Example: 1.2.3.4

    :type: str
    """

    detailed_billing_enabled = BooleanType(serialized_name="detailed_billing.enabled")
    """VMware Detailed Billing enable flag

    :type: bool
    """

    @property
    def cloud_account(self):
        """Returns Scalr cloud account"""
        return self.username

    @property
    def auth_url(self):
        """ URL to the authentication | endpoint service """
        return self.url

    def to_pyvmomi(self):
        """Builds kwargs for `pyVim.connect.SmartConnect`"""
        return {
            "host": six.moves.urllib.parse.urlparse(self.url).netloc,
            "user": self.username,
            "pwd": self.password,
        }

    def to_libcloud(self, **kwargs):  # pylint: disable=arguments-differ
        """
        Returns credentials for libcloud
        :rtype: ``LibcloudCredentials``
        """
        creds = super(VMwareCloudCredentials, self).to_libcloud(**kwargs)
        # pylint: disable=no-member
        timeout = creds.kwargs.get("timeout")
        creds.args = (self.username, self.password)
        kwargs = {"url": self.url, "timeout": timeout}
        creds.provider = compute_types.Provider.VSPHERE
        creds.kwargs.update(kwargs)
        return creds


class AzureCloudCredentials(CloudCredentials):
    """Credentials from Azure services."""

    _encrypted_fields = ["app_client_id", "app_secret_key", "tenant_id"]

    _unique_key_fields = ["account_type", "client_id", "tenant_id", "subscription_id"]

    _subject_key = "client_id"

    cloud = "azure"

    tenant_id = StringType(required=True)
    """Azure account ID.

    :type: str
    """

    subscription_id = StringType(required=True)
    """Azure subscription ID.

    :type: str
    """

    client_id = StringType(required=True, serialized_name="app_client_id")
    """Azure client ID.

    :type: str
    """

    secret_key = StringType(required=True, serialized_name="app_secret_key")
    """Azure password.

    :type: str
    """

    access_token = StringType()
    """Azure access token.

    :type: str
    """

    account_type = StringType(
        choices=("public", "germany", "china", "govcloud"), default="public", required=True
    )
    """Azure cloud environment.

    :type: str
    """

    offer_id = StringType()
    """Azure Offer ID

    :type: bool
    """

    detailed_billing_enabled = BooleanType(serialized_name="detailed_billing.enabled")
    """Azure Detailed Billing enable flag

    :type: bool
    """

    @property
    def cloud_account(self):
        """Returns Scalr cloud account"""
        return self.subscription_id

    def to_libcloud(self, location=None, api="compute", retry_delay=None, timeout=None):
        """Returns credentials for libcloud

        :param str location: Azure region
        :param str api: libcloud api type ('compute', 'storage', 'dns' etc..)

        :rtype: ``LibcloudCredentials``
        """
        cloud_environment = {
            "public": "default",
            "china": "AzureChinaCloud",
            "germany": "AzureGermanCloud",
            "govcloud": "AzureUSGovernment",
        }.get(self.account_type)
        creds = super(AzureCloudCredentials, self).to_libcloud(
            location=location, retry_delay=retry_delay, timeout=timeout
        )
        if api == "compute":
            creds.provider = compute_types.Provider.AZURE_ARM
        elif api == "storage":
            creds.provider = storage_types.Provider.AZURE_BLOBS
        else:
            raise exc.NotSupportedError("Libcloud API '{}' is not supported.".format(api))

        creds.args = (self.tenant_id, self.subscription_id, self.client_id, self.secret_key)
        # pylint: disable=no-member
        creds.kwargs.update({"cloud_environment": cloud_environment})
        return creds

    @classmethod
    def from_db(cls, data, **kwds):
        """Returns model instance from raw(database) data.

        :param list data: list of dicts with raw data from database
        :param dict kwds: additional keyword arguments for `schematics.Model` class
        :rtype: AzureCloudCredentials
        """
        credentials = super(AzureCloudCredentials, cls).from_db(data, **kwds)
        # pylint: disable=no-member
        if not credentials.client_id:
            credentials.client_id = config.azure["app_client_id"]
            credentials.secret_key = config.azure["app_secret_key"]
        return credentials


class OpenstackProvider(Enum):
    """Enum of openstack providers."""

    cisco = "cisco"
    hpcloud = "hpcloud"
    mirantis = "mirantis"
    nebula = "nebula"
    ocs = "ocs"
    otc = "otc"
    openstack = "openstack"
    vio = "vio"


class OpenstackCredentials(CloudCredentials):
    """Credentials from OpenStack clouds family (including HP cloud)."""

    _encrypted_fields = [
        "username",
        "password",
        "keystone_url",
        "api_key",
        "tenant_name",
        "domain_name",
    ]

    _unique_key_fields = ["keystone_url", "username", "tenant_name", "domain_name"]

    cloud = "openstack"

    auth_url_key = "keystone_url"

    username = StringType(required=True)
    """Username.

    :type: str
    """

    password = StringType()
    """Password

    :type: str
    """

    api_key = StringType()
    """API key

    :type: str
    """

    tenant_name = StringType()
    """Tenant / Project name

    :type: str
    """

    domain_name = StringType()
    """Custom cloud domain name

    :type: str
    """

    keystone_url = StringType(required=True)
    """Authentication | endpoint service (Keystone) API URL.

    :type: str
    """

    ssl_verify = BooleanType(default=True)
    """Verify SSL connection

    Default is True

    :type: bool
    """

    @property
    def cloud_account(self):
        """Returns Scalr cloud account"""
        return self.username

    @property
    def cloud_id(self):
        """ Returns unique id of the cloud """
        if self.keystone_url:
            return self._parse_url()

    @property
    def auth_url(self):
        """ URL to the authentication | endpoint service """
        return self.keystone_url

    def validate(self, *args, **kwargs):  # pylint: disable=arguments-differ
        """OpenstackCredentials specific validation"""
        super(OpenstackCredentials, self).validate(*args, **kwargs)
        if self.password is None and self.api_key is None:
            raise ValidationError("OpenstackCredentials password or api_key fields are required.")

    def to_libcloud(self, location=None, api="compute", retry_delay=None, timeout=None):
        """Returns credentials for libcloud

        :param str location: GCE datacenter
        :param str api: libcloud api type ('compute', 'storage', 'dns' etc..)

        :rtype: ``LibcloudCredentials``
        """
        creds = super(OpenstackCredentials, self).to_libcloud(
            location=location, retry_delay=retry_delay, timeout=timeout
        )
        if api == "compute":
            creds.provider = compute_types.Provider.OPENSTACK
        elif api == "storage":
            creds.provider = storage_types.Provider.OPENSTACK_SWIFT
        else:
            raise exc.NotSupportedError("Libcloud API '{}' is not supported.".format(api))

        creds.args = (self.username, self.password or self.api_key)
        kwargs = dict.fromkeys(
            [
                "ex_force_auth_url",
                "ex_force_base_url",
                "ex_force_auth_version",
                "ex_force_service_type",
                "ex_force_service_name",
                "ex_force_service_region",
                "ex_tenant_name",
                "ex_domain_name",
            ]
        )
        kwargs["ex_force_service_region"] = location
        keystone_url = str(self.keystone_url)
        if self.password:
            if "v3" in keystone_url:  # pylint: disable=unsupported-membership-test
                kwargs["ex_force_auth_version"] = "3.x_password"
            else:
                kwargs["ex_force_auth_version"] = "2.0_password"
        else:
            if "v2" not in keystone_url:  # pylint: disable=unsupported-membership-test
                raise ValueError("API Key Auth allowed only for Keystone v2")
            kwargs["ex_force_auth_version"] = "2.0_apikey"
        # pylint: disable=no-member
        if not keystone_url.rstrip("/").endswith("/tokens"):
            if kwargs["ex_force_auth_version"] == "3.x_password":
                keystone_url = os.path.join(keystone_url, "auth", "tokens")
            else:
                keystone_url = os.path.join(keystone_url, "tokens")
        if self.tenant_name:
            kwargs["ex_tenant_name"] = self.tenant_name
        if kwargs["ex_force_auth_version"] == "3.x_password" and self.domain_name:
            kwargs["ex_domain_name"] = self.domain_name

        url = six.moves.urllib.parse.urlparse(keystone_url)
        kwargs["ex_force_auth_url"] = url.geturl()
        kwargs["ex_force_base_url"] = "%s://%s" % (url.scheme, url.netloc)

        # pylint: disable=no-member
        creds.kwargs.update(kwargs)
        return creds

    def _parse_url(self):
        """Remove scheme from url. Also removing '/' character
        from the end if exist.
        """
        parsed_url = six.moves.urllib.parse.urlparse(self.keystone_url)
        scheme = "%s://" % parsed_url.scheme
        parsed_url = parsed_url.geturl().replace(scheme, "", 1)
        if parsed_url.endswith("/"):
            parsed_url = parsed_url[:-1]
        return parsed_url

    def get_nova_driver(self, location=None, retry_delay=None, timeout=None):
        """
        Returns connection driver to Compute (Nova) service. It uses
        `to_libcloud()` to get initial credentials to Keystone and then
        generates credentials for Nova and creates the driver.

        :returns libcloud.compute.drivers.openstack.OpenStackNodeDriver:
            driver with a connection to Nova.
        """
        libcloud_credentials = self.to_libcloud(
            location=location, retry_delay=retry_delay, timeout=timeout
        )
        keystone_driver = libcloudutil.create_driver(libcloud_credentials)
        service_catalog = keystone_driver.connection.get_service_catalog()

        nova = None
        for service in service_catalog.get_entries():
            if service.service_type.startswith("compute"):
                nova = service
                break
        else:
            raise exc.Error("Compute service not found. {}".format(self))

        nova_endpoint = None
        for endpoint in nova.endpoints:
            if endpoint.endpoint_type == "external" and (
                location is None or endpoint.region == location
            ):
                nova_endpoint = endpoint
                break
        else:
            raise exc.Error("External endpoint not found. {}".format(self))

        kwargs = {}
        kwargs["ex_force_service_type"] = nova.service_type
        kwargs["ex_force_service_name"] = nova.service_name
        kwargs["ex_force_service_region"] = nova_endpoint.region
        kwargs["ex_force_base_url"] = nova_endpoint.url

        # pylint: disable=no-member
        libcloud_credentials.kwargs.update(kwargs)

        return libcloudutil.create_driver(libcloud_credentials)


# Class for each OpenStack provider
class MirantisCredentials(OpenstackCredentials):
    """Mirantis Openstack-based credentials"""

    cloud = OpenstackProvider.mirantis


class HpCloudCredentials(OpenstackCredentials):
    """HP Cloud Openstack-based credentials"""

    cloud = OpenstackProvider.hpcloud


class NebulaCredentials(OpenstackCredentials):
    """Nebula Openstack-based credentials"""

    cloud = OpenstackProvider.nebula


class OcsCredentials(OpenstackCredentials):
    """OCS Openstack-based credentials"""

    cloud = OpenstackProvider.ocs


class OtcCredentials(OpenstackCredentials):
    """OTC Openstack-based credentials"""

    cloud = OpenstackProvider.otc


class VioCredentials(OpenstackCredentials):
    """Vio Openstack-based credentials"""

    cloud = OpenstackProvider.vio


class Clouds(str, enum.Enum):
    GCE = "gce"
    AWS = "ec2"
    AZURE = "azure"
    VMWARE = "vmware"


_cloud_to_cloud_credentials_factory = {
    "ec2": AwsCloudCredentials,
    "azure": AzureCloudCredentials,
    "gce": GceCloudCredentials,
    "vmware": VMwareCloudCredentials,
    # openstack family
    "openstack": OpenstackCredentials,
    "mirantis": MirantisCredentials,
    "hpcloud": HpCloudCredentials,
    "nebula": NebulaCredentials,
    "ocs": OcsCredentials,
    "otc": OtcCredentials,
    "vio": VioCredentials,
}


def get_cloud_credentials_factory(cloud):
    """Returns factory to instantiate CloudCredentials subclass for a given ``cloud``."""
    try:
        return _cloud_to_cloud_credentials_factory[cloud]
    except KeyError:
        raise NotImplementedError("No cloud credentials class for {!r} platform".format(cloud))


class Storage(Model):
    """
    A base for Volume and Snapshot classes.
    """

    id = StringType(default=None)
    """
    Unique storage id.

    This builtin override is required in order to this model irreplaceable
    with libcloud's ``StorageVolume`` model, which has an id attribute as well.

    :type: str
    """

    scalr_storage_id = StringType(default=None)
    """
    A Scalr managed storage id.

    :type: string
    """

    # FAM-897
    # type = StringType(required=True)
    type = StringType()
    """
    Storage type.

    : type: str
    """

    tags = DictType(StringType, default={})
    """
    Tags to apply on a cloud.

    Default is {}

    :type: dict
    """

    @property
    def volume_id(self):
        """
        Volume id.

        :rtype: str
        """
        return self.__dict__["id"]

    @property
    def volume_type(self):
        """
        Volume type (e.g. ebs, cinder, etc...).

        :rtype: str
        """
        return self.__dict__["type"]


class Snapshot(Storage):
    def __init__(self, raw_data=None, **kwargs):
        if raw_data is None:
            raw_data = {}
        if not raw_data.get("id"):
            raw_data["id"] = mixutil.generate_uid(prefix="snap-", length=8)
        super(Snapshot, self).__init__(raw_data=raw_data, **kwargs)

    volume_id = StringType()
    """The ID of the volume that was used to create the snapshot.

    :type: str
    """

    state = StringType()
    """The snapshot state.

    :type: str
    """

    created = DateTimeType()
    """A datetime object that represents when the snapshot was created.

    :type: :class:`datetime.datetime`
    """

    name = StringType()
    """A snapshot name generated by cloud.

    :type: str
    """

    @classmethod
    def from_libcloud_snapshot(cls, libcloud_snapshot, volume_id=None, **kwargs):
        """
        Creates :class:`Snapshot` from :class:`libcloud.common.base.VolumeSnapshot`.

        :type libcloud_snapshot: :class:`libcloud.common.base.VolumeSnapshot`
        :type volume_id: str

        :rtype: :class:`Snapshot`
        """
        return cls(
            dict(
                id=libcloud_snapshot.id,
                volume_id=volume_id,
                state=libcloud_snapshot.state,
                created=libcloud_snapshot.created,
                size=libcloud_snapshot.size,
                name=libcloud_snapshot.name or libcloud_snapshot.id,
                **kwargs,
            )
        )

    def __str__(self):
        return "VolumeSnapshot (id={id}, volume_id={volume_id})".format(
            id=self.id, volume_id=self.volume_id
        )


class EbsSnapshot(Snapshot):

    encrypted = BooleanType(default=False)
    """Specifies whether the volume should be encrypted.

    :type: str
    """

    kms_key_id = StringType()
    """The full ARN of the AWS Key Management Service (AWS KMS)
    customer master key (CMK) to use when creating the encrypted volume.
    Example: arn:aws:kms:us-east-1:012345678910:key/abcd1234-a123-456a-a12b-a123b4cd56ef.
    Only used if encrypted is set to True

    :type: str
    """

    @classmethod
    def from_libcloud_snapshot(cls, libcloud_snapshot, volume_id=None, **kwargs):
        """
        Creates :class:`EbsSnapshot` from :class:`libcloud.common.base.VolumeSnapshot`.

        :type libcloud_snapshot: :class:`libcloud.common.base.VolumeSnapshot`
        :type volume_id: str

        :rtype: :class:`EbsSnapshot`
        """
        kwargs["encrypted"] = libcloud_snapshot.extra["encrypted"]
        kwargs["kms_key_id"] = libcloud_snapshot.extra["kms_key_id"]
        kwargs["tags"] = libcloud_snapshot.extra["tags"]

        return super(EbsSnapshot, cls).from_libcloud_snapshot(
            libcloud_snapshot, volume_id=volume_id, **kwargs
        )


class Volume(Storage):
    """
    A base volume class.
    """

    type = StringType(
        choices=(
            "base",
            "cinder",
            "csvol",
            "ebs",
            "azure",
            "azure_vhd",
            "ec2_ephemeral",
            "gce_ephemeral",
            "gce_persistent",
            "aws_efs",
        ),
        required=True,
    )
    """Volume type (depends on cloud provider).

    :type: str
    """

    snap = ModelType(Snapshot)
    """Snapshot object, to be used for volume recreation.

    :type :class:`Snapshot`
    """

    device = StringType()
    """Attachment device name.

    :type: str
    """

    fstype = StringType()
    """Filesystem type name.

    :type: str
    """

    mpoint = StringType()
    """Mount point name.

    :type: str
    """

    label = StringType(default=None)
    """Optional volume label, windows only.

    :type: str
    """

    mount_options = ListType(StringType, default=[])
    """Mount options.

    :type: list(str)
    """

    recreate_if_missing = BooleanType(default=False)
    """Allow storage volume recreation, if this volume's id is missing.

    Default is False

    :type: bool
    """

    template = DictType(BaseType, default={})
    """A serializable configuration to be used for new
    volume creation when `recreate_if_missing` is `True`.

    :type: dict
    """

    def __str__(self):
        return "StorageVolume (id={id})".format(id=self.id)


class EbsVolume(Volume):

    snap = ModelType(EbsSnapshot)
    """Snapshot object, to be used for volume recreation.

    :type :class:`EbsSnapshot`
    """

    size = IntType(required=True)
    """Size of volume in gigabytes.

    :type: int
    """

    volume_type = StringType()
    """Type of volume to create. eg: gp2, gp3, io1, standard

    :type: str
    """

    iops = IntType()
    """The number of I/O operations per second (IOPS) that the volume supports.
    Only used if volume_type is io1 or gp3.

    :type: int
    """

    throughput = IntType()
    """The throughput to provision for a volume, with a maximum of 1,000 MiB/s.
    This parameter is valid only for gp3 volumes.
    Valid Range: Minimum value of 125. Maximum value of 1000.

    :type: int
    """

    encrypted = BooleanType(default=False)
    """Specifies whether the volume should be encrypted.

    :type: str
    """

    kms_key_id = StringType()
    """The full ARN of the AWS Key Management Service (AWS KMS)
    customer master key (CMK) to use when creating the encrypted volume.
    Example: arn:aws:kms:us-east-1:012345678910:key/abcd1234-a123-456a-a12b-a123b4cd56ef.
    Only used if encrypted is set to True

    :type: str
    """

    avail_zone = StringType()
    """Datacenter name in which to create a volume in.  (e.g. "eu-west-1b")

    :type: str
    """

    name = StringType()
    """EBS attachment device name (ex: /dev/xvdc)
    """


class AzureVolume(Volume):
    """
    Azure managed volumes class.
    """

    size = IntType(required=True)
    """Size of volume in gigabytes.

    :type: int
    """

    resource_group = StringType(required=True)
    """
    The name of the resource group.

    :type: str
    """

    account_type = StringType(default=None)
    """
    Azure storage account type (e.g. ``Standard_LRS``, ``Premium_LRS``, etc.).

    :type: str or None
    """

    zones = ListType(StringType)
    """
    Azure availability zone list.

    :type: list(str) or None
    """

    disk_encryption_set = StringType(default=None, required=False)
    """
    Disk encryption set for encryption data disk with customer-managed key.
    Used only for data disk.
    https://docs.microsoft.com/en-us/rest/api/compute/disks/createorupdate#encryption
    """

    @property
    def name(self):
        """
        Returns name of the volume.
        """
        if self.id is not None:
            return os.path.basename(self.id)


class AzureVHDVolume(Volume):
    """
    Azure VHD (unmanaged) volumes class.
    """

    size = IntType(required=True)
    """
    Size of volume in gigabytes.

    :type: int
    """

    managed_by = StringType()
    """
    The ID of the owner (instance) of the volume.

    :type: str
    """

    vhd_uri = StringType(required=True)
    """
    VHD blob URI.

    :type: str
    """

    resource_group = StringType(required=True)
    """
    The name of the resource group.

    :type: str
    """

    @property
    def name(self):
        """
        Returns name of the volume.
        """
        return self.id

    def __str__(self):
        return "StorageVolume (vhd_uri={vhd_uri})".format(vhd_uri=self.vhd_uri)


class EFSVolume(Volume):
    """
    Amazon Elastic File System volume class.
    """

    mount_target = StringType()
    """
    NFS remote share address.

    :type: str
    """


class VolumeUpdateParameters(Model):
    """ Parameters of a volume that need to be updated. """

    size = IntType()
    """
    Size of volume in gigabytes.
    """


class EbsVolumeUpdateParameters(VolumeUpdateParameters):
    dry_run = BooleanType(default=False)
    """
    Checks if you have the permission to do the operation without actually doing volume update.
    """

    iops = IntType()
    """
    The number of I/O operations per second (IOPS) that the volume supports.
    Only used if volume_type is io1 or gp3.
    """

    throughput = IntType()
    """
    The throughput to provision for a volume, with a maximum of 1,000 MiB/s.
    This parameter is valid only for gp3 volumes.
    Valid Range: Minimum value of 125. Maximum value of 1000.
    """

    volume_type = StringType()
    """
    Type of volume. eg: gp2, gp3, io1, standard
    """


class StorageGlobals(Model):
    """
    Storage workflows input data.
    """

    server_id = StringType(required=True)
    """
    Scalr managed server-id.
    """

    volume = ModelType(Volume, required=True)
    """
    A volume configuration to apply.
    """

    cloud_server = ModelType(CloudServer, required=False)
    """
    A cloud server to operate on.
    """

    credentials = ModelType(CloudCredentials, required=False)
    """
    Cloud credentials object.
    """


class StorageCreateGlobals(StorageGlobals):
    """
    storage.create workflow input data.
    """


class StorageUpdateGlobals(StorageGlobals):
    """
    storage.update workflow input data.
    """

    update_parameters = ModelType(VolumeUpdateParameters)
    """
    Parameters of volume to update.
    """


class StorageLocals(Model):
    """'storage.create' workflow locals."""

    create_snapshot = BooleanType(default=False)
    """If True new storage snapshot will be created

    Default is False

    :type: bool
    """

    create_volume = BooleanType(default=False)
    """If True new storage volume will be created

    Default is False

    :type: bool
    """

    create_filesystem = BooleanType(default=False)
    """If True a filesystem will be created for a given storage

    Default is False

    :type: bool
    """

    mount_filesystem = BooleanType(default=False)
    """If True a filesystem will be mounted for a given storage

    Default is False

    :type: bool
    """

    libcloud_credentials = ModelType(LibcloudCredentials, required=True)
    """Libcloud credentials

    :type: class: `LibcloudCredentials`
    """

    volume = ModelType(Volume, required=True)
    """A volume to process

    :type: class: `Volume`
    """

    attach_volume = BooleanType(default=False)
    """If True, then volume attach operation should be executed

    Default is False

    :type: bool
    """

    detach_volume = BooleanType(default=False)
    """If True, then volume detach operation should be executed

    Default is False

    :type: bool
    """


class StorageContext(Model):
    """
    Context for storage tasks.
    """

    volume = ModelType(Volume, required=True)
    """
    Volume object to process.

    :type :class:`Volume`
    """

    cloud_server = ModelType(CloudServer, required=True)
    """
    A cloud server to operate on.

    :type :class:`CloudServer`
    """

    driver = BaseType()
    """
    A libcloud driver instance.

    :type :class:`libcloud.compute.base.NodeDriver`
    """

    libcloud_credentials = ModelType(LibcloudCredentials, required=True)
    """
    A libcloud-adapted set of credentials for current account.

    :type :class:`LibcloudCredentials`
    """


# Update workflow data types


class Package(Model):
    """A Fatmouse agent package."""

    name = StringType(required=True)
    """A package name
    (as it known in a repository index)
    :type: str
    """

    version = StringType()
    """Package version.
    :type: str
    """

    deprecated_since_version = StringType()
    """Deprecation package version.
    :type: str
    """

    arch = StringType()
    """Architecture (x86_64, i386).
    :type: str
    """


class RepositoryType(Enum):
    """Enum of repository types."""

    deb = "deb"
    rpm = "rpm"
    suse = "suse"
    win = "win"
    docker = "docker"


class Repository(Model):
    """A Fatmouse agent repository."""

    repo_type = StringType(choices=RepositoryType.values(), required=True)
    """Repository type (rpm, deb, win, suse).
    :type: str
    """

    url = StringType(required=True)
    """Repository URL.
    For example: http://repo.scalr.net/rpm/stable/rhel/$releasever/$basearch/
    :type: str
    """

    packages = ListType(ModelType(Package), required=True)
    """Available packages list
    :type: list(:class:`Package`)
    """


class UpdateChannel(Model):
    """A collection of repositories for all supported operating systems."""

    name = StringType(required=True)
    """Channel name (stable, latest)
    :type: str
    """

    repos = ListType(ModelType(Repository), required=True)
    """Repositories list.
    :type: list(:class:`Repository`)
    """


class UpdateServer(Model):
    """UpdateServer information."""

    url = StringType(required=True)
    """API endpoint.
    :type: str
    """

    crypto_key = StringType(required=True)
    """An API crypto key base64encoded string.
    :type: str
    """

    scalr_id = StringType(required=True)
    """This Scalr installation ID.
    :type: str
    """

    scalr_version = StringType(required=True)
    """This Scalr version.
    :type: str
    """


class AgentPackage(Model):
    """An "agent.package" fact data type."""

    name = StringType(required=True)
    """Installed package name.
    :type: str
    """

    installed_version = StringType(required=True)
    """Installed package version.
    :type: str
    """

    service_version = StringType(required=True)
    """Source version in a package (from version file).
    :type: str
    """

    repo_type = StringType(required=True, choices=RepositoryType.values())
    """Agent repository type, package installed from.
    :type: str
    """

    downgrade_blacklist = ListType(DictType(StringType), required=True)
    """List of backward incompatible versions for the package.
    :type: list
    """


class ScriptLogs(Model):
    stdout = StringType()
    """STDOUT log string
    :type: str
    """

    stderr = StringType()
    """STDERR log string
    :type: str
    """


class FirewallRule(Model):
    comment = StringType(required=True)
    """
    A name of a rule to be set. Should be unique and not `all`
    On windows this should be unique as it will become a name of a rule.
    """

    from_port = IntType(required=True, min_value=0, max_value=65535)
    """Port range lower boundary, or single port value"""

    to_port = IntType(required=True, min_value=0, max_value=65535)
    """Port range upper boundary"""

    direction = StringType(choices=("in", "out"), default="in", required=True)
    """Dataflow direction to apply rule for"""

    source = StringType(required=True)
    """Dataflow source. And ip, hostname or range."""

    protocol = StringType(choices=("tcp", "udp", "all"), default="all", required=True)
    """Networking protocol"""

    action = StringType(choices=("allow", "deny"), required=True)
    """Rule action"""

    def _get_inner(self, name):
        """
        Converts field to its inner representation for comparison.
        """
        value = self.get(name)
        if value is None:
            return value
        elif name == "source":
            if self._is_plain_ip(value):
                value = "{}/32".format(value)
            elif value.endswith("/0"):
                value = "0.0.0.0/0"
        return value

    @staticmethod
    def _is_plain_ip(address):
        if address.count(".") != 3:
            return False
        return all(octet.isdigit() and 0 <= int(octet) <= 255 for octet in address.split("."))

    def __eq__(self, other):
        excluded_fields = ("comment",)
        # pylint: disable=protected-access, no-member
        return all(
            self._get_inner(field) == other._get_inner(field)
            for field in self._fields
            if field not in excluded_fields
        )

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.flatten())


class AgentHello(Model):
    new_password = StringType(required=True)
    """Permanent password generated by agent."""

    boot_id = StringType(required=True)
    """Boot id used to determinate server reboot."""

    server_id = StringType(required=True)
    """Scalr server-id."""


class SCSIAddress(Model):
    """Class for encapsulating disk SCSI address"""

    port = IntType(serialized_name="SCSIPort")
    bus = IntType(serialized_name="SCSIBus")
    target_id = IntType(serialized_name="SCSITargetId")
    logical_unit = IntType(serialized_name="SCSILogicalUnit")

    class Options:
        serialize_when_none = False

    def __init__(self, *args, **kwds):
        """
        :param Win32_DiskDrive disk_drive: disk drive to init address from
        """
        disk_drive = kwds.get("disk_drive")
        if disk_drive:
            kwds["raw_data"] = {
                "port": disk_drive.SCSIPort,
                "bus": disk_drive.SCSIBus,
                "target_id": disk_drive.SCSITargetId,
                "logical_unit": disk_drive.SCSILogicalUnit,
            }
            kwds.pop("disk_drive")
        super(SCSIAddress, self).__init__(*args, **kwds)

    def __str__(self):
        return "<SCSIAddress {}>".format(self.packed_address)

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash((self.port, self.bus, self.target_id, self.logical_unit))

    @property
    def packed_address(self) -> str:
        """
        Return packed SCSI address (e.g `3:0:0:1`).
        """
        format_dict = {"SCSIPort": "", "SCSIBus": "", "SCSITargetId": "", "SCSILogicalUnit": ""}
        format_dict.update(self.to_native())
        return "{SCSIPort}:{SCSIBus}:{SCSITargetId}:{SCSILogicalUnit}".format(**format_dict)

    @classmethod
    def from_packed_address(cls, packed_address: str) -> "SCSIAddress":
        """
        Initialize :class:`types.SCSIAddress` from SCSI address string.

        :param packed_address: packed SCSI address (e.g `3:0:0:1`).
        """
        address = cls()
        try:
            return address.import_data(
                {
                    key: int(value)
                    for key, value in zip(address.keys(), packed_address.split(":"))
                    if value != ""
                }
            )
        except:
            raise exc.Error("Invalid SCSI address string: {}".format(packed_address))

    def __eq__(self, other):
        if isinstance(other, SCSIAddress):
            return self.packed_address == other.packed_address
        if isinstance(other, str):
            return self.packed_address == other
        return False


class AgentStatEvent(Model):
    """Agent event with OS related statistics."""

    class Options:
        serialize_when_none = False

    timestamp = BaseType()
    """local sent time (see celery docs)
    """

    local_received = BaseType()
    """local received time (see celery docs)
    """

    facts = DictType(BaseType)
    """Dictionary of system facts.
    """

    farm_id = StringType(required=True)
    """Identifier of the Farm.
    """

    farm_role_id = StringType(required=True)
    """Identifier of the FarmRole.
    """

    server_id = StringType(required=True)
    """Identifier of the Server.
    """

    server_index = IntType(required=True)
    """Server index in the Farm.
    """

    env_id = StringType(required=True)
    """Identifier of the Environment.
    """

    cloud = StringType(required=True)
    """Name of the cloud where the instance  is launched (aws, openstack etc).
    """

    cloud_location = StringType(required=True)
    """Region when the instance is launched (us-east-1, RegionOne etc).
    """

    agent_version = StringType(required=True)
    """Installed Scalarizr agent version.
    """


class TupleType(ListType):
    def to_native(self, value, context=None):
        return tuple(super().to_native(value, context=context))

    # pylint: disable=arguments-differ
    def export_loop(self, *args, **kwds):
        return tuple(super().export_loop(*args, **kwds))


class SetType(ListType):
    def to_native(self, value, context=None):
        return set(super().to_native(value, context=context))
