"""
.. py:module:: common.types.flow
   :synopsis: Data types for a workflow engine

.. py:moduleauthor:: Marat Komarov <marat@scalr.com>
"""
from schematics.exceptions import ValidationError
from schematics.types import BaseType
from schematics.types import DateTimeType
from schematics.types import StringType
from schematics.types import UUIDType
from schematics.types.compound import ListType
from schematics.types.compound import ModelType

from common.types import Enum
from common.types import Error
from common.types import Model


class WorkflowErrorTrace(Model):
    """A class for task error traceback."""

    state = StringType(required=True)
    """A FSM state name (ex: 'create image')

    :type: str
    """

    event = StringType(required=True)
    """A FSM event name (ex: 'then')

    :type: str
    """


class TaskState(Enum):
    """Values for :attr:`Task.state`"""

    pending = "pending"
    running = "running"
    completed = "completed"
    failed = "failed"


class WorkflowState(TaskState):
    """Values for :attr:`Workflow.state`"""

    half_completed = "half-completed"


class BaseTask(Model):
    """This is a base class for `Task` and `Workflow`"""

    task_id = StringType(required=True)
    """A Celery task-id

    :type: str
    """

    name = StringType(required=True)
    """A celery task name

    :type: str
    """

    state = StringType(required=True, choices=TaskState.values())
    """A state value (from :class:`TaskState` attributes)

    :type: str
    """

    result = BaseType(required=True)
    """A result value or error

    :type: mixed result value
    :type: :class:`common.types.Error` when state is :attr:`TaskState.failed`
    """

    def validate_result(self, data, value):
        if data["state"] == TaskState.failed and not isinstance(value, Error):
            raise ValidationError("Error object is expected but got: {}".format(value))


class Task(BaseTask):
    """A task result."""

    date_start = DateTimeType()
    """UTC datetime of a start

    :type: str or None
    """

    date_end = DateTimeType()
    """UTC datetime of a finish

    :type: str or None
    """

    title = StringType()
    """A task fuman friendly title.

    It's actually a FSM state name (ex: 'create image') from a workflow definition.

    :type: str
    """

    server_id = UUIDType()
    """Scalr server-id with which task is associated.

    Tasks running on Scalarizr, always have this property equal to their Scalr server_id
    But there are some Fatmouse tasks that are only associated with a particular server.

    You should understand this property as a task context, not an actual execution target.

    :type: str
    :type: None when task was not associated with a server
    """

    # XXX: models are a server scope. move this method
    @classmethod
    def from_model(cls, model):
        """Creates object from Task database model

        :rtype: :class:`Task`
        """
        return Task(
            {
                "task_id": model.celery_task_id,
                "name": model.celery_task_name,
                "state": model.state,
                "result": model.result,
                "date_start": model.date_start,
                "date_end": model.date_end,
                "title": model.fsm_state,
                "server_id": model.server_id,
            }
        )


class Workflow(BaseTask):
    """A workflow result."""

    state = StringType(required=True, choices=WorkflowState.values())
    """A state value (from :class:`WorkflowState` attributes)

    :type: str
    """

    completed_tasks = ListType(ModelType(Task), default=[])
    """A list of completed tasks.

    (i.e with state :attr:`TaskState.completed`)

    :type: list(:class:`Task`)
    """

    failed_tasks = ListType(ModelType(Task), default=[])
    """A list of failed tasks

    (i.e. with state :attr:`TaskState.failed`)

    :type: list(:class:`Task`)
    """
