from schematics.types import BaseType
from schematics.types import BooleanType
from schematics.types import IntType
from schematics.types import StringType
from schematics.types.compound import DictType
from schematics.types.compound import ListType
from schematics.types.compound import ModelType

from common.types.base import Model
from common.types.data import CloudServer


class Ec2BlockDeviceMapping(Model):
    """Server block device mapping container."""

    IOPS_TYPES = (
        "io1",
        "gp3",
    )

    THROUGHPUT_TYPES = ("gp3",)

    delete_on_termination = BooleanType(default=True)
    device_name = StringType(required=True)
    iops = BaseType()
    throughput = BaseType()
    no_device = StringType()
    snapshot_id = StringType()
    virtual_name = StringType()
    volume_size = IntType()
    volume_type = StringType()

    def for_libcloud(self):
        """Returns block device mapping in dict format acceptable by Apache Libcloud"""
        if self.virtual_name:
            ebs_map = None
        else:
            ebs_map = {
                "DeleteOnTermination": "true" if self.delete_on_termination else "false",
                "Iops": self.iops if self.volume_type in self.IOPS_TYPES else None,
                "Throughput": (
                    self.throughput if self.volume_type in self.THROUGHPUT_TYPES else None
                ),
                "SnapshotId": self.snapshot_id,
                "VolumeSize": self.volume_size,
                "VolumeType": self.volume_type,
            }
            ebs_map = {k: v for k, v in ebs_map.items() if v is not None}
        bdm_dict = {
            "DeviceName": self.device_name,
            "Ebs": ebs_map,
            "NoDevice": "" if self.no_device else None,
            "VirtualName": self.virtual_name,
        }
        return {k: v for k, v in bdm_dict.items() if v is not None}


class ImageCreation(Model):
    """Image creation parameters container."""


class ImageCreationLocals(Model):
    """
    Image creation workflow parameters container.
    They are used to return data back to server from agent.
    """


class Ec2BundleCreation(ImageCreation):
    """Container for bundle creation parameters."""

    destination = StringType(default="/mnt")  # ec2-bundle-vol -d
    """Destination where snapshot of image will be placed.

    :type: str
    """

    ami_cloudfs_path = StringType(required=True)  # ec2-upload-bundle --bucket (but strip s3://)
    """Destination where to put image snapshot

    :type: str
    """

    excludes = ListType(StringType)


class Ec2ImageCreation(ImageCreation):
    """Container for image creation parameters."""

    root_device_mapping = DictType(StringType)
    """Custom root device mapping to use when making an image.

    :type: dict
    """

    no_reboot = BooleanType(default=False)
    """If True do not reboot server after image creation.

    Default is False

    :type: bool
    """

    name = StringType(required=True)

    description = StringType()

    sysprep = BooleanType(required=True, default=False)

    bundle_creation = ModelType(Ec2BundleCreation)


class Ec2BundleLocals(Ec2BundleCreation):
    """Container for instance store bundle parameters."""

    environ = DictType(StringType)

    cmd_path = StringType(required=True)

    size = IntType(required=True)

    manifest_path = StringType(required=True)  # ec2-upload-bundle -m
    """Destination where manifest will be created.

    :type: str
    """

    name = StringType(required=True)


class Ec2ImageCreationLocals(ImageCreationLocals):
    """
    Container for image creation parameters determined in real time.
    """

    block_device_mapping = ListType(ModelType(Ec2BlockDeviceMapping), required=True)
    """Block device mapping to be used when creating image.

    :type: :class:`Ec2BlockDeviceMapping`
    """

    libcloud_credentials = BaseType(required=True)
    """Libcloud credentials

    :type: :class:`LibcloudCredentials`
    """

    cloud_server = ModelType(CloudServer, required=True)
    """Cloud server info

    :type: :class:`CloudServer`
    """

    image_creation = ModelType(ImageCreation, required=True)
    """Image creation parameters

    :type: :class:`ImageCreation`
    """

    bundle = ModelType(Ec2BundleLocals)
    """Instance store parameters

    :type: :class:`Ec2BundleLocals`
    """
