"""
.. module:: augeas
   :platform: Linux
   :synopsis: Augeas utils

.. moduleauthor:: Scalr Inc. <http://scalr.com>


"""

import os

import augeas as _augeas

from common.utils import exc
from common.utils import facts
from common.utils import logutil

LOG = logutil.get_logger(__name__)

Augeas = _augeas.Augeas


class Error(exc.Error):
    pass


def _check_error(aug, path):
    """
    Function to check Augeas error

    :param :class:`Augeas` aug: Augeas object

    :param str path: <PATH> param in augtool command
    """

    context = aug.get("/augeas/context")
    filename = os.path.join(context, path)

    if filename.startswith("/files"):
        filename = filename[len("/files") :]
    while filename and filename != "/":
        if os.path.isfile(filename):
            break
        filename = os.path.dirname(filename)

    for error_path in aug.match("/augeas//error"):
        if filename in error_path:
            code = aug.get(error_path)
            message = aug.get(error_path + "/message")
            msg = "{0} {1}: {2}".format(filename, code, message)
            raise Error(msg)


def _insert(aug, *args):
    if args[1] == "before":
        ret = aug.insert(args[2], args[0], before=True)
    elif args[1] == "after":
        ret = aug.insert(args[2], args[0], before=False)
    else:
        msg = "Use 'insert <LABEL> <WHERE> <PATH>'. <WHERE> must be either 'before' or 'after'"
        raise Error(msg)
    _check_error(aug, args[2])
    return ret


def _transform(aug, *args):
    if args[1] == "incl":
        ret = aug.transform(args[0], args[2], excl=False)
    elif args[1] == "excl":
        ret = aug.transform(args[0], args[2], excl=True)
    else:
        msg = "Use 'transform <LENS> <FILTER> <FILE>'. <FILTER> must be either 'incl' or 'excl'"
        raise Error(msg)
    return ret


def augeas(context=None, root=None, autoload=True, lens=None, incl=None, loadpath=None):
    """
    :param str context: Sets /augeas/context.

    :param str root: Tree root path

    :param bool autoload: If set, Augeas will automatically load all lenses and
        parse all known file paths. Otherwise you'll need to specify
        lens' and modules by editing '/augeas/' subtree.

    :param str lens: Lens to load. Forces autoload=False

    :param list(str) incl: Files that will be added to <lens>/incl (will be parsed by it)

    :param str loadpath: is a colon-spearated list of directories that modules should be
        searched in. This is in addition to the standard load path and the
        directories in AUGEAS_LENS_LIB.
    """
    aug = Augeas(
        root=root,
        flags=Augeas.NONE if autoload and not lens else Augeas.NO_MODL_AUTOLOAD,
        loadpath=loadpath,
    )

    # TESTS-ONLY
    if facts.fact["container"] == "docker":
        execute(['set /augeas/save/copy_if_rename_fails "True"'], aug)

    if context:
        changes = ["set /augeas/context {}".format(context), "get /augeas/context"]
        cur_context = execute(changes, aug)[1]
        if context != cur_context:
            raise Error(
                'Failed to set context to "{0}", '
                'current context is "{1}"'.format(context, cur_context)
            )

    if lens:
        lens_name = os.path.splitext(lens)[0]
        changes = ["set /augeas/load/{0}/lens {0}.lns".format(lens_name)]
        for i, file_path in enumerate(incl or []):
            changes.append("set /augeas/load/{0}/incl[{1}] {2}".format(lens_name, i + 1, file_path))
        changes.append("load")
        execute(changes, aug)

    return aug


def edit(path, changes, lens=None):
    changes = list(changes)
    changes.append("save")
    return execute(changes, augeas(context="/files" + path, lens=lens, incl=[path]))


def execute(changes, aug=None, **kwds):
    """
    Helper function to make Augeas changes

    :param changes: List of changes
    :type changes: list

    :returns: output
    :rtype: list
    """
    if not aug:
        aug = augeas(**kwds)
    ret, paths_for_check = list(), list()
    for change in changes:
        splitted_change = change.strip().split()
        cmd = splitted_change[0]
        args = splitted_change[1:]
        try:
            if cmd == "defnode":
                ret.append(aug.defnode(*args))
                paths_for_check.append(args[1])
            elif cmd == "defvar":
                ret.append(aug.defvar(*args))
            elif cmd == "get":
                ret.append(aug.get(*args))
                paths_for_check.append(args[0])
            elif cmd in ["insert", "ins"]:
                ret.append(_insert(aug, *args))
            elif cmd == "label":
                ret.append(aug.label(*args))
                paths_for_check.append(args[0])
            elif cmd == "load":
                ret.append(aug.load(*args))
            elif cmd == "match":
                ret.append(aug.match(*args))
                paths_for_check.append(args[0])
            elif cmd in ["move", "mv"]:
                ret.append(aug.move(*args))
                paths_for_check.append(args[0])
                paths_for_check.append(args[1])
            elif cmd == "rename":
                ret.append(aug.rename(*args))
                paths_for_check.append(args[0])
            elif cmd == "rm":
                ret.append(aug.remove(*args))
                paths_for_check.append(args[0])
            elif cmd == "set":
                ret.append(aug.set(*args))
                paths_for_check.append(args[0])
            elif cmd == "setm":
                ret.append(aug.setm(*args))
            elif cmd == "span":
                ret.append(aug.span(*args))
                paths_for_check.append(args[0])
            elif cmd == "retrieve":
                ret.append(aug.text_retrieve(*args))
                paths_for_check.append(args[1])
                paths_for_check.append(args[2])
                paths_for_check.append(args[3])
            elif cmd == "store":
                ret.append(aug.text_store(*args))
                paths_for_check.append(args[1])
                paths_for_check.append(args[2])
            elif cmd == "transform":
                ret.append(_transform(aug, *args))
            elif cmd == "save":
                try:
                    aug.save(*args)
                except (
                    _augeas.AugeasIOError,
                    _augeas.AugeasRuntimeError,
                    _augeas.AugeasValueError,
                ):
                    for path in paths_for_check:
                        _check_error(aug, path)
                    # if error not found raise IOError
                    raise
                else:
                    num_files = len(aug.match("/augeas/events/saved"))
                    if num_files:
                        ret.append("Saved {} file(s)".format(num_files))
                    else:
                        ret.append(None)
            else:
                msg = "Unsupported Augeas command: {0}".format(cmd)
                raise Error(msg)
            for path in paths_for_check:
                _check_error(aug, path)
        except (TypeError, IndexError) as err:
            msg = "Failed to execute command: {0}. Wrong number of arguments".format(cmd)
            raise Error(msg) from err
    return ret
