"""Module that provides celery utils for server and agent"""
import glob
import os
import time
import uuid

from celery import schedules

from common.utils import exc
from common.utils import logutil
from common.utils import sysutil

LOG = logutil.get_logger(__name__)

try:
    import server.config as config
except ImportError:
    import agent.config as config


def agent_queue(server_id):
    """agent tasks queue name"""
    return "agent.{}".format(server_id)


def agent_results_queue(server_id):
    """agent tasks results queue name"""
    return "agent.{}.results".format(server_id)


def create_schedule_from_crontab(str_crontab):
    """Create celery beat schedule from GNU crontab string format.

    :type str_crontab: str
    :rtype: ``celery.schedules.crontab``
    """
    cron_options = ("minute", "hour", "day_of_month", "month_of_year", "day_of_week")
    task_cron_values = str_crontab.split()

    if len(task_cron_values) != len(cron_options):
        raise exc.MalformedError("Wrong crontab format: {!r}".format(str_crontab))

    return schedules.crontab(**dict(zip(cron_options, task_cron_values)))


class TaskLocalDir:
    """
    Class that implements celery task extension for agent and server.
    Created to separate celery internals from agent/server ones.

    Tasks data directory structure:

    ${config.workflow_engine['home_dir']}/tasks
    |__ ${task_id}/      - celery task ID
        |__ ${func_id}/
            |__ bin/     - executable directory for scripts
            |__ data/    - task own data files are here.
            |__ log/     - logs directory
            |__ meta/    - task meta data files: e.g: cloud credentials
    """

    DIRECTORY_MODE = 0o755

    def __init__(self):
        self.fn_seq = 0
        self.stack = []
        self.task_id = None

    @property
    def func_id(self):
        """Returns ID of the current function."""
        return self.stack[-1]

    @property
    def task_dir(self):
        """Returns this task directory."""
        return os.path.join(self._get_home_dir(), self.task_id)

    @property
    def meta_dir(self):
        """Returns meta directory."""
        return os.path.join(self.task_dir, "meta")

    @property
    def func_dir(self):
        """Returns function directory."""
        return os.path.join(self.task_dir, self.func_id)

    @property
    def bin_dir(self):
        """Returns bin directory."""
        return os.path.join(self.func_dir, "bin")

    @property
    def data_dir(self):
        """Returns data directory."""
        return os.path.join(self.func_dir, "data")

    @property
    def log_dir(self):
        """Returns log directory."""
        return os.path.join(self.func_dir, "log")

    @classmethod
    def _get_home_dir(cls):
        """Returns home directory for all tasks data."""
        return os.path.join(config.workflow_engine["home_dir"], "tasks")

    @classmethod
    def generate_task_id(cls):
        return str(uuid.uuid4()).replace("-", "")

    def task_pre_run(self, task_id):
        """Runs before task call."""
        LOG.debug("task_pre_run task_id=%s", task_id)
        self.task_id = task_id
        self.create_task_dirs()

    def task_post_run(self):
        """Runs after task call."""
        LOG.debug("task_post_run task_id=%s", self.task_id)
        # bad idea to move directories. What if i'm tailing logs and they disapper?
        # self.remove_task_dirs()

    def func_pre_run(self, func_name):
        """Runs before task call."""
        LOG.debug("func_pre_run func_name=%s", func_name)
        self.fn_seq += 1
        func_id = "{:03d}-{}".format(self.fn_seq, func_name)
        self.stack.append(func_id)
        self.create_func_dirs()

    def func_post_run(self):
        """Runs after task call."""
        LOG.debug("func_post_run stack=%s", self.stack)
        if not config.workflow_engine["devbox"]["keep_task_data"]:
            sysutil.rm_rf(self.bin_dir)
            sysutil.rm_rf(self.data_dir)
            # NOTE: self.log_dir contents remain for reason.
        self.stack.pop()

    def create_task_dirs(self):
        """Creates task directories."""
        # FAM-281
        # sysutil.rm_rf(self.task_dir)  # cleanup previous (possible) run attempt
        sysutil.mkdir_p(self.task_dir, self.DIRECTORY_MODE)
        sysutil.mkdir_p(self.meta_dir, self.DIRECTORY_MODE)

    def create_func_dirs(self):
        """Creates func directories."""
        sysutil.mkdir_p(self.func_dir, self.DIRECTORY_MODE)
        sysutil.mkdir_p(self.bin_dir, self.DIRECTORY_MODE)
        sysutil.mkdir_p(self.data_dir, self.DIRECTORY_MODE)
        sysutil.mkdir_p(self.log_dir, self.DIRECTORY_MODE)

    def remove_task_dirs(self):
        """Removes task directory."""
        sysutil.rm_rf(self.task_dir)

    def set_current_func(self, task_name):
        """Sets current function."""
        task_dirs = glob.glob(os.path.join(self.task_dir, "[0-9]*-{}*".format(task_name)))
        if task_dirs:
            func_id = sorted([os.path.split(i)[1] for i in task_dirs], reverse=True).pop()
            self.stack.append(func_id)
            return True

    @classmethod
    def gc(cls):
        """Drop state tasks directories."""
        home_dir = cls._get_home_dir()
        if not os.path.exists(home_dir):
            return

        now = time.time()
        num_removed = 0
        for name in os.listdir(home_dir):
            path = os.path.join(home_dir, name)
            if now - os.stat(path).st_ctime >= config.workflow_engine["gc_keep_time"]:
                sysutil.rm_rf(path)
                num_removed += 1
        if num_removed:
            LOG.debug("Drop %s stale directories from %s", num_removed, home_dir)
