import copy
import functools
import json
import logging
import os
import re
import ssl
import urllib.parse

import requests.certs
import yaml
from six.moves import configparser

from common.utils import exc
from common.utils import mixutil
from common.utils import six
from common.utils import sysutil

LOG = logging.getLogger(__name__)


def _reduce_dict(obj, key):
    """Return value of the composite key.
    >>> d = {'a': {'b': 1}}
    >>> assert _reduce_dict(d, 'a.b') == 1

    :param obj: dict with embedded keys.
    :param key: composite key: key1.key2.key3 = obj.get('key1', {}).get('key2', {}).get('key3', {})
    :return: value of the last dict key;
    """

    def dict_get(d, key):
        return d.get(key, {})

    return functools.reduce(dict_get, key.split("."), obj)


def load_config_from_file(cfg_file, key=None):
    """
    :param cfg_file: Path to YML or JSON config file
    :type  cfg_file: str

    :param key: Specifies subsection from cfg_file to load.
        key can be composite, for example 'section1.section2'
    :type key: str
    """
    config = {}
    if os.path.exists(cfg_file) and os.stat(cfg_file).st_size:
        LOG.debug('Loading configuration from file "%s"', cfg_file)
        with open(cfg_file) as fp:
            file_ext = os.path.splitext(cfg_file)[-1]
            try:
                unserialize = {".json": json.load, ".yml": yaml.safe_load}[file_ext]
            except KeyError:
                raise exc.MalformedError(
                    (
                        "Don't know how to read configuration from {}. "
                        "Supported file extensions are .yml and .json"
                    ).format(cfg_file)
                )
            config = unserialize(fp)
            if not isinstance(config, dict):
                raise exc.MalformedError(
                    "Configuration file {} has type {}. Expected: mapping".format(
                        cfg_file, type(config)
                    )
                )
            if key is not None:
                config = _reduce_dict(config, key)
    return config


def load_config_from_env(cfg_env, key=None):
    """
    :param cfg_env: Environment variable name. Value should contain valid JSON.
    :type  cfg_env: str

    :param key: Specifies subsection from cfg_env to load.
        key can be composite, for example 'section1.section2'
    :type key: str
    """
    config = {}
    if cfg_env in os.environ:
        LOG.debug("Loading configuration from ENV $%s", cfg_env)
        try:
            config = json.loads(os.environ[cfg_env])
            assert isinstance(config, dict)
        except ValueError:
            raise exc.MalformedError(
                "Env variable ${} exists, but is not valid json. Value: {}".format(
                    cfg_env, os.environ.get(cfg_env)
                )
            )
        except AssertionError:
            raise exc.MalformedError(
                "Configuration from env variable {} has type {}. Expected: mapping".format(
                    cfg_env, type(config)
                )
            )
        if key is not None:
            config = _reduce_dict(config, key)
    return config


def _is_primitive(var):
    # pylint: disable=no-member
    return isinstance(var, (dict, list, tuple, six.integer_types, float, bool, six.string_types))


def merge_dict(cfg_object, cfg_dict, keys=None):
    """Update config object using values from dict.
    :param cfg_object: Configuration object (usually module), attributes are options.
    :type  cfg_object: object
    :param cfg_dict: dict object with values to merge.
    :type  cfg_dict: dict
    :param keys: list of top level keys to merge
    :type  keys: list of str
    """
    updated = False

    def sanitize(data):
        if not isinstance(data, dict):
            return data

        retval = copy.deepcopy(data)
        mask = "******"

        if retval.get("workflow_engine", {}).get("rabbitmq", {}).get("ssl_cacert"):
            retval["workflow_engine"]["rabbitmq"]["ssl_cacert"] = mask
        if retval.get("workflow_engine", {}).get("celery", {}).get("broker_url"):
            url = retval["workflow_engine"]["celery"]["broker_url"]
            r = list(urllib.parse.urlsplit(url))
            r[1] = re.sub(r":([^@:]+)@", ":" + mask + "@", r[1])
            retval["workflow_engine"]["celery"]["broker_url"] = urllib.parse.urlunsplit(r)

        return retval

    # Select only public dict attributes of cfg_object to make copy.deepcopy possible
    cfg_object_dict = {
        k: v for k, v in cfg_object.__dict__.items() if not k.startswith("__") and _is_primitive(v)
    }
    origin = copy.deepcopy(cfg_object_dict)

    if keys:
        for key in keys:
            cfg_object_dict_reduced = _reduce_dict(cfg_object_dict, key)
            cfg_dict_reduced = _reduce_dict(cfg_dict, key)
            # handle primitive values in config
            if not isinstance(cfg_object_dict_reduced, dict) or not isinstance(
                cfg_dict_reduced, dict
            ):
                if cfg_object_dict_reduced != cfg_dict_reduced:
                    LOG.debug(
                        "Overriding config object %s. Key: %s. Update: %s",
                        cfg_object,
                        key,
                        sanitize(cfg_dict.get(key)),
                    )
                    setattr(cfg_object, key, cfg_dict.get(key))
                    updated = True
            else:
                origin_reduced = _reduce_dict(origin, key)
                if cfg_dict_reduced:
                    LOG.debug(
                        "Overriding config object %s. Key: %s. Update: %s",
                        cfg_object,
                        key,
                        sanitize(cfg_dict_reduced),
                    )
                merged = mixutil.merge_recursive(cfg_object_dict_reduced, cfg_dict_reduced)
                if merged != origin_reduced:
                    cfg_object_dict_reduced.update(merged)
                    updated = True
    else:
        LOG.debug("Overriding config object %s. Update: %s", cfg_object, sanitize(cfg_dict))
        merged = mixutil.merge_recursive(cfg_object_dict, cfg_dict)
        if merged != origin:
            cfg_object.__dict__.update(merged)
            updated = True

    return updated


def configure_ssl(cfg_object):
    """
    Configures SSL to the broker. It reads `cfg_object.workflow_engine['rabbitmq']['ssl_cacert']`
    which can be a string containing raw SSL certificate or a file path to the certificate file.

    :param object cfg_object: config object with workflow_engine field.
    """
    if cfg_object.workflow_engine["devbox"]["enabled"]:
        return

    if not cfg_object.workflow_engine["rabbitmq"]["ssl_verify"]:
        cfg_object.workflow_engine["celery"]["broker_use_ssl"] = True
        LOG.debug("SSL is configured with disabled certificate validation.")
        return

    ssl_cacert = cfg_object.workflow_engine["rabbitmq"]["ssl_cacert"]

    if not ssl_cacert:
        # it is necessary to update certifi from time to time if we're using this option a lot
        # http://docs.python-requests.org/en/master/user/advanced/#ca-certificates
        LOG.debug("Using python-requests' CA-Certificates bundle for Rabbitmq SSL connection.")
        ssl_cacert_path = requests.certs.where()

    elif "\n" in ssl_cacert:
        # If there is \n in ssl_cacert string, assuming it's a raw certificate
        # Saving raw CA certificate to location described in ssl_cacert_store_path.
        ssl_cacert_path = cfg_object.workflow_engine["rabbitmq"]["ssl_cacert_store_path"]
        sysutil.mkdir_p(os.path.dirname(ssl_cacert_path))
        with open(ssl_cacert_path, "w") as fp:
            fp.write(ssl_cacert)

    elif os.path.exists(ssl_cacert):
        ssl_cacert_path = ssl_cacert

    else:
        # If ssl_cacert doesn't look like raw certificate,
        # assuming it's a path to a file that doesn't exist
        raise exc.Error('Certificate file "{}" does not exist.'.format(ssl_cacert))

    cfg_object.workflow_engine["celery"]["broker_use_ssl"] = {
        "ca_certs": ssl_cacert_path,
        "cert_reqs": ssl.CERT_REQUIRED,
    }
    LOG.debug("SSL is configured.")


def merge(
    cfg_object,
    cfg_file=None,
    cfg_env=None,
    keys=None,
    cfg_file_root_key=None,
    cfg_env_root_key=None,
):
    """Update/create config file using values from config file and environment variable.

    Returns boolean value, which indicates if configuration file was updated.

    :param cfg_object: Configuration object (usually module), attributes are options.
    :type  cfg_object: object

    :param cfg_file: Path to YML or JSON config file
    :type  cfg_file: str

    :param cfg_env: Environment variable name. Value should contain valid JSON.
    :type  cfg_env: str

    :param keys: Specifies subsections that contains configuration to merge.
        keys is applied for both cfg_object and cfg_file/cfg_env.
    :type   key: list of str

    :param cfg_file_root_key: Specifies root subsections of cfg_file to merge.
        cfg_file_root_key can be composite, for example 'section1.section2'
    :type cfg_file_root_key: str

    :param cfg_env_root_key: Specifies root subsections of cfg_env to merge.
        cfg_env_root_key can be composite, for example 'section1.section2'
    :type cfg_env_root_key: str

    Example:
    cfg_object as dict
        {
            '1': 'old one',
            '2': {'2.1': 2.1},
        }
    cfg_file as dict
        {
            'root1': {
                'root2': {
                    '1': 'new one',
                    '2': {'2.2': 2.2},
                },
            },
        }
    after merge(cfg_object, cfg_file, keys=['2'], cfg_file_root_key='root1.root2')
    cfg_object will be
        {
            '1': 'old one',
            '2': {'2.1': 2.1, '2.2': 2.2},
        }
    """
    config = load_config_from_file(cfg_file, key=cfg_file_root_key)
    config_env = load_config_from_env(cfg_env, key=cfg_env_root_key)
    mixutil.merge_recursive(config, config_env)
    return merge_dict(cfg_object, config, keys=keys)


class MultiKeyConfigParser(configparser.ConfigParser):
    """
    Extension for ConfigParser that gives possibility
    to add multiple values for same option
    """

    def _read(self, *args, **kwargs):
        # pylint: disable=arguments-differ,abstract-method
        raise NotImplementedError("Currently read() is not supported")

    def add_option_value(self, section, option, value=None):
        """Add an option."""
        # pylint: disable=no-member
        if not section or section == configparser.DEFAULTSECT:
            sectdict = self._defaults
        else:
            try:
                sectdict = self._sections[section]
            except KeyError:
                raise configparser.ConfigParser.NoSectionError(section)
        if sectdict.get(option):
            sectdict[option].append(value)
        else:
            sectdict[option] = [value]

    def write(self, fp):  # pylint: disable=arguments-differ
        """Write an .ini-format representation of the configuration state."""
        if self._defaults:
            fp.write("[%s]\n" % configparser.DEFAULTSECT)
            for (key, value) in self._defaults.items():
                fp.write("%s=%s\n" % (key, str(value).replace("\n", "\n\t")))
            fp.write("\n")
        for section in self._sections:
            fp.write("[%s]\n" % section)
            for (key, value) in self._sections[section].items():
                if key == "__name__":
                    continue
                for entry in value:
                    if (entry is not None) or (self._optcre == self.OPTCRE):
                        key_entry = "=".join((key, str(entry).replace("\n", "\\\n\t")))
                        fp.write("%s\n" % (key_entry))
                    else:
                        fp.write("%s\n" % (key))
            fp.write("\n")

    def remove_option_value(self, section, option, value):
        """Remove an option value."""
        if not section or section == configparser.DEFAULTSECT:
            sectdict = self._defaults
        else:
            try:
                sectdict = self._sections[section]
            except KeyError:
                raise configparser.NoSectionError(section)
        if option in sectdict and value in sectdict[option]:
            sectdict[option].remove(value)
            return value


class UnitFile(MultiKeyConfigParser):  # pylint: disable=abstract-method
    # Method '_read' is abstract in class 'MultiKeyConfigParser' but is not overridden (abstract-method)
    """
    Class gives possibility generate and write systemd unit file
    """
    unit_section = "Unit"
    service_section = "Service"
    install_section = "Install"

    def __init__(self):
        """Create default section for unit object"""
        # pylint: disable=no-member
        MultiKeyConfigParser.__init__(self)  # old-style class
        self.add_section(self.unit_section)
        self.add_section(self.service_section)
        self.add_section(self.install_section)

    def set_description(self, value):
        """
        Set description
        :param value: string value
        """
        self.add_option_value(self.unit_section, "Description", value)

    def set_documentation(self, value):
        """
        Add documentation url
        :param value: string value
        """
        self.add_option_value(self.unit_section, "Documentation", value)

    def set_env_variable(self, var_name, var_value):
        """
        Set environment variable
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#Command%20lines
        :param var_name: name of variable
        :param var_value: value of variable
        """
        value = "=".join((str(var_name), '"' + str(var_value) + '"'))
        self.add_option_value(self.service_section, "Environment", value)

    def set_success_exit_status(self, codes_list):
        """
        Set list of successful exit status
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#SuccessExitStatus=
        :param codes_list:
        """
        value = " ".join([str(c) for c in codes_list])
        self.add_option_value(self.service_section, "SuccessExitStatus", value)

    def set_type(self, value):
        """
        Define a type of service unit
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#Type=
        :param value: type
        """
        self.add_option_value(self.service_section, "Type", value)

    def set_working_directory(self, value):
        """
        Sets the working directory for executed processes.
        url: https://www.freedesktop.org/software/systemd/man/systemd.exec.html#WorkingDirectory=
        :param value: path to directory
        """
        self.add_option_value(self.service_section, "WorkingDirectory", value)

    def set_remain_after_exit(self, value):
        """
        Specifies whether the service shall be considered active
        even when all its processes exited
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#RemainAfterExit=
        :param value: str: true/false
        """
        self.add_option_value(self.service_section, "RemainAfterExit", value)

    def set_stdout(self, value):
        """
        Control stdout of process
        url: https://www.freedesktop.org/software/systemd/man/systemd.exec.html#StandardOutput=
        """
        self.add_option_value(self.service_section, "StandardOutput", value)

    def set_stderr(self, value):
        """
        Control stderr of process
        url: https://www.freedesktop.org/software/systemd/man/systemd.exec.html#StandardError=
        """
        self.add_option_value(self.service_section, "StandardError", value)

    def set_start_timeout(self, timeout):
        """
        Configures the time to wait for start-up.
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#TimeoutStartSec=
        :param timeout: str: in sec
        """
        self.add_option_value(self.service_section, "TimeoutStartSec", timeout)

    def set_user(self, value):
        """
        Set user that the processes are executed as.
        url: https://www.freedesktop.org/software/systemd/man/systemd.exec.html#User=
        :param value: user name
        """
        self.add_option_value(self.service_section, "User", value)

    def set_start_exec(self, value):
        """
        Commands that are executed when this service is started.
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#ExecStart=
        :param value: command
        """
        self.add_option_value(self.service_section, "ExecStart", value)

    def set_stop_exec(self, value):
        """
        Commands to execute to stop the service started via ExecStart=.
        url: https://www.freedesktop.org/software/systemd/man/systemd.service.html#ExecStop=
        :param value: command
        """
        self.add_option_value(self.service_section, "ExecStop", value)
