"""
Created on Apr 7, 2010

@author: marat
@author: spike
"""
import base64
import binascii
import hashlib
import hmac
import os
import time

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import modes

from common.utils import exc
from common.utils import six

DATE_FMT = "%a %d %b %Y %H:%M:%S UTC"

"""
RFC 3339 date format (UTC) for the messaging services.

python: `time.strftime(RFC_3339_DATE_FMT, time.gmtime())`
php: `gmdate("c", time())`
"""
RFC_3339_UTC_DATE_FMT = "%Y-%m-%dT%H:%M:%S+00:00"  # type: str


def keygen(length=40):
    """Generate random bytes of a given `length` and encode in BASE64."""
    return binascii.b2a_base64(os.urandom(length))


def pwgen(length=10):
    """Generate password of a given `length`."""
    result = binascii.b2a_base64(os.urandom(int(length + length * 0.2))).decode()
    return result.replace("/", "").replace("+", "")[:length]


def pkcs5_pad(data, block_size):
    """Custom PKCS5 padding for compatibility with Scalr AES implementation."""
    pad = block_size // 8 - (len(data) % (block_size // 8))
    return data + six.b(chr(pad) * pad)


def pkcs5_unpad(data, block_size):
    """Custom PKCS5 unpadding for compatibility with Scalr AES implementation"""
    offset = ord(data[-1]) if six.PY2 else data[-1]
    return data[0:-offset]


def generate_openssh_public_key(private_key):
    """Generate public OpenSSH key from PEM private key

    :param private_key:
    :type private_key: bytes
    :return: OpenSSH public key
    :rtype: bytes
    """
    key = serialization.load_pem_private_key(private_key, None, default_backend())
    return key.public_key().public_bytes(
        serialization.Encoding.OpenSSH, serialization.PublicFormat.OpenSSH
    )


def encrypt_scalr(data, crypto_key):
    """Encrypt `data` with AES with custom PKCS5 padding and encode in BASE64.

    :type crypto_key: bytes
    :type data: bytes
    :rtype: bytes
    """
    if not crypto_key:
        raise ValueError("Crypto key is empty")
    data = pkcs5_pad(data, algorithms.AES.block_size)
    iv = crypto_key[-algorithms.AES.block_size // 8 :]
    cipher = Cipher(
        algorithms.AES(crypto_key[0:32]), mode=modes.CFB8(iv), backend=default_backend()
    )
    encryptor = cipher.encryptor()
    cipher_bytes = encryptor.update(data) + encryptor.finalize()
    return binascii.b2a_base64(cipher_bytes).strip()


def decrypt_scalr(data, crypto_key):
    """Decrypt string encrypted with `encrypt_scalr()`.

    :type crypto_key: bytes
    :type data: bytes
    :rtype: bytes
    """
    if not crypto_key:
        raise ValueError("Crypto key is empty")
    iv = crypto_key[-algorithms.AES.block_size // 8 :]
    cipher = Cipher(
        algorithms.AES(crypto_key[0:32]), mode=modes.CFB8(iv), backend=default_backend()
    )
    decryptor = cipher.decryptor()
    cipher_bytes = binascii.a2b_base64(data)
    decrypted = decryptor.update(cipher_bytes) + decryptor.finalize()
    return pkcs5_unpad(decrypted, algorithms.AES.block_size)


def encrypt_scalarizr(data, crypto_key):
    """Encrypt `data` with TripleDES with PKCS7 padding and encode in BASE64.

    :type crypto_key: bytes
    :type data: bytes
    :rtype: bytes
    """
    iv = crypto_key[-algorithms.TripleDES.block_size // 8 :]
    cipher = Cipher(
        algorithms.TripleDES(crypto_key[0:24]), modes.CBC(iv), backend=default_backend()
    )
    encryptor = cipher.encryptor()
    padder = padding.PKCS7(64).padder()
    padded = padder.update(data) + padder.finalize()
    cipher_bytes = encryptor.update(padded) + encryptor.finalize()
    return binascii.b2a_base64(cipher_bytes).strip()


def decrypt_scalarizr(data, crypto_key):
    """Decrypt string encrypted with `encrypt_scalarizr()`.

    :type crypto_key: bytes
    :type data: str
    :rtype: str
    """
    iv = crypto_key[-algorithms.TripleDES.block_size // 8 :]
    cipher = Cipher(
        algorithms.TripleDES(crypto_key[0:24]), modes.CBC(iv), backend=default_backend()
    )
    decryptor = cipher.decryptor()
    unpadder = padding.PKCS7(64).unpadder()
    cipher_bytes = binascii.a2b_base64(data)
    decrypted = decryptor.update(cipher_bytes) + decryptor.finalize()
    return (unpadder.update(decrypted) + unpadder.finalize()).decode()


def _construct_canonical_headers(headers):
    """
    Constructs canonical headers string from the dict of headers.

    The following pseudocode describes how to construct the canonical list of headers::
        CanonicalHeaders =  HeadersEntry0 + HeadersEntry1 + ... + HeadersEntryN
        HeadersEntry = Lowercase(HeaderName) + ':' + Trimall(HeaderValue) + '\n'

    `Lowercase` represents a function that converts all characters to lowercase.
    The `Trimall` function removes excess white space before and after values,
    and converts sequential spaces to a single space.

    The headers must be sorted by character code.

    See: https://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html

    :type headers: dict
    :rtype: str
    """
    formatted_headers = (
        (
            name.strip().replace(" ", "").lower(),
            " ".join(str(value).split()) if value is not None else "",
        )
        for name, value in headers.items()
    )
    formatted_headers = sorted(formatted_headers, key=lambda item: item[0])

    return "".join("{0}:{1}\n".format(name, value) for name, value in formatted_headers)


def _construct_canonical_string(params):
    """
    Constructs the string from the dict keys and values.

    :type params: dict
    :rtype: str
    """
    if not params:
        return ""
    return "".join((str(key) + str(value) for key, value in sorted(params.items())))


def sign_http_request(
    payload,
    crypto_key,
    headers=None,
    timestamp=None,
    version=1,
    digestmod=hashlib.sha1,
    date_fmt=DATE_FMT,
):
    """
    Calculates Scalr signature header.

    Structure of string to sign::
        StringToSign =
            CanonicalHeaders + \\n + (optional)
            RequestPayload +
            RequestTimestamp

    :param payload: The request payload to be signed.
    :type payload: str | dict
    :param crypto_key: The cryptographic key.
    :type crypto_key: bytes
    :param headers: HTTP/HTTPS headers to include in the signature.
    :param headers: dict | None
    :param timestamp: :class:`time.struct_time` in UTC (returned by `time.gmtime()`).
    :type timestamp: :class:`time.struct_time`
    :param version: Signature version:
        - 1. `Base64(HashAlgo(StringToSign))`
        - 2. `HexEncode(HashAlgo(StringToSign))`
    :type version: int
    :param digestmod: The cryptographic hash function (default ``hashlib.sha1``).
    :type digestmod: callable
    :param date_fmt: The datetime format (by default `cryptoutil.DATE_FMT` const).
    :type date_fmt: str

    :return: Signature and Date.
    :rtype: (str, str)
    """
    string_to_sign = b""

    # append canonical headers (optional)
    if headers:
        string_to_sign += "{}\n".format(_construct_canonical_headers(headers)).encode()

    # append request payload
    if isinstance(payload, dict):
        payload = _construct_canonical_string(payload)
    string_to_sign += payload.encode() if isinstance(payload, str) else payload

    # append request timestamp
    sign_date = time.strftime(date_fmt, timestamp or time.gmtime())
    string_to_sign += sign_date.encode()

    if version == 1:
        digest = hmac.new(crypto_key, string_to_sign, digestmod).digest()
        sign = binascii.b2a_base64(digest)
        if sign.endswith(b"\n"):
            sign = sign[:-1]
    elif version == 2:
        sign = hmac.new(crypto_key, string_to_sign, digestmod).hexdigest().encode()
    else:
        raise exc.Error(
            ("Unsupported `version` argument for sign_http_request(): {}").format(version)
        )
    return sign.decode(), sign_date


def base64url_decode(msg):
    """Decodes JWT string

    :type msg: str
    :rtype: str
    """
    msg = msg.encode()
    if len(msg) % 4:
        msg += b"=" * (4 - len(msg) % 4)
    return base64.urlsafe_b64decode(msg).decode()


def base64url_encode(msg):
    """Encodes string for JWT as base64 encoded and removes alignment '=' symbols

    :type msg: str
    :rtype: str
    """
    return base64.urlsafe_b64encode(msg.encode()).rstrip(b"=").decode()


def sign_jwt_content(content, secret):
    """Creates JWT sign using HS256 algorithm

    :type content: str
    :type secret: str
    :rtype: str
    """
    content, secret = content.encode(), secret.encode()
    sign = hmac.new(secret, content, digestmod=hashlib.sha256).digest()
    return base64.urlsafe_b64encode(sign).rstrip(b"=").decode()


def sha256(msg):
    """Calculates SHA256 digest from `msg` str

    :param str msg:
    :rtype: str
    """
    return hashlib.sha256(msg.encode()).hexdigest()
