"""
This modules provides helpers for date & time manipulations
"""
import datetime

from . import exc

RFC3339 = "%Y-%m-%dT%H:%M:%S%z"
RFC3339_WITH_MS = "%Y-%m-%dT%H:%M:%S.%f%z"

MINUTE = 60
HOUR = 3600
DAY = 86400
WEEK = 3600


def get_utc_now_formatted(dt_format=RFC3339):
    """
    Returns current UTC datetime string, formatted according to dt_format template

    :param dt_format: datetime format string
    :type dt_format: str
    :return: formatted ``datetime.datetime.utcnow()``
    :rtype: str
    """
    return datetime.datetime.utcnow().replace(microsecond=0).strftime(dt_format)


def rfc3339_timestamp_format(timestamp):
    """
    Converts timestamp to utc datetime string and formats it according to RFC3339 format

    :param timestamp:
    :type timestamp: float
    :return: datetime string representation of ``timestamp`` according to `RFC3339`
    :rtype: str
    """
    return datetime.datetime.utcfromtimestamp(timestamp).replace(microsecond=0).strftime(RFC3339)


def from_rfc3339(timestring):
    """
    Converts string in format like '2017-10-16T01:37:52.630-07:00'. Python's
    datetime doesn't have an option to split timezone part to hours and minutes,
    so it is done manually.

    :param str timestring: string in format '%Y-%m-%dT%H:%M:%S.%f<timezone_hour>:<timezone_minute>'
    :returns datetime.datetime:
    """
    if timestring[-3] == ":":
        timestring = timestring[:-3] + timestring[-2:]

    if "." in timestring:
        format_ = RFC3339_WITH_MS
        msecs = timestring.split(".")[1]
        msecs = msecs.split("+" if "+" in msecs else "-")[0]

        # 6 is the length of microsecond part according to doc. But some clouds
        # have a different number of digits after `.`
        # https://docs.python.org/3/library/datetime.html#tzinfo-objects
        if len(msecs) > 6:
            timestring = timestring.replace(msecs, msecs[:6])
    else:
        format_ = RFC3339

    return datetime.datetime.strptime(timestring, format_)


def to_naive_utc(dtime):
    """
    Converts tz-aware datetime object or rfc3339 string into naive one assuming timezone is UTC.
    Does nothing for already naive datetimes.

    :param (datetime.datetime|str) dtime: datetime object with tzinfo or string in rfc3339.
    :return datetime.datetime: datetime object without tzinfo.
    """
    if isinstance(dtime, str):
        dtime = from_rfc3339(dtime)
    if dtime.tzinfo:
        return datetime.datetime.utcfromtimestamp(dtime.timestamp())
    return dtime


def round_dtime(dtime: datetime.datetime, round_to: int) -> datetime.datetime:
    """Rounds dtime to nearest value multiply by round_to

    :param datetime.datetime dtime: dtime to round
    :param int round_to: round basis
    """
    if round_to > 60 * 60 * 24:
        raise exc.NotSupportedError("Round to more than 24 hours is not supported")
    seconds = (dtime.replace(tzinfo=None) - dtime.min).seconds
    rounding = (seconds + round_to / 2) // round_to * round_to
    return dtime + datetime.timedelta(seconds=rounding - seconds, microseconds=-dtime.microsecond)


def round_timedelta(delta: datetime.timedelta, round_to: int) -> datetime.timedelta:
    """Rounds timedelta to nearest value multiply by round_to

    :param datetime.timedelta delta: datetime delta to round
    :param int round_to: round basis
    """
    seconds = delta.total_seconds()
    return datetime.timedelta(seconds=(seconds + round_to / 2) // round_to * round_to)
