import contextlib
import itertools
import sys
import threading
import time

import monotonic
import peewee
import playhouse
import playhouse.db_url
import playhouse.pool
from pymysql import cursors

from common.utils import exc
from common.utils import logutil

LOG = logutil.get_logger(__name__)
CONNECT_TIMEOUT = 30
MAX_CONNECTIONS = 50
POOL_CONNECT_RETRIES = 10
POOL_CONNECT_RETRY_INTERVAL = 1


# Backported playhouse.shortcuts.RetryOperationalError, since it was removed in peewee >= 3.0.0
# see <https://github.com/coleifer/peewee/issues/1472>
class RetryOperationalErrorMixin:
    def execute_sql(self, sql, params=None, commit=True):
        try:
            cursor = super(RetryOperationalErrorMixin, self).execute_sql(sql, params, commit)
        except peewee.OperationalError:
            if not self.is_closed():
                self.close()
            with peewee.__exception_wrapper__:
                cursor = self.cursor()
                cursor.execute(sql, params or ())
                if commit and not self.in_transaction():
                    self.commit()
        return cursor


class PooledMySQLDatabaseEx(playhouse.pool.PooledMySQLDatabase, RetryOperationalErrorMixin):
    """Pooled database that checks, you're using connection per thread"""

    # pylint: disable=abstract-method

    def __init__(self, *args, **kwds):
        super(PooledMySQLDatabaseEx, self).__init__(*args, **kwds)
        self._thread_local = threading.local()

    def _connect(self):
        for retry in itertools.count(1):
            try:
                conn = playhouse.pool.PooledMySQLDatabase._connect(self)
                with conn.cursor() as cur:
                    cur.execute("set @@session.time_zone = '+00:00'")

                # It is important to set the connection charset according to the MySQL
                # server charset as PHP client does.
                # See: https://scalr-labs.atlassian.net/browse/SCALRCORE-14922
                if conn.server_charset and conn.server_charset != conn.charset:
                    LOG.debug(
                        "Changing charset from %s to %s according to the MySQL server settings...",
                        conn.charset,
                        conn.server_charset,
                    )
                    with conn.cursor() as cur:
                        cur.execute(f"set names {conn.server_charset}")
                    conn.charset = conn.server_charset

                return conn
            except ValueError as e:
                if str(e) == "Exceeded maximum connections.":
                    if retry > POOL_CONNECT_RETRIES:
                        raise exc.TimeoutError(
                            (
                                "Max retries={} exceeded for _connect(): "
                                "Pool exceeded max_connections={}"
                            ).format(POOL_CONNECT_RETRIES, self._max_connections)
                        )
                    LOG.debug(
                        "Pool exceeded max_connections=%s, retry (%s/%s) in %s second(s) ...",
                        self._max_connections,
                        retry,
                        POOL_CONNECT_RETRIES,
                        POOL_CONNECT_RETRY_INTERVAL,
                    )
                    time.sleep(POOL_CONNECT_RETRY_INTERVAL)
                else:
                    raise

    def cursor(self, *args, **kwds):  # pylint: disable=arguments-differ
        """Returns cursor instance"""
        conn = self.connection()
        original_cursorclass = conn.cursorclass
        conn.cursorclass = getattr(self._thread_local, "cursorclass", original_cursorclass)
        try:
            cursor = conn.cursor()
        finally:
            conn.cursorclass = original_cursorclass
        return cursor

    @contextlib.contextmanager
    def dict_cursor(self):
        """Sets cursor class to `cursors.DictCursor` for context lifetime"""
        self._thread_local.cursorclass = cursors.DictCursor
        try:
            yield
        finally:
            del self._thread_local.cursorclass

    def execute_sql(self, sql, *args, **kwargs):  # pylint: disable=arguments-differ
        """Executes sql query within execution context"""
        if not self.transaction_depth():
            raise Exception("You should use only ORM methods inside ExecutionContext")
        return super(PooledMySQLDatabaseEx, self).execute_sql(dedent_query(sql), *args, **kwargs)

    def fetchall_while(self, query, params=None, limit=1):
        """Executes sql with limit parameter"""
        i = 0
        while True:
            _query = "{} LIMIT {}, {}".format(query, i * limit, limit)
            results = self.execute_sql(_query, params).fetchall()
            if not results:
                break
            yield results
            if len(results) < limit:
                break
            i += 1


def connector(config, cursorclass=cursors.Cursor):
    """Returns right database instance

    :param config: user configurable database settings.
    :type config: dict

    :param cursorclass: internal application level setting
    :type cursorclass: type

    All user configurable settings must be defined in config param
    All other internal settings must be defined as additional kwds params
    """
    if config["host"].startswith("sqlite"):
        retval = peewee.SqliteDatabase(**playhouse.db_url.parse(config["host"]))
    else:
        LOG.debug(
            "Creating PooledMySQLDatabaseEx %r (host=%s port=%s)",
            config["name"],
            config["host"],
            config["port"],
        )
        retval = PooledMySQLDatabaseEx(
            config["name"],
            host=config["host"],
            port=int(config["port"]),
            user=config["user"],
            passwd=config["pass"],
            connect_timeout=config.get("connect_timeout", CONNECT_TIMEOUT),
            max_connections=config.get("max_connections", MAX_CONNECTIONS),
            cursorclass=cursorclass,
            binary_prefix=True  # <https://github.com/PyMySQL/PyMySQL/blob
            # /3674bc6fd064bf88524e839c07690e8c35223709/CHANGELOG#L61>
        )
    return retval


def db(connector):
    """Connect to database and check connection is valid"""
    # pylint: disable=protected-access
    if not hasattr(db, "_lock"):
        db._lock = threading.Lock()
    with db._lock:
        if connector not in db._instances:
            _db = connector()
            if isinstance(_db, PooledMySQLDatabaseEx):
                # XXX: WTF connect_params ?
                timeout_time = monotonic.monotonic() + _db.connect_params.get(
                    "connect_timeout", CONNECT_TIMEOUT
                )
                error = None
                LOG.debug("Connecting to database ...")
                while monotonic.monotonic() < timeout_time:
                    try:
                        _db.connect()
                        _db.close()
                        db._instances[connector] = _db
                        LOG.debug("Connected")
                        break
                    except peewee.OperationalError:
                        error = sys.exc_info()[1]
                        time.sleep(0.5)
                else:
                    msg = "Error connecting to database, connect_timeout={} exceeded. Last error: {}".format(
                        _db.connect_params.get("connect_timeout", 30), error
                    )
                    raise Exception(msg)
            elif isinstance(_db, peewee.SqliteDatabase):
                db._instances[connector] = _db
    return db._instances[connector]


get_or_create_connection = db  # modern name
db._instances = {}  # pylint: disable=protected-access


def clear_cache():
    """Clear db instances"""
    db._instances = {}  # pylint: disable=protected-access


def sql_args(count, repeat=1):
    """Returns comma separated pymysql query args template.
    Examples:
    >>> sql_args(3)
    '(%s, %s, %s)'
    >>> sql_args(3, 2)
    '(%s, %s, %s), (%s, %s, %s)'
    >>> sql_args(4, 1)
    '(%s, %s, %s, %s)'

    :param int count: Number of elements in query template.
    :param int repeat: Number of groups in query template.
    :return str: Template string.
    """
    return (("(%s), " % ("%s, " * count).rstrip(", ")) * repeat).rstrip(", ")


def dedent_query(query):
    """
    Transforms multiline query to one line query for debug purposes

    :param str query:
    :return: query formatted to one line
    :rtype: str
    """
    return " ".join(filter(lambda line: line, [line.strip() for line in query.split("\n") if line]))
