"""Common exceptions
"""


class Error(Exception):
    """A base class for Fatmouse exceptions"""

    def __init__(self, *args, **data):
        super(Error, self).__init__(*args)
        if hasattr(self, "message_fmt"):
            self.message = getattr(self, "message_fmt").format(*args, self=self)
        self.__dict__.update(data)

    def __str__(self):
        if getattr(self, "message", None):
            return "{}".format(self.message)
        else:
            return super(Error, self).__str__()


class TimeoutError(Error):  # pylint: disable=redefined-builtin
    # XXX: Pick new name for TimeoutError
    """When operation/function/task/workflow has been timed out."""


class NotAuthorizedError(Error):
    """When operation is not authorized for given context."""


class NotFoundError(Error):
    """Some thing (e.g resource, object, utility, file) was not found."""

    @classmethod
    def for_command(cls, command):
        return NotFoundError(
            (
                "{0!r} was not found. "
                "Please make sure that {0!r} is installed and available in $PATH."
            ).format(command)
        )


class NotSupportedError(Error):
    """Some action/function is not supported in this environment"""


class TryAgainError(Error):
    """Some action doesn't work, but it may work if you'll try again"""


class ConstraintError(Error):
    """When constraints between things (e.g. params, definitions) are violated"""


class MalformedError(Error):
    """When parameter violates validation rules"""


class DuplicateError(Error):
    """During creation, when creating thing already exists"""


class UpdateRepositoryError(Error):
    """When update can't proceed because something wrong with repository or no
    connection to repository
    """


class UpdateRollbackFailureError(Error):
    """When update fails and we can't rollback into a working state"""


class UpdateRollbackedError(Error):
    """When update was rollbacked"""


class UpdateDeniedError(Error):
    """When update server denies update or when task running on agent is
    unsafe to interrupt
    """


class UpdateIntegrityError(Error):
    """When expected version and actual version of running agent mismatch after
    update
    """


class StorageError(Error):
    """When storage operation failed"""


class AlreadyMountedError(StorageError):
    """When storage is already mounted"""

    def __init__(self, *args, **data):
        self.device = data.pop("device", None)
        super(AlreadyMountedError, self).__init__(*args, **data)


class LimitExceeded(Error):
    """When limit has been exceeded."""
