import importlib
import os
import platform
import re
import sys

import psutil

from common.utils import mixutil

WINDOWS = sys.platform == "win32"
SYSTEMD_PID = 1  # When used as init system, systemd will always have PID==1


class FactError(Exception):
    pass


class CachedFunction(object):
    """Adds cache to function, so multiple launches won't perform hard calculations"""

    def __init__(self, function):
        self.function = function
        self.value = None

    def __call__(self):
        if not self.value:
            self.value = self.function()
        return self.value

    def invalidate(self):
        """Deletes cache so next launch will perform actual function call"""
        self.value = None


class CompositeNode(list):
    """Container for trees|dicts that could be merged to make Node element"""


class CacheTree(dict):
    """
    Represents tree datastructure with leaves may be CachedFunction objects which are
    evalueted on access.
    """

    def has_leaf(self, key):
        """
        Returns True if node is leaf - value, or is composite and has leaf among its parts.
        :param hashable key: node key
        """
        node = self.get_raw(key)
        if isinstance(node, CompositeNode):
            return any(CacheTree._is_leaf(node_part) for node_part in node)
        else:
            return CacheTree._is_leaf(node)

    @classmethod
    def _is_leaf(cls, node):
        """Returns True if node is not a subtree or composite container"""
        return not isinstance(node, cls) and not isinstance(node, CompositeNode)

    @classmethod
    def _get_leaves(self, node, acc):
        """
        Recursively walks the tree and gathers leaves into acc set
        :param node: root of a tree to walk
        :param set acc: accumulator to store leaves
        """
        if CacheTree._is_leaf(node):
            acc.add(node)
        elif isinstance(node, CompositeNode):
            for node_part in node:
                CacheTree._get_leaves(node_part, acc)
        else:
            for subnode in node.raw_values():
                CacheTree._get_leaves(subnode, acc)

    def _get_leaves_raw(self, key=None, acc=None):
        """Returns all leaves under given node"""
        node = self.get_raw(key) if key else self
        acc = acc or set()
        CacheTree._get_leaves(node, acc)
        return acc

    def get_raw(self, key):
        """Returns node without expanding it to actual data"""
        return super(CacheTree, self).__getitem__(key)

    def raw_values(self):
        """Like values() but uses get_raw instead of __getitem__"""
        return [self.get_raw(k) for k in self.keys()]

    def invalidate(self, key=None):
        """
        Invalidates cache of all underlying CahedFunctions.
        :param hashable key: subtree/node point
        """
        for f in self._get_leaves_raw(key):
            if isinstance(f, CachedFunction):
                f.invalidate()

    def add_subtree(self, key):
        """
        Adds point to join new nodes
        :param hashable key: point where subtree will be created
        """
        if key in self:
            current_value = self.get_raw(key)
            subtree = CacheTree()
            if isinstance(current_value, CompositeNode):
                current_value.append(subtree)
            else:
                super(CacheTree, self).__setitem__(key, CompositeNode([current_value, subtree]))
        else:
            subtree = CacheTree()
            super(CacheTree, self).__setitem__(key, subtree)
        return subtree

    def __getitem__(self, key):
        """Returns value of leaf or subtree"""
        value = super(CacheTree, self).__getitem__(key)
        if isinstance(value, CompositeNode):
            # if node has more than one value, it implies that this subtree
            # is constructed from more than one subtrees, so we need to
            # merge them into one
            result = CacheTree()
            for node_part in value:
                if callable(node_part):
                    # if it has function-leaf among these values, it implies
                    # that this leaf returns dict
                    node_part = node_part()
                result.update(node_part)
            return result
        if callable(value):
            return value()
        return value

    def __setitem__(self, key, value):
        """
        It doesn't override values, but merges them.
        In that case values must be dicts or CachedFunctions that return dicts
        """
        if key in self:
            current_value = super(CacheTree, self).__getitem__(key)
            if isinstance(current_value, CompositeNode):
                current_value.append(value)
            else:
                super(CacheTree, self).__setitem__(key, CompositeNode([current_value, value]))
        else:
            super(CacheTree, self).__setitem__(key, value)

    def items(self):
        """
        Overloaded to use __getitem__() so CompositeNodes and CachedFunctions are expanded.
        """
        return [(k, self[k]) for k in self]

    def to_native(self):
        """
        Expands underlying CompositeNodes and CachedFunctions and returns nested dict.

        :returns: dict
        """
        return {k: self[k].to_simple() if isinstance(self[k], CacheTree) else self[k] for k in self}

    def __str__(self):
        return "{}".format(self.to_native())

    def __repr__(self):
        return self.__str__()


class Fact(object):
    """
    A singleton class for lazy-getting an information from fact
    functions-collectors.
    Works in two ways:
        - When used as a decorator binds a function to a binding specified by
        a binding argumet
        - When used as a dict like singleton object - evaluates a corresponding
        fact collector or immediatley returns a value from a cache.
    Usage:
        >>>fact = Fact()
        >>>
        >>>@fact(binding='some_binding'):
        >>>def some_fact_collector():
        >>>    return {'facts': ['some facts']}
        >>>
        >>>fact['some_binding']['facts']
        ['some facts']
        >>>@fact(binding='another_binding'):
        >>>def some_fact_collector():
        >>>    return 'some_value'
        >>>fact['another_binding']
        'some_value'
    """

    path_delimeter = "."

    def __init__(self):
        """
        self._tree structure:
        It's a CacheTree object, that has special subtree key '' which subtrees will be
        accessible as if they are lying at root point. Functions that are registered at this
        binding must return dicts.
        Example:
        self._tree == {'': [
                <function_returns_dict={'key1': 'value1'}>,
                <another_one={'key2': 'value2'}>],
            ...}
        # expand self:
        self == {'key1': 'value1', 'key2': 'value2', ...}
        """
        self._tree = CacheTree()
        self.includes = []

    def __call__(self, binding="", no_cache=False):
        def wrapper(function):
            # If it's a new function for this binding,
            # add it to a list of functions, our singleton is aware of.
            # This will only be triggered when a class instace is
            # acting as a decorator
            if function is not None:
                points = binding.split(self.path_delimeter)

                current = self._tree
                for point in points[:-1]:
                    if point not in current or current.has_leaf(point):
                        # creating tree structure
                        # if current point has leaf, it means that from this point
                        # fact is constructed from dicts and inner bindings
                        current = current.add_subtree(point)
                    else:
                        current = current[point]

                # last point is where the leaf-function is placed
                current[points[-1]] = function if no_cache else CachedFunction(function)
            return function

        return wrapper

    def __getitem__(self, key):
        points = key.split(self.path_delimeter)
        if points[0] not in self._tree and "" in self._tree:
            # '' is exceptional binding because we can skip pointing it in
            # provided key, so we need to add it if needed
            points.insert(0, "")
        current = self._tree
        for point in points:
            try:
                current = current[point]
            except KeyError as e:
                raise KeyError("{}, fact requested '{}'".format(str(e), key))
        return current

    def invalidate(self, binding=None):
        """
        Clears cached values of facts under binding.
        :param hashable binding: fact binding
        """
        if binding and self.path_delimeter in binding:
            path, name = binding.rsplit(self.path_delimeter, 1)
            self[path].invalidate(name)
        else:
            self._tree.invalidate(binding)

    def load(self):
        """Loads modules with fact definitions"""
        for module in self.includes:
            importlib.import_module(module)


fact = Fact()


@fact(binding="os")
def os_fact(rootfs_dir="/"):
    """
    Returns a dictionary containing os info.
    """
    lsb = {}
    ret = {}

    def _full_file_path(relative_path):
        return os.path.join(rootfs_dir, relative_path)

    def _file_exists(relative_path):
        if os.path.isfile(_full_file_path(relative_path)):
            return True
        else:
            return False

    if WINDOWS:
        ret["family"] = "windows"
        ret["name"] = "windows"
        v = mixutil.Version(platform.win32_ver()[1])
        if v >= (10, 0, 20348):
            ret["release"] = "2022"
        elif v >= (10, 0, 17623):
            ret["release"] = "2019"
        elif v >= (10, 0):
            ret["release"] = "2016"
        elif v >= (6, 2):
            ret["release"] = "2012"
        elif v >= (5, 2):
            ret["release"] = "2008"
        else:
            ret["release"] = "2003"
        ret["codename"] = ret["release"] + "Server"
        ret["arch"] = "x86_64" if platform.uname()[4].lower() == "amd64" else "i386"
    elif _file_exists("etc/lsb-release"):
        for line in open(_full_file_path("etc/lsb-release")).readlines():
            # Matches any possible format:
            #     DISTRIB_ID="Ubuntu"
            #     DISTRIB_ID='Mageia'
            #     DISTRIB_ID=Fedora
            #     DISTRIB_RELEASE='10.10'
            #     DISTRIB_CODENAME='squeeze'
            #     DISTRIB_DESCRIPTION='Ubuntu 10.10'
            regex = re.compile(
                r"^DISTRIB_((?:ID|RELEASE|CODENAME|DESCRIPTION))="
                r'(?:\'|")?([\w\s\.-_]+)(?:\'|")?'
            )
            match = regex.match(line)
            if match:
                # Adds: lsb_{id,release,codename,description}
                lsb[match.groups()[0].lower()] = match.groups()[1].rstrip()

        if "codename" in lsb:
            ret["codename"] = lsb["codename"]
        if "release" in lsb:
            ret["release"] = lsb["release"]

    if _file_exists("etc/arch-release"):
        ret["name"] = "arch"
        ret["family"] = "arch"
    elif _file_exists("etc/debian_version"):
        ret["name"] = "debian"
        ret["family"] = "debian"
        if "id" in lsb:
            if "GCEL" in lsb["id"]:
                ret["name"] = "gcel"
            if "Ubuntu" in lsb["id"]:
                ret["name"] = "ubuntu"
            elif (
                _file_exists("etc/issue.net")
                and "Ubuntu" in open(_full_file_path("etc/issue.net")).readline()
            ):
                ret["name"] = "ubuntu"
    elif _file_exists("etc/gentoo-release"):
        ret["name"] = "gentoo"
        ret["family"] = "gentoo"
    elif _file_exists("etc/fedora-release"):
        ret["name"] = "fedora"
        ret["family"] = "redhat"
    elif _file_exists("etc/mandriva-version"):
        ret["name"] = "mandriva"
        ret["family"] = "mandriva"
    elif _file_exists("etc/mandrake-version"):
        ret["name"] = "mandrake"
        ret["family"] = "mandriva"
    elif _file_exists("etc/mageia-version"):
        ret["name"] = "mageia"
        ret["family"] = "mageia"
    elif _file_exists("etc/meego-version"):
        ret["name"] = "meego"
        ret["family"] = "meego"
    elif _file_exists("etc/vmware-version"):
        ret["name"] = "vmwareesx"
        ret["family"] = "vmware"
    elif _file_exists("etc/bluewhite64-version"):
        ret["name"] = "bluewhite64"
        ret["family"] = "bluewhite"
    elif _file_exists("etc/slamd64-version"):
        ret["name"] = "slamd64"
        ret["family"] = "slackware"
    elif _file_exists("etc/slackware-version"):
        ret["name"] = "slackware"
        ret["family"] = "slackware"
    elif _file_exists("etc/oracle-release"):
        ret["family"] = "redhat"
        ret["name"] = "oel"
    elif _file_exists("etc/redhat-release"):
        ret["family"] = "redhat"
        data = open(_full_file_path("etc/redhat-release"), "r").read()
        if "centos" in data.lower():
            ret["name"] = "centos"
        elif "scientific" in data.lower():
            ret["name"] = "scientific"
        elif "goose" in data.lower():
            ret["name"] = "goose"
        else:
            ret["name"] = "redhat"
    elif _file_exists("etc/system-release"):
        ret["family"] = "redhat"
        data = open(_full_file_path("etc/system-release"), "r").read().strip()
        if data.lower().startswith("amazon"):
            ret["name"] = "amazon"
            ret["release"] = data[-7:]
            # examples:
            # Amazon Linux AMI release 2018.03
            # Amazon Linux release 2 (Karoo)
            release_match = re.search(r"release\s((\d+\.)?(\d+))", data)
            codename_match = re.search(r"\((.*?)\)$", data)
            if release_match:
                ret["release"] = release_match.group(1)
            if codename_match:
                ret["codename"] = codename_match.group(1)
    elif _file_exists("etc/SuSE-release"):
        ret["family"] = "suse"
        data = open(_full_file_path("etc/SuSE-release"), "r").read()
        if "SUSE Linux Enterprise Server" in data:
            ret["name"] = "sles"
        elif "SUSE Linux Enterprise Desktop" in data:
            ret["name"] = "sled"
        elif "openSUSE" in data:
            ret["name"] = "opensuse"
        else:
            ret["name"] = "suse"

    elif _file_exists("etc/os-release"):
        with open(_full_file_path("etc/os-release")) as f:
            release_info = f.read()
            if "CoreOS" in release_info:
                ret["name"] = "coreos"
                ret["family"] = "coreos"
            elif "SUSE Linux Enterprise Server" in release_info:
                ret["family"] = "suse"
                ret["name"] = "sles"
            elif "SUSE Linux Enterprise Desktop" in release_info:
                ret["family"] = "suse"
                ret["name"] = "sled"
            elif "openSUSE" in release_info:
                ret["family"] = "suse"
                ret["name"] = "opensuse"

            if ret["family"] == "suse":
                f.seek(0)
                for line in f.readlines():
                    if line.startswith("VERSION_ID="):
                        ret["release"] = line.replace('"', "").split("=")[1].strip()
                        break

    # TODO: dist is deprecated
    name, release, codename = platform.dist()  # pylint: disable=deprecated-method
    if "name" not in ret:
        ret["name"] = name
    if "release" not in ret:
        ret["release"] = release
    ret["release"] = mixutil.Version(ret["release"])
    ret["version"] = ret["release"]
    if "codename" not in ret:
        if not codename and release and "debian" == ret["name"]:  # FAM-751
            major_version = int(float(release))
            debian_codenames = {6: "squeeze", 7: "wheezy", 8: "jessie", 9: "stretch"}
            ret["codename"] = debian_codenames.get(major_version)
        else:
            ret["codename"] = codename

    if "name" not in ret:
        ret["name"] = "unknown {}".format(ret["kernel"]).lower()
        ret["family"] = "unknown"
    if "family" not in ret:
        ret["family"] = "unknown"

    if ret["family"] != "windows":
        arch = os.uname()[-1]
        if re.search(r"i\d86", arch):
            arch = "i386"
        ret["arch"] = arch

    if ret["family"] != "windows":
        ret["init"] = "sysv"
        try:
            systemd_process = psutil.Process(SYSTEMD_PID)
            if systemd_process.name() == "systemd":
                ret["init"] = "systemd"
        except psutil.NoSuchProcess:
            pass

    return {
        "name": ret["name"],
        "family": ret["family"],
        "release": ret["release"],
        "codename": ret["codename"],
        "arch": ret["arch"],
        "init": ret.get("init"),
    }


@fact(binding="container")
def container():
    container = None
    if WINDOWS and os.environ.get("TURBOCONTAINER") == "1":
        container = "turbo"
    else:
        cgroup_file = "/proc/1/cgroup"
        if os.path.exists(cgroup_file):
            with open(cgroup_file) as fp:
                for line in fp:
                    cgroup = line.rsplit(":")[-1].strip()
                    if cgroup.startswith("/docker"):
                        container = "docker"
                        break
    return container


@fact(binding="manage_container_host")
def manage_container_host():
    """
    Return True if scalarizr run inside container
    """
    if (
        fact["container"] == "docker"
        and os.path.isdir("/rootfs")
        and os.environ.get("FAM_MANAGE_CONTAINER_HOST")
    ):
        return True
    else:
        return False


@fact(binding="container_host.os")
def container_host_os():
    """
    Return host os_fact in case when scalarizr run inside container
    """
    if fact["manage_container_host"]:
        return os_fact("/rootfs")
