import httplib2
import requests.utils


class ProxyInfo(httplib2.ProxyInfo):
    """
    This class extends :class:`httplib2.ProxyInfo` to imrove "no_proxy" support.
    """

    def __init__(self, *args, no_proxy: str = None, **kwargs) -> None:
        """
        :param no_proxy: The comma separated list of hosts to exclude from proxying.
        """
        super().__init__(*args, **kwargs)
        self.__no_proxy = no_proxy

    def bypass_host(self, hostname: str) -> bool:
        """
        Has this host been excluded from the proxy config.
        """
        no_proxy = (self.__no_proxy or "").strip()
        if not no_proxy:
            return False
        if no_proxy == "*":
            return True

        safe_url = hostname.strip().rstrip("/")
        if "://" not in safe_url:
            # the requests library expects a URL with scheme
            safe_url = f"any://{safe_url}"

        bypass: bool = requests.utils.should_bypass_proxies(safe_url, no_proxy)  # type: ignore
        return bypass


class HTTPSConnectionWithTimeout(httplib2.HTTPSConnectionWithTimeout):
    """
    This class extends :class:`httplib2.HTTPSConnectionWithTimeout`
    to fix "no_proxy" support.
    """

    def connect(self):
        """Connect to a host on a given (SSL) port."""
        if self.proxy_info and self.proxy_info.isgood():
            if not self.proxy_info.applies_to(self.host):
                self.proxy_info = None  # drop proxy setting for current request
        return super(HTTPSConnectionWithTimeout, self).connect()


def patch_httplib2(no_proxy: bool = True) -> None:
    """Patch httplib to add new features or fixes.

    Features:
      - no_proxy: Fix no_proxy support for HTTPS connections.
    """
    if no_proxy:
        httplib2.SCHEME_TO_CONNECTION["https"] = HTTPSConnectionWithTimeout
