"""
Various utilites for libcloud
"""
# pylint: disable=import-error
import itertools
import os
import re

import libcloud.backup
import libcloud.common.base
import libcloud.common.google
import libcloud.compute
import libcloud.container
import libcloud.dns
import libcloud.storage
import libcloud.storage.drivers.s3
from libcloud import http
from libcloud.common import exceptions as libcloud_exc
from libcloud.compute import base
from libcloud.compute.drivers import ec2

from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import six

LOG = logutil.get_logger(__name__)
POLLING_INTERVAL = 1
EC2_NOT_FOUND_CODE_REGEX = r"Invalid.*\.NotFound"
AZURE_NOT_FOUND_CODE_REGEX = r"\[(NotFound|ResourceNotFound)\]"

# TESTS-ONLY
# we need to set ec2.API_VERSION, ec2.NAMESPACE, ec2.EC2Connection.version
# to that value in order to get aws api mock (moto) working
if os.environ.get("FAM_EC2_API_VERSION"):
    ec2.API_VERSION = os.environ["FAM_EC2_API_VERSION"]
    ec2.NAMESPACE = "http://ec2.amazonaws.com/doc/%s/" % os.environ["FAM_EC2_API_VERSION"]
    ec2.EC2Connection.version = os.environ["FAM_EC2_API_VERSION"]


class LoggingConnection(http.LibcloudConnection):
    """
    Debug class to log all HTTP(s) requests.
    """

    log_level = "DEBUG"

    def getresponse(self):
        """
        Logging libcloud response data.
        """
        response = super(LoggingConnection, self).getresponse()
        message = "Response:\n%s\nHeaders: %s\nBody: %s"
        log = getattr(LOG, self.log_level.lower(), None)
        if log:
            # pylint: disable=protected-access,not-callable
            body = response.text if response._content is not False else "<data stream>"
            log(message, response.url, response.headers, body)
        return response

    def request(self, method, url, **kwargs):  # pylint: disable=arguments-differ
        """
        Logging libcloud request data
        """
        if self.http_proxy_used:
            if self.proxy_username and self.proxy_password:
                proxy_url = f"http://****:****@{self.proxy_host}:{self.proxy_port}"
            else:
                proxy_url = f"http://{self.proxy_host}:{self.proxy_port}"
        else:
            proxy_url = None

        message = "Request:\n%s %s%s\nHeaders: %s\nBody: %s\nProxy: %s"
        log = getattr(LOG, self.log_level.lower(), None)
        if log:
            # pylint: disable=not-callable
            log(
                message,
                method,
                self.host,
                url,
                kwargs.get("headers"),
                kwargs.get("body"),
                proxy_url,
            )
        try:
            return super(LoggingConnection, self).request(method, url, **kwargs)
        except Exception as err:
            LOG.error("Request failed: %s", err)
            raise


libcloud.common.base.Connection.conn_class = LoggingConnection


class GoogleOAuth2CredentialInMemoryCache(libcloud.common.google.GoogleOAuth2Credential):
    """Removes caching of auth token to file"""

    def __init__(self, *args, **kwargs):
        """Adds in-memory storage"""
        self.token_storage = {}
        super(GoogleOAuth2CredentialInMemoryCache, self).__init__(*args, **kwargs)

    def _write_token_to_file(self):
        """Saves token to in-memory storage"""
        self.token_storage[self.user_id] = self.token

    def _get_token_from_file(self):
        """Retrieves token from in-memory storage. Returns `token` or `None`"""
        return self.token_storage.get(self.user_id)


libcloud.common.google.GoogleOAuth2Credential = GoogleOAuth2CredentialInMemoryCache


def morph_action_hook(self, action):
    """Returns full request path without splitting and joining url.

    urllib.parse.urljoin removes '//' from file path, so any s3 file url
    like https://amazonaws.com/bucket-name/subdir//file.ext will be invalid after
    transformations
    """
    return self.request_path + action


def patch_S3SignatureV4Connection():
    """Patches libcloud.storage.drivers.s3.S3SignatureV4Connection class"""
    libcloud.storage.drivers.s3.S3SignatureV4Connection.morph_action_hook = morph_action_hook


class Error(exc.Error):
    pass


TYPE_TO_LIST_METHOD = {
    base.Node: "list_nodes",
    base.NodeLocation: "list_locations",
    base.NodeImage: "list_images",
    base.StorageVolume: "list_volumes",
    base.VolumeSnapshot: "list_snapshots",
}


def create_driver(credentials, api="compute"):
    """
    Creates libcloud driver using provided LibcloudCredentials

    :param :class:`types.LibcloudCredentials` credentials: Cloud credentials.
    :param str api: A libcloud api section to get a driver from.
        Can be one of ('compute', 'backup', 'container', 'dns', 'storage')

    :rtype :class:`libcloud.compute.base.NodeDriver`
    """
    if api == "compute":
        libcloud_module = libcloud.compute
    elif api == "container":
        libcloud_module = libcloud.container
    elif api == "dns":
        libcloud_module = libcloud.dns
    elif api == "storage":
        libcloud_module = libcloud.storage
    elif api == "backup":
        libcloud_module = libcloud.backup
    else:
        raise Error("Unknown libcloud api {}".format(api))
    driver_cls = libcloud_module.providers.get_driver(credentials.provider)
    return driver_cls(*credentials.args, **credentials.kwargs)


def find(
    object_type=None,
    driver=None,
    query=None,
    credentials=None,
    max_results=1,
    ignore_case=False,
    api_call_kwds=None,
):
    """
    Get a list of objects corresponding to given `query` constraints.

    :param class object_type: One of `libcloud.comput.base` libcloud object types
                        (see `libcloudutil.TYPE_TO_LIST_METHOD` mapping)

    :param driver: A driver object to used for finding an object.
                If this is set then `node` should be none
    :type driver: None or class:`libcloud.compute.base.NodeDriver`

    :param query: A set of filters to apply. If not specified, no filters will be applied.
    :type query: dict or None

    :returns : List of objects representing storage volumes.
    :rtype : list
    """
    api_call_kwds = api_call_kwds or {}
    listmethod = None
    for type_ in TYPE_TO_LIST_METHOD:
        if issubclass(object_type, type_):
            listmethod = TYPE_TO_LIST_METHOD[type_]
            break
    else:
        raise TypeError("{} is not associated with any list method".format(object_type))

    if not driver and credentials:
        driver = create_driver(credentials)

    finder = getattr(driver, listmethod)
    if query is None:
        try:
            return finder(**api_call_kwds)
        except libcloud_exc.BaseHTTPError as e:
            # XXX: For GCE we need to catch a different exception
            if re.search(EC2_NOT_FOUND_CODE_REGEX, e.message) or re.search(
                AZURE_NOT_FOUND_CODE_REGEX, e.message
            ):
                raise exc.NotFoundError(e.message)
            raise

    elif not isinstance(query, dict):
        raise AssertionError("`query` kwarg is expected to be `None` " "or an instance of `dict` ")

    def filter_fn(obj):
        get_obj_attr = lambda x: getattr(obj, x, getattr(obj, "extra", {}).get(x))
        is_equal = lambda x, y: x.lower() == y.lower() if (ignore_case and x and y) else x == y
        return all(is_equal(get_obj_attr(key), value) for key, value in query.items())

    try:
        object_iterator = six.moves.filter(filter_fn, finder(**api_call_kwds))
    except libcloud_exc.BaseHTTPError as e:
        # XXX: For GCE we need to catch a different exception
        if re.search(EC2_NOT_FOUND_CODE_REGEX, e.message) or re.search(
            AZURE_NOT_FOUND_CODE_REGEX, e.message
        ):
            raise exc.NotFoundError(e.message)
        raise
    return list(itertools.islice(object_iterator, max_results))


def find_one(*args, **kwargs):
    """
    A wrapper around `common.utils.libcloudutil.find`
    """
    try:
        return find(*args, **kwargs)[0]
    except IndexError:
        query = kwargs.get("query")
        object_type = kwargs.get("object_type", args[0] if args else "Unknown")
        message = "No {}s found".format(object_type)
        if query:
            message = "No {}s matching search criteria found. Query: {}".format(object_type, query)
        raise exc.NotFoundError(message)


def wait_state_change(
    object_type=None,
    object_id=None,
    object_getter=None,
    case_insensitive_id=False,
    from_state=None,
    to_state=None,
    abort_on=None,
    timeout=None,
    driver=None,
    state_getter=None,
    ignore_errors=None,
):
    """
    Waits a libcloud object by polling for a given state, then exits.
    If objects transits to a state in abort_on list, then Exception is raised.

    :param object_type: One of `libcloud.compute.base` libcloud object types
        (see `libcloudutil.TYPE_TO_LIST_METHOD` mapping)
    :type object_type: class

    :param function object_getter: Function which will be used to get object (eg. call libcloud
        getter method). Must not accept any parameters.

    :param  driver: A libcloud driver object.
    :type driver: :class:`libcloud.compute.base.NodeDriver`

    :param str object_id: object's  unique identificator (typically `id` attribute)

    :param str from_state: State to transit from. If `to_state` is not specified
        transaction to any state which is not `from_state` or `abort_on`  will be
        assumed a positive result.
    :type from_state: str or None

    :param str to_state: State to wait for.
        This can be used without `from_state` being specified.
    :type to_state: str or None

    :param tuple[str] abort_on: When any of `abort_on` states is encountered
        an error will be raised.

    :param tuple[class] ignore_errors: A list of exceptions to be ignored.
    """
    ignore_errors = ignore_errors or tuple()

    try_msg = "Waiting {} state change".format(object_type.__name__)
    if from_state:
        try_msg += ' from "{}"'.format(from_state)
    if to_state:
        try_msg += ' to "{}"'.format(to_state)
    try_msg += " ({})".format(object_id)
    LOG.info(try_msg)

    def poller(state_getter=state_getter):
        try:
            if object_getter:
                obj = object_getter()
            else:
                obj = find_one(
                    object_type,
                    query={"id": object_id},
                    driver=driver,
                    ignore_case=case_insensitive_id,
                )
            current_state = state_getter(obj) if state_getter else obj.state

            if current_state is None:
                LOG.debug("No state information received, assuming state is 'unknown'")
            else:
                LOG.debug("Current state is %s", current_state)

            if abort_on and current_state in abort_on:
                raise Error(
                    "Aborting! {} state is '{}' ({})".format(
                        object_type.__name__, current_state, object_id
                    )
                )

            if (to_state and current_state == to_state) or (
                from_state and current_state != from_state
            ):
                LOG.info("%s state is '%s' (%s)", object_type.__name__, current_state, object_id)
                return True
        except ignore_errors as e:
            # pylint: disable=no-member
            LOG.debug("Ignored error: %s", six.text_type(e))

    mixutil.try_until(poller, try_msg=try_msg, interval=POLLING_INTERVAL, timeout=timeout)
