"""
All filesystem-related utilities belong here.
"""
import collections
import os
import pathlib
import shutil
import sys

from common.utils import exc
from common.utils import logutil
from common.utils import six
from common.utils import subprocess2
from common.utils.facts import fact

is_windows = sys.platform == "win32"
if not is_windows:
    from common.utils import augeasutil

LOG = logutil.get_logger(__name__)

__all__ = [
    "FSTAB_PATH",
    "MOUNTS_PATH",
    "ENOFS",
    "BlkidCmd",
    "DDCmd",
    "FileCmd",
    "LsblkCmd",
    "MountCmd",
    "SyncCmd",
    "UmountCmd",
    "blkid",
    "chmod_r",
    "chown_r",
    "dd",
    "fstab",
    "get_fstype",
    "has_partition_table",
    "mount",
    "mounts",
    "statvfs",
    "sync",
    "umount",
    "df",
    "disk_usage",
]

if fact["manage_container_host"]:
    FSTAB_PATH = "/rootfs/etc/fstab"
    MOUNTS_PATH = "/rootfs/proc/mounts"
else:
    FSTAB_PATH = "/etc/fstab"
    MOUNTS_PATH = "/proc/mounts"
ENOFS = "NoFilesystem"


FstabLikeEntry = collections.namedtuple(
    "_FstabLikeEntry", "device mpoint fstype options dump fsck_order"
)
BlkidEntry = collections.namedtuple("BlkidEntry", "device uuid fstype")
DFEntry = collections.namedtuple("df", "device size used available used_percentage mpoint")

MountCmd = subprocess2.new_command("MountCmd", binary="/bin/mount")
UmountCmd = subprocess2.new_command("UmountCmd", binary="/bin/umount")
BlkidCmd = subprocess2.new_command("BlkidCmd", binary="/sbin/blkid")
LsblkCmd = subprocess2.new_command("LsblkCmd", binary="/bin/lsblk")
SyncCmd = subprocess2.new_command("SyncCmd", binary="/bin/sync")
FileCmd = subprocess2.new_command("FileCmd", binary="/usr/bin/file")
DDCmd = subprocess2.new_command("DDCmd", binary="/bin/dd", long_prefix="", long_jumper="=")
DFCmd = subprocess2.new_command("DFCmd", binary="/bin/df")


def chown_r(path, user=None, group=None):
    if six.PY3:
        # pylint: disable=no-member
        return shutil.chown(path, user, group)
    else:
        owner = "{}:{}".format(user or "", group or "")
        if owner.endswith(":"):
            owner = owner[:-1]

    subprocess2.call(
        ("/bin/chown", "--recursive", owner, path),
        stdout=subprocess2.DEVNULL,
        stderr=subprocess2.STDOUT,
    )


def chmod_r(path, mode):
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for name in dirs + files:
                os.chmod(os.path.join(root, name), mode)
    else:
        os.chmod(path, mode)


def blkid():
    result = {}
    # blkid out:
    # /dev/xvdb: UUID="50309f91-b92e-49b9-bbac-dceed65cb311" TYPE="ext3"
    # /dev/block/202:1: UUID="64d674f8-35c3-497b-b46b-156c7d86386e" TYPE="xfs"
    out = subprocess2.check_output([BlkidCmd.binary, "-sUUID", "-sTYPE"])
    for line in out.splitlines():
        if ":" not in line:
            continue
        device, data = [e.strip() for e in line.rsplit(":", 1)]
        entries = [e for e in data.split(" ") if "=" in e]
        entrydict = dict([(i.strip('"') or None) for i in e.split("=")] for e in entries)
        result[device] = BlkidEntry(device, entrydict.get("UUID"), entrydict.get("TYPE"))

    return result


def sync():
    subprocess2.check_call([SyncCmd.binary])


def dd(**kwds):
    return subprocess2.check_call(DDCmd().args(**kwds).build())


class FstabLikeDict(six.moves.Mapping):
    # pylint: disable=no-member

    _space_escape_seq = "\\040"

    def __init__(self, filename):
        self.filename = filename
        self._reload()

    def __getitem__(self, device_or_mpoint):
        return self._get_item_fn(
            lambda entry: entry.device == device_or_mpoint or entry.mpoint == device_or_mpoint,
            device_or_mpoint,
        )

    def __iter__(self):
        return iter(self.entries)

    def __len__(self):
        return len(self.entries)

    def __contains__(self, entry):
        return entry in self.mpoints or entry in self.devices

    @property
    def mpoints(self):
        return [entry.mpoint for entry in self.entries]

    @property
    def devices(self):
        return [entry.device for entry in self.entries]

    def get_device(self, device, *default):
        return self._get_item_fn(lambda entry: entry.device == device, device, *default)

    def get_mpoint(self, mpoint, *default):
        return self._get_item_fn(lambda entry: entry.mpoint == mpoint, mpoint, *default)

    def put(
        self, device, mpoint, fstype="auto", options=None, dump="0", fsck_order="0", index=None
    ):
        if device in self.devices:
            index = self.devices.index(device) + 1
        elif mpoint in self.mpoints:
            index = self.mpoints.index(mpoint) + 1
        if index is None:
            index = len(self.entries) + 1

        mpoint = self._escape_mpoint(mpoint)

        # validate mount options
        if isinstance(options, six.string_types):
            options = options.split(",")
        options = options or ["defaults"]

        def normopt(opt):
            if isinstance(opt, str):
                return opt.split("=", 1) if "=" in opt else (opt, None)
            return opt  # we believe in option to be ``tuple(optionname, optionvalue)``

        options = tuple(map(normopt, options))

        changes = list()
        if index == 0:
            changes.append("ins 0 before /files/{}/1".format(self._augeas_filename))
        changes.append("set /augeas/context /files/{}/{}".format(self._augeas_filename, index))
        changes.append("set spec {}".format(device))
        changes.append("set file {}".format(mpoint))
        changes.append("set vfstype {}".format(fstype))
        changes.append("set dump {}".format(dump))
        changes.append("set passno {}".format(fsck_order))

        changes.append("rm opt")
        for n, (option, value) in enumerate(options, 1):
            changes.append("ins opt before dump")
            changes.append("set opt[{}] {}".format(n, option))
            if value:
                changes.append("set opt[{}]/value {}".format(n, value))

        changes.append("save")

        try:
            return augeasutil.execute(changes, self.augeas)
        finally:
            self._reload()

    def remove(self, device_or_mpoint):
        if not self.remove_device(device_or_mpoint):
            self.remove_mpoint(device_or_mpoint)

    __delitem__ = remove

    def remove_device(self, device):
        return self._remove("spec", device)

    def remove_mpoint(self, mpoint):
        return self._remove("file", mpoint)

    def _remove(self, attrname, attrvalue):
        nodes = self.augeas.match(
            '/files/{}/*[{} = "{}"]'.format(self._augeas_filename, attrname, attrvalue)
        )
        if not nodes:
            return
        try:
            changes = []
            changes.extend("rm {}".format(node) for node in nodes)
            changes.append("save")
            return augeasutil.execute(changes, self.augeas)
        finally:
            self._reload()

    @property
    def _augeas_filename(self):
        return self.filename.strip("/")

    def _reload(self):
        self.augeas = self._create_fstab_augeas()
        self.entries = []

        for path in self.augeas.match("/files/{0}/*".format(self._augeas_filename)):
            device = self.augeas.get(path + "/spec")
            mpoint = self._unescape_mpoint(self.augeas.get(path + "/file"))
            fstype = self.augeas.get(path + "/vfstype")
            options = [
                (self.augeas.get(p), self.augeas.get(p + "/value"))
                for p in self.augeas.match(path + "/opt")
            ]
            dump = self.augeas.get(path + "/dump")
            fsck_order = self.augeas.get(path + "/passno")

            entry = [device, mpoint, fstype, options, dump, fsck_order]
            if any(entry):
                self.entries.append(FstabLikeEntry(*entry))

    def _create_fstab_augeas(self):
        augeas = augeasutil.augeas(autoload=False)
        augeas.set("/augeas/load/Fstab/lens", "Fstab.lns")
        augeas.set("/augeas/load/Fstab/incl[1]", self._augeas_filename)
        augeas.load()
        return augeas

    def _get_item_fn(self, fn, key, *default):
        try:
            return list(filter(fn, self.entries))[0]
        except IndexError:
            if default:
                return default[0]
            else:
                raise KeyError(key)

    def _escape_mpoint(self, mpoint):
        if mpoint:
            return mpoint.replace(" ", self._space_escape_seq)

    def _unescape_mpoint(self, mpoint):
        if mpoint:
            return mpoint.replace(self._space_escape_seq, " ")


def mounts(filename=None):
    """
    Manages the contents of the /proc/mounts.
    """
    return FstabLikeDict(filename or MOUNTS_PATH)


def _extract_device_name(device):
    """Extract device name from its full path.

    For RAID arrays that are represented as logical volumes read the link target:
    /dev/mapper/d79278f9--0007--444c--845a--fa3706ff06e7-lvol0 -> /dev/dm-0
    :param device: device path.
    :return: device name.
    """
    with open("/proc/diskstats") as diskstas_file:
        devices = [l.split()[2] for l in diskstas_file.readlines()]
    device_name = os.path.basename(device)
    if device_name not in devices and os.path.islink(device):
        device_name = os.path.basename(os.readlink(device))
    return device_name


def statvfs(mpoints):
    """Return size/free_space for each mount point in bytes.

    :arg mpoints: list of mounting points to get info about.("/mnt/dbstorage", "/media/mpoint",..)
    :return: dict, where key is requested mount points and value is dict with total/free disc usage:
    {
       "/mnt/dbstorage": {
         "total_bytes" : 10000,
         "free_bytes" : 5000,
         "device": "/dev/sda1",
         "fstype": "ext4"
       },
       "/media/mpoint": {
         "total_bytes" : 20000,
         "free_bytes" : 1000,
         "device": "/dev/sdb2",
         "fstype": "ext3"
       },
       "/non/existing/mpoint" : None
    }
    """
    res = dict.fromkeys(mpoints)
    all_mounts = mounts()
    for fstab_entry in all_mounts:
        mpoint = fstab_entry.mpoint
        if mpoint in mpoints:
            if fact["manage_container_host"]:
                mpoint_stat = os.statvfs("/rootfs{}".format(mpoint))
            else:
                mpoint_stat = os.statvfs(mpoint)

            res[mpoint] = {}
            res[mpoint]["total_bytes"] = mpoint_stat.f_bsize * mpoint_stat.f_blocks
            res[mpoint]["free_bytes"] = mpoint_stat.f_bsize * mpoint_stat.f_bavail
            res[mpoint]["fstype"] = fstab_entry.fstype
            res[mpoint]["device"] = _extract_device_name(fstab_entry.device)
    return res


def fstab(filename=None):
    """
    Manages the contents of the /etc/fstab.
    """
    return FstabLikeDict(filename or FSTAB_PATH)


def mount(device, mpoint, raise_already_mounted=False, *args, **kwds):
    # TODO: remove with python2 drop.
    # pylint: disable=keyword-arg-before-vararg
    """
    Mounts the device to the mpoint.
    """
    cmd = MountCmd()
    cmd.args(*args, **kwds)
    cmd.args(device, mpoint)

    subprocess2.check_output_host("/bin/mkdir", "-p", mpoint)
    try:
        subprocess2.check_output_host(*cmd.build())
    except subprocess2.CalledProcessError as e:
        if "already mounted" in e.stderr:
            if raise_already_mounted:
                raise exc.AlreadyMountedError(
                    "Device {} is already mounted on {}".format(device, mpoint), device=device,
                ) from None
            return
        if "you must specify the filesystem type" in e.stderr:
            e.errorcode = ENOFS
        raise


def umount(device, *args, **kwds):
    cmd = UmountCmd().args("-f").args(*args, **kwds).args(device)
    try:
        subprocess2.check_output_host(*cmd.build())
    except subprocess2.CalledProcessError as e:
        if "not mounted" in e.stderr or "not found" in e.stderr:
            return
        raise


def _parse_int(val):
    try:
        return int(val)
    except ValueError:
        return 0


def df():
    """
    Parses df command output into DFEntry structure.

    Example of df output.
    Source:
    Filesystem     1K-blocks    Used Available Use% Mounted on
    /dev/xvda1      10484288 1565128   8386580  16% /
    Result:
    [DFEntry('/dev/xvda1', 10484288, 1565128, 8386580, 16, '/'')]
    """
    out = subprocess2.check_output(DFCmd().build())
    result = []
    words = []
    for line in out.splitlines()[1:]:
        words += line.split()
        if len(words) < 6:
            continue
        result.append(
            DFEntry(
                words[0],
                _parse_int(words[1]),
                _parse_int(words[2]),
                _parse_int(words[3]),
                _parse_int(words[4].replace("%", "")),
                words[5],
            )
        )
        words = []
    return result


def get_fstype(device):
    """
    Returns filesystem type of device.

    :param device: The path to device (e.g: /dev/sda).
    :type device: str

    :return: The type of the file system (example: ``ext4``, ``xfs``)
    :rtype: str or None
    """
    blkid_devices = blkid()
    if device in blkid_devices:
        return blkid_devices[device].fstype

    device = os.path.realpath(device)
    if device in blkid_devices:
        return blkid_devices[device].fstype


def has_partition_table(device):
    """
    Checks if device has partition table.

    :param device: The path to device (e.g: /dev/sda).
    :type device: str

    :rtype: bool
    """
    try:
        out = subprocess2.check_output(["sfdisk", "-V", "--force", device])
    except subprocess2.CalledProcessError as e:
        out = str(e)
        LOG.warning(out)
    return "unrecognized partition table" not in out


def disk_usage(path: str) -> int:
    """Return size of the file or directory in bytes."""
    if not os.path.exists(path):
        raise OSError(f"File {path} doesn't exist.")
    if os.path.isdir(path):
        return sum(f.stat().st_size for f in pathlib.Path(path).glob("**/*") if f.is_file())
    else:
        return pathlib.Path(path).stat().st_size
