import ctypes
import os
import sys

from common.utils import facts
from common.utils import subprocess2

__author__ = "spike"
__all__ = [
    "daemonize",
    "free",
    "get_boot_id",
    "losetup_all",
    "LosetupCmd",
    "modprobe",
    "ModprobeCmd",
    "reboot",
    "runlevel",
    "uptime",
    "nvme_admin_cmd",
    "NVME_IOCTL_ADMIN_CMD",
]

# constant for linux portability
_IOC_NRBITS = 8
_IOC_TYPEBITS = 8

# architecture specific
_IOC_SIZEBITS = 14
_IOC_DIRBITS = 2

_IOC_NRMASK = (1 << _IOC_NRBITS) - 1
_IOC_TYPEMASK = (1 << _IOC_TYPEBITS) - 1
_IOC_SIZEMASK = (1 << _IOC_SIZEBITS) - 1
_IOC_DIRMASK = (1 << _IOC_DIRBITS) - 1

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS

_IOC_NONE = 0
_IOC_WRITE = 1
_IOC_READ = 2


def _IOC(dir_, type_, nr, size):  # pylint: disable=invalid-name
    return (
        dir_ << _IOC_DIRSHIFT
        | type_ << _IOC_TYPESHIFT
        | nr << _IOC_NRSHIFT
        | size << _IOC_SIZESHIFT
    )


def _IO(type_, nr):  # pylint: disable=invalid-name
    return _IOC(_IOC_NONE, type_, nr, 0)


def _IOR(type_, nr, size):  # pylint: disable=invalid-name
    return _IOC(_IOC_READ, type_, nr, size)


def _IOW(type_, nr, size):  # pylint: disable=invalid-name
    return _IOC(_IOC_WRITE, type_, nr, size)


def _IOWR(type_, nr, size):  # pylint: disable=invalid-name
    return _IOC(_IOC_READ | _IOC_WRITE, type_, nr, size)


class nvme_admin_cmd(ctypes.Structure):  # pylint: disable=invalid-name
    # see https://github.com/linux-nvme/nvme-cli/blob/master/linux/nvme_ioctl.h
    _fields_ = [
        ("opcode", ctypes.c_uint8),
        ("flags", ctypes.c_uint8),
        ("rsvd1", ctypes.c_uint16),
        ("nsid", ctypes.c_uint32),
        ("cdw2", ctypes.c_uint32),
        ("cdw3", ctypes.c_uint32),
        ("metadata", ctypes.c_uint64),
        ("addr", ctypes.c_uint64),
        ("metadata_len", ctypes.c_uint32),
        ("data_len", ctypes.c_uint32),
        ("cdw10", ctypes.c_uint32),
        ("cdw11", ctypes.c_uint32),
        ("cdw12", ctypes.c_uint32),
        ("cdw13", ctypes.c_uint32),
        ("cdw14", ctypes.c_uint32),
        ("cdw15", ctypes.c_uint32),
        ("timeout_ms", ctypes.c_uint32),
        ("result", ctypes.c_uint32),
    ]


NVME_IOCTL_ADMIN_CMD = _IOWR(ord("N"), 0x41, ctypes.sizeof(nvme_admin_cmd))


class ModprobeCmd(subprocess2.Command):

    binary = "/sbin/modprobe"

    @classmethod
    def enabled(cls):
        return bool(facts.fact["kernel"]["modules"])


def modprobe(name, *args, **kwds):
    if ModprobeCmd.enabled():
        cmd = ModprobeCmd().args(*args, **kwds).args(name)
        return subprocess2.check_output(cmd.build())


modprobe.enabled = ModprobeCmd.enabled


LosetupCmd = subprocess2.new_command("LosetupCmd", binary="/sbin/losetup")


def losetup_all(flip=False):
    """Alias for losetup --all with parsing output into python dict.
    When flip=True filenames becomes keys and devices are values

    Example:
        {'/dev/loop0': '/mnt/loop0',
         '/dev/loop1': '/mnt/loop1',
         '/dev/loop2': '/mnt/loop2'}
    """
    ret = []
    cmd = LosetupCmd().args("-a")
    out = subprocess2.check_output(cmd.build()).strip()
    for line in out.splitlines():
        cols = line.split()
        device = cols[0][:-1]
        filename = cols[-1][1:-1]
        ret.append((device, filename) if not flip else (filename, device))
    return dict(ret)


def free():
    cmd = ("/usr/bin/free", "-m")
    out_lines = subprocess2.run(cmd).stdout.splitlines()
    free_ram = int(out_lines[1].split()[3])
    free_swap = int(out_lines[3].split()[3])
    return free_ram, free_swap


def runlevel():
    r_level = subprocess2.check_output("who -r | awk '{print $2}'", shell=True).strip()
    return int(r_level)


def uptime():
    """Implements `common.sysutil.uptime()` on Linux."""
    with open("/proc/uptime") as fp:
        return float(fp.read().strip().split()[0])


def daemonize():
    """
    Creates a daemon.

    Do the UNIX fork-decouple-fork magic.
    """
    # Create process fork
    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    # Decouple from parent environment
    os.chdir("/")
    os.setsid()
    os.umask(0o0022)

    # This second fork guarantees that the child is no longer a session leader (PID != SID),
    # preventing the daemon from ever acquiring a controlling terminal.
    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    # Redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    si = open(os.devnull, "r")
    so = open(os.devnull, "a+")
    se = open(os.devnull, "a+")
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


def reboot(reason=""):
    """Implements `common.sysutil.reboot()` on Linux."""
    if os.fork():
        return
    daemonize()
    subprocess2.call_host("reboot", shell=True)
    exit(0)


def get_boot_id():
    """Implements `common.sysutil.get_boot_id()` on Linux."""
    with open("/proc/sys/kernel/random/boot_id") as fp:
        return fp.read().strip()
