# pylint: disable=import-error
import codecs
import os
import socket

from schematics.types import IntType
from schematics.types import StringType

from common.types.base import Model
from common.utils import subprocess2

__all__ = [
    "systemctl_status",
    "systemctl_daemon_reload",
    "systemctl_start",
    "systemctl_stop",
    "sd_notify",
]


class SystemdServiceStatus(Model):
    """Model to store systemd service status"""

    pid = StringType()
    """Pid of the main process of the daemon

    :type: str
    """

    exit_code = IntType()
    """Exit code if main process of the daemon finished

    :type: int
    """

    active_state = StringType(required=True)
    """Active state of service unit

    :type: str
    """

    sub_state = StringType(required=True)
    """Sub state of service unit

    :type: str
    """

    result = StringType(required=True)
    """Result state of service unit

    :type: str
    """


def systemctl_status(service_name):
    """
    Get service status
    :param service_name: service name
    :return: service status
    """
    cmd = (
        "systemctl show -p ExecStart -p SubState -p ActiveState "
        "-p ExecMainStatus -p ExecMainPID -p Result %s" % service_name
    )
    status_data = subprocess2.check_output_host(cmd)
    service_status = {}
    for row in status_data.split("\n"):
        if row.startswith("SubState"):
            service_status["sub_state"] = row.split("=")[1]
        elif row.startswith("ActiveState"):
            service_status["active_state"] = row.split("=")[1]
        elif row.startswith("ExecMainStatus"):
            service_status["exit_code"] = int(row.split("=")[1])
        elif row.startswith("ExecMainPID"):
            service_status["pid"] = row.split("=")[1]
        elif row.startswith("Result"):
            service_status["result"] = row.split("=")[1]

    return SystemdServiceStatus(service_status)


def systemctl_daemon_reload():
    """Reload systemd unit files"""
    subprocess2.check_output_host("systemctl daemon-reload")


def systemctl_start(service_name):
    """
    Start systemd service
    :param service_name: service name
    """
    subprocess2.check_output_host("systemctl start %s" % service_name)


def systemctl_stop(service_name):
    """
    Stop systemd service
    :param service_name: service name
    """
    subprocess2.check_output_host("systemctl stop %s" % service_name)


def sd_notify(state):
    """
    Send a notification to systemd.
    :param str state: The notification. See
        the man page of sd_notify (http://www.freedesktop.org/software/systemd/man/sd_notify.html)
        for a description of the allowable values.
    """
    s = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    addr = os.getenv("NOTIFY_SOCKET")
    if not addr:
        return
    if addr[0] == "@":
        addr = "\0" + addr[1:]
    s.connect(addr)
    s.sendall(codecs.latin_1_encode(state)[0])
