"""
Logging utils
"""
import contextlib
import datetime
import json
import logging
import logging.config
import logging.handlers
import os
import socket
import sys
import threading
import time

import requests

from common.utils import dateutil
from common.utils import exc
from common.utils import six

DEBUG2 = logging.DEBUG - 1
DEBUG = logging.DEBUG
INFO = logging.INFO
PROGRESS = logging.INFO + 1
WARN = logging.WARNING
WARNING = logging.WARNING
ERROR = logging.ERROR


def add_level(level, name):
    def log_method(self, message, *args, **kwds):
        if self.isEnabledFor(level):
            # pylint: disable=protected-access
            self._log(level, message, args, **kwds)

    logging.addLevelName(level, name)
    setattr(logging.Logger, name.lower(), log_method)


def parse_level(level):
    if isinstance(level, int):
        return level
    elif isinstance(level, six.string_types):
        uppercase_level = level.upper()
        self_module = sys.modules[__name__]
        if hasattr(self_module, uppercase_level):
            return getattr(self_module, uppercase_level)
    raise TypeError("Unknown log level: {}".format(level))


add_level(DEBUG2, "DEBUG2")
add_level(PROGRESS, "PROGRESS")

get_logger = logging.getLogger

root = logging.root

LOG = get_logger(__name__)
LOGGING_FORMAT = "%(asctime)s | %(levelname)-7s | %(name)s: %(message)s"


def basic_config(level=None, stream=None):
    if level is not None:
        level = parse_level(level)
    logging.basicConfig(
        format=LOGGING_FORMAT, level=level or logging.DEBUG, stream=stream or sys.stderr
    )
    tweak()


def file_config(fileobj):
    logging.config.fileConfig(fileobj)
    tweak()


def tweak():
    f = LevelChangingFilter(logging.INFO, logging.DEBUG)
    for cat in (
        "requests.packages.urllib3.connectionpool",
        "oauth2client.client",
        "googleapiclient.discovery",
    ):
        logging.getLogger(cat).addFilter(f)

    # discard annoying heartbeats
    logging.getLogger("amqp").addFilter(_AMQPFilter())
    logging.getLogger("peewee").addFilter(PeeweeFilter())

    # disable trash on ipython autocomplete
    logging.getLogger("parso.cache").disabled = True
    logging.getLogger("parso.python.diff").disabled = True
    logging.getLogger("chardet.universaldetector").setLevel(logging.INFO)
    logging.getLogger("chardet.charsetprober").setLevel(logging.INFO)

    # disable flask's file watcher
    logging.getLogger("watchdog").setLevel(logging.INFO)


@contextlib.contextmanager
def lower_level(new_level=WARN):
    fltr = lambda record: record.levelno >= new_level
    try:
        for hdlr in root.handlers:
            hdlr.addFilter(fltr)
        yield
    finally:
        for hdlr in root.handlers:
            hdlr.removeFilter(fltr)


def die(message, code=1):
    """Print error ``message`` and exit with ``code``."""
    print("ERROR: {}".format(message))
    exit(code)


def set_console_log_level(category, level):
    for hdlr in get_logger(category).handlers:
        if isinstance(hdlr, logging.StreamHandler) and hdlr.stream == sys.stderr:
            hdlr.setLevel(level)


def get_process_index():
    try:
        # monitor:1            -> 1
        # tf-api               -> 0
        # not under supervisor -> 0
        return int(os.environ.get("SUPERVISOR_PROCESS_NAME", 0))
    except ValueError:
        return 0


class RotatingFileHandler(logging.handlers.RotatingFileHandler):
    """RotatingFileHandler with chmod support."""

    def __init__(self, filename, mode, maxBytes, backupCount, chmod=0o600, encoding=None):
        logging.handlers.RotatingFileHandler.__init__(
            self, filename, mode, maxBytes, backupCount, encoding
        )
        try:
            if sys.platform != "win32":
                os.chown(self.baseFilename, os.getuid(), os.getgid())
                os.chmod(self.baseFilename, chmod)
        except OSError:
            pass


class LevelChangingFilter(object):
    """Filter changing log level ``from_level`` into ``to_level``."""

    def __init__(self, from_level, to_level):
        self.from_level = from_level
        self.to_level = to_level

    def filter(self, record):
        if record.levelno == self.from_level:
            record.levelno = self.to_level
            record.levelname = logging.getLevelName(record.levelno)
        return logging.getLogger(record.name).isEnabledFor(record.levelno)


class PeeweeFilter:
    """Internal filter for Peewee logs."""

    max_msg_param_len = 64

    @staticmethod
    def check_max_length(params: list):
        """Count total length of Peewee logs outputted fields

        :param params: Peewee logs outputted fields
        """
        log_len_sum = 0
        for param in params:
            if isinstance(param, (str, bytes)):
                log_len_sum += len(param)
        return log_len_sum

    def filter(self, record):
        if isinstance(record.msg, tuple):
            sql, params = record.msg
            if params and self.check_max_length(params) > self.max_msg_param_len * 4:
                record.msg = (
                    sql,
                    [
                        f"{p[:29]}..."
                        if isinstance(p, (bytes, str)) and len(p) > self.max_msg_param_len
                        else p
                        for p in params
                    ],
                )
        return True


class _AMQPFilter(object):
    """Internal filter for AMQP heartbeat logs."""

    def filter(self, record):
        if isinstance(record.msg, six.string_types):
            return not record.msg.startswith("heartbeat_tick")
        return True


class NoStacktraceFormatter(logging.Formatter):
    # Moved from scalarizr/util/log.py@d8a8384
    """Formatter removing stacktrace.

    For logs, where showing programming internals is scary.
    """

    def formatException(self, ei):
        return ""


class UtcOffsetFormatter(logging.Formatter):
    # Moved from scalarizr/util/log.py@d8a8384
    """time.strftime("%Y-%m-%d %H:%M:%S %z", time.localtime()) doesn't work in Python < 3.3. See:

    http://bugs.python.org/issue1493676
    http://bugs.python.org/issue1667546

    So we cant just specify a custom `datefmt` paramete r with the `%z` option in the logging config,
    but need to create a separate formatter and compute the offset themselves,
    taking into account various edge cases:

    http://bz.selenic.com/show_bug.cgi?id=2511"""

    def formatTime(self, record, datefmt=None):
        original = super(UtcOffsetFormatter, self).formatTime(record, datefmt)

        delta = datetime.datetime.utcfromtimestamp(record.created) - (
            datetime.datetime.fromtimestamp(record.created)
        )
        offset_seconds = delta.days * 86400 + delta.seconds

        sign = "-" if offset_seconds > 0 else "+"
        utc_offset = time.strftime("%H:%M", time.gmtime(abs(offset_seconds)))

        return "".join([original, sign, utc_offset])


class DebugFormatter(UtcOffsetFormatter):
    # Moved from scalarizr/util/log.py@d8a8384
    pass


class UserFormatter(UtcOffsetFormatter, NoStacktraceFormatter):
    # Moved from scalarizr/util/log.py@d8a8384
    pass


class BaseBackend(object):
    """This is a base class for log backends (logstash/fluentd)."""

    PROTO_HTTPS = "https"
    PROTO_HTTP = "http"
    PROTO_UNIX = "unix"
    PROTO_TCP = "tcp"
    PROTO_UDP = "udp"

    connected = None
    """Objects, that need notification about reconnect to backend, should .wait() this event.

    :type: :class:`threading.Event`
    """

    def __init__(self, url, timeout=3.0, reconnect_interval=0.5, auth=None, ssl_verify=True):
        """Allows to send data to backend via: http, unix socket and tcp/udp

        :param url: http://{net_loc}:{port}/{path}, unix://{path}}, tcp(udp)://{net_loc}:{port}
        :type url: str

        :param timeout: timeout for connect and send
        :type timeout: float

        :param auth: Username and password should use to authenticate to logstash/fluentd.
        :type username: tuple

        :param ssl_verify: Sets agent to verify ssl.
        :type ssl_verify: bool
        """
        self.url = url
        self.timeout = timeout
        self.socket = None
        self.reconnect_interval = reconnect_interval

        self._lock = threading.Lock()
        self._parse_connection_url()
        self._emit = (
            self._send_http
            if self._proto in (self.PROTO_HTTP, self.PROTO_HTTPS)
            else self._send_socket
        )
        self._reconnect_thread = None
        self.connected = threading.Event()

        self.auth = auth
        self.ssl_verify = ssl_verify

    def wait_connected(self, timeout):
        """Method to wait `timeout` seconds for event"""
        self.connected.wait(timeout)

    def _reconnect(self):
        args = (self.__class__.__name__, self.url, self.reconnect_interval)
        LOG.info("%s reconnecting (%s) in loop with interval %ss.", *args)
        while True:
            try:
                with self._lock:
                    self._socket_connect()
                self._reconnect_thread = None
                LOG.info("%s connection restored.", self.__class__.__name__)
                break
            except socket.error:
                time.sleep(self.reconnect_interval)

    def _parse_connection_url(self):
        res = six.moves.urllib_parse.urlparse(self.url)
        self._proto = res.scheme
        self._host = res.netloc.split(":")[0]
        self._port = res.port

    def _close(self):
        if self.socket:
            try:
                self.socket.close()
            except:  # pylint: disable=bare-except
                pass
            finally:
                self.socket = None
                self.connected.clear()

    def _socket_connect(self):
        if self.socket:
            return
        if self._proto == self.PROTO_UNIX:
            unix_socket_path = self.url[len("unix://") :]
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock_conn_args = unix_socket_path
        elif self._proto == self.PROTO_TCP:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock_conn_args = (self._host, self._port)
        elif self._proto == self.PROTO_UDP:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock_conn_args = (self._host, self._port)
        else:
            raise Exception("Unsupported protocol: {}".format(self._proto))

        sock.settimeout(self.timeout)
        sock.connect(sock_conn_args)
        self.socket = sock
        self.connected.set()

    def _format_data(self, data):
        return json.dumps(data).encode("utf-8")

    def _send_socket(self, data):
        with self._lock:
            if self._reconnect_thread is not None:
                return False
            try:
                self._socket_connect()
                data_to_send = self._format_data(data) + b"\r\n"
                self.socket.sendall(data_to_send)
            except socket.error:
                self._close()
                self._reconnect_thread = threading.Thread(target=self._reconnect)
                self._reconnect_thread.daemon = True
                self._reconnect_thread.start()
                raise
        return True

    def _prepare_request_kwargs(self, data):
        kwargs = dict(json=data, timeout=self.timeout)
        if self.auth:
            kwargs["auth"] = self.auth
        if self._proto == self.PROTO_HTTPS:
            kwargs["verify"] = self.ssl_verify
        return kwargs

    def _send_http(self, data):
        resp = requests.post(self.url, **self._prepare_request_kwargs(data))
        resp.raise_for_status()
        return True

    def emit(self, data):
        """
        Sends data to backend via _send_socket or _send_http method

        :param data: data to send
        :type data: dict

        :return: bool
        """
        return self._emit(data)


class LogstashBackend(BaseBackend):
    """Sends log to logstash."""

    def emit(self, data):
        """
        Sends data to Logstash backend via _send_socket or _send_http method

        :param data: data to send
        :type data: dict

        :return: bool
        """
        logstash_data = {
            "@source": "Scalr",
            "@tags": data["extra"]["tags"],
            "@fields": data["extra"],
            "@timestamp": data["extra"]["timestamp"],
        }
        return super(LogstashBackend, self).emit(logstash_data)


class FluentdBackend(BaseBackend):
    """Sends logs to fluentd."""

    def _send_socket(self, data):
        tag = "{tag}.{event}".format(**data)
        _data = [tag, data["extra"]["timestamp"], data["extra"]]
        return super(FluentdBackend, self)._send_socket(_data)

    def _send_http(self, data):
        url = "{}/{}".format(self.url, data["tag"] or "")
        resp = requests.post(url, **self._prepare_request_kwargs(data))
        resp.raise_for_status()
        return True


class ScalrLogger(object):
    """Scalr Logger protocol implementation."""

    name_to_backend = {"logstash": LogstashBackend, "fluentd": FluentdBackend}

    def __init__(self, url, tag, extra=None, backend=None, auth=None, ssl_verify=True, timeout=3.0):
        """
        :param url: http://{net_loc}:{port}/{path}, unix://{path}}, tcp(udp)://{net_loc}:{port}
        :type url: str

        :param tag: logger tag
        :type tag: str

        :param extra: default extra data
        :type extra: dict

        :param backend: backend type, possible ['fluentd', 'logstash'], default fluentd
        :type backend: str

        :param auth: Username and password should use to authenticate to logstash/fluentd.
        :type username: tuple

        :param ssl_verify: Sets agent to verify ssl.
        :type ssl_verify: bool
        """
        backend = backend or "fluentd"
        try:
            backend_cls = ScalrLogger.name_to_backend[backend]
        except KeyError:
            six.reraise(
                exc.NotSupportedError(
                    "Cannot instantiate ScalrLogger with unsupported backend {}".format(backend)
                ),
                exc.NotSupportedError,
                sys.exc_info()[2],
            )
        self.backend = backend_cls(url, auth=auth, ssl_verify=ssl_verify, timeout=timeout)
        self.tag = tag
        self.extra = extra or {}

    def wait_backend_connected(self, timeout):
        """wrapper for `backend.wait_connected` method"""
        self.backend.wait_connected(timeout)

    def _serialize(self, prefix, **kwargs):
        """Recursive method which serializes **kwargs data into Scalr Log Record."""
        data = {}
        for key, value in kwargs.items():
            key = ".".join((prefix, str(key).lower()))
            if isinstance(value, dict):
                data.update(self._serialize(key, **value))
                continue
            data[key] = value
        return data

    def emit(self, record_event_name, **kwargs):
        """Serializes event **kwargs data and sends it to backend.

        :return: True, when event was delivered to backend.
        :rtype: bool
        """
        try:
            extra = self._serialize(record_event_name, **kwargs)
            extra["tags"] = [self.tag, record_event_name]
            extra["timestamp"] = dateutil.get_utc_now_formatted()
            extra.update(self.extra)

            data = {
                "tag": self.tag,
                "event": record_event_name,
                "extra": extra,
            }

            return self.backend.emit(data)
        except Exception as e:
            LOG.warning("Failed to send Scalr log event: %s", str(e))
            return False


class LazyJoin:
    """
    Join which allows to use iterative elements in logging function in the lazy manner.
    """

    def __init__(self, s, items):
        """
        :param s: Separator between elements in iterable.
        :param items: Iterable which contains strings.
        """
        self.s = s
        self.items = items

    def __str__(self):
        """
        :return: A string which is the concatenation of the strings in the
        iterable.
        """
        return self.s.join(self.items)
