"""
A collection of useful utilities
"""
from __future__ import absolute_import

import abc
import binascii
import contextlib
import copy
import distutils.version
import enum
import functools
import hashlib
import itertools
import logging
import re
import secrets  # cryptographically secure random bytes
import sys
import threading
import time
import typing as tp
import uuid as _uuid

import monotonic
import requests.exceptions
import semver

from common.utils import exc
from common.utils import logutil
from common.utils import six

try:
    import agent.config as config
except ImportError:
    import server.config as config


LOG = logutil.get_logger(__name__)

UUID_RE = r"^[A-Fa-f0-9]{8}-([A-Fa-f0-9]{4}-){3}[A-Fa-f0-9]{12}$"


def error_logged(preface, reraise=True):
    """Decorator to wrap a function, any unhandled exception in which, should be logged,
    and (optionally) propagated further.
    """

    def decorator(func):
        @six.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except (SystemExit, KeyboardInterrupt):  # pylint: disable=try-except-raise
                raise
            except BaseException as e:
                t, e, _ = sys.exc_info()
                LOG.exception("%s: %s: %s", preface, t.__name__, e)
                if reraise:
                    raise

        return wrapper

    return decorator


def try_until(
    func,
    try_msg=None,
    error_msg=None,
    log=None,
    interval: tp.Union[int, tp.Iterable] = 1,
    timeout=0,
    times=None,
):
    """
    repeat call func while it returns False
    raises exc.TimeoutError if timeout expired or call times reached
    """
    log = log or LOG
    is_interval_iterated = isinstance(interval, tp.Iterable)
    if is_interval_iterated:
        interval = iter(interval)
    timeout = timeout or config.workflow_engine["global_timeout"]
    begin_msg = "Trying {0} until (timeout: {1} interval: {2} times: {3})".format(
        func, timeout, "iterable" if is_interval_iterated else interval, times or "unlimited"
    )
    try_msg = try_msg or "{0} returns false".format(func)
    error_msg = error_msg or "Try {0} Failed!".format(try_msg)

    start_time = monotonic.monotonic()
    log.debug(begin_msg)
    for num in itertools.count(1):
        log.debug("%s (%s) ...", try_msg, num)
        try:
            result = func()
            if result:
                return result

            sleep_period = next(interval) if is_interval_iterated else interval
        except StopIteration:
            msg = "{0}: exceeded interval list".format(error_msg)
            raise exc.TimeoutError(msg)
        except Exception as e:
            msg = "{0}: got error: {1}".format(error_msg, e)
            six.reraise(exc.Error, exc.Error(msg), sys.exc_info()[2])
        else:
            if monotonic.monotonic() - start_time > timeout:
                msg = "{0}: timeout {1} seconds exceeded".format(error_msg, timeout)
                raise exc.TimeoutError(msg)
            if times and num >= times:
                msg = "{0}: call count {1} times exceeded".format(error_msg, times)
                raise exc.TimeoutError(msg)
        log.debug("Wait {:.2f} seconds before the next attempt".format(sleep_period))
        time.sleep(sleep_period)


def _get_fqname(obj):
    """Get object fully qualified name(module + name)."""
    qualname = obj.__qualname__ if hasattr(obj, "__qualname__") else obj.__name__
    return ".".join((obj.__module__, qualname))


def retries_error(
    except_expr,
    condition_func=None,
    times=1,
    interval=1,
    except_log_level=logutil.DEBUG,
    error_message=None,
    silent=False,
):
    """
    Retries function when a specific error occurs.

    :param except_expr: Error(s) to retry.
    :type except_expr: Exception class or tuple of exception classes.
    :param func condition_func: Optional callback that can refine error to retry, because sometimes
        exceptions are too broad. Should take exception instance as argument and return bool.
    :param int times: Number of retires. If 0, retry infinitely.
    :param interval: Interval between retries in seconds.
    :type interval: int or iterable of ints.
    :param except_log_level: write log messages with specified level.
    :param error_message: The message that will be displayed when a function fails to call. Defaults
        to None.
    :param silent: Disable logging.

    :decorator:
    """

    def inner(func):
        @six.wraps(func)
        def wrapper(*args, **kwargs):
            intervals = iter((interval,) if isinstance(interval, int) else interval)
            countdown = 1
            for num in itertools.count(1):
                try:
                    countdown = next(intervals)
                except StopIteration:
                    # keep last interval
                    pass
                try:
                    if not silent:
                        LOG.debug(
                            'Calling "%s" (retry: %s/%s) ...',
                            _get_fqname(func),
                            num,
                            times or "infinity",
                        )
                    return func(*args, **kwargs)
                except except_expr as e:
                    if condition_func and not condition_func(e):
                        raise

                    should_fail = times and num >= times
                    if not silent:
                        msg = error_message or 'Error calling "{0}"'.format(_get_fqname(func))
                        recap = (
                            "Max retries reached."
                            if should_fail
                            else "Trying again in {} seconds...".format(countdown)
                        )
                        LOG.log(except_log_level, "%s: %s. %s", msg, requests_exc_to_str(e), recap)
                    if should_fail:
                        raise

                time.sleep(countdown)

        return wrapper

    return inner


def timeit(func):
    """
    Measure execution time of small code snippets
    """

    @six.wraps(func)
    def wrapper(*args, **kwds):
        time0 = time.time()
        try:
            return func(*args, **kwds)
        finally:
            LOG.info("{0!r} finished in {1:.8f}s".format(func.__name__, time.time() - time0))

    return wrapper


class Observable(object):
    """Implementation of the Observable design pattern"""

    def __init__(self):
        self._listeners = {}
        self.suspended = False

    def __iter__(self):
        return iter(self._listeners.keys())

    def fire(self, event: tp.Union[str, tp.Tuple], *args, **kwargs):
        """Fire event.

        All registered listeners will be called as ``ln(*args, **kwargs)`` in a loop

        :param event: Event name (or any key that used as event) to trigger
        """
        if not self.suspended:
            lns = copy.copy(self._listeners.get("*", [])) + copy.copy(
                self._listeners.get(event, [])
            )
            for ln in lns:
                ln(*args, **kwargs)

    def on(self, *args, **kwargs):
        """Add listener

        Attach one or many listeners to a single event::
            obj.on('event_name', func_1, func_2)

        Attach listeners for many events::
            obj.on(add=func_1, remove=func_2, change=func_3)
        """
        if len(args) >= 2:
            event = args[0]
            if event not in self._listeners:
                self._listeners[event] = []
            for listener in args[1:]:
                if listener not in self._listeners[event]:
                    self._listeners[event].append(listener)
        elif kwargs:
            for event in kwargs.keys():
                self.on(event, kwargs[event])

    def un(self, event, listener):
        """Remove listener"""
        if event in self._listeners and listener in self._listeners[event]:
            self._listeners[event].remove(listener)

    def suspend(self):
        """Suspend events.

        :meth:`~Observable.fire` won't trigger listeners.
        """
        self.suspended = True

    def resume(self):
        """Resume events, previously suspended with :meth:`~Observable.suspend`."""
        self.suspended = False


class Proxy(object):
    """Implementation of the Proxy design pattern."""

    def __init__(self, initiator):
        object.__setattr__(self, "__factory", initiator)
        object.__setattr__(self, "__instance", None)
        object.__setattr__(self, "__initialized", None)

    def _lazy_init(self):
        if not object.__getattribute__(self, "__initialized"):
            factory = object.__getattribute__(self, "__factory")
            instance = factory() if hasattr(factory, "__call__") else factory
            object.__setattr__(self, "__instance", instance)
            object.__setattr__(self, "__initialized", True)

    def __getattribute__(self, name):
        object.__getattribute__(self, "_lazy_init")()
        return getattr(object.__getattribute__(self, "__instance"), name)

    def __delattr__(self, name):
        object.__getattribute__(self, "_lazy_init")()
        delattr(object.__getattribute__(self, "__instance"), name)

    def __setattr__(self, name, value):
        object.__getattribute__(self, "_lazy_init")()
        setattr(object.__getattribute__(self, "__instance"), name, value)

    def __call__(self, *args, **kwargs):
        # FIXME: this does everyone callable
        object.__getattribute__(self, "_lazy_init")()
        return object.__getattribute__(self, "__instance")(*args, **kwargs)


class DictProxy(dict):
    """
    Implements attribute access to dictionary.
    """

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            # to conform with __getattr__ spec
            raise AttributeError(key)

    def __setattr__(self, key, value):
        self[key] = value

    # def __delattr__(self, key):
    #     try:
    #         del self[key]
    #     except KeyError:
    #         raise AttributeError(key)


def sortable_uuid(prior_seconds=0) -> _uuid.UUID:
    """Generate sortable UUID.
    `Spec: <https://github.com/Scalr/int-scalr/blob/master/docs/coding_standards/ids.rst>`_
    """
    version = "6"
    now = str(time.time() - prior_seconds)
    nowhex = format(int(now[:10] + now[11:17].ljust(8, "0")), "x")
    rand = binascii.hexlify(secrets.token_bytes(8)).decode()
    return _uuid.UUID(
        f"{nowhex[:8]}-{nowhex[8:12]}-{version}{nowhex[12:16]}-{rand[:4]}-{rand[4:16]}"
    )


def generate_uid(prefix=None, length=8):
    """
    Generates unique identifier.
    """
    uid = _uuid.uuid4().hex[0:length]
    return uid if prefix is None else "{}{}".format(prefix, uid)


class Singleton(type):
    """Implementation of the Singleton design pattern."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


# pylint: disable=invalid-name
@contextlib.contextmanager
def empty_context_manager():
    """Empty context manager that does nothing. Useful when
    we need to apply some CM only based on condition.
    """
    yield


# pylint: disable=protected-access
def memorize(threadsafe=False):
    """Memorize decorator with optional support for multithreaded access.

    :param bool threadsafe: If enabled, it locks subsequent calls
        until the first one returns a value.

    :decorator:
    """

    def inner(func):
        lock = threading.Lock() if threadsafe else None

        @six.wraps(func)
        def wrapper(*args, **kwds):
            with lock if threadsafe else empty_context_manager():
                if not hasattr(func, "_memorized"):
                    func._memorized = func(*args, **kwds)
            return func._memorized

        return wrapper

    return inner


def ellipsis(value, length=None):
    """
    appends ellipsis to an object.
    """
    length = length or config.workflow_engine["ellipsis_length"]
    # pylint: disable=no-member
    if isinstance(value, (six.string_types, six.binary_type)):
        if len(value) < length:
            return value
        if isinstance(value, six.string_types):
            return value[0:length] + "..."
        else:
            return value[0:length] + b"..."
    elif isinstance(value, six.moves.Sequence):
        return tuple(ellipsis(val) for val in value)
    elif isinstance(value, six.moves.Mapping):
        return {key: ellipsis(val) for key, val in value.items()}
    else:
        return value


def validate_url(url):
    """Validates HTTP URL by scheme"""
    parse_result = six.moves.urllib.parse.urlparse(url)
    if parse_result.scheme not in ("http", "https"):
        raise exc.MalformedError('Invalid URL: Bad scheme is provided: "%s"' % parse_result.scheme)
    if not parse_result.hostname:
        raise exc.MalformedError("Invalid URL: No hostname provided")


class Version(distutils.version.LooseVersion):
    """
    Extended :class:`distutils.version.LooseVersion` with ability to compare
    versions specified by tuple / list.

    Example:
        >>> Version(debian_version) >= (7, 4)
    """

    VersionUnion = tp.Union[
        "Version", str, int, tp.Tuple[tp.Union[str, int], ...], tp.List[tp.Union[str, int]],
    ]

    def __init__(self, vstring: tp.Optional[str] = None, vtuple: tp.Optional[tuple] = None) -> None:
        self.version: tp.Tuple[tp.Union[str, int], ...] = tuple()
        self.vstring: str = ""

        if vtuple is not None:
            if vstring is not None:
                raise ValueError("'vstring' and 'vtuple' arguments are mutually exclusive")
            self.version = tuple(vtuple)
            self.vstring = ".".join(map(str, self.version))

        super(Version, self).__init__(vstring=vstring)

    def _cmp(self, other: tp.Any) -> int:  # type: ignore
        if not isinstance(other, Version):
            if isinstance(other, str):
                other = Version(vstring=other)
            elif isinstance(other, int):
                other = Version(vtuple=(other,))
            elif isinstance(other, (tuple, list)):
                other = Version(vtuple=tuple(other))
            else:
                raise TypeError(f"Unable to compare {self.__class__} with {type(other)}")

        try:
            return super(Version, self)._cmp(other)
        except TypeError as err:
            raise TypeError(
                ("Unable to compare the version {} with the other " "version {}: {}").format(
                    self.version, other.version, str(err)
                )
            )

    def parse(self, vstring: str) -> None:  # type: ignore
        self.vstring = vstring
        components: tp.List = [
            item for item in self.component_re.split(vstring) if item and item not in (".", "-")
        ]
        for index, item in enumerate(components):
            try:
                components[index] = int(item)
            except ValueError:
                pass
        self.version = tuple(components)

    # pylint: disable=useless-super-delegation
    def __lt__(self, other: VersionUnion) -> bool:
        return super(Version, self).__lt__(other)  # type: ignore

    def __le__(self, other: VersionUnion) -> bool:
        return super(Version, self).__le__(other)  # type: ignore

    def __gt__(self, other: VersionUnion) -> bool:
        return super(Version, self).__gt__(other)  # type: ignore

    def __ge__(self, other: VersionUnion) -> bool:
        return super(Version, self).__ge__(other)  # type: ignore

    # pylint: enable=useless-super-delegation


class SemVersion(semver.VersionInfo):
    """
    Extended :class:`semver.VersionInfo` with constructor, which accepts string version and
    possibility to compare with string representation of semantic version.

    Example:
        >>> version = SemVersion("1.3.4")
        >>> version
        SemVersion(major=1, minor=3, patch=4, prerelease=None, build=None)
        >>> version > "1.3.2"
        True
        >>> version < "1.3.1"
        False
        >>> version == "1.3.4"
        True
    """

    comparable_types = (semver.VersionInfo, dict, tuple, str)
    CompTypes = tp.Union[comparable_types]

    def __init__(self, version: str) -> None:
        """
        :param version: String representation of semantic version.
        :raises: class: ValueError if version is not of a valid type.
        """
        super().__init__(**semver.parse(version))

    @classmethod
    def _convert_to_supported(cls, value: tp.Any) -> "CompTypes":
        """
        Converts value to one of the supported types.

        :raises: class: TypeError if value is not of supported types.
        """
        if not isinstance(value, cls.comparable_types):
            raise TypeError(
                f"Impossible to compare {type(cls)!r} and {type(value)!r}. "
                f"Types must be in {cls.comparable_types!r}."
            )
        if isinstance(value, str):
            value: semver.VersionInfo = semver.parse_version_info(value)
        return value

    def __eq__(self, other):
        return super().__eq__(self._convert_to_supported(other))

    def __ne__(self, other):
        return super().__ne__(self._convert_to_supported(other))

    def __lt__(self, other):
        return super().__lt__(self._convert_to_supported(other))

    def __le__(self, other):
        return super().__le__(self._convert_to_supported(other))

    def __gt__(self, other):
        return super().__gt__(self._convert_to_supported(other))

    def __ge__(self, other):
        return super().__ge__(self._convert_to_supported(other))


class classproperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, instance, owner):
        return self.fget(owner)


def map_recursive(dct, func):
    """
    Maps function to dictionary values recursively

    :param dict dct: Dict to be applied to a function
    :param function func: mapping function
    :return: dict
    """
    for k, v in dct.items():
        if isinstance(v, dict):
            dct[k] = map_recursive(v, func)
        else:
            dct[k] = func(v)
    return dct


def merge_recursive(dst, src):
    """Recursive merge two dicts (vs .update which overwrites the hashes at the
    root level)
    Note: This updates dst.
    Copied from checkmate.utils
    """
    stack = [(dst, src)]
    while stack:
        current_dst, current_src = stack.pop()
        for key in current_src:
            source = current_src[key]
            if key not in current_dst:
                current_dst[key] = source
            else:
                dest = current_dst[key]
                if isinstance(source, dict) and isinstance(dest, dict):
                    stack.append((dest, source))
                elif isinstance(source, list) and isinstance(dest, list):
                    # Make them the same size
                    r = dest[:]
                    s = source[:]
                    if len(dest) > len(source):
                        s.extend([None for i in range(len(dest) - len(source))])
                    elif len(dest) < len(source):
                        r.extend([None for i in range(len(source) - len(dest))])
                    # Merge lists
                    for index, value in enumerate(r):
                        if (not value) and s[index]:
                            r[index] = s[index]
                        elif isinstance(value, dict) and isinstance(s[index], dict):
                            stack.append((dest[index], source[index]))
                        else:
                            dest[index] = s[index]
                    current_dst[key] = r
                else:
                    current_dst[key] = source
    return dst


if six.PY2:
    _find_unsafe = re.compile(r"[^\w@%+=:,./-]").search

    def quote(value):
        """
        Return a shell-escaped version of the string.
        :param str value: string to quote
        :return: str
        """
        if not value:
            return "''"
        if _find_unsafe(value) is None:
            return value

        # use single quotes, and put single quotes into double quotes
        # the string $'b is then quoted as '$'"'"'b'
        return "'" + value.replace("'", "'\"'\"'") + "'"


else:
    import shlex

    # pylint: disable=no-member
    quote = shlex.quote


def chunks(iterable, chunk_size):
    """Slice iterable into chunks"""
    if isinstance(iterable, set):
        iterable = list(iterable)
    if isinstance(iterable, (list, tuple, six.string_types, six.binary_type)):
        for i in six.moves.xrange(0, len(iterable), chunk_size):
            yield iterable[i : i + chunk_size]
    else:
        while True:
            chunk = list(itertools.islice(iterable, 0, chunk_size))
            if not chunk:
                break
            yield chunk


def firstmatched(func, iterable):
    """
    Returns first item in `iterable`, for which `fn(item)` returns True.
    """
    for item in iterable:
        if func(item):
            return item
    return None


def inverted_dict(d):
    """
    Inverts a dictionary with value as a key of original dictionary.

    :type d: dict
    :rtype: dict
    """
    return {v: k for k, v in d.items()}


def camelize(underscored, upper=False):
    """
    Transforms string with underscores params/words to camel case

    :param str underscored: string with underscores

    :param bool upper: if True will produce UpperCamelCase, otherwise lowerCamelCase.

    :returns string: camel-cased string
    """
    words = underscored.split("_")
    first = words[0].capitalize() if upper else words[0]
    rest = [x.capitalize() for x in words[1:]]
    rest.insert(0, first)
    return "".join(rest)


def camel_to_underscore(camelized):
    """
    Transforms camel-case string to underscore-case.

    :param str camelized: CamelCase string

    :return string: underscored string.
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", camelized)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


class NestedDict(dict):
    """
    Dict extension that works as defaultdict, but add possibility to control of dict depth
    """

    def __init__(self, *args, **kwargs):
        self._depth = kwargs.pop("_depth", None)
        self._last_return_func = kwargs.pop("_last_return_func", {})
        super(NestedDict, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        if self._depth is None:
            self[key] = NestedDict()
        elif self._depth == 0:
            if callable(self._last_return_func):
                self[key] = self._last_return_func()
            else:
                self[key] = self._last_return_func
        else:
            self[key] = NestedDict(_depth=self._depth - 1, _last_return_func=self._last_return_func)

        return self[key]


@six.add_metaclass(abc.ABCMeta)
class ABCCache(object):
    """
    Base class for cached dict like object that have ttl for each values
    """

    class NONE:
        pass

    @abc.abstractmethod
    def __setitem__(self, key, value):
        pass

    @abc.abstractmethod
    def __getitem__(self, key):
        pass

    @abc.abstractmethod
    def invalidate(self, key):
        """
        invalidate key from cache
        :param key: key value
        :return:
        """

    def get(self, key, default=None):
        """
        Return the value for key if key is in dict like object, else default. If default is not given,
        it defaults to None, so that this method never raises a KeyError.
        """
        try:
            return self[key]
        except KeyError:
            return default


class DictCache(dict, ABCCache):
    """
    Cached dict like object that have ttl for each values
    """

    capacity_limit = 10000

    def __init__(self, ttl=0):
        super(DictCache, self).__init__()
        self.ttl = ttl

    def __setitem__(self, key, value):
        now = monotonic.monotonic()
        super(DictCache, self).__setitem__(key, (now, value))

    def __getitem__(self, key):
        last_update, value = super(DictCache, self).__getitem__(key)
        if self._is_expired(last_update):
            del self[key]
            raise KeyError(key)
        return value

    def invalidate(self, key):
        """
        Invalidate value from cache by given key
        :param key: key value
        """
        try:
            del self[key]
        except KeyError:
            pass

    def _validate_cache(self):
        if len(self) >= self.capacity_limit:
            for key, value in list(self.items()):
                last_update = value[0]
                if self._is_expired(last_update):
                    del self[key]
            if len(self) >= self.capacity_limit:
                LOG.warning("Cleaning cache is unsatisfying. Reset all cache.")
                for key in list(self.keys()):
                    del self[key]

    def _is_expired(self, last_update):
        if self.ttl:
            now = monotonic.monotonic()
            return now - last_update > self.ttl
        return False

    def get(self, key, default=None):
        """
        Return the value for key if key is in dict like object, else default. If default is not given,
        it defaults to None, so that this method never raises a KeyError.
        """
        self._validate_cache()
        try:
            return self[key]
        except KeyError:
            return default


class timed_cache(object):
    """
    Memorizing decorator, caches function's return value, expires after ttl time in second.
    """

    def __init__(self, ttl=3, cache_backend=DictCache, classless=False, threadsafe=False):
        """
        :param ttl: cache value time to live
        :param cache_backend: cache backend
        :param classless: if True remove class object from cache key format
        """
        self._cache = cache_backend(ttl)
        self.ttl = ttl
        self.classless = classless
        self.threadsafe = threadsafe
        if threadsafe:
            self._lock = threading.Lock()

    def invalidate(self, func_name, *args, **kwargs):
        """
        Invalidate value from cache
        :param func_name: fully qualified name(module + name)
        :param args: function args
        :param kwargs: function kwargs
        """
        key = self._md5(func_name + str(args) + str(kwargs))
        self._cache.invalidate(key)

    def _md5(self, value):
        return hashlib.md5(value.encode()).hexdigest()

    def __call__(self, func):
        @six.wraps(func)
        def wrapper(*args, **kwargs):
            args_str = str(args[1:]) if self.classless else str(args)
            key = self._md5(_get_fqname(func) + args_str + str(kwargs))
            with self._lock if self.threadsafe else empty_context_manager():
                value = self._cache.get(key, self._cache.NONE)
                if value is self._cache.NONE:
                    value = func(*args, **kwargs)
                    self._cache[key] = value
                return value

        return wrapper


_textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})


def guess_binary(bs):
    """
    Returns true when input is not an ascii / utf-8 text.

    :param bytes bs: bytes to analyze
    """
    return bool(bs.translate(None, _textchars))


def build_url(scheme, host, port=None, user=None, password=None) -> str:
    """
    :param str scheme:
    :param str host:
    :param str port:
    :param str user:
    :param str password:
    :return:
    :rtype: str
    """
    user_pass = ""
    if user:
        password = ":{}".format(password) if password else ""
        user_pass = "{user}{password}@".format(user=user, password=password)
    port = ":{port}".format(port=port) if port else ""
    return "{scheme}://{user_pass}{host}{port}".format(
        scheme=scheme, user_pass=user_pass, host=host, port=port
    )


def requests_exc_to_str(exception):
    """
    Transforms `exception` instance to string representation

    :param Exception exception:
    :return:
    :rtype: str
    """
    exc_str = str(exception)
    if not isinstance(exception, requests.exceptions.RequestException):
        return exc_str
    regexp = r".* *(<urllib3.connection.(.*)HTTP(S?)Connection object at .*>:? *).*"
    match = re.match(regexp, exc_str)
    if match:
        return exc_str.replace(match.groups()[0], "")
    return exc_str


def int_to_bytes(n, length, byteorder="little"):
    """
    Converts int type to bytes.
    """
    if not isinstance(n, int):
        n = int(n)
    if n < 0:
        n += 1 << (8 * length)
    if six.PY3:
        # pylint: disable=no-member
        return n.to_bytes(length, byteorder=byteorder)
    h = "%x" % n
    s = ("0" * (len(h) % 2) + h).zfill(length * 2).decode("hex")
    return s if byteorder == "big" else s[::-1]


def pipeline(func):
    """Syntax sugar to avoid next(gen) when creating generator that waits for value"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        coro = func(*args, **kwargs)
        next(coro)
        return coro

    return wrapper


def is_uuid(value: str) -> bool:
    """
    Checks that value is valid `UUID` string.
    """
    regexp = re.compile(UUID_RE, re.IGNORECASE)
    return bool(regexp.match(value))


def is_jwe_token(value: str) -> bool:
    """
    Checks that a value corresponding to the JWE Compact Serialization format.

    see: https://tools.ietf.org/html/rfc7516#section-7.1
    """
    return value and value.count(".") == 4


def is_semver(value: str) -> bool:
    """
    Checks that a value corresponding to the Semantic Versioning format.

    see: https://semver.org
    """
    return semver.VersionInfo.isvalid(value)


@contextlib.contextmanager
def time_report(message, log=LOG, level=logging.DEBUG):
    """
    Reports the execution time of its code block.

    :param str message: Message pattern with one %s for execution time substitution.
    :param logging.Logger log: Logger.
    :param int level: Report logging level.
    """
    start_time = time.monotonic()
    try:
        log.log(level, "%s ...", message)
        yield
    finally:
        execution_time = time.monotonic() - start_time
        log.log(level, "%s finished in %.6f seconds.", message, execution_time)


# Base32 encoding/decoding must be done in Python
_b32hexalphabet = b"0123456789abcdefghijklmnopqrstuv"
_b32hextab2 = None


def _base32_hex_encode(s):
    """
    Encode the bytes-like object s using Base32hex and return a bytes object.
    See: `Implementation for base32
    <https://github.com/python/cpython/blob/b64c2c66e5cfe6d138b342ee58ee3b13a8d7ef16/Lib/base64.py#L137-L168>`_

    :param bytes|bytearray|str s:
    :return: Encoded string
    :rtype: bytes
    """
    global _b32hextab2

    if _b32hextab2 is None:
        b32hextab = [bytes((i,)) for i in _b32hexalphabet]
        _b32hextab2 = [a + b for a in b32hextab for b in b32hextab]
        b32hextab = None

    if not isinstance(s, (bytes, bytearray)):
        s = memoryview(s).tobytes()
    leftover = len(s) % 5
    # Pad the last quantum with zero bits if necessary
    if leftover:
        s = s + b"\x00" * (5 - leftover)  # Don't use += !
    encoded = bytearray()
    from_bytes = int.from_bytes
    b32hextab = _b32hextab2
    for i in range(0, len(s), 5):
        c = from_bytes(s[i : i + 5], "big")
        encoded += (
            b32hextab[c >> 30]
            + b32hextab[(c >> 20) & 0x3FF]  # bits 1 - 10
            + b32hextab[(c >> 10) & 0x3FF]  # bits 11 - 20
            + b32hextab[c & 0x3FF]  # bits 21 - 30  # bits 31 - 40
        )

    # Adjust for any leftover partial quanta
    if leftover == 1:
        encoded = encoded[:-6]
    elif leftover == 2:
        encoded = encoded[:-4]
    elif leftover == 3:
        encoded = encoded[:-3]
    elif leftover == 4:
        encoded = encoded[:-1]

    return bytes(encoded)


def scalr_id(prefix: str) -> str:
    """Generate `Scalr ID <https://github.com/Scalr/int-scalr/blob/master/docs/coding_standards/ids.rst>`_.

    :param prefix: str Prefix for Scalr id
    :return: Generated id
    """
    validation = re.compile("[a-z0-9]{1,6}")
    if not validation.fullmatch(prefix):
        raise ValueError(
            "Prefix must be 1-6 length and contain only lowercase latin letters or digits"
        )

    time_mark = str(time.time())
    time_prefix = format(int(time_mark[:10] + time_mark[11:15].ljust(4, "0")), "x").encode()
    st = b"".join([time_prefix, binascii.hexlify(secrets.token_bytes(4))[1:]])

    return prefix + "-" + _base32_hex_encode(binascii.unhexlify(st)).decode()


class OrderedEnum(enum.Enum):
    # pylint: disable=comparison-with-callable
    # https://github.com/PyCQA/pylint/issues/2306
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented
