"""
Minimalist package manager we use to install fatmouse agents
and maintain a cache of them for rollback purposes.
"""
import functools
import gzip
import hashlib
import io
import itertools
import json
import logging
import os
import posixpath
import re
import shutil
import sys
import tempfile
import time
import typing as tp
import uuid
import xml.etree.ElementTree as ET
from abc import ABCMeta
from abc import abstractmethod
from collections import Sized
from collections import defaultdict
from collections import namedtuple
from datetime import datetime
from operator import attrgetter

import requests
import urllib3
from requests import adapters

from common.utils import exc
from common.utils import mixutil
from common.utils import six
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact

LOG = logging.getLogger(__name__)
ARCHIVE_SIZE = 2
MSI_ERROR_INSTALL_ALREADY_RUNNING = 1618

# 5 (interval) x 480 (times) = ~2400s - default hostup timeout
RETRY_LOCKED_INTERVAL: int = 5
RETRY_LOCKED_TIMES: int = 480


def fix_version_format(version1, version2):
    """XXX: temporary fix for HostedScalr"""
    is_version1_devel = ".b" in version1
    is_version2_devel = ".b" in version2
    if is_version1_devel != is_version2_devel:
        # Mixing release and devel version formats. Needs fixing.
        if is_version1_devel:
            version1 = version1[: version1.index(".b")] + ".0"
        else:
            version2 = version2[: version2.index(".b")] + ".0"
    return version1, version2


class DpkgCmd(subprocess2.Command):
    """Command wrapper around dpkg."""

    binary = "/usr/bin/dpkg"


class RpmCmd(subprocess2.Command):
    """Command wrapper around rpm."""

    binary = "/bin/rpm"


class Error(exc.Error):
    """Base Error exception."""


class PackageNotFoundError(Error):
    """When package is not found in repository."""


class PackageDownloadError(Error):
    """When package download fails."""


class PackageInstallError(Error):
    """When package can't be installed."""


Package = namedtuple("Package", "name, version, location, hash, hash_type")
HistoryEntry = namedtuple("HistoryEntry", "hash, version, download_date, install_date")


def requests_get(
    url: str,
    stream: bool = False,
    verify: bool = True,
    proxies: tp.Optional[dict] = None,
    headers: tp.Optional[dict] = None,
    max_retries: int = 8,
    backoff_factor: float = 0.2,
    exc_class: tp.Type[Error] = None,
    exc_message: tp.Optional[str] = None,
) -> requests.Response:
    """
    A wrapper around `requests.get` which handles exceptions nicely.

    :param url: The url to fetch.
    :param stream: Stream response content.
    :param verify: SSL Verification.
    :param proxies: Dictionary mapping protocol or protocol and host to the
        URL of the proxy (e.g. {'http': 'foo.bar:3128', 'http://host.name': 'foo.bar:4012'}).
    :param max_retries: The maximum number of retries each connection should attempt.
    :param backoff_factor: Backoff interval factor.
    :param exc_class: The optional exception class for an error.
    :param exc_message: The optional message for an error.
    """
    if exc_class is None:
        exc_class = Error
    if exc_message is None:
        exc_message = f"GET request to {url} is failed"

    # interval: {backoff factor} * (2 ** ({number of total retries} - 1))
    retry_strategy = urllib3.util.retry.Retry(
        total=max_retries,
        backoff_factor=backoff_factor,  # 0.2 => 0.1, 0.2, 0.4, 0.8, 1.6, 3.2, 6.4, 12.8, 25.6
        status_forcelist=[429, 500, 502, 503, 504],
    )
    adapter = adapters.HTTPAdapter(max_retries=retry_strategy)
    session = requests.Session()
    session.proxies = proxies or {}
    session.stream = stream
    session.headers = headers

    session.mount("https://", adapter)
    session.mount("http://", adapter)

    LOG.debug("GET %s (stream=%s, verify=%s, proxies=%s)", url, stream, verify, bool(proxies))
    try:
        # NOTE: We must pass `verify` directly, rather though the `session.verify`
        # because we use REQUESTS_CA_BUNDLE and https://github.com/psf/requests/issues/3829
        # affects us.
        response = session.get(url, verify=verify)
    except requests.RequestException as err:
        raise exc_class(f"{exc_message.rstrip('.')}. Requests error: {str(err)}") from err

    if not response.ok:
        raise exc_class(f"{exc_message.rstrip('.')}. HTTP error code: {response.status_code}")

    return response


class CacheDirMetadata(object):

    """
    A simple json db to store metadata for fetched packages.
    """

    def __init__(self, path):
        self._metadata = {}
        self._metadata_path = os.path.join(path, "metadata")
        self.exists = os.path.exists(self._metadata_path)
        if self.exists:
            try:
                with open(self._metadata_path) as f:
                    LOG.debug("Loading metadata")
                    self._metadata = json.load(f)
            except ValueError:
                six.reraise(
                    Error,
                    Error("Invalid metadata file: {0}".format(self._metadata_path)),
                    sys.exc_info()[2],
                )

    def get(self, value, default=None):
        """Retrieve metadata value."""
        return self._metadata.get(value, default)

    def set(self, **kwargs):
        """Change metadata value."""
        self._metadata.update(kwargs)
        with open(self._metadata_path, "w") as f:
            LOG.debug("Saving metadata")
            json.dump(self._metadata, f)


class PackageList(Sized):
    """
    Packages stored as a dict of lists. Need this because we can
    have many versions of the same package in the repo.
    """

    def __init__(self, packages):
        self._packages = defaultdict(list)
        for package in packages:
            self._packages[package.name].append(package)

    def __len__(self):
        return sum(len(entry) for entry in self._packages.values())

    def all_versions(self, name):
        """Return all versions of the given package."""
        packages = self._packages.get(name)
        if not packages:
            raise PackageNotFoundError("Package {0} can't be found in the repository".format(name))
        return packages

    def last_version(self, name):
        """Return newest version of the package with given name."""
        packages = self.all_versions(name)
        return max(packages, key=lambda pkg: mixutil.Version(pkg.version))

    def exact_version(self, name, version):
        """Return exact package of the given version."""
        packages = self.all_versions(name)
        if version:
            for pkg in packages:
                if pkg.version == version:
                    return pkg
            raise PackageNotFoundError(
                "Package {0} version {1} can't be found in the repository".format(name, version)
            )


@six.add_metaclass(ABCMeta)
class ABCManager(object):
    """
    Common functionality for fetching, installing, querying status and updating history.
    History code is platform-independent and other methods use a lot of template methods
    that do the job on the actual platforms (rpm/dpkg).
    """

    def __init__(
        self,
        repo_string,
        storage_dir=None,
        proxies=None,
        ssl_verify_peer=True,
        arch="x86_64",
        agent=True,
    ):
        """
        :param str repo_string: Link to the repository in the repository type
            (rpm, dpkg) specific format.
        :param str storage_dir: Package manager cache directory.
        :param list proxies: list of proxies understood by Requests.
        :param bool ssl_verify_peer: Package manager ssl verify setting
        :param str arch: An architecture of the packages we are interested in.
        :param bool agent: Determine who instantiate the the class. If True(default) - agent,
                           otherwise server.
        """
        self.on_agent = agent
        self.packages = None
        self.baseurl = None
        self.storage_dir = storage_dir
        self._proxies = proxies
        self._ssl_verify_peer = ssl_verify_peer
        self._arch = arch
        self._set_repo_string(repo_string)
        LOG.debug("Constructed %s", self)

    @abstractmethod
    def _set_repo_string(self, repo_string):
        pass

    @abstractmethod
    def _fetch_repo_metadata(self):
        pass

    @abstractmethod
    def _parse_repo_metadata(self, metadata):
        pass

    @abstractmethod
    def _fetch_packages_metadata(self, path, checksum):
        pass

    @abstractmethod
    def _parse_packages_metadata(self, metadata):
        pass

    @abstractmethod
    def _install(self, packages_path, metadata, no_deps=False):
        pass

    @classmethod
    @abstractmethod
    def get_installed_version(cls, name):
        """Return installed version of the package."""

    def _get_history(self, name):
        LOG.debug("Getting history")
        history = []
        package_dir = os.path.join(self.storage_dir, name)
        if os.path.exists(package_dir):
            for package_hash in os.listdir(package_dir):
                metadata = CacheDirMetadata(os.path.join(package_dir, package_hash))
                if not metadata.exists:
                    continue
                history_entry = HistoryEntry(
                    hash=package_hash,
                    version=metadata.get("version"),
                    download_date=datetime.fromtimestamp(metadata.get("download_date")),
                    install_date=datetime.fromtimestamp(metadata.get("install_date"))
                    if metadata.get("install_date")
                    else None,
                )
                history.append(history_entry)
        return history

    def _update_fetch_history(self, package_storage_dir, package, deps):
        LOG.debug("Updating history")
        # Add index to allow access to our fetched files only knowing the hash
        index = os.path.join(self.storage_dir, ".index")
        if not os.path.exists(index):
            os.makedirs(index)
        symlink = os.path.join(index, package.hash)
        if not os.path.exists(symlink):
            sysutil.symlink(package_storage_dir, symlink, target_is_directory=True)

        # Write down version, date and package order
        CacheDirMetadata(package_storage_dir).set(
            version=package.version,
            download_date=time.time(),
            packages=list(map(lambda pkg: posixpath.basename(pkg.location), deps + [package])),
        )

    def _remove_cache_entries(self, entries, name):
        installed_version = self.get_installed_version(name)
        for entry in entries:
            if entry.version == installed_version:
                continue
            path = os.path.join(self.storage_dir, name, entry.hash)
            LOG.debug("Removing old entry from the pkgmgr cache: %s", path)
            shutil.rmtree(path)
            symlink = os.path.join(self.storage_dir, ".index", entry.hash)
            if os.path.lexists(symlink):
                LOG.debug("Removing symlink to the old entry in the pkgmgr cache: %s", symlink)
                sysutil.rm_rf(symlink)

        # We can have a directory without metadata file, which means
        # UpdClient was interrupted during fetch. Let's clean this up.
        # [SCALARIZR-1854]
        pkg_dir = os.path.join(self.storage_dir, name)
        if os.path.exists(pkg_dir):
            broken_entries = [
                os.path.join(pkg_dir, dir)
                for dir in os.listdir(pkg_dir)
                if "metadata" not in os.listdir(os.path.join(pkg_dir, dir))
            ]

            for entry in broken_entries:
                LOG.debug("Removing broken entry from the pkgmgr cache: %s", entry)
                shutil.rmtree(entry)

    def _cleanup_history(self, name):
        # We only keep last `ARCHIVE_SIZE` versions in our cache. Remove the rest.
        history = sorted(
            self._get_history(name), key=lambda history: history.download_date, reverse=True
        )
        old_entries = list(history)[ARCHIVE_SIZE:]
        self._remove_cache_entries(old_entries, name)

    @staticmethod
    def _update_install_history(metadata):
        metadata.set(install_date=time.time())

    def _get_package_list(self):
        repo_metadata = self._fetch_repo_metadata()
        packages_path, checksum = self._parse_repo_metadata(repo_metadata)
        packages_metadata = self._fetch_packages_metadata(packages_path, checksum)
        packages = self._parse_packages_metadata(packages_metadata)
        return PackageList(packages)

    def updatedb(self):
        """Update packages list from repository."""
        LOG.info("Syncing repository info")
        self.packages = self._get_package_list()

    def search(self, package):
        """Search for a package from the list obtained on updatedb."""
        packages = self._get_package_list()
        return packages.last_version(package)

    def fetch(self, name, version, deps=None):
        """
        Directory structure for fetched files:

        STORAGE_DIR/
            .index/
                $hash1 (symlinks to the actual folders below)
                $hash2
                $hash3
            $name1/
                $hash1/
                    package1.rpm
                    package2.rpm
                    metadata
                $hash2/
                    package1.rpm
                    package2.rpm
                    metadata
            $name1/
                $hash3/
                    package1.rpm
                    package2.rpm
                    metadata

        Where `metadata` is a json containing file where we store version,
        download date, install date and package installation order.

        :param name: Name of the package to fetch from repository
        :type name: str

        :param version: Version of the package
        :type version: str

        :param deps: Optional dependencies which must be fetched along with the main package
        :type deps: list

        :returns: Hash of the fetched packages
        :rtype: str
        """

        def _fetch(pkg, base_dir):
            hasher = hashlib.new(pkg.hash_type)
            path = os.path.join(base_dir, posixpath.basename(pkg.location))
            # pylint: disable=no-member
            url = posixpath.join(self.baseurl, pkg.location)
            LOG.debug("Fetching package from %s", url)
            r = requests_get(
                url,
                stream=True,
                verify=self._ssl_verify_peer,
                proxies=self._proxies,
                exc_class=PackageDownloadError,
                exc_message="Cannot fetch packages from the repository",
            )
            with open(path, "wb") as f:
                chunk_size = 1024
                downloaded = 0
                for chunk in r.iter_content(chunk_size):
                    f.write(chunk)
                    hasher.update(chunk)
                    downloaded += len(chunk)
                if r.headers.get("Content-Length"):
                    LOG.debug(
                        "Fetched %d bytes out of expected %s bytes.",
                        downloaded,
                        r.headers.get("Content-Length"),
                    )
                else:
                    LOG.debug("Fetched %d bytes.", downloaded)
            actual_hash = hasher.hexdigest()
            if pkg.hash != actual_hash:
                os.unlink(path)
                raise Error(
                    "Package {} checksum does not match (actual={} expected={})".format(
                        pkg.name, actual_hash, pkg.hash
                    )
                )

        self._cleanup_history(name)
        LOG.info("Fetching packages")
        package = self.packages.exact_version(name, version)
        package_storage_dir = os.path.join(self.storage_dir, name, package.hash)
        if not os.path.exists(package_storage_dir):
            os.makedirs(package_storage_dir)

        deps = (
            [self.packages.exact_version(dep["name"], dep["version"]) for dep in deps]
            if deps
            else []
        )
        for pkg in deps + [package]:
            _fetch(pkg, package_storage_dir)

        self._update_fetch_history(package_storage_dir, package, deps)
        return package.hash

    def status(self, name):
        """
        Returns installed and available versions of the requested package,
        and also its install history.

        :param name: Name of the package we are interested it
        :type name: str

        :rtype: dict
        """
        package_versions = self.packages.all_versions(name)

        versions_available = list(
            sorted(map(attrgetter("version"), package_versions), key=mixutil.Version)
        )
        version_installed = self.get_installed_version(name)
        history = self._get_history(name)
        result = {
            "name": name,
            "installed": version_installed,
            "available": versions_available,
            "history": history,
        }

        max_available = self.packages.last_version(name).version

        # XXX: temporary fix for HS
        available_version, installed_version = fix_version_format(
            max_available, version_installed or "0"
        )

        update_available = mixutil.Version(available_version) > mixutil.Version(installed_version)
        result["candidate"] = max_available if update_available else None

        for entry in history:
            if entry.version == version_installed:
                result["installed_hash"] = entry.hash
                break
        return result

    def install(self, package_hash, no_deps=False):
        """
        Installs the fetched package from the cache dir.

        :param package_hash: Hash of the package to install (returned by :py:meth:`.fetch`)
        :type package_hash: str

        :param no_deps: Disables all dependency logic (deb/rpm based only).
        :type no_deps: bool
        """
        LOG.info("Installing packages")
        packages_path = os.path.join(self.storage_dir, ".index", package_hash)
        LOG.debug("Installing packages from %s", packages_path)
        if not os.path.exists(packages_path):
            raise PackageNotFoundError(
                "Package hashed as {0} can't be found in the cache directory {1}".format(
                    package_hash, self.storage_dir
                )
            )
        metadata = CacheDirMetadata(packages_path)
        self._install(packages_path, metadata, no_deps=no_deps)
        self._update_install_history(metadata)

    def __repr__(self):
        return "<{} url={} ssl_verify_peer={}>".format(
            self.__class__.__name__, self.baseurl, self._ssl_verify_peer
        )


class RpmManager(ABCManager):
    """Rpm package manager."""

    REPO_XMLNS = "{http://linux.duke.edu/metadata/repo}"
    COMMON_XMLNS = "{http://linux.duke.edu/metadata/common}"

    @staticmethod
    def normalize_repo_string(url, os_name, releasever, arch):
        """
        We might get an url in this format:
        http://rpm-delayed.scalr.net/rpm/rhel/$releasever/$basearch
        Substitute $releasever and $basearch with actual values if they are present.
        """
        if os_name not in ("amazon", "suse"):
            releasever = releasever.split(".")[0]
        url = re.sub(r"\$releasever", releasever, url)
        url = re.sub(r"\$basearch", arch, url)
        return url

    def _set_repo_string(self, repo_string):
        self.baseurl = repo_string

    def _fetch_repo_metadata(self):
        r = requests_get(
            posixpath.join(self.baseurl, "repodata/repomd.xml"),
            verify=self._ssl_verify_peer,
            proxies=self._proxies,
            exc_class=Error,
            exc_message="Cannot retrieve repo metadata from {0}.".format(self.baseurl),
        )
        return r.text

    def _parse_repo_metadata(self, metadata):
        ns = RpmManager.REPO_XMLNS
        tree = ET.fromstring(metadata)
        data = tree.findall(ns + "data")
        for el in data:
            if el.attrib.get("type") == "primary":
                location = el.find(ns + "location")
                primary_path = location.attrib["href"]
                checksum = el.find(ns + "checksum")
                checksum_type = checksum.attrib["type"].lower()
                # From the createrepo man: `The older default was "sha", which is actually "sha1"`
                checksum_type = "sha1" if checksum_type == "sha" else checksum_type
        return primary_path, {"type": checksum_type, "value": checksum.text}

    def _fetch_packages_metadata(self, path, checksum):
        r = requests_get(
            posixpath.join(self.baseurl, path),
            verify=self._ssl_verify_peer,
            proxies=self._proxies,
            exc_class=Error,
            exc_message="Cannot get packages metadata",
        )
        hasher = hashlib.new(checksum["type"])
        hasher.update(r.content)
        actual_hash = hasher.hexdigest()
        if checksum["value"] != actual_hash:
            raise Error(
                "Packages file has wrong sha1 sum (actual={} expected={})".format(
                    actual_hash, checksum["value"]
                )
            )
        metadata = gzip.GzipFile(fileobj=io.BytesIO(r.content)).read()
        return metadata

    def _parse_packages_metadata(self, metadata):
        ns = RpmManager.COMMON_XMLNS
        tree = ET.fromstring(metadata)
        entries = tree.findall(ns + "package")

        def _make_package(element):
            name = element.find(ns + "name").text
            version = element.find(ns + "version").attrib["ver"]
            location = element.find(ns + "location").attrib["href"]
            package_hash = element.find(ns + "checksum")
            hash_type = package_hash.attrib["type"].lower()
            hash_type = "sha1" if hash_type == "sha" else hash_type
            return Package(name, version, location, package_hash.text, hash_type)

        return [_make_package(e) for e in entries]

    @classmethod
    def get_installed_version(cls, name):
        r = RpmCmd().args("-q", "--queryformat", "%{VERSION}", name).run()
        version = None if r.returncode else r.stdout
        return version

    def _install(self, packages_path, metadata, no_deps=False):
        packages = tuple(metadata.get("packages"))
        rpm_args = ["-U", "--nosignature", "--replacepkgs", "--oldpackage"]
        if no_deps:
            rpm_args.append("--nodeps")
        r = RpmCmd().args(*rpm_args).args(*packages).run(cwd=packages_path)
        if r.returncode:
            raise PackageInstallError(r.stderr)
        # The only way to detect the failure of a postinstall script,
        # because rpm doesn't forward the return code.
        if "scriptlet failed" in r.stderr:
            raise PackageInstallError(r.stderr)

    @staticmethod
    def uninstall(name):
        """
        :param name: Name of the package to uninstall
        :type name: str
        """
        LOG.info("Uninstalling packages")
        try:
            RpmCmd().args("-e", "--noscripts", "--nodeps", name).check_call()
        except subprocess2.CalledProcessError as e:
            six.reraise(PackageInstallError, PackageInstallError(e.stderr), sys.exc_info()[2])

    def purge(self, name):
        """
        :param name: Name of the package to purge
        :type name: str
        """
        LOG.info("Uninstalling packages")
        self.uninstall(name)


retry_locked_dpkg = functools.partial(
    mixutil.retries_error,
    except_expr=subprocess2.CalledProcessError,
    condition_func=lambda e: "is locked by another process" in e.stderr,
    interval=RETRY_LOCKED_INTERVAL,
    times=RETRY_LOCKED_TIMES,
)


class DpkgManager(ABCManager):
    """Dpkg package manager."""

    # Default checksum to use for verifing files
    HASH_TYPE = "SHA1"

    def _set_repo_string(self, repo_string):
        parts = repo_string.split()
        if len(parts) == 2:
            # APT-plain repo
            # "http://stridercd.scalr-labs.com/apt-plain stable"
            self.plain = True
        elif len(parts) == 3:
            # APT-pool repo
            # "http://stridercd.scalr-labs.com/apt/develop master main"
            self.plain = False
            self.component = parts[2]
        else:
            raise Error("Invalid repo string: {0}".format(repo_string))
        self.baseurl, self.distr = parts[0], parts[1].lstrip("/")
        if self.on_agent:
            self._try_distributive_repository()

    def _try_distributive_repository(self):
        """Try to use distributive path for non-develop repos.

        For instance: 'deb http://repo.scalr.net/apt-plain latest/' =>
                      'deb http://repo.scalr.net/apt-plain latest/ubuntu/16.04/'
        """
        if not self.baseurl.endswith("develop"):
            path = posixpath.join(self.distr, fact["os"]["name"], str(fact["os"]["release"]) + "/")
            try:
                LOG.debug(
                    "Trying to use repository for %s %s...",
                    fact["os"]["name"],
                    str(fact["os"]["release"]),
                )
                resp = requests.get(posixpath.join(self.baseurl, path))
                if resp.ok:
                    self.distr = path  # pylint: disable=attribute-defined-outside-init
            except requests.RequestException:
                LOG.warning("Unable to use %r path. Using %r", path, self.distr)
        else:
            LOG.debug("Using development repository. OS name and version will not be injected.")

    def _fetch_repo_metadata(self):
        if self.plain:
            url = posixpath.join(self.baseurl, self.distr, "Release")
        else:
            url = posixpath.join(self.baseurl, "dists", self.distr, "Release")
        r = requests_get(
            url,
            verify=self._ssl_verify_peer,
            proxies=self._proxies,
            exc_class=Error,
            exc_message="Cannot retrieve repo metadata from {0}.".format(url),
        )
        return r.text.split("\n")

    def _parse_repo_metadata(self, metadata):
        if self.plain:
            packages_path = "Packages"
            full_packges_path = posixpath.join(self.baseurl, self.distr, packages_path)
        else:
            packages_path = posixpath.join(self.component, "binary-{}/Packages".format(self._arch))
            full_packges_path = posixpath.join(self.baseurl, "dists", self.distr, packages_path)

        hashes = itertools.dropwhile(
            lambda s: not s.strip().startswith("{0}:".format(DpkgManager.HASH_TYPE)), metadata
        )
        packages = itertools.dropwhile(lambda s: not s.strip().endswith(packages_path), hashes)
        try:
            packages_hash = list(packages)[0].strip().split()[0]
        except IndexError:
            LOG.debug("Repository metadata does not contain checksums for Packages file.")
            packages_hash = None
        return full_packges_path, packages_hash

    def _fetch_packages_metadata(self, path, checksum):
        r = requests_get(
            path,
            verify=self._ssl_verify_peer,
            proxies=self._proxies,
            exc_class=Error,
            exc_message="Cannot retrieve packages metadata",
        )
        if checksum:
            hasher = hashlib.new(DpkgManager.HASH_TYPE)
            hasher.update(r.content)
            actual_hash = hasher.hexdigest()
            if checksum != actual_hash:
                raise Error(
                    "Packages file has wrong checksum (actual={} expected={})".format(
                        actual_hash, checksum
                    )
                )
        return r.text

    def _parse_packages_metadata(self, metadata):
        def _make_package(entry):
            arch = re.search(r"^Architecture:(.*)", entry, re.M).group(1).strip()
            self_arch = "amd64" if self._arch == "x86_64" else "i386"
            if arch not in (self_arch, "all"):
                return None

            name = re.search(r"^Package:(.*)", entry, re.M).group(1).strip()
            version = re.search(r"^Version:(.*)", entry, re.M).group(1).strip()
            location = re.search(r"^Filename:(.*)", entry, re.M).group(1).strip()
            package_hash = (
                re.search(r"^{0}:(.*)".format(DpkgManager.HASH_TYPE), entry, re.M).group(1).strip()
            )
            return Package(name, version, location, package_hash, DpkgManager.HASH_TYPE)

        entries = metadata.strip().split("\n\n")
        return [pkg for pkg in map(_make_package, entries) if pkg]

    @classmethod
    def get_installed_version(cls, name):
        r = subprocess2.run(("dpkg-query", "-W", "-f=${Version} ${Status}", name))
        if r.returncode == 0:
            version, status = r.stdout.split(" ", 1)
            if status.startswith("install ok"):
                return version
        return None

    def _install(self, packages_path, metadata, no_deps=False):
        packages = (
            subprocess2.check_output(("dpkg-query", "-W", "-f=${Status}|${Package}\n"))
            .strip()
            .split("\n")
        )
        problem_packages = (
            package.split("|")[1]
            for package in packages
            if package.split("|")[0] in ("install ok unpacked", "install ok half-configured")
        )

        # delete postinst scripts of problem packages
        for name in problem_packages:
            postinst_path = "/var/lib/dpkg/info/{0}.postinst".format(name)
            sysutil.rm_rf(postinst_path)

        # Restore system state if we have half-configured packages before install
        @retry_locked_dpkg()
        def do_reconfigure():
            DpkgCmd().args("--configure", "-a").check_output()

        try:
            do_reconfigure()
        except subprocess2.CalledProcessError as e:
            six.reraise(Error, Error(e.stderr), sys.exc_info()[2])

        packages = tuple(metadata.get("packages"))
        dpkg_args = ["-i", "--force-downgrade"]
        if no_deps:
            dpkg_args.append("--force-depends")

        @retry_locked_dpkg()
        def do_install():
            DpkgCmd().args(*dpkg_args).args(*packages).check_output(cwd=packages_path)

        try:
            do_install()
        except subprocess2.CalledProcessError as e:
            six.reraise(PackageInstallError, PackageInstallError(e.stderr), sys.exc_info()[2])

    @staticmethod
    def uninstall(name):
        """
        :param name: Name of the package to uninstall
        :type name: str
        """
        LOG.info("Uninstalling package")

        @retry_locked_dpkg()
        def do_uninstall():
            DpkgCmd().args("-r", "--force-depends", name).check_call()

        try:
            do_uninstall()
        except subprocess2.CalledProcessError as e:
            six.reraise(Error, Error(e.stderr), sys.exc_info()[2])

    @staticmethod
    def purge(name):
        """
        :param name: Name of the package to purge
        :type name: str
        """
        LOG.info("Purging package")

        @retry_locked_dpkg()
        def do_purge():
            DpkgCmd().args("-P", "--force-depends", name).check_call()

        try:
            do_purge()
        except subprocess2.CalledProcessError as e:
            six.reraise(Error, Error(e.stderr), sys.exc_info()[2])


retry_locked_msiexec = functools.partial(
    mixutil.retries_error,
    except_expr=subprocess2.CalledProcessError,
    condition_func=lambda e: e.returncode == MSI_ERROR_INSTALL_ALREADY_RUNNING,
    interval=RETRY_LOCKED_INTERVAL,
    times=RETRY_LOCKED_TIMES,
)


class MsiManager(ABCManager):
    """Msi package manager."""

    # Default checksum to use for verifing files
    HASH_TYPE = "md5"
    # Default install dir of the main package
    INSTALL_DIR = os.path.join("C:\\opt")

    def __init__(self, repo_string, *args, **kwargs):
        super(MsiManager, self).__init__(repo_string, *args, **kwargs)
        self.repo_type = None
        self._log_file_name = "scalr-repo-log.txt"

    def _set_repo_string(self, repo_string):
        self.baseurl = repo_string

    def _fetch_repo_metadata(self):
        r = requests_get(
            posixpath.join(self.baseurl, "index.json"),
            verify=self._ssl_verify_peer,
            proxies=self._proxies,
            exc_class=Error,
            exc_message="Cannot retrieve repo metadata",
        )
        return json.loads(r.text)

    def _parse_repo_metadata(self, metadata):
        self.repo_type = metadata["repository"]
        return None, None

    def _fetch_packages_metadata(self, path, checksum):
        return self._fetch_repo_metadata()

    def _parse_packages_metadata(self, metadata):
        def _make_package(entry):
            name = entry["name"]
            version = entry["version"]
            location = entry["path"]
            package_hash = entry["md5"]
            return Package(name, version, location, package_hash, self.HASH_TYPE)

        return [_make_package(pkg) for pkg in metadata["packages"]]

    @classmethod
    def get_installed_version(cls, name):
        """
        It's implied that all packages provided by scalr win repo are omnibus-builded.
        So we can always get version from the omnibus Manifest file
        """
        manifest_path = os.path.join(cls.INSTALL_DIR, name, "current", "version-manifest.txt")
        omnibus_packages = None
        if os.path.exists(manifest_path):
            with open(manifest_path, "r") as fp:
                omnibus_packages = fp.readlines()
            for l in omnibus_packages:
                pkg_info = l.split()
                if pkg_info[0] == name:
                    return pkg_info[1]
        return ""

    def _install_single(self, package, packages_path):
        log_file = os.path.join(self.storage_dir, self._log_file_name)
        post_inst_error_file = os.path.join(
            tempfile.gettempdir(), f"scalarizr_post_install_error_{uuid.uuid4()}.log"
        )

        @retry_locked_msiexec()
        def do_install():
            msi_cmd = (
                "msiexec",
                "/i",
                package,
                "/qn",
                "/norestart",
                "/l*vx",
                log_file,
                f"POST_INSTALL_SEQUENCE_ERROR_LOG={post_inst_error_file}",
            )
            subprocess2.check_call(msi_cmd, shell=True, cwd=packages_path)

        try:
            do_install()
        except subprocess2.CalledProcessError as err:
            err_code = err.returncode
            err_msg = err.stderr or f"Error during {package} package installation"

            # handle post-install errors
            if os.path.exists(post_inst_error_file):
                err_msg = f"The post-install sequence failed"

                # optionally read an error from the log
                try:
                    with open(post_inst_error_file, "r", errors="ignore") as fp:
                        post_inst_err_msg = fp.read().strip()
                    if post_inst_err_msg:
                        err_msg = f"{err_msg} with an error: {post_inst_err_msg}"
                except Exception as err:
                    LOG.debug(
                        "Failed to read the log %s. Reason: %s", post_inst_error_file, str(err)
                    )
                sysutil.rm_rf_silent(post_inst_error_file)

            raise PackageInstallError(
                f"[{err_code}] {err_msg}. " f"See installation log for details: {log_file}"
            ) from None

    def _install(self, packages_path, metadata, no_deps=False):
        # Restore system state if we have half-configured packages before install
        packages = tuple(metadata.get("packages"))
        for package in packages:
            self._install_single(package, packages_path)

    def _uninstall_single(self, package, packages_path, no_log=False):
        log_file = os.path.join(self.storage_dir, self._log_file_name)
        cmd = ("msiexec", "/qn", "/x", package)
        if not no_log:
            cmd += ("/l*", log_file)

        @retry_locked_msiexec()
        def do_uninstall():
            subprocess2.check_call(cmd, shell=True, cwd=packages_path)

        try:
            do_uninstall()
        except subprocess2.CalledProcessError as err:
            err_code = err.returncode
            err_msg = err.stderr or f"Error during {package} package uninstallation"
            raise Error(f"[{err_code}] {err_msg}") from err

    def uninstall(self, name):
        """
        :param name: Name of the package to uninstall
        :type name: str
        """
        LOG.info("Uninstalling package %s", name)
        pkg_status = self.status(name)
        hash_ = pkg_status.get("installed_hash")
        if hash_:
            packages_path = os.path.join(self.storage_dir, ".index", hash_)
            if not os.path.exists(packages_path):
                raise PackageNotFoundError(
                    "Package installer at {0} can't be found in the cache "
                    "directory {1}".format(hash_, self.storage_dir)
                )
            metadata = CacheDirMetadata(packages_path)
            for package in tuple(metadata.get("packages")):
                self._uninstall_single(package, packages_path)
        else:
            LOG.info('Package "%s" is not installed', name)

    def purge(self, name):
        """
        :param name: Name of the package to purge
        :type name: str
        """
        self.uninstall(name)


class DockerManager(ABCManager):
    """
    Implement package manager interface in case when scalarizr runs inside docker container
    """

    DEFAULT_SCHEME = "https"
    HASH_TYPE = "sha256"
    auth = None

    def _set_repo_string(self, repo_string):
        _repo_components = repo_string.lower().strip("/").split("/")
        self.image = _repo_components[-1]
        self.namespace = "/".join(_repo_components[1:-1])
        self.registry_path = _repo_components[0]
        self.baseurl = self.DEFAULT_SCHEME + "://" + self.registry_path
        if self.registry_path == "docker.io":
            self.auth = True
            self.baseurl = "https://registry.hub.docker.com"

    def _fetch_repo_metadata(self):
        # Just need to exist for compatibility with ABCManager interface
        # DockerManager works without repository metadata
        return

    def _parse_repo_metadata(self, metadata):
        # Just need to exist for compatibility with ABCManager interface
        return None, None

    def _fetch_packages_metadata(self, path, checksum):
        packages_metadata = []
        url = posixpath.join(self.baseurl, "v2", self.namespace, self.image, "tags", "list")
        r = requests_get(
            url,
            verify=self._ssl_verify_peer,
            proxies=self._proxies,
            exc_class=Error,
            headers=self._get_authz_headers(),
            exc_message="Cannot retrieve package metadata",
        )
        packages_metadata.append(json.loads(r.text))
        return packages_metadata

    @mixutil.timed_cache(ttl=180)
    def _get_registry_token(self, registry_path, repo):
        if registry_path == "docker.io":
            auth_url = (
                "https://auth.docker.io/token?service="
                "registry.docker.io&scope=repository:{}:pull".format(repo)
            )
            auth_data = requests_get(
                auth_url,
                proxies=self._proxies,
                exc_class=Error,
                exc_message="Cannot retrieve repo auth data",
            )
            token = json.loads(auth_data.text.strip())["token"]
            return token

    def _get_authz_headers(self):
        if self.auth:
            repo = posixpath.join(self.namespace, self.image)
            header = {
                "Authorization": "Bearer {}".format(
                    self._get_registry_token(self.registry_path, repo)
                )
            }
            return header

    def _hash(self, data):
        h = hashlib.new(self.HASH_TYPE)
        h.update(data.encode("utf-8"))
        return h.hexdigest()

    def _parse_packages_metadata(self, metadata):
        def _make_package(entry):
            packages = []
            location = entry["name"]
            name = location.split("/")[-1]
            # First two octet digit than anything
            # 5.5.5, 5.5.sds-132, 5.5.b4069.fb332ef - ok
            # latest, 5.34ffdf, 34d.2.32 - not ok
            # pylint: disable=anomalous-backslash-in-string
            pattern = re.compile("^(\*|\d+(\.\d+){1,3}(\..*)?)$")
            digit_tag = filter(pattern.match, entry["tags"])
            for tag in digit_tag:
                version = tag
                package_hash = self._hash(self.image + ":" + version)
                packages.append(Package(name, version, location, package_hash, self.HASH_TYPE))
            return packages

        packages_metadata = []
        for m in metadata:
            packages_metadata += _make_package(m)
        return packages_metadata

    @classmethod
    def get_installed_version(cls, name):
        """
        Return a version of installed image
        :param name: name of image that currently installed
        :return: version string
        """
        try:
            docker_format = "'{{ index .Config.Labels \"com.scalr.agent.version\"}}'"
            cmd = "docker inspect --format {} scalr/scalarizr".format(docker_format)
            version = subprocess2.check_output_host(cmd).strip()
            return version
        except subprocess2.CalledProcessError:
            e = sys.exc_info()[1]
            LOG.error("Could not fetch container %s version. Error: %s", name, e.stderr)

    def _install(self, packages_path, metadata, no_deps=False):
        image = None
        cmd = 'docker images --digests --format "{{.Repository}}|{{.Tag}}"'
        images_data = subprocess2.check_output_host(cmd).split("\n")
        fetched_version = metadata.get("version")
        fetched_package = metadata.get("packages")[0]
        for i in images_data:
            if i:
                image_param = i.split("|")
                repo = image_param[0]
                version = image_param[1]
                if fetched_package in repo and version == fetched_version:
                    image = repo + ":" + version
                    break

        installed_version = self.get_installed_version(fetched_package)
        LOG.info(
            "Installing package scalr/scalarizr:%s over scalr/scalarizr:%s",
            installed_version,
            image.split(":")[1],
        )
        subprocess2.check_output_host("docker tag {} scalr/scalarizr".format(image))
        subprocess2.check_output_host("docker run --rm -v /:/rootfs scalr/scalarizr --install")
        subprocess2.check_output_host("systemctl daemon-reload")

    def _cleanup_history(self, name):
        # We only keep last `ARCHIVE_SIZE` versions in our cache. Remove the rest.
        history = sorted(
            self._get_history(name), key=lambda history: history.download_date, reverse=True
        )
        old_entries = list(history)[ARCHIVE_SIZE:]
        self._remove_cache_entries(old_entries, name)

        for entry in old_entries:
            try:
                # TODO: need additional testing, can have a problem with image dependency
                cmd = "docker rmi {}:{}".format(name, entry.version)
                subprocess2.check_output_host(cmd)
            except subprocess2.CalledProcessError:
                e = sys.exc_info()[1]
                LOG.error("Could not remove image %s:%s. Error: %s", name, entry.hash, e.stderr)
            else:
                LOG.error("Image %s:%s successfully removed.", name, entry.hash)

    def fetch(self, name, version, deps=None):
        def _fetch(pkg):
            registry_path = posixpath.join(self.registry_path, pkg.location + ":" + pkg.version)
            cmd = "docker pull {}".format(registry_path)
            subprocess2.check_output_host(cmd)
            LOG.debug("Container %s:%s successfully fetched.", pkg.name, pkg.version)

        self._cleanup_history(name)
        package = self.packages.exact_version(name, version)
        storage_dir = os.path.join(self.storage_dir, name, package.hash)
        if not os.path.exists(storage_dir):
            os.makedirs(storage_dir)

        deps = (
            [self.packages.exact_version(dep["name"], dep["version"]) for dep in deps]
            if deps
            else []
        )
        for p in deps + [package]:
            _fetch(p)

        self._update_fetch_history(storage_dir, package, deps)
        return package.hash

    def search(self, package):
        """Search for a package from the list obtained on updatedb."""
        try:
            return super(DockerManager, self).search(package)
        except PackageNotFoundError:
            err_msg = (
                "Docker image {image_name} can't be found in the registry '{registry_url}'"
            ).format(
                image_name=package,
                registry_url="/".join([self.registry_path, self.namespace, self.image]),
            )
            six.reraise(PackageNotFoundError, PackageNotFoundError(err_msg), sys.exc_info()[2])


def get_pkgmgr(repo_type):
    """
    A pkgutil factory. Returns a class for the desired platform.


    :param repo_type: What type of pkgutil we need
    :type repo_type: str

    :returns: A class of pkgutil
    :rtype: class
    """
    if repo_type in ("suse", "rpm"):
        return RpmManager
    elif repo_type == "win":
        return MsiManager
    elif repo_type == "deb":
        return DpkgManager
    elif repo_type == "docker":
        return DockerManager
    else:
        raise Error("Unsupported type of the pkgmgr: {}".format(repo_type))


# pylint: disable=line-too-long
if __name__ == "__main__":
    # APT-pool
    # apt = DkpgManager("deb http://stridercd.scalr-labs.com/apt/develop feature-omnibus-integration main")
    # apt.updatedb()
    # print apt.status('scalarizr')
    # hash = apt.fetch('scalarizr', version='3.1.b7073.271f12f-1', deps=[{'name': 'scalarizr-ec2', 'version': '3.1.b7073.271f12f-1'}])
    # apt.install(hash)
    # print apt.status('scalarizr')

    # APT-plain
    # apt = DpkgManager("deb http://stridercd.scalr-labs.com/apt-plain stable")
    # apt.updatedb()
    # print apt.status('scalarizr')
    # hash = apt.fetch('scalarizr', version)
    # apt.install(package.hash)
    # print apt.status('scalarizr')

    # RPM
    rpm = RpmManager("http://rpm.scalr.net/rpm/rhel/latest/x86_64/")
    rpm.updatedb()
    print(rpm.status("python-argparse"))
    pkg_hash = rpm.fetch("python-argparse", version="1.0.1")
    rpm.install(pkg_hash)
    print(rpm.status("python-argparse"))
    rpm.uninstall("python-argparse")
