"""
JSON-RPC over HTTP.
"""
import abc
import binascii
import hashlib
import hmac
import json
import posixpath
import sys
import time
from threading import local

import requests

from common.utils import cryptoutil
from common.utils import exc
from common.utils import logutil
from common.utils import mixutil
from common.utils import six

LOG = logutil.get_logger(__name__)


class Error(exc.Error):
    """General rpcutil error."""


class ServiceError(Error):
    """General service error."""

    NAMESPACE_NOT_FOUND = -32099
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_PARAMS = -32602
    INTERNAL = -32603
    PARSE = -32700

    code = None
    message = None

    def __init__(self, data):
        self.data = data
        super(ServiceError, self).__init__()

    def as_dict(self):
        """Formats service error as a dict."""
        return {"code": self.code, "message": self.message, "data": self.data}

    def __str__(self):
        return repr(self.as_dict())


class NamespaceNotFoundError(ServiceError):
    """When service namespace is not found."""

    code = ServiceError.NAMESPACE_NOT_FOUND
    message = "Namespace not found"


class MethodNotFoundError(ServiceError):
    """When api method is not found."""

    code = ServiceError.METHOD_NOT_FOUND
    message = "Method not found"


class ParseError(ServiceError):
    """When reqeust can't be parsed."""

    code = ServiceError.PARSE
    message = "Parse error"


class InvalidRequestError(ServiceError):
    """When request is wrongly composed."""

    code = ServiceError.INVALID_REQUEST
    message = "Invalid Request"


class InvalidParamsError(ServiceError):
    """When request has invalid parameters."""

    code = ServiceError.INVALID_PARAMS
    message = "Invalid params"


class InternalError(ServiceError):
    """When an error is encountered during request processing."""

    code = ServiceError.INTERNAL
    message = "Internal error"


class Security(object):
    """Helper class that implements encrypting/decrypting and signing of requests."""

    DATE_FORMAT = "%a %d %b %Y %H:%M:%S UTC"

    def __init__(self, crypto_key_path=None, crypto_key=None):
        """
        :param crypto_key_path str: Path to a crypto key on disk.
        :param crypto_key str: Actual crypto_key.
        """
        assert (
            crypto_key_path is not None or crypto_key is not None
        ), "You must specifiy either crypto_key_path or crypto_key"
        self.crypto_key_path = crypto_key_path
        self._crypto_key = crypto_key

    @property
    def crypto_key(self):
        """Load from file or return saved in memory crypto key."""
        if self.crypto_key_path:
            with open(self.crypto_key_path) as f:
                crypto_key = f.read().strip()
        else:
            crypto_key = self._crypto_key
        return binascii.a2b_base64(crypto_key)

    def _adjust_data(self, data):
        """Converts data to bytes for crypto operations if required.

        :param data: String, bytes or json dict data
        :type data: str or bytes or dict
        :return: bytes
        """
        if not isinstance(data, bytes):
            data = (
                data.encode("utf-8")
                if isinstance(data, six.string_types)
                else json.dumps(data).encode("utf-8")
            )
        return data

    def sign(self, data, timestamp=None):
        """Sign json data with corresponding crypto key.

        Seed timestamp optionally (will be generated otherwise).

        In regular use cases this method should not be used directly,
        but called by exchange instead.

        :param data: A string(or bytes) containing data to be signed.
        :type data: str or bytes
        :param timestamp: an optional timestamp string to be used for seeding sha generation.
        :type timestamp: str or None

        :rtype tuple(str, str)
        """
        str_date = time.strftime(self.DATE_FORMAT, timestamp or time.gmtime())
        date = str_date.encode(encoding="utf-8")
        data = self._adjust_data(data)

        canonical_string = b"".join((data, date))
        digest = hmac.new(self.crypto_key, canonical_string, hashlib.sha1).digest()
        signature = binascii.b2a_base64(digest)

        if signature.endswith(b"\n"):
            signature = signature[:-1]
        return signature.decode(), str_date

    def check_signature(self, signature, data, timestamp):
        """Check data signature.

        :param str signature: Expected signature
        :param str data: Data to sign and check
        :param str timestamp: Timestamp used to generate original signature
        """
        calc_signature, _ = self.sign(data, time.strptime(timestamp, self.DATE_FORMAT))
        assert signature == calc_signature, "Signature doesn't match"

    def decrypt_data(self, data):
        """Decrypt data using Security instance crypto_key."""
        try:
            return cryptoutil.decrypt_scalarizr(self._adjust_data(data), self.crypto_key)
        except Exception as e:
            raise InvalidRequestError("Failed to decrypt data. Error: {}".format(six.text_type(e)))

    def encrypt_data(self, data):
        """Encrypt data using Security instance crypto_key."""
        try:
            return cryptoutil.encrypt_scalarizr(self._adjust_data(data), self.crypto_key)
        except Exception as e:
            raise InvalidRequestError("Failed to encrypt data. Error: {}".format(six.text_type(e)))


@six.add_metaclass(abc.ABCMeta)
class ABCRequestHandler(object):
    """ABC for a request handler suitable for use with our RPC server."""

    @abc.abstractmethod
    def handle_request(self, data, namespace):
        """Request handling logic."""


class Server(object):
    """Implements server-side remote procedure call(RPC) communication routines."""

    def __init__(self, request_handler, crypto_key_path):
        self.request_handler = request_handler
        self.security = Security(crypto_key_path=crypto_key_path)

    def __call__(self, environ, start_response):
        try:
            try:
                length = int(environ.get("CONTENT_LENGTH"))
                data = environ["wsgi.input"].read(length)

                self.security.check_signature(
                    environ["HTTP_X_SIGNATURE"], data, environ["HTTP_DATE"]
                )
                data = self.security.decrypt_data(data)
            except Exception as e:
                start_response("400 Bad request", [], sys.exc_info())
                return [six.text_type(e)]

            result = self.request_handler.handle_request(
                data, namespace=environ["PATH_INFO"][1:] or None
            )
            result = self.security.encrypt_data(result)
            signature, date = self.security.sign(result)
            headers = [
                ("Content-type", "application/json"),
                ("Content-length", str(len(result))),
                ("X-Signature", signature),
                ("Date", date),
            ]
            start_response("200 OK", headers)
            return [result]
        except Exception as e:
            start_response("500 Internal Server Error", [], sys.exc_info())
            LOG.exception("Unhandled exception")
            return [six.text_type(e)]


class Client(object):
    """Implements client-side remote procedure call (RPC) communication routines.

    Example:

        self.client = rpcutil.Client(
            'http://update.scalr.net:8011',
            '<crypto key>',
            '7e1b6c89-b371-4192-a717-e4e5694a84b0')
        params = {
            'server_id': '7e1b6c89-b371-4192-a717-e4e5694a84b0',
            'version': '4.1.14-1',
            'package': 'scalarizr-ec2'
        }
        update_allowed = self.client.update_allowed(**params)

    """

    def __init__(
        self,
        endpoint,
        crypto_key,
        server_id=None,
        sign_only=False,
        headers=None,
        timeout=None,
        proxies=None,
    ):
        """Creates RPC client object.

        :param str endpoint: a server url to communicate with
        :param str crypto_key: a key to use for communication with Update server
        :param server_id: a uniquie identificator of Scalr-manageable node
        :param proxies: dict of http/https proxies
        :type server_id: str or None
        """
        self.local = local()
        self.endpoint = endpoint
        self.server_id = server_id
        self.sign_only = sign_only
        self.security = Security(crypto_key=crypto_key)
        self.headers = headers
        self.timeout = timeout
        self.proxies = proxies

    def __getattr__(self, name):
        try:
            self.__dict__["local"].method.append(name)
        except AttributeError:
            self.__dict__["local"].method = [name]
        return self

    def __call__(self, **kwargs):
        try:
            resp = self.exchange(method=self.local.method[-1], params=kwargs)
            return resp["result"]
        finally:
            self.local.method = []

    def __repr__(self):
        return "<rpcutil.Client endpoint={!r}, server_id={!r}>".format(
            self.endpoint, self.server_id
        )

    def prepare_request(self, method, params=None):
        """
        Prepares RPC request.
        :param dict params: params of RPC request.
        :return dict: requests.post() params in a dict format:
            {'url': <url>,
             'data': <request_data>,
             'headers': <headers>,
             'timeout': <timeout>,
             'proxies': <proxies>}
        """
        id_ = time.time()
        jsonrpc_req = json.dumps({"method": method, "params": params or {}, "id": id_})
        if not self.sign_only:
            jsonrpc_req = self.security.encrypt_data(jsonrpc_req)
        signature, date = self.security.sign(jsonrpc_req)
        headers = {"Date": date, "X-Signature": signature}
        if self.server_id:
            headers["X-Server-Id"] = self.server_id
        if self.headers:
            headers.update(self.headers)
        namespace = self.local.method[0] if len(self.local.method) > 1 else ""
        url = posixpath.join(self.endpoint, namespace)

        return {
            "url": url,
            "data": jsonrpc_req,
            "headers": headers,
            "timeout": self.timeout,
            "proxies": self.proxies,
            "id": id_,
        }

    def decode_response(self, resp):
        """
        Decodes RPC response.
        :param (str|requests.Response) resp: Response from RPC server.
        :return dict: response decoded and translate into dict.
        """
        if isinstance(resp, requests.Response):
            jsonrpc_resp = resp.content
        else:
            jsonrpc_resp = resp
        try:
            if not self.sign_only:
                jsonrpc_resp = self.security.decrypt_data(jsonrpc_resp)
            if isinstance(jsonrpc_resp, bytes):
                jsonrpc_resp = jsonrpc_resp.decode()
            try:
                return json.loads(jsonrpc_resp)
            except ValueError as e:
                raise Error("Cannot understand server response: {!r}".format(e))
        except Exception:
            if isinstance(resp, requests.Response):
                LOG.debug(
                    "Response Headers: %s\nResponse Body:%s",
                    resp.headers,
                    mixutil.ellipsis(resp.content, 1000),
                )
            raise

    def exchange(self, method, params):
        """
        Perform a request/reply exchange
        :param str params: request parameters.
        :return dict:
        """
        request_data = self.prepare_request(method, params)
        resp = requests.post(
            request_data["url"],
            data=request_data["data"],
            headers=request_data["headers"],
            timeout=request_data["timeout"],
            proxies=request_data["proxies"],
        )
        decoded_resp = self.decode_response(resp)
        if "error" in decoded_resp:
            rpc_request = {"method": method, "params": params, "id": request_data["id"]}
            LOG.debug(
                "Call failed %s. Request -> %s. Response <- %s", self, rpc_request, decoded_resp
            )
            error = decoded_resp["error"]
            raise Error(message=error.get("message"))
        return decoded_resp
