import os
import sys
import threading

import psutil

from common.utils import logutil
from common.utils import six
from common.utils.facts import fact

mswindows = sys.platform == "win32"
if six.PY2 and not mswindows:
    # pylint: disable=import-error
    import subprocess32 as subprocess
else:
    # PY3 and PY2 Windows
    import subprocess


LOG = logutil.get_logger(__name__)
PIPE = subprocess.PIPE
STDOUT = subprocess.STDOUT
DEVNULL = object()


if not mswindows:
    from subprocess import select

    _has_poll = hasattr(select, "poll")
    _PIPE_BUF = getattr(select, "PIPE_BUF", 512)

term = None


class CompletedProcess(object):
    """A process that has finished running.
    This is returned by run().
    Attributes:
      args: The list or str args passed to run().
      returncode: The exit code of the process, negative for signals.
      stdout: The standard output (None if not captured).
      stderr: The standard error (None if not captured).

    Backported from Python 3.5.
    """

    def __init__(self, args, returncode, cwd=None, stdout=None, stderr=None):
        self.args = args
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.cwd = cwd

    def __repr__(self):
        args = ["args={!r}".format(self.args), "returncode={!r}".format(self.returncode)]
        if self.stdout is not None:
            args.append("stdout={!r}".format(self.stdout))
        if self.stderr is not None:
            args.append("stderr={!r}".format(self.stderr))
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def check_returncode(self):
        """Raise CalledProcessError if the exit code is non-zero."""
        if self.returncode:
            raise CalledProcessError(self.returncode, self.args, self.cwd, self.stdout, self.stderr)


class CalledProcessError(subprocess.CalledProcessError):
    """
    Augment the standard exception with more data.
    """

    # exit codes meanings manual http://www.tldp.org/LDP/abs/html/exitcodes.html
    EX_CMD_CANNOT_EXEC = 126
    EX_CMD_NOT_FOUND = 127
    # signals man http://man7.org/linux/man-pages/man7/signal.7.html
    EX_SIGNAL_BASE = 128
    EX_SIGINT = 130
    EX_SIGKILL = 137
    EX_SIGNAL_NUM = 27

    def __init__(self, returncode, cmd, cwd, stdout, stderr):
        super(CalledProcessError, self).__init__(returncode, cmd)
        self.stdout = stdout
        self.stderr = stderr
        self.cwd = cwd

    def __str__(self):
        # pylint: disable=no-member
        err_msg = "Command {0!r} returned non-zero exit status {1}"
        err_details = None
        if self.returncode == self.EX_CMD_CANNOT_EXEC:
            err_details = (
                "Command {0!r} cannot execute. "
                "Permission problem or command is not an executable"
            )
        elif self.returncode == self.EX_CMD_NOT_FOUND:
            err_details = "Command {0!r} not found"
        elif self.returncode == self.EX_SIGINT:
            err_details = "Command {0!r} terminated by interrupt from keyboard"
        elif self.returncode == self.EX_SIGKILL:
            err_details = "Command {0!r} terminated by kill signal"
        elif self.returncode > self.EX_SIGNAL_BASE and self.returncode <= (
            self.EX_SIGNAL_BASE + self.EX_SIGNAL_NUM
        ):
            err_details = "Command {0!r} received fatal error signal " + (
                "{}".format(self.returncode - self.EX_SIGNAL_BASE)
            )

        if err_details:
            err_msg += " (Bash: " + err_details + ")"

        result = six.u(err_msg).format(self.cmd, self.returncode)
        if self.cwd:
            result += " in " + self.cwd
        result += six.u("\nstdout: {0}\nstderr: {1}").format(self.stdout, self.stderr)
        return result


class TimeoutExpired(Exception):

    """This exception is raised when the timeout expires while waiting for a
    child process.
    Backported from Python3.3
    """

    def __init__(self, cmd, timeout, stdout, stderr):
        # pylint: disable=super-init-not-called
        self.cmd = cmd
        self.timeout = timeout
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        # pylint: disable=no-member
        result = six.u("Command {!r} timed out after {} seconds").format(self.cmd, self.timeout)
        result += six.u("\nstdout: {0}\nstderr: {1}").format(self.stdout, self.stderr)
        return result


class Popen(subprocess.Popen):

    """
    Extend stdlib's subprocess.Popen with ability to report progress.
    Accepts two additional kwargs in a constructor:
        - progress_cb:
                a function to be called each config.workflow_engine['progress_tick'] and report
                correspondingly(see progress_cb function below in this module)
        - signal:
            if set to True - creates a process probing tread, which will send
            a SIGUSR1 to process described by self.args every
            config.workflow_engine['progress_tick'] seconds.
            This is required for processes having no progress output by default
    """

    def __init__(self, *args, **kwds):
        self.terminated = False
        if not mswindows:
            env = kwds.get("env") or os.environ
            apply_english_env(env)
            kwds["env"] = env
        if isinstance(args, six.string_types):
            self.cmd = args
        else:
            self.cmd = subprocess.list2cmdline(args)

        for name in ("stdout", "stderr"):
            if kwds.get(name) == DEVNULL:
                kwds[name] = open(os.devnull, "w")

        kwds.setdefault("stdout", PIPE)
        kwds.setdefault("stderr", PIPE)

        self.progress_cb = kwds.pop("progress_cb", None)
        self.progress_signal = kwds.pop("progress_signal", None)
        self.shell = kwds.get("shell", None)
        self.cwd = kwds.get("cwd", None)
        self.args = args
        self._real_pid = None
        if self.progress_signal is not None:
            if not progress_cb:
                raise AttributeError("Cannot use progress_signal with a progress callback unset")
            if not isinstance(self.progress_signal, int):
                raise TypeError("progress_signal should be int or None")
        LOG.debug("Executing subprocess: %s", args)
        super(Popen, self).__init__(*args, **kwds)

    @property
    def real_pid(self):
        if self._real_pid:
            return self._real_pid
        if self.shell:
            cmd = "ps -ef | awk '$3 == \"%s\" {print $2}'" % self.pid
            pids = subprocess.check_output(cmd, shell=True)
            if pids and b"\n" not in pids:
                self._real_pid = int(pids)
            elif len(pids) > 1:
                pid = pids.split(b"\n")[0]
                if pid.isalnum():
                    self._real_pid = int(pid)
        if not self._real_pid:
            self._real_pid = self.pid
        return self._real_pid

    def communicate_with_timer(self, input=None, timeout=None):
        # pylint: disable=redefined-builtin
        timed_out = False

        if timeout is None:
            out, err = self.communicate(input=input)
            return out, err, timed_out

        timer = threading.Timer(timeout, self.terminate, kwargs=dict(recursive=True))
        try:
            timer.start()
            out, err = self.communicate(input=input)
            timer.cancel()
            timer.join()
        finally:
            if self.terminated:
                timed_out = True
                out, err = self.communicate()
        return out, err, timed_out

    def run(self, timeout=None, input=None):  # pylint: disable=redefined-builtin
        """
        Emulates subprocess.run(), supports timeouts.

        Note that if you want to send data to the process's stdin, you need to create the Popen object with stdin=PIPE.
        Similarly, to get anything other than None in the result tuple, you need to give stdout=PIPE and/or stderr=PIPE too.

        Retruns CompletedProcess.
        """
        timed_out = False
        if input is not None and not isinstance(input, six.binary_type):
            input = input.encode("utf-8")
        if mswindows:
            out, err, timed_out = self.communicate_with_timer(input=input, timeout=timeout)
        else:
            try:
                out, err = self.communicate(input=input, timeout=timeout)
            except subprocess.TimeoutExpired:
                # terminate process tree and call communicate() again to grab out/err
                timed_out = True
                self.terminate(recursive=True)
                out, err = self.communicate()
        if isinstance(out, six.binary_type):
            out = out.decode("utf-8", errors="ignore")
        if isinstance(err, six.binary_type):
            err = err.decode("utf-8", errors="ignore")
        # pylint: disable=no-member
        if out:
            LOG.debug(six.u("Subprocess stdout: {0}").format(out))
        if err:
            LOG.debug(six.u("Subprocess stderr: {0}").format(err))
        if self.returncode:
            LOG.debug(six.u("Subprocess returncode: {0}").format(self.returncode))

        if timed_out:
            raise TimeoutExpired(self.args, timeout, out, err)
        return CompletedProcess(self.args, self.returncode, self.cwd, out, err)

    def terminate(self, recursive=False):
        if not recursive:
            return super(Popen, self).terminate()
        else:
            _kill_tree(self.pid)

        self.terminated = True

    def kill(self, recursive=False):
        _kill_tree(self.pid, soft=False)


def _kill_tree(pid, soft=True):
    try:
        proc = psutil.Process(pid=pid)
    except psutil.NoSuchProcess:
        # Process has already terminated.
        return

    try:
        proc.suspend()
        for child_process in proc.children():
            _kill_tree(child_process.pid, soft=soft)
        if soft:
            proc.terminate()
        else:
            proc.kill()
    finally:
        try:
            proc.resume()
        except psutil.NoSuchProcess:
            # Either windows platform or process terminated very quickly.
            pass


def run(args, **kwds):
    """
    Emulates subprocess.run(), supports timeouts.

    :rtype: :class:`CompletedProcess`
    """
    timeout = kwds.pop("timeout", None)
    input_ = kwds.pop("input", None)
    if input_:
        kwds["stdin"] = PIPE
    proc = Popen(args, **kwds)
    return proc.run(timeout=timeout, input=input_)


def call(args, **kwds):
    """
    Emulates subprocess.call().

    Automatically convert stdout=PIPE or stderr=PIPE to DEVNULL.
    In no case they can be returned since no code path raises
    subprocess.CalledProcessError.
    """
    if not kwds.get("stdout") or kwds.get("stdout") == PIPE:
        kwds["stdout"] = DEVNULL
    if not kwds.get("stderr") or kwds.get("stderr") == PIPE:
        kwds["stderr"] = DEVNULL
    return run(args, **kwds).returncode


def check_call(args, **kwds):
    """
    Emulates subprocess.check_call().
    """
    returncode, out, err = call(args, **kwds), "", ""
    if returncode:
        raise CalledProcessError(returncode, args, kwds.get("cwd"), out, err)
    return 0


def check_output(args, **kwds):
    """
    Emulates subprocess.check_output().
    CalledProcessError also has 'stdout' and 'stderr' properties.

    :rtype: str or None
    """
    r = run(args, **kwds)
    if r.returncode:
        raise CalledProcessError(r.returncode, args, kwds.get("cwd"), r.stdout, r.stderr)
    return r.stdout


def which(program):
    if program and mswindows:
        r = run("where {}".format(program))
        if not r.returncode:
            path = r.stdout.splitlines()[0]  # first line in result
            if os.access(path, os.X_OK):
                return path
    elif program and program.startswith("/") and os.access(program, os.X_OK):
        return program
    for where in set(os.environ["PATH"].split(os.pathsep)):
        path = os.path.join(where, program)
        if os.access(path, os.X_OK):
            return path


def apply_english_env(env):
    """
    Forces LANG and/or LANGUAGE to be English.
    Forces encoding to UTF-8 for subprocesses.
    """
    for name in ("LANG", "LANGUAGE"):
        if not env.get(name, "").startswith("en"):
            env[name] = "en_US.UTF-8"


class Command(object):

    """
    Extend this base class to implement your own command builder

    Example 1

        class IptablesCmd(Command):
            binary = '/sbin/iptables'

        iptables = IptablesCmd()
        iptables. \
            args(protocol='tcp'). \
            args(match='tcp'). \
            args(dport=8014, insert='INPUT', jump='ACCEPT')

        subprocess.communicate(iptables.build())


    Example 2

        class RsyncCmd(Command):
            binary = '/usr/bin/rsync'

        rsync = RsyncCmd()
        rsync. \
            args('-a', '-v')
        for exclude in excludes:
            rsync.args(exclude=exclude)
        rsync.args(source, destination)

        proc = subprocess.Popen(rsync.build(), stdout=subprocess.PIPE)


    Example 3

        VGCreate = new_command('VGCreate', binary='/usr/bin/vgcreate')
        vgcreate = VGCreate()
        vgcreate. \
            args('-s', '8M'). \
            args('media'). \
            args(*physical_volumes)

    """

    binary = None
    long_prefix = "--"
    long_jumper = None

    def __init__(self):
        if not self.binary:
            raise ValueError(f"Missing binary for {self.__class__}")
        self._args = [self.binary]

    @classmethod
    def exists(cls):
        return which(cls.binary)

    def args(self, *args, **long_args):
        return self._format_args(args=args, long_args=long_args)

    def _format_args(self, args=None, long_args=None):
        if args:
            for arg in args:
                self._args.append(str(arg))
        if long_args:
            for name, value in list(long_args.items()):
                name = name.replace("_", "-")
                if self.long_prefix:
                    name = "{0}{1}".format(self.long_prefix, name)
                if self.long_jumper:
                    if type(value) == bool and value:
                        self._args.append(name)
                    else:
                        self._args.append("".join([name, self.long_jumper, str(value)]))
                else:
                    self._args.append(name)
                    if type(value) == bool and value:
                        continue
                    self._args.append(str(value))
        return self

    def build(self):
        return list(self._args)

    def check_output(self, **kwds):
        return check_output(self.build(), **kwds)

    def check_call(self, **kwds):
        return check_call(self.build(), **kwds)

    def call(self, **kwds):
        return call(self.build(), **kwds)

    def run(self, **kwds):
        return run(self.build(), **kwds)


class RemoteCommand(Command):
    """
    Extent subprocess2 to execute remote shell command using default ssh interface
    """

    SSH_CONTROL_SOCKET_PATH = "/tmp/.ssh-host-control-sock"
    SESSION_ARGS = [
        "-o",
        "ControlMaster=auto",
        "-o",
        "ControlPersist=5m",
        "-o",
        "ControlPath=%s" % SSH_CONTROL_SOCKET_PATH,
    ]

    def __init__(self, **kwargs):
        # pylint: disable=super-init-not-called
        self.binary = which("ssh")
        self.shell = kwargs.get("shell")
        self.hostname = kwargs.get("hostname")
        self.keep_session = kwargs.get("keep_session")
        self._args = [self.binary]
        self._args += [
            "-4",
            "-o",
            "PreferredAuthentications=publickey",
            "-o",
            "UserKnownHostsFile=/dev/null",
        ]

        options = {
            "user": ["-o", "User={}"],
            "port": ["-o", "Port={}"],
            "host_key_checking": ["-o", "StrictHostKeyChecking={}"],
            "private_key_file": ["-o", "IdentityFile={}"],
            "config_file": ["-F", "{}"],
            "subsystem": ["-s", "{}"],
        }

        if self.keep_session:
            self._args += self.SESSION_ARGS

        for k, v in kwargs.items():
            if k in options:
                self._args += [arg.format(v) for arg in options[k]]

        self._args += [self.hostname]
        self._create_socket()

    def _create_socket(self):
        if self.keep_session:
            if not os.path.exists(self.SSH_CONTROL_SOCKET_PATH):
                call(self._args + [" "])

    def build(self):
        if self.shell:
            return " ".join(self._args)
        else:
            return list(self._args)

    def check_output(self, **kwds):
        if self.keep_session:
            self._create_socket()
        return super(RemoteCommand, self).check_output(**kwds)


def new_command(name, **kwds):
    return type(name, (Command,), kwds)


def progress_cb(stdout, stderr):
    """
    This is an example fn to be passed to extended Popen constructor.
    """


def check_output_host(*args, **kwargs):
    """
    Interface to execute system calls
    :param args: called arguments, can be list or just string
    :return: stdout
    """
    if fact["manage_container_host"]:
        rsh = RemoteCommand(
            hostname=fact["routes"]["default"]["gw"],
            host_key_checking="no",
            shell=kwargs.get("shell"),
            keep_session=True,
        )
        rsh.args(*args)
        return rsh.check_output(**kwargs)
    else:
        return check_output(args, **kwargs)


def call_host(*args, **kwargs):
    """
    Interface to execute system calls
    :param args: called arguments, can be list or just string
    :return: exit code
    """
    if fact["manage_container_host"]:
        rsh = RemoteCommand(
            hostname=fact["routes"]["default"]["gw"],
            host_key_checking="no",
            shell=kwargs.get("shell"),
            keep_session=True,
        )
        rsh.args(*args)
        return rsh.call(**kwargs)
    else:
        return call(args, **kwargs)
