import io
import os
import re
import shutil
import stat

from common.utils import facts
from common.utils import six
from common.utils import subprocess2

try:
    import agent.config as config
except ImportError:
    import server.config as config
if facts.WINDOWS:
    from common.utils import winutil as _nativeutil
else:
    from common.utils import linuxutil as _nativeutil

__all__ = [
    "clean_dir",
    "create_file",
    "get_boot_id",
    "is_binary",
    "mkdir_p",
    "reboot",
    "rm_rf",
    "symlink",
    "tail",
    "tail_s",
    "touch",
    "unlink",
    "uptime",
    "sanitize_filename",
    "has_partition_table",
    "get_fstype",
]

has_partition_table = _nativeutil.has_partition_table
"""
.. py:function:: has_partition_table()

   :return: True if the device has partition table, otherwise False.
   :rtype: bool
"""

get_fstype = _nativeutil.get_fstype
"""
.. py:function:: get_fstype()

   :return: filesystem type of volume
   :rtype: str
"""

uptime = _nativeutil.uptime
"""
.. py:function:: uptime()

   :return: OS uptime in seconds
   :rtype: float
"""


reboot = _nativeutil.reboot
"""
.. py:function:: reboot()

   Reboot operation system
"""


get_boot_id = _nativeutil.get_boot_id
"""
.. py:function:: get_boot_id()

   :return: OS current boot ID
   :rtype: str
"""


mounts = _nativeutil.mounts
"""
.. py:function:: mounts()
   :return: List of system mountpoints.
   :rtype: list
"""

statvfs = _nativeutil.statvfs
"""
.. py:function:: statvfs(mpoints=None)
   :return: Total/free bytes for provided mountpoints.
   :rtype: dict
"""


def symlink(src, dst, target_is_directory=False, dir_fd=None):
    """
    Create a symbolic link pointing to src named dst.

    See https://docs.python.org/3/library/os.html#os.symlink
    """
    if six.PY3:
        return os.symlink(src, dst, target_is_directory=target_is_directory, dir_fd=dir_fd)
    else:
        if facts.WINDOWS:
            is_dir = "/D" if target_is_directory else ""
            subprocess2.check_call(
                'cmd.exe /c mklink {} "{}" "{}"'.format(is_dir, dst, src), shell=True
            )
        else:
            return os.symlink(src, dst)


def unlink(path):
    """
    os.unlink() that also works right on Windows
    """
    if os.path.exists(path):
        if facts.WINDOWS and os.path.isdir(path):
            # see FAM-151
            os.rmdir(path)
        else:
            os.unlink(path)


def rm_rf(path):
    """
    `rm -rf` for Python
    """

    def onerror(func, path, exc_info):
        """shutil.rmtree onerror handler."""
        # pylint: disable=misplaced-bare-raise
        if not os.access(path, os.W_OK):
            # Is the error an access error?
            os.chmod(path, stat.S_IWUSR)
            func(path)
        else:
            raise

    if os.path.exists(path) or os.path.lexists(path):
        if os.path.islink(path):
            unlink(path)
        elif os.path.isdir(path):
            shutil.rmtree(path, onerror=onerror)
        else:
            os.remove(path)


def rm_rf_silent(path):
    """
    `rm -rf` for Python. Ignore any errors.
    """
    try:
        rm_rf(path)
    except Exception:
        pass


def touch(path):
    """
    `touch` for Python
    """
    open(path, "w+").close()


def mkdir_p(path, mode=0o777):
    """
    `mkdir -p` for Python
    """
    if not os.path.exists(path):
        os.makedirs(path, mode)
    return path


def clean_dir(path):
    """
    Cleans directory contents.

    :type path: str
    :param path: Directory path to clean
    """
    if not os.path.isdir(path):
        raise IOError("Not a directory: {0}".format(path))
    for name in os.listdir(path):
        rm_rf(os.path.join(path, name))


def create_file(filename, content, mode=0o644, **variables):
    """
    File creation helper. Creates `filename` with a given `mode`
    and write `content` interpolated with `**variables` using `str.format()`

    :type filename: str
    :param filename: Filename path

    :type content: str
    :param content: Content string/template

    :type mode: int
    :param mode: File octal mode

    :return: Filename
    :rtype: str
    """
    mkdir_p(os.path.dirname(filename))
    with open(filename, "wb") as fp:
        if variables:
            content = content.format(**variables)
        fp.write(content.encode() if isinstance(content, str) else content)
    os.chmod(filename, mode)
    return filename


def is_binary(path):
    """
    Checks whether the `path` is a binary file

    :return: True when file is binary
    :rtype: bool
    """
    with open(path, "rb") as fp:
        bts = fp.read(1024)
    if bts:
        textchars = bytearray([7, 8, 9, 10, 12, 13, 27]) + bytearray(range(0x20, 0x100))
        return bool(bts.translate(None, textchars))
    return False


def tail(filename, num_bytes=None, num_lines=None, from_head=False):
    """Read up to `num_bytes` from the tail or `from_head=True`
    and return up to `num_lines` of text.

    :param filename: file path
    :type filename: str

    :param num_bytes: maximum number of bytes to read
                      (default is `agent.config.workflow_engine['tail_num_bytes']`)
    :type num_bytes: int

    :param num_lines: maximum number of lines to return from the read
                      (default is `agent.config.workflow_engine['tail_num_lines']`)
    :type num_lines: int

    :param from_head: When `True` read from the begining of the file
    :type from_head: bool

    :return: truncated file content
    :rtype: str
    """
    num_bytes = num_bytes or config.workflow_engine["tail_num_bytes"]

    with io.open(filename, "rb") as fp:
        fp.seek(0, 2)
        offset = num_bytes if num_bytes < fp.tell() else 0
        if offset != 0 and not from_head:
            fp.seek(-offset, 2)  # pylint: disable=invalid-unary-operand-type
        else:
            fp.seek(0)

        # FIXME: Find better way to decode unicode byte range
        content = fp.read(num_bytes).decode("utf-8", "ignore")
        return tail_s(
            string=content, num_lines=num_lines, from_head=from_head, force_ellipsis=offset != 0
        )


def tail_s(string, num_lines=0, from_head=False, force_ellipsis=False):
    if num_lines is None:
        num_lines = config.workflow_engine["tail_num_lines"]
    result_is_smaller = False
    if num_lines:
        lines = string.splitlines(True)
        result_is_smaller = num_lines < len(lines)
        if result_is_smaller:
            # bug in pylint
            # pylint: disable=invalid-unary-operand-type
            string = "".join(lines[:num_lines] if from_head else lines[-num_lines:])

    ellipsis_needed = force_ellipsis or result_is_smaller
    ellipsis = "...\n" if ellipsis_needed else ""
    return string + ellipsis if from_head else ellipsis + string


def sanitize_filename(filename):
    """Sanitize reserved characters from file name.

    :see: https://en.wikipedia.org/wiki/Filename
    :see: https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
    """
    return re.sub(r'[<>?*|"\\/:]+', "", filename)
