import ctypes
import ctypes.wintypes
import enum
import typing as tp

# pylint: disable=import-error
import win32api
import win32con
import win32file
import winioctlcon

from common.utils import exc

# pylint: enable=import-error


class Error(exc.Error):
    pass


class GptPartitionType(str, enum.Enum):
    """Some of the GPT partitions types.
    Though we are interested only in basic data partition type here.

    There is a direct one-to-one correlation between a basic data partition
    and a drive letter or mount point, other volume device object, or both.
    Each basic data partition is represented in Windows as a volume device object,
    and optionally as a mount point or a drive letter.
    """

    unused = "{00000000-0000-0000-0000-000000000000}"
    reserved = "{e3c9e316-0b5c-4db8-817d-f92df00215ae}"
    basic_data = "{ebd0a0a2-b9e5-4433-87c0-68b6b72699c7}"


class GUID(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/guiddef/ns-guiddef-guid>
    """

    _fields_ = [
        ("Data1", ctypes.wintypes.DWORD),
        ("Data2", ctypes.wintypes.WORD),
        ("Data3", ctypes.wintypes.WORD),
        ("Data4", ctypes.c_ubyte * 8),
    ]

    def __str__(self) -> str:
        return "{{{:08x}-{:04x}-{:04x}-{:02x}{:02x}-{:02x}{:02x}{:02x}{:02x}{:02x}{:02x}}}".format(
            self.Data1, self.Data2, self.Data3, *self.Data4
        )


class PARTITION_INFORMATION_MBR(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-partition_information_mbr>
    """

    _fields_ = [
        ("PartitionType", ctypes.wintypes.BYTE),
        ("BootIndicator", ctypes.wintypes.BOOLEAN),
        ("RecognizedPartition", ctypes.wintypes.BOOLEAN),
        ("HiddenSectors", ctypes.wintypes.DWORD),
    ]


class PARTITION_INFORMATION_GPT(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-partition_information_gpt>
    """

    _fields_ = [
        ("PartitionType", GUID),
        ("PartitionId", GUID),
        ("Attributes", ctypes.c_ulonglong),
        ("Name", ctypes.wintypes.WCHAR * 36),
    ]


class PARTITION_INFORMATION_UNION(ctypes.Union):
    """A union for MBR or GPT partition information"""

    _fields_ = [("Mbr", PARTITION_INFORMATION_MBR), ("Gpt", PARTITION_INFORMATION_GPT)]


class PARTITION_INFORMATION_EX(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-partition_information_ex>

    The size of PARTITION_INFORMATION_EX record is 144 bytes.
    """

    _fields_ = [
        ("PartitionStyle", ctypes.wintypes.BYTE),
        ("StartingOffset", ctypes.wintypes.LARGE_INTEGER),
        ("PartitionLength", ctypes.wintypes.LARGE_INTEGER),
        ("PartitionNumber", ctypes.wintypes.INT),
        ("RewritePartition", ctypes.wintypes.BOOLEAN),
        ("IsServicePartition", ctypes.wintypes.BOOLEAN),
        ("PartitionInformation", PARTITION_INFORMATION_UNION),
    ]


class DRIVE_LAYOUT_INFORMATION_MBR(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-drive_layout_information_mbr>
    """

    _fields_ = [("Signature", ctypes.wintypes.DWORD), ("CheckSum", ctypes.wintypes.DWORD)]


class DRIVE_LAYOUT_INFORMATION_GPT(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-drive_layout_information_gpt>
    """

    _fields_ = [
        ("DiskId", GUID),
        ("StartingUsableOffset", ctypes.wintypes.LARGE_INTEGER),
        ("UsableLength", ctypes.wintypes.LARGE_INTEGER),
        ("MaxPartitionCount", ctypes.wintypes.DWORD),
    ]


class DRIVE_LAYOUT_INFORMATION_UNION(ctypes.Union):
    """A union for MBR or GPT layout information"""

    _fields_ = [("Mbr", DRIVE_LAYOUT_INFORMATION_MBR), ("Gpt", DRIVE_LAYOUT_INFORMATION_GPT)]


class DRIVE_LAYOUT_INFORMATION_EX(ctypes.Structure):
    """See
    <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ns-winioctl-drive_layout_information_ex>

    Total size of this structure is 192 bytes, considering it having
    single PartitionEntry record of size 144.
    """

    _fields_ = [
        ("PartitionStyle", ctypes.wintypes.DWORD),
        ("PartitionCount", ctypes.wintypes.DWORD),
        ("DriveLayoutInformation", DRIVE_LAYOUT_INFORMATION_UNION),
        ("PartitionEntry", PARTITION_INFORMATION_EX),
    ]


def _get_drive_layout_information_ex_sized(size: int) -> tp.Type[ctypes.Structure]:
    """Given PartitionEntry is a variable-sized array of PARTITION_INFORMATION_EX structures,
    we redeclare DRIVE_LAYOUT_INFORMATION_EX once partition count is known
    to be able to unpack entire buffer into our structures.
    """

    class DRIVE_LAYOUT_INFORMATION_EX_SIZED(ctypes.Structure):
        _fields_ = [
            ("PartitionStyle", ctypes.wintypes.DWORD),
            ("PartitionCount", ctypes.wintypes.DWORD),
            ("DriveLayoutInformation", DRIVE_LAYOUT_INFORMATION_UNION),
            ("PartitionEntry", PARTITION_INFORMATION_EX * size),
        ]

    return DRIVE_LAYOUT_INFORMATION_EX_SIZED


def _as_type(ctype, buf):
    """Returns a new instance of ctype which points to the same memory block as buf"""
    ctype_instance = ctypes.cast(ctypes.pointer(buf), ctypes.POINTER(ctype)).contents
    return ctype_instance


def get_drive_layout_information(disk_drive) -> DRIVE_LAYOUT_INFORMATION_EX:
    """Returns ``DRIVE_LAYOUT_INFORMATION_EX`` structure for given disk drive.

    :type disk_drive: wmi._wmi_object - A ``Win32_DiskDrive`` WMI object
    """

    buffer = ctypes.create_string_buffer(
        ctypes.sizeof(DRIVE_LAYOUT_INFORMATION_EX) + ctypes.sizeof(PARTITION_INFORMATION_EX) * 3
    )
    # We expect single partition per disk, but reserve buffer size for additional partitions -
    # there will be one extra of 'Microsoft reserved partition' type for GPT disk
    # and for MBR disk it can return data area with 4 partition information blocks
    # (which is a primary partitions number limit for MBR partitioning style)
    # even if there is actually only one, buffer for other three will be just zeroed.
    #
    # The code supports unpacking any number of partitions to structure.

    handle = win32file.CreateFile(
        disk_drive.DeviceId,
        win32file.GENERIC_READ | win32file.GENERIC_EXECUTE,
        win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,
        None,
        win32file.OPEN_EXISTING,
        win32file.FILE_ATTRIBUTE_NORMAL,
        None,
    )

    if not handle:
        # handle is positive integer in case of success
        raise Error(f"Failed creating {disk_drive.DeviceId} handle")

    try:
        ret = win32file.DeviceIoControl(
            handle, winioctlcon.IOCTL_DISK_GET_DRIVE_LAYOUT_EX, None, buffer, None
        )
    except Exception as err:
        raise Error(f"Failed performing DeviceIoControl WinAPI call. {err}")
    finally:
        win32api.CloseHandle(handle)  # pylint: disable=no-member

    if not ret:
        # if the operation completes successfully, the return value is nonzero
        raise Error("Failed performing DeviceIoControl WinAPI call")

    layout_info = _as_type(DRIVE_LAYOUT_INFORMATION_EX, buffer)
    # We got partitions count now, so we can extend our structure
    # to hold all PartitionEntry records
    layout_info = _as_type(
        _get_drive_layout_information_ex_sized(layout_info.PartitionCount), buffer
    )

    return layout_info
