"""
Windows utils.
"""
from __future__ import absolute_import

import binascii
import contextlib
import enum
import itertools
import logging
import multiprocessing.pool
import os
import re
import threading
import time
import uuid

# pylint: disable=import-error
import ntsecuritycon
import pythoncom
import win32api
import win32security
import winerror
import wmi

from common import types
from common.utils import exc
from common.utils import mixutil
from common.utils import six
from common.utils import subprocess2
from common.utils import win32native

# pylint: enable=import-error


LOG = logging.getLogger(__name__)

DISKPART_POLL_INTERVAL = 1
DISKPART_RETRIES_NUM = 10
DISKPART_UNAVAIL_ERR = -2146959355


class DiskPartitionStyle(enum.Enum):
    """Partition styles enumeration.

    See <https://docs.microsoft.com/en-us/windows/win32/api/winioctl/ne-winioctl-partition_style>
    """

    mbr = 0
    gpt = 1
    raw = 2


@contextlib.contextmanager
def coinitialized_context():
    need_init = threading.current_thread().name != "MainThread"
    if need_init:
        pythoncom.CoInitialize()
    try:
        yield
    finally:
        if need_init:
            pythoncom.CoUninitialize()


def coinitialized(fn):
    @six.wraps(fn)
    def decorator(*args, **kwargs):
        with coinitialized_context():
            return fn(*args, **kwargs)

    return decorator


@contextlib.contextmanager
def adjusted_priveleges(priveleges):
    flags = ntsecuritycon.TOKEN_ADJUST_PRIVILEGES | ntsecuritycon.TOKEN_QUERY
    htoken = win32security.OpenProcessToken(win32api.GetCurrentProcess(), flags)
    idd = win32security.LookupPrivilegeValue(None, priveleges)
    enabled_priveleges = [(idd, ntsecuritycon.SE_PRIVILEGE_ENABLED)]
    win32security.AdjustTokenPrivileges(htoken, 0, enabled_priveleges)

    try:
        yield
    finally:
        idd = win32security.LookupPrivilegeValue(None, priveleges)
        disabled_priveleges = [(idd, 0)]
        win32security.AdjustTokenPrivileges(htoken, 0, disabled_priveleges)


@coinitialized
def uptime():
    """
    Implements `common.utils.sysutil.uptime()` on Windows.
    """
    # pylint: disable=W0603
    win_os = wmi.WMI().Win32_OperatingSystem()[0]
    local_time, tz_op, tz_hh60mm = re.split(r"(\+|\-)", win_os.LastBootUpTime)
    local_time = local_time.split(".")[0]
    local_time = time.mktime(time.strptime(local_time, "%Y%m%d%H%M%S"))
    tz_seconds = int(tz_hh60mm) * 60
    if tz_op == "+":
        return time.time() - local_time + tz_seconds
    else:
        return time.time() - local_time - tz_seconds


@coinitialized
def reboot(reason=""):
    """
    Implements `common.sysutil.reboot()` on Windows.
    """
    wmi_svc = wmi.WMI()
    with adjusted_priveleges(ntsecuritycon.SE_SHUTDOWN_NAME):
        current_os = wmi_svc.Win32_OperatingSystem()[0]
        current_os.Reboot(None, reason, 0, 1, 1)


def wait_boot():
    """
    In a threadpool wait for registry keys:
    - hklm:\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Setup\\State ImageState=IMAGE_STATE_COMPLETE
    - hklm:\\SYSTEM\\Setup\\Status\\SysprepStatus GeneralizationState=7

    See:
     - https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/how-configuration-passes-work
     - ImageState: https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/windows-setup-states
     - GeneralizationState: https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/update-windows-settings-and-scripts-create-your-own-answer-file-sxs#answer-file-settings

    Note: The system MAY reboot while polling the GeneralizationState (4 -> 7), but ImageState will
        only reach `IMAGE_STATE_COMPLETE` after a reboot.
    """
    winreg = six.moves.winreg

    def poll_winreg(key, sub_key, value_name, value):
        seen = set()
        LOG.debug("Winreg waiting until %s=%s (in key=%s)", value_name, value, sub_key)
        prev = None
        hk = None
        while True:
            try:
                if not hk:
                    hk = winreg.OpenKey(key, sub_key, 0, winreg.KEY_READ)
                v = winreg.QueryValueEx(hk, value_name)[0]
                if v != prev:
                    LOG.debug("Winreg seen %s=%s", value_name, v)
                    prev = v
                seen.add(v)
                if v == value:
                    return seen
                time.sleep(1)
            except WindowsError as e:  # pylint: disable=undefined-variable
                # On Windows 2016 we sometimes get ERROR_WRITE_PROTECT error here.
                # We need to wait until Registry becomes readable
                if e.winerror == winerror.ERROR_WRITE_PROTECT:
                    wait_seconds = 1
                    LOG.debug(
                        (
                            "Got %s while polling winreg. We need to wait until registry becomes readable. "
                            "Retry after %s second(s)"
                        ),
                        e,
                        wait_seconds,
                    )
                    time.sleep(wait_seconds)
                elif e.winerror == winerror.ERROR_FILE_NOT_FOUND:
                    LOG.warning("Got %s while polling winreg.", e)
                    return seen
                else:
                    raise e
            finally:
                if hk:
                    hk.Close()

    pool = multiprocessing.pool.ThreadPool(processes=2)
    try:
        image_state = pool.apply_async(
            poll_winreg,
            args=(
                winreg.HKEY_LOCAL_MACHINE,
                "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Setup\\State",
                "ImageState",
                "IMAGE_STATE_COMPLETE",
            ),
        )
        gen_state = pool.apply_async(
            poll_winreg,
            args=(
                winreg.HKEY_LOCAL_MACHINE,
                "SYSTEM\\Setup\\Status\\SysprepStatus",
                "GeneralizationState",
                7,
            ),
        )
    finally:
        pool.close()
    image_state.get()
    gen_state.get()


@coinitialized
def get_boot_id():
    """
    Implements `common.sysutil.get_boot_id()` on Windows.
    """
    win_os = wmi.WMI().Win32_OperatingSystem()[0]
    return win_os.LastBootUpTime


def get_hklm_key(sub_key, value_name):
    """
    Get a value of existing HKLM key.

    :param str sub_key:  path to reg 'directory' containing the key.
    :param str value_name:  key name.
    """
    winreg = six.moves.winreg
    with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, sub_key, 0, winreg.KEY_READ) as registry_key:
        return winreg.QueryValueEx(registry_key, value_name)[0]


def get_hklm_values(sub_key):
    """
    Return dict with key-value data from HKLM registry key.

    :param str sub_key: path to registry key.
    """
    winreg = six.moves.winreg
    result = {}
    all_keys = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, sub_key, 0, winreg.KEY_ALL_ACCESS)
    try:
        for i in itertools.count():
            name, value, _ = winreg.EnumValue(all_keys, i)
            result[name] = value
    except WindowsError:  # pylint: disable=undefined-variable
        pass
    return result


def call_diskpart(stdin):
    """
    Calls diskpart with given script passed through stdin.

    If it returns strange error code ``DISKPART_UNAVAIL_ERR``,
    retries the call.

    :rtype: :class:`subprocess2.CompletedProcess`
    """

    def _do_diskpart():
        completed_process = subprocess2.run(["diskpart"], input=stdin)
        if completed_process.returncode != DISKPART_UNAVAIL_ERR:
            return completed_process

    return mixutil.try_until(
        _do_diskpart,
        interval=DISKPART_POLL_INTERVAL,
        times=DISKPART_RETRIES_NUM,
        error_msg="diskpart cannot complete the task",
    )


def _get_disk_drive_attribute(scsi_address, attr_name):
    """
    Returns attribute value of `Win32_DiskDrive`.

    :type scsi_address: SCSI address of disk drive.
    :param scsi_address: :class:`types.SCSIAddress`

    :type attr_name: str
    :param attr_name: Attribute to take value from
    """
    with coinitialized_context():
        disk = _scsi_address_to_disk_drive(scsi_address)
        if disk:
            return getattr(disk, attr_name)
        return None


def _get_logical_disk_attribute(scsi_address, attr_name):
    """
    Returns attribute value of ``Win32_LogicalDisk``.

    :type scsi_address: SCSI address of disk drive.
    :param scsi_address: :class:`types.SCSIAddress`

    :type attr_name: str
    :param attr_name: attribute to take value from
    """
    with coinitialized_context():
        logical_disk = _scsi_address_to_logical_disk(scsi_address)
        return getattr(logical_disk, attr_name)


def _get_volume_attribute(scsi_address, attr_name):
    """
    Returns attribute value of ``Win32_Volume``.

    :type scsi_address: SCSI address of disk drive.
    :param scsi_address: :class:`types.SCSIAddress`

    :type attr_name: str
    :param attr_name: attribute to take value from
    """
    with coinitialized_context():
        volume = _scsi_address_to_volume(scsi_address)
        return getattr(volume, attr_name)


def _get_disk_drive_partition_style(disk_index: int) -> DiskPartitionStyle:
    """Get partitioning style by disk index - GPT or MBR.

    It is supposed that disk already has created partition.
    """
    wmi_api = wmi.WMI()
    part_type = wmi_api.Win32_DiskPartition(Index=0, DiskIndex=disk_index)[0].Type
    return DiskPartitionStyle.gpt if part_type.lower().startswith("gpt") else DiskPartitionStyle.mbr


def get_partition_index(disk_index: int) -> int:
    """Returns partition index for given disk drive index to use in diskpart commands.

    We support single partition per disk, so this was supposed to be always '1',
    but as we switched to GPT partitioning style, this is not a rule anymore -
    creating partition on GPT disk always creates a hidden 'Microsoft reserved' partition
    which can not be seen via WMI, but shows up in diskpart.
    """
    return 2 if _get_disk_drive_partition_style(disk_index) == DiskPartitionStyle.gpt else 1


def _disk_drive_letter_cmd(cmd, device, letter):
    """
    Letter assignment/removing for dynamic disks with one simple volume.

    :rtype: :class:`subprocess2.CompletedProcess`
    """
    if letter is None and cmd == "remove":
        letter = "all"
    elif letter == "" or letter == "auto":
        letter = ""
    else:
        letter = "letter={}".format(letter)

    # if device given is just one letter, then try to select volume by this letter
    if isinstance(device, six.string_types) and len(device) == 1:
        select_cmd = """
            select volume {0}
            """.format(
            device
        )
    else:
        disk_index = get_disk_drive_index(device)
        part_index = get_partition_index(disk_index)
        select_cmd = """
            select disk {}
            select partition {}
            """.format(
            disk_index, part_index
        )
    commands = """{0}
           {1} {2}
           exit
        """.format(
        select_cmd, cmd, letter
    )
    return call_diskpart(stdin=commands)


def _assign_volume_letter(scsi_address, letter, raise_already_mounted=False):
    """
    Assigns letter (mpoint) to volume. If it is taken, tries to move volume
    that occupies needed letter to next available one.

    :rtype: str
    """
    if letter == get_logical_disk_letter(scsi_address):
        msg = "The letter '%s' is already assigned to Win32_Volume (scsi_address=%s)."
        if raise_already_mounted:
            raise exc.AlreadyMountedError(msg % (letter, scsi_address))
        LOG.debug(msg, letter, scsi_address)
        return

    LOG.debug("Trying to move device which taken given letter.")
    _disk_drive_letter_cmd("assign", letter, "auto")

    LOG.debug("Assigning letter '%s' to Win32_Volume (scsi_address=%s)", letter, scsi_address)
    stdout = _disk_drive_letter_cmd("assign", scsi_address, letter).stdout

    if stdout and "specified drive letter is not free to be assigned" in stdout:
        raise exc.StorageError("Letter {} is taken and can't be released.".format(letter))
    return stdout


def _remove_volume_letter(scsi_address, letter=None):
    """
    Removes volume letter (or mount point) assignment.
    """
    msg_letter = "letter {}".format(letter) if letter else "all letters"
    LOG.debug("Removing %s from Win32_Volume (scsi_address=%s).", msg_letter, scsi_address)
    output = _disk_drive_letter_cmd("remove", scsi_address, letter).stdout
    if output and "error" in output:
        raise exc.StorageError(
            "Can't remove {} from Win32_Volume (scsi_address={}).".format(msg_letter, scsi_address)
        )


def _volume_mount_point(win_32_volume):
    """
    Returns ``Win32_Volume`` mount point (drive letter or directory).

    Note: Must be called in ``coinitialized_context``.

    :rtype: str or None
    """
    win_32_volume_name = win_32_volume.Name
    if ":" in win_32_volume_name and os.path.exists(win_32_volume_name):
        return os.path.abspath(win_32_volume_name)


@coinitialized
def _assign_volume_mount_point(scsi_address, letter, directory, raise_already_mounted=False):
    """
    Add a mount point directory for ``Win32_Volume``.
    """
    drive_path = "{}:".format(letter)
    mount_path = os.path.abspath(os.path.join(drive_path, directory))

    win_32_volume = _scsi_address_to_volume(scsi_address)
    win_32_volume_mpoint = _volume_mount_point(win_32_volume)
    if win_32_volume_mpoint is not None:
        if win_32_volume_mpoint == mount_path:
            msg = "Win32Volume is already mounted to directory (scsi_address=%s, directory=%s)."
            if raise_already_mounted:
                raise exc.AlreadyMountedError(msg % (scsi_address, mount_path))
            LOG.debug(msg, scsi_address, mount_path)
            return
        else:
            _remove_volume_letter(scsi_address)

    if not os.path.exists(drive_path):
        raise exc.StorageError(
            ("Cannot mount the volume to {}. " "The specified drive '{}' does not exist.").format(
                mount_path, drive_path
            )
        )

    if os.path.exists(mount_path):
        if not os.path.isdir(mount_path):
            raise exc.StorageError(
                (
                    "Cannot mount the volume to {}. " "A file with the same name already exists."
                ).format(mount_path)
            )
        if os.listdir(mount_path):
            raise exc.StorageError(
                ("Cannot mount the volume to {}. Directory is not empty.").format(mount_path)
            )
    else:
        os.makedirs(mount_path)

    LOG.debug("Create Win32_MountPoint (scsi_address=%s, directory=%s).", scsi_address, mount_path)
    win_32_volume.AddMountPoint(mount_path)


def _scsi_address_to_disk_drive(scsi_address):
    """
    Returns Win32_DiskDrive (wmi object) instance for corresponding SCSI address.
    MUST BE CALLED AND Win32_DiskDrive RESULT USED ONLY IN coinitialized_context!!!
    """
    if not isinstance(scsi_address, types.SCSIAddress):
        raise exc.StorageError(
            "Unexpected argument '{}', expect SCSIAddress object.".format(scsi_address)
        )

    LOG.debug("Searching for Win32_DiskDrive (scsi_address=%s).", scsi_address)
    wmi_api = wmi.WMI()
    try:
        return wmi_api.Win32_DiskDrive(**scsi_address.to_native())[0]
    except (KeyError, IndexError):
        raise exc.StorageError("Win32_DiskDrive (scsi_address={}) not found.".format(scsi_address))


def _scsi_address_to_logical_disk(scsi_address):
    """
    Returns ``Win32_LogicalDisk`` (wmi object) instance for corresponding SCSI address.

    Note: Must be called and ``Win32_LogicalDisk`` result used only in ``coinitialized_context``.
    """
    try:
        disk = _scsi_address_to_disk_drive(scsi_address)
        partition = disk.associators("Win32_DiskDriveToDiskPartition")[0]
        return partition.associators("Win32_LogicalDiskToPartition")[0]
    except (KeyError, IndexError):
        raise exc.StorageError(
            "Win32_LogicalDisk (scsi_address={}) not found.".format(scsi_address)
        )


def _scsi_address_to_volume(scsi_address):
    """
    Returns ``Win32_Volume`` (wmi object) instance for corresponding SCSI address.

    Note: Must be called and ``Win32_Volume`` result used only in ``coinitialized_context``.
    """
    if not isinstance(scsi_address, types.SCSIAddress):
        raise exc.StorageError(
            ("Unexpected argument '{}', expect 'types.SCSIAddress'.").format(scsi_address)
        )

    LOG.debug("Searching for Win32_Volume (scsi_address=%s).", scsi_address)
    wmi_api = wmi.WMI()
    volume_uid_pattern = re.compile(r"Volume{([A-Za-z0-9\-]+)}")
    try:
        disk_drive = wmi_api.Win32_DiskDrive(**scsi_address.to_native())[0]
        volumes_map = {
            re.findall(volume_uid_pattern, volume.DeviceID)[0]: volume
            for volume in wmi_api.Win32_Volume()
            if re.search(volume_uid_pattern, volume.DeviceID)
        }

        disk_layout_info = win32native.get_drive_layout_information(disk_drive)
        disk_partition_style = DiskPartitionStyle(disk_layout_info.PartitionStyle)
        LOG.debug(
            "Win32_DiskDrive (scsi_address=%s) is partitioned in %s style",
            scsi_address,
            disk_partition_style.name.upper(),
        )

        if disk_partition_style == DiskPartitionStyle.gpt:
            for entry in disk_layout_info.PartitionEntry:
                if (
                    str(entry.PartitionInformation.Gpt.PartitionType)
                    == win32native.GptPartitionType.basic_data
                ):
                    partition_guid = uuid.UUID(str(entry.PartitionInformation.Gpt.PartitionId))
                    return volumes_map[str(partition_guid)]
        else:
            # Fallback to old volume uid calculation algorithm for MBR drive
            partition = disk_drive.associators("Win32_DiskDriveToDiskPartition")[0]
            signature = mixutil.int_to_bytes(disk_drive.Signature, 4, byteorder="little")
            signature_big = mixutil.int_to_bytes(disk_drive.Signature, 4, byteorder="big")
            starting_offset = mixutil.int_to_bytes(partition.StartingOffset, 8, byteorder="little")

            # win-server >2012 way
            # UID of the volume (16 byte entry) comprises reversed signature of the
            # disk drive (4 bytes), empty 4 bytes and the partition offset (8 bytes).
            # Example:
            #   disk signature: 'b2 f6 b2 7e'
            #   partition offset: '00 01 00 00 00 00 00 00'
            #   result: '7e b2 f6 b2 / 00 00 00 00 / 00 01 00 00 00 00 00 00'
            volume_uid_flat = binascii.hexlify(signature_big + (b"\x00" * 4) + starting_offset)
            if six.PY3:
                volume_uid_flat = volume_uid_flat.decode("utf-8")
            volume_uid = str(uuid.UUID(volume_uid_flat))
            if volume_uid in volumes_map:
                return volumes_map[volume_uid]

            # win-server <= 2012
            # Each partition/device allocated in windows will have two entries
            # (Values) in the HKLM\SYSTEM\MountedDevices key:
            #  - \??\Volume{guid} - e.g. \??\Volume{f95cc040-a528-11de-bbbb-806e6f6e6963}
            #  - \DosDevices\#:- e.g. \DosDevices\C:
            # Both of the above Values will contain the same data - a 12 byte binary entry.
            # The first four bytes contain the disk signature of the disk containing the
            # partition, the other eight bytes represent the partition offset.
            full_signature = signature + starting_offset
            for key, value in get_hklm_values("SYSTEM\\MountedDevices").items():
                if re.search(volume_uid_pattern, key) and value == full_signature:
                    return volumes_map[re.findall(volume_uid_pattern, key)[0]]
    except (KeyError, IndexError):
        pass
    raise exc.StorageError("Win32_Volume (scsi_address={}) not found.".format(scsi_address))


@coinitialized
def get_volume_label(scsi_address):
    volume = _scsi_address_to_volume(scsi_address)
    return volume.Label


@coinitialized
def set_volume_label(scsi_address, label):
    volume = _scsi_address_to_volume(scsi_address)
    volume.Label = label


def get_logical_disk_letter(scsi_address):
    """
    Returns disk drive letter.
    """
    try:
        device_id = _get_logical_disk_attribute(scsi_address, "DeviceID")
        if device_id:
            return device_id.replace(":", "")
    except exc.StorageError:
        return None


def get_disk_drive_index(scsi_address):
    """
    Returns disk drive index.
    """
    return _get_disk_drive_attribute(scsi_address, "Index")


@coinitialized
def serial_number_to_scsi_address(serial_number):
    """
    Finds Win32_DiskDrive by its serial number. Returns SCSI address.

    :param str serial_number: Serial number of a disk.
    :return types.SCSIAddress: SCSI address of a disk.
    :raises exc.StorageError: Raises exception if disk was not found.
    """
    wmi_api = wmi.WMI()

    for disk in wmi_api.Win32_DiskDrive():
        if hasattr(disk, "AdapterSerialNumber"):
            LOG.debug("Disk has AdapterSerialNumber %s", disk.AdapterSerialNumber)
            disk_serial_number = getattr(disk, "AdapterSerialNumber")
        elif hasattr(disk, "SerialNumber"):
            disk_serial_number = getattr(disk, "SerialNumber")
        else:
            raise exc.StorageError(
                "Cannot determine Win32_DiskDrive (scsi_address={}) serial number.".format(
                    types.SCSIAddress(disk_drive=disk)
                )
            )

        if serial_number in disk_serial_number:
            return types.SCSIAddress(disk_drive=disk)

    raise exc.StorageError("Win32_DiskDrive (serial_number={}) not found.".format(serial_number))


@coinitialized
def has_partition_table(scsi_address):
    """
    Checks that `Win32_DiskDrive` have no more than one partition with one volume.

    :type scsi_address: SCSI address of disk drive.
    :param scsi_address: :class:`types.SCSIAddress`

    :rtype: bool
    """
    physical_disk = _scsi_address_to_disk_drive(scsi_address)

    partitions = physical_disk.associators("Win32_DiskDriveToDiskPartition")
    if len(partitions) == 0:
        return False
    if len(partitions) > 1:
        return True

    logical_volumes = partitions[0].associators("Win32_LogicalDiskToPartition")
    if len(logical_volumes) > 1:
        return True

    return False


def get_fstype(scsi_address):
    """
    Returns filesystem type of ``Win32_Volume``.

    :param scsi_address: SCSI address of device.
    :type scsi_address: :class:`types.SCSIAddress`

    :return: The type of the file system (example: ``ntfs``)
    :rtype: str
    """
    try:
        fs_type = _get_volume_attribute(scsi_address, "FileSystem")
    except exc.StorageError:
        return None
    if isinstance(fs_type, six.string_types):
        fs_type = fs_type.lower()
    return fs_type


def mount(scsi_address, mpoint, raise_already_mounted=False):
    """
    Assigns mount point (drive letter or directory) to the ``Win32_Volume``.

    :param scsi_address: SCSI address of the ``Win32_DiskDrive``.
    :type scsi_address: str or :class:`types.SCSIAddress`

    :param mpoint: The mount point for the ``Win32_Volume``.
    :type mpoint: str

    :param raise_already_mounted: If True, raises an error if device is already mounted.
    :type raise_already_mounted: bool
    """
    # unpack SCSI address from string format
    if isinstance(scsi_address, six.string_types):
        scsi_address = types.SCSIAddress.from_packed_address(scsi_address)

    invalid_mpoint_err = ("Mount point must be a valid path or single letter. Given: {}.").format(
        mpoint
    )

    # validate mpoint and split 'C://mount/dir/' into 'C' and '\\mount\\dir'
    if len(mpoint) > 1:
        if ":" not in mpoint:
            raise ValueError(invalid_mpoint_err)
        mpoint = os.path.abspath(mpoint)
        letter, directory = os.path.splitdrive(mpoint)
        letter = letter.rstrip(":")
    else:
        letter, directory = mpoint, None

    # validate disk letter
    if not re.match(r"^[a-zA-Z]$", letter):
        raise ValueError(invalid_mpoint_err)

    if directory:
        _assign_volume_mount_point(
            scsi_address, letter, directory, raise_already_mounted=raise_already_mounted
        )
    else:
        _assign_volume_letter(scsi_address, letter, raise_already_mounted=raise_already_mounted)


def umount(scsi_address):
    """
    Removes all mount point (drive letter or directory) assignment.

    :param scsi_address: SCSI address of the ``Win32_DiskDrive``.
    :type scsi_address: str or :class:`types.SCSIAddress`
    """
    # unpack SCSI address from string format
    if isinstance(scsi_address, six.string_types):
        scsi_address = types.SCSIAddress.from_packed_address(scsi_address)

    LOG.debug("Umounting Win32_Volume (scsi_address=%s) ...", scsi_address)
    with coinitialized_context():
        try:
            win_32_volume = _scsi_address_to_volume(scsi_address)
        except exc.StorageError:
            LOG.debug("Win32_Volume (scsi_address=%s) not found, nothing to umount.", scsi_address)
            return
        if _volume_mount_point(win_32_volume) is None:
            LOG.debug("Win32_Volume (scsi_address=%s) is not mounted.", scsi_address)
            return
        _remove_volume_letter(scsi_address)


def bring_disk_online(scsi_address):
    """
    Brings given disk online.

    :param scsi_address: SCSI address of device.
    :type scsi_address: :class:`types.SCSIAddress`

    :raises: :class:`exc.TimeoutError`
    """
    disk_index = get_disk_drive_index(scsi_address)

    LOG.debug(
        "Bringing Win32_DiskDrive (index=%s, scsi_address=%s) online", disk_index, scsi_address
    )
    commands = """select disk {index}
           online disk noerr
           attributes disk clear readonly noerr
           import noerr
           exit
        """.format(
        index=disk_index
    )
    call_diskpart(stdin=commands)


@coinitialized
def mounts():
    """Get all windows volumes names(including disks and mounted directories).

    The format of volumes:
    Disk: C:\\
    Folder: D:\\some_folder

    :return: list of all available volume names.
    """
    ret = []
    connection = wmi.WMI()
    for volume in connection.Win32_Volume():
        ret.append(volume.Name)
    return ret


@coinitialized
def statvfs(mpoints):
    """Return size/free_space for each mount point(Volume) in bytes.

    :param mpoints: iterable with mount points("C:\\", "D:\\", "C:\\Folder")
    :return: dict with mpoint as a key and dict {'free_bytes': int, 'total_bytes': int} as a
    value.
    """
    ret = {}
    connection = wmi.WMI()
    for volume in connection.Win32_Volume():
        mpoint = volume.Name
        if mpoint in mpoints:
            ret[mpoint] = {
                "total_bytes": int(volume.Capacity) if volume.Capacity else None,
                "free_bytes": int(volume.FreeSpace) if volume.FreeSpace else None,
                "device": volume.Name,
                "fstype": volume.FileSystem,
            }
    return ret


@coinitialized
def disk_drive_exists(scsi_address):
    """
    Returns True if disk drive exists in OS. False otherwise.
    :param :class:`types.SCSIAddress` scsi_address: SCSI address of a disk
    :return bool: True if exists.
    """
    try:
        _scsi_address_to_disk_drive(scsi_address)
        return True
    except exc.StorageError as e:
        if "not found" in str(e):
            return False
        raise


@coinitialized
def logical_disk_exists(scsi_address):
    """
    Returns True if logical disk exists in OS. False otherwise.
    :param :class:`types.SCSIAddress` scsi_address: SCSI address of a disk
    :return bool: True if exists.
    """
    try:
        _scsi_address_to_logical_disk(scsi_address)
        return True
    except exc.StorageError as e:
        if "not found" in str(e):
            return False
        raise
