import inspect
import urllib.request
from xml.dom import minidom
import json as json_module
from xml.etree import ElementTree as ET
import yaml

from scalarizr.adm.command import Command
from scalarizr.adm.command import CommandError
from scalarizr.adm.util import make_table
from scalarizr.queryenv import xml2dict
from scalarizr.queryenv import QueryEnvError
from scalarizr.queryenv import QueryEnvService
from common.utils import six
from agent.scalarizr.hub import hub
from agent.scalarizr import metadata


def transpose(l):
    return [list(x) for x in zip(*l)]


class Queryenv(Command):
    """
    Launches queryenv methods.

    Usage:
      queryenv [--format=(xml|json|yaml)] <method> [<args>...]

    Options:
      -f <format>, --format=<format>  Output format: xml (default), json or yaml.
    """

    aliases = ['q']
    method_aliases = {'list-virtual-hosts': ['list-virtualhosts']}

    def __init__(self):
        super(Queryenv, self).__init__()

    @classmethod
    def get_method_name(cls, alias):
        """
        Returns method name if alias to supported method exists,
        None otherwise.
        """
        if alias in cls.get_supported_oldstyle_methods():
            return alias
        for method, aliases in cls.method_aliases.items():
            if alias in aliases:
                return method
        return None

    @classmethod
    def supports_oldstyle_method(cls, method_or_alias):
        """
        Returns True if method is supported or is alias of supported method
        with table output.
        """
        return cls.get_method_name(method_or_alias) is not None


    @classmethod
    def get_supported_oldstyle_methods(cls):
        """
        Returns list of methods that are supported for old-style calling with
        szradm <queryenv_method> [args...] with table output
        """
        return ['get-https-certificate',
            'get-latest-version',
            'list-ebs-mountpoints',
            'list-roles',
            'list-virtual-hosts']

    @classmethod
    def queryenv(cls):
        if not hasattr(cls, '_queryenv'):
            if not hub.general.queryenv_url:
                # hub is not intitalizated yet
                user_data = metadata.user_data()
                hub.setup_from_user_data(user_data)

            cls._queryenv = QueryEnvService.create_latest(
                hub.general.queryenv_url,
                server_id=hub.general.server_id,
                key_path=hub.crypto_key_path,
                autoretry=False)
        return cls._queryenv

    def _display_get_https_certificate(self, out):
        headers = ['cert', 'pkey', 'cacert']
        print(make_table(out, headers))

    def _display_get_latest_version(self, out):
        print(make_table([[out]], ['version']))

    def _display_list_ebs_mountpoints(self, out):
        headers = ['name', 'dir', 'createfs', 'isarray', 'volume-id', 'device']
        table_data = []
        for d in out:
            volume_params = [(v.volume_id, v.device) for v in d.volumes]
            volumes, devices = transpose(volume_params)
            table_data.append([d.name, d.dir, d.create_fs, d.is_array, volumes, devices])
        print(make_table(table_data, headers))

    def _display_list_roles(self, out):
        headers = ['behaviour',
           'name',
           'farm-role-id',
           'index',
           'internal-ip',
           'external-ip',
           'replication-master']
        table_data = []
        for d in out:
            behaviour = ', '.join(d.behaviour)
            for host in d.hosts:
                table_data.append([behaviour,
                                   d.name,
                                   d.farm_role_id,
                                   str(host.index),
                                   host.internal_ip,
                                   host.external_ip,
                                   str(host.replication_master)])
        print(make_table(table_data, headers))

    def _display_list_virtual_hosts(self, out):
        headers = ['hostname', 'https', 'type', 'raw']
        table_data = [[d.hostname, d.https, d.type, d.raw] for d in out]
        print(make_table(table_data, headers))

    def _display_list_global_variables(self, out):
        headers = ['key', 'value']
        table_data = list(out['public'].items())
        print(make_table(table_data, headers))

    def _display_fetch(self, out, format='xml'):
        if isinstance(out, six.string_types):
            out = out.encode('utf-8')
        if format == 'xml':
            print(minidom.parseString(out).toprettyxml())
        elif format == 'json':
            out_dict = xml2dict(ET.XML(out))
            print(json_module.dumps(out_dict, indent=4, sort_keys=True, ensure_ascii=False))
        elif format == 'yaml':
            out_dict = xml2dict(ET.XML(out))
            print(yaml.dump(out_dict, default_flow_style=False, allow_unicode=True))
        else:
            raise CommandError('Unknown output format.\nAvailable formats: xml, json, yaml')

    def _display_out(self, method, out, format='xml'):
        """
        General display method. Searches for certain display method and calls it
        with `out` or prints out in table form or raw. Custom display methods
        should be named by next pattern '_display_<method_name>' where
        <method_name> is queryenv method name with hyphens replaced with
        underscores.
        """
        all_display_methods = [m for m in dir(self) if m.startswith('_display')]
        display_method = None
        if method:
            for m in all_display_methods:
                if method.replace('-', '_') in m:
                    display_method = getattr(self, m)
                    break

        if display_method:
            display_kwds = {}
            argspec = inspect.getargspec(display_method)
            argnames = argspec.args
            if 'format' in argnames:
                display_kwds['format'] = format
            display_method(out, **display_kwds)
        elif isinstance(out, list) and isinstance(out[0], list) and method != 'fetch':
            print(make_table(out))
        else:
            print(out)

    def _run_queryenv_method(self, method, kwds, kwds_mapping=None):
        """
        Executes queryenv method with given `kwds`. `kwds` can contain excessive
        key-value pairs, this method filters it and passes only acceptable
        kwds by target method. `kwds_mapping` defines how `kwds` keys will be
        renamed when passed to queryenv method.
        """
        if not method:
            return
        if not kwds_mapping:
            kwds_mapping = {}

        method = method.replace('-', '_')
        queryenv_svc = self.queryenv()
        try:
            m = getattr(queryenv_svc, method)
        except AttributeError:
            raise CommandError('Unknown QueryEnv method\n')
        argspec = inspect.getargspec(m)
        argnames = argspec.args
        filtered_kwds = {}
        for k, v in kwds.items():
            arg_name = kwds_mapping.get(k, k)
            if argspec.keywords or arg_name in argnames:
                filtered_kwds[arg_name] = v

        if method == 'fetch':
            kwds = {k: v for k, v in kwds.items() if k not in filtered_kwds}
            filtered_kwds['params'] = kwds
        try:
            return m(**filtered_kwds)
        except (QueryEnvError, urllib.request.HTTPError) as e:
            if isinstance(e, urllib.request.HTTPError) and method == 'fetch':
                message = '%s method is not supported' % filtered_kwds['command']
            else:
                message = str(e)
            raise CommandError(message)

    def __call__(self, method=None, format=None, args=None, shortcut=False, **kwds):
        """
        All work is made by fetch method, other methods display methods and
        support remains only for backward compatibility.
        """
        if not args:
            args = []

        if not format:
            format = "xml"

        # we need to find method in kwds because parser places it there
        # param method is presented only when default parsing rule is applied
        # (queryenv <method> [<args>...])
        if not method:
            for kwd in kwds.keys():
                hyphen_kwd = kwd.replace('_', '-')
                _method = self.get_method_name(hyphen_kwd)
                if _method:
                    method = _method
                    kwds.pop(kwd)
                    break
        else:
            # if it's alias - taking real method name
            # it can be not known method so get_method_name() can return None
            # for valid method (such as 'fetch')
            method = self.get_method_name(method) or method

        for pair in args:
            if not pair.startswith('-') and '=' in pair:
                k, v = pair.split('=', 1)
                kwds[k] = v

        kwds_mapping = {}
        if method == 'list-roles':
            kwds_mapping = {'with_initializing': 'with_init'}

        if not shortcut and method != 'fetch':
            kwds['command'] = method
            method = 'fetch'

        out = self._run_queryenv_method(method, kwds, kwds_mapping)

        self._display_out(method, out, format)


commands = [Queryenv]
