from xml.dom import minidom
import json as json_module
import yaml

from agent.scalarizr.hub import hub

from scalarizr.adm.command import Command
from scalarizr.adm.command import CommandError
from scalarizr.adm.command import TAB_SIZE
from scalarizr.util import initdv2
from scalarizr.api.apache import ApacheAPI
from scalarizr.api.chef import ChefAPI
from scalarizr.api.haproxy import HAProxyAPI
from scalarizr.api.memcached import MemcachedAPI
from scalarizr.api.nginx import NginxAPI
from scalarizr.api.tomcat import TomcatAPI


service_apis = {
    'apache': ApacheAPI,
    'chef': ChefAPI,
    'haproxy': HAProxyAPI,
    'memcached': MemcachedAPI,
    'nginx': NginxAPI,
    'tomcat': TomcatAPI,
}


service_to_behavior = {
    'nginx': 'www',
    'apache': 'app'
}


class ReturnCode(object):
    RUNNING = 0
    STOPPED = 3
    UNKNOWN = 4
    MIXED = 150


class Service(Command):
    """
    Scalarizr service control.

    Usage:
      service nginx (start | stop) [(<index> | --port=<port>)]
      service nginx status [--format=(xml|json|yaml)] [(<index> | --port=<port>)]
      service <service> (start | stop)
      service <service> status [--format=(xml|json|yaml)]

    Options:
      -p <port>, --port=<port>
      -f <format>, --format=<format>
    """
    aliases = ['s']

    def help(self):
        doc = super(Service, self).help()
        services = [(' '*TAB_SIZE) + s for s in service_apis.keys()]
        return doc + '\nSupported services:\n' + '\n'.join(services)

    def _start_service(self, service, **kwds):
        api = service_apis[service]()
        for key, value in kwds.items():
            if value == None:
                kwds.pop(key)
        try:
            print('Starting %s' % service)
            api.start_service(**kwds)
        except (BaseException, Exception) as e:
            print('Service start failed.\n%s' % e)
            return int(CommandError())

    def _stop_service(self, service, **kwds):
        api = service_apis[service]()
        for key, value in kwds.items():
            if value == None:
                kwds.pop(key)
        try:
            print('Stopping %s' % service)
            api.stop_service(**kwds)
        except (BaseException, Exception) as e:
            print('Service stop failed.\n%s' % e)
            return int(CommandError())

    def _dict_to_xml(self, d, name):
        doc = minidom.Document()
        container = doc.createElement(name)
        doc.appendChild(container)
        for k, v in d.items():
            el = doc.createElement(str(k))
            text = doc.createTextNode(str(v))
            el.appendChild(text)
            container.appendChild(el)
        return doc

    def _dict_list_to_xml(self, list_of_dicts, list_name, element_name):
        doc = minidom.Document()
        container = doc.createElement(list_name)
        doc.appendChild(container)
        for element in list_of_dicts:
            child_doc = self._dict_to_xml(element, element_name)
            container.appendChild(child_doc.childNodes[0])
        return doc

    def _format_status(self, status_struct, format='xml'):
        if format == 'xml':
            if isinstance(status_struct, dict):
                doc = self._dict_to_xml(status_struct, 'status')
            else:
                doc = self._dict_list_to_xml(status_struct, 'statuses', 'status')
            return doc.toprettyxml(encoding='utf-8')
        elif format == 'json':
            return json_module.dumps(status_struct, indent=4, sort_keys=True, ensure_ascii=False) \
                .encode('utf-8')
        elif format == 'yaml':
            return yaml.dump(status_struct, default_flow_style=False, allow_unicode=True)
        else:
            raise CommandError('Unknown output format.\nAvailable formats: xml, json, yaml')

    def _display_service_status(self, service, print_format='xml', **kwds):
        api = service_apis[service]()
        for key, value in kwds.items():
            if value == None:
                kwds.pop(key)
        status = api.get_service_status(**kwds)

        if service == 'redis':
            return 'Redis automation was sunsetted'
        # TODO: make xml, json or yaml and dump it to out
        status_message = '%s service is stopped' % service
        code = ReturnCode.STOPPED
        if status == initdv2.Status.RUNNING:
            status_message = '%s service is running' % service
            code = ReturnCode.RUNNING
        elif status == initdv2.Status.UNKNOWN:
            status_message = '%s service has unknown status' % service
            code = ReturnCode.UNKNOWN
        # print(service + status_message)

        status_dict = {'code': code, 'message': status_message}
        print(self._format_status(status_dict, print_format))

        return code

    def __call__(self,
        service=None,
        start=False,
        stop=False,
        status=False,
        format='xml',

        redis=False,
        index=None,
        port=None,
        ):
        if service not in service_apis:
            print('Unknown service/behavior.')
            return ReturnCode.UNKNOWN

        if service not in hub.behaviors and \
            service_to_behavior.get(service) not in hub.behaviors:
            print('Not installed service/behavior.')
            return ReturnCode.UNKNOWN

        if start:
            return self._start_service(service, indexes=index, ports=port)
        elif stop:
            return self._stop_service(service, indexes=index, ports=port)
        elif status:
            return self._display_service_status(service,
                indexes=index,
                ports=port,
                print_format=format)


commands = [Service]
