"""
Created on Nov 25, 2011
@author: marat
"""

import binascii
import glob
import logging
from multiprocessing import pool
import os
import signal
import subprocess
import threading
import time
import weakref

from agent import config
from agent.scalarizr import rpc
from agent.scalarizr.hub import hub
from common.utils import six

import scalarizr
from scalarizr import handlers
from scalarizr.api import operation as operation_api
from scalarizr.handlers.union import base as union_base
from scalarizr.linux import mount
from scalarizr.linux import Version
from scalarizr.queryenv import ScalingMetric
from scalarizr.util import kill_childs
from scalarizr.util import Singleton

from common.utils import exc
from common.utils.facts import fact
from agent.tasks import sys as sys_tasks
from agent.tasks import execute as execute_tasks
from common.utils import sysutil
from common.utils import subprocess2
from common.utils import dockerutil


LOG = logging.getLogger(__name__)
CALL_TIMEOUT = 2
RC_NOT_FOUND = 127
max_log_size = 5*1024*1024
SUPPORTED_DOCKER_VERSION = (1, 9, 0)


class _ScalingMetricStrategy(object):
    """Strategy class for custom scaling metric"""

    @staticmethod
    def _get_execute(metric):
        if not os.access(metric.path, os.X_OK):
            raise BaseException("File is not executable: '%s'" % metric.path)

        exec_timeout = 10
        close_fds = fact['os']['name'] != 'windows'
        proc = subprocess.Popen(
            metric.path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=close_fds)

        timeout_time = time.time() + exec_timeout
        while time.time() < timeout_time:
            if proc.poll() is None:
                time.sleep(0.2)
            else:
                break
        else:
            kill_childs(proc.pid)
            if hasattr(proc, 'terminate'):
                # python >= 2.6
                proc.terminate()
            else:
                os.kill(proc.pid, signal.SIGTERM)
            raise BaseException('Timeouted')

        stdout, stderr = proc.communicate()
        if isinstance(stdout, six.binary_type):
            stdout = stdout.decode('utf-8', errors='ignore')
        if isinstance(stderr, six.binary_type):
            stderr = stderr.decode('utf-8', errors='ignore')


        if proc.returncode > 0:
            raise BaseException(stderr if stderr else 'exitcode: %d' % proc.returncode)

        return stdout.strip()

    @staticmethod
    def _get_read(metric):
        try:
            with open(metric.path, 'r') as fp:
                value = fp.readline()
        except IOError:
            raise BaseException("File is not readable: '%s'" % metric.path)

        return value.strip()

    @staticmethod
    def get(metric):
        error = ''
        try:
            if metric.retrieve_method == ScalingMetric.RetriveMethod.EXECUTE:
                value = _ScalingMetricStrategy._get_execute(metric)
            elif metric.retrieve_method == ScalingMetric.RetriveMethod.READ:
                value = _ScalingMetricStrategy._get_read(metric)
            else:
                raise BaseException('Unknown retrieve method %s' % metric.retrieve_method)
            try:
                value = float(value)
            except:
                raise ValueError("Can't convert metric value to float: {!r}".format(value))
        except (Exception, BaseException) as e:
            value = 0.0
            error = str(e)[0:255]

        return {'id': metric.id, 'name': metric.name, 'value': value, 'error': error}


@six.add_metaclass(Singleton)
class SystemAPI(object):
    """
    Pluggable API to get system information similar to SNMP, Facter(puppet), Ohai(chef).
    Namespace::
        system
    """

    _HOSTNAME = '/etc/hostname'
    _DISKSTATS = '/proc/diskstats'
    _PATH = ['/usr/bin/', '/usr/local/bin/']
    _CPUINFO = '/proc/cpuinfo'
    _NETSTATS = '/proc/net/dev'
    _LOG_FILE = '/var/log/scalarizr.log'
    _DEBUG_LOG_FILE = '/var/log/scalarizr_debug.log'
    _UPDATE_LOG_FILE = '/var/log/scalarizr_update.log'
    _CENTOS_NETWORK_CFG = '/etc/sysconfig/network'

    def __init__(self):
        self._op_api = operation_api.OperationAPI()
        self._container_stats = {}
        self._docker_version = None

    def _readlines(self, path):
        with open(path, "r") as fp:
            return fp.readlines()

    def add_extension(self, extension):
        """
        :type extension: object
        :param extension: Object with some callables to extend SysInfo public interface
        Note::
            Duplicates are resolved by overriding old function with a new one
        """

        for name in dir(extension):
            attr = getattr(extension, name)
            if not name.startswith('_') and callable(attr):
                if hasattr(self, name):
                    LOG.warn('Duplicate attribute %s. Overriding %s with %s',
                            name, getattr(self, name), attr)
                setattr(self, name, attr)

    @rpc.query_method
    def call_auth_shutdown_hook(self):
        """
        .. warning::
            Deprecated.
        """
        script_path = '/usr/local/scalarizr/hooks/auth-shutdown'
        LOG.debug("Executing %s" % script_path)
        if os.access(script_path, os.X_OK):
            process = subprocess.Popen(
                script_path, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, close_fds=True)
            out, _ = process.communicate()
            if isinstance(out, six.binary_type):
                out = out.decode('utf-8', errors='ignore')
            return out.strip()
        else:
            raise Exception('File not exists: %s' % script_path)

    @rpc.command_method
    def force_resume(self):
        """Forces init after reboot.

        Works only when last RebootFinish date is greater then uptime,
        there is no such message in history.
        """
        LOG.info('Scalarizr resumed after reboot (forced!)')
        from agent.scalarizr import app
        app.reboot_finish(force=True)
        return True

    @rpc.command_method
    def reboot(self):
        sysutil.reboot()

    @rpc.command_method
    def set_hostname(self, hostname=None, domain_name=None):
        """
        Updates server's FQDN.
        :param str hostname: Fully Qualified Domain Name to set for this host
        Example::
            api.system.set_hostname(hostname = "myhostname.com")
        """
        sys_tasks.set_hostname(hostname, domain_name=domain_name, reboot=False)

    @rpc.query_method
    def get_hostname(self):
        """
        :return: server's FQDN.
        :rtype: str
        Example::
            "ec2-50-19-134-77.compute-1.amazonaws.com"
        """
        return fact['hostname']

    @rpc.query_method
    def get_domain_name(self):
        """
        :return: server's domain name.
        :rtype: str
        Example::
            "compute-1.amazonaws.com"
        """
        return fact['domain_name']

    @rpc.query_method
    def block_devices(self):
        """
        :return: List of block devices including ramX and loopX
        :rtype: list
        """
        return fact['block_devices']


    @rpc.query_method
    def uptime(self):
        """
        :return: uptime value in seconds
        :rtype: float
        """
        return sysutil.uptime()


    @rpc.query_method
    def cpu_stat(self):

        """
        :return: CPU stat from /proc/stat.
        :rtype: dict

        Example::
            {
                'user': 8416,
                'nice': 0,
                'system': 6754,
                'idle': 147309
            }
        """
        return fact['stat']['cpu']

    @rpc.query_method
    def mem_info(self):
        """
        :return: Memory information from /proc/meminfo.
        :rtype: dict

        Example::
             {
                'total_swap': 0,
                'avail_swap': 0,
                'total_real': 604364,
                'total_free': 165108,
                'shared': 168,
                'buffer': 17832,
                'cached': 316756
            }
        """
        return fact['stat']['memory']

    @rpc.query_method
    def load_average(self):
        """
        :return: Load average (1, 5, 15) in 3 items list.
        :rtype: list
        Example::
            [
               0.0,      // LA1
               0.01,     // LA5
               0.05      // LA15
            ]
        """
        return fact['stat']['load_average']

    @rpc.query_method
    def disk_stats(self):
        """
        :return: Disks I/O statistics.
        Data format::
            {
            <device>: {
                <read>: {
                    <num>: total number of reads completed successfully
                    <sectors>: total number of sectors read successfully
                    <bytes>: total number of bytes read successfully
                }
                <write>: {
                    <num>: total number of writes completed successfully
                    <sectors>: total number of sectors written successfully
                    <bytes>: total number of bytes written successfully
                },
            ...
            }
        See more at http://www.kernel.org/doc/Documentation/iostats.txt
        """
        return fact['stat']['disk']

    @rpc.query_method
    def net_stats(self):
        """
        :return: Network I/O statistics.
        Data format::
            {
                <iface>: {
                    <receive>: {
                        <bytes>: total received bytes
                        <packets>: total received packets
                        <errors>: total receive errors
                    }
                    <transmit>: {
                        <bytes>: total transmitted bytes
                        <packets>: total transmitted packets
                        <errors>: total transmit errors
                    }
                },
                ...
            }
        """
        return fact['stat']['net']

    def _disk_usage_adapter(self, statfvs_data, device_info=False):
        """This method is used for compatibility with old Scalr servers.

        :param device_info: if True, include 'device' and 'fstype' to result dict. Default False.
        New version of `statvfs` and `mounts` returns data in bytes. Old instances of scalr rely on
        data in kbytes.
        """
        result = {}
        for mpoint, usage in statfvs_data.items():
            result[mpoint] = {
                'total': usage['total_bytes'] / 1024,  # kb
                'free': usage['free_bytes'] / 1024,  # kb
            }
            if device_info:
                result[mpoint]['device'] = usage.get('device')
                result[mpoint]['fstype'] = usage.get('fstype')
        return result

    @rpc.query_method
    def statvfs(self, mpoints=None):
        """
        :return: Information about provided mounted file systems (total size and free space).
        Request::
            {
                "mpoints": [
                    "/mnt/dbstorage",
                    "/media/mpoint",
                    "/non/existing/mpoint"
                ]
            }
        Response::
            {
               "/mnt/dbstorage": {
                 "total" : 10000,
                 "free" : 5000
               },
               "/media/mpoint": {
                 "total" : 20000,
                 "free" : 1000
               },
               "/non/existing/mpoint" : null
            }
        """
        if not isinstance(mpoints, list):
            raise Exception("Argument 'mpoints' should be a list of strings, "
                            "not {}.".format(type(mpoints)))
        return self._disk_usage_adapter(sysutil.statvfs(mpoints))

    @rpc.query_method
    def mounts(self):
        """Disks total/free bytes(usage).

        :return dict in format:
        {
            <mount_point>: {
                'total': int total kbytes,
                'free': int free kbytes
            },..
        }
        """
        return self._disk_usage_adapter(fact['stat']['mounts'], device_info=True)

    @rpc.command_method
    def scaling_metrics(self):
        """
        :return: list of scaling metrics
        :rtype: list

        Example::
            [{
                'id': 101011,
                'name': 'jmx.scaling',
                'value': 1,
                'error': None
            }, {
                'id': 202020,
                'name': 'app.poller',
                'value': None,
                'error': 'Couldnt connect to host'
            }]
        """

        # Obtain scaling metrics from Scalr.
        scaling_metrics = hub.queryenv.get_scaling_metrics()
        if not scaling_metrics:
            return []

        if not hasattr(threading.current_thread(), '_children'):
            threading.current_thread()._children = weakref.WeakKeyDictionary()

        wrk_pool = pool.ThreadPool(processes=10)

        try:
            return wrk_pool.map_async(_ScalingMetricStrategy.get, scaling_metrics).get()
        finally:
            wrk_pool.close()
            wrk_pool.join()

    @rpc.command_method
    def execute_scripts(self, scripts=None, global_variables=None, event_name=None,
            role_name=None, msg_body=None, async=False):
        def do_execute_scripts(op):
            msg = lambda: None
            msg.name = event_name
            msg.role_name = role_name
            msg.body = msg_body or {}
            msg.body.update({
                'scripts': scripts or [],
                'global_variables': global_variables or []
            })
            hdlr = union_base.get_handlers()[0]
            hdlr(msg)

        return self._op_api.run('system.execute_scripts', do_execute_scripts, async=async)

    @rpc.query_method
    def get_script_logs(self, exec_script_id, maxsize=max_log_size):
        """
        :return: stdout and stderr scripting logs
        :rtype: dict(stdout: base64encoded, stderr: base64encoded)
        """
        logs = execute_tasks.get_logs(exec_script_id.replace('-', ''))

        err_rotated = (b'Log file already rotated and no more exists on server. '
                       b'You can increase "Rotate scripting logs" setting under "Advanced" tab'
                       b' in Farm Designer')

        stdout = binascii.b2a_base64(
            logs.stdout.encode('utf-8') if logs.stdout is not None else err_rotated).decode()
        stderr = binascii.b2a_base64(
            logs.stderr.encode('utf-8') if logs.stderr is not None else err_rotated).decode()

        return dict(stdout=stdout, stderr=stderr)

    @rpc.query_method
    def get_debug_log(self):
        """
        :return: scalarizr debug log (/var/log/scalarizr.debug.log on Linux)
        :rtype: str
        """
        return binascii.b2a_base64(_get_log(self._DEBUG_LOG_FILE, -1).encode('utf-8')).decode()

    @rpc.query_method
    def get_update_log(self):
        """
        :return: scalarizr update log (/var/log/scalarizr.update.log on Linux)
        :rtype: str
        """
        return binascii.b2a_base64(_get_log(self._UPDATE_LOG_FILE, -1).encode('utf-8')).decode()

    @rpc.query_method
    def get_log(self):
        """
        :return: scalarizr info log (/var/log/scalarizr.log on Linux)
        :rtype: str
        """
        return binascii.b2a_base64(_get_log(self._LOG_FILE, -1).encode('utf-8')).decode()

    @rpc.query_method
    def container_stats(self):
        """
        :return list(dict): docker container stats. Example:
            [
                {
                'container_id': '12587a2f455c1d0b8b82eca2dbe7d4176497717a906cd26e7f9abd507038617b',
                'created': '2016-05-17 18:19:52',
                'image': 'alpine:stable',
                'network': 'default',
                'name': 'guesto',
                'privileged': False,
                'labels': {
                    'build-date': '20160701',
                    'license': 'GPLv2',
                    'name': 'CentOS Base Image',
                    'vendor': 'CentOS'
                },
                'volumes': [{'source': '/mnt/media', 'destination': '/mnt/media'}],
                'ports': {
                    'tcp/80': [
                        {'port': 8080, 'host_ip': '0.0.0.0'},
                         {'port': 9090, 'host_ip': '1.1.1.1'}],
                    '443/tcp': None}},
                {
                'container_id': '21ea45e5946323066d9bd55e6ad5fe351cc18f15d9ba180925fecd67a9cc2a6f',
                'created': '2016-06-03 11:13:34',
                'image': '04c5d3b7b0656168630d3ba35d8889bd0e9caafcaeb3004d2bfbc47e7c5d35d2',
                'network': 'default',
                'name': 'spy',
                'privileged': True,
                'labels': {
                    'build-date': '20150208',
                    'license': 'GPLv2'
                },
                'volumes': [],
                'ports': {}}]
        """
        return fact['stat']['container']

    @rpc.query_method
    def get_facts(self, keys=None, invalidate_bindings=None):
        facts = {}
        if isinstance(keys, (list, tuple)) and 'stat' in keys:
            keys = list(keys)
            facts = {'stat': self._get_stat_facts(invalidate_bindings=invalidate_bindings)}
            keys.remove('stat')

        if keys:
            facts.update(sys_tasks.get_facts(keys=keys,
                                             invalidate_bindings=invalidate_bindings,
                                             raise_on_fact_error=False))
        return facts

    def _get_stat_facts(self, invalidate_bindings=None):
        # scalrpy compatibility
        stat_keys = config.get_stat_metrics() + ['stat.load_average']
        stat_facts = sys_tasks.get_facts(keys=stat_keys,
                                         invalidate_bindings=invalidate_bindings,
                                         raise_on_fact_error=False)
        return {key.replace('stat.', ''): value for key, value in stat_facts.items()}


def _get_log(logfile, maxsize=max_log_size):
    if maxsize != -1 and (os.path.getsize(logfile) > maxsize):
        return 'Unable to fetch Log file %s: file is larger than %s bytes' % (logfile, maxsize)
    try:
        with open(logfile, "r", encoding="utf-8") as fp:
            return fp.read(int(maxsize))
    except IOError:
        return 'Log file %s is not readable' % logfile


if fact['os']['name'] == 'windows':

    class WindowsSystemAPI(SystemAPI):

        _LOG_FILE = os.path.join(hub.log_dir, 'scalarizr.log')
        _DEBUG_LOG_FILE = os.path.join(hub.log_dir, 'scalarizr_debug.log')
        _UPDATE_LOG_FILE = os.path.join(hub.log_dir, 'scalarizr_update.log')
        _pending_hostname = None

        @rpc.command_method
        def set_hostname(self, hostname=None, domain_name=None):
            super(WindowsSystemAPI, self).set_hostname(hostname, domain_name=domain_name)
            self._pending_hostname = hostname

        @rpc.query_method
        def get_hostname(self):
            if self._pending_hostname:
                return self._pending_hostname
            return super(WindowsSystemAPI, self).get_hostname()

        @rpc.query_method
        def get_domain_name(self):
            """
            :return: server's domain name.
            :rtype: str
            Example::
                "compute-1.amazonaws.com"
            """
            return fact['domain_name']

        @rpc.query_method
        def load_average(self):
            raise exc.NotSupportedError('LA is not available on Windows')

        @rpc.query_method
        def container_stats(self, pool_size=8):
            raise exc.NotSupportedError('Docker is not available on Windows')


    SystemAPI = WindowsSystemAPI
