import os
import sys
import socket
import glob
import logging

from agent.tasks import service
from agent.scalarizr import rpc
from agent.scalarizr.hub import hub
from common.utils import six

from scalarizr import linux
from scalarizr.linux import pkgmgr
from scalarizr.util import Singleton
from scalarizr import exceptions
from scalarizr.util import initdv2
from scalarizr.util import firstmatched
from scalarizr.api import BehaviorAPI
from scalarizr.api import SoftwareDependencyError

LOG = logging.getLogger(__name__)


CATALINA_HOME_DIR = None
JAVA_HOME = firstmatched(
    lambda path: os.access(path, os.X_OK),
    [linux.system('echo $JAVA_HOME', shell=True)[0].strip(), '/usr/java/default'],
    '/usr')
CONFIG_DIR = None
INSTALL_TYPE = None


def augload():
    path = CONFIG_DIR
    return [
        'set /augeas/load/Xml/incl[last()+1] "{0}/*.xml"'.format(path),
        'load',
        'defvar service /files{0}/server.xml/Server/Service'.format(path)
    ]


def augtool(script_lines):
    augscript = augload() + script_lines
    augscript = '\n'.join(augscript)
    LOG.debug('augscript: %s', augscript)
    return linux.system(('augtool', ), stdin=augscript)[0].strip()


class CatalinaInitScript(initdv2.ParametrizedInitScript):
    def __init__(self):
        initdv2.ParametrizedInitScript.__init__(self, 'tomcat',
                CATALINA_HOME_DIR + '/bin/catalina.sh')
        self.server_port = None

    def status(self):
        if not self.server_port:
            out = augtool(
                ['print /files{0}/server.xml/Server/#attribute/port'.format(CONFIG_DIR)])
            self.server_port = out.split(' = ')[-1]
            self.server_port = int(self.server_port.strip('"'))

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect(('', self.server_port))
            return initdv2.Status.RUNNING
        except socket.error:
            return initdv2.Status.NOT_RUNNING
        finally:
            try:
                sock.close()
            except:
                pass


@six.add_metaclass(Singleton)
class TomcatAPI(BehaviorAPI):
    """
    Basic API for managing Tomcat service.

    Namespace::

        tomcat
    """
    behavior = 'tomcat'

    @classmethod
    def catalina_home_dir(cls):
        return linux.system('echo $CATALINA_HOME', shell=True)[0].strip()

    @staticmethod
    def _find_file_name_by_patterns(directory, name_patterns, ext=None):
        """Returns the first match.
        """
        if ext and not ext.startswith('.'):
            ext = u'.{}'.format(ext)
        for pattern in name_patterns:
            paths = glob.glob(u'{}/{}{}'.format(
                directory, pattern, ext or ''))
            if paths:
                return os.path.basename(paths[0]).rstrip(ext)

    def _find_service(self):
        # try to read CATALINA_HOME from environment
        global CATALINA_HOME_DIR
        global INSTALL_TYPE
        global CONFIG_DIR

        CATALINA_HOME_DIR = self.catalina_home_dir()
        if not CATALINA_HOME_DIR:
            # try to locate CATALINA_HOME in /opt/apache-tomcat*
            try:
                CATALINA_HOME_DIR = glob.glob('/opt/apache-tomcat*')[0]
            except IndexError:
                pass

        tomcat_name_patterns = (
            'tomcat',
            'tomcat[1-9]',
            'tomcat[1-9][0-9]'
        )

        # determine tomcat service name
        service_name = None
        if os.path.exists('/etc/init.d/'):
            service_name = self._find_file_name_by_patterns(
                '/etc/init.d', tomcat_name_patterns
            )
        if not service_name and os.path.exists('/usr/lib/systemd/system'):
            service_name = self._find_file_name_by_patterns(
                '/usr/lib/systemd/system', tomcat_name_patterns, ext='service'
            )
        service_name = service_name or self.behavior

        # try to get tomcat system service
        try:
            tomcat_service = service.Service(service_name)
            tomcat_service.status()  # raises `service.NotInstalledError`
        except service.NotInstalledError:
            tomcat_service = None

        if CATALINA_HOME_DIR:
            INSTALL_TYPE = 'binary'
            CONFIG_DIR = '{0}/conf'.format(CATALINA_HOME_DIR)
            if not tomcat_service:
                tomcat_service = CatalinaInitScript()
        else:
            INSTALL_TYPE = 'package'
            config_dir_name = self._find_file_name_by_patterns(
                '/etc', tomcat_name_patterns
            )
            if config_dir_name:
                config_path = os.path.join('/etc', config_dir_name)
                if config_path and os.path.exists(config_path):
                    CONFIG_DIR = config_path

        if not CONFIG_DIR:
            raise exceptions.NotFound("Tomcat config directory not found")

        if not tomcat_service:
            raise exceptions.NotFound("Tomcat service not found")
        return tomcat_service

    def __init__(self):
        self.service = self._find_service()

    @rpc.command_method
    def start_service(self):
        """
        Starts Tomcat service.

        Example::

            api.tomcat.start_service()
        """
        self.service.start()

    @rpc.command_method
    def stop_service(self):
        """
        Stops Tomcat service.

        Example::

            api.tomcat.stop_service()
        """
        self.service.stop()

    @rpc.command_method
    def reload_service(self):
        """
        Reloads Tomcat configuration.

        Example::

            api.tomcat.reload_service()
        """
        self.service.reload()

    @rpc.command_method
    def restart_service(self):
        """
        Restarts Tomcat service.

        Example::

            api.tomcat.restart_service()
        """
        self.service.restart()

    @rpc.command_method
    def get_service_status(self):
        """
        Checks Tomcat service status.

        RUNNING = 0
        DEAD_PID_FILE_EXISTS = 1
        DEAD_VAR_LOCK_EXISTS = 2
        NOT_RUNNING = 3
        UNKNOWN = 4

        :return: Status num.
        :rtype: int


        Example::

            >>> api.tomcat.get_service_status()
            0
        """
        status = self.service.status()
        if type(status) == int:
            return status

        if status == service.ServiceStatus.running:
            return initdv2.Status.RUNNING
        elif status == service.ServiceStatus.stopped:
            return initdv2.Status.NOT_RUNNING
        else:
            return initdv2.Status.UNKNOWN

    @classmethod
    def do_check_software(cls, system_packages=None):
        os_name = linux.os['name'].lower()
        os_vers = linux.os['version']
        if cls.catalina_home_dir():
            return
        if linux.os.debian_family:
            requirements_main = None
            if os_name == 'ubuntu':
                if os_vers >= '12':
                    requirements_main = ['tomcat7']
                    requirements_dependencies = ['tomcat7-admin']
                elif os_vers >= '10':
                    requirements_main = ['tomcat6']
                    requirements_dependencies = ['tomcat6-admin']
            elif os_name == 'debian':
                if os_vers >= '7':
                    requirements_main = ['tomcat7']
                    requirements_dependencies = ['tomcat7-admin']
                elif os_vers >= '6':
                    requirements_main = ['tomcat6']
                    requirements_dependencies = ['tomcat6-admin']
            if requirements_main is None:
                raise exceptions.UnsupportedBehavior(
                        cls.behavior,
                        "Not supported on {0} os family".format(linux.os['family']))
            installed = pkgmgr.check_software(requirements_main, system_packages)[0]
            try:
                pkgmgr.check_software(requirements_dependencies, system_packages)
                return installed
            except pkgmgr.NotInstalledError:
                e = sys.exc_info()[1]
                raise SoftwareDependencyError(e.args[0])
        elif linux.os.redhat_family:
            requirements = [
                ['tomcat', 'tomcat-admin-webapps'],
                ['tomcat6', 'tomcat6-admin-webapps'],
                ['tomcat7', 'tomcat7-admin-webapps'],
            ]
            errors = list()
            for requirement in requirements:
                try:
                    installed = pkgmgr.check_software(requirement[0], system_packages)[0]
                    try:
                        pkgmgr.check_software(requirement[1:], system_packages)
                        return installed
                    except pkgmgr.NotInstalledError:
                        e = sys.exc_info()[1]
                        raise SoftwareDependencyError(e.args[0])
                except:
                    e = sys.exc_info()[1]
                    errors.append(e)
            for cls in [pkgmgr.VersionMismatchError, SoftwareDependencyError, pkgmgr.NotInstalledError]:
                for error in errors:
                    if isinstance(error, cls):
                        raise error
        else:
            raise exceptions.UnsupportedBehavior(
                    cls.behavior,
                    "Not supported on {0} os family".format(linux.os['family']))
