import os
import logging
import threading
import pprint
import sys
import traceback
import codecs
import platform as platform_module

from agent.tasks import firewall
from agent.scalarizr import msg
from agent.scalarizr.hub import hub
from common.utils import exc
from common.utils import six

import scalarizr
from scalarizr import util, linux, api, exceptions
from scalarizr.api import operation
from scalarizr.util import initdv2
from scalarizr.linux import pkgmgr
from scalarizr.service import CnfPresetStore, CnfPreset, PresetType

LOG = logging.getLogger(__name__)


class Handler(object):

    _service_name = behaviour = None
    _logger = logging.getLogger(__name__)

    def __init__(self):
        pass

    def new_message(self, msg_name, msg_body=None, msg_meta=None, broadcast=False, include_pad=False):
        message = msg.create(msg_name, msg_body, msg_meta)

        if broadcast:
            self._broadcast_message(message)
        if include_pad:
            message.body['platform_access_data'] = hub.platform.get_access_data()
        return message

    def send_message(self, msg_name,
                     msg_body=None,
                     msg_meta=None,
                     broadcast=False,
                     queue='control',
                     handle_host_init=False,
                     handle_before_host_up=False,
                     new_crypto_key=None):
        message = msg_name if not isinstance(msg_name, str) else \
            self.new_message(msg_name, msg_body, msg_meta, broadcast)
        msg.send(message, queue=queue)

    def send_int_message(self, host, msg_name, msg_body=None, msg_meta=None, broadcast=False,
                         include_pad=False, queue='control'):
        message = self.new_message(msg_name, msg_body, msg_meta, broadcast, include_pad)
        msg.send_internal(message, host)

    def send_result_error_message(self, msg_name, error_text=None, exc_info=None, body=None):
        body = body or {}
        if not exc_info:
            exc_info = sys.exc_info()
        body['status'] = 'error'
        body['last_error'] = ''
        if error_text:
            body['last_error'] += error_text + '. '
        body['last_error'] += str(exc_info[1])
        body['trace'] = ''.join(traceback.format_tb(exc_info[2]))

        LOG.error(body['last_error'], exc_info=exc_info)
        self.send_message(msg_name, body)

    def _broadcast_message(self, message):
        message.local_ip = hub.platform.get_private_ip()
        message.remote_ip = hub.platform.get_public_ip()
        message.behaviour = hub.behaviors
        message.role_name = hub.general.role_name

    def accept(self, message, queue, behaviour=None, platform=None, os=None, dist=None):
        return False

    def __call__(self, message):
        fn = "on_" + message.name
        if hasattr(self, fn) and callable(getattr(self, fn)):
            getattr(self, fn)(message)
        else:
            raise HandlerError("Handler %s has no method %s", self.__class__.__name__, fn)

    def _setup_firewall(self):
        rules = self._iter_firewall_rules()
        if not rules:
            return

        if firewall_management_enabled():
            for rule in rules:
                firewall.ensure(rule)
        else:
            self._logger.debug((
                "Firewall management is disabled in Scalr. "
                "Skipping adding firewall rules: {}."
            ).format(list(rules)))

    @staticmethod
    def _iter_firewall_rules():
        return []


class HandlerError(BaseException):
    pass


def async(fn):
    def decorated(*args, **kwargs):
        t = threading.Thread(target=fn, args=args, kwargs=kwargs)
        t.start()

    return decorated


class ServiceCtlHandler(Handler):
    _logger = None
    _cnf_ctl = None
    _init_script = None
    _preset_store = None
    _service_name = None
    initial_preset = None

    def __init__(self, service_name, init_script=None, cnf_ctl=None):
        '''
        XXX: When migrating to the new preset system
        do not forget that self._service_name is essential for
        Handler.get_ready_behaviours() and should be overloaded
        in every ServiceCtlHandler child.

        '''
        self._service_name = service_name
        self._cnf_ctl = cnf_ctl
        self._init_script = init_script
        self._logger = logging.getLogger(__name__)
        self._preset_store = CnfPresetStore(self._service_name)

        Handler.__init__(self)

        hub.on('init', self.sc_on_init)

    def on_UpdateServiceConfiguration(self, message):
        if self._service_name != message.behaviour:
            return

        result = self.new_message(Messages.UPDATE_SERVICE_CONFIGURATION_RESULT)
        result.behaviour = message.behaviour

        # Obtain current configuration preset
        if message.reset_to_defaults == '1':
            new_preset = self._preset_store.load(PresetType.DEFAULT)
        else:
            new_preset = self._obtain_current_preset()
        if new_preset:
            result.preset = new_preset.name

        # Apply current preset
        try:
            LOG.info("Applying preset '%s' to %s %s service restart",
                                            new_preset.name, self._service_name,
                                            'with' if message.restart_service == '1' else 'without')
            self._cnf_ctl.apply_preset(new_preset)
            if message.restart_service == '1' or message.reset_to_defaults == '1':
                self._stop_service(reason="Applying preset '%s'" % new_preset.name)
                self._start_service_with_preset(new_preset)
            result.status = 'ok'
        except (Exception, BaseException) as e:
            result.status = 'error'
            result.last_error = str(e)

        # Send result
        self.send_message(result)

    def _start_service(self):
        if not self._init_script.running:
            LOG.info("Starting %s" % self._service_name)
            try:
                self._init_script.start()
            except BaseException as e:
                if not self._init_script.running:
                    raise
                LOG.warning(str(e))
            LOG.debug("%s started" % self._service_name)

    def _stop_service(self, reason=None):
        if self._init_script.running:
            LOG.info("Stopping %s%s", self._service_name, '. (%s)' % reason if reason else '')
            try:
                self._init_script.stop()
            except:
                if self._init_script.running:
                    raise
            LOG.debug("%s stopped", self._service_name)

    def _restart_service(self, reason=None):
        LOG.info("Restarting %s%s", self._service_name, '. (%s)' % reason if reason else '')
        self._init_script.restart()
        LOG.debug("%s restarted", self._service_name)

    def _reload_service(self, reason=None):
        LOG.info("Reloading %s%s", self._service_name, '. (%s)' % reason if reason else '')
        try:
            self._init_script.reload()
            hub.fire(self._service_name + '_reload')
        except initdv2.InitdError as e:
            if e.code == initdv2.InitdError.NOT_RUNNING:
                LOG.debug('%s not running', self._service_name)
            else:
                raise
        LOG.debug("%s reloaded", self._service_name)

    def _obtain_current_preset(self):
        service_conf = hub.queryenv.get_service_configuration(self._service_name)
        if service_conf.new_engine:
            '''New sheriff in town. No need to calculate or apply old preset'''
            return None

        cur_preset = CnfPreset(service_conf.name, service_conf.settings)
        if cur_preset.name == 'default':
            try:
                cur_preset = self._preset_store.load(PresetType.DEFAULT)
            except IOError as e:
                if e.errno == 2:
                    cur_preset = self._cnf_ctl.current_preset()
                    self._preset_store.save(cur_preset, PresetType.DEFAULT)
                else:
                    raise
        return cur_preset

    def _start_service_with_preset(self, preset):
        '''
        TODO: Revise method carefully
        '''
        try:
            if self._init_script.running:
                self._restart_service('applying new service settings from configuration preset')
            else:
                self._start_service()
        except BaseException as e:
            LOG.error('Cannot start %s with current configuration preset. ' % self._service_name
                            + '[Reason: %s] ' % str(e)
                            + 'Rolling back to the last successful preset')
            preset = self._preset_store.load(PresetType.LAST_SUCCESSFUL)
            self._cnf_ctl.apply_preset(preset)
            self._start_service()

        LOG.debug("Set %s configuration preset '%s' as last successful", self._service_name, preset.name)
        self._preset_store.save(preset, PresetType.LAST_SUCCESSFUL)

    def sc_on_init(self):
        hub.on(
                start=self.sc_on_start,
                service_configured=self.sc_on_configured,
                before_host_down=self.sc_on_before_host_down
        )

    def sc_on_start(self):
        if hub.state == 'running':
            if self._cnf_ctl:
                # Obtain current configuration preset
                cur_preset = self._obtain_current_preset()
                if not cur_preset:
                    LOG.info('New configuration preset engine is used. Skipping old presets.')
                    return
                # Apply current preset
                my_preset = self._cnf_ctl.current_preset()
                if not self._cnf_ctl.preset_equals(cur_preset, my_preset):
                    LOG.info("Applying '%s' preset to %s", cur_preset.name, self._service_name)
                    self._cnf_ctl.apply_preset(cur_preset)
                    # Start service with updated configuration
                    self._start_service_with_preset(cur_preset)
                else:
                    LOG.debug("%s configuration satisfies current preset '%s'", self._service_name, cur_preset.name)
                    self._start_service()

            else:
                self._start_service()


    def sc_on_before_host_down(self, msg):
        self._stop_service('instance goes down')

    def sc_on_configured(self, service_name, **kwargs):
        if self._service_name != service_name:
            return

        # Fetch current configuration preset
        service_conf = hub.queryenv.get_service_configuration(self._service_name)

        if service_conf.new_engine:
            LOG.debug('New configuration presets engine is available.')
            response = None
            settings = {}
            LOG.debug('Initial preset from HostInitResponse: %s' % self.initial_preset)

            if self.initial_preset:
                LOG.debug('initial_preset = %s' % self.initial_preset)
                for preset in self.initial_preset:
                    for f, data in preset.items():
                        kv = {}
                        for setting in data['settings']:
                            k = setting['setting']['name']
                            v = setting['setting']['value']
                            kv[k] = v
                        settings.update({data['name']:kv})
                LOG.debug('Got settings from initial preset: %s' % settings)

            else:
                response = hub.queryenv.list_farm_role_params(hub.general.farm_role_id)
            if response and service_name in response and 'preset' in response[service_name]:
                settings = response[service_name]['preset']
                LOG.debug('list_farm_role_params returned settings: %s' % settings)
            if settings:
                manifest = self.preset_provider.get_manifest(service_name)
                if manifest:
                    LOG.debug('Applying configuration preset')
                    self.preset_provider.set_preset(settings, manifest)
                    LOG.debug('Configuration preset has been successfully applied.')
                else:
                    LOG.WARNING('Cannot apply preset: Manifest not found.')

        else:
            if self._cnf_ctl:
                LOG.info('Applying configuration preset')

                # Backup default configuration
                my_preset = self._cnf_ctl.current_preset()
                self._preset_store.save(my_preset, PresetType.DEFAULT)

                # Stop service if it's already running
                self._stop_service('Applying configuration preset')

                cur_preset = CnfPreset(service_conf.name, service_conf.settings, self._service_name)
                self._preset_store.copy(PresetType.DEFAULT, PresetType.LAST_SUCCESSFUL, override=False)

                if cur_preset.name == 'default':
                    # Scalr respond with default preset
                    LOG.debug('%s configuration is default', self._service_name)
                    #self._preset_store.copy(PresetType.DEFAULT, PresetType.LAST_SUCCESSFUL)
                    self._start_service()
                    return

                elif self._cnf_ctl.preset_equals(cur_preset, my_preset):
                    LOG.debug("%s configuration satisfies current preset '%s'", self._service_name, cur_preset.name)
                    self._start_service()
                    return

                else:
                    LOG.info("Applying '%s' preset to %s", cur_preset.name, self._service_name)
                    self._cnf_ctl.apply_preset(cur_preset)

                LOG.info('Start %s with configuration preset', service_name)
                self._start_service_with_preset(cur_preset)
            else:
                LOG.info('Start %s' % service_name)
                self._start_service()

        hub.fire(self._service_name + '_configure', **kwargs)


    def _get_preset(self, preset_data, config_fname):
        p = {}
        for preset in preset_data:
            file = preset['file']
            if 'name' in file and file['name'] == config_fname and 'settings' in file:
                settings = file['settings']
                if settings:
                    for setting in settings:
                        variable = setting['setting']['name']
                        value = setting['setting']['value']
                        p[variable] = value
                break
        return {config_fname : p}


class Messages(object):
    ###
    # Scalarizr events
    ###

    HELLO = "Hello"
    """
    Fires when Scalarizr wants to remind Scalr of himself
    @ivar behaviour
    @ivar local_ip
    @ivar remote_ip
    @ivar role_name
    """

    HOST_INIT = "HostInit"
    """
    @broadcast
    Fires when scalarizr is initialized and ready to be configured
    @ivar behaviour
    @ivar local_ip
    @ivar remote_ip
    @ivar role_name
    """

    BEFORE_HOST_UP = "BeforeHostUp"
    """
    @broadcast
    @blocking
    Fires before HostUp
    @ivar behaviour
    @ivar local_ip
    @ivar remote_ip
    @ivar role_name
    """

    HOST_UP = "HostUp"
    """
    @broadcast
    Fires when server is ready to play it's role
    """

    HOST_DOWN = "HostDown"
    """
    @broadcast
    Fires when server is terminating
    """

    HOST_UPDATE = "HostUpdate"
    """
    Agent to Scalr message to notify about changes in server state.
    """

    REBOOT_START = "RebootStart"
    """
    Fires when scalarizr is going to reboot
    """

    REBOOT_FINISH = "RebootFinish"
    """
    @broadcast
    Fires when scalarizr is resumed after reboot
    """

    RESTART = "Restart"
    """
    @broadcast
    Fires when server is resumed after stop
    """

    BLOCK_DEVICE_ATTACHED = "BlockDeviceAttached"
    """
    Fires when block device was attached
    """

    BLOCK_DEVICE_DETACHED = "BlockDeviceDetached"
    """
    Fires when block device was detached
    """

    BLOCK_DEVICE_MOUNTED = "BlockDeviceMounted"
    """
    Fires when block device was mounted
    """

    EXEC_SCRIPT_RESULT = "ExecScriptResult"
    """
    Fires after script execution
    """

    REBUNDLE_RESULT = "RebundleResult"
    """
    Fires after rebundle task finished
    """

    DEPLOY_RESULT = 'DeployResult'
    '''
    Fires after deployment finished
    '''

    OPERATION_DEFINITION = 'OperationDefinition'
    '''
    Fires before long timed operation
    '''

    OPERATION_PROGRESS = 'OperationProgress'
    '''
    Log message described operation progress
    '''

    OPERATION_RESULT = 'OperationResult'
    '''
    Operation result message
    '''

    UPDATE_CONTROL_PORTS = "UpdateControlPorts"

    REBUNDLE_LOG = "RebundleLog"

    DEPLOY_LOG = "DeployLog"

    LOG = "Log"

    ###
    # Scalr events
    ###

    VHOST_RECONFIGURE = "VhostReconfigure"

    MOUNTPOINTS_RECONFIGURE = "MountPointsReconfigure"

    HOST_INIT_RESPONSE = "HostInitResponse"

    REBUNDLE = "Rebundle"

    DEPLOY = 'Deploy'

    SCALARIZR_UPDATE_AVAILABLE = "ScalarizrUpdateAvailable"

    BEFORE_HOST_TERMINATE = "BeforeHostTerminate"

    BEFORE_INSTANCE_LAUNCH = "BeforeInstanceLaunch"

    DNS_ZONE_UPDATED = "DNSZoneUpdated"

    IP_ADDRESS_CHANGED = "IPAddressChanged"

    SCRIPTS_LIST_UPDATED = "ScriptsListUpdated"

    EXEC_SCRIPT = "ExecScript"

    UPDATE_SERVICE_CONFIGURATION = "UpdateServiceConfiguration"

    UPDATE_SERVICE_CONFIGURATION_RESULT = "UpdateServiceConfigurationResult"

    SSL_CERTIFICATE_UPDATE = "SSLCertificateUpdate"

    ###
    # Internal events
    ###

    INT_BLOCK_DEVICE_UPDATED = "IntBlockDeviceUpdated"
    """
    Fired by scripts/udev.py when block device was added/updated/removed
    """

    UPDATE_SSH_AUTHORIZED_KEYS = "UpdateSshAuthorizedKeys"

    WIN_PREPARE_BUNDLE = "Win_PrepareBundle"

    WIN_PREPARE_BUNDLE_RESULT = "Win_PrepareBundleResult"

    WIN_HOST_DOWN = "Win_HostDown"




class FarmSecurityMixin(object):
    def __init__(self):
        self.__enabled = False

    def init_farm_security(self, ports_ranges):
        self._logger = logging.getLogger(__name__)
        self._ports_ranges = ports_ranges
        if firewall_management_enabled():
            self.__setup_firewall()
            self.__enabled = True


    def on_HostInit(self, message):
        if not self.__enabled:
            return
        # Append new server to allowed list
        for port_range in self._ports_ranges:
            accept_rules = self.__accept_host(message.local_ip, message.remote_ip, port_range)
            for rule in accept_rules:
                firewall.ensure(rule)


    def on_HostDown(self, message):
        if not self.__enabled:
            return
        # Remove terminated server from allowed list
        for port_range in self._ports_ranges:
            accept_rules = self.__accept_host(message.local_ip, message.remote_ip, port_range)
            for rule in accept_rules:
                firewall.remove(rule)


    @staticmethod
    def __create_rule(action, port_range, source=None):
        """
        Creates INPUT rule for tcp protocol.
        """
        from_port, to_port = port_range
        return dict(
            action=action,
            direction='in',
            protocol='tcp',
            from_port=from_port,
            to_port=to_port,
            source=source or '0.0.0.0/0',
            comment="Farm security rule: {}/{}:{}".format(action, from_port, to_port),
        )


    def __create_allow_rule(self, port_range, source):
        return self.__create_rule('allow', port_range, source=source)


    def __create_deny_rule(self, port_range):
        return self.__create_rule('deny', port_range)


    def __accept_host(self, local_ip, public_ip, port_range):
        ret = []
        if local_ip == hub.platform.get_private_ip():
            ret.append(self.__create_allow_rule(port_range, '127.0.0.1'))
        if local_ip:
            ret.append(self.__create_allow_rule(port_range, local_ip))
        ret.append(self.__create_allow_rule(port_range, public_ip))
        return ret


    def __setup_firewall(self, *args, **kwds):
        # Collect farm servers IP-s
        hosts = []
        for role in hub.queryenv.list_roles(with_init=True):
            for host in role.hosts:
                hosts.append((host.internal_ip, host.external_ip))

        allow_rules = []
        for ports_range in self._ports_ranges:
            # TODO: this will be duplicated, because current host is in the
            # hosts list too
            # TODO: this also duplicates the rules, inserted in on_HostInit
            # for the current host
            allow_rules += self.__accept_host(hub.platform.get_private_ip(),
                                              hub.platform.get_public_ip(), ports_range)
            for local_ip, public_ip in hosts:
                allow_rules += self.__accept_host(local_ip, public_ip, ports_range)
        for allow_rule in allow_rules:
            firewall.ensure(allow_rule)

        # Deny from all
        for ports_range in self._ports_ranges:
            deny_rule = self.__create_deny_rule(ports_range)
            firewall.ensure(deny_rule, append=True)


def build_tags(purpose=None, state=None, set_owner=True, **kwargs):
    tags = dict()

    if purpose:
        tags['scalr-purpose'] = purpose

    if state:
        tags['scalr-status'] = state

    if set_owner:
        for tag, opt_name in {
            'scalr-farm-id': 'farm_id',
            'scalr-farm-role-id': 'farm_role_id',
            'scalr-env-id': 'env_id'
        }.items():
            tags[tag] = getattr(hub.general, opt_name, None)

        tags['scalr-owner'] = hub.general.owner_email

    if kwargs:
        # example: tmp = 1
        tags.update(kwargs)

    excludes = []
    for k, v in tags.items():
        if not v:
            excludes.append(k)
            del tags[k]
        else:
            try:
                tags[k] = str(v)
            except:
                excludes.append(k)

    LOG.debug('Prepared tags: %s. Excluded empty tags: %s' % (tags, excludes))
    return tags


def transfer_result_to_backup_result(mnf):
    base = os.path.dirname(mnf.cloudfs_path)
    files_sizes = list((os.path.join(base, chunk[0]), chunk[2])
                                    for file_ in mnf['files']
                                    for chunk in file_['chunks'])
    return list(dict(path=path, size=size) for path, size in files_sizes)


def _choose_host_ip(host, network=None):
    this_server_location = hub.general.region
    if network == None:
        use_internal_ip = this_server_location == host.cloud_location
    else:
        use_internal_ip = network == 'private'
    return host.internal_ip if use_internal_ip else host.external_ip


def get_role_servers(role_id=None, role_name=None, network=None):
    """ Method is used to get role servers from scalr """
    if type(role_id) is int:
        role_id = str(role_id)

    roles = hub.queryenv.list_roles(farm_role_id=role_id, role_name=role_name)
    servers = []
    for role in roles:
        ips = [_choose_host_ip(h, network) for h in role.hosts]
        servers.extend(ips)

    return servers


def check_supported_behaviors(*args):
    # *args to adaptee as a handler function
    if linux.os.windows:
        return
    system_packages = pkgmgr.package_mgr().list()
    for behavior in hub.behaviors:
        if behavior == 'base' or behavior not in api.api_routes.keys():
            continue
        try:
            api_cls = util.import_class(api.api_routes[behavior])
            api_cls.check_software(system_packages)
        except exceptions.NotFound as e:
            LOG.error(e)
        except exceptions.UnsupportedBehavior as e:
            if e.args[0] == 'chef':
                # We pass it, cause a lot of roles has chef behavior without chef installed on them
                continue
            msg.send(msg.create(
                'RuntimeError',
                body={
                    'code': 'UnsupportedBehavior',
                    'message': str(e)}))
            six.reraise(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])


def firewall_management_enabled():
    """
    Checks that:
      1. Firewall management is enabled on Scalr side.
      2. Firewall service is enabled on host.

    :rtype bool
    """
    try:
        if int(hub.base.disable_firewall_management):
            LOG.debug('base.disable_firewall_management: 1, skipping')
            return False
    except (ValueError, TypeError):
        pass
    return firewall.enabled()
