'''
Created on Oct 13, 2011

@author: marat
'''

import logging
import os
import sys

from agent.scalarizr.hub import hub
if sys.platform != 'win32':
    from common.utils import linuxutil
else:
    linuxutil = None
from common.utils.facts import fact

from scalarizr import handlers
from scalarizr import linux
from scalarizr import platform
from scalarizr import storage2
from scalarizr.linux import mount
from scalarizr.handlers import Messages
from scalarizr.util import wait_until

LOG = logging.getLogger(__name__)


class BlockDeviceHandler(handlers.Handler):
    _vol_type = None
    _config = None

    def __init__(self, vol_type):
        super(BlockDeviceHandler, self).__init__()
        self._vol_type = vol_type
        self._volumes = []

        hub.on(init=self.on_init)

    def accept(self, message, queue, **kwds):
        return message.name in (
            Messages.INT_BLOCK_DEVICE_UPDATED,
            Messages.MOUNTPOINTS_RECONFIGURE,
            Messages.BEFORE_HOST_TERMINATE)

    def on_init(self):
        hub.on(
            host_init_response=self.on_host_init_response,
            before_host_up=self.on_before_host_up,
            block_device_mounted=self.on_block_device_mounted
        )

        self.add_udev_rules()
        if hub.state != 'running':
            return

        # remount volumes after restart the Scalarizr service
        volumes = hub.queryenv.list_farm_role_params(
            hub.general.farm_role_id
        ).get('params', {}).get('volumes') or []
        original_volumes = []
        updated_volumes = []
        error_message = "Can't ensure volume %s. Error: %s"

        for original_volume in self._sorted_volumes(volumes):
            volume = storage2.volume(original_volume)
            volume_ensured = False
            try:
                volume.ensure(mount=bool(volume.mpoint))
            except platform.NoCredentialsError:
                pass
            except storage2.StorageError as err:
                LOG.warning(error_message, dict(volume), str(err))
            except Exception as err:
                # we shouldn't stop initialization
                LOG.warning(error_message, dict(volume), str(err))
            else:
                volume_ensured = True

            if volume_ensured \
                    and original_volume.get('id') \
                    and original_volume.get('device') and volume.device \
                    and original_volume['device'] != volume.device:
                LOG.debug(
                    "StorageVolume (id=%s) device changed from %s to %s",
                    original_volume['id'],
                    original_volume['device'],
                    volume.device)
                updated_volumes.append(dict(volume))
            else:
                original_volumes.append(dict(original_volume))

        # optionally send updated configuration to Scalr
        if updated_volumes:
            LOG.debug(
                "The volumes configuration has been modified, "
                "sending updated configuration to Scalr ...")
            try:
                self.send_message(Messages.HOST_UPDATE, {
                    'volumes': original_volumes + updated_volumes
                })
            except Exception as err:
                LOG.warning("Failed to send HostUpdate message: %s", str(err))

    def add_udev_rules(self):
        if linux.os.windows_family:
            return
        LOG.debug("Ensure udev rule for EBS devices")
        try:
            script_path = os.path.join(hub.scripts_dir, 'udev')

            if fact['manage_container_host']:
                ebs_rule_path = '/rootfs/etc/udev/rules.d/84-ebs.rules'
            else:
                ebs_rule_path = '/etc/udev/rules.d/84-ebs.rules'

            with open(ebs_rule_path, "w+") as fp:
                fp.write('KERNEL=="sd*", ACTION=="add|remove", RUN+="' + script_path + '"\n')
                fp.write('KERNEL=="xvd*", ACTION=="add|remove", RUN+="' + script_path + '"\n')
                fp.write('KERNEL=="nvme*n1", ACTION=="add|remove", RUN+="' + script_path + '"')
        except (OSError, IOError) as e:
            LOG.error("Cannot add udev rule into '/etc/udev/rules.d' Error: %s", str(e))
            raise

    def on_host_init_response(self, hir):
        LOG.info('Configuring storage volumes')
        # volumes from QueryEnv.list_ebs_mountpoints()
        wait_until(self._plug_old_style_volumes, sleep=10)
        # volumes assigned to this role on Farm Designer
        volumes = hir.body.get('volumes', []) or []
        self._plug_new_style_volumes(volumes)

    def on_before_host_up(self, hostup):
        if self._volumes:
            LOG.debug('update hostup with volumes. HostUp message body: %s', hostup.body)
            hostup.body['volumes'] = self._volumes

    def on_block_device_mounted(self, volume):
        cnf = volume.config()
        self.send_message(Messages.BLOCK_DEVICE_MOUNTED, {
            'device_name': cnf['device'],
            'volume_id': cnf['id'],
            'mountpoint': cnf['mpoint']
            })

    @staticmethod
    def _sorted_volumes(volumes):
        # windows letter mount points (F) should be processed before directories (F:\LABEL)
        if linux.os.windows_family:
            volumes = sorted(volumes, key=lambda vol: (vol.get('mpoint') or '').count('\\'))
        for volume in volumes:
            yield volume

    def _plug_new_style_volumes(self, volumes):
        for vol in self._sorted_volumes(volumes):
            vol = storage2.volume(**vol)
            #vol.tags.update(build_tags()) # [UI-343]
            self._log_ensure_volume(vol)
            vol.ensure(mount=bool(vol.mpoint), mkfs=True)
            self._volumes.append(dict(vol))

    def _log_ensure_volume(self, vol):
        '''
        Log messages I want to see:

        'Ensure ebs: take vol-12345678, mount to /mnt/volume'
        'Ensure ebs: create volume, create ext3 filesystem, mount to /mnt/volume'
        'Ensure ebs: create volume from snap-12345678'
        '''

        common_actions = {
            'mkfs': 'make {0} filesystem',
            'mount': 'mount to {0}'
        }
        persistent_actions = {
            'take': 'take {0}',
            'new': 'create volume',
            'snap': 'create volume from {0}'
        }

        acts = []
        if vol.id:
            act = persistent_actions['take'].format(vol.id)
            acts.append(act)
        elif vol.snap:
            act = persistent_actions['snap'].format(vol.snap['id'])
            acts.append(act)
        else:
            act = persistent_actions['new']
            acts.append(act)
            if vol.mpoint:
                act = common_actions['mkfs'].format(vol.fstype)
                acts.append(act)
        if vol.mpoint:
            act = common_actions['mount'].format(vol.mpoint)
            acts.append(act)

        msg = 'Ensure {0}: '.format(vol.type)
        msg += ', '.join(acts)
        LOG.info(msg)

    def _plug_old_style_volumes(self):
        unplugged = 0
        plugged_names = []
        for qe_mpoint in hub.queryenv.list_ebs_mountpoints():
            if qe_mpoint.name in plugged_names:
                continue
            if qe_mpoint.name == 'vol-creating':
                unplugged += 1
            else:
                self._plug_volume(qe_mpoint)
                plugged_names.append(qe_mpoint.name)
        return not unplugged

    def _plug_volume(self, qe_mpoint):
        try:
            assert len(qe_mpoint.volumes), 'Invalid mpoint info %s. Volumes list is empty' % qe_mpoint
            qe_volume = qe_mpoint.volumes[0]
            mpoint = qe_mpoint.dir or None
            assert qe_volume.volume_id, 'Invalid volume info %s. volume_id should be non-empty' % qe_volume

            vol = storage2.volume(
                type=self._vol_type,
                id=qe_volume.volume_id,
                name=qe_volume.device,
                mpoint=mpoint
            )
            LOG.info("Plugging volume with tags: %s" % str(vol.tags))
            if mpoint:
                LOG.info('Ensure %s: take %s, mount to %s', self._vol_type, vol.id, vol.mpoint)

                vol.ensure(mount=True, mkfs=True, fstab=True)

            # vol._create_tags_async(qe_volume.volume_id, build_tags())  # [SCALARIZR-1012] upd [UI-343]

        except:
            LOG.exception("Can't attach volume")

    def get_devname(self, devname):
        return devname

    def on_MountPointsReconfigure(self, message):
        LOG.info("Reconfiguring mountpoints")
        for qe_mpoint in hub.queryenv.list_ebs_mountpoints():
            self._plug_volume(qe_mpoint)
        LOG.debug("Mountpoints reconfigured")

    def send_message(self, msg_name, *args, **kwargs):
        if hub.state == 'importing':
            LOG.debug(
                "Silencing the %s message because the server is in an importing state",
                msg_name)
            return
        return super(BlockDeviceHandler, self).send_message(msg_name, *args, **kwargs)

    def on_IntBlockDeviceUpdated(self, message):
        if not message.devname:
            return

        if message.action == "add":
            LOG.debug("udev notified me that block device %s was attached", message.devname)

            self.send_message(
                Messages.BLOCK_DEVICE_ATTACHED,
                {"device_name" : self.get_devname(message.devname)},
                broadcast=True
            )

            hub.fire("block_device_attached", device=message.devname)

        elif message.action == "remove":
            LOG.debug("udev notified me that block device %s was detached", message.devname)
            if hub.volumes and any(v.get('device') == message.devname for v in hub.volumes):
                fstab = linuxutil.fstab()
                if fstab.remove_device(message.devname):
                    LOG.info('Removed %s from the fstab', message.devname)

            self.send_message(
                Messages.BLOCK_DEVICE_DETACHED,
                {"device_name" : self.get_devname(message.devname)},
                broadcast=True
            )

            hub.fire("block_device_detached", device=message.devname)


    def on_BeforeHostTerminate(self, message):
        if message.local_ip != hub.platform.get_private_ip():
            return

        suspend = message.body.get('suspend')
        suspend = suspend and int(suspend) or False

        if suspend:
            return

        volumes = message.body.get('volumes', [])
        volumes = volumes or []

        if volumes:
            LOG.info('Detaching Storage volumes...')
            for volume in volumes:
                try:
                    vol = storage2.volume(volume)
                    vol.umount()
                    vol.detach()
                except:
                    LOG.warn('Failed to detach volume %s: %s',
                             volume['id'], sys.exc_info()[1])

