import os
import time
import logging
import platform
import functools

from common.utils import sysutil
from agent.scalarizr.hub import hub

from scalarizr import util, linux
from scalarizr.util import firstmatched
from scalarizr.linux import coreutils
from scalarizr.storage2.util import gce as gce_util
from scalarizr.handlers import Handler, HandlerError
from scalarizr.handlers import Messages
from scalarizr.handlers.rebundle import RebundleHandler


def get_handlers():
    if linux.os.windows:
        return [GceRebundleWindowsHandler()]
    else:
        return [GceRebundleLinuxHandler()]


LOG = logging.getLogger(__name__)


def uniq_name(name):
    return "{}-{}".format(name, int(time.time())).replace('_', '-')


class GceRebundleLinuxHandler(RebundleHandler):
    """
    GCE rebundle on Linux is performed mainly via cloud API calls, through the following steps:

    1. Finding instance's root disk.
    2. Making a snapshot from it.
    3. Making a temporary persistent disk from the snapshot.
    4. Creating an image from the temporary disk.
    """

    def rebundle(self):
        zone = os.path.basename(hub.platform.get_zone())
        project_id = hub.platform.get_project_id()
        instance_id = hub.platform.get_instance_id()
        connection = hub.platform.get_compute_conn()

        wait_zone_operation = functools.partial(gce_util.wait_for_operation,
                                                connection,
                                                project_id,
                                                zone=zone)
        wait_global_operation = functools.partial(gce_util.wait_for_operation,
                                                  connection,
                                                  project_id)

        temporary_disk_name = None
        snapshot_name = None

        try:
            LOG.info('Taking root disk')
            instance = connection.instances().list(
                project=project_id,
                zone=zone,
                filter='id eq {}'.format(instance_id)
            ).execute()['items'][0]
            root_disk_info = firstmatched(lambda d: d['index'] == 0, instance['disks'])
            root_disk_name = root_disk_info['source'].split('/')[-1]
            root_disk = connection.disks().get(project=project_id, zone=zone, disk=root_disk_name).execute()

            LOG.info('Creating root disk snapshot (disk=%s)', root_disk['name'])
            coreutils.sync()
            snapshot_name = uniq_name(root_disk['name'])
            snapshot_operation = connection.disks().createSnapshot(
                project=project_id, zone=zone, disk=root_disk['name'], body={'name': snapshot_name}).execute()
            wait_zone_operation(snapshot_operation['name'])
            snapshot = connection.snapshots().get(project=project_id, snapshot=snapshot_name).execute()

            LOG.info('Creating root disk clone (snapshot=%s)', snapshot_name)
            temporary_disk_name = uniq_name(root_disk['name'])
            disk_operation = connection.disks().insert(
                project=project_id,
                zone=zone,
                body={'name': temporary_disk_name,
                      'sizeGb': snapshot['diskSizeGb'],
                      'sourceSnapshot': snapshot['selfLink']}).execute()
            wait_zone_operation(disk_operation['name'])
            temporary_disk = connection.disks().get(project=project_id, zone=zone, disk=temporary_disk_name).execute()

            image_name = uniq_name(self._role_name)
            LOG.info('Creating image %s from disk %s', image_name, temporary_disk)
            image_operation = connection.images().insert(
                project=project_id,
                body={'sourceDisk': temporary_disk['selfLink'],
                      'name': image_name}).execute()
            wait_global_operation(image_operation['name'])

            return '{}/images/{}'.format(project_id, image_name)

        finally:
            try:
                if temporary_disk_name:
                    LOG.info('Deleting root disk clone %s', temporary_disk_name)
                    delete_operation = connection.disks().delete(
                        project=project_id,
                        zone=zone,
                        disk=temporary_disk_name).execute()
                    wait_zone_operation(delete_operation['name'])
            except BaseException as e:
                LOG.exception("Could not delete temporary disk")

            try:
                if snapshot_name:
                    LOG.debug('Deleting shapshot %s', snapshot_name)
                    snapshot_delete_operation = connection.snapshots().delete(
                        project=project_id,
                        snapshot=snapshot_name).execute()
                    wait_global_operation(snapshot_delete_operation['name'])
            except BaseException as e:
                LOG.exception("Could not delete temporary snapshot")


class GceRebundleWindowsHandler(Handler):
    logger = None

    def __init__(self):
        self._logger = logging.getLogger(__name__)

    def accept(self, message, queue, behaviour=None, platform=None, os=None, dist=None):
        return message.name == Messages.WIN_PREPARE_BUNDLE

    def on_Win_PrepareBundle(self, message):
        try:
            linux.system('gcesysprep', shell=True)
            os_info = {}
            uname = platform.uname()
            os_info['version'] = uname[2]
            os_info['string_version'] = ' '.join(uname).strip()

            self.send_message(Messages.WIN_PREPARE_BUNDLE_RESULT, dict(
                status="ok",
                bundle_task_id=message.bundle_task_id,
                os=os_info
            ))

        except (Exception, BaseException) as e:
            self._logger.exception(e)
            last_error = hasattr(e, "error_message") and e.error_message or str(e)
            self.send_message(Messages.WIN_PREPARE_BUNDLE_RESULT, dict(
                status="error",
                last_error=last_error,
                bundle_task_id=message.bundle_task_id
            ))

