import logging

from agent.scalarizr.hub import hub

from scalarizr import handlers
from agent.tasks import chef
from agent.tasks import pkgmgr
from agent.tasks import service
from common.utils import subprocess2
from common.utils.facts import fact
from common import types


LOG = logging.getLogger(__name__)


def get_handlers():
    return [RouterHandler()]


class RouterHandler(handlers.Handler):

    def __init__(self):
        self._data = None
        hub.on(init=self.on_init, start=self.on_start)
        super(RouterHandler, self).__init__()

    def on_init(self):
        hub.on(
            host_init_response=self.on_host_init_response,
            before_host_up=self.on_before_host_up)

    def on_host_init_response(self, hir):
        """
        Accept HostInitResponse configuration

        .. code-block:: xml

            <router>
                <cidr>10.0.0.0/16</cidr>
                <whitelist>
                    <item></item>
                </whitelist>
                <scalr_addr>https://my.scalr.net<scalr_addr>
            <router>
        """
        if not hir.body.get('router'):
            msg = "HostInitResponse message for Router behavior " \
                "must have 'router' property"
            raise handlers.HandlerError(msg)

        self._data = hir.router

    def on_start(self):
        if hub.state != 'running':
            return
        self._patch_selinux()
        lfrp = hub.queryenv.list_farm_role_params(
            farm_role_id=hub.general.farm_role_id)['params']
        self._data = lfrp.get('router', {})
        if self._data:
            self._configure()
        else:
            service.start('nginx', already_running_ok=True)

    def on_before_host_up(self, hostup):
        self._configure()
        self._patch_selinux()

    def _configure(self):
        pkgmgr.install(
            'augeas-tools' if fact['os']['family'] == 'debian' else 'augeas',
            updatedb=True)
        augscript = '\n'.join([
            'set /files/etc/sysctl.conf/net.ipv4.ip_forward 1',
            'rm /files/etc/sysctl.conf/net.bridge.bridge-nf-call-ip6tables',
            'rm /files/etc/sysctl.conf/net.bridge.bridge-nf-call-iptables',
            'rm /files/etc/sysctl.conf/net.bridge.bridge-nf-call-arptables',
            'save'
        ])
        subprocess2.check_call(('augtool',), input=augscript)
        subprocess2.check_call(('sysctl', '-p'))

        if self._data.get('cidr'):
            cidr = self._data['cidr']
            if '/' not in cidr:
                cidr += '/32'
            subprocess2.check_call((
                'iptables',
                '-t', 'nat',
                '-A', 'POSTROUTING',
                '-s', cidr,
                '!', '-d', cidr,
                '-j', 'MASQUERADE'))

        LOG.info('Running chef. Runlist: recipe[scalarizr_proxy]')
        chef_action = types.ChefSoloAction({
            'action_type': 'chef.solo',
            'json_attributes': {
                'run_list': ['recipe[scalarizr_proxy]'],
                'scalarizr_proxy': {
                    'scalr_addr': self._data['scalr_addr'],
                    'whitelist': self._data['whitelist']}},
            'cookbooks': {
                'source_type': 'repository',
                'repo': {
                    'url': "https://github.com/Scalr/cookbooks.git"},
                'relative_path': 'cookbooks'}})
        pkgmgr.updatedb()
        chef.solo(action=chef_action)

    def _patch_selinux(self):
        if fact['os']['family'] != 'redhat':
            return
        try:
            pkgmgr.install('policycoreutils-python')
            subprocess2.check_call(('semanage', 'permissive', '-a', 'httpd_t'))
        except subprocess2.CalledProcessError as e:
            LOG.warning('Error setting permissive SELinux policy for httpd_t context: %s', e)
