import binascii
import copy
import functools
import getpass
import io
import json
import logging
import os
import re
import sys
import threading
import time
import typing as tp
import uuid

from agent.scalarizr import msg
from agent.scalarizr.hub import hub
from agent.tasks import ansibletower
from agent.tasks import chef
from agent.tasks import execute
from agent.tasks import pkgmgr
from agent.tasks import service
from common.utils import exc
from common.utils import subprocess2
from common.utils import sysutil
from common.utils.facts import fact
from scalarizr import handlers
from scalarizr.handlers import Messages

LOG = logging.getLogger(__name__)
FOREVER_TIMEOUT = 999999999
PYTHON_ENV_KEYS = [
    'PYTHONPATH',
    'PYTHONHOME',
    'PYTHONSTARTUP',
    'PYTHONOPTIMIZE',
    'PYTHONDEBUG',
    'PYTHONINSPECT',
    'PYTHONUNBUFFERED',
    'PYTHONVERBOSE',
    'PYTHONCASEOK',
    'PYTHONDONTWRITEBYTECODE',
    'PYTHONIOENCODING',
    'PYTHONNOUSERSITE',
    'PYTHONUSERBASE',
    'PYTHONEXECUTABLE',
    'PYTHONWARNINGS',
    'PYTHONFAULTHANDLER',
    'PYTHONTRACEMALLOC',
    'PYTHONASYNCIODEBUG',
    'PYTHONMALLOC',
    'PYTHONMALLOCSTATS',
    'PYTHONLEGACYWINDOWSFSENCODING',
    'PYTHONLEGACYWINDOWSSTDIO',
    'PYTHONTHREADDEBUG',
    'PYTHONDUMPREFS']


def get_handlers():
    return [BaseBehaviorHandler()]


class BaseBehaviorHandler(handlers.Handler):
    def __init__(self):
        hub.on(start=self.on_start)

    def accept(self, message, queue, **kwds):
        return message.name

    def __call__(self, message):
        if message.body.get('scripts'):
            execute_scripts(message)
        if message.name == Messages.HOST_INIT_RESPONSE:
            if message.body.get('chef'):
                # TODO: I now believe it's better to append chef data into HostInitResponse as 'scripts'
                # and let execute_scripts() handle this,
                # rather then virtualizing special 'chef' orchestration message.
                # See setup_chef_automation() for more.
                hub.on(before_host_up=functools.partial(setup_chef_automation, message))
            if 'host_init_response' not in hub._listeners:
                hub.on(host_init_response=install_behaviors)
            else:
                hub._listeners['host_init_response'].insert(0, install_behaviors)
            hub.on(host_init_response=handlers.check_supported_behaviors)

    def on_start(self, *args, **kwds):
        hub.periodic_executor.add_task(
            rotate_scripting_logs, 3600,
            title='Rotate scripting logs')
        if hub.state == 'running':
            farm_role_params = hub.queryenv.list_farm_role_params(hub.general.farm_role_id)
            chef_data = farm_role_params['params'].get('chef')
            if chef_data and int(chef_data.get('daemonize', False)):
                try:
                    service.enable('chef-client')
                    service.start('chef-client')
                except:
                    msg = "Can't daemonize Chef: {}".format(sys.exc_info()[1])
                    LOG.warn(msg, exc_info=sys.exc_info())


def normalize_env_name(name: str) -> str:
    """
    Normalizes an environment variable name.

    Note: On Windows, environment variable names are case-insensitive.
    """
    return name.upper() if fact['os']['family'] == 'windows' else name


def merge_env(base_env: tp.Dict[str, str], override: tp.Dict[str, str]) -> tp.Dict[str, str]:
    """
    Merges environment variables.
    """
    env = copy.deepcopy(base_env)
    for name, value in override.items():
        normalized_name = normalize_env_name(name)
        if normalized_name in env:
            env.pop(normalized_name)
        env[name] = value or ''
    return env


def get_env(message_or_script: tp.Union[msg.Message, dict]) -> tp.Dict[str, str]:
    """
    Returns environment variables for a message (global) or a script (local).

    :param message: The message or the script body.
    :return dict: Environment for the script launch.
    """
    if isinstance(message_or_script, msg.Message):
        variables = message_or_script.body.get('global_variables') or []
        env: tp.Dict[str, str] = {
            name: value for name, value in os.environ.copy().items()
            if normalize_env_name(name) not in PYTHON_ENV_KEYS}
    elif isinstance(message_or_script, dict):
        variables = message_or_script.get('global_variables') or []
        env = {}
    else:
        raise ValueError(
            f"Cannot get environment variables for '{type(message_or_script)}' "
            f"object")

    return merge_env(env, {
        var['name']: var['value'] or ''
        for var in variables})


def _get_user(user):
    """User to run process under.
    Default is Linux - root / Windows - Administrator
    :type: str
    """
    if fact['os']['family'] == 'windows':
        if user is None:
            user = 'Administrator'
        if user != 'Administrator':
            msg = ("Windows can't execute scripts remotely under "
                   "user other than Administrator. Got user: {0}").format(user)
            raise exc.ConstraintError(msg)
        return user
    else:
        return user or getpass.getuser()


def execute_scripts(message, before_call=None, reraise=False):
    scripts = message.body.get('scripts') or []
    if not scripts:
        return

    event_id = message.body.get('event_id')
    event_name = message.event_name \
        if message.name == Messages.EXEC_SCRIPT \
        else message.name
    role_name = message.body.get('role_name', 'unknown_role')
    kwds_defaults = {
        'env': get_env(message),
        'return_stdout': True,
        'return_stderr': True
    }

    countdown = {'counter': len(scripts)}

    def watch_work_done():
        while countdown['counter']:
            time.sleep(0.1)
        return msg.send(msg.create('OperationResult', body={
            'id': event_id,
            'name': 'system.orchestration',
            'status': 'ok'}))
    threading.Thread(target=watch_work_done, daemon=True).start()

    for script in scripts:
        script['event_name'] = event_name
        script['role_name'] = role_name
        script['run_as'] = _get_user(script.get('run_as'))

        kwds = kwds_defaults.copy()
        kwds['timeout'] = int(script['timeout'])
        kwds['run_as'] = script.get('run_as')

        handler = prepare_script_call
        if script.get('chef'):
            handler = prepare_chef_call
        elif script.get('ansibletower'):
            handler = prepare_ansibletower_job_call
        fn, kwds = handler(script, kwds)

        symlink_execution_logs_before(script, fn, kwds)
        if before_call:
            before_call(fn, kwds)
        LOG.info("Executing '{}' in {} mode".format(
            script['name'],
            'non-blocking' if int(script['asynchronous']) else 'blocking'))
        if int(script['asynchronous']):
            t = threading.Thread(
                target=execute_one_script,
                args=(message, script, fn, kwds, reraise, countdown),
                daemon=True)
            t.start()
        else:
            execute_one_script(message, script, fn, kwds, reraise, countdown)


def execute_one_script(message, script, fn, kwds, reraise, countdown):
    try:
        # Use execution_id as celery task_id for persistence purposes
        # https://scalr-labs.atlassian.net/browse/FAM-910
        fn.request.id = script['execution_id'].replace('-', '')
        result = fn(action=kwds)
    except Exception as e:
        if not isinstance(e, execute.ExecuteError):
            LOG.debug('Error occurred in script execution', exc_info=sys.exc_info())
        handle_script_result(message, script, fn, e)
        if reraise or fail_init_maybe.yes:
            fail_init_maybe.yes = False
            raise
    else:
        handle_script_result(message, script, fn, result)
    finally:
        countdown['counter'] -= 1


def handle_script_result(message, script, fn, result):
    try:
        symlink_execution_logs_after(message, script, fn, result)
        send_exec_script_result(message, script, fn, result)
        fail_init_maybe(message, script, fn, result)
    except:
        LOG.debug('Error occurred in handle script result', exc_info=sys.exc_info())


def send_exec_script_result(message, script, fn, result):
    if isinstance(result, Exception):
        if isinstance(result, execute.ExecuteError):
            data = copy.deepcopy(result.__dict__)
        else:
            data = {
                'stdout': '',
                'stderr': u'{}'.format(result),
                'return_code': 1}
    elif fn == chef.bootstrap:
        data = result.executed_script.to_primitive()
    else:
        data = result.to_primitive()
    data.pop('interpreter', None)
    data.pop('path', None)
    data.pop('command', None)
    data.pop('success_codes', None)
    data.pop('stdout_total_bytes', None)
    data.pop('stderr_total_bytes', None)
    data['stdout'] = binascii.b2a_base64(data['stdout'].encode('utf-8'))
    data['stderr'] = binascii.b2a_base64(data['stderr'].encode('utf-8'))
    data['script_id'] = script.get('script_id')
    data['script_version'] = script.get('script_version')
    data['script_name'] = script['name'] if 'job_template_name' not in data \
        else data['job_template_name']
    data['script_path'] = script.get('path')
    data['run_as'] = script.get('run_as')
    data['event_name'] = script['event_name']
    data['event_server_id'] = message.body.get('server_id')
    data['event_id'] = message.body.get('event_id')
    data['execution_id'] = script['execution_id']
    if hub.base.logger.enabled:  # FAM-1110
        try:
            logger_data = dict()
            logger_data['event'] = dict(
                id=data['event_id'],
                triggered_by=data['event_server_id'],
                name=data['event_name'])
            logger_data['script'] = dict(
                name=data['script_name'],
                id=data['script_id'],
                version=data['script_version'])
            logger_data['target'] = dict(
                env_id=hub.general.env_id,
                farm_id=hub.general.farm_id,
                farm_role_id=hub.general.farm_role_id,
                server_id=data['event_server_id'])

            stdout_log_path = data.get('stdout_log_path')
            stderr_log_path = data.get('stderr_log_path')
            for key, fpath in (('stdout', stdout_log_path), ('stderr', stderr_log_path)):
                try:
                    with open(fpath, 'r') as fp:
                        logger_data[key] = fp.read()
                except (OSError, TypeError):
                    logger_data[key] = binascii.a2b_base64(data[key])

            logger_data['return_code'] = data.get('return_code')
            logger_data['execution_id'] = data['execution_id']
            logger_data['time_elapsed'] = round(data['time_elapsed'], 3) if data.get('time_elapsed') else None
            LOG.debug("Sending scripting logs to logstash/fluentd")
            hub.scalr_logger.emit("orchestration.result", **logger_data)
        except Exception as e:
            LOG.error("ScalrLogger failed with error: '%s'.", e)
    msg.send(msg.create('ExecScriptResult', body=data), queue='log')


def get_script_logs_prefix(script):
    if script['execution_id']:
        retval = '{name}.{event_name}.{execution_id}'.format(**script)
    else:
        retval = '{name}.{event_name}.{role_name}.{}'.format(str(uuid.uuid4()), **script)
    return os.path.join(hub.scripting_log_dir, retval)


def symlink_execution_logs_before(script, fn, kwds):
    name_prefix = get_script_logs_prefix(script)
    sysutil.mkdir_p(hub.scripting_log_dir)
    kwds['stdout_symlink_path'] = '{}-out.log'.format(name_prefix)
    kwds['stderr_symlink_path'] = '{}-err.log'.format(name_prefix)


def symlink_execution_logs_after(message, script, fn, result):
    if isinstance(result, Exception) and not isinstance(result, execute.ExecuteError):
        # Scalr will query for logs despeate the fact that script wasn't executed.
        # We should make fake logs with message from exception.
        name_prefix = get_script_logs_prefix(script)
        err_log_file = '{}-err.log'.format(name_prefix)
        with io.open(err_log_file, 'w+', encoding='utf-8') as fp:
            fp.write(u'{}'.format(result))
        out_log_file = '{}-out.log'.format(name_prefix)
        sysutil.touch(out_log_file)


def fail_init_maybe(message, script, fn, result):
    if isinstance(result, Exception) \
            and hub.state == 'initializing' \
            and hub.base.abort_init_on_script_fail \
            and message.body.get('server_id') == hub.general.server_id \
            and script['event_name'] in ('HostInit', 'BeforeHostUp'):
        fail_init_maybe.yes = True
fail_init_maybe.yes = False


def rotate_scripting_logs():
    if not os.path.exists(hub.scripting_log_dir):
        return
    for name in os.listdir(hub.scripting_log_dir):
        path = os.path.join(hub.scripting_log_dir, name)
        if os.path.islink(path) and not os.path.exists(os.path.realpath(path)):
            LOG.debug('Remove {}'.format(path))
            os.unlink(path)


_matches_url = re.compile(r'^http(s)?://').search


def prepare_script_call(script, kwds):
    fn = execute.script
    kwds['action_type'] = 'script'
    if _matches_url(script.get('path', '')):
        kwds['url'] = script['path']
    elif script.get('path'):
        kwds['path'] = script['path']
    elif script.get('repo'):
        kwds['repo'] = {}
        kwds['repo'].update(script['repo'])
        kwds['relative_path'] = kwds['repo'].pop('relative_path')
    else:
        kwds['code'] = script['body']
        kwds['name'] = script['name']
    if script.get('global_variables'):
        kwds['env'] = merge_env(kwds.get('env') or {}, get_env(script))
    return fn, kwds


def prepare_ansibletower_job_call(script, kwds):
    fn = ansibletower.job
    kwds['action_type'] = 'ansibletower.job'
    kwds.update(script['ansibletower'])
    return fn, kwds


def setup_chef_automation(hir, hostup, *args):
    data = hir.body['chef'].to_primitive()
    # execute chef bootstrap script
    hi_chef = msg.Message(
        name=data.get('event_name') or 'HostInit',
        body={
            'role_name': hir.role_name,
            'server_id': hub.general.server_id,
            'global_variables': hir.body['global_variables'],
            'scripts': [{
                'name': data.get('script_name') or '[Scalr built-in] Chef bootstrap',
                'execution_id': data.get('execution_id') or str(uuid.uuid4()),
                'asynchronous': 0,
                # do not limit chef-client run with timeout
                'timeout': FOREVER_TIMEOUT,
                'chef': data
            }]
        }
    )

    def before_call(fn, kwds):
        """This callback sets evaluated chef node_name into HostUp data.

        It also send HostUpdate to let Scalr delete chef node/client
        when init failed and HostUp never comes
        """
        if fn == chef.bootstrap:
            data['node_name'] = kwds['node_name']
            msg.send(msg.create('HostUpdate', body={'chef': data}))

    execute_scripts(hi_chef, before_call, reraise=True)
    hostup.body['chef'] = data


def prepare_chef_call(script, kwds):
    fn = None
    chef_data = script['chef']
    if chef_data.get('ssl_verify_mode', 'chef_auto') != 'chef_auto':
        kwds['ssl_verify_mode'] = chef_data.get('ssl_verify_mode')
    if chef_data.get('log_level'):
        kwds['log_level'] = chef_data['log_level']
    run_list = extract_chef_run_list(chef_data)
    kwds['json_attributes'] = extract_chef_json_attributes(chef_data)
    if chef_data.get('cookbook_url'):
        fn = chef.solo
        kwds['action_type'] = 'chef.solo'
        kwds['solo_rb_template'] = chef_data.get('solo_rb_template')
        if run_list:
            kwds['json_attributes']['run_list'] = run_list
        kwds['cookbooks'] = {}
        if chef_data.get('cookbook_url_type') == 'git':
            kwds['cookbooks']['source_type'] = 'repository'
            kwds['cookbooks']['repo'] = {
                'url': chef_data['cookbook_url'],
                'ssh_key': chef_data.get('ssh_private_key')}
            if chef_data.get('relative_path'):
                kwds['cookbooks']['relative_path'] = chef_data['relative_path']
        elif chef_data.get('cookbook_url_type') == 'http':
            if chef_data['cookbook_url'].startswith('file://'):
                kwds['cookbooks']['source_type'] = 'path'
                kwds['cookbooks']['path'] = chef_data['cookbook_url'][7:]
            else:
                kwds['cookbooks']['source_type'] = 'url'
                kwds['cookbooks']['url'] = chef_data['cookbook_url']
        else:
            raise exc.MalformedError('Chef cookbook_url_type is not valid: {}'.format(
                chef_data['cookbook_url_type']))
    else:
        if chef_data.get('server_url'):
            # Bootstrap
            fn = chef.bootstrap
            kwds['action_type'] = 'chef.bootstrap'
            kwds['client_rb_template'] = chef_data.get('client_rb_template')
            if run_list:
                kwds['json_attributes']['run_list'] = run_list
            kwds['server_url'] = chef_data['server_url']
            kwds['validation_key'] = chef_data['validator_key']
            kwds['validation_client_name'] = chef_data['validator_name']
            if chef_data.get('environment'):
                kwds['environment'] = chef_data['environment']
            kwds['node_name'] = chef_data.get('node_name') or default_chef_node_name()
            kwds['daemonize'] = bool(int(chef_data.get('daemonize', 0)))
        elif run_list:
            # Override run_list
            fn = chef.override
            kwds['action_type'] = 'chef.override'
            kwds['run_list'] = run_list
        else:
            # Re-converge
            fn = chef.reconverge
            kwds['action_type'] = 'chef.reconverge'
    return fn, kwds


def default_chef_node_name():
    hostname = hub.base.hostname or fact['hostname']
    if hostname:
        return re.sub('\s+', '-', hostname)
    else:
        return '{0}-{1}-{2}'.format(
            hub.platform.name,
            hub.platform.get_private_ip(),
            time.time())


def extract_chef_run_list(chef_data):
    if chef_data.get('run_list'):
        try:
            return json.loads(chef_data['run_list'])
        except ValueError as e:
            raise exc.MalformedError("Chef run list is not a valid JSON: {0}".format(e))
    elif chef_data.get('role'):
        return ['role[{}]'.format(chef_data['role'])]
    return None


def extract_chef_json_attributes(chef_data):
    """
    Extract json attributes dictionary from scalr formatted structure
    """
    try:
        return json.loads(chef_data.get('json_attributes') or "{}")
    except ValueError as e:
        raise exc.MalformedError("Chef attributes is not a valid JSON: {0}".format(e))


# Role-builer uses the same mapping
BEHAVIORS_RECIPES_MAP = dict(
    apache="recipe[apache2]",
    haproxy="recipe[haproxy]",
    memcached="recipe[memcached]",
    nginx="recipe[nginx]",
    tomcat="recipe[tomcat]"
)
BEHAVIORS_RECIPES_MAP['www'] = BEHAVIORS_RECIPES_MAP['nginx']
BEHAVIORS_RECIPES_MAP['app'] = BEHAVIORS_RECIPES_MAP['apache']

SCALR_COOKBOOKS_GIT_URL = "https://github.com/Scalr/cookbooks.git"


def install_behaviors(hir_message):
    behaviors = (hir_message.base.install.behaviors
                if hir_message.base.install else [])
    install_chef = behaviors or (hir_message.body.get('chef') and hub.platform.name == 'azure')
    unknown_bhs = set(behaviors).difference(set(BEHAVIORS_RECIPES_MAP.keys()))
    if unknown_bhs:
        raise Exception('Unknown behaviors: {}'.format(list(unknown_bhs)))

    # install chef
    if install_chef and not subprocess2.which('chef-client'):
        chef.install(version='14.13.11')
    # install software
    if not behaviors:
        return
    LOG.info('Installing software for behaviors: {}'.format(behaviors))
    chef_action = {
        'action_type': 'chef.solo',
        'json_attributes': {
            'run_list': [BEHAVIORS_RECIPES_MAP[b] for b in behaviors]
            },
        'cookbooks': {
            'source_type': 'repository',
            'repo': {
                'url': SCALR_COOKBOOKS_GIT_URL
                },
            'relative_path': 'cookbooks'
            }
        }
    pkgmgr.updatedb()
    chef.solo(action=chef_action)
