'''
Created on Aug 27, 2012

@author: marat
'''

import re
import os
import collections
import itertools

from scalarizr import linux
from common.utils import subprocess2
from common.utils.facts import fact


class NoFileSystem(linux.LinuxError):
    pass


_MountEntry = collections.namedtuple('_MountEntry', 'device mpoint fstype options dump fsck_order')
MountCmd = subprocess2.new_command('MountCmd', binary='/bin/mount')
UmountCmd = subprocess2.new_command('UmountCmd', binary='/bin/umount')

if fact['manage_container_host']:
    FSTAB_PATH = '/rootfs/etc/fstab'
    MOUNTS_PATH = '/rootfs/proc/1/mounts'
else:
    FSTAB_PATH = '/etc/fstab'
    MOUNTS_PATH = '/proc/mounts'


class _Mounts(object):

    filename = MOUNTS_PATH

    _space_escape_seq = '\\040'
    _entries = None
    _entry_re = None

    def __init__(self, filename=None):
        if filename:
            self.filename = filename
        self._entry_re = re.compile("\\s+")
        self._reload()

    def _reload(self):
        self._entries = []
        for entry_line in open(self.filename):
            if entry_line[0] != "#":
                values = [x for x in self._entry_re.split(entry_line) if x]
                if values:
                    values.extend(itertools.repeat('', 6 - len(values)))
                    if os.path.islink(values[0]):
                        values[0] = os.path.realpath(values[0])
                    values[1] = self._unescape_mpoint(values[1])
                    self._entries.append(_MountEntry(*values))

    def __getitem__(self, device_or_mpoint):
        matched = [entry for entry in self
                   if self._entry_matches(entry, device_or_mpoint)]
        if matched:
            return matched[0]
        raise KeyError(device_or_mpoint)

    def __contains__(self, device_or_mpoint):
        return any([self._entry_matches(entry, device_or_mpoint)
                    for entry in self])

    def __len__(self):
        self._reload()
        return len(self._entries)

    def __iter__(self):
        self._reload()
        return iter(self._entries)

    list_entries = __iter__

    def __delitem__(self, device_or_mpoint):
        tmpname = self.filename + '.tmp'
        with open(tmpname, 'w+') as fp:
            for entry in self:
                if not self._entry_matches(entry, device_or_mpoint):
                    self._write_entry(entry, fp)
        os.rename(tmpname, self.filename)

    remove = __delitem__

    def __setitem__(self, device, entry):
        pass

    def add(self, device, mpoint, fstype, options='auto', dump=0, fsck_order=0):

        mpoint = self._escape_mpoint(mpoint)

        with open(self.filename, 'a+') as fp:
            fp.seek(0)
            if not fp.read().endswith('\n'):
                fp.write('\n')
            self._write_entry(_MountEntry(device, mpoint, fstype, options, dump, fsck_order), fp)

    def _entry_matches(self, entry, device_or_mpoint):
        return entry.device == device_or_mpoint or \
                entry.mpoint == device_or_mpoint

    def _write_entry(self, entry, fp):
        line = ' '.join(map(str, entry))
        fp.write(line)
        fp.write('\n')

    def _escape_mpoint(self, mpoint):
        if mpoint:
            return mpoint.replace(' ', self._space_escape_seq)

    def _unescape_mpoint(self, mpoint):
        if mpoint:
            return mpoint.replace(self._space_escape_seq, ' ')


class _Fstab(_Mounts):
    filename = FSTAB_PATH


def mounts(filename=MOUNTS_PATH):
    return _Mounts(filename)


def fstab(filename=FSTAB_PATH):
    if not os.path.exists(filename):
        with open(filename, 'w') as fp:
            fp.write('# <file system> <dir> <type> <options> <dump> <pass>')
    return _Fstab(filename)


def mount(device, mpoint, *args, **kwds):

    cmd = MountCmd()
    cmd.args(*args, **kwds)
    cmd.args(device, mpoint)

    subprocess2.check_output_host('/bin/mkdir', '-p', mpoint)
    try:
        subprocess2.check_output_host(*cmd.build())
    except subprocess2.CalledProcessError as e:
        if 'you must specify the filesystem type' in e.stderr:
            raise NoFileSystem(device)
        raise


def umount(device_or_mpoint, *args, **kwds):
    cmd = UmountCmd().args('-f').args(*args, **kwds).args(device_or_mpoint)
    try:
        subprocess2.check_output_host(*cmd.build())
    except subprocess2.CalledProcessError as e:
        if 'not mounted' in e.stderr or 'not found' in e.stderr:
            return
        raise


class MountError(BaseException):
    NO_FS = -100
    CANNOT_MOUNT = -101
    CANNOT_UMOUNT = -102
    CANNOT_CREATE_FS = -103

    message = None
    code = None

    def __init__(self, *args):
        BaseException.__init__(self, *args)
        self.message = args[0]
        if len(args) > 1:
            self.code = args[1]


def mount_ex(device,
             mpoint='/mnt',
             options=None,
             make_fs=False,
             fstype='ext3',
             auto_mount=False):
    if not os.path.exists(mpoint):
        os.makedirs(mpoint)

    options = options or ('-t', 'auto')

    if make_fs:
        from scalarizr import storage2
        storage2.filesystem(fstype).mkfs(device)

    out = mount(device, mpoint, *options)[0]

    if " ".join(options).find("loop") == -1:
        mtab = mounts()
        if not mtab.contains(device):
            raise MountError("Cannot mount device '%s'. %s" % (device, out),
                             MountError.CANNOT_MOUNT)

    if auto_mount:
        _fstab = fstab()
        if not _fstab.contains(device, mpoint=mpoint, reload=True):
            opts = "defaults"
            if linux.os.ubuntu and linux.os['version'] >= (10, 4):
                opts += ',comment=cloudconfig,nobootwait'
            _fstab.append(device, mpoint, options=opts)
