"""
Created on Dec 24, 2009

@author: marat
"""

import os
import re
import socket
import urllib.request
import logging
import platform
import sys
import struct
import array
import threading
import time

from agent.scalarizr.hub import hub

from scalarizr import linux
from agent.scalarizr import metadata
from scalarizr.util import LocalPool, NullPool
if linux.os.windows_family:
    import win32com.client
else:
    import fcntl

from common.utils.facts import fact
from common.utils import six


class PlatformError(BaseException):
    pass


class UserDataOptions:
    FARM_ID = "farmid"
    SERVER_ID = "serverid"
    ROLE_NAME = "realrolename"
    BEHAVIOUR = 'behaviors'
    CRYPTO_KEY = "szr_key"
    QUERYENV_URL = "queryenv_url"
    MESSAGE_SERVER_URL = "p2p_producer_endpoint"
    FARM_HASH = "hash"
    CLOUD_STORAGE_PATH = 'cloud_storage_path'
    ENV_ID = 'env_id'
    FARMROLE_ID = 'farm_roleid'
    ROLE_ID = 'roleid'
    REGION = 'region'
    MESSAGE_FORMAT = 'message_format'
    OWNER_EMAIL = 'owner_email'


class ConnectionError(Exception):
    pass


class NoCredentialsError(ConnectionError):
    pass


class InvalidCredentialsError(ConnectionError):
    pass


class ConnectionProxy(object):

    _logger = logging.getLogger(__name__)
    max_retries = 3

    def __init__(self, conn_pool, obj=None):
        self.obj = obj or conn_pool.get()
        self.conn_pool = conn_pool

    def __getattribute__(self, name):
        if re.search('^__.*__$', name):
            return getattr(object.__getattribute__(self, 'obj'), name)
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            connProxyCls = object.__getattribute__(self, '__class__')
            return connProxyCls(
                    self.conn_pool,
                    obj=getattr(object.__getattribute__(self, 'obj'), name))

    def __call__(self, *args, **kwds):
        for retry in range(self.max_retries):
            try:
                return self.invoke(*args, **kwds)
            except ConnectionError:
                exc_info = sys.exc_info()
                break
            except:
                exc_info = sys.exc_info()
                if retry < self.max_retries - 1:
                    time.sleep(1)
        self.conn_pool.dispose_local()
        six.reraise(exc_info[0], exc_info[1], exc_info[2])

    def invoke(self, *args, **kwds):
        '''Invoke wrapped object (override in subclass)'''
        return self.obj(*args, **kwds)


class PlatformFactory(object):
    _platforms = {}

    def new_platform(self, name):
        if name not in self._platforms:
            pl = __import__("scalarizr.platform." + name,
                globals(),
                locals(),
                fromlist=["get_platform"])
            self._platforms[name] = pl.get_platform()

        return self._platforms[name]


class PlatformFeatures:
    VOLUMES = 'volumes'
    SNAPSHOTS = 'snapshots'


class Platform(object):
    name = None
    _arch = None
    _userdata = None
    _ip_addr = None
    _logger = logging.getLogger(__name__)
    features = []
    scalrfs = None

    def __init__(self):
        self.scalrfs = self._scalrfs(self)
        hub.access_data = {}

    def get_private_ip(self):
        """Returns private ip address, as ipv4 address of default network interface.
        """
        return fact['private_ip']

    get_public_ip = get_private_ip

    def get_user_data(self, key=None):
        if key:
            return metadata.user_data().get(key)
        else:
            return metadata.user_data()

    def set_access_data(self, access_data):
        hub.access_data = access_data

    def get_access_data(self, prop=None):
        if prop:
            try:
                return hub.access_data[prop]
            except (TypeError, KeyError):
                raise PlatformError("Platform access data property '%s' doesn't exists" % (prop,))
        else:
            return hub.access_data

    @staticmethod
    def has_access_data():
        return bool(hub.access_data)

    def clear_access_data(self):
        hub.access_data = {}

    def get_architecture(self):
        """
        @return Architectures
        """
        if self._arch is None:

            if linux.os.windows_family:
                if '32' in platform.architecture()[0]:
                    self._arch = Architectures.I386
                else:
                    self._arch = Architectures.X86_64
            else:
                uname = os.uname()
                if re.search("^i\\d86$", uname[4]):
                    self._arch = Architectures.I386
                elif re.search("^x86_64$", uname[4]):
                    self._arch = Architectures.X86_64
                else:
                    self._arch = Architectures.UNKNOWN
        return self._arch

    @property
    def cloud_storage_path(self):
        return hub.general.cloud_storage_path

    def _parse_user_data(self, raw_userdata):
        userdata = {}
        for k, v in re.findall("([^=]+)=([^;]*);?", raw_userdata):
            userdata[k] = v
        return userdata

    def _raise_no_access_data(self):
        msg = 'There are no credentials from cloud services: %s' % self.name
        raise NoCredentialsError(msg)

    def get_cloud_server(self):
        """
        Returns cloud server.

        :rtype: :class:`common.types.data.CloudServer`
        """
        raise NotImplementedError()

    def get_cloud_credentials(self):
        """
        Returns credentials from cloud provider.

        :rtype: :class:`common.types.data.CloudCredentials`
        """
        raise NotImplementedError()

    class _scalrfs(object):

        def __init__(self, platform):
            self.platform = platform


        def root(self):
            scalr_id = hub.queryenv.get_global_config()['params'].get('scalr.id', '')
            if scalr_id:
                scalr_id = '-' + scalr_id

            if self.platform.cloud_storage_path:
                schema = self.platform.cloud_storage_path.split('://')[0]
            else:
                # Fix for instance-store bundle
                schema = 's3'

            return '%s://scalr%s-%s-%s' % (
                schema, scalr_id,
                hub.general.env_id,
                hub.general.region)

        def images(self):
            if hub.general.scalr_version >= (3, 1, 0):
                return os.path.join(self.root(), 'images/')
            else:
                return '%s://scalr2-images-%s-%s' % (
                        self.platform.cloud_storage_path.split('://')[0],
                        hub.general.region,
                        self.platform.get_account_id()
                )


        def backups(self, service):
            if hub.general.scalr_version >= (3, 1, 0):
                path = 'backups/%s/%s/%s-%s' % (
                    hub.general.farm_id,
                    service,
                    hub.general.farm_role_id,
                    hub.general.role_name)
                return os.path.join(self.root(), path)
            else:
                return os.path.join(self.root(), '%s-backup' % service)


class Ec2LikePlatform(Platform):

    _meta_url = "http://169.254.169.254/"
    _userdata_key = 'latest/user-data'
    _metadata_key = 'latest/meta-data'
    _metadata = {}
    _userdata = None
    _metadata_token = None

    def __init__(self):
        Platform.__init__(self)
        self._logger = logging.getLogger(__name__)

    def _get_property(self, name):
        if name not in self._metadata:
            full_name = self._metadata_key + "/" + name
            self._metadata[name] = self._fetch_metadata(full_name)
        return self._metadata[name]

    def _fetch_metadata(self, key):
        url = self._meta_url + key
        try:
            req = urllib.request.Request(
                url=url,
                method="GET",
                headers=self._get_metadata_token_headers()
            )
            r = urllib.request.urlopen(req)
            return r.read().decode('utf-8').strip()
        except IOError as e:
            if isinstance(e, urllib.request.HTTPError):
                if e.code == 404:
                    return ""
                if e.code == 401:
                    # Needed to handle retry mechanism
                    self._metadata_token = None
                    raise urllib.request.HTTPError(url=e.url, code=403, msg=e.msg, hdrs=e.hdrs, fp=e.fp)
            raise PlatformError("Cannot fetch %s metadata url '%s'. Error: %s" %
                (self.name, url, e))

    def get_private_ip(self):
        return self._get_property("local-ipv4")

    def get_public_ip(self):
        return self._get_property("public-ipv4")

    def get_public_hostname(self):
        return self._get_property("public-hostname")

    def get_instance_id(self):
        return self._get_property("instance-id")

    def get_instance_type(self):
        return self._get_property("instance-type")

    def get_ami_id(self):
        return self._get_property("ami-id")

    def get_ancestor_ami_ids(self):
        return self._get_property("ancestor-ami-ids").split("\n")

    def get_kernel_id(self):
        return self._get_property("kernel-id")

    def get_ramdisk_id(self):
        return self._get_property("ramdisk-id")

    def get_avail_zone(self):
        return self._get_property("placement/availability-zone")

    def get_region(self):
        return self.get_avail_zone()[0:-1]

    def get_block_device_mapping(self):
        keys = self._get_property("block-device-mapping").split("\n")
        ret = {}
        for key in keys:
            ret[key] = self._get_property("block-device-mapping/" + key)
        return ret

    def block_devs_mapping(self):
        keys = self._get_property("block-device-mapping").split("\n")
        ret = list()
        for key in keys:
            ret.append((key, self._get_property("block-device-mapping/" + key)))
        return ret

    def get_ssh_pub_key(self):
        return self._get_property("public-keys/0/openssh-key")

    def _get_metadata_token_headers(self):
        if not self._metadata_token:
            self._logger.debug("Fetching metadata token...")
            req = urllib.request.Request(
                url=f"{self._meta_url}latest/api/token",
                method='PUT',
                headers={"X-aws-ec2-metadata-token-ttl-seconds": "21600"}
            )
            r = urllib.request.urlopen(req)
            self._metadata_token = r.read().decode('utf-8').strip()

        return {"X-aws-ec2-metadata-token": self._metadata_token}

class Architectures:
    I386 = "i386"
    X86_64 = "x86_64"
    UNKNOWN = "unknown"


def is_private_ip(ipaddr):
    return any([ipaddr.startswith(x) for x in ('10.', '172.', '192.168.')])


def is_local_ip(ipaddr):
    return ipaddr.startswith('127.')


def is_public_ip(ipaddr):
    return not (is_private_ip(ipaddr) or is_local_ip(ipaddr))
