import os
import re
import sys
import json
import logging
import urllib.request
import contextlib

import mock
import novaclient
import cinderclient.exceptions
import novaclient.exceptions
import keystoneclient.exceptions
from cinderclient import client as cinder_client
from keystoneclient.auth.identity import v3 as auth_v3
from keystoneclient import session
from novaclient import client as nova_client

from agent.scalarizr.hub import hub

from scalarizr import platform
from scalarizr import linux
from scalarizr.util import LocalPool
from scalarizr.platform import PlatformError
from scalarizr.platform import NoCredentialsError, InvalidCredentialsError
from common.utils import six


LOG = logging.getLogger(__name__)
CINDER_SUPPORTED_APIS = (
    {'version': '3', 'service_type': 'volumev3'},
    {'version': '2', 'service_type': 'volumev2'})


def _create_keystone_v3_session():
    """
    Creates Keystone v3 session using platform acces data.

    :return `session.Session`: Keystone session.
    """
    pl = hub.platform

    try:
        domain_name = pl.get_access_data('domain_name')
    except PlatformError:
        raise InvalidCredentialsError('No domain_name provided for Keystone v3 auth.')

    auth = auth_v3.Password(
        username=pl.get_access_data('username'),
        password=pl.get_access_data('password'),
        project_name=pl.get_access_data('tenant_name'),
        auth_url=pl.get_access_data('keystone_url'),
        user_domain_name=domain_name,
        project_domain_name=domain_name)
    return session.Session(auth=auth, verify=bool(pl.get_access_data('ssl_verify_peer')))


def _get_client_kwds(service_type):
    pl = hub.platform
    kwds = {
        'region_name': pl.get_access_data('cloud_location'),
        'service_type': service_type}

    auth_url = pl.get_access_data('keystone_url')

    # Ending slash is important. See: FAM-1277
    # Volume v3 doesnt work with session object
    if 'v3' in auth_url.rstrip('/').split('/')[-1]:
        LOG.debug('Using Keystone v3.')
        kwds['session'] = _create_keystone_v3_session()
    else:
        kwds['username'] = pl.get_access_data('username')
        kwds['api_key'] = pl.get_access_data('api_key') or pl.get_access_data('password')
        kwds['project_id'] = pl.get_access_data('tenant_name')
        kwds['auth_url'] = auth_url

    if not bool(pl.get_access_data('ssl_verify_peer')):
        kwds['insecure'] = True

    return kwds


def _create_nova_connection():
    try:
        kwds = _get_client_kwds('compute')
        import novaclient  # NameError: name 'novaclient' is not defined
        if hasattr(novaclient, '__version__') and os.environ.get('OS_AUTH_SYSTEM'):
            try:
                import novaclient.auth_plugin
                auth_plugin = novaclient.auth_plugin.load_plugin(os.environ['OS_AUTH_SYSTEM'])
                kwds['auth_plugin'] = auth_plugin
            except ImportError:
                pass
        conn = nova_client.Client('2', **kwds)
    except PlatformError:
        raise NoCredentialsError(sys.exc_info()[1])
    return conn


def _create_cinder_connection():
    try:
        last_exception = None
        for api in CINDER_SUPPORTED_APIS:
            kwds = _get_client_kwds(api['service_type'])
            try:
                if kwds.get('session'):
                    kwds['session'].get_endpoint(service_type=api['service_type'])
                conn = cinder_client.Client(api['version'], **kwds)
                conn.authenticate()
                break
            except (cinderclient.exceptions.AuthorizationFailure,
                    cinderclient.exceptions.Unauthorized,
                    cinderclient.exceptions.Forbidden,
                    keystoneclient.exceptions.EndpointNotFound) as e:
                last_exception = e
        else:
            raise last_exception

    except PlatformError:
        raise NoCredentialsError(sys.exc_info()[1])
    return conn


@contextlib.contextmanager
def use_proxy(proxy_cfg=None):
    """Mocks requests.utils.os.environ dictionary in the module <module_path>,
    so `requests` library in that module uses proxy <proxy_url>
    """
    if proxy_cfg is not None:
        proxy_proto = {0: 'http', 4: 'socks4', 5: 'socks5'}.get(proxy_cfg['type']) or 'http'

        auth = ':'.join(filter(None, (proxy_cfg.get('user'), proxy_cfg.get('pass'))))
        address = ':'.join(filter(None, (proxy_cfg.get('host'),
                                         proxy_cfg.get('port') and str(proxy_cfg['port']))))
        netloc = '@'.join(filter(None, (auth, address)))
        proxy_url = "{}://{}".format(proxy_proto, netloc)

        patch_env = dict(HTTP_PROXY=proxy_url, HTTPS_PROXY=proxy_url)
        proxy_patcher = mock.patch.dict('os.environ', patch_env)

        with proxy_patcher:
            yield
    else:
        yield


class NovaConnectionProxy(platform.ConnectionProxy):

    def invoke(self, *args, **kwds):
        with use_proxy(hub.platform.get_access_data().get('proxy')):
            try:
                return super(NovaConnectionProxy, self).invoke(*args, **kwds)
            except (novaclient.exceptions.Unauthorized, novaclient.exceptions.Forbidden) as e:
                raise InvalidCredentialsError(e)


class CinderConnectionProxy(platform.ConnectionProxy):

    def invoke(self, *args, **kwds):
        with use_proxy(hub.platform.get_access_data().get('proxy')):
            try:
                return super(CinderConnectionProxy, self).invoke(*args, **kwds)
            except (cinderclient.exceptions.Unauthorized, cinderclient.exceptions.Forbidden) as e:
                raise InvalidCredentialsError(e)


class OpenstackPlatform(platform.Platform):

    _meta_url = "http://169.254.169.254/openstack/latest/meta_data.json"
    _metadata = {}
    _userdata = None

    features = ['volumes', 'snapshots']
    name = 'openstack'

    def __init__(self):
        platform.Platform.__init__(self)
        if not linux.os.windows_family:
            # Work over [Errno -3] Temporary failure in name resolution
            # http://bugs.centos.org/view.php?id=4814
            os.chmod('/etc/resolv.conf', 0o755)
        self._nova_conn_pool = LocalPool(_create_nova_connection)
        self._cinder_conn_pool = LocalPool(_create_cinder_connection)

    def _get_property(self, name):
        """
        Get property from user-data (provided by Scalr).

        :raises: KeyError
        """
        if self._userdata is None \
                or name not in self._userdata:
            self._userdata = self.get_user_data()
        if name not in self._userdata:
            raise KeyError("No {} property in userdata".format(name))
        return self._userdata[name]

    def _set_property(self, name, value):
        if not self._userdata:
            self._userdata = self.get_user_data()
        self._userdata[name] = value

    def get_server_id(self):
        if hub.general.farm_role_id:
            global_variables = hub.queryenv.list_global_variables()
            return global_variables['public']['SCALR_CLOUD_SERVER_ID']
        else:
            nova = self.get_nova_conn()
            servers = nova.servers.list()
            my_ip = self.get_private_ip()
            for server in servers:
                ips = []
                ip_addr = 'private' in server.addresses and server.addresses['private'][0]['addr']
                if ip_addr:
                    ips.append(ip_addr)
                else:
                    ips = [address['addr']
                                for network in server.addresses.values()
                                for address in network]
                if my_ip in ips:
                    return server.id
            raise BaseException("Can't get server_id because we can't get "
                                "server private ip")

    def get_avail_zone(self):
        try:
            avail_zone = self._get_property('avail_zone')
        except KeyError:
            avail_zone = None

        if not avail_zone:
            nova_client = self.get_nova_conn()
            server = nova_client.servers.get(self.get_server_id())
            avail_zone = server.to_dict()['OS-EXT-AZ:availability_zone']
            self._set_property('avail_zone', avail_zone)
        return avail_zone

    def get_ssh_pub_key(self):
        return self._get_property('public_keys')  # TODO: take one key

    def _fetch_metadata(self):
        """
        Fetches whole metadata dict. Unlike Ec2LikePlatform,
        which fetches data for concrete key.
        """

        try:
            try:
                self._logger.debug('fetching meta-data from %s', self._meta_url)
                r = urllib.request.urlopen(self._meta_url)
                response = r.read().strip()
                meta = json.loads(response)
            except:
                self._logger.debug('failed to fetch meta-data: %s', sys.exc_info()[1])
            else:
                if meta.get('meta'):
                    return meta
                else:
                    self._logger.debug('meta-data fetched, but has empty user-data (a "meta" key),'
                                       ' try next method')

            return {'meta': self._fetch_metadata_from_file()}
        except:
            six.reraise(platform.PlatformError, 'failed to fetch meta-data', sys.exc_info()[2])

    def _fetch_metadata_from_file(self):
        self._logger.debug('fetching meta-data from files')
        if self._userdata is None:
            private_dir_ud_path = os.path.join(hub.private_dir, '.user-data')
            for path in ('/etc/.scalr-user-data', private_dir_ud_path):
                if os.path.exists(path):
                    self._logger.debug('using file %s', path)
                    rawmeta = None
                    with open(path, 'r') as fp:
                        rawmeta = fp.read()
                    if not rawmeta:
                        raise platform.PlatformError("Empty user-data")
                    self._logger.info('Use user-data from %s', path)
                    return self._parse_user_data(rawmeta)
        return self._userdata

    def get_nova_conn(self):
        return NovaConnectionProxy(self._nova_conn_pool)

    def get_cinder_conn(self):
        return CinderConnectionProxy(self._cinder_conn_pool)


def get_platform():
    # Filter keystoneclient* log messages
    class FalseFilter(object):
        def filter(self, record):
            return False
    log = logging.getLogger('keystoneclient')
    log.addFilter(FalseFilter())

    return OpenstackPlatform()


# Logging

class OpenStackCredentialsLoggerFilter(object):

    request_re = re.compile('(X-Auth[^:]+:)([^\'"])+')
    response_re = re.compile('(.*)({["\']access.+})(.*)')

    def filter(self, record):
        message = record.getMessage()
        record.args = ()

        if "passwordCredentials" in message:
            record.msg = 'Requested authentication, credentials are hidden'
            return True

        search_res = re.search(self.response_re, message)
        if search_res:
            try:
                response_part_str = search_res.group(2)
                response = json.loads(response_part_str)
                response['access']['token'] = '<HIDDEN>'
                response['access']['user'] = '<HIDDEN>'
                altered_resp = json.dumps(response)
                record.msg = search_res.group(1) + altered_resp + search_res.group(3)
                return True
            except:
                return False

        if "X-Auth" in message:
            record.msg = re.sub(self.request_re, r'\1 <HIDDEN>', message)
            return True


openstack_filter = OpenStackCredentialsLoggerFilter()
for logger_name in ('keystoneclient.client', 'novaclient.client', 'cinderclient.client'):
    logger = logging.getLogger(logger_name)
    logger.addFilter(openstack_filter)
