import logging

import celery

from common import types
from common.utils import exc
from agent.scalarizr.hub import hub

from scalarizr import storage2
from scalarizr.storage2.volumes import base
from scalarizr import platform


LOG = logging.getLogger(__name__)


class EFSVolume(base.Volume):
    """
    Amazon Elastic File System volume class.
    """

    def __init__(self, mount_target=None, **kwargs):
        super(EFSVolume, self).__init__(mount_target=mount_target, **kwargs)

        self._volume = types.EFSVolume({
            'id': self.id,
            'scalr_storage_id': self.scalr_storage_id,
            'mount_target': self.mount_target,
            'tags': self.tags,
            'type': self.type,
            'snap': self.snap or None,
            'fstype': self.fstype,
            'mpoint': self.mpoint,
            'mount_options': self.mount_options,
            'recreate_if_missing': self.recreate_if_missing,
            'template': self.template
        })

    def ensure(self, mount=False, mkfs=False, fstab=True, **kwargs):
        cloud_platform = hub.platform

        if not cloud_platform.has_access_data():
            return platform.NoCredentialsError(
                "Unable to connect to EC2 cloud due to missing credentials.")

        try:
            ensured_volume = celery.current_app.send_task(
                'workflows.storage.create',
                kwargs={
                    'server_id': hub.general.server_id,
                    'volume': self._volume,
                    'cloud_server': cloud_platform.get_cloud_server(),
                    'credentials': cloud_platform.get_cloud_credentials(),
                    'mount': mount,
                    'fstab': fstab,
                    'mkfs': False,
                }
            ).get()
            self._config.update(ensured_volume)
        except exc.StorageError as e:
            if hasattr(e, 'volume'):
                self._config.update(e.volume)
            raise

        return self.config()

    def snapshot(self, description=None, tags=None, **kwargs):
        raise exc.NotSupportedError()

    def detach(self, force=False, **kwargs):
        raise NotImplementedError()

    def destroy(self, force=False, **kwargs):
        # pylint: disable=no-self-use
        raise exc.NotSupportedError()

    def apply_tags(self, tags, async=True):
        raise exc.NotSupportedError()


storage2.volume_types['aws_efs'] = EFSVolume
