import os
import logging

import celery
from common import types
from common.utils import exc
from agent.scalarizr.hub import hub

from scalarizr import storage2
from scalarizr.storage2.volumes import base
from scalarizr import platform

LOG = logging.getLogger(__name__)


class BaseAzureVolume(base.Volume):
    """
    Base Azure volume class.
    """

    _volume = None

    def ensure(self, mount=False, mkfs=False, fstab=True, **kwargs):
        cloud_platform = hub.platform
        try:
            if cloud_platform.has_access_data():
                credentials = cloud_platform.get_cloud_credentials()
                cloud_server = cloud_platform.get_cloud_server()
            else:
                credentials = None
                cloud_server = None
            ensured_volume = celery.current_app.send_task(
                'workflows.storage.create',
                kwargs={
                    'server_id': hub.general.server_id,
                    'volume': self._volume,
                    'cloud_server': cloud_server,
                    'credentials': credentials,
                    'mount': mount,
                    'fstab': fstab,
                    'mkfs': mkfs,
                }
            ).get()
            self._config.update(ensured_volume)
        except exc.StorageError as e:
            if hasattr(e, 'volume'):
                self._config.update(e.volume)
            raise

        return self.config()

    def _snapshot(self, description, tags, **kwargs):
        raise NotImplementedError()

    def _detach(self, **kwargs):
        raise NotImplementedError()

    def _destroy(self, force, **kwargs):
        raise NotImplementedError()

    def _clone(self, config):
        raise NotImplementedError()

    def apply_tags(self, tags, async=True):
        raise NotImplementedError()


class AzureVolume(BaseAzureVolume):
    """
    Azure managed volume class.
    """

    def __init__(self, *args, **kwargs):
        super(AzureVolume, self).__init__(*args, **kwargs)

        self._volume = types.AzureVolume({
            'id': self.id,
            'scalr_storage_id': self.scalr_storage_id,
            'resource_group': os.path.basename(self.resource_group),
            'account_type': self.account_type,
            'disk_encryption_set': getattr(self, "disk_encryption_set", None),
            'tags': self.tags,
            'type': self.type,
            'snap': self.snap or None,
            'device': self.device,
            'fstype': self.fstype,
            'mpoint': self.mpoint,
            'mount_options': self.mount_options,
            'recreate_if_missing': self.recreate_if_missing,
            'template': self.template,
            'size': int(self.size),
            'label': getattr(self, 'label', None),  # windows-only
            'zones': getattr(self, 'zones', None),
        })


storage2.volume_types['azure'] = AzureVolume
