import glob
import os
import re
import string  # pylint: disable=deprecated-module
import sys
import uuid

import celery
import six

from agent.scalarizr.hub import hub
from common import types
from common.utils import exc
from common.utils import subprocess2
from common.utils.facts import fact
from scalarizr import linux
from scalarizr import storage2
from scalarizr import util
from scalarizr.libs import bases
from scalarizr.linux import coreutils
from scalarizr.linux import mount as mod_mount

if linux.os.windows:
    import wmi
    from common.utils import winutil
    from agent.tasks.storage.utils import winutil as storage_winutil
else:
    from common.utils import linuxutil


LOG = storage2.LOG


class Base(bases.ConfigDriven):

    def __init__(self, version='2.0', type='base', id=None, tags=None, **kwds):
        super(Base, self).__init__(version=version, type=type, id=id,
                                   tags=tags or {}, **kwds)
        self.error_messages.update({
            'restore_unsupported': 'Restores from snapshot not supported '
                                   'by this volume type: %s',
        })

    def _genid(self, prefix=''):
        return '%s%s-%s' % (prefix, self.type, uuid.uuid4().hex[0:8])


class WinVolumeMixin(Base):
    """
    Mixin class with methods for Windows.
    """

    _avail_timeout = 60
    _polling_interval = 1
    _attach_timeout = 300
    _attach_polling_interval = 10

    def config(self):
        result = super(WinVolumeMixin, self).config()
        if isinstance(self.device, types.SCSIAddress):
            result['device'] = self.device.packed_address
        return result

    def _verify_device_type(self):
        if isinstance(self.device, types.SCSIAddress):
            return

        try:
            self.device = types.SCSIAddress.from_packed_address(self.device)
            return
        except exc.Error:
            pass

        try:
            self.device = storage_winutil.ec2_device_name_to_scsi_address(self.device)
        except exc.StorageError:
            return

    def detect_fstype(self):
        """
        Returns FS type of volume.
        """
        self._verify_device_type()

        LOG.debug('Searching for fs on %s', self.device)
        fs = winutil.get_fstype(self.device)

        if fs is None:
            LOG.debug('FS is not found on %s', self.device)
        else:
            LOG.debug('Found FS %s', fs)
        return fs

    def mounted_to(self):
        """
        Returns volume mount point.
        """
        self._verify_device_type()
        try:
            with winutil.coinitialized_context():
                win32_volume = winutil._scsi_address_to_volume(self.device)
                return winutil._volume_mount_point(win32_volume)
        except exc.StorageError:
            return None

    def mount(self):
        """
        Assigns mount point (drive letter or directory) and label (if presented) to volume.
        """
        self._check(mpoint=True)
        self._verify_device_type()
        if self.mpoint != self.mounted_to():
            storage_winutil.mount_device(
                self.device,
                self.mpoint,
                self.fstype,
                label=getattr(self, 'label', None))
            hub.fire("block_device_mounted", volume=self)

    def umount(self):
        """
        Removes mount point (drive letter or directory).
        """
        self._verify_device_type()
        try:
            self._check(fstype=False, device=True)
        except:
            return
        winutil.umount(self.device)

    def is_fs_created(self):
        """
        Checks if FS is created on volume. Returns True if yes, False otherwise.
        """
        self._check()
        self._verify_device_type()
        fstype = self.detect_fstype()
        if fstype:
            self.fstype = fstype
            return True
        return False

    def attach_win(self, *args, **kwds) -> types.SCSIAddress:
        """
        Performs pre and post attach actions on Windows to get attached disk online and
        available to work through WMI.
        """
        # list Win32_DiskDrive
        LOG.debug('Attaching volume drive on Windows')
        with winutil.coinitialized_context():
            c = wmi.WMI()

            scsi_addrs = set(types.SCSIAddress(disk_drive=d) for d in c.Win32_DiskDrive())
            LOG.debug('Win32_DiskDrive before attach %s', scsi_addrs)

            def check_disks():
                current_scsi_addrs = set(
                    types.SCSIAddress(disk_drive=d)
                    for d in c.Win32_DiskDrive())
                LOG.debug('Polling Win32_DiskDrive, found: %s', current_scsi_addrs)
                return len(scsi_addrs) < len(current_scsi_addrs)

            self._attach_volume(*args, **kwds)
            util.wait_until(
                check_disks,
                start_text='Waiting for disk to be available in system.',
                timeout=self._attach_timeout,
                sleep=self._attach_polling_interval,
                error_text='Disk is not available.')

            # list Win32_DiskDrive again
            # find difference and assert it as a disk attached
            new_scsi_addrs = set(types.SCSIAddress(disk_drive=d) for d in c.Win32_DiskDrive())
            LOG.debug('Win32_DiskDrive after attach %s', new_scsi_addrs)
            if len(new_scsi_addrs) <= len(scsi_addrs):
                raise storage2.StorageError('No new disks appeared after the attach.')

            attached_scsi_addr = (new_scsi_addrs - scsi_addrs).pop()
            LOG.debug(
                'New disk with %s appeared. Assuming it is a newly attached disk.',
                attached_scsi_addr)
            return self._bring_online(attached_scsi_addr)

    def _bring_online(self, scsi_address) -> types.SCSIAddress:
        """
        Performs actions to bring attached disk online.

        Should be ran only in coinitialized_context.
        """
        # perform all the actions that are needed to bring
        # disk online as in EBS disk attach method
        LOG.debug('Bringing disk %s online', scsi_address)

        def _check_available():
            try:
                winutil.bring_disk_online(scsi_address)
                return True
            except storage2.StorageError:
                return False

        util.wait_until(
            _check_available,
            start_text='Checking that disk with SCSIAddress "{}" is available in OS'
                .format(scsi_address),
            timeout=self._avail_timeout,
            sleep=self._polling_interval,
            error_text='No Win32_LogicalDisk available for disk with SCSIAddress "{}", '
                'although it should be online.'.format(scsi_address))

        # onlining can restore unwanted volume letters and mount points
        winutil.umount(scsi_address)

        return scsi_address

    def destroy(self, force=False, **kwds):
        raise NotImplementedError()

    def detach(self, force=False, **kwds):
        raise NotImplementedError()


class LinuxVolumeMixin(object):
    """
    Mixin class with methods for Linux systems.
    """

    def destroy(self, force=False, **kwds):
        LOG.debug('Destroying Volume (id=%s)', self.id)
        if self.device:
            self.detach(force, **kwds)
        self._destroy(force, **kwds)
        LOG.debug('Volume (id=%s) destroyed', self.id)

    def detach(self, force=False, **kwds):
        LOG.debug('Detaching Volume (id=%s)', self.id)
        if not self.device:
            LOG.debug('Volume (id=%s) has no device, nothing to detach', self.id)
            return
        self.umount()

        fstab = linuxutil.fstab()
        fstab.remove_device(self.device)
        self._detach(force, **kwds)
        if self.features['detach']:
            self.device = None
        LOG.debug('Volume (id=%s) detached', self.id)

    def mount(self):
        self._check(mpoint=True)
        mounted_to = self.mounted_to()
        if mounted_to == self.mpoint:
            return
        elif mounted_to:
            LOG.debug('Umounting Volume (id=%s) from %s', self.id, mounted_to)
            self.umount()

        short_args = []
        if self.mount_options:
            short_args += ['-o', ','.join(self.mount_options)]
        LOG.debug('Mounting Volume (id=%s, device=%s) to %s', self.id, self.device, self.mpoint)
        try:
            mod_mount.mount(self.device, self.mpoint, *short_args)
        except (linux.LinuxError, subprocess2.CalledProcessError) as e:
            # XXX: SCALARIZR-1974 is for complete solution
            if linux.os.ubuntu and \
                re.search(r'already mounted or /mnt/(dbstorage|pgstorage|redisstorage) busy', str(e)):
                LOG.debug('upstart already mounted database storage volume {} to {}'.format(
                    self.device, self.mpoint))
            else:
                raise

        hub.fire("block_device_mounted", volume=self)

    def umount(self):
        try:
            self._check(fstype=False, device=True)
        except:
            return

        mod_mount.umount(self.device)

    def mounted_to(self):
        try:
            self._check(fstype=False, device=True)
        except:
            return False

        try:
            # mounts() resolve symlinks in MountEntry (e.g. /dev/group/lvol becomes /dev/md-N)
            # we need to do the same to prevent KeyError for mounted device
            device = os.path.realpath(self.device)
            return mod_mount.mounts()[device].mpoint
        except KeyError:
            return False

    def is_fs_created(self):
        self._check()
        fstype = coreutils.blkid(self.device).get('type')

        if fstype is None:
            return False
        else:
            if fstype != self.fstype:
                LOG.warn(('Device {} already formatted into {!r}. '
                    'Ignoring {!r} fstype').format(self.device, fstype, self.fstype))
            self.fstype = fstype
            return True

    def detect_fstype(self):
        self._check_attr('device')
        blk_info = coreutils.blkid(self.device)
        return blk_info.get('type')


VolumeMixin = WinVolumeMixin if linux.os.windows else LinuxVolumeMixin


class Volume(VolumeMixin, Base):
    """
    Base class for all volume types

    self.device on Windows is SCSITargetId and has int type
    """
    MAX_SIZE = None

    def __init__(self, device=None, fstype='ext3', mpoint=None, throughput=None,
                 mount_options=None, snap=None, recreate_if_missing=False,
                 template=None, **kwds):

        # Get rid of fscreated flag
        kwds.pop('fscreated', None)

        # Backwards compatibility with block_device handler
        from_template_if_missing = kwds.pop('from_template_if_missing', False)

        recreate_if_missing = recreate_if_missing or from_template_if_missing

        kwds['scalr_storage_id'] = kwds.get('scalr_storage_id', None)

        super(Volume, self).__init__(
            device=device,
            fstype=fstype,
            mpoint=mpoint,
            throughput=throughput,
            mount_options=mount_options or [],
            snap=snap,
            recreate_if_missing=recreate_if_missing,
            template=template,
            **kwds)
        self.features.update({'restore': True, 'grow': False, 'detach': True})

    def mkfs(self, force=False):
        self._check()
        if not force and self.is_fs_created():
            raise storage2.OperationError('Filesystem on device %s is already created' % self.device)

        fs = storage2.filesystem(self.fstype)
        LOG.info('Creating filesystem on %s. fstype param is %s', self.device, self.fstype)
        fs.mkfs(self.device)

    def ensure(self, mount=False, mkfs=False, fstab=True, **updates):
        """
        Make sure that volume is attached and ready for use.

        :param mount: if set, volume eventually will be mounted to it's mpoint
        :param mkfs: if set, volume will have corresponding fs eventually
        :return:
        """
        platform = hub.platform

        if self.snap and isinstance(self.snap, Snapshot):  # moved to fix SnapRestore._run()
            self.snap = self.snap.config()

        if self.type == 'ebs':

            cloud_server = types.CloudServer({
                'id': platform.get_instance_id(),
                'location': platform.get_region(),
            })

            if platform.has_access_data():
                access_key, secret_key = platform.get_access_keys()
                account_id = platform.get_account_id()
                proxy = platform.get_access_data().get('proxy')
                credentials = types.AwsCloudCredentials({
                    'account_id': account_id,
                    'access_key': access_key,
                    'secret_key': secret_key,
                    'proxy': proxy
                })
            else:
                credentials = None

            if isinstance(self.snap, dict):  # XXX: corev2: snapshot is always a Snapshot object
                self.snap = types.EbsSnapshot(dict(id=self.snap['id'], type=self.snap['type']))
            else:
                self.snap = None

            ebs_volume = types.EbsVolume({
                'id': self.id,
                'scalr_storage_id': self.scalr_storage_id,
                'tags': self.tags,
                'type': self.type,
                'snap': self.snap,
                'device': self.device,
                'name': self.name,
                'fstype': self.fstype,
                'mpoint': self.mpoint,
                'mount_options': self.mount_options,
                'recreate_if_missing': self.recreate_if_missing,
                'template': self.template,
                'volume_type': self.volume_type,
                'size': int(self.size) if self.size else 0,
                'iops': int(self.iops) if self.iops else None,
                'throughput': int(self.throughput) if self.throughput else None,
                'encrypted': self.encrypted,
                'kms_key_id': self.kms_key_id or None,
                'avail_zone': platform.get_avail_zone(),
                'label': getattr(self, 'label', None),  # windows-only
            })

            try:
                ebs_volume = celery.current_app.send_task(
                    'workflows.storage.create',
                    kwargs={
                        'server_id': hub.general.server_id,
                        'volume': ebs_volume,
                        'cloud_server': cloud_server,
                        'credentials': credentials,
                        'mount': mount,
                        'mkfs': mkfs,
                        'fstab': fstab,
                    }
                ).get()

                self._config.update(ebs_volume)
            except exc.StorageError as e:
                if hasattr(e, 'volume'):
                    self._config.update(e.volume)
                raise

            return self.config()

        if not self.features['restore']:
            self._check_restore_unsupported()

        # volume.snap is always `None` after _ensure()
        # pylint: disable=access-member-before-definition
        snap_id = self.snap.get('id') if self.snap else None
        try:
            self._ensure()
        except (storage2.VolumeNotExistsError,
                exc.NotFoundError) as e:
            if self.recreate_if_missing:
                LOG.warning(e)
                LOG.info('Re-creating %s volume %s from template', self.type, self.id)
                template = dict(self.template or self.clone())
                LOG.debug('Template: %s', template)
                vol = storage2.volume(**template)
                vol.ensure(mount=bool(vol.mpoint), mkfs=True)
                self._config = vol.config()
                return self.config()
            else:
                raise

        if not linux.os.windows and snap_id and coreutils.has_partition_table(self.device):
            msg = 'Snapshot {} was taken from EBS with a partition table. ' \
                  'Partitioned devices are not supported by Scalr ' \
                  'Storage.'.format(snap_id)
            raise storage2.OperationError(msg)

        self._check_attr('device')
        if not self.id:
            self.id = self._genid('vol-')
        if mount:
            if self.type == 'ec2_ephemeral' and not mkfs:
                # After instance stop all data will gone.
                # And if volume was mounted before, we must signal
                # with InstanceStoreDataLost exception to uplevel,
                # so it will resolve the situation adequately.
                #
                # XXX: it's impossible to handle such condition in ec2_ephemeral.py
                # because it's unaware of ``mkfs`` flag. that's why condition sits here.

                info = subprocess2.check_output(f'file -s {self.device}', shell=True)
                if 'filesystem' not in info:
                    LOG.warn(
                        "Instance store data was lost: %s. "
                        "This occurs when the instance stops or the underlying disk drive fails.",
                        self.mpoint)
                    mkfs = True

            if not self.is_fs_created() and mkfs:
                try:
                    self.mkfs()
                except storage2.OperationError as e:
                    if 'already' not in str(e):
                        raise

            try:
                __import__('scalarizr.storage2.filesystems.%s' % self.fstype)
            except ImportError:
                raise exc.StorageError((
                    "StorageVolume (id={id}) has unknown filesystem {fstype} and "
                    "can't be mounted to {mpoint}."
                ).format(id=self.id, fstype=self.fstype, mpoint=self.mpoint))
            self.mount()

            if fstab and not linux.os.windows:
                fstab_mgr = linuxutil.fstab()
                if self.device in fstab_mgr.devices:
                    fstab_entry = fstab_mgr[self.device]
                    if fstab_entry.mpoint == self.mpoint:
                        return self.config()
                    LOG.debug(
                        'According to volume config, device {} should be mounted to {}, '
                        'but fstab mount point is {}. Correcting fstab...'.format(
                            fstab_entry.device, self.mpoint, fstab_entry.mpoint))
                os_fact = fact['container_host.os' if fact['manage_container_host'] else 'os']
                mount_options = list(self.mount_options) if self.mount_options else ['defaults']
                if os_fact['name'] == 'ubuntu' and os_fact['release'] < (16, 4):
                    mount_options.append('nobootwait')
                elif os_fact['family'] != 'coreos':
                    # coreos can't boot with a nofail option
                    mount_options.append('nofail')
                mount_options.append('comment=scalr')
                fstab_mgr.put(self.device, self.mpoint, self.fstype, ','.join(mount_options))
        return self.config()

    def snapshot(self, description=None, tags=None, **kwds):
        return self._snapshot(description, tags, **kwds)

    def clone(self):
        config = self._config.copy()
        config.pop('id', None)
        config.pop('fscreated', None)
        config.pop('device', None)
        self._clone(config)
        return storage2.volume(config)

    def grow(self, **growth):
        """
        Grow (and/or alternate, e.g.: change ebs type to io1) volume and fs.
        Method creates clone of current volume, increases it's size and
        attaches it to the same place. In case of error, old volume attaches back.

        Old volume detached, but not destroyed.

        :param growth: Volume type-dependent rules for volume growth
        :type growth: dict
        :param resize_fs: Resize fs on device after it's growth or not
        :type resize_fs: bool
        :return: New, bigger (or altered) volume instance
        :rtype: Volume
        """

        if not self.features.get('grow'):
            raise storage2.StorageError("%s volume type does not support grow." % self.type)

        # No id, no growth
        if not self.id:
            raise storage2.StorageError('Failed to grow volume: volume has no id.')

        # Resize_fs is true by default
        resize_fs = growth.pop('resize_fs', True)

        self.check_growth(**growth)
        was_mounted = self.mounted_to() if self.device else False

        new_vol = None
        try:
            LOG.info("Marking current volume archived")
            self.apply_tags({"scalr-status": "archived"})
            LOG.info('Detaching volume %s', self.id)
            self.detach()
            new_vol = self.clone()
            self._grow(new_vol, **growth)
            if resize_fs:
                fs_created = new_vol.detect_fstype()

                if self.fstype:
                    LOG.info('Resizing filesystem')
                    fs = storage2.filesystem(fstype=self.fstype)
                    umount_on_resize = fs.features.get('umount_on_resize')

                    if fs_created:
                        if umount_on_resize:
                            if new_vol.mounted_to():
                                new_vol.umount()
                            fs.resize(new_vol.device)
                            if was_mounted:
                                new_vol.mount()
                        else:
                            new_vol.mount()
                            fs.resize(new_vol.device)
                            if not was_mounted:
                                new_vol.umount()

        except:
            err_type, err_val, trace = sys.exc_info()
            LOG.warn('Failed to grow volume: %s. Trying to attach old volume', err_val)
            try:
                if new_vol and getattr(new_vol, 'id', False):
                    try:
                        new_vol.destroy(force=True, remove_disks=True)
                    except:
                        destr_err = sys.exc_info()[1]
                        LOG.error('Enlarged volume destruction failed: %s' % destr_err)

                self.ensure(mount=bool(was_mounted))
                LOG.info("Marking old volume active again.")
                self.apply_tags({"scalr-status": "active"})
            except:
                e = sys.exc_info()[1]
                err_val = str(err_val) + '\nFailed to restore old volume: %s' % e

            err_val = 'Volume growth failed: %s' % err_val
            six.reraise(storage2.StorageError, err_val, trace)
        else:
            LOG.info("The process of growing the storage volume completed.")

        return new_vol

    def _grow(self, bigger_vol, **growth):
        """
        Create, attach and do everything except mount.
        All cleanup procedures and artifact removal should be
        performed in this method

        :param growth: Type-dependant config for volume growth
        :type growth: dict
        :rtype: Volume
        """
        pass

    def extend_fs(self):
        """
        Extends FS on the volume.
        :return Volume: Updated volume config.
        """
        celery.current_app.send_task(
            'filesystem.extend',
            kwargs={
                'fstype': self.fstype,
                'device': self.device}
        ).get()
        return self

    def _check(self, fstype=True, device=True, **kwds):
        if fstype:
            self._check_attr('fstype')
        if device:
            self._check_attr('device')
        for name in kwds:
            self._check_attr(name)

    def _check_attr(self, name):
        assert hasattr(self, name) and getattr(self, name) is not None,  \
                        self.error_messages['empty_attr'] % name

    def _check_restore_unsupported(self):
        if self.snap:
            msg = self.error_messages['restore_unsupported'] % self.type
            #FIXME: eph volume in NewMasterUp raises error here
            LOG.debug(msg)
            LOG.debug('Some details: features=%s, config=%s', self.features, self.config())
            #raise NotImplementedError(msg)

    def check_growth(self, **growth):
        pass

    def _ensure(self):
        # Base volume doesn't guarantee that device 'self.device' exists
        # TODO: Add explanatory comment
        pass

    def _snapshot(self, description, tags, **kwds):
        raise NotImplementedError()

    def _detach(self, force, **kwds):
        pass

    def _destroy(self, force, **kwds):
        pass

    def _clone(self, config):
        pass

    def apply_tags(self, tags, async=True):
        pass


storage2.volume_types['base'] = Volume


class Snapshot(Base):

    QUEUED = 'queued'
    IN_PROGRESS = 'in-progress'
    COMPLETED = 'completed'
    FAILED = 'failed'
    UNKNOWN = 'unknown'

    def __init__(self, **kwds):
        super(Snapshot, self).__init__(**kwds)
        if not self._config.get('id'):
            self._config['id'] = self._genid('snap-')

    def restore(self):
        vol = storage2.volume(type=self.type, snap=self)
        vol.ensure()
        return vol

    def destroy(self):
        return self._destroy()

    def status(self):
        return self._status()

    def _destroy(self):
        pass

    def _status(self):
        pass


storage2.snapshot_types['base'] = Snapshot


def taken_devices():
    devs = glob.glob('/dev/xvd*') + glob.glob('/dev/vd*') + glob.glob('/dev/sd*')
    devs = [os.path.realpath(x) for x in devs if x[-1] in string.ascii_lowercase]
    return set(devs)


def taken_letters():
    lets = [x[-1] for x in taken_devices()]
    return set(lets)


def rescan_scsi_host():
    scsi_host = '/sys/class/scsi_host'
    if os.path.exists(scsi_host):
        for name in os.listdir(scsi_host):
            scsi_scan_file = scsi_host + '/' + name + '/scan'
            LOG.debug('Poking %s', scsi_scan_file)
            with open(scsi_scan_file, 'w') as fp:
                fp.write('- - -')


def wait_device_plugged(volume_id, taken_before):
    '''
    This is called after attaching volume to instance
    '''
    def device_plugged():
        rescan_scsi_host()
        return taken_devices() > taken_before

    util.wait_until(device_plugged,
            start_text='Checking that volume %s is available in OS' % volume_id,
            timeout=30,
            sleep=1,
            error_text='Volume %s attached but not available in OS' % volume_id)

    devices = list(taken_devices() - taken_before)
    if len(devices) > 1:
        msg = ("While polling for attached device, got multiple new devices: {}. "
            "Don't know which one to select").format(devices)
        raise Exception(msg)
    device = devices[0]
    LOG.debug('Volume %s mapped to %s', volume_id, device)
    return device
