import logging
import os
import re
import sys

import celery

from agent.scalarizr.hub import hub
from common import types
from scalarizr import linux
from scalarizr import storage2
from scalarizr.storage2.volumes import base


LOG = logging.getLogger(__name__)


def name2device(name):
    if not name.startswith('/dev'):
        if linux.os.windows:
            return re.sub(r'^sd', 'xvd', name)
        name = os.path.join('/dev', name)
    if name.startswith('/dev/xvd'):
        return name
    name = name.replace('/sd', '/xvd')
    if storage2.RHEL_DEVICE_ORDERING_BUG:
        name = name[0:8] + chr(ord(name[8])+4) + name[9:]
    return name


def device2name(device):
    if device.startswith('/dev/sd') or re.search(r'/dev/xvd[b-c][a-z]', device):
        return device
    elif storage2.RHEL_DEVICE_ORDERING_BUG:
        device = device[0:8] + chr(ord(device[8])-4) + device[9:]
    return device.replace('/xvd', '/sd')


IOPS_SUPPORT_TYPES = ['io1', 'gp3']


def is_iops_supported(volume_type):
    return volume_type in IOPS_SUPPORT_TYPES


THROUGHPUT_SUPPORT_TYPES = ['gp3']


def is_throughput_supported(volume_type):
    return volume_type in THROUGHPUT_SUPPORT_TYPES


class EbsVolume(base.Volume):

    _global_timeout = 3600

    def __init__(
            self,
            name=None,
            avail_zone=None,
            size=None,
            volume_type='standard',
            iops=None,
            throughput=None,
            encrypted=False,
            kms_key_id=None,
            **kwds):

        size = int(size) if size else None
        base.Volume.__init__(
            self,
            name=name,
            avail_zone=avail_zone,
            size=size,
            volume_type=volume_type,
            iops=iops,
            throughput=throughput,
            encrypted=encrypted,
            kms_key_id=kms_key_id,
            **kwds)

        self.error_messages.update({
            'no_id_or_conn': 'Volume has no ID and EC2 connection '
                'required for volume construction is not available'
        })
        self.features.update({'grow': True})

    def _clone(self, config):
        config.pop('device', None)
        config.pop('avail_zone', None)

    def _init_adapters(self):
        platform = hub.platform
        cloud_server = types.CloudServer({
            'id': platform.get_instance_id(),
            'location': platform.get_region()})
        credentials = platform.get_cloud_credentials()
        return credentials, cloud_server, types.EbsVolume({
            'id': self.id,
            'scalr_storage_id': self.scalr_storage_id,
            'tags': self.tags,
            'type': self.type,
            'snap': self.snap or None,
            'device': self.device,
            'fstype': self.fstype,
            'mpoint': self.mpoint,
            'mount_options': self.mount_options,
            'recreate_if_missing': self.recreate_if_missing,
            'template': self.template,
            'volume_type': self.volume_type,
            'size': int(self.size) if self.size else 0,
            'iops': int(self.iops) if self.iops else None,
            'throughput': int(self.throughput) if self.throughput else None,
            'encrypted': self.encrypted,
            'kms_key_id': self.kms_key_id,
            'avail_zone': platform.get_avail_zone()
        })


    def extend_fs(self):
        platform = hub.platform

        if platform.get_access_data():
            credentials, cloud_server, ebs_volume = self._init_adapters()

            celery.current_app.send_task(
                'storage.wait_volume_update',
                kwargs={
                    'volume': ebs_volume,
                    'cloud_server': cloud_server,
                    'credentials': credentials}).get()
        return super().extend_fs()

    def update(self, parameters):
        credentials, cloud_server, ebs_volume = self._init_adapters()
        update_parameters = types.EbsVolumeUpdateParameters(parameters)

        celery.current_app.send_task(
            'workflows.storage.update',
            kwargs={
                'server_id': hub.general.server_id,
                'volume': ebs_volume,
                'update_parameters': update_parameters,
                'cloud_server': cloud_server,
                'credentials': credentials}).get()

        if update_parameters.size:
            self.size = update_parameters.size
        if update_parameters.iops:
            self.iops = update_parameters.iops
        if update_parameters.throughput:
            self.throughput = update_parameters.throughput
        if update_parameters.volume_type:
            self.volume_type = update_parameters.volume_type

        return self

    def grow(self, **growth):
        platform = hub.platform
        if not platform.get_access_data() or linux.os.windows:
            return super(EbsVolume, self).grow(**growth)
        return self.update({k: growth.get(k) for k in ('size', 'volume_type', 'iops', 'throughput')})

    def _grow(self, new_vol, **growth):
        """
        :param new_vol: New volume instance (almost empty)
        :type new_vol: EbsVolume
        :param growth: Growth rules for ebs with size, ebs type, (optionally) iops and (optionally) throughput
        :type growth: dict
        :return: New, bigger, ready to volume instance
        :rtype: EbsVolume
        """
        size = growth.get('size')
        ebs_type = growth.get('volume_type')
        iops = growth.get('iops')
        throughput = growth.get('throughput')

        LOG.info('Creating volume snapshot')
        snap = self.snapshot('Temporary snapshot for volume growth', tags=new_vol.tags, nowait=False)
        try:
            new_vol.snap = snap
            new_vol.size = size if size is not None else self.size
            new_vol.volume_type = ebs_type if ebs_type is not None else self.volume_type
            new_vol.iops = iops if iops is not None else self.iops
            new_vol.throughput = throughput if throughput is not None else self.throughput
            LOG.info('Creating new volume from snapshot')
            new_vol.ensure()
        finally:
            try:
                snap.destroy()
            except:
                e = sys.exc_info()[1]
                LOG.error('Temporary snapshot desctruction failed: %s' % e)

    def check_growth(self, **growth):
        size = growth.get('size')
        target_size = int(size or self.size)

        ebs_type = growth.get('volume_type')
        target_type = ebs_type or self.volume_type

        iops = growth.get('iops')
        target_iops = iops or self.iops

        throughput = growth.get('throughput')
        target_throughput = throughput or self.throughput

        change_type = ebs_type and ebs_type != self.volume_type
        change_size = size and size != self.size
        change_iops = iops and iops != self.iops
        change_throughput = throughput and throughput != self.throughput

        if not (change_size or change_type or change_iops or change_throughput):
            raise storage2.NoOpError('New ebs volume configuration is equal'
                                     ' to present. Nothing to do.')

        if target_iops and not is_iops_supported(target_type):
            raise storage2.StorageError(
                'EBS iops can only be used with %s volume types' % ' or '.join(IOPS_SUPPORT_TYPES)
            )

        if is_iops_supported(target_type) and not target_iops:
            raise storage2.StorageError('Iops parameter must be specified for %s volume' % target_type)

        if target_throughput and not is_throughput_supported(target_type):
            raise storage2.StorageError(
                'EBS throughput can only be used with %s volume types' % ' or '.join(THROUGHPUT_SUPPORT_TYPES)
            )

        if size and int(size) < self.size:
            raise storage2.StorageError('New size is smaller than old.')

    def _ensure(self):
        # XXX moved to corev2
        raise NotImplementedError()

    def _attach_volume(self):
        # XXX moved to corev2
        raise NotImplementedError()

    def _snapshot(self, description, tags, **kwds):
        '''
        @type nowait: bool
        @param nowait: Wait for snapshot completion. Default: True
        '''
        credentials, cloud_server, ebs_volume = self._init_adapters()

        ctx = celery.current_app.send_task(
            'storage.create_snapshot',
            kwargs={
                'volume': ebs_volume,
                'cloud_server': cloud_server,
                'credentials': credentials,
                'wait_available': not kwds.get('nowait', True),
                'description': description,
                'tags': tags}).get()

        return storage2.snapshot(
                        type='ebs',
                        id=ctx.volume.snap.id,
                        description=description,
                        tags=tags)

    def _detach(self, force, **kwds):
        raise NotImplementedError()


    def _destroy(self, force, **kwds):
        raise NotImplementedError()



class EbsSnapshot(base.Snapshot):

    def _status(self):
        return base.Snapshot.COMPLETED

    def _destroy(self):
        pass



storage2.volume_types['ebs'] = EbsVolume
storage2.snapshot_types['ebs'] = EbsSnapshot
