"""
The module to manage the EC2 ephemeral volumes.
"""
import glob
import logging
import os
import sys
import typing as tp

import requests

from agent.scalarizr.hub import hub
from common.utils import six
from common.utils.facts import fact
from scalarizr import storage2
from scalarizr.storage2.volumes import base
from scalarizr.storage2.volumes import ebs
from scalarizr.util import firstmatched

if fact['os']['family'] != 'windows':
    from common.utils import linuxutil
    from agent.tasks.storage.utils import linuxutil as storage_linuxutil
else:
     # pylint: disable=invalid-name
    linuxutil = None
    storage_linuxutil = None

LOG = logging.getLogger(__name__)


def _get_root_devices() -> tp.List[str]:
    """
    Returns the list of devices mounted to the root.

    Note: linux-only
    """
    if linuxutil is None:
        raise NotImplementedError('Not available on Windows')

    entries = linuxutil.mounts().entries
    return list(str(entry.device) for entry in entries if entry.mpoint == '/')


def _is_nitro_based_istance() -> bool:
    """
    Checks that root storage is NVMe SSD disk.

    Note: linux-only
    """
    if fact['os']['name'] == 'windows':
        raise NotImplementedError('Not available on Windows')
    return any(dev.startswith('/dev/nvme') for dev in _get_root_devices())


def _is_instance_storage_nvme(device_path: str) -> bool:
    """
    Checks that given device path is the Instance Storage NVMe disk.

    Note: linux-only
    """
    if fact['os']['name'] == 'windows':
        raise NotImplementedError('Not available on Windows')

    if not (device_path \
            and device_path.startswith('/dev/nvme') \
            and os.path.exists(device_path)):
        return False
    nvme_info: storage_linuxutil.Ec2NvmeInfo = storage_linuxutil.ec2_get_nvme_info(device_path)
    return nvme_info.manufacturer and 'instance storage' in nvme_info.manufacturer.lower()


def _get_nvme_index(path: str) -> tp.Optional[int]:
    """Parse index of the NVME device.
    """
    if not path.startswith("/dev/nvme"):
        return None
    index = path[len("/dev/nvme"):].split("n")[0]
    return int(index)


class Ec2EphemeralVolume(base.Volume):
    """
    The EC2 ephemeral volume.
    """

    def __init__(self, name: tp.Optional[str] = None, **kwargs) -> None:
        """
        :param name: Ephemeral disk name. Valid values: 'ephemeral{0-23}'.
            On EC2 up to 4 ephemeral devices may be available on instance.
            It depends from instance type.
        """
        super(Ec2EphemeralVolume, self).__init__(name=name, **kwargs)
        self.features.update({
            'restore': False,
            'detach': False,
        })

    @property
    def _index(self) -> int:
        """
        Retutn the ephemeral volume index.

        :raises: :class:`storage2.StorageError`
        """
        disk_index = self.name[len('ephemeral'):]
        try:
            return int(disk_index)
        except ValueError:
            raise storage2.StorageError(
                f"Invalid ephemeral device name, must be "
                f"'ephemeral[0-23]', got '{self.name}''"
            ) from None

    def _ensure(self) -> None:
        self._check_attr('name')
        path = f"block-device-mapping/{self.name}"

        try:
            device = fact["ec2"][f"block-device-mapping_{self.name}"]
        except Exception as err:
            msg = f"Failed to fetch device name for instance store '{self.name}'. {err}"
            six.reraise(storage2.StorageError, msg, sys.exc_info()[2])

        device = ebs.name2device(device)
        if fact['os']['name'] != 'windows':

            # lookup for /dev/sd* and /dev/xvd* devices
            devices_to_lookup = [device]
            if not os.path.exists(device) and device.startswith('/dev/xvd'):
                devices_to_lookup.append(device.replace('/xvd', '/sd'))  # SCALARIZR-2302

            # lookup for /dev/nvme* devices
            if _is_nitro_based_istance():
                all_device_indexes = set(map(_get_nvme_index, glob.glob("/dev/nvme*")))
                root_device_indexes = set(map(_get_nvme_index, _get_root_devices()))
                non_root_device_indexes = list(all_device_indexes - root_device_indexes - {None})

                try:
                    nvme_index = sorted(non_root_device_indexes)[self._index]
                except KeyError:
                    nvme_index = None
            else:
                nvme_index = self._index

            if nvme_index is not None and _is_instance_storage_nvme(f'/dev/nvme{nvme_index}n1'):
                devices_to_lookup.append(f'/dev/nvme{nvme_index}n1')

            device = firstmatched(os.path.exists, devices_to_lookup)
            if not device:
                raise Exception(
                    f"Instance store device {self.name} seems doesn't exist (agent has "
                    f"has searched for {', '.join(devices_to_lookup)}). Please check that instance "
                    f"type {hub.platform.get_instance_type()} supports it")

            device = storage_linuxutil.ec2_get_persistent_device_path(device)

        self.device = device

    def _snapshot(self, *args, **kwargs):
        raise NotImplementedError()


storage2.volume_types['ec2_ephemeral'] = Ec2EphemeralVolume
