import os
import logging

from scalarizr import storage2
from scalarizr.storage2.volumes import base
from common import types
from common.utils.facts import fact
from common.utils import six

if fact['os']['family'] == 'windows':
    import wmi
    from common.utils import winutil
else:
    wmi = None
    winutil = None


LOG = logging.getLogger(__name__)


class VmdkVolume(base.Volume):
    """
    Manages storage devices on VMware.
    """

    def __init__(self, pci_slot_number=None, unit_number=None, **kwds):
        super(VmdkVolume, self).__init__(**kwds)

        if pci_slot_number is None:
            raise storage2.StorageError("No pci_slot_number specified for VMware volume.")
        if unit_number is None:
            raise storage2.StorageError("No unit_number specified for VMware volume.")
        self.pci_slot_number = pci_slot_number
        self.unit_number = unit_number

    if fact['os']['family'] == 'windows':
        @winutil.coinitialized
        def _find_device(self):
            """
            Finds disk device based on SCSI controllers' PCI slot number
            and disks' SCSI unit number.

            It uses next WMI classes: Win32_SCSIController, Win32_PnPEntity, Win32_DiskDrive.

            Using Win32_SCSIController.DeviceID and Windows Registry
            we can find SCSI Controller where UINumber equals self.pci_slot_number.
            Next, we list all Win32_PnPEntity on the Controller and find the according
            Win32_DiskDrive where SCSITargetId equals self.unit_number.
            """
            c = wmi.WMI()
            winreg = six.moves.winreg
            # Listing all SCSI controllers
            for scsi_controller in c.Win32_SCSIController():
                registry_path = "SYSTEM\\CurrentControlSet\\Enum\\" + scsi_controller.DeviceID
                # UINumber in this registry path is the controllers pci_slot_number
                try:
                    if winutil.get_hklm_key(registry_path, 'UINumber') != self.pci_slot_number:
                        continue
                except winreg.error:
                    continue

                # Listing all devices that are "plugged" into scsi_controller.
                for entity in scsi_controller.associators(wmi_result_class="Win32_PnPEntity"):
                    # If device|entity is not a disk, continue
                    if entity.Service != 'disk':
                        continue

                    # Gathering associated Win32_DiskDrive and cheching its SCSITargetId
                    disk_drive = entity.associators(wmi_result_class="Win32_DiskDrive")[0]
                    if disk_drive.SCSITargetId == self.unit_number:
                        return types.SCSIAddress(disk_drive=disk_drive)

                raise storage2.StorageError(
                    "Device with SCSI unit number {} "
                    "was not found on controller with PCI slot {}.".format(
                        self.unit_number,
                        self.pci_slot_number))

            raise storage2.StorageError("Controller with PCI slot {} was not found.".format(
                self.pci_slot_number))

    else:
        def _find_device(self):
            """
            Finds disk device based on SCSI controllers' PCI slot number
            and disks' SCSI unit number.
            """
            pci_addr_path = "/sys/bus/pci/slots/{}/address".format(self.pci_slot_number)
            if not os.path.exists(pci_addr_path):
                raise storage2.StorageError("Controller with PCI slot {} was not found.".format(
                    self.pci_slot_number))
            with open(pci_addr_path, 'r') as fp:
                pci_addr = fp.read().strip()
            device = "/dev/disk/by-path/pci-{pci_addr}.0-scsi-0:0:{unit_number}:0".format(
                pci_addr=pci_addr,
                unit_number=self.unit_number)
            if not os.path.exists(device):
                LOG.debug('Existing devices: %s', os.listdir("/dev/disk/by-path/"))
                raise storage2.StorageError(
                    "Device with SCSI unit number {} "
                    "was not found on controller with PCI slot {}.".format(
                        self.unit_number,
                        self.pci_slot_number))
            return device

    def _ensure(self):
        LOG.debug("Ensuring VMDK volume.")
        self.device = self._find_device()
        if fact["os"]["family"] == "windows" and self.mounted_to() is None:
            self._bring_online(self.device)
        LOG.debug("VMware volume %s is mapped to device %s.", self.id, self.device)


storage2.volume_types["vmdk"] = VmdkVolume
