'''
Created on Aug 29, 2010

@author: marat
@author: spike
'''
import socket
import os
import time
import re
import logging
import platform
import time

from threading import local

from scalarizr import linux
from scalarizr import exceptions
from scalarizr.util import system2, PopenError
if platform.system() == 'Windows':
    import win32con
    import win32service


LOG = logging.getLogger(__name__)

_services = dict()
_instances = dict()


# TODO: error codes not used
class InitdError(BaseException):
    GENERIC_ERR = 1
    INVALID_ARG = 2
    UNIMPLEMENTED = 3
    INSUFFICIENT_PRIVILEGE = 4
    NOT_INSTALLED = 5
    NOT_CONFIGURED = 6
    NOT_RUNNING = 7

    @property
    def code(self):
        return len(self.args) > 1 and self.args[1] or None

    @property
    def message(self):
        return self.args[0]


class Status:
    RUNNING = 0
    DEAD_PID_FILE_EXISTS = 1
    DEAD_VAR_LOCK_EXISTS = 2
    NOT_RUNNING = 3
    UNKNOWN = 4


class InitScript(object):
    name = None
    pid_file = None
    lock_file = None

    def start(self):
        '''
        @raise InitdError:
        '''
        pass

    def stop(self):
        '''
        @raise InitdError:
        '''
        pass

    def restart(self):
        '''
        @raise InitdError:
        '''
        pass

    def reload(self):
        '''
        @raise InitdError:
        '''
        pass

    def status(self):
        '''
        @return: Service status
        @rtype: scalarizr.util.initdv2.Status
        '''
        if self.pid_file:
            if not os.path.exists(self.pid_file):
                return Status.NOT_RUNNING
            pid = None
            with open(self.pid_file, 'r') as fp:
                pid = fp.read().strip()
            if os.path.isfile('/proc/%s/status' % pid):
                try:
                    fp = open('/proc/%s/status' % pid)
                    status = fp.read()
                except:
                    return Status.NOT_RUNNING
                finally:
                    fp.close()

                if status:
                    pid_state = re.search('State:\s+(?P<state>\w)', status).group('state')
                    if pid_state in ('T', 'Z'):
                        return Status.NOT_RUNNING
            else:
                return Status.NOT_RUNNING
        if self.socks:
            try:
                for sock in self.socks:
                    timeout = sock.timeout
                    sock.timeout = 1
                    try:
                        wait_sock(sock)
                    finally:
                        sock.timeout = timeout
            except InitdError:
                return Status.NOT_RUNNING

        return Status.RUNNING

    def configtest(self, path=None):
        """
        @raise InitdError:
        """
        pass

    def trans(self, enter=None, exit=None):
        return self

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return None


class SockParam:

    def __init__(self, port=None, family=socket.AF_INET, type=socket.SOCK_STREAM, conn_address=None, timeout=5):

        self.family = family
        self.type = type
        self.conn_address = (conn_address or '127.0.0.1', int(port))
        self.timeout = timeout


class ParametrizedInitScript(InitScript):
    name = None

    def __init__(self, name, initd_script, pid_file=None, lock_file=None, socks=None):
        '''
        if isinstance(initd_script, str):
            if not os.path.exists(initd_script):
                raise InitdError("Can't find %s init script at %s. Make sure that %s is installed" % (
                        name, initd_script, name))
            if not os.access(initd_script, os.X_OK):
                raise InitdError("Permission denied to execute %s" % (initd_script))
        '''

        self.name = name
        self.initd_script = initd_script
        self.pid_file = pid_file
        self.lock_file = lock_file
        self.socks = socks
        self.local = local()

        '''
        @param socks: list(SockParam)
        '''

    def _start_stop_reload(self, action):
        try:

            args = [self.initd_script] \
                if isinstance(self.initd_script, str) \
                else list(self.initd_script)
            args.append(action)
            out, err, returncode = system2(args, close_fds=True, preexec_fn=os.setsid)
        except PopenError as e:
            raise InitdError("Popen failed with error %s" % (e,))

        if returncode:
            raise InitdError("Cannot %s %s. output= %s. %s" % (action, self.name, out, err), returncode)

        if self.socks and (action != "stop" and not (action == 'reload' and not self.running)):
            for sock in self.socks:
                wait_sock(sock)

        return True

    def start(self):
        return self._start_stop_reload('start')

    def stop(self):
        return self._start_stop_reload('stop')

    def restart(self):
        return self._start_stop_reload('restart')

    def reload(self):
        if not self.running:
            raise InitdError('Service "%s" is not running' % self.name, InitdError.NOT_RUNNING)
        return self._start_stop_reload('reload')

    @property
    def running(self):
        return self.status() == Status.RUNNING

    def running_on_exit(self):
        self.local.on_exit = Status.RUNNING
        return self

    def running_on_enter(self):
        self.local.on_enter = Status.RUNNING
        return self

    def __enter__(self):
        self._ctxmgr_ensure_status('on_enter')
        return self

    def __exit__(self, *args):
        self._ctxmgr_ensure_status('on_exit')

    def _ctxmgr_ensure_status(self, status_attr, reason_attr=None):
        if hasattr(self.local, status_attr):
            cur_status = self.status()
            status = getattr(self.local, status_attr)
            if status != cur_status:
                if status == Status.RUNNING:
                    if cur_status == Status.NOT_RUNNING:
                        self.start()
                    else:
                        self.restart()
                else:
                    self.stop(getattr(self.local, reason_attr))

            delattr(self.local, status_attr)
            if reason_attr:
                delattr(self.local, reason_attr)


class Daemon(object):

    '''
    Alternate implementation (from updclient project)
    TODO: we should merge Daemon and ParametrizedInitScript classes and update clients code
    '''

    def __init__(self, name):
        self.name = name

    if linux.os.windows_family:
        def ctl(self, command, raise_exc=True):
            # sending start signal via service control will not fail if
            # a service can accept start command but cannot change its state into 'Running'
            # we should instead poll a service state until we succeed or timeout
            if command.lower() == 'start':
                RUNNING = 4
                TIMEOUT = 3

                GENERIC_READ = win32con.GENERIC_READ
                SC_MANAGER_ALL_ACCESS = win32service.SC_MANAGER_ALL_ACCESS

                __handle = win32service.OpenSCManager(None, None, GENERIC_READ)
                handle = win32service.OpenService(__handle, self.name, SC_MANAGER_ALL_ACCESS)
                try:
                    LOG.debug("Starting service '{}' ...".format(self.name))
                    win32service.StartService(handle, None)
                except Exception as e:
                    raise InitdError('Could not complete StartService command.\n, {0}'.format(e))
                elapsed = 0
                LOG.debug("Waiting until server '{}' becomes running".format(self.name))
                while elapsed < TIMEOUT and win32service.QueryServiceStatusEx(handle)['CurrentState'] != RUNNING:
                    time.sleep(0.1)
                    elapsed += 0.1

                if win32service.QueryServiceStatusEx(handle)['CurrentState'] == RUNNING:
                    LOG.debug("Service '{}' started".format(self.name))
                    return

                raise exceptions.Timeout(
                    "{0} service failed to change its state"
                    " into 'Running' after {1}-second timeout.".format(self.name, TIMEOUT))

            # handles stop on windows
            return linux.system(('sc', command, self.name), raise_exc=raise_exc)
    else:
        def ctl(self, command, raise_exc=True):
            if linux.os['init'] == 'systemd':
                args = ['systemctl', command, "{}.service".format(self.name)]
            elif linux.os['name'] == 'Ubuntu' and linux.os['release'] >= (10, 4):
                args = ['service', self.name, command]
            else:
                args = ['/etc/init.d/' + self.name, command]

            LOG.debug('Daemon.ctl (command: {})'.format(' '.join(args)))
            return linux.system(args, raise_exc=raise_exc, close_fds=True, preexec_fn=os.setsid)

    def restart(self):
        LOG.info('Restarting %s', self.name)
        if linux.os.windows_family:
            self.ctl('stop')
            time.sleep(2)
            self.ctl('start')
        else:
            self.ctl('restart')
        return self._wait_running()

    def forcerestart(self):
        LOG.info('Forcefully restarting %s', self.name)
        if linux.os['init'] == 'systemd':
            self.ctl('kill')
            self.restart()
        else:
            self.ctl('stop')
            try:
                out = linux.system('ps -C %s --noheaders -o pid' % self.name)[0]
                for pid in out.strip().splitlines():
                    LOG.debug('Killing process %s', pid)
                    os.kill(pid, 9)
            finally:
                self.ctl('start')
        return self._wait_running()

    def condrestart(self):
        LOG.info('Conditional restarting %s', self.name)
        self.ctl('condrestart')
        return self._wait_running()

    def start(self):
        LOG.info('Starting %s', self.name)
        self.ctl('start')
        return self._wait_running()

    def stop(self):
        LOG.info('Stopping %s', self.name)
        self.ctl('stop')

    @property
    def running(self):
        if linux.os.windows_family:
            out = self.ctl('query')[0]
            lines = [l.strip() for l in out.splitlines() if l.strip()]
            for line in lines:
                name, value = [x.strip() for x in line.split(':', 1)]
                if name.lower() == 'state':
                    return value.lower().endswith('running')
        else:
            cmd = 'is-active' if linux.os['init'] == 'systemd' else 'status'
            return not self.ctl(cmd, raise_exc=False)[2]

    def _wait_running(self, timeout=10):
        if linux.os.windows_family:
            return self.running
        else:
            cmd = 'is-active' if linux.os['init'] == 'systemd' else 'status'
            for _ in range(timeout):
                status = self.ctl(cmd, raise_exc=False)[0].strip()
                if status == 'activating':
                    time.sleep(1)
                    continue
                elif status == 'active':
                    return True
                else:
                    return False
            return False


def explore(name, init_script_cls):
    _services[name] = init_script_cls


def extend_instances(name, instance):
    _instances[name] = instance


def lookup(name):
    '''
    Lookup init script object by service name
    '''
    if name in _instances:
        return _instances[name]

    if name not in _services:
        raise InitdError('No service has been explored with name %s ' % name)

    _instances[name] = _services[name]()
    return _instances[name]


def wait_sock(sock=None):
    if not isinstance(sock, SockParam):
        raise InitdError('Socks parameter must be instance of SockParam class')

    time_start = time.time()
    while time.time() - time_start < sock.timeout:
        try:
            s = socket.socket(sock.family, sock.type)
            s.connect(sock.conn_address)
            s.shutdown(2)
            del s
            return
        except:
            time.sleep(1)
            pass
    raise InitdError("Service unavailable after %d seconds of waiting" % sock.timeout)
