"""
Created on Sep 10, 2010

@author: marat
"""
import glob
import os
import platform
import re
import zipfile

from common.utils.facts import fact
from common.utils import mixutil
from scalarizr import linux
from scalarizr.linux import pkgmgr, which
from scalarizr.util import system2, PopenError


__all__ = ('all_installed', 'software_info', 'explore', 'which')

def all_installed():
    ret = []
    for getinfo_func in software_list.itervalues():
        try:
            ret.append(getinfo_func())
        except:
            pass
    return ret


def software_info(name):
    if name not in software_list:
        raise Exception("Unknown software: %s" % name)
    return software_list[name]()

def explore(name, lookup_fn):
    if name in software_list.keys():

        raise Exception("'%s' software has been already explored" % name)
    software_list[name] = lookup_fn


def system_info(verbose=False):
    if fact['manage_container_host']:
        os_info = fact['container_host']['os']
    else:
        os_info = fact['os']

    ret = {}
    ret['os'] = {}
    ret['os']['version'] = '{0} {1} {2}'.format(os_info['name'], os_info['release'], os_info['codename']).strip()
    ret['os']['string_version'] = ' '.join(platform.uname()).strip()

    ret['dist'] = {
        'distributor': os_info['name'].lower(),
        'release': str(os_info['release']),
        'codename': os_info['codename']}

    ret['os']['arch'] = linux.os['arch']

    return ret


class SoftwareError(BaseException):
    pass

class SoftwareInfo(object):
    name = None
    version = None
    '''
    @param version: tuple(major, minor, bugfix)
    '''
    string_version = None

    def __init__(self, name, version, string_version):
        self.name = name
        self.string_version = string_version
        self.version = mixutil.Version(version)

software_list = dict()


def nginx_software_info():
    binary = which('nginx', path_append='/usr/local/nginx/sbin')
    if not binary:
        raise SoftwareError("Can't find executable for Nginx server")

    out = system2((binary, '-V'))[1]
    if not out:
        raise SoftwareError

    version_string = out.splitlines()[0]
    res = re.search('[\d\.]+', version_string)
    if res:
        version = res.group(0)
        return SoftwareInfo('nginx', version, out)
    raise SoftwareError


explore('nginx', nginx_software_info)

def memcached_software_info():
    binary = which('memcached')
    if not binary:
        raise SoftwareError("Can't find executable for Memcached")

    out = system2((binary, '-h'))[0]
    if not out:
        raise SoftwareError

    version_string = out.splitlines()[0]

    res = re.search('memcached\s+([\d\.]+)', version_string)
    if res:
        version = res.group(1)
        return SoftwareInfo('memcached', version, version_string)
    raise SoftwareError

explore('memcached', memcached_software_info)

def php_software_info():
    binary = which('php')
    if not binary:
        raise SoftwareError("Can't find executable for php interpreter")

    out = system2((binary, '-v'))[0]
    if not out:
        raise SoftwareError

    version_string = out.splitlines()[0]

    res = re.search('PHP\s+([\d\.]+)', version_string)

    if res:
        version = res.group(1)
        return SoftwareInfo('php', version, out)
    raise SoftwareError

explore('php', php_software_info)

def python_software_info():
    binary = which('python')
    if not binary:
        raise SoftwareError("Can't find executable for python interpreter")

    version_string = system2((binary, '-V'))[1].strip()
    if not version_string:
        raise SoftwareError

    version_string = version_string.splitlines()[0]

    res = re.search('Python\s+([\d\.]+)', version_string)

    if res:
        version = res.group(1)
        return SoftwareInfo('python', version, version_string)

    raise SoftwareError

explore('python', python_software_info)


def apache_software_info():

    binary_name = "httpd" if linux.os.redhat_family else "apache2ctl"
    binary = which(binary_name)
    if not binary:
        raise SoftwareError("Can't find executable for apache http server")

    try:
        out = system2((binary, '-V'))[0]
    except PopenError:
        pkg_mgr = pkgmgr.package_mgr()
        version_string = pkg_mgr.info('apache2')['installed']
    else:
        if not out:
            raise SoftwareError
        version_string = out.splitlines()[0]

    res = re.search('[\d\.]+', version_string)
    if res:
        version = res.group(0)
        return SoftwareInfo('apache', version, version_string)

    raise SoftwareError


explore('apache', apache_software_info)


def tomcat_software_info():
    catalina_home = linux.system('echo $CATALINA_HOME', shell=True)[0].strip()
    if not catalina_home:
        try:
            catalina_home = glob.glob('/opt/apache-tomcat*')[0]
        except IndexError:
            pass
    if not catalina_home:
        try:
            catalina_home = glob.glob('/usr/share/*tomcat*')[0]
        except IndexError:
            msg = (
                "Can't find Tomcat installation\n"
                " - CATALINA_HOME env variable is unset\n"
                " - /opt/apache-tomcat*\n"
                " - /usr/share/*tomcat* search is empty\n"
            )
            raise SoftwareError(msg)

    catalina_jar = os.path.join(catalina_home, 'lib/catalina.jar')
    if not os.path.exists(catalina_jar):
        msg = "Can't get Tomcat version: file {0} not exists".format(catalina_jar)
        raise SoftwareError(msg)

    catalina = zipfile.ZipFile(catalina_jar, 'r')
    try:
        properties_file = 'org/apache/catalina/util/ServerInfo.properties'
        if not properties_file in catalina.namelist():
            raise SoftwareError("ServerInfo.properties file isn't in catalina.jar")

        properties = catalina.read(properties_file)
        properties = re.sub(re.compile('^#.*$', re.M), '', properties).strip()

        res = re.search('^server.info=Apache\s+Tomcat/([\d\.]+)', properties, re.M)
        if res:
            version = res.group(1)
            return SoftwareInfo('tomcat', version, properties)
        raise SoftwareError
    finally:
        catalina.close()

explore('tomcat', tomcat_software_info)

def varnish_software_info():
    binary = which('varnishd')
    if not binary:
        raise SoftwareError("Can't find executable for varnish HTTP accelerator")

    out = system2((binary, '-V'))[1].strip()
    if not out:
        raise SoftwareError

    version_string = out.splitlines()[0]

    res = re.search('varnish-([\d\.]+)', version_string)

    if res:
        version = res.group(1)
        return SoftwareInfo('varnish', version, out)

    raise SoftwareError

explore('varnish', varnish_software_info)

def rails_software_info():
    binary = which('gem')

    if not binary:
        raise SoftwareError("Can't find executable for ruby gem packet manager")

    out = system2((binary, 'list', 'rails'))[0].strip()

    if not out:
        raise SoftwareError

    res = re.search('\(([\d\.]+)\)', out)

    if res:
        version = res.group(1)
        return SoftwareInfo('rails', version, '')

    raise SoftwareError

explore('rails', rails_software_info)


def haproxy_software_info():

    binary_name = "haproxy"
    binary = which(binary_name)
    if not binary:
        raise SoftwareError("Can't find executable for HAProxy")

    out = system2((binary, '-v'))[0]
    if not out:
        raise SoftwareError()

    version_string = out.splitlines()[0]
    res = re.search('[\d\.]+', version_string)
    if res:
        version = res.group(0)

        return SoftwareInfo('haproxy', version, out)
    raise SoftwareError
explore('haproxy', haproxy_software_info)


def chef_software_info():
    binary = which('chef-client')
    if not binary:
        raise SoftwareError("Can't find executable for chef client")

    try:
        version_string = linux.system((binary, '-v'), shell=bool(linux.os.windows))[0].strip()
    except OSError as e:
        if e.errno == 2:
            # This happen when chef-client has a non existed path in shebang.
            version_string = None
        else:
            raise
    if not version_string:
        raise SoftwareError

    res = re.search('Chef:\s+([\d\.]+)', version_string)

    if res:
        version = res.group(1)
        return SoftwareInfo('chef', version, version_string)

    raise SoftwareError
explore('chef', chef_software_info)


